var require = meteorInstall({"imports":{"server":{"accounts":{"ldap.js":["ldapjs",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/accounts/ldap.js                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  LDAP: function () {                                                                                                  // 1
    return LDAP;                                                                                                       // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var ldapjs = void 0;                                                                                                   // 1
module.import("ldapjs", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    ldapjs = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 1
                                                                                                                       //
// At a minimum, set up LDAP_DEFAULTS.url and .dn according to                                                         // 4
// your needs. url should appear as 'ldap://your.url.here'                                                             // 5
// dn should appear in normal ldap format of comma separated attribute=value                                           // 6
// e.g. 'uid=someuser,cn=users,dc=somevalue'                                                                           // 7
LDAP_DEFAULTS = {                                                                                                      // 8
  url: false,                                                                                                          // 9
  port: "389",                                                                                                         // 10
  dn: false,                                                                                                           // 11
  searchDN: false,                                                                                                     // 12
  searchCredentials: false,                                                                                            // 13
  createNewUser: true,                                                                                                 // 14
  defaultDomain: false,                                                                                                // 15
  searchResultsProfileMap: false,                                                                                      // 16
  base: null,                                                                                                          // 17
  search: "(objectclass=*)",                                                                                           // 18
  ldapsCertificate: false                                                                                              // 19
}; /**                                                                                                                 // 8
    @class LDAP                                                                                                        //
    @constructor                                                                                                       //
    */                                                                                                                 //
                                                                                                                       //
function LDAP() {                                                                                                      // 26
  // Set options                                                                                                       // 27
  this.options = _.clone(LDAP_DEFAULTS);                                                                               // 28
}                                                                                                                      // 29
                                                                                                                       //
; /**                                                                                                                  // 29
   * Attempt to bind (authenticate) ldap                                                                               //
   * and perform a dn search if specified                                                                              //
   *                                                                                                                   //
   * @method ldapCheck                                                                                                 //
   *                                                                                                                   //
   * @param {Object} options  Object with username, ldapPass and overrides for LDAP_DEFAULTS object.                   //
   * Additionally the searchBeforeBind parameter can be specified, which is used to search for the DN                  //
   * if not provided.                                                                                                  //
   */                                                                                                                  //
                                                                                                                       //
LDAP.prototype.ldapCheck = function (db, options) {                                                                    // 41
  var _this = this;                                                                                                    // 43
                                                                                                                       //
  options = options || {};                                                                                             // 45
                                                                                                                       //
  if (options.hasOwnProperty("username") && options.hasOwnProperty("ldapPass") || options.hasOwnProperty("searchUsername")) {
    _this.options.base = db.getLdapBase();                                                                             // 49
    _this.options.url = db.getLdapUrl();                                                                               // 50
    _this.options.searchBeforeBind = {};                                                                               // 51
    _this.options.searchBeforeBind[options.searchUsernameField || db.getLdapSearchUsername()] = options.searchUsername || options.username;
    _this.options.filter = db.getLdapFilter() || "(objectclass=*)";                                                    // 54
    _this.options.searchBindDn = db.getLdapSearchBindDn();                                                             // 55
    _this.options.searchBindPassword = db.getLdapSearchBindPassword();                                                 // 56
    var resolved = false;                                                                                              // 58
    var ldapAsyncFut = new Future(); // Create ldap client                                                             // 59
                                                                                                                       //
    var fullUrl = db.getLdapUrl();                                                                                     // 62
    var client = null;                                                                                                 // 63
                                                                                                                       //
    var errorFunc = function (err) {                                                                                   // 65
      if (err) {                                                                                                       // 66
        if (resolved) return;                                                                                          // 67
        resolved = true;                                                                                               // 68
        ldapAsyncFut.return({                                                                                          // 69
          error: err                                                                                                   // 70
        });                                                                                                            // 69
      }                                                                                                                // 72
    };                                                                                                                 // 73
                                                                                                                       //
    if (fullUrl.indexOf("ldaps://") === 0) {                                                                           // 75
      var tlsOptions = {};                                                                                             // 76
      var cert = db.getLdapCaCert();                                                                                   // 77
                                                                                                                       //
      if (cert) {                                                                                                      // 78
        tlsOptions.ca = cert;                                                                                          // 79
      }                                                                                                                // 80
                                                                                                                       //
      client = ldapjs.createClient({                                                                                   // 82
        url: fullUrl,                                                                                                  // 83
        tlsOptions: tlsOptions                                                                                         // 84
      }, errorFunc);                                                                                                   // 82
    } else {                                                                                                           // 86
      client = ldapjs.createClient({                                                                                   // 87
        url: fullUrl                                                                                                   // 88
      }, errorFunc);                                                                                                   // 87
    }                                                                                                                  // 90
                                                                                                                       //
    client.on("error", errorFunc);                                                                                     // 92
    var username = options.username;                                                                                   // 94
    var domain = _this.options.defaultDomain;                                                                          // 95
                                                                                                                       //
    if (!options.hasOwnProperty("searchUsername")) {                                                                   // 97
      // Slide @xyz.whatever from username if it was passed in                                                         // 98
      // and replace it with the domain specified in defaults                                                          // 99
      var emailSliceIndex = options.username.indexOf("@"); // If user appended email domain, strip it out              // 100
      // And use the defaults.defaultDomain if set                                                                     // 103
                                                                                                                       //
      if (emailSliceIndex !== -1) {                                                                                    // 104
        username = options.username.substring(0, emailSliceIndex);                                                     // 105
        domain = domain || options.username.substring(emailSliceIndex + 1, options.username.length);                   // 106
      } else {                                                                                                         // 107
        username = options.username;                                                                                   // 108
      }                                                                                                                // 109
    }                                                                                                                  // 110
                                                                                                                       //
    if (_this.options.searchBindDn) {                                                                                  // 112
      var ldapBindFut = new Future();                                                                                  // 113
      client.bind(_this.options.searchBindDn, _this.options.searchBindPassword, function (err) {                       // 114
        if (err) {                                                                                                     // 116
          ldapBindFut.throw(err);                                                                                      // 117
        } else {                                                                                                       // 118
          ldapBindFut.return();                                                                                        // 119
        }                                                                                                              // 120
      });                                                                                                              // 121
                                                                                                                       //
      try {                                                                                                            // 124
        ldapBindFut.wait();                                                                                            // 125
      } catch (err) {                                                                                                  // 126
        return {                                                                                                       // 127
          error: err                                                                                                   // 127
        };                                                                                                             // 127
      }                                                                                                                // 128
    } // initialize result                                                                                             // 129
                                                                                                                       //
                                                                                                                       //
    var retObject = {                                                                                                  // 131
      username: username,                                                                                              // 132
      email: domain ? username + "@" + domain : false,                                                                 // 133
      emptySearch: true,                                                                                               // 134
      searchResults: {}                                                                                                // 135
    };                                                                                                                 // 131
    var filter = _this.options.filter;                                                                                 // 138
    Object.keys(_this.options.searchBeforeBind).forEach(function (searchKey) {                                         // 139
      filter = "&" + filter + "(" + searchKey + "=" + _this.options.searchBeforeBind[searchKey] + ")";                 // 140
    });                                                                                                                // 141
    var searchOptions = {                                                                                              // 143
      scope: "sub",                                                                                                    // 144
      sizeLimit: 1,                                                                                                    // 145
      filter: filter                                                                                                   // 146
    }; // perform LDAP search to determine DN                                                                          // 143
                                                                                                                       //
    client.search(_this.options.base, searchOptions, function (err, res) {                                             // 150
      if (err) {                                                                                                       // 151
        if (resolved) return;                                                                                          // 152
        resolved = true;                                                                                               // 153
        ldapAsyncFut.return({                                                                                          // 154
          error: err                                                                                                   // 155
        });                                                                                                            // 154
        return;                                                                                                        // 157
      }                                                                                                                // 158
                                                                                                                       //
      retObject.emptySearch = true;                                                                                    // 160
      res.on("searchEntry", function (entry) {                                                                         // 161
        retObject.dn = entry.objectName;                                                                               // 162
        retObject.username = retObject.dn;                                                                             // 163
        retObject.emptySearch = false;                                                                                 // 164
        retObject.searchResults = _.omit(entry.object, "userPassword");                                                // 166
                                                                                                                       //
        if (options.hasOwnProperty("searchUsername")) {                                                                // 168
          // This was only a search, return immediately                                                                // 169
          resolved = true;                                                                                             // 170
          ldapAsyncFut.return(retObject);                                                                              // 171
          return;                                                                                                      // 172
        } // use the determined DN to bind                                                                             // 173
                                                                                                                       //
                                                                                                                       //
        client.bind(entry.objectName, options.ldapPass, function (err) {                                               // 176
          try {                                                                                                        // 177
            if (err) {                                                                                                 // 178
              throw new Meteor.Error(err.code, err.message);                                                           // 179
            } else {                                                                                                   // 180
              resolved = true;                                                                                         // 181
              ldapAsyncFut.return(retObject);                                                                          // 182
            }                                                                                                          // 183
          } catch (e) {                                                                                                // 184
            if (resolved) return;                                                                                      // 186
            resolved = true;                                                                                           // 187
            ldapAsyncFut.return({                                                                                      // 188
              error: e                                                                                                 // 189
            });                                                                                                        // 188
          }                                                                                                            // 191
        });                                                                                                            // 192
      });                                                                                                              // 193
      res.on("error", errorFunc); // If no dn is found, return as is.                                                  // 195
                                                                                                                       //
      res.on("end", function (result) {                                                                                // 198
        if (retObject.dn === undefined) {                                                                              // 199
          resolved = true;                                                                                             // 200
          ldapAsyncFut.return(retObject);                                                                              // 201
        }                                                                                                              // 202
      });                                                                                                              // 203
    });                                                                                                                // 204
    return ldapAsyncFut.wait();                                                                                        // 206
  } else {                                                                                                             // 208
    throw new Meteor.Error(403, "Missing LDAP Auth Parameter");                                                        // 209
  }                                                                                                                    // 210
};                                                                                                                     // 212
                                                                                                                       //
LDAP.prototype.updateUserQuota = function (db, user) {                                                                 // 214
  var fallback = {                                                                                                     // 215
    storage: user.cachedStorageQuota || 0,                                                                             // 216
    grains: Infinity,                                                                                                  // 217
    compute: Infinity                                                                                                  // 218
  };                                                                                                                   // 215
  var setting = db.collections.settings.findOne({                                                                      // 221
    _id: "quotaLdapAttribute"                                                                                          // 221
  });                                                                                                                  // 221
  if (!setting || !setting.value) return fallback; // TODO(someday): don't just assume the first login identity is the primary identity?
                                                                                                                       //
  var email = db.getPrimaryEmail(user._id, user.loginIdentities[0].id);                                                // 225
  if (!email) return fallback;                                                                                         // 226
  var ldapUser = void 0;                                                                                               // 228
                                                                                                                       //
  try {                                                                                                                // 229
    ldapUser = this.ldapCheck(db, {                                                                                    // 230
      searchUsername: email,                                                                                           // 230
      searchUsernameField: "mail"                                                                                      // 230
    });                                                                                                                // 230
  } catch (err) {                                                                                                      // 231
    console.error("Error looking up quota from LDAP");                                                                 // 232
    return fallback;                                                                                                   // 233
  }                                                                                                                    // 234
                                                                                                                       //
  if (!ldapUser || !ldapUser.searchResults) return fallback;                                                           // 236
  var newStorageQuota = +ldapUser.searchResults[setting.value];                                                        // 238
                                                                                                                       //
  if (newStorageQuota !== user.cachedStorageQuota) {                                                                   // 239
    Meteor.users.update({                                                                                              // 240
      _id: user._id                                                                                                    // 240
    }, {                                                                                                               // 240
      $set: {                                                                                                          // 240
        cachedStorageQuota: newStorageQuota                                                                            // 240
      }                                                                                                                // 240
    });                                                                                                                // 240
  } // TODO(someday): cache timestamp as well and only check/update if greater than 60s ago                            // 241
                                                                                                                       //
                                                                                                                       //
  return {                                                                                                             // 244
    storage: newStorageQuota,                                                                                          // 245
    grains: Infinity,                                                                                                  // 246
    compute: Infinity                                                                                                  // 247
  };                                                                                                                   // 244
};                                                                                                                     // 249
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"picture.js":["meteor/http",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/accounts/picture.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  userPictureUrl: function () {                                                                                        // 1
    return userPictureUrl;                                                                                             // 1
  },                                                                                                                   // 1
  fetchPicture: function () {                                                                                          // 1
    return fetchPicture;                                                                                               // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var HTTP = void 0;                                                                                                     // 1
module.import("meteor/http", {                                                                                         // 1
  "HTTP": function (v) {                                                                                               // 1
    HTTP = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var userPictureUrl = function (user) {                                                                                 // 3
  if (user.services && !(user.profile && user.profile.picture)) {                                                      // 4
    // Try to determine user's avatar URL from login service.                                                          // 5
    var google = user.services.google;                                                                                 // 7
                                                                                                                       //
    if (google && google.picture) {                                                                                    // 8
      return google.picture;                                                                                           // 9
    }                                                                                                                  // 10
                                                                                                                       //
    var github = user.services.github;                                                                                 // 12
                                                                                                                       //
    if (github && github.id) {                                                                                         // 13
      return "https://avatars.githubusercontent.com/u/" + github.id;                                                   // 14
    } // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be                    // 15
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                          // 18
    // Github are different because they are actually the identity providers, so they already know                     // 19
    // the user logged in.                                                                                             // 20
                                                                                                                       //
  }                                                                                                                    // 21
};                                                                                                                     // 22
                                                                                                                       //
var fetchPicture = function (db, url) {                                                                                // 24
  try {                                                                                                                // 25
    var result = HTTP.get(url, {                                                                                       // 26
      npmRequestOptions: {                                                                                             // 27
        encoding: null                                                                                                 // 27
      },                                                                                                               // 27
      timeout: 5000                                                                                                    // 28
    });                                                                                                                // 26
    var metadata = {};                                                                                                 // 31
    metadata.mimeType = result.headers["content-type"];                                                                // 33
                                                                                                                       //
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 && metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                                  // 36
    }                                                                                                                  // 37
                                                                                                                       //
    var enc = result.headers["content-encoding"];                                                                      // 39
                                                                                                                       //
    if (enc && enc !== "identity") {                                                                                   // 40
      metadata.encoding = enc;                                                                                         // 41
    }                                                                                                                  // 42
                                                                                                                       //
    return db.addStaticAsset(metadata, result.content);                                                                // 44
  } catch (err) {                                                                                                      // 45
    console.error("failed to fetch user profile picture:", url, err.stack);                                            // 46
  }                                                                                                                    // 47
};                                                                                                                     // 48
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"saml-utils.js":["crypto","querystring","url","xml2js","xmlbuilder","xml-crypto","xmldom","zlib",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/accounts/saml-utils.js                                                                               //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  SAML: function () {                                                                                                  // 1
    return SAML;                                                                                                       // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var crypto = void 0;                                                                                                   // 1
module.import("crypto", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var querystring = void 0;                                                                                              // 1
module.import("querystring", {                                                                                         // 1
  "default": function (v) {                                                                                            // 1
    querystring = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Url = void 0;                                                                                                      // 1
module.import("url", {                                                                                                 // 1
  "default": function (v) {                                                                                            // 1
    Url = v;                                                                                                           // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var xml2js = void 0;                                                                                                   // 1
module.import("xml2js", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    xml2js = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var xmlbuilder = void 0;                                                                                               // 1
module.import("xmlbuilder", {                                                                                          // 1
  "default": function (v) {                                                                                            // 1
    xmlbuilder = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var xmlCrypto = void 0;                                                                                                // 1
module.import("xml-crypto", {                                                                                          // 1
  "default": function (v) {                                                                                            // 1
    xmlCrypto = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var xmldom = void 0;                                                                                                   // 1
module.import("xmldom", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    xmldom = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
var zlib = void 0;                                                                                                     // 1
module.import("zlib", {                                                                                                // 1
  "default": function (v) {                                                                                            // 1
    zlib = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 7);                                                                                                                 // 1
var HOSTNAME = Url.parse(process.env.ROOT_URL).hostname;                                                               // 10
                                                                                                                       //
var SAML = function (options) {                                                                                        // 12
  this.options = this.initialize(options);                                                                             // 13
};                                                                                                                     // 14
                                                                                                                       //
SAML.prototype.initialize = function (options) {                                                                       // 16
  if (!options) {                                                                                                      // 17
    options = {};                                                                                                      // 18
  }                                                                                                                    // 19
                                                                                                                       //
  if (!options.protocol) {                                                                                             // 21
    options.protocol = "https://";                                                                                     // 22
  }                                                                                                                    // 23
                                                                                                                       //
  if (!options.path) {                                                                                                 // 25
    options.path = "/saml/consume";                                                                                    // 26
  }                                                                                                                    // 27
                                                                                                                       //
  if (!options.issuer) {                                                                                               // 29
    options.issuer = "onelogin_saml";                                                                                  // 30
  }                                                                                                                    // 31
                                                                                                                       //
  if (options.identifierFormat === undefined) {                                                                        // 33
    options.identifierFormat = "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent";                                 // 34
  }                                                                                                                    // 35
                                                                                                                       //
  return options;                                                                                                      // 37
};                                                                                                                     // 38
                                                                                                                       //
SAML.prototype.generateUniqueID = function () {                                                                        // 40
  var chars = "abcdef0123456789";                                                                                      // 41
  var uniqueID = "";                                                                                                   // 42
                                                                                                                       //
  for (var i = 0; i < 20; i++) {                                                                                       // 43
    uniqueID += chars.substr(Math.floor(Math.random() * 15), 1);                                                       // 44
  }                                                                                                                    // 45
                                                                                                                       //
  return uniqueID;                                                                                                     // 47
};                                                                                                                     // 48
                                                                                                                       //
SAML.prototype.generateInstant = function () {                                                                         // 50
  var date = new Date();                                                                                               // 51
  return date.getUTCFullYear() + "-" + ("0" + (date.getUTCMonth() + 1)).slice(-2) + "-" + ("0" + date.getUTCDate()).slice(-2) + "T" + ("0" + date.getUTCHours()).slice(-2) + ":" + ("0" + date.getUTCMinutes()).slice(-2) + ":" + ("0" + date.getUTCSeconds()).slice(-2) + "Z";
};                                                                                                                     // 53
                                                                                                                       //
SAML.prototype.signRequest = function (xml) {                                                                          // 55
  var signer = crypto.createSign("RSA-SHA1");                                                                          // 56
  signer.update(xml);                                                                                                  // 57
  return signer.sign(this.options.privateCert, "base64");                                                              // 58
};                                                                                                                     // 59
                                                                                                                       //
SAML.prototype.generateAuthorizeRequest = function (req) {                                                             // 61
  var id = "_" + this.generateUniqueID();                                                                              // 62
  var instant = this.generateInstant();                                                                                // 63
  var callbackUrl = void 0; // Post-auth destination                                                                   // 65
                                                                                                                       //
  if (this.options.callbackUrl) {                                                                                      // 67
    callbackUrl = this.options.callbackUrl;                                                                            // 68
  } else {                                                                                                             // 69
    callbackUrl = this.options.protocol + req.headers.host + this.options.path;                                        // 70
  }                                                                                                                    // 71
                                                                                                                       //
  if (this.options.id) id = this.options.id;                                                                           // 73
  var request = "<samlp:AuthnRequest xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" ID=\"" + id + "\" Version=\"2.0\" IssueInstant=\"" + instant + "\" ProtocolBinding=\"urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST\" AssertionConsumerServiceURL=\"" + callbackUrl + "\" Destination=\"" + this.options.entryPoint + "\">" + "<saml:Issuer xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\">" + this.options.issuer + "</saml:Issuer>\n";
                                                                                                                       //
  if (this.options.identifierFormat) {                                                                                 // 82
    request += "<samlp:NameIDPolicy xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" Format=\"" + this.options.identifierFormat + "\" AllowCreate=\"true\"></samlp:NameIDPolicy>\n";
  }                                                                                                                    // 85
                                                                                                                       //
  request += "</samlp:AuthnRequest>";                                                                                  // 87
  return request;                                                                                                      // 89
};                                                                                                                     // 90
                                                                                                                       //
SAML.prototype.generateLogoutRequest = function (req) {                                                                // 92
  var id = "_" + this.generateUniqueID();                                                                              // 93
  var instant = this.generateInstant(); //samlp:LogoutRequest xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol" xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"
  // ID="_135ad2fd-b275-4428-b5d6-3ac3361c3a7f" Version="2.0" Destination="https://idphost/adfs/ls/"                   // 97
  //IssueInstant="2008-06-03T12:59:57Z"><saml:Issuer>myhost</saml:Issuer><NameID Format="urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"
  //NameQualifier="https://idphost/adfs/ls/">myemail@mydomain.com</NameID<samlp:SessionIndex>_0628125f-7f95-42cc-ad8e-fde86ae90bbe
  //</samlp:SessionIndex></samlp:LogoutRequest>                                                                        // 100
                                                                                                                       //
  var request = "<samlp:LogoutRequest xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" " + "xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\" ID=\"" + id + "\" Version=\"2.0\" IssueInstant=\"" + instant + "\" Destination=\"" + this.options.entryPoint + "\">" + "<saml:Issuer xmlns:saml=\"urn:oasis:names:tc:SAML:2.0:assertion\">" + this.options.issuer + "</saml:Issuer>" + "<saml:NameID Format=\"" + req.user.nameIDFormat + "\">" + req.user.nameID + "</saml:NameID>" + "</samlp:LogoutRequest>";
  return request;                                                                                                      // 108
};                                                                                                                     // 109
                                                                                                                       //
SAML.prototype.requestToUrl = function (request, operation, callback) {                                                // 111
  var _this = this;                                                                                                    // 112
                                                                                                                       //
  zlib.deflateRaw(request, function (err, buffer) {                                                                    // 113
    if (err) {                                                                                                         // 114
      return callback(err);                                                                                            // 115
    }                                                                                                                  // 116
                                                                                                                       //
    var base64 = buffer.toString("base64");                                                                            // 118
    var target = _this.options.entryPoint;                                                                             // 119
                                                                                                                       //
    if (operation === "logout") {                                                                                      // 121
      if (_this.options.logoutUrl) {                                                                                   // 122
        target = _this.options.logoutUrl;                                                                              // 123
      }                                                                                                                // 124
    }                                                                                                                  // 125
                                                                                                                       //
    if (target.indexOf("?") > 0) target += "&";else target += "?";                                                     // 127
    var samlRequest = {                                                                                                // 132
      SAMLRequest: base64                                                                                              // 133
    };                                                                                                                 // 132
                                                                                                                       //
    if (_this.options.privateCert) {                                                                                   // 136
      samlRequest.SigAlg = "http://www.w3.org/2000/09/xmldsig#rsa-sha1";                                               // 137
      samlRequest.Signature = _this.signRequest(querystring.stringify(samlRequest));                                   // 138
    }                                                                                                                  // 139
                                                                                                                       //
    target += querystring.stringify(samlRequest);                                                                      // 141
    callback(null, target);                                                                                            // 143
  });                                                                                                                  // 144
};                                                                                                                     // 145
                                                                                                                       //
SAML.prototype.getAuthorizeUrl = function (req, callback) {                                                            // 147
  var request = this.generateAuthorizeRequest(req);                                                                    // 148
  this.requestToUrl(request, "authorize", callback);                                                                   // 150
};                                                                                                                     // 151
                                                                                                                       //
SAML.prototype.getLogoutUrl = function (req, callback) {                                                               // 153
  var request = this.generateLogoutRequest(req);                                                                       // 154
  this.requestToUrl(request, "logout", callback);                                                                      // 156
};                                                                                                                     // 157
                                                                                                                       //
SAML.prototype.certToPEM = function (cert) {                                                                           // 159
  cert = cert.match(/.{1,64}/g).join("\n");                                                                            // 160
  cert = "-----BEGIN CERTIFICATE-----\n" + cert;                                                                       // 161
  cert = cert + "\n-----END CERTIFICATE-----\n";                                                                       // 162
  return cert;                                                                                                         // 163
};                                                                                                                     // 164
                                                                                                                       //
SAML.prototype.validateSignature = function (xml, cert) {                                                              // 166
  var _this = this;                                                                                                    // 167
                                                                                                                       //
  var doc = new xmldom.DOMParser().parseFromString(xml);                                                               // 168
  var signature = xmlCrypto.xpath(doc, "//*[local-name(.)='Signature' and namespace-uri(.)='http://www.w3.org/2000/09/xmldsig#']")[0];
  var sig = new xmlCrypto.SignedXml();                                                                                 // 170
  sig.keyInfoProvider = {                                                                                              // 171
    getKeyInfo: function (key) {                                                                                       // 172
      return "<X509Data></X509Data>";                                                                                  // 173
    },                                                                                                                 // 174
    getKey: function (keyInfo) {                                                                                       // 176
      return _this.certToPEM(cert);                                                                                    // 177
    }                                                                                                                  // 178
  };                                                                                                                   // 171
  sig.loadSignature(signature.toString());                                                                             // 180
  return sig.checkSignature(xml);                                                                                      // 181
};                                                                                                                     // 182
                                                                                                                       //
SAML.prototype.getElement = function (parentElement, elementName) {                                                    // 184
  if (parentElement["saml:" + elementName]) {                                                                          // 185
    return parentElement["saml:" + elementName];                                                                       // 186
  } else if (parentElement["samlp:" + elementName]) {                                                                  // 187
    return parentElement["samlp:" + elementName];                                                                      // 188
  } else if (parentElement["saml2p:" + elementName]) {                                                                 // 189
    return parentElement["saml2p:" + elementName];                                                                     // 190
  } else if (parentElement["saml2:" + elementName]) {                                                                  // 191
    return parentElement["saml2:" + elementName];                                                                      // 192
  }                                                                                                                    // 193
                                                                                                                       //
  return parentElement[elementName];                                                                                   // 195
};                                                                                                                     // 196
                                                                                                                       //
SAML.prototype.parseLogoutRequest = function (xml, callback) {                                                         // 198
  var _this = this;                                                                                                    // 199
                                                                                                                       //
  var parser = new xml2js.Parser({                                                                                     // 200
    explicitRoot: true                                                                                                 // 200
  });                                                                                                                  // 200
  parser.parseString(xml, function (err, doc) {                                                                        // 201
    if (err) {                                                                                                         // 202
      callback(err);                                                                                                   // 203
      return;                                                                                                          // 204
    }                                                                                                                  // 205
                                                                                                                       //
    var request = _this.getElement(doc, "LogoutRequest");                                                              // 207
                                                                                                                       //
    var nameId = _this.getElement(request, "NameID")[0];                                                               // 208
                                                                                                                       //
    callback(null, nameId && nameId._);                                                                                // 209
  });                                                                                                                  // 210
};                                                                                                                     // 211
                                                                                                                       //
SAML.prototype.validateResponse = function (samlResponse, callback) {                                                  // 213
  var _this = this;                                                                                                    // 214
                                                                                                                       //
  var xml = new Buffer(samlResponse, "base64").toString("utf8");                                                       // 215
  var parser = new xml2js.Parser({                                                                                     // 216
    explicitRoot: true                                                                                                 // 216
  });                                                                                                                  // 216
  parser.parseString(xml, function (err, doc) {                                                                        // 217
    // Verify signature                                                                                                // 218
    if (_this.options.cert && !_this.validateSignature(xml, _this.options.cert)) {                                     // 219
      return callback(new Error("Invalid signature"), null, false, xml);                                               // 220
    }                                                                                                                  // 221
                                                                                                                       //
    var response = _this.getElement(doc, "Response");                                                                  // 223
                                                                                                                       //
    if (response) {                                                                                                    // 224
      var assertion = _this.getElement(response, "Assertion");                                                         // 225
                                                                                                                       //
      if (!assertion) {                                                                                                // 226
        return callback(new Error("Missing SAML assertion"), null, false, xml);                                        // 227
      }                                                                                                                // 228
                                                                                                                       //
      profile = {};                                                                                                    // 230
                                                                                                                       //
      if (response.$ && response.$.InResponseTo) {                                                                     // 232
        profile.inResponseToId = response.$.InResponseTo;                                                              // 233
      }                                                                                                                // 234
                                                                                                                       //
      if (response.$ && response.$.Destination) {                                                                      // 236
        if (!response.$.Destination.startsWith(process.env.ROOT_URL)) {                                                // 237
          return callback(new Error("SAML Response received with invalid Destination: " + response.$.Destination), null, false, xml);
        }                                                                                                              // 240
      }                                                                                                                // 241
                                                                                                                       //
      var issuer = _this.getElement(assertion[0], "Issuer");                                                           // 243
                                                                                                                       //
      if (issuer) {                                                                                                    // 244
        profile.issuer = issuer[0];                                                                                    // 245
      }                                                                                                                // 246
                                                                                                                       //
      var subject = _this.getElement(assertion[0], "Subject");                                                         // 248
                                                                                                                       //
      if (subject) {                                                                                                   // 249
        var nameID = _this.getElement(subject[0], "NameID");                                                           // 250
                                                                                                                       //
        if (nameID) {                                                                                                  // 251
          profile.nameID = nameID[0]._;                                                                                // 252
                                                                                                                       //
          if (nameID[0].$.Format) {                                                                                    // 254
            profile.nameIDFormat = nameID[0].$.Format;                                                                 // 255
                                                                                                                       //
            if (profile.nameIDFormat.toLowerCase().indexOf("transient") !== -1) {                                      // 256
              return callback(new Error("SAML returned a transient NameID. Sandstorm requires a persistent NameID. " + "Please check your IdP config."), null, false, xml);
            }                                                                                                          // 260
          }                                                                                                            // 261
        }                                                                                                              // 262
                                                                                                                       //
        var subjectConfirmation = _this.getElement(subject[0], "SubjectConfirmation");                                 // 264
                                                                                                                       //
        if (subjectConfirmation) {                                                                                     // 265
          var subjectConfirmationData = _this.getElement(subjectConfirmation[0], "SubjectConfirmationData")[0];        // 266
                                                                                                                       //
          if (subjectConfirmationData) {                                                                               // 268
            var recipient = subjectConfirmationData.$.Recipient;                                                       // 269
                                                                                                                       //
            if (recipient && !recipient.startsWith(process.env.ROOT_URL)) {                                            // 270
              return callback(new Error("SAML sent to wrong recipient"), null, false, xml);                            // 271
            }                                                                                                          // 272
                                                                                                                       //
            var nowMs = Date.now();                                                                                    // 274
            var notBefore = subjectConfirmationData.$.NotBefore;                                                       // 275
                                                                                                                       //
            if (notBefore && nowMs < Date.parse(notBefore)) {                                                          // 276
              return callback(new Error("SAML assertion was signed for the future."), null, false, xml);               // 277
            }                                                                                                          // 278
                                                                                                                       //
            var notOnOrAfter = subjectConfirmationData.$.NotOnOrAfter;                                                 // 280
                                                                                                                       //
            if (notOnOrAfter && nowMs >= Date.parse(notOnOrAfter)) {                                                   // 281
              return callback(new Error("SAML assertion was signed for the past."), null, false, xml);                 // 282
            }                                                                                                          // 283
          }                                                                                                            // 284
        }                                                                                                              // 285
      }                                                                                                                // 286
                                                                                                                       //
      var conditions = _this.getElement(assertion[0], "Conditions")[0];                                                // 288
                                                                                                                       //
      if (conditions) {                                                                                                // 289
        for (var key in meteorBabelHelpers.sanitizeForInObject(conditions)) {                                          // 290
          if (conditions.hasOwnProperty(key)) {                                                                        // 291
            var value = conditions[key];                                                                               // 292
                                                                                                                       //
            if (key === "$") {                                                                                         // 293
              var _nowMs = Date.now();                                                                                 // 294
                                                                                                                       //
              var _notBefore = value.NotBefore;                                                                        // 295
                                                                                                                       //
              if (_notBefore && _nowMs < Date.parse(_notBefore)) {                                                     // 296
                return callback(new Error("SAML condition NotBefore is in the future."), null, false, xml);            // 297
              }                                                                                                        // 299
                                                                                                                       //
              var _notOnOrAfter = value.NotOnOrAfter;                                                                  // 301
                                                                                                                       //
              if (_notOnOrAfter && _nowMs >= Date.parse(_notOnOrAfter)) {                                              // 302
                return callback(new Error("SAML condition notOnOrAfter is in the past."), null, false, xml);           // 303
              }                                                                                                        // 305
            } else if (key.endsWith("AudienceRestriction") || key.endsWith("OneTimeUse") || key.endsWith("ProxyRestriction")) {
              continue; // Do nothing.                                                                                 // 309
              // We already check both Destination and SubjectConfirmation.Recipient, so                               // 311
              // the Audience constraint isn't as important, and is tricky to handle                                   // 312
              // correctly.                                                                                            // 313
              // OneTimeUse is actually very tricky to get right, and requires a cache of                              // 314
              // previously seen assertions (or at least their hashes). In Sandstorm's case,                           // 315
              // we would probably need to stuff it in Mongo to make it robust across multiple                         // 316
              // front-ends. It is a very rarely used constraint anyways.                                              // 317
              // ProxyRestriction constraints only govern the future use of the assertion by the                       // 318
              // SP. We don't fall under this constraint so it's meaningless to us. As per the                         // 319
              // spec, it is always considered valid.                                                                  // 320
            } else {                                                                                                   // 321
              return callback(new Error("Unrecognized SAML constraint: " + key), null, false, xml);                    // 322
            }                                                                                                          // 323
          }                                                                                                            // 324
        }                                                                                                              // 325
      }                                                                                                                // 326
                                                                                                                       //
      var attributeStatement = _this.getElement(assertion[0], "AttributeStatement");                                   // 328
                                                                                                                       //
      if (attributeStatement) {                                                                                        // 329
        var attributes = _this.getElement(attributeStatement[0], "Attribute");                                         // 330
                                                                                                                       //
        if (attributes) {                                                                                              // 332
          attributes.forEach(function (attribute) {                                                                    // 333
            var value = _this.getElement(attribute, "AttributeValue");                                                 // 334
                                                                                                                       //
            if (typeof value[0] === "string") {                                                                        // 335
              profile[attribute.$.Name] = value[0];                                                                    // 336
            } else {                                                                                                   // 337
              profile[attribute.$.Name] = value[0]._;                                                                  // 338
            }                                                                                                          // 339
          });                                                                                                          // 340
        }                                                                                                              // 341
                                                                                                                       //
        if (!profile.mail && profile["urn:oid:0.9.2342.19200300.100.1.3"]) {                                           // 343
          // See http://www.incommonfederation.org/attributesummary.html for definition of attribute OIDs              // 344
          profile.mail = profile["urn:oid:0.9.2342.19200300.100.1.3"];                                                 // 345
        }                                                                                                              // 346
                                                                                                                       //
        if (!profile.email && profile.mail) {                                                                          // 348
          profile.email = profile.mail;                                                                                // 349
        }                                                                                                              // 350
                                                                                                                       //
        var microsoftEmail = profile["http://schemas.xmlsoap.org/ws/2005/05/identity/claims/emailaddress"];            // 352
                                                                                                                       //
        if (!profile.email && microsoftEmail) {                                                                        // 353
          profile.email = microsoftEmail;                                                                              // 354
        }                                                                                                              // 355
                                                                                                                       //
        var azureAdEmail = profile["http://schemas.xmlsoap.org/ws/2005/05/identity/claims/name"];                      // 357
                                                                                                                       //
        if (!profile.email && azureAdEmail) {                                                                          // 358
          profile.email = azureAdEmail;                                                                                // 359
        }                                                                                                              // 360
                                                                                                                       //
        var microsoftDisplayName = profile["http://schemas.microsoft.com/identity/claims/displayname"];                // 362
                                                                                                                       //
        if (!profile.displayName && microsoftDisplayName) {                                                            // 363
          profile.displayName = microsoftDisplayName;                                                                  // 364
        }                                                                                                              // 365
      }                                                                                                                // 366
                                                                                                                       //
      if (!profile.email && profile.nameID && profile.nameIDFormat && profile.nameIDFormat.indexOf("emailAddress") >= 0) {
        profile.email = profile.nameID;                                                                                // 369
      }                                                                                                                // 370
                                                                                                                       //
      callback(null, profile, false, xml);                                                                             // 372
    } else {                                                                                                           // 373
      var logoutResponse = _this.getElement(doc, "LogoutResponse");                                                    // 374
                                                                                                                       //
      if (logoutResponse) {                                                                                            // 376
        callback(null, null, true, xml);                                                                               // 377
      } else {                                                                                                         // 378
        return callback(new Error("Unknown SAML response message"), null, false, xml);                                 // 379
      }                                                                                                                // 380
    }                                                                                                                  // 382
  });                                                                                                                  // 384
};                                                                                                                     // 385
                                                                                                                       //
SAML.prototype.generateServiceProviderMetadata = function () {                                                         // 387
  var entityId = this.options.issuer;                                                                                  // 388
  var metadata = {                                                                                                     // 389
    "EntityDescriptor": {                                                                                              // 390
      "@xmlns": "urn:oasis:names:tc:SAML:2.0:metadata",                                                                // 391
      "@xmlns:ds": "http://www.w3.org/2000/09/xmldsig#",                                                               // 392
      "@entityID": entityId,                                                                                           // 393
      "@ID": entityId.replace(/\W/g, "_"),                                                                             // 394
      "SPSSODescriptor": {                                                                                             // 395
        "@protocolSupportEnumeration": "urn:oasis:names:tc:SAML:2.0:protocol"                                          // 396
      }                                                                                                                // 395
    }                                                                                                                  // 390
  };                                                                                                                   // 389
  var decryptionCert = this.options.localCertPublic; // TODO(someday); support local certs                             // 401
                                                                                                                       //
  if (decryptionCert) {                                                                                                // 402
    decryptionCert = decryptionCert.replace(/-+BEGIN CERTIFICATE-+\r?\n?/, "");                                        // 403
    decryptionCert = decryptionCert.replace(/-+END CERTIFICATE-+\r?\n?/, "");                                          // 404
    decryptionCert = decryptionCert.replace(/\r\n/g, "\n");                                                            // 405
    metadata.EntityDescriptor.SPSSODescriptor.KeyDescriptor = {                                                        // 407
      "ds:KeyInfo": {                                                                                                  // 408
        "ds:X509Data": {                                                                                               // 409
          "ds:X509Certificate": {                                                                                      // 410
            "#text": decryptionCert                                                                                    // 411
          }                                                                                                            // 410
        }                                                                                                              // 409
      },                                                                                                               // 408
      "EncryptionMethod": [// this should be the set that the xmlenc library supports                                  // 415
      {                                                                                                                // 417
        "@Algorithm": "http://www.w3.org/2001/04/xmlenc#aes256-cbc"                                                    // 417
      }, {                                                                                                             // 417
        "@Algorithm": "http://www.w3.org/2001/04/xmlenc#aes128-cbc"                                                    // 418
      }, {                                                                                                             // 418
        "@Algorithm": "http://www.w3.org/2001/04/xmlenc#tripledes-cbc"                                                 // 419
      }]                                                                                                               // 419
    };                                                                                                                 // 407
  }                                                                                                                    // 422
                                                                                                                       //
  metadata.EntityDescriptor.SPSSODescriptor.SingleLogoutService = {                                                    // 424
    "@Binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",                                                  // 425
    "@Location": Meteor.absoluteUrl("saml/logout/default")                                                             // 426
  };                                                                                                                   // 424
  metadata.EntityDescriptor.SPSSODescriptor.NameIDFormat = this.options.identifierFormat;                              // 429
  metadata.EntityDescriptor.SPSSODescriptor.AssertionConsumerService = {                                               // 430
    "@index": "1",                                                                                                     // 431
    "@isDefault": "true",                                                                                              // 432
    "@Binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",                                                      // 433
    "@Location": Meteor.absoluteUrl("_saml/validate/default")                                                          // 434
  };                                                                                                                   // 430
  return xmlbuilder.create(metadata).end({                                                                             // 437
    pretty: true,                                                                                                      // 437
    indent: "  ",                                                                                                      // 437
    newline: "\n"                                                                                                      // 437
  });                                                                                                                  // 437
};                                                                                                                     // 438
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"async-helpers.js":["meteor/meteor",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/async-helpers.js                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  inMeteor: function () {                                                                                              // 1
    return inMeteor;                                                                                                   // 1
  },                                                                                                                   // 1
  promiseToFuture: function () {                                                                                       // 1
    return promiseToFuture;                                                                                            // 1
  },                                                                                                                   // 1
  waitPromise: function () {                                                                                           // 1
    return waitPromise;                                                                                                // 1
  },                                                                                                                   // 1
  onInMeteor: function () {                                                                                            // 1
    return onInMeteor;                                                                                                 // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 3
                                                                                                                       //
var inMeteorListener = undefined;                                                                                      // 5
                                                                                                                       //
var onInMeteor = function (callback) {                                                                                 // 6
  inMeteorListener = callback;                                                                                         // 7
}; // Meteor context <-> Async Node.js context adapters                                                                // 8
                                                                                                                       //
                                                                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {                                                    // 11
  callback();                                                                                                          // 12
});                                                                                                                    // 13
                                                                                                                       //
var inMeteor = function (callback) {                                                                                   // 15
  if (inMeteorListener) {                                                                                              // 16
    inMeteorListener();                                                                                                // 17
  } // Calls the callback in a Meteor context.  Returns a Promise for its result.                                      // 18
                                                                                                                       //
                                                                                                                       //
  return new Promise(function (resolve, reject) {                                                                      // 21
    inMeteorInternal(function () {                                                                                     // 22
      try {                                                                                                            // 23
        resolve(callback());                                                                                           // 24
      } catch (err) {                                                                                                  // 25
        reject(err);                                                                                                   // 26
      }                                                                                                                // 27
    });                                                                                                                // 28
  });                                                                                                                  // 29
};                                                                                                                     // 30
                                                                                                                       //
var promiseToFuture = function (promise) {                                                                             // 32
  var result = new Future();                                                                                           // 33
  promise.then(result.return.bind(result), result.throw.bind(result));                                                 // 34
  return result;                                                                                                       // 35
};                                                                                                                     // 36
                                                                                                                       //
var waitPromise = function (promise) {                                                                                 // 38
  return promiseToFuture(promise).wait();                                                                              // 39
};                                                                                                                     // 40
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"auth.js":["meteor/meteor","meteor/check","crypto","fs","/imports/server/constants.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/auth.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  checkAuth: function () {                                                                                             // 1
    return checkAuth;                                                                                                  // 1
  },                                                                                                                   // 1
  clearAdminToken: function () {                                                                                       // 1
    return clearAdminToken;                                                                                            // 1
  },                                                                                                                   // 1
  tokenIsValid: function () {                                                                                          // 1
    return tokenIsValid;                                                                                               // 1
  },                                                                                                                   // 1
  tokenIsSetupSession: function () {                                                                                   // 1
    return tokenIsSetupSession;                                                                                        // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Match = void 0,                                                                                                    // 1
    check = void 0;                                                                                                    // 1
module.import("meteor/check", {                                                                                        // 1
  "Match": function (v) {                                                                                              // 1
    Match = v;                                                                                                         // 1
  },                                                                                                                   // 1
  "check": function (v) {                                                                                              // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Crypto = void 0;                                                                                                   // 1
module.import("crypto", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    Crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var Fs = void 0;                                                                                                       // 1
module.import("fs", {                                                                                                  // 1
  "default": function (v) {                                                                                            // 1
    Fs = v;                                                                                                            // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var SANDSTORM_VARDIR = void 0;                                                                                         // 1
module.import("/imports/server/constants.js", {                                                                        // 1
  "SANDSTORM_VARDIR": function (v) {                                                                                   // 1
    SANDSTORM_VARDIR = v;                                                                                              // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
// TODO(cleanup): globalDb is still an unbound global, but extracting it is Hard.                                      // 6
var ADMIN_TOKEN_EXPIRATION_TIME = 60 * 60 * 1000;                                                                      // 8
var SANDSTORM_ADMIN_TOKEN = SANDSTORM_VARDIR + "/adminToken";                                                          // 9
                                                                                                                       //
function clearAdminToken(token) {                                                                                      // 11
  if (tokenIsSetupSession(token)) {                                                                                    // 12
    var hash = Crypto.createHash("sha256").update(token).digest("base64");                                             // 13
    globalDb.collections.setupSession.remove({                                                                         // 14
      _id: "current-session",                                                                                          // 15
      hashedSessionId: hash                                                                                            // 16
    });                                                                                                                // 14
  }                                                                                                                    // 18
                                                                                                                       //
  if (tokenIsValid(token)) {                                                                                           // 20
    Fs.unlinkSync(SANDSTORM_ADMIN_TOKEN);                                                                              // 21
    console.log("Admin token deleted.");                                                                               // 22
  }                                                                                                                    // 23
}                                                                                                                      // 24
                                                                                                                       //
function tokenIsValid(token) {                                                                                         // 26
  if (token && Fs.existsSync(SANDSTORM_ADMIN_TOKEN)) {                                                                 // 27
    var stats = Fs.statSync(SANDSTORM_ADMIN_TOKEN);                                                                    // 28
    var expireTime = new Date(Date.now() - ADMIN_TOKEN_EXPIRATION_TIME);                                               // 29
                                                                                                                       //
    if (stats.mtime < expireTime) {                                                                                    // 30
      return false;                                                                                                    // 31
    } else {                                                                                                           // 32
      return Fs.readFileSync(SANDSTORM_ADMIN_TOKEN, {                                                                  // 33
        encoding: "utf8"                                                                                               // 33
      }) === token;                                                                                                    // 33
    }                                                                                                                  // 34
  } else {                                                                                                             // 35
    return false;                                                                                                      // 36
  }                                                                                                                    // 37
}                                                                                                                      // 38
                                                                                                                       //
function tokenIsSetupSession(token) {                                                                                  // 40
  if (token) {                                                                                                         // 41
    var setupSession = globalDb.collections.setupSession.findOne({                                                     // 42
      _id: "current-session"                                                                                           // 42
    });                                                                                                                // 42
                                                                                                                       //
    if (setupSession) {                                                                                                // 43
      var hash = Crypto.createHash("sha256").update(token).digest("base64");                                           // 44
      var now = new Date();                                                                                            // 45
      var sessionLifetime = 24 * 60 * 60 * 1000; // length of setup session validity, in milliseconds: 1 day           // 46
                                                                                                                       //
      if (setupSession.hashedSessionId === hash && now - setupSession.creationDate < sessionLifetime) {                // 47
        return true;                                                                                                   // 48
      }                                                                                                                // 49
    }                                                                                                                  // 50
  }                                                                                                                    // 51
                                                                                                                       //
  return false;                                                                                                        // 53
}                                                                                                                      // 54
                                                                                                                       //
function checkAuth(token) {                                                                                            // 56
  check(token, Match.OneOf(undefined, null, String));                                                                  // 57
                                                                                                                       //
  if (!isAdmin() && !tokenIsValid(token) && !tokenIsSetupSession(token)) {                                             // 58
    throw new Meteor.Error(403, "User must be admin or provide a valid token");                                        // 59
  }                                                                                                                    // 60
}                                                                                                                      // 61
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"backend.js":["babel-runtime/helpers/classCallCheck","/imports/server/constants.js","/imports/server/async-helpers.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/backend.js                                                                                           //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  SandstormBackend: function () {                                                                                      // 1
    return SandstormBackend;                                                                                           // 1
  },                                                                                                                   // 1
  shouldRestartGrain: function () {                                                                                    // 1
    return shouldRestartGrain;                                                                                         // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var SANDSTORM_ALTHOME = void 0;                                                                                        // 1
module.import("/imports/server/constants.js", {                                                                        // 1
  "SANDSTORM_ALTHOME": function (v) {                                                                                  // 1
    SANDSTORM_ALTHOME = v;                                                                                             // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var inMeteor = void 0,                                                                                                 // 1
    promiseToFuture = void 0,                                                                                          // 1
    waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  "promiseToFuture": function (v) {                                                                                    // 1
    promiseToFuture = v;                                                                                               // 1
  },                                                                                                                   // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 20
                                                                                                                       //
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 21
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 22
                                                                                                                       //
var storageUsageUnimplemented = false;                                                                                 // 24
                                                                                                                       //
function generateSessionId(grainId, userId, packageSalt, clientSalt) {                                                 // 26
  var sessionParts = [grainId, clientSalt];                                                                            // 27
                                                                                                                       //
  if (userId) {                                                                                                        // 28
    sessionParts.push(userId);                                                                                         // 29
  }                                                                                                                    // 30
                                                                                                                       //
  if (packageSalt) {                                                                                                   // 32
    sessionParts.push(packageSalt);                                                                                    // 33
  }                                                                                                                    // 34
                                                                                                                       //
  var sessionInput = sessionParts.join(":");                                                                           // 36
  return Crypto.createHash("sha256").update(sessionInput).digest("hex");                                               // 37
}                                                                                                                      // 38
                                                                                                                       //
var shouldRestartGrain = function (error, retryCount) {                                                                // 40
  // Given an error thrown by an RPC call to a grain, return whether or not it makes sense to try                      // 41
  // to restart the grain and retry. `retryCount` is the number of times that the request has                          // 42
  // already gone through this cycle (should be zero for the first call).                                              // 43
  return error.kjType === "disconnected" && retryCount < 1;                                                            // 44
};                                                                                                                     // 45
                                                                                                                       //
var SandstormBackend = function () {                                                                                   //
  function SandstormBackend(db, backendCap) {                                                                          // 48
    (0, _classCallCheck3.default)(this, SandstormBackend);                                                             // 48
    this._db = db;                                                                                                     // 49
    this._backendCap = backendCap;                                                                                     // 50
  }                                                                                                                    // 51
                                                                                                                       //
  SandstormBackend.prototype.cap = function () {                                                                       //
    function cap() {                                                                                                   //
      return this._backendCap;                                                                                         // 54
    }                                                                                                                  // 55
                                                                                                                       //
    return cap;                                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.deleteUser = function () {                                                                //
    function deleteUser(userId) {                                                                                      //
      return waitPromise(this._backendCap.deleteUser(userId));                                                         // 58
    }                                                                                                                  // 59
                                                                                                                       //
    return deleteUser;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.shutdownGrain = function () {                                                             //
    function shutdownGrain(grainId, ownerId, keepSessions) {                                                           //
      if (!keepSessions) {                                                                                             // 62
        Sessions.remove({                                                                                              // 63
          grainId: grainId                                                                                             // 63
        });                                                                                                            // 63
      }                                                                                                                // 64
                                                                                                                       //
      var grain = this._backendCap.getGrain(ownerId, grainId).supervisor;                                              // 66
                                                                                                                       //
      return grain.shutdown().then(function () {                                                                       // 67
        grain.close();                                                                                                 // 68
        throw new Error("expected shutdown() to throw disconnected");                                                  // 69
      }, function (err) {                                                                                              // 70
        grain.close();                                                                                                 // 72
                                                                                                                       //
        if (err.kjType !== "disconnected") {                                                                           // 73
          throw err;                                                                                                   // 74
        }                                                                                                              // 75
      });                                                                                                              // 76
    }                                                                                                                  // 77
                                                                                                                       //
    return shutdownGrain;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.deleteGrain = function () {                                                               //
    function deleteGrain(grainId, ownerId) {                                                                           //
      // We leave it up to the caller if they want to actually wait, but some don't so we report                       // 80
      // exceptions.                                                                                                   // 81
      return this._backendCap.deleteGrain(ownerId, grainId).catch(function (err) {                                     // 82
        console.error("problem deleting grain " + grainId + ":", err.message);                                         // 83
        throw err;                                                                                                     // 84
      });                                                                                                              // 85
    }                                                                                                                  // 86
                                                                                                                       //
    return deleteGrain;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.maybeRetryUseGrain = function () {                                                        //
    function maybeRetryUseGrain(grainId, cb, retryCount, err) {                                                        //
      var _this = this;                                                                                                // 88
                                                                                                                       //
      if (shouldRestartGrain(err, retryCount)) {                                                                       // 89
        return inMeteor(function () {                                                                                  // 90
          return cb(_this.continueGrain(grainId).supervisor).catch(_this.maybeRetryUseGrain.bind(_this, grainId, cb, retryCount + 1));
        });                                                                                                            // 93
      } else {                                                                                                         // 94
        throw err;                                                                                                     // 95
      }                                                                                                                // 96
    }                                                                                                                  // 97
                                                                                                                       //
    return maybeRetryUseGrain;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.useGrain = function () {                                                                  //
    function useGrain(grainId, cb) {                                                                                   //
      var _this2 = this;                                                                                               // 99
                                                                                                                       //
      // This will open a grain for you, handling restarts if needed, and call the passed function with                // 100
      // the supervisor capability as the only parameter. The callback must return a promise that used                 // 101
      // the supervisor, so that we can check if a disconnect error occurred, and retry if possible.                   // 102
      // This function returns the same promise that your callback returns.                                            // 103
      //                                                                                                               // 104
      // This function is NOT expected to be run in a meteor context.                                                  // 105
      return inMeteor(function () {                                                                                    // 106
        return cb(_this2.continueGrain(grainId).supervisor).catch(_this2.maybeRetryUseGrain.bind(_this2, grainId, cb, 0));
      });                                                                                                              // 109
    }                                                                                                                  // 110
                                                                                                                       //
    return useGrain;                                                                                                   //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.continueGrain = function () {                                                             //
    function continueGrain(grainId) {                                                                                  //
      var grain = Grains.findOne(grainId);                                                                             // 113
                                                                                                                       //
      if (!grain) {                                                                                                    // 114
        throw new Meteor.Error(404, "Grain Not Found", "Grain ID: " + grainId);                                        // 115
      } // If a DevPackage with the same app ID is currently active, we let it override the installed                  // 116
      // package, so that the grain runs using the dev app.                                                            // 119
                                                                                                                       //
                                                                                                                       //
      var devPackage = DevPackages.findOne({                                                                           // 120
        appId: grain.appId                                                                                             // 120
      });                                                                                                              // 120
      var isDev = void 0;                                                                                              // 121
      var mountProc = void 0;                                                                                          // 122
      var pkg = void 0;                                                                                                // 123
                                                                                                                       //
      if (devPackage) {                                                                                                // 124
        isDev = true;                                                                                                  // 125
        pkg = devPackage;                                                                                              // 126
        mountProc = pkg.mountProc;                                                                                     // 127
      } else {                                                                                                         // 128
        pkg = Packages.findOne(grain.packageId);                                                                       // 129
                                                                                                                       //
        if (!pkg) {                                                                                                    // 130
          throw new Meteor.Error(500, "Grain's package not installed", "Package ID: " + grain.packageId);              // 131
        }                                                                                                              // 133
      }                                                                                                                // 134
                                                                                                                       //
      var manifest = pkg.manifest;                                                                                     // 136
      var packageId = pkg._id;                                                                                         // 137
                                                                                                                       //
      if (!("continueCommand" in manifest)) {                                                                          // 139
        throw new Meteor.Error(500, "Package manifest defines no continueCommand.", "Package ID: " + packageId);       // 140
      }                                                                                                                // 142
                                                                                                                       //
      var result = this.startGrainInternal(packageId, grainId, grain.userId, manifest.continueCommand, false, isDev, mountProc);
      result.packageSalt = isDev ? pkg._id : grain.packageSalt;                                                        // 146
      return result;                                                                                                   // 147
    }                                                                                                                  // 148
                                                                                                                       //
    return continueGrain;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.startGrainInternal = function () {                                                        //
    function startGrainInternal(packageId, grainId, ownerId, command, isNew, isDev, mountProc) {                       //
      // Starts the grain supervisor.  Must be executed in a Meteor context.  Blocks until grain is                    // 151
      // started. Returns a promise for an object containing two fields: `owner` (the ID of the owning                 // 152
      // user) and `supervisor` (the supervisor capability).                                                           // 153
      if (isUserExcessivelyOverQuota(Meteor.users.findOne(ownerId))) {                                                 // 155
        throw new Meteor.Error(402, "Cannot start grain because owner's storage is exhausted.\n" + "Please ask them to upgrade.");
      } // Ugly: Stay backwards-compatible with old manifests that had "executablePath" and "args" rather              // 159
      //   than just "argv".                                                                                           // 162
                                                                                                                       //
                                                                                                                       //
      if ("args" in command) {                                                                                         // 163
        if (!("argv" in command)) {                                                                                    // 164
          command.argv = command.args;                                                                                 // 165
        }                                                                                                              // 166
                                                                                                                       //
        delete command.args;                                                                                           // 168
      }                                                                                                                // 169
                                                                                                                       //
      if ("executablePath" in command) {                                                                               // 171
        if (!("deprecatedExecutablePath" in command)) {                                                                // 172
          command.deprecatedExecutablePath = command.executablePath;                                                   // 173
        }                                                                                                              // 174
                                                                                                                       //
        delete command.executablePath;                                                                                 // 176
      }                                                                                                                // 177
                                                                                                                       //
      return this._backendCap.startGrain(ownerId, grainId, packageId, command, isNew, isDev, mountProc);               // 179
    }                                                                                                                  // 180
                                                                                                                       //
    return startGrainInternal;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.updateLastActive = function () {                                                          //
    function updateLastActive(grainId, userId, identityId) {                                                           //
      // Update the lastActive date on the grain, any relevant API tokens, and the user,                               // 183
      // and also update the user's storage usage.                                                                     // 184
      var storagePromise = undefined;                                                                                  // 186
      var ownerId = undefined;                                                                                         // 187
                                                                                                                       //
      if (this._db.isQuotaEnabled() && !storageUsageUnimplemented) {                                                   // 188
        ownerId = Grains.findOne(grainId).userId;                                                                      // 189
        storagePromise = this._backendCap.getUserStorageUsage(ownerId);                                                // 190
      }                                                                                                                // 191
                                                                                                                       //
      var now = new Date();                                                                                            // 193
                                                                                                                       //
      if (Grains.update(grainId, {                                                                                     // 194
        $set: {                                                                                                        // 194
          lastUsed: now                                                                                                // 194
        }                                                                                                              // 194
      }) === 0) {                                                                                                      // 194
        // Grain must have been deleted. Ignore.                                                                       // 195
        return;                                                                                                        // 196
      }                                                                                                                // 197
                                                                                                                       //
      if (userId) {                                                                                                    // 199
        Meteor.users.update(userId, {                                                                                  // 200
          $set: {                                                                                                      // 200
            lastActive: now                                                                                            // 200
          }                                                                                                            // 200
        });                                                                                                            // 200
      }                                                                                                                // 201
                                                                                                                       //
      if (identityId) {                                                                                                // 203
        Meteor.users.update({                                                                                          // 204
          _id: identityId                                                                                              // 204
        }, {                                                                                                           // 204
          $set: {                                                                                                      // 204
            lastActive: now                                                                                            // 204
          }                                                                                                            // 204
        }); // Update any API tokens that match this user/grain pairing as well                                        // 204
                                                                                                                       //
        ApiTokens.update({                                                                                             // 206
          grainId: grainId,                                                                                            // 206
          "owner.user.identityId": identityId                                                                          // 206
        }, {                                                                                                           // 206
          $set: {                                                                                                      // 207
            lastUsed: now                                                                                              // 207
          }                                                                                                            // 207
        }, {                                                                                                           // 207
          multi: true                                                                                                  // 208
        });                                                                                                            // 208
      }                                                                                                                // 209
                                                                                                                       //
      if (storagePromise) {                                                                                            // 211
        try {                                                                                                          // 212
          var size = parseInt(waitPromise(storagePromise).size);                                                       // 213
          Meteor.users.update(ownerId, {                                                                               // 214
            $set: {                                                                                                    // 214
              storageUsage: size                                                                                       // 214
            }                                                                                                          // 214
          }); // TODO(security): Consider actively killing grains if the user is excessively over quota?               // 214
          //   Otherwise a constantly-active grain could consume arbitrary space without being stopped.                // 216
        } catch (err) {                                                                                                // 217
          if (err.kjType === "unimplemented") {                                                                        // 218
            storageUsageUnimplemented = true;                                                                          // 219
          } else {                                                                                                     // 220
            console.error("error getting user storage usage:", err.stack);                                             // 221
          }                                                                                                            // 222
        }                                                                                                              // 223
      }                                                                                                                // 224
    }                                                                                                                  // 225
                                                                                                                       //
    return updateLastActive;                                                                                           //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormBackend.prototype.openSessionInternal = function () {                                                       //
    function openSessionInternal(grainId, userId, identityId, title, apiToken, cachedSalt, sessionFields) {            //
      // Start the grain if it is not running. This is an optimization: if we didn't start it here,                    // 228
      // it would start on the first request to the session host, but we'd like to get started before                  // 229
      // the round trip.                                                                                               // 230
      var _continueGrain = this.continueGrain(grainId),                                                                // 227
          supervisor = _continueGrain.supervisor,                                                                      // 227
          packageSalt = _continueGrain.packageSalt;                                                                    // 227
                                                                                                                       //
      this.updateLastActive(grainId, userId, identityId);                                                              // 233
      cachedSalt = cachedSalt || Random.id(22);                                                                        // 235
      var sessionId = generateSessionId(grainId, userId, packageSalt, cachedSalt);                                     // 236
      var session = Sessions.findOne({                                                                                 // 237
        _id: sessionId                                                                                                 // 237
      });                                                                                                              // 237
                                                                                                                       //
      if (session) {                                                                                                   // 238
        // TODO(someday): also do some more checks for anonymous sessions (sessions without a userId).                 // 239
        if (session.identityId && session.identityId !== identityId || session.grainId !== grainId) {                  // 240
          var e = new Meteor.Error(500, "Duplicate SessionId");                                                        // 242
          console.error(e);                                                                                            // 243
          throw e;                                                                                                     // 244
        } else {                                                                                                       // 245
          return {                                                                                                     // 246
            supervisor: supervisor,                                                                                    // 247
            methodResult: {                                                                                            // 248
              sessionId: session._id,                                                                                  // 249
              title: title,                                                                                            // 250
              grainId: grainId,                                                                                        // 251
              hostId: session.hostId,                                                                                  // 252
              tabId: session.tabId,                                                                                    // 253
              salt: cachedSalt                                                                                         // 254
            }                                                                                                          // 248
          };                                                                                                           // 246
        }                                                                                                              // 257
      } // TODO(someday): Allow caller to specify the parent session from which to inherit the tab ID, or              // 258
      //   something.                                                                                                  // 261
                                                                                                                       //
                                                                                                                       //
      session = {                                                                                                      // 263
        _id: sessionId,                                                                                                // 264
        grainId: grainId,                                                                                              // 265
        hostId: Crypto.createHash("sha256").update(sessionId).digest("hex").slice(0, 32),                              // 266
        tabId: Crypto.createHash("sha256").update("tab:").update(sessionId).digest("hex").slice(0, 32),                // 267
        timestamp: new Date().getTime(),                                                                               // 268
        hasLoaded: false                                                                                               // 269
      };                                                                                                               // 263
                                                                                                                       //
      if (userId) {                                                                                                    // 272
        session.userId = userId;                                                                                       // 273
      }                                                                                                                // 274
                                                                                                                       //
      if (identityId) {                                                                                                // 276
        session.identityId = identityId;                                                                               // 277
      }                                                                                                                // 278
                                                                                                                       //
      if (apiToken) {                                                                                                  // 280
        session.hashedToken = apiToken._id;                                                                            // 281
      }                                                                                                                // 282
                                                                                                                       //
      if (sessionFields) {                                                                                             // 284
        _.extend(session, sessionFields);                                                                              // 285
      }                                                                                                                // 286
                                                                                                                       //
      Sessions.insert(session);                                                                                        // 288
      return {                                                                                                         // 290
        supervisor: supervisor,                                                                                        // 291
        methodResult: {                                                                                                // 292
          sessionId: session._id,                                                                                      // 293
          title: title,                                                                                                // 294
          grainId: grainId,                                                                                            // 295
          hostId: session.hostId,                                                                                      // 296
          tabId: session.tabId,                                                                                        // 297
          salt: cachedSalt                                                                                             // 298
        }                                                                                                              // 292
      };                                                                                                               // 290
    }                                                                                                                  // 301
                                                                                                                       //
    return openSessionInternal;                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  return SandstormBackend;                                                                                             //
}();                                                                                                                   //
                                                                                                                       //
;                                                                                                                      // 302
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"constants.js":["meteor/meteor","url",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/constants.js                                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  SANDSTORM_ALTHOME: function () {                                                                                     // 1
    return SANDSTORM_ALTHOME;                                                                                          // 1
  },                                                                                                                   // 1
  SANDSTORM_LOGDIR: function () {                                                                                      // 1
    return SANDSTORM_LOGDIR;                                                                                           // 1
  },                                                                                                                   // 1
  SANDSTORM_VARDIR: function () {                                                                                      // 1
    return SANDSTORM_VARDIR;                                                                                           // 1
  },                                                                                                                   // 1
  staticAssetHost: function () {                                                                                       // 1
    return staticAssetHost;                                                                                            // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Url = void 0;                                                                                                      // 1
module.import("url", {                                                                                                 // 1
  "default": function (v) {                                                                                            // 1
    Url = v;                                                                                                           // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;                                                       // 22
var SANDSTORM_LOGDIR = (SANDSTORM_ALTHOME || "") + "/var/log";                                                         // 23
var SANDSTORM_VARDIR = (SANDSTORM_ALTHOME || "") + "/var/sandstorm"; // TODO: dedupe this and the equivalent copy in sandstorm-db
                                                                                                                       //
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                                       // 27
var staticAssetHost = Url.parse(process.env.ROOT_URL).protocol + "//" + wildcardHost[0] + "static" + wildcardHost[1];  // 29
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"desktop-notifications.js":["meteor/check",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/desktop-notifications.js                                                                             //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  createAppActivityDesktopNotification: function () {                                                                  // 1
    return createAppActivityDesktopNotification;                                                                       // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Match = void 0,                                                                                                    // 1
    check = void 0;                                                                                                    // 1
module.import("meteor/check", {                                                                                        // 1
  "Match": function (v) {                                                                                              // 1
    Match = v;                                                                                                         // 1
  },                                                                                                                   // 1
  "check": function (v) {                                                                                              // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var createAppActivityDesktopNotification = function (options) {                                                        // 19
  check(options, {                                                                                                     // 20
    userId: String,                                                                                                    // 21
    notificationId: String,                                                                                            // 22
    appActivity: {                                                                                                     // 23
      user: Match.OneOf(undefined, {                                                                                   // 24
        identityId: String,                                                                                            // 27
        name: String,                                                                                                  // 28
        avatarUrl: String                                                                                              // 29
      }, {                                                                                                             // 26
        anonymous: true                                                                                                // 31
      }),                                                                                                              // 31
      grainId: String,                                                                                                 // 33
      path: String,                                                                                                    // 34
      body: Match.ObjectIncluding({                                                                                    // 35
        defaultText: String                                                                                            // 36
      }),                                                                                                              // 35
      actionText: Match.ObjectIncluding({                                                                              // 38
        defaultText: String                                                                                            // 39
      })                                                                                                               // 38
    }                                                                                                                  // 23
  });                                                                                                                  // 20
  globalDb.collections.desktopNotifications.insert({                                                                   // 44
    userId: options.userId,                                                                                            // 45
    notificationId: options.notificationId,                                                                            // 46
    creationDate: new Date(),                                                                                          // 47
    appActivity: options.appActivity                                                                                   // 48
  });                                                                                                                  // 44
};                                                                                                                     // 50
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"email.js":["babel-runtime/helpers/typeof","meteor/meteor","nodemailer","nodemailer-smtp-pool",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/email.js                                                                                             //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _typeof2 = require("babel-runtime/helpers/typeof");                                                                //
                                                                                                                       //
var _typeof3 = _interopRequireDefault(_typeof2);                                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  send: function () {                                                                                                  // 1
    return send;                                                                                                       // 1
  },                                                                                                                   // 1
  rawSend: function () {                                                                                               // 1
    return rawSend;                                                                                                    // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var nodemailer = void 0;                                                                                               // 1
module.import("nodemailer", {                                                                                          // 1
  "default": function (v) {                                                                                            // 1
    nodemailer = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var smtpPool = void 0;                                                                                                 // 1
module.import("nodemailer-smtp-pool", {                                                                                // 1
  "default": function (v) {                                                                                            // 1
    smtpPool = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 5
                                                                                                                       //
var getSmtpConfig = function () {                                                                                      // 7
  var config = Settings.findOne({                                                                                      // 8
    _id: "smtpConfig"                                                                                                  // 8
  });                                                                                                                  // 8
  return config && config.value;                                                                                       // 9
};                                                                                                                     // 10
                                                                                                                       //
var makePool = function (mailConfig) {                                                                                 // 12
  if (!mailConfig.hostname) {                                                                                          // 13
    throw new Error("This Sandstorm server has not been configured to send email.");                                   // 14
  }                                                                                                                    // 15
                                                                                                                       //
  var auth = false;                                                                                                    // 17
                                                                                                                       //
  if (mailConfig.auth && (mailConfig.auth.user || mailConfig.auth.pass)) {                                             // 18
    auth = mailConfig.auth;                                                                                            // 19
  }                                                                                                                    // 20
                                                                                                                       //
  var secure = mailConfig.port === 465;                                                                                // 22
  var tlsOptions = {                                                                                                   // 23
    // Previously, node 0.10 did not attempt to validate certificates received when connecting                         // 24
    // with STARTTLS, so to avoid regressing we need to preserve that behavior here for now.                           // 25
    rejectUnauthorized: false                                                                                          // 26
  };                                                                                                                   // 23
  var pool = nodemailer.createTransport(smtpPool({                                                                     // 29
    host: mailConfig.hostname,                                                                                         // 30
    port: mailConfig.port,                                                                                             // 31
    secure: secure,                                                                                                    // 32
    tls: tlsOptions,                                                                                                   // 33
    auth: auth                                                                                                         // 34
  }));                                                                                                                 // 29
  pool._futureWrappedSendMail = _.bind(Future.wrap(pool.sendMail), pool);                                              // 38
  return pool;                                                                                                         // 39
}; // We construct the SMTP pool at the first call to Email.send, so that                                              // 40
// other code like migrations can modify the SMTP configuration.                                                       // 43
                                                                                                                       //
                                                                                                                       //
var pool = void 0;                                                                                                     // 44
var configured = false;                                                                                                // 45
Meteor.startup(function () {                                                                                           // 47
  Settings.find({                                                                                                      // 48
    _id: "smtpConfig"                                                                                                  // 48
  }).observeChanges({                                                                                                  // 48
    removed: function () {                                                                                             // 49
      configured = false;                                                                                              // 50
    },                                                                                                                 // 51
    changed: function () {                                                                                             // 53
      configured = false;                                                                                              // 54
    },                                                                                                                 // 55
    added: function () {                                                                                               // 57
      configured = false;                                                                                              // 58
    }                                                                                                                  // 59
  });                                                                                                                  // 48
});                                                                                                                    // 61
                                                                                                                       //
var getPool = function (smtpConfig) {                                                                                  // 63
  if (smtpConfig) {                                                                                                    // 64
    return makePool(smtpConfig);                                                                                       // 65
  } else if (!configured) {                                                                                            // 66
    configured = true;                                                                                                 // 67
    var config = getSmtpConfig();                                                                                      // 68
                                                                                                                       //
    if (config) {                                                                                                      // 69
      pool = makePool(config);                                                                                         // 70
    }                                                                                                                  // 71
  }                                                                                                                    // 72
                                                                                                                       //
  return pool;                                                                                                         // 74
};                                                                                                                     // 75
                                                                                                                       //
var smtpSend = function (pool, mailOptions) {                                                                          // 77
  pool._futureWrappedSendMail(mailOptions).wait();                                                                     // 78
}; // From http://emailregex.com/, which claims this is the W3C standard for the HTML input element,                   // 79
// although their link is broken and I can find no evidence that this is a standard. The page                          // 82
// lists several regexes, ostensibly in syntaxes intended for different programming languages,                         // 83
// but each regex is in fact substantially different for no apparent reason.                                           // 84
//                                                                                                                     // 85
// The most important thing here is that we disallow separators that might allow a user to confuse                     // 86
// nodemailer into thinking the address is a list. Unfortunately, nodemailer will happily separate                     // 87
// strings into lists splitting on all kinds of separator characters, such as commas, semicolons,                      // 88
// etc. This regex should accomplish that both by disallowing the separators, and by disallowing                       // 89
// multiple @ signs. The rest is for show.                                                                             // 90
                                                                                                                       //
                                                                                                                       //
var EMAIL_REGEX = /^[a-zA-Z0-9.!#$%&’*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*$/;                              // 91
                                                                                                                       //
function validateEmail(email) {                                                                                        // 93
  if (email instanceof Array) {                                                                                        // 94
    email.forEach(validateEmail);                                                                                      // 95
  } else if ((typeof email === "undefined" ? "undefined" : (0, _typeof3.default)(email)) === "object" && "address" in email) {
    validateEmail(email.address);                                                                                      // 97
  } else if (email) {                                                                                                  // 98
    check(email, String);                                                                                              // 99
                                                                                                                       //
    if (!email.match(EMAIL_REGEX)) {                                                                                   // 101
      throw new Meteor.Error(400, "invalid e-mail address");                                                           // 102
    }                                                                                                                  // 103
  }                                                                                                                    // 104
}                                                                                                                      // 105
                                                                                                                       //
var rawSend = function (mailOptions, smtpConfig) {                                                                     // 107
  // Sends an email mailOptions object structured as described in                                                      // 108
  // https://github.com/nodemailer/mailcomposer#e-mail-message-fields                                                  // 109
  // across the transport described by smtpConfig.                                                                     // 110
  // For fields that are supposed to be lists of addresses, if only a single string is provided,                       // 112
  // wrap it in an array. This prevents nodemailer from interpreting the address as a                                  // 113
  // comma-separated list.                                                                                             // 114
  ["from", "to", "cc", "bcc", "replyTo"].forEach(function (field) {                                                    // 115
    validateEmail(mailOptions[field]);                                                                                 // 116
  });                                                                                                                  // 117
  var pool = getPool(smtpConfig);                                                                                      // 119
                                                                                                                       //
  if (pool) {                                                                                                          // 120
    smtpSend(pool, mailOptions);                                                                                       // 121
  } else {                                                                                                             // 122
    throw new Error("SMTP pool is misconfigured.");                                                                    // 123
  }                                                                                                                    // 124
}; // Old comment below                                                                                                // 125
/**                                                                                                                    // 128
 * Send an email.                                                                                                      //
 *                                                                                                                     //
 * Connects to the mail server configured via the MAIL_URL environment                                                 //
 * variable. If unset, prints formatted message to stdout. The "from" option                                           //
 * is required, and at least one of "to", "cc", and "bcc" must be provided;                                            //
 * all other options are optional.                                                                                     //
 *                                                                                                                     //
 * @param options                                                                                                      //
 * @param options.from {String} RFC5322 "From:" address                                                                //
 * @param options.to {String|String[]} RFC5322 "To:" address[es]                                                       //
 * @param options.cc {String|String[]} RFC5322 "Cc:" address[es]                                                       //
 * @param options.bcc {String|String[]} RFC5322 "Bcc:" address[es]                                                     //
 * @param options.replyTo {String|String[]} RFC5322 "Reply-To:" address[es]                                            //
 * @param options.subject {String} RFC5322 "Subject:" line                                                             //
 * @param options.text {String} RFC5322 mail body (plain text)                                                         //
 * @param options.html {String} RFC5322 mail body (HTML)                                                               //
 * @param options.headers {Object} custom RFC5322 headers (dictionary)                                                 //
 */ // New API doc comment below                                                                                       //
/**                                                                                                                    // 149
 * @summary Send an email. Throws an `Error` on failure to contact mail server                                         //
 * or if mail server returns an error. All fields should match                                                         //
 * [RFC5322](http://tools.ietf.org/html/rfc5322) specification.                                                        //
 * @locus Server                                                                                                       //
 * @param {Object} options                                                                                             //
 * @param {String} options.from "From:" address (required)                                                             //
 * @param {String|String[]} options.to,cc,bcc,replyTo                                                                  //
 *   "To:", "Cc:", "Bcc:", and "Reply-To:" addresses                                                                   //
 * @param {String} [options.subject]  "Subject:" line                                                                  //
 * @param {String} [options.text|html] Mail body (in plain text or HTML)                                               //
 * @param {Object} [options.headers] Dictionary of custom headers                                                      //
 * @param {Object} [options.smtpConfig] SMTP server to use. Otherwise defaults to configured one.                      //
 * @param {String} [options.smtpConfig.hostname] SMTP server hostname.                                                 //
 * @param {Number} [options.smtpConfig.port] SMTP server port.                                                         //
 * @param {Object} [options.smtpConfig.auth] SMTP server authentication tokens.  Optional.                             //
 * @param {String} [options.smtpConfig.auth.user] Username of user to log in to SMTP server as.  Optional.             //
 * @param {String} [options.smtpConfig.auth.pass] Password of user to log in to SMTP server as.  Optional.             //
 * @param {Object} [options.attachments] Attachments. See:                                                             //
 *   https://github.com/nodemailer/mailcomposer/tree/v0.1.15#add-attachments                                           //
 * @param {String} [options.envelopeFrom] Envelope sender.                                                             //
 */                                                                                                                    //
                                                                                                                       //
var send = function (options) {                                                                                        // 171
  // Unpack options                                                                                                    // 172
  var from = options.from,                                                                                             // 171
      to = options.to,                                                                                                 // 171
      cc = options.cc,                                                                                                 // 171
      bcc = options.bcc,                                                                                               // 171
      replyTo = options.replyTo,                                                                                       // 171
      subject = options.subject,                                                                                       // 171
      text = options.text,                                                                                             // 171
      html = options.html,                                                                                             // 171
      envelopeFrom = options.envelopeFrom,                                                                             // 171
      headers = options.headers,                                                                                       // 171
      attachments = options.attachments,                                                                               // 171
      smtpConfig = options.smtpConfig;                                                                                 // 171
  var opts = {                                                                                                         // 188
    from: from,                                                                                                        // 189
    to: to,                                                                                                            // 190
    cc: cc,                                                                                                            // 191
    bcc: bcc,                                                                                                          // 192
    replyTo: replyTo,                                                                                                  // 193
    subject: subject,                                                                                                  // 194
    text: text,                                                                                                        // 195
    html: html,                                                                                                        // 196
    headers: headers,                                                                                                  // 197
    attachments: attachments                                                                                           // 198
  };                                                                                                                   // 188
                                                                                                                       //
  if (envelopeFrom) {                                                                                                  // 201
    opts.envelope = {                                                                                                  // 202
      from: envelopeFrom,                                                                                              // 203
      to: to,                                                                                                          // 204
      cc: cc,                                                                                                          // 205
      bcc: bcc                                                                                                         // 206
    };                                                                                                                 // 202
  }                                                                                                                    // 208
                                                                                                                       //
  rawSend(opts, smtpConfig);                                                                                           // 210
};                                                                                                                     // 211
                                                                                                                       //
// TODO(cleanup): Remove this once BlackrockPayments code finds a better way to import it.                             // 215
global.SandstormEmail = {                                                                                              // 216
  send: send                                                                                                           // 216
};                                                                                                                     // 216
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"frontend-ref.js":["babel-runtime/helpers/classCallCheck","./persistent.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/frontend-ref.js                                                                                      //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  FrontendRefRegistry: function () {                                                                                   // 1
    return FrontendRefRegistry;                                                                                        // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var checkRequirements = void 0;                                                                                        // 1
module.import("./persistent.js", {                                                                                     // 1
  "checkRequirements": function (v) {                                                                                  // 1
    checkRequirements = v;                                                                                             // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var FrontendRefRegistry = function () {                                                                                //
  function FrontendRefRegistry() {                                                                                     // 20
    (0, _classCallCheck3.default)(this, FrontendRefRegistry);                                                          // 20
    this._frontendRefHandlers = {};                                                                                    // 21
    this._typeIdHandlers = {};                                                                                         // 22
  }                                                                                                                    // 23
                                                                                                                       //
  FrontendRefRegistry.prototype.create = function () {                                                                 //
    function create(db, frontendRef, requirements) {                                                                   //
      // Construct a fresh capability based on a `frontendRef` value.                                                  // 26
      checkRequirements(requirements);                                                                                 // 28
      var saveTemplate = {                                                                                             // 30
        frontendRef: frontendRef                                                                                       // 30
      };                                                                                                               // 30
                                                                                                                       //
      if (requirements) {                                                                                              // 31
        saveTemplate.requirements = requirements;                                                                      // 32
      }                                                                                                                // 33
                                                                                                                       //
      return this.restore(db, saveTemplate, frontendRef);                                                              // 35
    }                                                                                                                  // 36
                                                                                                                       //
    return create;                                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  FrontendRefRegistry.prototype.restore = function () {                                                                //
    function restore(db, saveTemplate, frontendRef) {                                                                  //
      // Restores a frontendRef capability using the appropriate registered handler.                                   // 39
      var keys = Object.keys(frontendRef);                                                                             // 41
                                                                                                                       //
      if (keys.length != 1) {                                                                                          // 42
        throw new Error("invalid frontendRef: " + JSON.stringify(frontendRef));                                        // 43
      }                                                                                                                // 44
                                                                                                                       //
      var key = keys[0];                                                                                               // 46
      var handler = this._frontendRefHandlers[key];                                                                    // 47
                                                                                                                       //
      if (!handler) {                                                                                                  // 48
        throw new Error("invalid frontendRef: " + JSON.stringify(frontendRef));                                        // 49
      }                                                                                                                // 50
                                                                                                                       //
      return handler.restore(db, saveTemplate, frontendRef[key]);                                                      // 52
    }                                                                                                                  // 53
                                                                                                                       //
    return restore;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  FrontendRefRegistry.prototype.query = function () {                                                                  //
    function query(db, userAccountId, tag) {                                                                           //
      // Performs a powerbox query using the appropriate registered handler.                                           // 56
      var handler = this._typeIdHandlers[tag.id];                                                                      // 58
                                                                                                                       //
      if (handler) {                                                                                                   // 59
        return handler.query(db, userAccountId, tag.value);                                                            // 60
      } else {                                                                                                         // 61
        return []; // no matches                                                                                       // 62
      }                                                                                                                // 63
    }                                                                                                                  // 64
                                                                                                                       //
    return query;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  FrontendRefRegistry.prototype.validate = function () {                                                               //
    function validate(db, session, frontendRefRequest) {                                                               //
      // Validates a powerbox request on the given `session` (a record from the Sessions table)                        // 67
      // requesting the creation of the given frontendRef type. `frontendRefRequest` is expected to                    // 68
      // come directly from the client; calling validate() verifies that it is well-formed and                         // 69
      // carries out any type-specific server-side work needed to construct the desired frontendRef.                   // 70
      // `validate()` also returns an array of MembraneRequirements that shall apply to the                            // 71
      // capability. These requirements will not yet have been checked, so the caller should check                     // 72
      // them before proceeding.                                                                                       // 73
      var keys = Object.keys(frontendRefRequest);                                                                      // 75
                                                                                                                       //
      if (keys.length != 1) {                                                                                          // 76
        throw new Error("invalid frontendRefRequest: " + JSON.stringify(frontendRefRequest));                          // 77
      }                                                                                                                // 78
                                                                                                                       //
      var key = keys[0];                                                                                               // 80
      var handler = this._frontendRefHandlers[key];                                                                    // 81
                                                                                                                       //
      if (!handler) {                                                                                                  // 82
        throw new Error("invalid frontendRefRequest: " + JSON.stringify(frontendRefRequest));                          // 83
      }                                                                                                                // 84
                                                                                                                       //
      if (!handler.validate) {                                                                                         // 86
        throw new Error("frontendRef type cannot be created via powerbox: " + JSON.stringify(frontendRefRequest));     // 87
      }                                                                                                                // 89
                                                                                                                       //
      var _handler$validate = handler.validate(db, session, frontendRefRequest[key]),                                  // 66
          descriptor = _handler$validate.descriptor,                                                                   // 66
          requirements = _handler$validate.requirements,                                                               // 66
          frontendRef = _handler$validate.frontendRef;                                                                 // 66
                                                                                                                       //
      var result = {                                                                                                   // 94
        descriptor: descriptor,                                                                                        // 94
        requirements: requirements,                                                                                    // 94
        frontendRef: {}                                                                                                // 94
      };                                                                                                               // 94
      result.frontendRef[key] = frontendRef;                                                                           // 95
      return result;                                                                                                   // 96
    }                                                                                                                  // 97
                                                                                                                       //
    return validate;                                                                                                   //
  }();                                                                                                                 //
                                                                                                                       //
  FrontendRefRegistry.prototype.register = function () {                                                               //
    function register(object) {                                                                                        //
      // Register callbacks related to a particular frontendRef type. The object has the fields:                       // 100
      //   `frontendRefField`: Name of the field of `ApiTokens.frontendRef` that is filled in for                      // 101
      //       this type. Only needed if `create` and/or `validate` handlers are defined.                              // 102
      //   `typeId`: Type ID of powerbox tags handled by the `query` callback. Stringified decimal                     // 103
      //       64-bit integer. Only needed if `query` is defined.                                                      // 104
      //   `restore`: Callback to construct a capability of this type when restoring a saved                           // 105
      //       capability. Has signature `(db, saveTemplate, value) -> capability`, where:                             // 106
      //     `value`: The value of the single field of `ApiTokens.frontendRef` for this capability.                    // 107
      //     `saveTemplate`: The token template to pass to the PersistentImpl constructor.                             // 108
      //     `capability` (returned): A Cap'n Proto capability implementing SystemPersistent along                     // 109
      //         with whatever other interfaces are appropriate for the ref type.                                      // 110
      //   `validate`: Callback to validate a powerbox request for a new capability of this type.                      // 111
      //       Has signature `(db, session, request) -> {descriptor, requirements, frontendRef}`,                      // 112
      //       where:                                                                                                  // 113
      //     `request` is type-specific information describing the requested capability. The                           // 114
      //         callback *must* type-check this value, and should throw an exception if it is                         // 115
      //         not valid.                                                                                            // 116
      //     `session` is the record from the Sessions table of the UI session where the powerbox                      // 117
      //         request occurred.                                                                                     // 118
      //     `descriptor` (returned) is the JSON-encoded PowerboxDescriptor for the capability. Note                   // 119
      //         that if the descriptor contains any `tag.value`s, they of course need to be                           // 120
      //         presented as capnp-encoded Buffers.                                                                   // 121
      //     `requirements` (returned) is an array of MembraneRequirements which should apply to the                   // 122
      //         new capability. Note that these requirements will be checked immediately and the                      // 123
      //         powerbox request will fail if they aren't met.                                                        // 124
      //     `frontendRef` (returned) is the value that will be written for the key specified                          // 125
      //         by `frontendRefField` in the single-key object `ApiTokens.frontendRef`.                               // 126
      //    `query`: Callback to populate options for a powerbox request for this type ID. Has                         // 127
      //        signature `(db, userAccountId, tagValue) -> options`, where:                                           // 128
      //      `tagValue`: A Buffer of the Cap'n-Proto-encoded `PowerboxDescriptor.Tag.value`.                          // 129
      //      `options` (returned): An array of objects representing the options that should be                        // 130
      //          offered to the user for this query. See the `powerboxOptions` Meteor publish in                      // 131
      //          powerbox-server.js for a full description of the fields of each option.                              // 132
      if (object.frontendRefField) {                                                                                   // 134
        if (object.frontendRefField in this._frontendRefHandlers) {                                                    // 135
          throw new Error("frontendRef handler already registered: " + object.frontendRefField);                       // 136
        }                                                                                                              // 137
                                                                                                                       //
        this._frontendRefHandlers[object.frontendRefField] = object;                                                   // 139
      }                                                                                                                // 140
                                                                                                                       //
      if (object.typeId) {                                                                                             // 142
        if (object.typeId in this._typeIdHandlers) {                                                                   // 143
          throw new Error("typeId handler already registered: " + object.typeId);                                      // 144
        }                                                                                                              // 145
                                                                                                                       //
        this._typeIdHandlers[object.typeId] = object;                                                                  // 147
      }                                                                                                                // 148
    }                                                                                                                  // 149
                                                                                                                       //
    return register;                                                                                                   //
  }();                                                                                                                 //
                                                                                                                       //
  return FrontendRefRegistry;                                                                                          //
}();                                                                                                                   //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"migrations.js":["meteor/meteor","meteor/underscore","meteor/check","/imports/server/accounts/picture.js","/imports/server/async-helpers.js","/imports/constants.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/migrations.js                                                                                        //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  migrateToLatest: function () {                                                                                       // 1
    return migrateToLatest;                                                                                            // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var _ = void 0;                                                                                                        // 1
                                                                                                                       //
module.import("meteor/underscore", {                                                                                   // 1
  "_": function (v) {                                                                                                  // 1
    _ = v;                                                                                                             // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Match = void 0;                                                                                                    // 1
module.import("meteor/check", {                                                                                        // 1
  "Match": function (v) {                                                                                              // 1
    Match = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var userPictureUrl = void 0,                                                                                           // 1
    fetchPicture = void 0;                                                                                             // 1
module.import("/imports/server/accounts/picture.js", {                                                                 // 1
  "userPictureUrl": function (v) {                                                                                     // 1
    userPictureUrl = v;                                                                                                // 1
  },                                                                                                                   // 1
  "fetchPicture": function (v) {                                                                                       // 1
    fetchPicture = v;                                                                                                  // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var PRIVATE_IPV4_ADDRESSES = void 0,                                                                                   // 1
    PRIVATE_IPV6_ADDRESSES = void 0;                                                                                   // 1
module.import("/imports/constants.js", {                                                                               // 1
  "PRIVATE_IPV4_ADDRESSES": function (v) {                                                                             // 1
    PRIVATE_IPV4_ADDRESSES = v;                                                                                        // 1
  },                                                                                                                   // 1
  "PRIVATE_IPV6_ADDRESSES": function (v) {                                                                             // 1
    PRIVATE_IPV6_ADDRESSES = v;                                                                                        // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 14
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 15
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 16
                                                                                                                       //
var updateLoginStyleToRedirect = function (db, backend) {                                                              // 18
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                           // 19
  ["google", "github"].forEach(function (serviceName) {                                                                // 20
    var config = configurations.findOne({                                                                              // 21
      service: serviceName                                                                                             // 21
    });                                                                                                                // 21
                                                                                                                       //
    if (config && config.loginStyle !== "redirect") {                                                                  // 22
      configurations.update({                                                                                          // 23
        service: serviceName                                                                                           // 23
      }, {                                                                                                             // 23
        $set: {                                                                                                        // 23
          loginStyle: "redirect"                                                                                       // 23
        }                                                                                                              // 23
      });                                                                                                              // 23
    }                                                                                                                  // 24
  });                                                                                                                  // 25
};                                                                                                                     // 26
                                                                                                                       //
var enableLegacyOAuthProvidersIfNotInSettings = function (db, backend) {                                               // 28
  // In the before time, Google and Github login were enabled by default.                                              // 29
  //                                                                                                                   // 30
  // This actually didn't make much sense, required the first user to configure                                        // 31
  // OAuth, and had some trust-the-first-user properties that weren't totally                                          // 32
  // secure.                                                                                                           // 33
  //                                                                                                                   // 34
  // Now, we have admin-token, so we wish to disable all logins by default                                             // 35
  // (since they need to be configured anyway) but since users may have never                                          // 36
  // explicitly told Sandstorm that Google or Github login should be enabled,                                          // 37
  // we can't just use the value in the Settings collection, since it might                                            // 38
  // never have been set.                                                                                              // 39
  //                                                                                                                   // 40
  // Thus, if the service is configured but absent from Settings, we should                                            // 41
  // explicitly enable it in Settings, and then the rest of the logic can just                                         // 42
  // depend on what value is in Settings and default to false without breaking                                         // 43
  // user installations.                                                                                               // 44
  var configurations = ServiceConfiguration.configurations;                                                            // 45
  ["google", "github"].forEach(function (serviceName) {                                                                // 46
    var config = configurations.findOne({                                                                              // 47
      service: serviceName                                                                                             // 47
    });                                                                                                                // 47
    var serviceConfig = db.collections.settings.findOne({                                                              // 48
      _id: serviceName                                                                                                 // 48
    });                                                                                                                // 48
                                                                                                                       //
    if (config && !serviceConfig) {                                                                                    // 49
      // Only explicitly enable the login service if:                                                                  // 50
      // 1) the service is already configured                                                                          // 51
      // 2) there is no sandstorm configuration already present (the user was                                          // 52
      //    using the previous default behavior).                                                                      // 53
      db.collections.settings.insert({                                                                                 // 54
        _id: serviceName,                                                                                              // 54
        value: true                                                                                                    // 54
      });                                                                                                              // 54
    }                                                                                                                  // 55
  });                                                                                                                  // 56
};                                                                                                                     // 57
                                                                                                                       //
var denormalizeInviteInfo = function (db, backend) {                                                                   // 59
  // When a user is invited via a signup token, the `signupKey` field of their user table entry                        // 60
  // has always been populated to indicate the key they used. This points into the SignupKeys table                    // 61
  // which has more information about the key, namely a freeform note entered by the admin when                        // 62
  // they created the key. In the case that the email invite form was used, the note has the form                      // 63
  // "E-mail invite to <address>".                                                                                     // 64
  //                                                                                                                   // 65
  // Later, we decided it was useful to indicate in the users table visible to the admin                               // 66
  // information about the invite terms. Namely, for email invites we want to show the address                         // 67
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                          // 68
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing                        // 69
  // from the SignupKeys table.                                                                                        // 70
  db.collections.users.find().forEach(function (user) {                                                                // 72
    if (user.signupKey && typeof user.signupKey === "string" && user.signupKey !== "admin") {                          // 73
      var signupInfo = db.collections.signupKeys.findOne(user.signupKey);                                              // 74
                                                                                                                       //
      if (signupInfo && signupInfo.note) {                                                                             // 75
        var newFields = {                                                                                              // 76
          signupNote: signupInfo.note                                                                                  // 76
        };                                                                                                             // 76
        var prefix = "E-mail invite to ";                                                                              // 78
                                                                                                                       //
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                               // 79
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                                // 80
        }                                                                                                              // 81
                                                                                                                       //
        db.collections.users.update(user._id, {                                                                        // 83
          $set: newFields                                                                                              // 83
        });                                                                                                            // 83
      }                                                                                                                // 84
    }                                                                                                                  // 85
  });                                                                                                                  // 86
};                                                                                                                     // 87
                                                                                                                       //
var mergeRoleAssignmentsIntoApiTokens = function (db, backend) {                                                       // 89
  db.collections.roleAssignments.find().forEach(function (roleAssignment) {                                            // 90
    db.collections.apiTokens.insert({                                                                                  // 91
      grainId: roleAssignment.grainId,                                                                                 // 92
      userId: roleAssignment.sharer,                                                                                   // 93
      roleAssignment: roleAssignment.roleAssignment,                                                                   // 94
      petname: roleAssignment.petname,                                                                                 // 95
      created: roleAssignment.created,                                                                                 // 96
      owner: {                                                                                                         // 97
        user: {                                                                                                        // 98
          userId: roleAssignment.recipient,                                                                            // 99
          title: roleAssignment.title                                                                                  // 100
        }                                                                                                              // 98
      }                                                                                                                // 97
    });                                                                                                                // 91
  });                                                                                                                  // 104
};                                                                                                                     // 105
                                                                                                                       //
var fixOasisStorageUsageStats = function (db, backend) {}; // This migration only pertained to Oasis and it was successfully applied there. Since it referred
// to some global variables that we later wanted to remove and/or rename, we've since replaced it                      // 109
// with a no-op.                                                                                                       // 110
                                                                                                                       //
                                                                                                                       //
var fetchProfilePictures = function (db, backend) {                                                                    // 112
  db.collections.users.find({}).forEach(function (user) {                                                              // 113
    var url = userPictureUrl(user);                                                                                    // 114
                                                                                                                       //
    if (url) {                                                                                                         // 115
      console.log("Fetching user picture:", url);                                                                      // 116
      var assetId = fetchPicture(db, url);                                                                             // 117
                                                                                                                       //
      if (assetId) {                                                                                                   // 118
        db.collections.users.update(user._id, {                                                                        // 119
          $set: {                                                                                                      // 119
            "profile.picture": assetId                                                                                 // 119
          }                                                                                                            // 119
        });                                                                                                            // 119
      }                                                                                                                // 120
    }                                                                                                                  // 121
  });                                                                                                                  // 122
};                                                                                                                     // 123
                                                                                                                       //
var assignPlans = function (db, backend) {// This was a one-time migration intended to be applied on Oasis to existing users.
  // It has run, so we only need this stub function here.                                                              // 127
};                                                                                                                     // 128
                                                                                                                       //
var removeKeyrings = function (db, backend) {                                                                          // 130
  // These blobs full of public keys were not intended to find their way into mongo and while                          // 131
  // harmless they slow things down because they're huge. Remove them.                                                 // 132
  db.collections.packages.update({                                                                                     // 133
    "manifest.metadata.pgpKeyring": {                                                                                  // 133
      $exists: true                                                                                                    // 133
    }                                                                                                                  // 133
  }, {                                                                                                                 // 133
    $unset: {                                                                                                          // 134
      "manifest.metadata.pgpKeyring": ""                                                                               // 134
    }                                                                                                                  // 134
  }, {                                                                                                                 // 134
    multi: true                                                                                                        // 135
  });                                                                                                                  // 135
};                                                                                                                     // 136
                                                                                                                       //
var useLocalizedTextInUserActions = function (db, backend) {                                                           // 138
  function toLocalizedText(newObj, oldObj, field) {                                                                    // 139
    if (field in oldObj) {                                                                                             // 140
      if (typeof oldObj[field] === "string") {                                                                         // 141
        newObj[field] = {                                                                                              // 142
          defaultText: oldObj[field]                                                                                   // 142
        };                                                                                                             // 142
      } else {                                                                                                         // 143
        newObj[field] = oldObj[field];                                                                                 // 144
      }                                                                                                                // 145
    }                                                                                                                  // 146
  }                                                                                                                    // 147
                                                                                                                       //
  db.collections.userActions.find({}).forEach(function (userAction) {                                                  // 149
    var fields = {};                                                                                                   // 150
    toLocalizedText(fields, userAction, "appTitle");                                                                   // 151
    toLocalizedText(fields, userAction, "title");                                                                      // 152
    toLocalizedText(fields, userAction, "nounPhrase");                                                                 // 153
    db.collections.userActions.update(userAction._id, {                                                                // 154
      $set: fields                                                                                                     // 154
    });                                                                                                                // 154
  });                                                                                                                  // 155
};                                                                                                                     // 156
                                                                                                                       //
var verifyAllPgpSignatures = function (db, backend) {                                                                  // 158
  db.collections.packages.find({}).forEach(function (pkg) {                                                            // 159
    try {                                                                                                              // 160
      console.log("checking PGP signature for package:", pkg._id);                                                     // 161
      var info = waitPromise(backend.cap().tryGetPackage(pkg._id));                                                    // 162
                                                                                                                       //
      if (info.authorPgpKeyFingerprint) {                                                                              // 163
        console.log("  " + info.authorPgpKeyFingerprint);                                                              // 164
        db.collections.packages.update(pkg._id, {                                                                      // 165
          $set: {                                                                                                      // 166
            authorPgpKeyFingerprint: info.authorPgpKeyFingerprint                                                      // 166
          }                                                                                                            // 166
        });                                                                                                            // 166
      } else {                                                                                                         // 167
        console.log("  no signature");                                                                                 // 168
      }                                                                                                                // 169
    } catch (err) {                                                                                                    // 170
      console.error(err.stack);                                                                                        // 171
    }                                                                                                                  // 172
  });                                                                                                                  // 173
};                                                                                                                     // 174
                                                                                                                       //
var splitUserIdsIntoAccountIdsAndIdentityIds = function (db, backend) {                                                // 176
  db.collections.users.find().forEach(function (user) {                                                                // 177
    var identity = {};                                                                                                 // 178
    var serviceUserId = void 0;                                                                                        // 179
                                                                                                                       //
    if ("devName" in user) {                                                                                           // 180
      identity.service = "dev";                                                                                        // 181
      serviceUserId = user.devName;                                                                                    // 182
    } else if ("expires" in user) {                                                                                    // 183
      identity.service = "demo";                                                                                       // 184
      serviceUserId = user._id;                                                                                        // 185
    } else if (user.services && "google" in user.services) {                                                           // 186
      identity.service = "google";                                                                                     // 187
                                                                                                                       //
      if (user.services.google.email && user.services.google.verified_email) {                                         // 188
        // jscs:ignore requireCamelCaseOrUpperCaseIdentifiers                                                          // 188
        identity.verifiedEmail = user.services.google.email;                                                           // 189
      }                                                                                                                // 190
                                                                                                                       //
      serviceUserId = user.services.google.id;                                                                         // 192
    } else if (user.services && "github" in user.services) {                                                           // 193
      identity.service = "github";                                                                                     // 194
      identity.unverifiedEmail = user.services.github.email;                                                           // 195
      serviceUserId = user.services.github.id;                                                                         // 196
    } else if (user.services && "emailToken" in user.services) {                                                       // 197
      identity.service = "emailToken";                                                                                 // 198
      identity.verifiedEmail = user.services.emailToken.email;                                                         // 199
      serviceUserId = user.services.emailToken.email;                                                                  // 200
    }                                                                                                                  // 201
                                                                                                                       //
    identity.id = Crypto.createHash("sha256").update(identity.service + ":" + serviceUserId).digest("hex");            // 203
                                                                                                                       //
    if (user.profile) {                                                                                                // 206
      if (user.profile.name) {                                                                                         // 207
        identity.name = user.profile.name;                                                                             // 208
      }                                                                                                                // 209
                                                                                                                       //
      if (user.profile.handle) {                                                                                       // 211
        identity.handle = user.profile.handle;                                                                         // 212
      }                                                                                                                // 213
                                                                                                                       //
      if (user.profile.picture) {                                                                                      // 215
        identity.picture = user.profile.picture;                                                                       // 216
      }                                                                                                                // 217
                                                                                                                       //
      if (user.profile.pronoun) {                                                                                      // 219
        identity.pronoun = user.profile.pronoun;                                                                       // 220
      }                                                                                                                // 221
                                                                                                                       //
      if (user.profile.email) {                                                                                        // 223
        identity.unverifiedEmail = user.profile.email;                                                                 // 224
      }                                                                                                                // 225
    }                                                                                                                  // 226
                                                                                                                       //
    identity.main = true;                                                                                              // 228
    db.collections.users.update(user._id, {                                                                            // 230
      $set: {                                                                                                          // 230
        identities: [identity]                                                                                         // 230
      }                                                                                                                // 230
    });                                                                                                                // 230
    db.collections.grains.update({                                                                                     // 232
      userId: user._id                                                                                                 // 232
    }, {                                                                                                               // 232
      $set: {                                                                                                          // 232
        identityId: identity.id                                                                                        // 232
      }                                                                                                                // 232
    }, {                                                                                                               // 232
      multi: true                                                                                                      // 232
    });                                                                                                                // 232
    db.collections.sessions.update({                                                                                   // 233
      userId: user._id                                                                                                 // 233
    }, {                                                                                                               // 233
      $set: {                                                                                                          // 233
        identityId: identity.id                                                                                        // 233
      }                                                                                                                // 233
    }, {                                                                                                               // 233
      multi: true                                                                                                      // 233
    });                                                                                                                // 233
    db.collections.apiTokens.update({                                                                                  // 234
      userId: user._id                                                                                                 // 234
    }, {                                                                                                               // 234
      $set: {                                                                                                          // 235
        identityId: identity.id                                                                                        // 235
      }                                                                                                                // 235
    }, {                                                                                                               // 235
      multi: true                                                                                                      // 236
    });                                                                                                                // 236
    db.collections.apiTokens.update({                                                                                  // 237
      "owner.user.userId": user._id                                                                                    // 237
    }, {                                                                                                               // 237
      $set: {                                                                                                          // 238
        "owner.user.identityId": identity.id                                                                           // 238
      }                                                                                                                // 238
    }, {                                                                                                               // 238
      multi: true                                                                                                      // 239
    });                                                                                                                // 239
    db.collections.apiTokens.update({                                                                                  // 240
      "owner.grain.introducerUser": user._id                                                                           // 240
    }, {                                                                                                               // 240
      $set: {                                                                                                          // 241
        "owner.grain.introducerIdentity": identity.id                                                                  // 241
      }                                                                                                                // 241
    }, {                                                                                                               // 241
      multi: true                                                                                                      // 242
    });                                                                                                                // 242
                                                                                                                       //
    while (db.collections.apiTokens.update({                                                                           // 244
      "requirements.permissionsHeld.userId": user._id                                                                  // 245
    }, {                                                                                                               // 244
      $set: {                                                                                                          // 247
        "requirements.$.permissionsHeld.identityId": identity.id                                                       // 247
      },                                                                                                               // 247
      $unset: {                                                                                                        // 248
        "requirements.$.permissionsHeld.userId": 1                                                                     // 248
      }                                                                                                                // 248
    }, {                                                                                                               // 246
      multi: true                                                                                                      // 250
    }) > 0) {} // The `$` operator modifies the first element in the array that matches the query. Since               // 249
    // there may be many matches, we need to repeat until no documents are modified.                                   // 253
                                                                                                                       //
  });                                                                                                                  // 255
  db.collections.apiTokens.remove({                                                                                    // 257
    userInfo: {                                                                                                        // 257
      $exists: true                                                                                                    // 257
    }                                                                                                                  // 257
  }); // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                          // 257
  // that this field could show up in the database was in this deprecated, no-longer-functional                        // 259
  // form of API token.                                                                                                // 260
};                                                                                                                     // 261
                                                                                                                       //
var appUpdateSettings = function (db, backend) {                                                                       // 263
  db.collections.settings.insert({                                                                                     // 264
    _id: "appMarketUrl",                                                                                               // 264
    value: "https://apps.sandstorm.io"                                                                                 // 264
  });                                                                                                                  // 264
  db.collections.settings.insert({                                                                                     // 265
    _id: "appIndexUrl",                                                                                                // 265
    value: "https://app-index.sandstorm.io"                                                                            // 265
  });                                                                                                                  // 265
  db.collections.settings.insert({                                                                                     // 266
    _id: "appUpdatesEnabled",                                                                                          // 266
    value: true                                                                                                        // 266
  });                                                                                                                  // 266
};                                                                                                                     // 267
                                                                                                                       //
var moveDevAndEmailLoginDataIntoIdentities = function (db, backend) {                                                  // 269
  db.collections.users.find().forEach(function (user) {                                                                // 270
    if (user.identities.length !== 1) {                                                                                // 271
      throw new Error("User does not have exactly one identity: ", user);                                              // 272
    }                                                                                                                  // 273
                                                                                                                       //
    var identity = user.identities[0];                                                                                 // 275
                                                                                                                       //
    if (Match.test(identity.service, Object)) {                                                                        // 276
      return;                                                                                                          // 276
    } // Already migrated.                                                                                             // 276
                                                                                                                       //
                                                                                                                       //
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");                   // 278
                                                                                                                       //
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                                    // 279
    var serviceObject = {};                                                                                            // 281
    var fieldsToUnset = {};                                                                                            // 282
                                                                                                                       //
    if (identity.service === "dev") {                                                                                  // 284
      serviceObject.name = user.devName;                                                                               // 285
      fieldsToUnset.devName = 1;                                                                                       // 286
    } else if (identity.service === "emailToken") {                                                                    // 287
      serviceObject.tokens = user.services.emailToken.tokens;                                                          // 288
      serviceObject.email = user.services.emailToken.email;                                                            // 289
      fieldsToUnset["services.emailToken"] = 1;                                                                        // 290
    }                                                                                                                  // 291
                                                                                                                       //
    newIdentity.service = {};                                                                                          // 293
    newIdentity.service[identity.service] = serviceObject;                                                             // 294
    var modifier = {                                                                                                   // 296
      $set: {                                                                                                          // 296
        identities: [newIdentity]                                                                                      // 296
      }                                                                                                                // 296
    };                                                                                                                 // 296
                                                                                                                       //
    if (Object.keys(fieldsToUnset).length > 0) {                                                                       // 297
      modifier.$unset = fieldsToUnset;                                                                                 // 298
    }                                                                                                                  // 299
                                                                                                                       //
    db.collections.users.update({                                                                                      // 301
      _id: user._id                                                                                                    // 301
    }, modifier);                                                                                                      // 301
  });                                                                                                                  // 302
};                                                                                                                     // 303
                                                                                                                       //
var repairEmailIdentityIds = function (db, backend) {                                                                  // 305
  db.collections.users.find({                                                                                          // 306
    "identities.service.emailToken": {                                                                                 // 306
      $exists: 1                                                                                                       // 306
    }                                                                                                                  // 306
  }).forEach(function (user) {                                                                                         // 306
    if (user.identities.length !== 1) {                                                                                // 307
      throw new Error("User does not have exactly one identity: ", user);                                              // 308
    }                                                                                                                  // 309
                                                                                                                       //
    var identity = user.identities[0];                                                                                 // 311
                                                                                                                       //
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail", "profile");               // 312
                                                                                                                       //
    newIdentity.service = {                                                                                            // 314
      email: identity.service.emailToken                                                                               // 314
    };                                                                                                                 // 314
    newIdentity.id = Crypto.createHash("sha256").update("email:" + identity.service.emailToken.email).digest("hex");   // 315
    db.collections.grains.update({                                                                                     // 318
      identityId: identity.id                                                                                          // 318
    }, {                                                                                                               // 318
      $set: {                                                                                                          // 318
        identityId: newIdentity.id                                                                                     // 318
      }                                                                                                                // 318
    }, {                                                                                                               // 318
      multi: true                                                                                                      // 318
    });                                                                                                                // 318
    db.collections.sessions.update({                                                                                   // 319
      identityId: identity.id                                                                                          // 319
    }, {                                                                                                               // 319
      $set: {                                                                                                          // 319
        identityId: newIdentity.id                                                                                     // 319
      }                                                                                                                // 319
    }, {                                                                                                               // 319
      multi: true                                                                                                      // 319
    });                                                                                                                // 319
    db.collections.apiTokens.update({                                                                                  // 320
      identityId: identity.id                                                                                          // 320
    }, {                                                                                                               // 320
      $set: {                                                                                                          // 321
        identityId: newIdentity.id                                                                                     // 321
      }                                                                                                                // 321
    }, {                                                                                                               // 321
      multi: true                                                                                                      // 322
    });                                                                                                                // 322
    db.collections.apiTokens.update({                                                                                  // 323
      "owner.user.identityId": identity.id                                                                             // 323
    }, {                                                                                                               // 323
      $set: {                                                                                                          // 324
        "owner.user.identityId": newIdentity.id                                                                        // 324
      }                                                                                                                // 324
    }, {                                                                                                               // 324
      multi: true                                                                                                      // 325
    });                                                                                                                // 325
    db.collections.apiTokens.update({                                                                                  // 326
      "owner.grain.introducerIdentity": identity.id                                                                    // 326
    }, {                                                                                                               // 326
      $set: {                                                                                                          // 327
        "owner.grain.introducerIdentity": newIdentity.id                                                               // 327
      }                                                                                                                // 327
    }, {                                                                                                               // 327
      multi: true                                                                                                      // 328
    });                                                                                                                // 328
                                                                                                                       //
    while (db.collections.apiTokens.update({                                                                           // 330
      "requirements.permissionsHeld.identityId": identity.id                                                           // 330
    }, {                                                                                                               // 330
      $set: {                                                                                                          // 331
        "requirements.$.permissionsHeld.identityId": newIdentity.id                                                    // 331
      }                                                                                                                // 331
    }, {                                                                                                               // 331
      multi: true                                                                                                      // 332
    }) > 0) {}                                                                                                         // 332
                                                                                                                       //
    db.collections.users.update({                                                                                      // 334
      _id: user._id                                                                                                    // 334
    }, {                                                                                                               // 334
      $set: {                                                                                                          // 334
        identities: [newIdentity]                                                                                      // 334
      }                                                                                                                // 334
    });                                                                                                                // 334
  });                                                                                                                  // 335
};                                                                                                                     // 336
                                                                                                                       //
var splitAccountUsersAndIdentityUsers = function (db, backend) {                                                       // 338
  db.collections.users.find({                                                                                          // 339
    identities: {                                                                                                      // 339
      $exists: true                                                                                                    // 339
    }                                                                                                                  // 339
  }).forEach(function (user) {                                                                                         // 339
    if (user.identities.length !== 1) {                                                                                // 340
      throw new Error("User does not have exactly one identity: ", user);                                              // 341
    }                                                                                                                  // 342
                                                                                                                       //
    var identity = user.identities[0];                                                                                 // 344
                                                                                                                       //
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires");                                             // 345
                                                                                                                       //
    identityUser._id = identity.id;                                                                                    // 346
    identityUser.profile = identity.profile;                                                                           // 347
                                                                                                                       //
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                                       // 348
                                                                                                                       //
    identityUser.profile.service = Object.keys(identity.service)[0]; // Updating this user needs to be a two step process because the `services` field typically
    // contains subfields that are constrained to be unique by Mongo indices.                                          // 352
                                                                                                                       //
    identityUser.stagedServices = _.omit(user.services, "resume");                                                     // 353
                                                                                                                       //
    if (identity.service.dev) {                                                                                        // 354
      identityUser.stagedServices.dev = identity.service.dev;                                                          // 355
    } else if (identity.service.email) {                                                                               // 356
      identityUser.stagedServices.email = identity.service.email;                                                      // 357
    }                                                                                                                  // 358
                                                                                                                       //
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires", "isAdmin", "signupKey", "signupNote", "signupEmail", "plan", "storageUsage", "isAppDemoUser", "appDemoId", "payments", "dailySentMailCount", "hasCompletedSignup");
                                                                                                                       //
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                            // 364
    accountUser.nonloginIdentities = [];                                                                               // 365
                                                                                                                       //
    if (user.services && user.services.resume) {                                                                       // 366
      accountUser.services = {                                                                                         // 367
        resume: user.services.resume                                                                                   // 367
      };                                                                                                               // 367
    }                                                                                                                  // 368
                                                                                                                       //
    accountUser.stashedOldUser = user;                                                                                 // 370
    db.collections.apiTokens.update({                                                                                  // 372
      identityId: identityUser._id                                                                                     // 373
    }, {                                                                                                               // 373
      $set: {                                                                                                          // 374
        accountId: user._id                                                                                            // 374
      }                                                                                                                // 374
    }, {                                                                                                               // 374
      multi: true                                                                                                      // 375
    });                                                                                                                // 375
    db.collections.users.upsert({                                                                                      // 377
      _id: identityUser._id                                                                                            // 377
    }, identityUser);                                                                                                  // 377
    db.collections.users.update({                                                                                      // 378
      _id: user._id                                                                                                    // 378
    }, accountUser);                                                                                                   // 378
  });                                                                                                                  // 379
  db.collections.users.find({                                                                                          // 381
    stagedServices: {                                                                                                  // 381
      $exists: true                                                                                                    // 381
    }                                                                                                                  // 381
  }).forEach(function (identity) {                                                                                     // 381
    db.collections.users.update({                                                                                      // 382
      _id: identity._id                                                                                                // 382
    }, {                                                                                                               // 382
      $unset: {                                                                                                        // 383
        stagedServices: 1                                                                                              // 383
      },                                                                                                               // 383
      $set: {                                                                                                          // 384
        services: identity.stagedServices                                                                              // 384
      }                                                                                                                // 384
    });                                                                                                                // 382
  });                                                                                                                  // 386
};                                                                                                                     // 387
                                                                                                                       //
var populateContactsFromApiTokens = function (db, backend) {                                                           // 389
  db.collections.apiTokens.find({                                                                                      // 390
    "owner.user.identityId": {                                                                                         // 391
      $exists: 1                                                                                                       // 391
    },                                                                                                                 // 391
    accountId: {                                                                                                       // 392
      $exists: 1                                                                                                       // 392
    }                                                                                                                  // 392
  }).forEach(function (token) {                                                                                        // 390
    var identityId = token.owner.user.identityId;                                                                      // 394
    var identity = db.getIdentity(identityId);                                                                         // 395
                                                                                                                       //
    if (identity) {                                                                                                    // 396
      var profile = identity.profile;                                                                                  // 397
      db.collections.contacts.upsert({                                                                                 // 398
        ownerId: token.accountId,                                                                                      // 398
        identityId: identityId                                                                                         // 398
      }, {                                                                                                             // 398
        ownerId: token.accountId,                                                                                      // 399
        petname: profile && profile.name,                                                                              // 400
        created: new Date(),                                                                                           // 401
        identityId: identityId,                                                                                        // 402
        profile: profile                                                                                               // 403
      });                                                                                                              // 398
    }                                                                                                                  // 405
  });                                                                                                                  // 406
};                                                                                                                     // 407
                                                                                                                       //
var cleanUpApiTokens = function (db, backend) {                                                                        // 409
  // The `splitUserIdsIntoAccountIdsAndIdentityIds()` migration only added `identityId` in cases                       // 410
  // where the user still existed in the database.                                                                     // 411
  db.collections.apiTokens.remove({                                                                                    // 412
    userId: {                                                                                                          // 413
      $exists: true                                                                                                    // 413
    },                                                                                                                 // 413
    identityId: {                                                                                                      // 414
      $exists: false                                                                                                   // 414
    }                                                                                                                  // 414
  });                                                                                                                  // 412
  db.collections.apiTokens.remove({                                                                                    // 416
    "owner.user.userId": {                                                                                             // 417
      $exists: true                                                                                                    // 417
    },                                                                                                                 // 417
    "owner.user.identityId": {                                                                                         // 418
      $exists: false                                                                                                   // 418
    }                                                                                                                  // 418
  }); // For a while we were accidentally setting `appIcon` instead of `icon`.                                         // 416
                                                                                                                       //
  db.collections.apiTokens.find({                                                                                      // 422
    "owner.user.denormalizedGrainMetadata.appIcon": {                                                                  // 423
      $exists: true                                                                                                    // 423
    }                                                                                                                  // 423
  }).forEach(function (apiToken) {                                                                                     // 422
    var icon = apiToken.owner.user.denormalizedGrainMetadata.appIcon;                                                  // 425
    db.collections.apiTokens.update({                                                                                  // 426
      _id: apiToken._id                                                                                                // 426
    }, {                                                                                                               // 426
      $set: {                                                                                                          // 427
        "owner.user.denormalizedGrainMetadata.icon": icon                                                              // 427
      },                                                                                                               // 427
      $unset: {                                                                                                        // 428
        "owner.user.denormalizedGrainMetadata.appIcon": true                                                           // 428
      }                                                                                                                // 428
    });                                                                                                                // 426
  }); // For a while the `identityId` field of child UiView tokens was not getting set.                                // 430
                                                                                                                       //
  function repairChain(parentToken) {                                                                                  // 433
    db.collections.apiTokens.find({                                                                                    // 434
      parentToken: parentToken._id,                                                                                    // 435
      grainId: {                                                                                                       // 436
        $exists: true                                                                                                  // 436
      },                                                                                                               // 436
      identityId: {                                                                                                    // 437
        $exists: false                                                                                                 // 437
      }                                                                                                                // 437
    }).forEach(function (childToken) {                                                                                 // 434
      db.collections.apiTokens.update({                                                                                // 439
        _id: childToken._id                                                                                            // 439
      }, {                                                                                                             // 439
        $set: {                                                                                                        // 439
          identityId: parentToken.identityId                                                                           // 439
        }                                                                                                              // 439
      });                                                                                                              // 439
      repairChain(childToken);                                                                                         // 440
    });                                                                                                                // 441
  }                                                                                                                    // 442
                                                                                                                       //
  db.collections.apiTokens.find({                                                                                      // 444
    grainId: {                                                                                                         // 445
      $exists: true                                                                                                    // 445
    },                                                                                                                 // 445
    identityId: {                                                                                                      // 446
      $exists: true                                                                                                    // 446
    },                                                                                                                 // 446
    parentToken: {                                                                                                     // 447
      $exists: false                                                                                                   // 447
    }                                                                                                                  // 447
  }).forEach(repairChain);                                                                                             // 444
};                                                                                                                     // 449
                                                                                                                       //
var initServerTitleAndReturnAddress = function (db, backend) {                                                         // 451
  var hostname = Url.parse(process.env.ROOT_URL).hostname;                                                             // 452
  db.collections.settings.insert({                                                                                     // 453
    _id: "serverTitle",                                                                                                // 453
    value: hostname                                                                                                    // 453
  });                                                                                                                  // 453
  db.collections.settings.insert({                                                                                     // 454
    _id: "returnAddress",                                                                                              // 454
    value: "no-reply@" + hostname                                                                                      // 454
  });                                                                                                                  // 454
};                                                                                                                     // 455
                                                                                                                       //
var sendReferralNotifications = function (db, backend) {                                                               // 457
  if (db.isReferralEnabled()) {                                                                                        // 458
    db.collections.users.find({                                                                                        // 459
      loginIdentities: {                                                                                               // 460
        $exists: true                                                                                                  // 460
      },                                                                                                               // 460
      expires: {                                                                                                       // 461
        $exists: false                                                                                                 // 461
      }                                                                                                                // 461
    }, {                                                                                                               // 459
      fields: {                                                                                                        // 462
        _id: 1                                                                                                         // 462
      }                                                                                                                // 462
    }).forEach(function (user) {                                                                                       // 462
      db.sendReferralProgramNotification(user._id);                                                                    // 463
    });                                                                                                                // 464
  }                                                                                                                    // 465
};                                                                                                                     // 466
                                                                                                                       //
var assignBonuses = function (db, backend) {// This was a one-time migration intended to be applied on Oasis to existing users.
  // It has run, so we only need this stub function here.                                                              // 470
};                                                                                                                     // 471
                                                                                                                       //
var splitSmtpUrl = function (db, backend) {                                                                            // 473
  var smtpUrlSetting = db.collections.settings.findOne({                                                               // 474
    _id: "smtpUrl"                                                                                                     // 474
  });                                                                                                                  // 474
  var smtpUrl = smtpUrlSetting ? smtpUrlSetting.value : process.env.MAIL_URL;                                          // 475
  var returnAddress = db.collections.settings.findOne({                                                                // 476
    _id: "returnAddress"                                                                                               // 476
  }); // Default values.                                                                                               // 476
                                                                                                                       //
  var smtpConfig = {                                                                                                   // 479
    hostname: "localhost",                                                                                             // 480
    port: "25",                                                                                                        // 481
    auth: undefined,                                                                                                   // 482
    returnAddress: returnAddress.value                                                                                 // 483
  };                                                                                                                   // 479
  var parsed = void 0;                                                                                                 // 486
                                                                                                                       //
  try {                                                                                                                // 487
    parsed = smtpUrl && Url.parse(smtpUrl);                                                                            // 488
  } catch (e) {}                                                                                                       // 489
                                                                                                                       //
  if (parsed) {                                                                                                        // 491
    // If there was a SMTP URL previously defined, import its data.                                                    // 492
    var auth = undefined;                                                                                              // 493
                                                                                                                       //
    if (parsed.auth) {                                                                                                 // 494
      var colonIndex = parsed.auth.indexOf(":");                                                                       // 495
      var user = undefined;                                                                                            // 496
      var pass = undefined;                                                                                            // 497
                                                                                                                       //
      if (colonIndex !== -1) {                                                                                         // 498
        user = parsed.auth.slice(0, colonIndex);                                                                       // 499
        pass = parsed.auth.slice(colonIndex + 1);                                                                      // 500
      }                                                                                                                // 501
                                                                                                                       //
      auth = {                                                                                                         // 503
        user: user,                                                                                                    // 504
        pass: pass                                                                                                     // 505
      };                                                                                                               // 503
    } // Override defaults with previous config's values.                                                              // 507
                                                                                                                       //
                                                                                                                       //
    smtpConfig.hostname = parsed.hostname || "localhost";                                                              // 510
    smtpConfig.port = parsed.port || "25";                                                                             // 511
    smtpConfig.auth = auth;                                                                                            // 512
  }                                                                                                                    // 513
                                                                                                                       //
  db.collections.settings.upsert({                                                                                     // 515
    _id: "smtpConfig"                                                                                                  // 515
  }, {                                                                                                                 // 515
    value: smtpConfig                                                                                                  // 515
  });                                                                                                                  // 515
  db.collections.settings.remove({                                                                                     // 516
    _id: "returnAddress"                                                                                               // 516
  });                                                                                                                  // 516
  db.collections.settings.remove({                                                                                     // 517
    _id: "smtpUrl"                                                                                                     // 517
  });                                                                                                                  // 517
};                                                                                                                     // 518
                                                                                                                       //
var smtpPortShouldBeNumber = function (db, backend) {                                                                  // 520
  var entry = db.collections.settings.findOne({                                                                        // 521
    _id: "smtpConfig"                                                                                                  // 521
  });                                                                                                                  // 521
                                                                                                                       //
  if (entry) {                                                                                                         // 522
    var setting = entry.value;                                                                                         // 523
                                                                                                                       //
    if (setting.port) {                                                                                                // 524
      setting.port = _.isNumber(setting.port) ? setting.port : parseInt(setting.port);                                 // 525
      db.collections.settings.upsert({                                                                                 // 526
        _id: "smtpConfig"                                                                                              // 526
      }, {                                                                                                             // 526
        value: setting                                                                                                 // 526
      });                                                                                                              // 526
    }                                                                                                                  // 527
  }                                                                                                                    // 528
};                                                                                                                     // 529
                                                                                                                       //
var consolidateOrgSettings = function (db, backend) {                                                                  // 531
  var settings = db.collections.settings;                                                                              // 532
  var orgGoogleDomain = settings.findOne({                                                                             // 533
    _id: "organizationGoogle"                                                                                          // 533
  });                                                                                                                  // 533
  var orgEmailDomain = settings.findOne({                                                                              // 534
    _id: "organizationEmail"                                                                                           // 534
  });                                                                                                                  // 534
  var orgLdap = settings.findOne({                                                                                     // 535
    _id: "organizationLdap"                                                                                            // 535
  });                                                                                                                  // 535
  var orgSaml = settings.findOne({                                                                                     // 536
    _id: "organizationSaml"                                                                                            // 536
  });                                                                                                                  // 536
  var orgMembership = {                                                                                                // 538
    google: {                                                                                                          // 539
      enabled: orgGoogleDomain ? !!orgGoogleDomain.value : false,                                                      // 540
      domain: orgGoogleDomain ? orgGoogleDomain.value : ""                                                             // 541
    },                                                                                                                 // 539
    email: {                                                                                                           // 543
      enabled: orgEmailDomain ? !!orgEmailDomain.value : false,                                                        // 544
      domain: orgEmailDomain ? orgEmailDomain.value : ""                                                               // 545
    },                                                                                                                 // 543
    ldap: {                                                                                                            // 547
      enabled: orgLdap ? orgLdap.value : false                                                                         // 548
    },                                                                                                                 // 547
    saml: {                                                                                                            // 550
      enabled: orgSaml ? orgSaml.value : false                                                                         // 551
    }                                                                                                                  // 550
  };                                                                                                                   // 538
  settings.upsert({                                                                                                    // 555
    _id: "organizationMembership"                                                                                      // 555
  }, {                                                                                                                 // 555
    value: orgMembership                                                                                               // 555
  });                                                                                                                  // 555
  settings.remove({                                                                                                    // 556
    _id: "organizationGoogle"                                                                                          // 556
  });                                                                                                                  // 556
  settings.remove({                                                                                                    // 557
    _id: "organizationEmail"                                                                                           // 557
  });                                                                                                                  // 557
  settings.remove({                                                                                                    // 558
    _id: "organizationLdap"                                                                                            // 558
  });                                                                                                                  // 558
  settings.remove({                                                                                                    // 559
    _id: "organizationSaml"                                                                                            // 559
  });                                                                                                                  // 559
};                                                                                                                     // 560
                                                                                                                       //
var unsetSmtpDefaultHostnameIfNoUsersExist = function (db, backend) {                                                  // 562
  // We don't actually want to have the default hostname "localhost" set.                                              // 563
  // If the user has already finished configuring their server, then this migration should do                          // 564
  // nothing (since we might break their deployment), but for new installs (which will have no users                   // 565
  // at the time this migration runs) we'll unset the hostname if it's still the previously-filled                     // 566
  // default value.                                                                                                    // 567
  var hasUsers = db.collections.users.findOne();                                                                       // 568
                                                                                                                       //
  if (!hasUsers) {                                                                                                     // 569
    var entry = db.collections.settings.findOne({                                                                      // 570
      _id: "smtpConfig"                                                                                                // 570
    });                                                                                                                // 570
    var smtpConfig = entry.value;                                                                                      // 571
                                                                                                                       //
    if (smtpConfig.hostname === "localhost") {                                                                         // 572
      smtpConfig.hostname = "";                                                                                        // 573
      db.collections.settings.upsert({                                                                                 // 574
        _id: "smtpConfig"                                                                                              // 574
      }, {                                                                                                             // 574
        value: smtpConfig                                                                                              // 574
      });                                                                                                              // 574
    }                                                                                                                  // 575
  }                                                                                                                    // 576
};                                                                                                                     // 577
                                                                                                                       //
var extractLastUsedFromApiTokenOwner = function (db, backend) {                                                        // 579
  // We used to store lastUsed as a field on owner.user.  It makes more sense to store lastUsed on                     // 580
  // the apiToken as a whole.  This migration hoists such values from owner.user onto the apiToken                     // 581
  // itself.                                                                                                           // 582
  db.collections.apiTokens.find({                                                                                      // 583
    "owner.user": {                                                                                                    // 583
      $exists: true                                                                                                    // 583
    }                                                                                                                  // 583
  }).forEach(function (token) {                                                                                        // 583
    var lastUsed = token.owner.user.lastUsed;                                                                          // 584
    db.collections.apiTokens.update({                                                                                  // 585
      _id: token._id                                                                                                   // 585
    }, {                                                                                                               // 585
      $set: {                                                                                                          // 586
        lastUsed: lastUsed                                                                                             // 586
      },                                                                                                               // 586
      $unset: {                                                                                                        // 587
        "owner.user.lastUsed": true                                                                                    // 587
      }                                                                                                                // 587
    });                                                                                                                // 585
  });                                                                                                                  // 589
};                                                                                                                     // 590
                                                                                                                       //
var setUpstreamTitles = function (db, backend) {                                                                       // 592
  // Initializes the `upstreamTitle` and `renamed` fields of `ApiToken.owner.user`.                                    // 593
  var apiTokensRaw = db.collections.apiTokens.rawCollection();                                                         // 595
  var aggregateApiTokens = Meteor.wrapAsync(apiTokensRaw.aggregate, apiTokensRaw); // First, construct a list of all *shared* grains. We will need to do a separate update()
  // for each of these, so we'd like to skip those which have nothing to update.                                       // 599
                                                                                                                       //
  var grainIds = aggregateApiTokens([{                                                                                 // 600
    $match: {                                                                                                          // 601
      "owner.user": {                                                                                                  // 601
        $exists: true                                                                                                  // 601
      },                                                                                                               // 601
      grainId: {                                                                                                       // 601
        $exists: true                                                                                                  // 601
      }                                                                                                                // 601
    }                                                                                                                  // 601
  }, {                                                                                                                 // 601
    $group: {                                                                                                          // 602
      _id: "$grainId"                                                                                                  // 602
    }                                                                                                                  // 602
  }]).map(function (grain) {                                                                                           // 602
    return grain._id;                                                                                                  // 603
  });                                                                                                                  // 603
  var count = 0;                                                                                                       // 605
  db.collections.grains.find({                                                                                         // 606
    _id: {                                                                                                             // 606
      $in: grainIds                                                                                                    // 606
    }                                                                                                                  // 606
  }, {                                                                                                                 // 606
    fields: {                                                                                                          // 606
      title: 1                                                                                                         // 606
    }                                                                                                                  // 606
  }).forEach(function (grain) {                                                                                        // 606
    if (count % 100 == 0) {                                                                                            // 607
      console.log(count + " / " + grainIds.length);                                                                    // 608
    }                                                                                                                  // 609
                                                                                                                       //
    ++count; // For ApiTokens whose petname titles do not match the upstream title, we need to set                     // 611
    // `upstreamTitle` and `renamed`. We have no way of knowing which of the two names was the                         // 614
    // original name, so we don't know whether it was the owner or the receiver who renamed their                      // 615
    // copy post-sharaing. We assume it was the owner, because:                                                        // 616
    // 1. It's probably unusual for people to try to rename a grain they don't own.                                    // 617
    // 2. This results in UI that is reasonably non-confusing even if we guessed wrong. Namely,                        // 618
    //    the user sees "Owner's title (was: User's title)", which is reasonably OK even if it                         // 619
    //    was the user who had renamed their copy. They can rename it again if they like. On the                       // 620
    //    other hand, if we guessed wrongly in the other direction, the user would see                                 // 621
    //    "User's title (renamed from: Owners title)", which would be wrong if it was in fact the                      // 622
    //    owner who renamed post-sharing.                                                                              // 623
                                                                                                                       //
    db.collections.apiTokens.update({                                                                                  // 624
      grainId: grain._id,                                                                                              // 625
      "owner.user.title": {                                                                                            // 626
        $exists: true,                                                                                                 // 626
        $ne: grain.title                                                                                               // 626
      }                                                                                                                // 626
    }, {                                                                                                               // 624
      $set: {                                                                                                          // 627
        "owner.user.upstreamTitle": grain.title                                                                        // 627
      }                                                                                                                // 627
    }, {                                                                                                               // 627
      multi: true                                                                                                      // 627
    });                                                                                                                // 627
  });                                                                                                                  // 628
};                                                                                                                     // 629
                                                                                                                       //
var markAllRead = function (db, backend) {                                                                             // 631
  // Mark as "read" all grains and tokens that predate the creation of read/unread status.                             // 632
  // Otherwise it's pretty annoying to see all your old grains look like they have activity.                           // 633
  db.collections.grains.update({}, {                                                                                   // 635
    $set: {                                                                                                            // 635
      ownerSeenAllActivity: true                                                                                       // 635
    }                                                                                                                  // 635
  }, {                                                                                                                 // 635
    multi: true                                                                                                        // 635
  });                                                                                                                  // 635
  db.collections.apiTokens.update({                                                                                    // 636
    "owner.user": {                                                                                                    // 636
      $exists: true                                                                                                    // 636
    }                                                                                                                  // 636
  }, {                                                                                                                 // 636
    $set: {                                                                                                            // 637
      "owner.user.seenAllActivity": true                                                                               // 637
    }                                                                                                                  // 637
  }, {                                                                                                                 // 637
    multi: true                                                                                                        // 638
  });                                                                                                                  // 638
};                                                                                                                     // 639
                                                                                                                       //
var clearAppIndex = function (db, backend) {                                                                           // 641
  // Due to a bug in the app update code, some app update notifications that the user accepted                         // 642
  // around July 9-16, 2016 may not have applied. We have no way of knowing exactly which updates                      // 643
  // the user accepted but didn't receive. Instead, to recover, we are clearing the local cache of                     // 644
  // the app index one time. This way, the next time the app index is re-fetched, the server will                      // 645
  // interpret all of the apps in the index as updated and will re-deliver notifications to                            // 646
  // everyone. This may mean users get notifications that they previously dismissed, but they can                      // 647
  // click "dismiss" again easily enough.                                                                              // 648
  db.collections.appIndex.remove({});                                                                                  // 650
};                                                                                                                     // 651
                                                                                                                       //
var assignEmailVerifierIds = function (db, backend) {                                                                  // 653
  // Originally, the ID of an EmailVerifier was actually the _id of the root token from which it                       // 654
  // was restored. This was broken, though: Conceptually, it meant that you couldn't have a working                    // 655
  // EmailVerifier that had not been restore()d from disk. In practice, that wasn't a problem due                      // 656
  // to the fact that powerbox selections always involve a restore(), but a different problem                          // 657
  // existed: the capability returned by claimRequest() would return one ID, but after save()ing                       // 658
  // and restore()ing that capability, the ID would be different, because the new capability would                     // 659
  // be a copy -- not a child -- of the original. Additionally, the whole thing made various code                      // 660
  // ugly because it was puncturing layers of abstraction. So, we switched to doing the right                          // 661
  // thing: assigning an ID to the EmailVerifier on first creation and storing it separately.                          // 662
  db.collections.apiTokens.find({                                                                                      // 664
    "frontendRef.emailVerifier": {                                                                                     // 664
      $exists: true                                                                                                    // 664
    }                                                                                                                  // 664
  }).forEach(function (token) {                                                                                        // 664
    db.collections.apiTokens.update(token._id, {                                                                       // 665
      $set: {                                                                                                          // 665
        "frontendRef.emailVerifier.id": token._id                                                                      // 665
      }                                                                                                                // 665
    });                                                                                                                // 665
  });                                                                                                                  // 666
};                                                                                                                     // 667
                                                                                                                       //
var startPreinstallingApps = function (db, backend) {                                                                  // 669
  // This isn't really a normal migration. It will run only on brand new servers, and it has to                        // 670
  // run after the `clearAppIndex` migration because it relies on populating AppIndex.                                 // 671
  var startPreinstallingAppsHelper = function () {                                                                     // 673
    db.updateAppIndex();                                                                                               // 674
    var preinstalledApps = db.collections.appIndex.find({                                                              // 676
      _id: {                                                                                                           // 676
        $in: db.getProductivitySuiteAppIds().concat(db.getSystemSuiteAppIds())                                         // 677
      }                                                                                                                // 676
    }).fetch();                                                                                                        // 676
                                                                                                                       //
    var appAndPackageIds = _.map(preinstalledApps, function (app) {                                                    // 680
      return {                                                                                                         // 681
        appId: app.appId,                                                                                              // 682
        packageId: app.packageId                                                                                       // 683
      };                                                                                                               // 681
    });                                                                                                                // 685
                                                                                                                       //
    db.setPreinstalledApps(appAndPackageIds);                                                                          // 687
  };                                                                                                                   // 688
                                                                                                                       //
  if (!Meteor.settings.public.isTesting && !db.allowDevAccounts()) {                                                   // 690
    // We want preinstalling apps to run async and not block startup.                                                  // 691
    Meteor.setTimeout(startPreinstallingAppsHelper, 0);                                                                // 692
  }                                                                                                                    // 693
};                                                                                                                     // 694
                                                                                                                       //
var setNewServer = function (db, backend) {                                                                            // 696
  // This migration only applies to "old" servers. New servers will set                                                // 697
  // new_server_migrations_applied to false before any migrations run.                                                 // 698
  if (!db.collections.migrations.findOne({                                                                             // 699
    _id: "new_server_migrations_applied"                                                                               // 699
  })) {                                                                                                                // 699
    db.collections.migrations.insert({                                                                                 // 700
      _id: "new_server_migrations_applied",                                                                            // 700
      value: true                                                                                                      // 700
    });                                                                                                                // 700
  }                                                                                                                    // 701
};                                                                                                                     // 702
                                                                                                                       //
var addMembraneRequirementsToIdentities = function (db, backend) {                                                     // 704
  var query = {                                                                                                        // 705
    "frontendRef.identity": {                                                                                          // 706
      $exists: true                                                                                                    // 706
    },                                                                                                                 // 706
    "owner.grain.grainId": {                                                                                           // 707
      $exists: true                                                                                                    // 707
    },                                                                                                                 // 707
    "requirements.0": {                                                                                                // 708
      $exists: false                                                                                                   // 708
    }                                                                                                                  // 708
  };                                                                                                                   // 705
  db.collections.apiTokens.find(query).map(function (apiToken) {                                                       // 711
    db.collections.apiTokens.update({                                                                                  // 712
      _id: apiToken._id                                                                                                // 713
    }, {                                                                                                               // 713
      $push: {                                                                                                         // 715
        requirements: {                                                                                                // 716
          permissionsHeld: {                                                                                           // 717
            grainId: apiToken.owner.grain.grainId,                                                                     // 718
            identityId: apiToken.frontendRef.identity,                                                                 // 719
            permissions: []                                                                                            // 720
          }                                                                                                            // 717
        }                                                                                                              // 716
      }                                                                                                                // 715
    });                                                                                                                // 714
  });                                                                                                                  // 726
};                                                                                                                     // 727
                                                                                                                       //
var addEncryptionToFrontendRefIpNetwork = function (db, backend) {                                                     // 729
  db.collections.apiTokens.find({                                                                                      // 730
    "frontendRef.ipNetwork": true                                                                                      // 730
  }).map(function (apiToken) {                                                                                         // 730
    db.collections.apiTokens.update({                                                                                  // 731
      _id: apiToken._id                                                                                                // 732
    }, {                                                                                                               // 732
      $set: {                                                                                                          // 733
        "frontendRef.ipNetwork": {                                                                                     // 733
          encryption: {                                                                                                // 733
            none: null                                                                                                 // 733
          }                                                                                                            // 733
        }                                                                                                              // 733
      }                                                                                                                // 733
    });                                                                                                                // 733
  });                                                                                                                  // 734
};                                                                                                                     // 735
                                                                                                                       //
function backgroundFillInGrainSizes(db, backend) {                                                                     // 737
  // Fill in sizes for all grains that don't have them. Since computing a grain size requires a                        // 738
  // directory walk, we don't want to do them all at once. Instead, we compute one a second until                      // 739
  // all grains have sizes.                                                                                            // 740
  try {                                                                                                                // 742
    var grain = db.collections.grains.findOne({                                                                        // 743
      size: {                                                                                                          // 743
        $exists: false                                                                                                 // 743
      }                                                                                                                // 743
    }, {                                                                                                               // 743
      fields: {                                                                                                        // 743
        _id: 1,                                                                                                        // 743
        userId: 1                                                                                                      // 743
      }                                                                                                                // 743
    });                                                                                                                // 743
                                                                                                                       //
    if (grain) {                                                                                                       // 745
      // Compute size!                                                                                                 // 746
      try {                                                                                                            // 747
        var result = waitPromise(backend.cap().getGrainStorageUsage(grain.userId, grain._id));                         // 748
        db.collections.grains.update({                                                                                 // 750
          _id: grain._id,                                                                                              // 750
          size: {                                                                                                      // 750
            $exists: false                                                                                             // 750
          }                                                                                                            // 750
        }, {                                                                                                           // 750
          $set: {                                                                                                      // 751
            size: parseInt(result.size)                                                                                // 751
          }                                                                                                            // 751
        });                                                                                                            // 751
      } catch (err) {                                                                                                  // 752
        if (err.kjType === "failed") {                                                                                 // 753
          // Backend had a problem. Maybe the grain doesn't actually exist on disk and the database                    // 754
          // is messed up. We'll set the size to zero and move on.                                                     // 755
          console.error("Error while backfilling grain size for", grain._id, ":", err.stack);                          // 756
          db.collections.grains.update({                                                                               // 757
            _id: grain._id,                                                                                            // 757
            size: {                                                                                                    // 757
              $exists: false                                                                                           // 757
            }                                                                                                          // 757
          }, {                                                                                                         // 757
            $set: {                                                                                                    // 757
              size: 0                                                                                                  // 757
            }                                                                                                          // 757
          });                                                                                                          // 757
        } else {                                                                                                       // 758
          // Rethrow on disconnected / overloaded / unimplemented.                                                     // 759
          throw err;                                                                                                   // 760
        }                                                                                                              // 761
      } // Do another one in a second.                                                                                 // 762
                                                                                                                       //
                                                                                                                       //
      Meteor.setTimeout(backgroundFillInGrainSizes.bind(this, db, backend), 1000);                                     // 765
    }                                                                                                                  // 766
  } catch (err) {                                                                                                      // 767
    // We'll just stop for now, to avoid spamming logs if this error persists. Next time the server                    // 768
    // restarts, the migration will continue.                                                                          // 769
    console.error("Error while backfilling grain sizes:", err.stack);                                                  // 770
  }                                                                                                                    // 771
}                                                                                                                      // 772
                                                                                                                       //
function removeFeatureKeys(db, backend) {                                                                              // 774
  // Remove obsolete data related to the Sandstorm for Work paywall, which was eliminated.                             // 775
  db.notifications.remove({                                                                                            // 777
    "admin.type": "cantRenewFeatureKey"                                                                                // 777
  });                                                                                                                  // 777
  db.notifications.remove({                                                                                            // 778
    "admin.type": "trialFeatureKeyExpired"                                                                             // 778
  });                                                                                                                  // 778
}                                                                                                                      // 779
                                                                                                                       //
function setIpBlacklist(db, backend) {                                                                                 // 781
  if (Meteor.settings.public.isTesting) {                                                                              // 782
    db.collections.settings.insert({                                                                                   // 783
      _id: "ipBlacklist",                                                                                              // 783
      value: "192.168.0.0/16"                                                                                          // 783
    });                                                                                                                // 783
  } else {                                                                                                             // 784
    var defaultIpBlacklist = PRIVATE_IPV4_ADDRESSES.concat(PRIVATE_IPV6_ADDRESSES).join("\n");                         // 785
    db.collections.settings.insert({                                                                                   // 786
      _id: "ipBlacklist",                                                                                              // 786
      value: defaultIpBlacklist                                                                                        // 786
    });                                                                                                                // 786
  }                                                                                                                    // 787
} // This must come after all the functions named within are defined.                                                  // 788
// Only append to this list!  Do not modify or remove list entries;                                                    // 791
// doing so is likely change the meaning and semantics of user databases.                                              // 792
                                                                                                                       //
                                                                                                                       //
var MIGRATIONS = [updateLoginStyleToRedirect, enableLegacyOAuthProvidersIfNotInSettings, denormalizeInviteInfo, mergeRoleAssignmentsIntoApiTokens, fixOasisStorageUsageStats, fetchProfilePictures, assignPlans, removeKeyrings, useLocalizedTextInUserActions, verifyAllPgpSignatures, splitUserIdsIntoAccountIdsAndIdentityIds, appUpdateSettings, moveDevAndEmailLoginDataIntoIdentities, repairEmailIdentityIds, splitAccountUsersAndIdentityUsers, populateContactsFromApiTokens, cleanUpApiTokens, initServerTitleAndReturnAddress, sendReferralNotifications, assignBonuses, splitSmtpUrl, smtpPortShouldBeNumber, consolidateOrgSettings, unsetSmtpDefaultHostnameIfNoUsersExist, extractLastUsedFromApiTokenOwner, setUpstreamTitles, markAllRead, clearAppIndex, assignEmailVerifierIds, setNewServer, addMembraneRequirementsToIdentities, addEncryptionToFrontendRefIpNetwork, setIpBlacklist];
var NEW_SERVER_STARTUP = [startPreinstallingApps];                                                                     // 829
                                                                                                                       //
var migrateToLatest = function (db, backend) {                                                                         // 833
  if (Meteor.settings.replicaNumber) {                                                                                 // 834
    // This is a replica. Wait for the first replica to perform migrations.                                            // 835
    console.log("Waiting for migrations on replica zero...");                                                          // 837
    var done = new Future();                                                                                           // 839
                                                                                                                       //
    var change = function (doc) {                                                                                      // 840
      console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                             // 841
      if (doc.value >= MIGRATIONS.length) done.return();                                                               // 842
    };                                                                                                                 // 843
                                                                                                                       //
    var observer = db.collections.migrations.find({                                                                    // 845
      _id: "migrations_applied"                                                                                        // 845
    }).observe({                                                                                                       // 845
      added: change,                                                                                                   // 846
      changed: change                                                                                                  // 847
    });                                                                                                                // 845
    var newServerDone = new Future();                                                                                  // 850
                                                                                                                       //
    var newServerChange = function (doc) {                                                                             // 851
      if (doc.value) {                                                                                                 // 852
        console.log("New server migrations applied elsewhere");                                                        // 853
        newServerDone.return();                                                                                        // 854
      }                                                                                                                // 855
    };                                                                                                                 // 856
                                                                                                                       //
    var newServerObserver = db.collections.migrations.find({                                                           // 858
      _id: "new_server_migrations_applied"                                                                             // 858
    }).observe({                                                                                                       // 858
      added: newServerChange,                                                                                          // 859
      changed: newServerChange                                                                                         // 860
    });                                                                                                                // 858
    done.wait();                                                                                                       // 863
    observer.stop();                                                                                                   // 864
    newServerDone.wait();                                                                                              // 865
    newServerObserver.stop();                                                                                          // 866
    console.log("Migrations have completed on replica zero.");                                                         // 867
  } else {                                                                                                             // 868
    var applied = db.collections.migrations.findOne({                                                                  // 869
      _id: "migrations_applied"                                                                                        // 869
    });                                                                                                                // 869
    var start = void 0;                                                                                                // 870
                                                                                                                       //
    if (!applied) {                                                                                                    // 871
      // Migrations table is not yet seeded with a value.  This means it has                                           // 872
      // applied 0 migrations.  Persist this.                                                                          // 873
      db.collections.migrations.insert({                                                                               // 874
        _id: "migrations_applied",                                                                                     // 874
        value: 0                                                                                                       // 874
      });                                                                                                              // 874
      start = 0; // This also means this is a brand new server                                                         // 875
                                                                                                                       //
      db.collections.migrations.insert({                                                                               // 878
        _id: "new_server_migrations_applied",                                                                          // 878
        value: false                                                                                                   // 878
      });                                                                                                              // 878
    } else {                                                                                                           // 879
      start = applied.value;                                                                                           // 880
    }                                                                                                                  // 881
                                                                                                                       //
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                                     // 883
                                                                                                                       //
    for (var i = start; i < MIGRATIONS.length; i++) {                                                                  // 885
      // Apply migration i, then record that migration i was successfully run.                                         // 886
      console.log("Applying migration " + (i + 1));                                                                    // 887
      MIGRATIONS[i](db, backend);                                                                                      // 888
      db.collections.migrations.update({                                                                               // 889
        _id: "migrations_applied"                                                                                      // 889
      }, {                                                                                                             // 889
        $set: {                                                                                                        // 889
          value: i + 1                                                                                                 // 889
        }                                                                                                              // 889
      });                                                                                                              // 889
      console.log("Applied migration " + (i + 1));                                                                     // 890
    }                                                                                                                  // 891
                                                                                                                       //
    if (!db.collections.migrations.findOne({                                                                           // 893
      _id: "new_server_migrations_applied"                                                                             // 893
    }).value) {                                                                                                        // 893
      // new_server_migrations_applied is guaranteed to exist since we have a migration that                           // 894
      // ensures it.                                                                                                   // 895
      for (var _i = 0; _i < NEW_SERVER_STARTUP.length; _i++) {                                                         // 896
        console.log("Running new server startup function " + (_i + 1));                                                // 897
                                                                                                                       //
        NEW_SERVER_STARTUP[_i](db, backend);                                                                           // 898
                                                                                                                       //
        console.log("Running new server startup function " + (_i + 1));                                                // 899
      }                                                                                                                // 900
                                                                                                                       //
      db.collections.migrations.update({                                                                               // 902
        _id: "new_server_migrations_applied"                                                                           // 902
      }, {                                                                                                             // 902
        $set: {                                                                                                        // 902
          value: true                                                                                                  // 902
        }                                                                                                              // 902
      });                                                                                                              // 902
    } // Start background migrations.                                                                                  // 903
                                                                                                                       //
                                                                                                                       //
    backgroundFillInGrainSizes(db, backend);                                                                           // 906
  }                                                                                                                    // 907
};                                                                                                                     // 908
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"networking.js":["babel-runtime/helpers/typeof","/imports/constants.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/networking.js                                                                                        //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _typeof2 = require("babel-runtime/helpers/typeof");                                                                //
                                                                                                                       //
var _typeof3 = _interopRequireDefault(_typeof2);                                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  ssrfSafeLookup: function () {                                                                                        // 1
    return ssrfSafeLookup;                                                                                             // 1
  },                                                                                                                   // 1
  ssrfSafeLookupOrProxy: function () {                                                                                 // 1
    return ssrfSafeLookupOrProxy;                                                                                      // 1
  },                                                                                                                   // 1
  monkeyPatchHttp: function () {                                                                                       // 1
    return monkeyPatchHttp;                                                                                            // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var SPECIAL_IPV4_ADDRESSES = void 0,                                                                                   // 1
    SPECIAL_IPV6_ADDRESSES = void 0;                                                                                   // 1
module.import("/imports/constants.js", {                                                                               // 1
  "SPECIAL_IPV4_ADDRESSES": function (v) {                                                                             // 1
    SPECIAL_IPV4_ADDRESSES = v;                                                                                        // 1
  },                                                                                                                   // 1
  "SPECIAL_IPV6_ADDRESSES": function (v) {                                                                             // 1
    SPECIAL_IPV6_ADDRESSES = v;                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Dns = Npm.require("dns");                                                                                          // 19
                                                                                                                       //
var Ip = Npm.require("ip");                                                                                            // 20
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 21
                                                                                                                       //
var lookupInFiber = Meteor.wrapAsync(Dns.lookup, Dns);                                                                 // 23
                                                                                                                       //
function getNetwork(addr, bits) {                                                                                      // 25
  // npm ip's "mask" and "cidr" functions are broken for ipv6. :(                                                      // 26
  var parsed = Ip.toBuffer(addr);                                                                                      // 28
                                                                                                                       //
  for (var i = Math.ceil(bits / 8); i < parsed.length; i++) {                                                          // 30
    parsed[i] = 0;                                                                                                     // 31
  }                                                                                                                    // 32
                                                                                                                       //
  var n = Math.floor(bits / 8);                                                                                        // 34
                                                                                                                       //
  if (n < parsed.length) {                                                                                             // 35
    parsed[n] = parsed[n] & 0xff << 8 - bits % 8;                                                                      // 36
  }                                                                                                                    // 37
                                                                                                                       //
  return parsed;                                                                                                       // 39
}                                                                                                                      // 40
                                                                                                                       //
function parseCidr(cidr) {                                                                                             // 42
  // Given a CIDR-format network specification, return a function which, given an address string,                      // 43
  // returns true if the address is in this network.                                                                   // 44
  // The "ip" NPM module's CIDR handling unfortunately is very broken for IPv6. Many bugs have been                    // 46
  // filed but it remains broken.                                                                                      // 47
  try {                                                                                                                // 49
    cidr = cidr.trim();                                                                                                // 50
    if (cidr === "") return null;                                                                                      // 51
    var parts = cidr.split("/");                                                                                       // 53
                                                                                                                       //
    if (parts.length === 1) {                                                                                          // 54
      // Bare address.                                                                                                 // 55
      return function (addr) {                                                                                         // 56
        return Ip.isEqual(cidr, addr);                                                                                 // 56
      };                                                                                                               // 56
    } else if (parts.length === 2) {                                                                                   // 57
      var bits = parseInt(parts[1], 10);                                                                               // 58
      if (bits !== bits) throw new Error("value after slash must be an integer");                                      // 59
      var network = getNetwork(parts[0], bits);                                                                        // 60
      return function (addr) {                                                                                         // 61
        return network.equals(getNetwork(addr, bits));                                                                 // 62
      };                                                                                                               // 63
    } else {                                                                                                           // 64
      throw new Error("too many slashes");                                                                             // 65
    }                                                                                                                  // 66
  } catch (err) {                                                                                                      // 67
    console.error("invalid network specification in IP blacklist:", cidr, err);                                        // 68
    return null;                                                                                                       // 69
  }                                                                                                                    // 70
}                                                                                                                      // 71
                                                                                                                       //
SPECIAL_FILTERS = SPECIAL_IPV4_ADDRESSES.concat(SPECIAL_IPV6_ADDRESSES).map(parseCidr);                                // 73
                                                                                                                       //
function ssrfSafeLookup(db, url) {                                                                                     // 75
  // Given an HTTP/HTTPS URL, look up the hostname, verify it doesn't point to a blacklisted IP,                       // 76
  // then return an object of {url, host}, where `url` has the original hostname substituted with                      // 77
  // an IP address, and `host` is the original hostname suitable for sending in the `Host` header.                     // 78
  var parsedUrl = Url.parse(url);                                                                                      // 80
                                                                                                                       //
  if (parsedUrl.protocol !== "http:" && parsedUrl.protocol !== "https:") {                                             // 82
    throw new Error("not an HTTP nor HTTPS URL: " + url);                                                              // 83
  }                                                                                                                    // 84
                                                                                                                       //
  var addresses = lookupInFiber(parsedUrl.hostname, {                                                                  // 86
    all: true,                                                                                                         // 86
    hints: Dns.ADDRCONFIG                                                                                              // 86
  }); // TODO(perf): Subscribe to blacklist changes so that we don't have to do a new lookup and                       // 86
  //   parse each time.                                                                                                // 89
                                                                                                                       //
  var blacklist = db.getSettingWithFallback("ipBlacklist", "").split("\n").map(parseCidr).filter(function (x) {        // 90
    return x;                                                                                                          // 91
  });                                                                                                                  // 91
                                                                                                                       //
  var _loop = function (i) {                                                                                           // 75
    var address = addresses[i];                                                                                        // 94
    if (address.family !== 4 && address.family !== 6) return "continue";                                               // 95
    var ok = true;                                                                                                     // 97
    blacklist.forEach(function (test) {                                                                                // 98
      if (test(address.address)) {                                                                                     // 98
        ok = false;                                                                                                    // 98
      }                                                                                                                // 98
    });                                                                                                                // 98
    SPECIAL_FILTERS.forEach(function (test) {                                                                          // 100
      if (test(address.address)) {                                                                                     // 100
        ok = false;                                                                                                    // 100
      }                                                                                                                // 100
    });                                                                                                                // 100
                                                                                                                       //
    if (ok) {                                                                                                          // 102
      var host = parsedUrl.host;                                                                                       // 103
      delete parsedUrl.host;                                                                                           // 104
      parsedUrl.hostname = address.address;                                                                            // 105
      return {                                                                                                         // 106
        v: {                                                                                                           // 106
          url: Url.format(parsedUrl),                                                                                  // 106
          host: host                                                                                                   // 106
        }                                                                                                              // 106
      };                                                                                                               // 106
    }                                                                                                                  // 107
  };                                                                                                                   // 75
                                                                                                                       //
  for (var i in meteorBabelHelpers.sanitizeForInObject(addresses)) {                                                   // 93
    var _ret = _loop(i);                                                                                               // 93
                                                                                                                       //
    switch (_ret) {                                                                                                    // 93
      case "continue":                                                                                                 // 93
        continue;                                                                                                      // 95
                                                                                                                       //
      default:                                                                                                         // 93
        if ((typeof _ret === "undefined" ? "undefined" : (0, _typeof3.default)(_ret)) === "object") return _ret.v;     // 93
    }                                                                                                                  // 93
  }                                                                                                                    // 108
                                                                                                                       //
  if (addresses.length > 0) {                                                                                          // 110
    throw new Meteor.Error(403, "can't connect to blacklisted private network address: " + parsedUrl.hostname + "; the Sandstorm server admin can change the blacklist in the admin settings");
  } else {                                                                                                             // 114
    throw new Meteor.Error(404, "host not found: " + parsedUrl.hostname);                                              // 115
  }                                                                                                                    // 116
}                                                                                                                      // 117
                                                                                                                       //
function ssrfSafeLookupOrProxy(db, url) {                                                                              // 119
  // If there is an HTTP proxy, then it will have to do the work of blacklisting IPs, because it's                     // 120
  // the proxy that does the DNS lookup.                                                                               // 121
  var httpProxy = process.env.HTTP_PROXY || process.env.http_proxy;                                                    // 122
  var httpsProxy = process.env.HTTPS_PROXY || process.env.https_proxy;                                                 // 123
                                                                                                                       //
  if (httpProxy && url.startsWith("http:")) {                                                                          // 125
    return {                                                                                                           // 126
      proxy: httpProxy                                                                                                 // 126
    };                                                                                                                 // 126
  } else if (httpsProxy && url.startsWith("https:")) {                                                                 // 127
    return {                                                                                                           // 128
      proxy: httpsProxy                                                                                                // 128
    };                                                                                                                 // 128
  } else {                                                                                                             // 129
    return ssrfSafeLookup(db, url);                                                                                    // 130
  }                                                                                                                    // 131
}                                                                                                                      // 132
                                                                                                                       //
function ssrfSafeHttp(originalHttpCall, db, method, url, options, callback) {                                          // 134
  if (typeof options === "function") {                                                                                 // 135
    callback = options;                                                                                                // 136
    options = undefined;                                                                                               // 137
  }                                                                                                                    // 138
                                                                                                                       //
  if (!options) options = {};                                                                                          // 140
                                                                                                                       //
  if (options.npmRequestOptions && options.npmRequestOptions.proxy) {                                                  // 142
    // Request already specifies a different proxy.                                                                    // 143
    return originalHttpCall(method, url, options, callback);                                                           // 144
  }                                                                                                                    // 145
                                                                                                                       //
  var safe = ssrfSafeLookupOrProxy(db, url);                                                                           // 147
                                                                                                                       //
  if (safe.proxy) {                                                                                                    // 149
    if (!options.npmRequestOptions) options.npmRequestOptions = {};                                                    // 150
    options.npmRequestOptions.proxy = safe.proxy;                                                                      // 151
    return originalHttpCall(method, url, options, callback);                                                           // 152
  } else {                                                                                                             // 153
    var _safe = ssrfSafeLookup(db, url);                                                                               // 154
                                                                                                                       //
    if (!options.headers) options.headers = {};                                                                        // 155
    options.headers.host = _safe.host;                                                                                 // 156
    options.servername = _safe.host.split(":")[0];                                                                     // 157
    return originalHttpCall(method, _safe.url, options, callback);                                                     // 158
  }                                                                                                                    // 159
}                                                                                                                      // 160
                                                                                                                       //
function monkeyPatchHttp(db, HTTP) {                                                                                   // 162
  var original = HTTP.call.bind(HTTP);                                                                                 // 163
  HTTP.call = ssrfSafeHttp.bind(this, original, db);                                                                   // 164
}                                                                                                                      // 165
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"observe-helpers.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/observe-helpers.js                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  fillUndefinedForChangedDoc: function () {                                                                            // 1
    return fillUndefinedForChangedDoc;                                                                                 // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
function fillUndefinedForChangedDoc(newDoc, oldDoc) {                                                                  // 17
  // In publishes, setting a field to `undefined` specifies that it is no longer present.                              // 18
  // Annoyingly, in the changed callback of observing a collection, it does not do this for you.                       // 19
  Object.keys(oldDoc).forEach(function (key) {                                                                         // 20
    if (!(key in newDoc)) {                                                                                            // 21
      newDoc[key] = undefined;                                                                                         // 22
    }                                                                                                                  // 23
  });                                                                                                                  // 24
}                                                                                                                      // 25
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"persistent.js":["babel-runtime/helpers/classCallCheck","/imports/server/async-helpers.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/persistent.js                                                                                        //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  PersistentImpl: function () {                                                                                        // 1
    return PersistentImpl;                                                                                             // 1
  },                                                                                                                   // 1
  hashSturdyRef: function () {                                                                                         // 1
    return hashSturdyRef;                                                                                              // 1
  },                                                                                                                   // 1
  generateSturdyRef: function () {                                                                                     // 1
    return generateSturdyRef;                                                                                          // 1
  },                                                                                                                   // 1
  checkRequirements: function () {                                                                                     // 1
    return checkRequirements;                                                                                          // 1
  },                                                                                                                   // 1
  fetchApiToken: function () {                                                                                         // 1
    return fetchApiToken;                                                                                              // 1
  },                                                                                                                   // 1
  insertApiToken: function () {                                                                                        // 1
    return insertApiToken;                                                                                             // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
var inMeteor = void 0;                                                                                                 // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 19
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 20
                                                                                                                       //
var privateDb = Symbol("PersistentImpl.db");                                                                           // 22
var privateTemplate = Symbol("PersistentImpl.template");                                                               // 23
var privateIsSaved = Symbol("PersistentImpl.isSaved");                                                                 // 24
                                                                                                                       //
var PersistentImpl = function () {                                                                                     //
  function PersistentImpl(db, saveTemplate) {                                                                          // 27
    (0, _classCallCheck3.default)(this, PersistentImpl);                                                               // 27
    this[privateDb] = db;                                                                                              // 28
    this[privateTemplate] = saveTemplate; // TODO(someday): This may or may not be quite right in cases where we're saving a root token
    //   by copy. Is the copy the "same" capability or not? Today, isSaved() is only used by                           // 32
    //   wakelock notification handles which are not issued through the powerbox, therefore do not                     // 33
    //   go through claimRequest(), therefore are not saved by copy, so the point is moot.                             // 34
                                                                                                                       //
    this[privateIsSaved] = !!saveTemplate.parentToken;                                                                 // 35
  }                                                                                                                    // 36
                                                                                                                       //
  PersistentImpl.prototype.isSaved = function () {                                                                     //
    function isSaved() {                                                                                               //
      // Can be called by subclasses to ask whether or not a record currently exists in the ApiTokens                  // 39
      // table representing this capability.                                                                           // 40
      return this[privateIsSaved];                                                                                     // 42
    }                                                                                                                  // 43
                                                                                                                       //
    return isSaved;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  PersistentImpl.prototype.save = function () {                                                                        //
    function save(params) {                                                                                            //
      var _this = this;                                                                                                // 45
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 46
        if (!params.sealFor) {                                                                                         // 47
          throw new Error("must specify 'sealFor'");                                                                   // 48
        }                                                                                                              // 49
                                                                                                                       //
        var db = _this[privateDb];                                                                                     // 51
                                                                                                                       //
        var newToken = _.clone(_this[privateTemplate]);                                                                // 53
                                                                                                                       //
        newToken.owner = params.sealFor;                                                                               // 54
                                                                                                                       //
        if (newToken.owner.user) {                                                                                     // 55
          if (!newToken.identityId) {                                                                                  // 56
            throw new Error("can't save non-UiView with user as owner");                                               // 57
          } // Only "identityId" and "title" are allowed to be passed to save().                                       // 58
                                                                                                                       //
                                                                                                                       //
          var userOwner = _.pick(newToken.owner.user, "identityId", "title"); // Fill in denormalizedGrainMetadata and upstreamTitle ourselves.
                                                                                                                       //
                                                                                                                       //
          userOwner.denormalizedGrainMetadata = db.getDenormalizedGrainInfo(newToken.grainId);                         // 64
          var grain = db.getGrain(newToken.grainId);                                                                   // 66
                                                                                                                       //
          if (grain && grain.title !== userOwner.title) {                                                              // 67
            userOwner.upstreamTitle = grain.title;                                                                     // 68
          }                                                                                                            // 69
                                                                                                                       //
          newToken.owner.user = userOwner;                                                                             // 71
        }                                                                                                              // 72
                                                                                                                       //
        newToken.created = new Date();                                                                                 // 74
        var sturdyRef = insertApiToken(db, newToken);                                                                  // 75
        _this[privateIsSaved] = true;                                                                                  // 77
        return {                                                                                                       // 78
          sturdyRef: new Buffer(sturdyRef)                                                                             // 78
        };                                                                                                             // 78
      });                                                                                                              // 79
    }                                                                                                                  // 80
                                                                                                                       //
    return save;                                                                                                       //
  }(); // TODO(someday): Implement SystemPersistent.addRequirements().                                                 //
                                                                                                                       //
                                                                                                                       //
  return PersistentImpl;                                                                                               //
}();                                                                                                                   //
                                                                                                                       //
function hashSturdyRef(sturdyRef) {                                                                                    // 85
  return Crypto.createHash("sha256").update(sturdyRef).digest("base64");                                               // 86
}                                                                                                                      // 87
                                                                                                                       //
;                                                                                                                      // 87
                                                                                                                       //
function generateSturdyRef() {                                                                                         // 89
  return Random.secret();                                                                                              // 90
}                                                                                                                      // 91
                                                                                                                       //
function cryptApiToken(key, entry, cryptIn, cryptOut) {                                                                // 93
  // Encrypts or decrypts all fields of an ApiToken.                                                                   // 94
  // `cryptIn` translates a token in the input to a buffer.                                                            // 95
  // `cryptOut` translates a buffer to a token in the output.                                                          // 96
  check(key, String);                                                                                                  // 98
  check(entry, Object);                                                                                                // 99
  var nonce0 = new Buffer(8);                                                                                          // 101
  nonce0.fill(0);                                                                                                      // 102
  var keyBuf = new Buffer(32);                                                                                         // 104
  keyBuf.fill(0);                                                                                                      // 105
  keyBuf.write(key, 0, 32, "base64");                                                                                  // 106
                                                                                                                       //
  function encrypt0(token) {                                                                                           // 108
    return cryptOut(Capnp.chacha20(cryptIn(token), nonce0, keyBuf));                                                   // 109
  }                                                                                                                    // 110
                                                                                                                       //
  if (entry.parentTokenKey) {                                                                                          // 112
    entry.parentTokenKey = encrypt0(entry.parentTokenKey);                                                             // 113
  } else if (entry.frontendRef && entry.frontendRef.http) {                                                            // 114
    var http = entry.frontendRef.http;                                                                                 // 115
                                                                                                                       //
    if (http.auth) {                                                                                                   // 116
      var auth = http.auth;                                                                                            // 117
                                                                                                                       //
      if (auth.bearer) {                                                                                               // 118
        auth.bearer = encrypt0(auth.bearer);                                                                           // 119
      } else if (auth.basic) {                                                                                         // 120
        auth.basic.password = encrypt0(auth.basic.password);                                                           // 121
      } else if (auth.refresh) {                                                                                       // 122
        auth.refresh = encrypt0(auth.refresh);                                                                         // 123
      }                                                                                                                // 124
    }                                                                                                                  // 125
  }                                                                                                                    // 126
}                                                                                                                      // 127
                                                                                                                       //
function fetchApiToken(db, key, moreQuery) {                                                                           // 129
  // Reads an ApiToken from the database and decrypts its encrypted fields.                                            // 130
  function bufferToString(buf) {                                                                                       // 132
    // un-pad short secrets                                                                                            // 133
    var size = buf.length;                                                                                             // 134
                                                                                                                       //
    while (size > 0 && buf[size - 1] == 0) {                                                                           // 135
      --size;                                                                                                          // 136
    }                                                                                                                  // 137
                                                                                                                       //
    return buf.slice(0, size).toString("utf8");                                                                        // 139
  }                                                                                                                    // 140
                                                                                                                       //
  var query = {                                                                                                        // 142
    _id: hashSturdyRef(key)                                                                                            // 142
  };                                                                                                                   // 142
  Object.assign(query, moreQuery || {});                                                                               // 143
  var entry = db.collections.apiTokens.findOne(query);                                                                 // 144
                                                                                                                       //
  if (entry) {                                                                                                         // 145
    cryptApiToken(key, entry, function (x) {                                                                           // 146
      return x;                                                                                                        // 146
    }, bufferToString);                                                                                                // 146
  }                                                                                                                    // 147
                                                                                                                       //
  return entry;                                                                                                        // 149
}                                                                                                                      // 150
                                                                                                                       //
function insertApiToken(db, entry, key) {                                                                              // 152
  // Adds a new ApiToken to the database. `key`, if specified, *must* be a base64-encoded 256-bit                      // 153
  // value. If omitted, a key will be generated. Either way, the key is returned, and entry._id                        // 154
  // is filled in. Also, as a side effect, some fields of `entry` will become encrypted, but                           // 155
  // ideally callers should not depend on this behavior.                                                               // 156
  function stringToBuffer(str) {                                                                                       // 158
    var buf = new Buffer(str, "utf8");                                                                                 // 159
    if (buf.length >= 32) return buf;                                                                                  // 160
    var padded = new Buffer(32);                                                                                       // 162
    padded.fill(0);                                                                                                    // 163
    buf.copy(padded);                                                                                                  // 164
    return padded;                                                                                                     // 165
  }                                                                                                                    // 166
                                                                                                                       //
  if (!key) key = generateSturdyRef();                                                                                 // 168
  entry._id = hashSturdyRef(key);                                                                                      // 169
  cryptApiToken(key, entry, stringToBuffer, function (x) {                                                             // 170
    return x;                                                                                                          // 170
  });                                                                                                                  // 170
  db.collections.apiTokens.insert(entry);                                                                              // 171
  return key;                                                                                                          // 172
}                                                                                                                      // 173
                                                                                                                       //
function checkRequirements(db, requirements) {                                                                         // 175
  // Checks if the given list of MembraneRequirements are all satisfied, returning true if so and                      // 176
  // false otherwise.                                                                                                  // 177
  // TODO(cleanup): SandstormPermissions has a different way of checking the same requirements.                        // 179
  //   Reuse?                                                                                                          // 180
  // TODO(security): Eventually we want checkRequirements(), when it passes, to produce an observer                    // 182
  //   that can be used to receive a notification when the requirements may no longer be satisfied,                    // 183
  //   in order to revoke the live object. See RequirementObserver in supervisor.capnp.                                // 184
  if (!requirements) {                                                                                                 // 186
    return true;                                                                                                       // 187
  }                                                                                                                    // 188
                                                                                                                       //
  requirements.forEach(function (requirement) {                                                                        // 190
    if (requirement.tokenValid) {                                                                                      // 191
      var token = db.collections.apiTokens.findOne({                                                                   // 192
        _id: requirement.tokenValid,                                                                                   // 193
        revoked: {                                                                                                     // 193
          $ne: true                                                                                                    // 193
        }                                                                                                              // 193
      }, {                                                                                                             // 193
        fields: {                                                                                                      // 194
          requirements: 1                                                                                              // 194
        }                                                                                                              // 194
      });                                                                                                              // 194
                                                                                                                       //
      if (!token) {                                                                                                    // 195
        throw new Meteor.Error(403, "Capability revoked because the link through which it was introduced has been " + "revoked or deleted.");
      }                                                                                                                // 199
                                                                                                                       //
      checkRequirements(db, token.requirements);                                                                       // 201
                                                                                                                       //
      if (token.parentToken) {                                                                                         // 203
        checkRequirements(db, [{                                                                                       // 204
          tokenValid: token.parentToken                                                                                // 204
        }]);                                                                                                           // 204
      }                                                                                                                // 205
    } else if (requirement.permissionsHeld) {                                                                          // 206
      var p = requirement.permissionsHeld;                                                                             // 207
      var viewInfo = db.collections.grains.findOne(p.grainId, {                                                        // 208
        fields: {                                                                                                      // 209
          cachedViewInfo: 1                                                                                            // 209
        }                                                                                                              // 209
      }).cachedViewInfo;                                                                                               // 209
      var vertex = void 0;                                                                                             // 210
                                                                                                                       //
      if (p.identityId) {                                                                                              // 211
        vertex = {                                                                                                     // 212
          grain: {                                                                                                     // 212
            _id: p.grainId,                                                                                            // 212
            identityId: p.identityId                                                                                   // 212
          }                                                                                                            // 212
        };                                                                                                             // 212
      } else {                                                                                                         // 213
        vertex = {                                                                                                     // 214
          token: {                                                                                                     // 214
            _id: p.tokenId,                                                                                            // 214
            grainId: p.grainId                                                                                         // 214
          }                                                                                                            // 214
        };                                                                                                             // 214
      }                                                                                                                // 215
                                                                                                                       //
      var currentPermissions = SandstormPermissions.grainPermissions(db, vertex, viewInfo || {}).permissions;          // 217
                                                                                                                       //
      if (!currentPermissions) {                                                                                       // 219
        throw new Meteor.Error(403, "Capability revoked because a user involved in introducing it no longer has " + "the necessary permissions.");
      }                                                                                                                // 223
                                                                                                                       //
      var requiredPermissions = p.permissions || [];                                                                   // 225
                                                                                                                       //
      for (var ii = 0; ii < requiredPermissions.length; ++ii) {                                                        // 226
        if (requiredPermissions[ii] && !currentPermissions[ii]) {                                                      // 227
          throw new Meteor.Error(403, "Capability revoked because a user involved in introducing it no longer has " + "the necessary permissions.");
        }                                                                                                              // 231
      }                                                                                                                // 232
    } else if (requirement.userIsAdmin) {                                                                              // 233
      if (!db.isAdminById(requirement.userIsAdmin)) {                                                                  // 234
        throw new Meteor.Error(403, "Capability revoked because the user who created it has lost their admin " + "rights.");
      }                                                                                                                // 238
    } else {                                                                                                           // 239
      throw new Meteor.Error(403, "Unknown requirement type.");                                                        // 240
    }                                                                                                                  // 241
  });                                                                                                                  // 242
}                                                                                                                      // 243
                                                                                                                       //
;                                                                                                                      // 243
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"static-asset.js":["babel-runtime/helpers/classCallCheck",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/server/static-asset.js                                                                                      //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
module.export({                                                                                                        // 1
  StaticAssetImpl: function () {                                                                                       // 1
    return StaticAssetImpl;                                                                                            // 1
  },                                                                                                                   // 1
  IdenticonStaticAssetImpl: function () {                                                                              // 1
    return IdenticonStaticAssetImpl;                                                                                   // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014-2016 Sandstorm Development Group, Inc. and contributors                                          // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Capnp = Npm.require("capnp");                                                                                      // 17
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 18
                                                                                                                       //
var StaticAsset = Capnp.importSystem("sandstorm/grain.capnp").StaticAsset;                                             // 19
var PROTOCOL = Url.parse(process.env.ROOT_URL).protocol;                                                               // 21
                                                                                                                       //
var StaticAssetImpl = function () {                                                                                    //
  function StaticAssetImpl(assetId) {                                                                                  // 24
    (0, _classCallCheck3.default)(this, StaticAssetImpl);                                                              // 24
    check(assetId, String);                                                                                            // 25
    this._protocol = PROTOCOL.slice(0, -1);                                                                            // 26
    this._hostPath = makeWildcardHost("static") + "/" + assetId;                                                       // 27
  }                                                                                                                    // 28
                                                                                                                       //
  StaticAssetImpl.prototype.getUrl = function () {                                                                     //
    function getUrl() {                                                                                                //
      return {                                                                                                         // 31
        protocol: this._protocol,                                                                                      // 31
        hostPath: this._hostPath                                                                                       // 31
      };                                                                                                               // 31
    }                                                                                                                  // 32
                                                                                                                       //
    return getUrl;                                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  return StaticAssetImpl;                                                                                              //
}();                                                                                                                   //
                                                                                                                       //
;                                                                                                                      // 33
                                                                                                                       //
var IdenticonStaticAssetImpl = function () {                                                                           //
  function IdenticonStaticAssetImpl(hash, size) {                                                                      // 36
    (0, _classCallCheck3.default)(this, IdenticonStaticAssetImpl);                                                     // 36
    check(hash, String);                                                                                               // 37
    check(size, Match.Integer);                                                                                        // 38
    this._protocol = PROTOCOL.slice(0, -1);                                                                            // 39
    this._hostPath = makeWildcardHost("static") + "/identicon/" + hash + "?s=" + size;                                 // 40
  }                                                                                                                    // 41
                                                                                                                       //
  IdenticonStaticAssetImpl.prototype.getUrl = function () {                                                            //
    function getUrl() {                                                                                                //
      return {                                                                                                         // 44
        protocol: this._protocol,                                                                                      // 44
        hostPath: this._hostPath                                                                                       // 44
      };                                                                                                               // 44
    }                                                                                                                  // 45
                                                                                                                       //
    return getUrl;                                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  return IdenticonStaticAssetImpl;                                                                                     //
}();                                                                                                                   //
                                                                                                                       //
;                                                                                                                      // 46
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"constants.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/constants.js                                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
  ACCOUNT_DELETION_SUSPENSION_TIME: function () {                                                                      // 1
    return ACCOUNT_DELETION_SUSPENSION_TIME;                                                                           // 1
  },                                                                                                                   // 1
  PRIVATE_IPV4_ADDRESSES: function () {                                                                                // 1
    return PRIVATE_IPV4_ADDRESSES;                                                                                     // 1
  },                                                                                                                   // 1
  PRIVATE_IPV6_ADDRESSES: function () {                                                                                // 1
    return PRIVATE_IPV6_ADDRESSES;                                                                                     // 1
  },                                                                                                                   // 1
  SPECIAL_IPV4_ADDRESSES: function () {                                                                                // 1
    return SPECIAL_IPV4_ADDRESSES;                                                                                     // 1
  },                                                                                                                   // 1
  SPECIAL_IPV6_ADDRESSES: function () {                                                                                // 1
    return SPECIAL_IPV6_ADDRESSES;                                                                                     // 1
  }                                                                                                                    // 1
});                                                                                                                    // 1
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var ACCOUNT_DELETION_SUSPENSION_TIME = 7 * 60 * 60 * 24 * 1000; // 7 days in ms                                        // 17
// Lists below developed from RFC6890, which is an overview of all special addresses.                                  // 19
                                                                                                                       //
var PRIVATE_IPV4_ADDRESSES = ["10.0.0.0/8", // RFC1918 reserved for internal network                                   // 20
"127.0.0.0/8", // RFC1122 loopback / localhost                                                                         // 22
"169.254.0.0/16", // RFC3927 "link local" (auto-configured LAN in absence of DHCP)                                     // 23
"172.16.0.0/12", // RFC1918 reserved for internal network                                                              // 24
"192.168.0.0/16"];                                                                                                     // 25
var PRIVATE_IPV6_ADDRESSES = ["::1/128", // RFC4291 loopback / localhost                                               // 28
"fc00::/7", // RFC4193 unique private network                                                                          // 30
"fe80::/10"];                                                                                                          // 31
var SPECIAL_IPV4_ADDRESSES = ["0.0.0.0/8", // RFC1122 "this host" / wildcard                                           // 34
"100.64.0.0/10", // RFC6598 "shared address space" for carrier-grade NAT                                               // 36
"192.0.0.0/24", // RFC6890 reserved for special protocols                                                              // 37
"192.0.2.0/24", // RFC5737 "example address" block 1 -- like example.com for IPs                                       // 38
"192.88.99.0/24", // RFC3068 6to4 relay                                                                                // 39
"198.18.0.0/15", // RFC2544 standard benchmarks                                                                        // 40
"198.51.100.0/24", // RFC5737 "example address" block 2 -- like example.com for IPs                                    // 41
"203.0.113.0/24", // RFC5737 "example address" block 3 -- like example.com for IPs                                     // 42
"224.0.0.0/4", // RFC1112 multicast                                                                                    // 43
"240.0.0.0/4", // RFC1112 multicast / reserved for future use                                                          // 44
"255.255.255.255/32"];                                                                                                 // 45
var SPECIAL_IPV6_ADDRESSES = ["::/128", // RFC4291 unspecified address / wildcard                                      // 48
"64:ff9b::/96", // RFC6052 IPv4-IPv6 translation                                                                       // 50
"::ffff:0:0/96", // RFC4291 IPv4-mapped address                                                                        // 51
// TODO(someday): I don't understand the difference between the above                                                  // 52
//     two. Both are described as mapping ip4 addresses into the ip6                                                   // 53
//     space. Perhaps this should be allowed, however, we'd need to                                                    // 54
//     filter the ip4 address against the ip4 blacklist, so special                                                    // 55
//     handling would be needed.                                                                                       // 56
"100::/64", // RFC6666 discard-only address block                                                                      // 57
"2001::/23", // RFC2928 reserved for special protocols                                                                 // 58
"2001:2::/48", // RFC5180 standard benchmarks                                                                          // 59
"2001:db8::/32", // RFC3849 "example address" block -- like example.com for IPs                                        // 60
"2001:10::/28", // RFC4843 ORCHID                                                                                      // 61
"2002::/16", // RFC3056 6to4 relay                                                                                     // 62
"ff00::/8"];                                                                                                           // 63
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"db-deprecated.js":["/imports/server/accounts/ldap.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/db-deprecated.js                                                                                            //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file imports symbols from the sandstorm-db package. New code should use SandstormDb                            // 17
// directly.                                                                                                           // 18
// TODO(cleanup): Over time, eliminate the use of each of these assignments by using `globalDb`                        // 20
//   directly. For collections, prefer to update SandstormDb to provide methods for querying the                       // 21
//   collection rather than use `globalDb.collections` directly. When code is moved into packages,                     // 22
//   the `globalDb` global should NOT go with it; the package should expect a SandstormDb to be                        // 23
//   passed in, thus allowing mocking the database for unit tests.                                                     // 24
var quotaManager = void 0;                                                                                             // 26
                                                                                                                       //
if (Meteor.isServer) {                                                                                                 // 27
  var LDAP = void 0;                                                                                                   // 1
  module.import("/imports/server/accounts/ldap.js", {                                                                  // 1
    "LDAP": function (v) {                                                                                             // 1
      LDAP = v;                                                                                                        // 1
    }                                                                                                                  // 1
  }, 0);                                                                                                               // 1
  // Imports are usually not allowed to occur in a block. However, it is the only way to do                            // 29
  // this under Meteor. Using // jscs:disable doesn't work for what it considers syntax violations,                    // 30
  // and so we've added this file to .jscsrc's excludedFiles explicitly.                                               // 31
  quotaManager = new LDAP();                                                                                           // 33
} else {                                                                                                               // 34
  quotaManager = {                                                                                                     // 35
    updateUserQuota: function (db, user) {                                                                             // 36
      return {                                                                                                         // 37
        storage: user.cachedStorageQuota || 0,                                                                         // 38
        grains: Infinity,                                                                                              // 39
        compute: Infinity                                                                                              // 40
      };                                                                                                               // 37
    }                                                                                                                  // 42
  };                                                                                                                   // 35
}                                                                                                                      // 44
                                                                                                                       //
globalDb = new SandstormDb(quotaManager);                                                                              // 46
Packages = globalDb.collections.packages;                                                                              // 48
DevPackages = globalDb.collections.devPackages;                                                                        // 49
UserActions = globalDb.collections.userActions;                                                                        // 50
Grains = globalDb.collections.grains;                                                                                  // 51
Contacts = globalDb.collections.contacts;                                                                              // 52
Sessions = globalDb.collections.sessions;                                                                              // 53
SignupKeys = globalDb.collections.signupKeys;                                                                          // 54
ActivityStats = globalDb.collections.activityStats;                                                                    // 55
DeleteStats = globalDb.collections.deleteStats;                                                                        // 56
FileTokens = globalDb.collections.fileTokens;                                                                          // 57
ApiTokens = globalDb.collections.apiTokens;                                                                            // 58
Notifications = globalDb.collections.notifications;                                                                    // 59
StatsTokens = globalDb.collections.statsTokens;                                                                        // 60
Misc = globalDb.collections.misc;                                                                                      // 61
Settings = globalDb.collections.settings;                                                                              // 62
currentUserGrains = globalDb.currentUserGrains.bind(globalDb);                                                         // 64
isDemoUser = globalDb.isDemoUser.bind(globalDb);                                                                       // 65
isSignedUp = globalDb.isSignedUp.bind(globalDb);                                                                       // 66
isSignedUpOrDemo = globalDb.isSignedUpOrDemo.bind(globalDb);                                                           // 67
isUserOverQuota = globalDb.isUserOverQuota.bind(globalDb);                                                             // 68
isUserExcessivelyOverQuota = globalDb.isUserExcessivelyOverQuota.bind(globalDb);                                       // 69
isAdmin = globalDb.isAdmin.bind(globalDb);                                                                             // 70
isAdminById = globalDb.isAdminById.bind(globalDb);                                                                     // 71
findAdminUserForToken = globalDb.findAdminUserForToken.bind(globalDb);                                                 // 72
matchWildcardHost = globalDb.matchWildcardHost.bind(globalDb);                                                         // 73
makeWildcardHost = globalDb.makeWildcardHost.bind(globalDb);                                                           // 74
allowDevAccounts = globalDb.allowDevAccounts.bind(globalDb);                                                           // 75
roleAssignmentPattern = globalDb.roleAssignmentPattern;                                                                // 76
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"demo.js":function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// imports/demo.js                                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.export({                                                                                                        // 1
                  allowDemo: function () {                                                                             // 1
                                    return allowDemo;                                                                  // 1
                  }                                                                                                    // 1
});                                                                                                                    // 1
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var allowDemo = Meteor.settings && Meteor.settings.public && Meteor.settings.public.allowDemoAccounts;                 // 17
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"server":{"accounts":{"email-token":{"token-server.js":["crypto","/imports/server/email.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/email-token/token-server.js                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var crypto = void 0;                                                                                                   // 1
module.import("crypto", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var sendEmail = void 0;                                                                                                // 1
module.import("/imports/server/email.js", {                                                                            // 1
  "send": function (v) {                                                                                               // 1
    sendEmail = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 5
                                                                                                                       //
var V1_ROUNDS = 4096; // Selected to take ~5msec at creation time (2016) on a developer's laptop.                      // 7
                                                                                                                       //
var V1_KEYSIZE = 32; // 256 bits / 8 bits/byte = 32 bytes                                                              // 8
                                                                                                                       //
var V1_HASHFUNC = "sha512"; // ^ hash function used with pbkdf2.  Chosen to be different from the function which maps the token
// to the value stored in the database.  Note that the first thing that pbkdf2 does is                                 // 11
// HMAC(HASHFUNC, key, salt), and the first thing that HMAC does is either pad or hash the key to                      // 12
// make it the appropriate width.  The result is that knowing sha256(key) and the salt is possibly                     // 13
// sufficient to reconstruct the output of pbkdf2().                                                                   // 14
                                                                                                                       //
var V1_CIPHER = "AES-256-CTR"; // cipher used                                                                          // 15
                                                                                                                       //
var TOKEN_EXPIRATION_MS = 60 * 60 * 1000;                                                                              // 17
                                                                                                                       //
var cleanupExpiredTokens = function () {                                                                               // 19
  Meteor.users.update({                                                                                                // 20
    "services.email.tokens.createdAt": {                                                                               // 21
      $lt: new Date(Date.now() - TOKEN_EXPIRATION_MS)                                                                  // 22
    }                                                                                                                  // 21
  }, {                                                                                                                 // 20
    $pull: {                                                                                                           // 25
      "services.email.tokens": {                                                                                       // 26
        createdAt: {                                                                                                   // 27
          $lt: new Date(Date.now() - TOKEN_EXPIRATION_MS)                                                              // 27
        }                                                                                                              // 27
      }                                                                                                                // 26
    }                                                                                                                  // 25
  }, {                                                                                                                 // 24
    multi: true                                                                                                        // 31
  });                                                                                                                  // 30
};                                                                                                                     // 33
                                                                                                                       //
Meteor.startup(cleanupExpiredTokens); // Tokens can actually last up to 2 * TOKEN_EXPIRATION_MS                        // 35
                                                                                                                       //
SandstormDb.periodicCleanup(TOKEN_EXPIRATION_MS, cleanupExpiredTokens);                                                // 37
                                                                                                                       //
var hashToken = function (token) {                                                                                     // 39
  return {                                                                                                             // 40
    digest: SHA256(token),                                                                                             // 41
    algorithm: "sha-256"                                                                                               // 42
  };                                                                                                                   // 40
};                                                                                                                     // 44
                                                                                                                       //
var checkToken = function (tokens, token) {                                                                            // 46
  // Looks for an object in `tokens` with `algorithm` and `digest` fields matching those in `token`.                   // 47
  // Returns the matching object, if one is found, or undefined if none match.                                         // 48
  var foundToken = undefined;                                                                                          // 49
  tokens.forEach(function (userToken) {                                                                                // 50
    if (userToken.algorithm === token.algorithm && userToken.digest === token.digest) {                                // 51
      foundToken = userToken;                                                                                          // 53
    }                                                                                                                  // 54
  });                                                                                                                  // 55
  return foundToken;                                                                                                   // 57
};                                                                                                                     // 58
                                                                                                                       //
var consumeToken = function (user, token) {                                                                            // 60
  var hashedToken = hashToken(token);                                                                                  // 61
  var foundToken = checkToken(user.services.email.tokens, hashedToken);                                                // 62
                                                                                                                       //
  if (foundToken !== undefined) {                                                                                      // 64
    Meteor.users.update({                                                                                              // 65
      _id: user._id                                                                                                    // 65
    }, {                                                                                                               // 65
      $pull: {                                                                                                         // 65
        "services.email.tokens": hashedToken                                                                           // 65
      }                                                                                                                // 65
    });                                                                                                                // 65
  }                                                                                                                    // 66
                                                                                                                       //
  return foundToken;                                                                                                   // 68
};                                                                                                                     // 69
                                                                                                                       //
var makeBox = function (token, plaintext) {                                                                            // 71
  // Encrypt plaintext symmetrically with a key derived from token.  Returns an object with                            // 72
  // ciphertext and associated data needed to decrypt later.                                                           // 73
  // Produce a symmetric key.  Note that the token itself does not have sufficient entropy to                          // 75
  // be used as a key directly, so we need to use a KDF with a strong random salt.                                     // 76
  // In the fullness of time, it might be nice to move away from using a KDF (which blocks the whole                   // 77
  // node process) in favor of the token itself having enough entropy to serve as the key itself.                      // 78
  // This would require lengthening the token, which would make the manual-code-entry workflow                         // 79
  // worse, so I'm punting on that for now.                                                                            // 80
  var salt = Random.secret(16);                                                                                        // 81
  var key = crypto.pbkdf2Sync(token, salt, V1_ROUNDS, V1_KEYSIZE, V1_HASHFUNC);                                        // 82
  var iv = crypto.randomBytes(16);                                                                                     // 83
  var cipher = crypto.createCipheriv(V1_CIPHER, key, iv);                                                              // 84
  var ciphertext = cipher.update(new Buffer(plaintext, "binary"));                                                     // 85
  return {                                                                                                             // 86
    version: 1,                                                                                                        // 87
    salt: salt,                                                                                                        // 88
    iv: iv.toString("base64"),                                                                                         // 89
    boxedValue: ciphertext.toString("base64")                                                                          // 90
  };                                                                                                                   // 86
};                                                                                                                     // 92
                                                                                                                       //
var tryUnbox = function (box, secret) {                                                                                // 94
  if (box) {                                                                                                           // 95
    if (box.version === 1) {                                                                                           // 96
      var key = crypto.pbkdf2Sync(secret, box.salt, V1_ROUNDS, V1_KEYSIZE, V1_HASHFUNC);                               // 97
      var iv = new Buffer(box.iv, "base64");                                                                           // 98
      var cipher = crypto.createDecipheriv(V1_CIPHER, key, iv);                                                        // 99
      var cipherText = new Buffer(box.boxedValue, "base64");                                                           // 100
      var plaintext = cipher.update(cipherText);                                                                       // 101
      return plaintext.toString("binary");                                                                             // 102
    }                                                                                                                  // 103
  } // If no box was provided, or it was of an unknown version, return no data.                                        // 104
                                                                                                                       //
                                                                                                                       //
  return;                                                                                                              // 107
}; // Handler to login with a token.                                                                                   // 108
                                                                                                                       //
                                                                                                                       //
Accounts.registerLoginHandler("email", function (options) {                                                            // 111
  if (!options.email) {                                                                                                // 112
    return undefined; // don't handle                                                                                  // 113
  }                                                                                                                    // 114
                                                                                                                       //
  if (!Accounts.identityServices.email.isEnabled()) {                                                                  // 116
    throw new Meteor.Error(403, "Email identity service is disabled.");                                                // 117
  }                                                                                                                    // 118
                                                                                                                       //
  options = options.email;                                                                                             // 120
  check(options, {                                                                                                     // 121
    email: String,                                                                                                     // 122
    token: String                                                                                                      // 123
  });                                                                                                                  // 121
  var user = Meteor.users.findOne({                                                                                    // 126
    "services.email.email": options.email                                                                              // 127
  }, {                                                                                                                 // 126
    fields: {                                                                                                          // 129
      "services.email": 1                                                                                              // 130
    }                                                                                                                  // 129
  });                                                                                                                  // 128
                                                                                                                       //
  if (!user) {                                                                                                         // 134
    console.error("User not found:", options.email);                                                                   // 135
    return {                                                                                                           // 136
      error: new Meteor.Error(403, "User not found")                                                                   // 137
    };                                                                                                                 // 136
  }                                                                                                                    // 139
                                                                                                                       //
  if (!user.services.email.tokens) {                                                                                   // 141
    console.error("User has no token set:", options.email);                                                            // 142
    return {                                                                                                           // 143
      error: new Meteor.Error(403, "User has no token set")                                                            // 144
    };                                                                                                                 // 143
  }                                                                                                                    // 146
                                                                                                                       //
  var tokenString = options.token.trim();                                                                              // 148
  var maybeToken = consumeToken(user, tokenString);                                                                    // 149
                                                                                                                       //
  if (!maybeToken) {                                                                                                   // 150
    console.error("Token not found:", options.email);                                                                  // 151
    return {                                                                                                           // 152
      error: new Meteor.Error(403, "Invalid authentication code")                                                      // 153
    };                                                                                                                 // 152
  } // Attempt to decrypt the resumePath, if provided.                                                                 // 155
                                                                                                                       //
                                                                                                                       //
  var resumePath = tryUnbox(maybeToken.secureBox, tokenString);                                                        // 158
  return {                                                                                                             // 160
    userId: user._id,                                                                                                  // 161
    options: {                                                                                                         // 162
      resumePath: resumePath                                                                                           // 163
    }                                                                                                                  // 162
  };                                                                                                                   // 160
});                                                                                                                    // 166
                                                                                                                       //
var makeTokenUrl = function (email, token, options) {                                                                  // 168
  if (options.linking) {                                                                                               // 169
    return options.rootUrl + "/_emailLinkIdentity/" + encodeURIComponent(email) + "/" + encodeURIComponent(token) + "/" + Meteor.userId() + "?allowLogin=" + options.linking.allowLogin;
  } else {                                                                                                             // 173
    return options.rootUrl + "/_emailLogin/" + encodeURIComponent(email) + "/" + encodeURIComponent(token);            // 174
  }                                                                                                                    // 175
}; ///                                                                                                                 // 176
/// EMAIL VERIFICATION                                                                                                 // 179
///                                                                                                                    // 180
                                                                                                                       //
                                                                                                                       //
var sendTokenEmail = function (db, email, token, options) {                                                            // 181
  var subject = void 0;                                                                                                // 182
  var text = void 0;                                                                                                   // 183
  var rootHostname = Url.parse(options.rootUrl).hostname;                                                              // 185
                                                                                                                       //
  if (!options.linking) {                                                                                              // 187
    subject = "Log in to " + rootHostname;                                                                             // 188
    text = "To confirm this email address on ";                                                                        // 189
  } else {                                                                                                             // 190
    subject = "Confirm this email address on " + rootHostname;                                                         // 191
    text = "To confirm this email address on ";                                                                        // 192
  }                                                                                                                    // 193
                                                                                                                       //
  text = text + rootHostname + ", click on the following link:\n\n" + makeTokenUrl(email, token, options) + "\n\n" + "Alternatively, enter the following one-time authentication code into the log-in form:\n\n" + token;
  var sendOptions = {                                                                                                  // 200
    to: email,                                                                                                         // 201
    from: {                                                                                                            // 202
      name: globalDb.getServerTitle(),                                                                                 // 202
      address: db.getReturnAddress()                                                                                   // 202
    },                                                                                                                 // 202
    subject: subject,                                                                                                  // 203
    text: text                                                                                                         // 204
  };                                                                                                                   // 200
  sendEmail(sendOptions);                                                                                              // 207
};                                                                                                                     // 208
                                                                                                                       //
var parsedRootUrl = Url.parse(process.env.ROOT_URL); ///                                                               // 210
/// CREATING USERS                                                                                                     // 212
///                                                                                                                    // 213
// returns the user id                                                                                                 // 214
                                                                                                                       //
var createAndEmailTokenForUser = function (db, email, options) {                                                       // 215
  check(email, String);                                                                                                // 216
  check(options, {                                                                                                     // 217
    resumePath: String,                                                                                                // 218
    linking: Match.Optional({                                                                                          // 219
      allowLogin: Boolean                                                                                              // 219
    }),                                                                                                                // 219
    rootUrl: String                                                                                                    // 220
  });                                                                                                                  // 217
  var parsedUrl = Url.parse(options.rootUrl);                                                                          // 223
                                                                                                                       //
  if ((parsedUrl.hostname !== parsedRootUrl.hostname || parsedUrl.protocol !== parsedRootUrl.protocol) && !db.hostIsStandalone(parsedUrl.hostname)) {
    // Ignore port and only check hostname/protocol since IE will differ from other browsers and                       // 227
    // sometimes include port 80/443 and sometimes won't                                                               // 228
    throw new Meteor.Error(400, "rootUrl is not valid");                                                               // 229
  }                                                                                                                    // 230
                                                                                                                       //
  var atIndex = email.indexOf("@");                                                                                    // 232
                                                                                                                       //
  if (atIndex === -1) {                                                                                                // 233
    throw new Meteor.Error(400, "No @ symbol was found in your email");                                                // 234
  }                                                                                                                    // 235
                                                                                                                       //
  var user = Meteor.users.findOne({                                                                                    // 237
    "services.email.email": email                                                                                      // 237
  }, {                                                                                                                 // 237
    fields: {                                                                                                          // 238
      "services.email": 1                                                                                              // 238
    }                                                                                                                  // 238
  });                                                                                                                  // 238
  var userId = void 0; // TODO(someday): make this shorter, and handle requests that try to brute force it.            // 239
  // Alternately, require using the link over copy/pasting the code, and crank up the entropy.                         // 242
                                                                                                                       //
  var token = Random.id(12);                                                                                           // 243
  var tokenObj = hashToken(token);                                                                                     // 244
  tokenObj.createdAt = new Date();                                                                                     // 245
  tokenObj.secureBox = makeBox(token, options.resumePath);                                                             // 246
                                                                                                                       //
  if (user) {                                                                                                          // 248
    if (user.services.email.tokens && user.services.email.tokens.length > 2) {                                         // 249
      throw new Meteor.Error("alreadySentEmailToken", "It looks like we sent a log in email to this address not long " + "ago. Please use the one that was already sent (check your spam folder if you can't find " + "it), or wait a while and try again.");
    }                                                                                                                  // 255
                                                                                                                       //
    userId = user._id;                                                                                                 // 257
    Meteor.users.update({                                                                                              // 259
      _id: user._id                                                                                                    // 259
    }, {                                                                                                               // 259
      $push: {                                                                                                         // 259
        "services.email.tokens": tokenObj                                                                              // 259
      }                                                                                                                // 259
    });                                                                                                                // 259
  } else {                                                                                                             // 260
    var _options = {};                                                                                                 // 261
    user = {                                                                                                           // 262
      services: {                                                                                                      // 263
        email: {                                                                                                       // 264
          tokens: [tokenObj],                                                                                          // 265
          email: email                                                                                                 // 266
        }                                                                                                              // 264
      }                                                                                                                // 263
    };                                                                                                                 // 262
    userId = Accounts.insertUserDoc(_options, user);                                                                   // 271
  }                                                                                                                    // 272
                                                                                                                       //
  sendTokenEmail(db, email, token, options);                                                                           // 274
  return userId;                                                                                                       // 276
};                                                                                                                     // 277
                                                                                                                       //
Meteor.methods({                                                                                                       // 279
  createAndEmailTokenForUser: function (email, options) {                                                              // 280
    // method for create user. Requests come from the client.                                                          // 281
    // This method will create a user if it doesn't exist, otherwise it will generate a token.                         // 282
    // It will always send an email to the user                                                                        // 283
    check(email, String);                                                                                              // 285
    check(options, {                                                                                                   // 286
      resumePath: String,                                                                                              // 287
      linking: Match.Optional({                                                                                        // 288
        allowLogin: Boolean                                                                                            // 288
      }),                                                                                                              // 288
      rootUrl: String                                                                                                  // 289
    });                                                                                                                // 286
                                                                                                                       //
    if (!Accounts.identityServices.email.isEnabled()) {                                                                // 292
      throw new Meteor.Error(403, "Email identity service is disabled.");                                              // 293
    } // Create user. result contains id and token.                                                                    // 294
                                                                                                                       //
                                                                                                                       //
    var user = createAndEmailTokenForUser(this.connection.sandstormDb, email, options);                                // 296
  },                                                                                                                   // 297
  linkEmailIdentityToAccount: function (email, token, allowLogin) {                                                    // 299
    // Links the email identity with address `email` and login token `token` to the current account.                   // 300
    check(email, String);                                                                                              // 301
    check(token, String);                                                                                              // 302
    check(allowLogin, Boolean);                                                                                        // 303
    var account = Meteor.user();                                                                                       // 304
                                                                                                                       //
    if (!account || !account.loginIdentities) {                                                                        // 305
      throw new Meteor.Error(403, "Must be logged in to an account to link an email identity.");                       // 306
    }                                                                                                                  // 307
                                                                                                                       //
    var identity = Meteor.users.findOne({                                                                              // 309
      "services.email.email": email                                                                                    // 309
    }, {                                                                                                               // 309
      fields: {                                                                                                        // 310
        "services.email": 1                                                                                            // 310
      }                                                                                                                // 310
    });                                                                                                                // 310
                                                                                                                       //
    if (!identity) {                                                                                                   // 311
      throw new Meteor.Error(403, "Invalid authentication code.");                                                     // 312
    }                                                                                                                  // 313
                                                                                                                       //
    var maybeToken = consumeToken(identity, token.trim());                                                             // 315
                                                                                                                       //
    if (!maybeToken) {                                                                                                 // 316
      throw new Meteor.Error(403, "Invalid authentication code.");                                                     // 317
    }                                                                                                                  // 318
                                                                                                                       //
    Accounts.linkIdentityToAccount(this.connection.sandstormDb, this.connection.sandstormBackend, identity._id, account._id, allowLogin); // Return the resume path, if we have one.
                                                                                                                       //
    var resumePath = tryUnbox(maybeToken.secureBox, token);                                                            // 324
    return resumePath;                                                                                                 // 325
  }                                                                                                                    // 326
});                                                                                                                    // 279
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"identity":{"identity-server.js":["/imports/server/backend.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/identity/identity-server.js                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var SandstormBackend = void 0;                                                                                         // 1
module.import("/imports/server/backend.js", {                                                                          // 1
  "SandstormBackend": function (v) {                                                                                   // 1
    SandstormBackend = v;                                                                                              // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var linkIdentityToAccountInternal = function (db, backend, identityId, accountId, allowLogin) {                        // 19
  // Links the identity to the account. If `allowLogin` is true, grants the identity login access                      // 20
  // if possible. Makes the account durable if it is a demo account.                                                   // 21
  check(db, SandstormDb);                                                                                              // 23
  check(backend, SandstormBackend);                                                                                    // 24
  check(identityId, String);                                                                                           // 25
  check(accountId, String);                                                                                            // 26
  var accountUser = Meteor.users.findOne({                                                                             // 28
    _id: accountId                                                                                                     // 28
  });                                                                                                                  // 28
                                                                                                                       //
  if (!accountUser) {                                                                                                  // 29
    throw new Meteor.Error(404, "No account found with ID " + accountId);                                              // 30
  }                                                                                                                    // 31
                                                                                                                       //
  if (accountUser.profile) {                                                                                           // 33
    throw new Meteor.Error(400, "Cannot link an identity to another identity.");                                       // 34
  }                                                                                                                    // 35
                                                                                                                       //
  if (!!_.findWhere(accountUser.loginIdentities, {                                                                     // 37
    id: identityId                                                                                                     // 37
  }) || !!_.findWhere(accountUser.nonloginIdentities, {                                                                // 37
    id: identityId                                                                                                     // 38
  })) {                                                                                                                // 38
    throw new Meteor.Error("alreadyLinked", "Cannot link an identity that's alread linked to this account.");          // 39
  }                                                                                                                    // 41
                                                                                                                       //
  var identityUser = Meteor.users.findOne({                                                                            // 43
    _id: identityId                                                                                                    // 43
  });                                                                                                                  // 43
                                                                                                                       //
  if (!identityUser) {                                                                                                 // 45
    throw new Meteor.Error(404, "No identity found with ID " + identityId);                                            // 46
  }                                                                                                                    // 47
                                                                                                                       //
  if (!identityUser.profile) {                                                                                         // 49
    throw new Meteor.Error(400, "Cannot link an account to another account");                                          // 50
  }                                                                                                                    // 51
                                                                                                                       //
  db.deleteUnusedAccount(backend, identityUser._id);                                                                   // 53
                                                                                                                       //
  if (Meteor.users.findOne({                                                                                           // 54
    "loginIdentities.id": identityUser._id                                                                             // 54
  })) {                                                                                                                // 54
    throw new Meteor.Error(403, "Cannot link an identity that can already log into another account");                  // 55
  }                                                                                                                    // 57
                                                                                                                       //
  var alreadyLinked = !!Meteor.users.findOne({                                                                         // 59
    "nonloginIdentities.id": identityUser._id                                                                          // 59
  });                                                                                                                  // 59
  var pushModifier = alreadyLinked || !allowLogin ? {                                                                  // 61
    nonloginIdentities: {                                                                                              // 62
      id: identityUser._id                                                                                             // 62
    }                                                                                                                  // 62
  } : {                                                                                                                // 62
    loginIdentities: {                                                                                                 // 63
      id: identityUser._id                                                                                             // 63
    }                                                                                                                  // 63
  };                                                                                                                   // 63
  var modifier = void 0;                                                                                               // 65
                                                                                                                       //
  if (accountUser.expires) {                                                                                           // 66
    if (alreadyLinked) {                                                                                               // 67
      throw new Meteor.Error(403, "Cannot create an account for an identity that's " + "already linked to another account.");
    }                                                                                                                  // 70
                                                                                                                       //
    modifier = {                                                                                                       // 72
      $push: pushModifier,                                                                                             // 73
      $unset: {                                                                                                        // 74
        expires: 1                                                                                                     // 74
      },                                                                                                               // 74
      $set: {                                                                                                          // 75
        upgradedFromDemo: Date.now()                                                                                   // 75
      }                                                                                                                // 75
    };                                                                                                                 // 72
                                                                                                                       //
    if (db.isReferralEnabled()) {                                                                                      // 77
      // Demo users never got the referral notification. Send it now:                                                  // 78
      db.sendReferralProgramNotification(accountUser._id);                                                             // 79
    }                                                                                                                  // 80
  } else {                                                                                                             // 82
    modifier = {                                                                                                       // 83
      $push: pushModifier                                                                                              // 83
    };                                                                                                                 // 83
  } // Make sure not to add the same identity twice.                                                                   // 84
                                                                                                                       //
                                                                                                                       //
  Meteor.users.update({                                                                                                // 87
    _id: accountUser._id,                                                                                              // 87
    "nonloginIdentities.id": {                                                                                         // 88
      $ne: identityUser._id                                                                                            // 88
    },                                                                                                                 // 88
    "loginIdentities.id": {                                                                                            // 89
      $ne: identityUser._id                                                                                            // 89
    }                                                                                                                  // 89
  }, modifier);                                                                                                        // 87
                                                                                                                       //
  if (accountUser.expires) {                                                                                           // 92
    var demoIdentityId = SandstormDb.getUserIdentityIds(accountUser)[0];                                               // 93
    Meteor.users.update({                                                                                              // 94
      _id: demoIdentityId                                                                                              // 94
    }, {                                                                                                               // 94
      $unset: {                                                                                                        // 95
        expires: 1                                                                                                     // 95
      },                                                                                                               // 95
      $set: {                                                                                                          // 96
        upgradedFromDemo: Date.now()                                                                                   // 96
      }                                                                                                                // 96
    }); // Mark the demo identity as nonlogin. It'd be nicer if the identity started out as nonlogin,                  // 95
    // but to get that to work we would need to adjust the account creation and first login logic.                     // 99
                                                                                                                       //
    Meteor.users.update({                                                                                              // 100
      _id: accountUser._id,                                                                                            // 100
      "loginIdentities.id": demoIdentityId,                                                                            // 101
      "nonloginIdentities.id": {                                                                                       // 102
        $not: {                                                                                                        // 102
          $eq: demoIdentityId                                                                                          // 102
        }                                                                                                              // 102
      }                                                                                                                // 102
    }, {                                                                                                               // 100
      $pull: {                                                                                                         // 103
        loginIdentities: {                                                                                             // 103
          id: demoIdentityId                                                                                           // 103
        }                                                                                                              // 103
      },                                                                                                               // 103
      $push: {                                                                                                         // 104
        nonloginIdentities: {                                                                                          // 104
          id: demoIdentityId                                                                                           // 104
        }                                                                                                              // 104
      }                                                                                                                // 104
    });                                                                                                                // 103
  }                                                                                                                    // 106
};                                                                                                                     // 107
                                                                                                                       //
Meteor.methods({                                                                                                       // 109
  loginWithIdentity: function (accountUserId) {                                                                        // 110
    // Logs into the account with ID `accountUserId`. Throws an exception if the current user is                       // 111
    // not an identity user listed in the account's `loginIdentities` field. This method is not                        // 112
    // intended to be called directly; client-side code should only invoke it through                                  // 113
    // `Meteor.loginWithIdentity()`, which additionally maintains the standard Meteor client-side                      // 114
    // login state.                                                                                                    // 115
    check(accountUserId, String);                                                                                      // 117
    var identityUser = Meteor.user();                                                                                  // 119
                                                                                                                       //
    if (!identityUser || !identityUser.profile) {                                                                      // 120
      throw new Meteor.Error(403, "Must be already logged in as an identity.");                                        // 121
    }                                                                                                                  // 122
                                                                                                                       //
    var accountUser = Meteor.users.findOne(accountUserId);                                                             // 124
                                                                                                                       //
    if (!accountUser) {                                                                                                // 125
      throw new Meteor.Error(404, "No such user found: " + accountUserId);                                             // 126
    }                                                                                                                  // 127
                                                                                                                       //
    var linkedIdentity = _.findWhere(accountUser.loginIdentities, {                                                    // 129
      id: identityUser._id                                                                                             // 129
    });                                                                                                                // 129
                                                                                                                       //
    if (!linkedIdentity) {                                                                                             // 131
      throw new Meteor.Error(403, "Current identity is not a login identity for account " + accountUserId);            // 132
    }                                                                                                                  // 134
                                                                                                                       //
    return Accounts._loginMethod(this, "loginWithIdentity", [accountUserId], "identity", function () {                 // 136
      return {                                                                                                         // 137
        userId: accountUserId                                                                                          // 137
      };                                                                                                               // 137
    });                                                                                                                // 137
  },                                                                                                                   // 138
  createAccountForIdentity: function () {                                                                              // 140
    // Creates a new account for the currently-logged-in identity.                                                     // 141
    var user = Meteor.user();                                                                                          // 143
                                                                                                                       //
    if (!(user && user.profile)) {                                                                                     // 144
      throw new Meteor.Error(403, "Must be logged in as an identity in order to create an account.");                  // 145
    }                                                                                                                  // 146
                                                                                                                       //
    if (Meteor.users.findOne({                                                                                         // 148
      $or: [{                                                                                                          // 149
        "loginIdentities.id": user._id                                                                                 // 150
      }, {                                                                                                             // 150
        "nonloginIdentities.id": user._id                                                                              // 151
      }]                                                                                                               // 151
    })) {                                                                                                              // 148
      throw new Meteor.Error(403, "Cannot create an account for an identity that's already " + "linked to another account.");
    }                                                                                                                  // 156
                                                                                                                       //
    var newUser = {                                                                                                    // 158
      loginIdentities: [{                                                                                              // 159
        id: user._id                                                                                                   // 159
      }],                                                                                                              // 159
      nonloginIdentities: []                                                                                           // 160
    };                                                                                                                 // 158
                                                                                                                       //
    if (user.services.dev) {                                                                                           // 162
      newUser.signupKey = "devAccounts";                                                                               // 163
                                                                                                                       //
      if (user.services.dev.isAdmin) {                                                                                 // 164
        newUser.isAdmin = true;                                                                                        // 165
      }                                                                                                                // 166
                                                                                                                       //
      if (user.services.dev.hasCompletedSignup) {                                                                      // 168
        newUser.hasCompletedSignup = true;                                                                             // 169
      }                                                                                                                // 170
    } else if (user.expires) {                                                                                         // 171
      // Demo user.                                                                                                    // 172
      newUser.expires = user.expires;                                                                                  // 173
                                                                                                                       //
      if (!!user.appDemoId) {                                                                                          // 174
        newUser.appDemoId = user.appDemoId;                                                                            // 175
      }                                                                                                                // 176
    }                                                                                                                  // 177
                                                                                                                       //
    var options = {}; // This will throw an error if the identity has been added as a login identity to some           // 179
    // other account while we were executing the body of this method.                                                  // 182
                                                                                                                       //
    return Accounts.insertUserDoc(options, newUser);                                                                   // 183
  },                                                                                                                   // 184
  linkIdentityToAccount: function (token) {                                                                            // 186
    // Links the identity of the current user to the account that has `token` as a resume token.                       // 187
    // If the account is a demo account, makes the account durable and gives the identity login                        // 188
    // access to it.                                                                                                   // 189
    check(token, String);                                                                                              // 191
                                                                                                                       //
    if (!this.userId) {                                                                                                // 193
      throw new Meteor.Error(403, "Cannot link to account if not logged in.");                                         // 194
    }                                                                                                                  // 195
                                                                                                                       //
    var hashed = Accounts._hashLoginToken(token);                                                                      // 197
                                                                                                                       //
    var accountUser = Meteor.users.findOne({                                                                           // 198
      "services.resume.loginTokens.hashedToken": hashed                                                                // 198
    });                                                                                                                // 198
    linkIdentityToAccountInternal(this.connection.sandstormDb, this.connection.sandstormBackend, this.userId, accountUser._id, true);
  },                                                                                                                   // 202
  unlinkIdentity: function (accountUserId, identityId) {                                                               // 204
    // Unlinks the identity with ID `identityId` from the account with ID `accountUserId`.                             // 205
    check(identityId, String);                                                                                         // 207
    check(accountUserId, String);                                                                                      // 208
                                                                                                                       //
    if (!this.userId) {                                                                                                // 210
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 211
    }                                                                                                                  // 212
                                                                                                                       //
    if (!this.connection.sandstormDb.userHasIdentity(this.userId, identityId)) {                                       // 214
      throw new Meteor.Error(403, "Current user does not own identity " + identityId);                                 // 215
    }                                                                                                                  // 216
                                                                                                                       //
    var identityUser = Meteor.users.findOne({                                                                          // 218
      _id: identityId                                                                                                  // 218
    });                                                                                                                // 218
    Meteor.users.update({                                                                                              // 219
      _id: accountUserId                                                                                               // 220
    }, {                                                                                                               // 219
      $pull: {                                                                                                         // 222
        nonloginIdentities: {                                                                                          // 223
          id: identityId                                                                                               // 223
        },                                                                                                             // 223
        loginIdentities: {                                                                                             // 224
          id: identityId                                                                                               // 224
        }                                                                                                              // 224
      }                                                                                                                // 222
    });                                                                                                                // 221
  },                                                                                                                   // 227
  setIdentityAllowsLogin: function (identityId, allowLogin) {                                                          // 229
    // Sets whether the current account allows the identity with ID `identityId` to log in.                            // 230
    check(identityId, String);                                                                                         // 232
    check(allowLogin, Boolean);                                                                                        // 233
                                                                                                                       //
    if (!this.userId) {                                                                                                // 234
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 235
    }                                                                                                                  // 236
                                                                                                                       //
    if (!this.connection.sandstormDb.userHasIdentity(this.userId, identityId)) {                                       // 238
      throw new Meteor.Error(403, "Current user does not own identity " + identityId);                                 // 239
    }                                                                                                                  // 240
                                                                                                                       //
    if (allowLogin) {                                                                                                  // 242
      Meteor.users.update({                                                                                            // 243
        _id: this.userId,                                                                                              // 243
        "nonloginIdentities.id": identityId,                                                                           // 244
        "loginIdentities.id": {                                                                                        // 245
          $not: {                                                                                                      // 245
            $eq: identityId                                                                                            // 245
          }                                                                                                            // 245
        }                                                                                                              // 245
      }, {                                                                                                             // 243
        $pull: {                                                                                                       // 246
          nonloginIdentities: {                                                                                        // 246
            id: identityId                                                                                             // 246
          }                                                                                                            // 246
        },                                                                                                             // 246
        $push: {                                                                                                       // 247
          loginIdentities: {                                                                                           // 247
            id: identityId                                                                                             // 247
          }                                                                                                            // 247
        }                                                                                                              // 247
      });                                                                                                              // 246
    } else {                                                                                                           // 248
      Meteor.users.update({                                                                                            // 249
        _id: this.userId,                                                                                              // 249
        "loginIdentities.id": identityId,                                                                              // 250
        "nonloginIdentities.id": {                                                                                     // 251
          $not: {                                                                                                      // 251
            $eq: identityId                                                                                            // 251
          }                                                                                                            // 251
        }                                                                                                              // 251
      }, {                                                                                                             // 249
        $pull: {                                                                                                       // 252
          loginIdentities: {                                                                                           // 252
            id: identityId                                                                                             // 252
          }                                                                                                            // 252
        },                                                                                                             // 252
        $push: {                                                                                                       // 253
          nonloginIdentities: {                                                                                        // 253
            id: identityId                                                                                             // 253
          }                                                                                                            // 253
        }                                                                                                              // 253
      });                                                                                                              // 252
    }                                                                                                                  // 254
  },                                                                                                                   // 255
  logoutIdentitiesOfCurrentAccount: function () {                                                                      // 257
    // Logs out all identities that are allowed to log in to the current account.                                      // 258
    var user = Meteor.user();                                                                                          // 259
                                                                                                                       //
    if (user && user.loginIdentities) {                                                                                // 260
      user.loginIdentities.forEach(function (identity) {                                                               // 261
        Meteor.users.update({                                                                                          // 262
          _id: identity.id                                                                                             // 262
        }, {                                                                                                           // 262
          $set: {                                                                                                      // 262
            "services.resume.loginTokens": []                                                                          // 262
          }                                                                                                            // 262
        });                                                                                                            // 262
      });                                                                                                              // 263
    }                                                                                                                  // 264
  }                                                                                                                    // 265
});                                                                                                                    // 109
                                                                                                                       //
Accounts.linkIdentityToAccount = function (db, backend, identityId, accountId, allowLogin) {                           // 268
  // Links the identity to the account. If the account is a demo account, makes it durable.                            // 269
  // If `allowLogin` is true, attempts to give the identity login access.                                              // 270
  check(db, SandstormDb);                                                                                              // 271
  check(backend, SandstormBackend);                                                                                    // 272
  check(identityId, String);                                                                                           // 273
  check(accountId, String);                                                                                            // 274
  check(allowLogin, Boolean);                                                                                          // 275
  linkIdentityToAccountInternal(db, backend, identityId, accountId, allowLogin);                                       // 276
};                                                                                                                     // 277
                                                                                                                       //
Meteor.publish("accountsOfIdentity", function (identityId) {                                                           // 279
  check(identityId, String);                                                                                           // 280
  if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return; // We maintain a map from identity IDs to live query handles that track profile changes.
                                                                                                                       //
  var loginIdentities = {};                                                                                            // 284
                                                                                                                       //
  var _this = this;                                                                                                    // 286
                                                                                                                       //
  function addIdentitiesOfAccount(account) {                                                                           // 287
    account.loginIdentities.forEach(function (identity) {                                                              // 288
      if (!(identity.id in loginIdentities)) {                                                                         // 289
        var user = Meteor.users.findOne({                                                                              // 290
          _id: identity.id                                                                                             // 290
        });                                                                                                            // 290
                                                                                                                       //
        if (user) {                                                                                                    // 291
          SandstormDb.fillInProfileDefaults(user);                                                                     // 292
          SandstormDb.fillInIntrinsicName(user);                                                                       // 293
          SandstormDb.fillInLoginId(user);                                                                             // 294
                                                                                                                       //
          var filteredUser = _.pick(user, "_id", "profile", "loginId");                                                // 295
                                                                                                                       //
          filteredUser.loginAccountId = account._id;                                                                   // 296
          filteredUser.sourceIdentityId = identityId;                                                                  // 297
                                                                                                                       //
          _this.added("loginIdentitiesOfLinkedAccounts", user._id, filteredUser);                                      // 298
        }                                                                                                              // 299
                                                                                                                       //
        loginIdentities[identity.id] = Meteor.users.find({                                                             // 301
          _id: identity.id                                                                                             // 302
        }, {                                                                                                           // 302
          fields: {                                                                                                    // 302
            profile: 1                                                                                                 // 302
          }                                                                                                            // 302
        }).observeChanges({                                                                                            // 302
          changed: function (id, fields) {                                                                             // 303
            _this.changed("loginIdentitiesOfLinkedAccounts", id, fields);                                              // 304
          }                                                                                                            // 305
        });                                                                                                            // 302
      }                                                                                                                // 307
    });                                                                                                                // 308
  }                                                                                                                    // 309
                                                                                                                       //
  var cursor = Meteor.users.find({                                                                                     // 311
    $or: [{                                                                                                            // 312
      "loginIdentities.id": identityId                                                                                 // 313
    }, {                                                                                                               // 313
      "nonloginIdentities.id": identityId                                                                              // 314
    }]                                                                                                                 // 314
  });                                                                                                                  // 311
  var handle = cursor.observe({                                                                                        // 318
    added: function (account) {                                                                                        // 319
      addIdentitiesOfAccount(account);                                                                                 // 320
    },                                                                                                                 // 321
    changed: function (newAccount, oldAccount) {                                                                       // 323
      addIdentitiesOfAccount(newAccount);                                                                              // 324
    },                                                                                                                 // 325
    removed: function (account) {                                                                                      // 327
      account.loginIdentities.forEach(function (identity) {                                                            // 328
        if (identity.id in loginIdentities) {                                                                          // 329
          _this.removed("loginIdentitiesOfLinkedAccounts", identity.id);                                               // 330
                                                                                                                       //
          loginIdentities[identity.id].stop();                                                                         // 331
          delete loginIdentities[identity.id];                                                                         // 332
        }                                                                                                              // 333
      });                                                                                                              // 334
    }                                                                                                                  // 335
  });                                                                                                                  // 318
  this.ready();                                                                                                        // 337
  this.onStop(function () {                                                                                            // 339
    handle.stop();                                                                                                     // 340
    Object.keys(loginIdentities).forEach(function (identityId) {                                                       // 341
      loginIdentities[identityId].stop();                                                                              // 342
      delete loginIdentities[identityId];                                                                              // 343
    });                                                                                                                // 344
  });                                                                                                                  // 345
});                                                                                                                    // 346
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"ldap":{"ldap-server.js":["/imports/server/accounts/ldap.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/ldap/ldap-server.js                                                                                 //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var LDAP = void 0;                                                                                                     // 1
module.import("/imports/server/accounts/ldap.js", {                                                                    // 1
  "LDAP": function (v) {                                                                                               // 1
    LDAP = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
// Register login handler with Meteor                                                                                  // 3
Accounts.registerLoginHandler("ldap", function (loginRequest) {                                                        // 4
  // If 'ldap' isn't set in loginRequest object,                                                                       // 5
  // then this isn't the proper handler (return undefined)                                                             // 6
  if (!loginRequest.ldap) {                                                                                            // 7
    return undefined;                                                                                                  // 8
  }                                                                                                                    // 9
                                                                                                                       //
  if (!Accounts.identityServices.ldap.isEnabled()) {                                                                   // 11
    throw new Meteor.Error(403, "LDAP service is disabled.");                                                          // 12
  }                                                                                                                    // 13
                                                                                                                       //
  check(loginRequest, {                                                                                                // 15
    ldap: true,                                                                                                        // 16
    username: String,                                                                                                  // 17
    ldapPass: String                                                                                                   // 18
  }); // Instantiate LDAP                                                                                              // 15
                                                                                                                       //
  var ldapObj = new LDAP(); // Call ldapCheck and get response                                                         // 22
                                                                                                                       //
  var ldapResponse = ldapObj.ldapCheck(this.connection.sandstormDb, loginRequest);                                     // 25
                                                                                                                       //
  if (ldapResponse.error) {                                                                                            // 27
    return {                                                                                                           // 28
      userId: null,                                                                                                    // 29
      error: ldapResponse.error                                                                                        // 30
    };                                                                                                                 // 28
  } else if (ldapResponse.emptySearch) {                                                                               // 32
    return {                                                                                                           // 33
      userId: null,                                                                                                    // 34
      error: new Meteor.Error(403, "User not found in LDAP")                                                           // 35
    };                                                                                                                 // 33
  } else {                                                                                                             // 37
    // Set initial userId and token vals                                                                               // 38
    return Accounts.updateOrCreateUserFromExternalService("ldap", {                                                    // 39
      id: ldapResponse.dn,                                                                                             // 40
      username: loginRequest.username,                                                                                 // 40
      rawAttrs: ldapResponse.searchResults                                                                             // 41
    }, {});                                                                                                            // 40
  }                                                                                                                    // 42
});                                                                                                                    // 44
Meteor.methods({                                                                                                       // 46
  updateQuota: function () {                                                                                           // 47
    return this.connection.sandstormDb.updateUserQuota(Meteor.user()); // This is a no-op if settings aren't enabled   // 48
  }                                                                                                                    // 50
});                                                                                                                    // 46
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"saml":{"saml-server.js":["url","zlib","/imports/server/accounts/saml-utils.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/saml/saml-server.js                                                                                 //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Url = void 0;                                                                                                      // 1
module.import("url", {                                                                                                 // 1
  "default": function (v) {                                                                                            // 1
    Url = v;                                                                                                           // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var zlib = void 0;                                                                                                     // 1
module.import("zlib", {                                                                                                // 1
  "default": function (v) {                                                                                            // 1
    zlib = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var SAML = void 0;                                                                                                     // 1
module.import("/imports/server/accounts/saml-utils.js", {                                                              // 1
  "SAML": function (v) {                                                                                               // 1
    SAML = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
var Fiber = Npm.require("fibers");                                                                                     // 5
                                                                                                                       //
if (!Accounts.saml) {                                                                                                  // 7
  Accounts.saml = {};                                                                                                  // 8
} // We need to use connect. Let's make sure we're using the same version as Meteor's WebApp module                    // 9
// uses. Fortunately, they let us extract it.                                                                          // 12
                                                                                                                       //
                                                                                                                       //
var connect = WebAppInternals.NpmModules.connect.module;                                                               // 13
RoutePolicy.declare("/_saml/", "network");                                                                             // 15
var HOSTNAME = Url.parse(process.env.ROOT_URL).hostname; // TODO(soon): This may need to be a Mongo collection in order to work when the frontend is
//   replicated (but currently SAML is not used on any Blackrock servers).                                             // 20
                                                                                                                       //
var _loginResultForCredentialToken = {};                                                                               // 21
                                                                                                                       //
var retrieveCredential = function (credentialToken) {                                                                  // 23
  var result = _loginResultForCredentialToken[credentialToken];                                                        // 24
  delete _loginResultForCredentialToken[credentialToken];                                                              // 25
  return result;                                                                                                       // 26
};                                                                                                                     // 27
                                                                                                                       //
Accounts.registerLoginHandler(function (loginRequest) {                                                                // 29
  if (!loginRequest.saml || !loginRequest.credentialToken) {                                                           // 30
    return undefined;                                                                                                  // 31
  }                                                                                                                    // 32
                                                                                                                       //
  if (!Accounts.identityServices.saml.isEnabled()) {                                                                   // 34
    throw new Meteor.Error(403, "SAML service is disabled.");                                                          // 35
  }                                                                                                                    // 36
                                                                                                                       //
  var loginResult = retrieveCredential(loginRequest.credentialToken);                                                  // 38
                                                                                                                       //
  if (!loginResult) {                                                                                                  // 39
    throw new Meteor.Error(500, "SAML login did not complete.");                                                       // 40
  } else if (loginResult.profile && loginResult.profile.email) {                                                       // 41
    var user = _.pick(loginResult.profile, "displayName", "email", "nameIDFormat");                                    // 42
                                                                                                                       //
    user.id = loginResult.profile.nameID;                                                                              // 43
    return Accounts.updateOrCreateUserFromExternalService("saml", user, {});                                           // 44
  } else {                                                                                                             // 45
    throw new Meteor.Error(500, "SAML profile did not contain an email address");                                      // 46
  }                                                                                                                    // 47
});                                                                                                                    // 48
                                                                                                                       //
var samlUrlToObject = function (url) {                                                                                 // 50
  // req.url will be "/_saml/<action>/<service name>/<credentialToken>"                                                // 51
  if (!url) return null;                                                                                               // 52
  var splitPath = url.split("/"); // Any non-saml request will continue down the default                               // 55
  // middlewares.                                                                                                      // 58
                                                                                                                       //
  if (splitPath[1] !== "_saml") return null;                                                                           // 59
  return {                                                                                                             // 62
    actionName: splitPath[2],                                                                                          // 63
    serviceName: splitPath[3],                                                                                         // 64
    credentialToken: splitPath[4]                                                                                      // 65
  };                                                                                                                   // 62
};                                                                                                                     // 67
                                                                                                                       //
var closePopup = function (res, err) {                                                                                 // 69
  res.writeHead(200, {                                                                                                 // 70
    "Content-Type": "text/html"                                                                                        // 70
  });                                                                                                                  // 70
  var content = "<html><head><script>window.close()</script></head></html>";                                           // 71
  if (err) content = "<html><body><h2>Sorry, an error occured</h2><div>" + err + '</div><a onclick="window.close();">Close Window</a> <a href="/admin/status" target="_blank">View system log (admin only)</a></body></html>';
  res.end(content, "utf-8");                                                                                           // 75
};                                                                                                                     // 76
                                                                                                                       //
var generateService = function () {                                                                                    // 78
  // TODO(cleanup): Inject the db.                                                                                     // 79
  var db = new SandstormDb();                                                                                          // 80
  var entityId = db.getSamlEntityId();                                                                                 // 82
  var service = {                                                                                                      // 83
    "provider": "default",                                                                                             // 84
    "entryPoint": db.getSamlEntryPoint(),                                                                              // 85
    "logoutUrl": db.getSamlLogout(),                                                                                   // 86
    // TODO(someday): find a better way to inject the DB                                                               // 87
    "issuer": entityId || HOSTNAME,                                                                                    // 88
    // If the certificate has "-----BEGIN CERTIFICATE-----" markers, automatically remove those.                       // 89
    "cert": db.getSamlPublicCert().replace(/-[^\n]*-/g, "").trim()                                                     // 90
  };                                                                                                                   // 83
  return service;                                                                                                      // 92
};                                                                                                                     // 93
                                                                                                                       //
var middleware = function (req, res, next) {                                                                           // 95
  // Make sure to catch any exceptions because otherwise we'd crash                                                    // 96
  // the runner                                                                                                        // 97
  try {                                                                                                                // 98
    var samlObject = samlUrlToObject(req.url);                                                                         // 99
                                                                                                                       //
    if (!samlObject || !samlObject.serviceName) {                                                                      // 100
      next();                                                                                                          // 101
      return;                                                                                                          // 102
    }                                                                                                                  // 103
                                                                                                                       //
    if (samlObject.actionName === "config") {                                                                          // 105
      var _saml = new SAML(generateService());                                                                         // 106
                                                                                                                       //
      res.writeHead(200, {                                                                                             // 107
        "Content-Type": "text/xml"                                                                                     // 107
      });                                                                                                              // 107
      res.end(_saml.generateServiceProviderMetadata());                                                                // 108
      return;                                                                                                          // 109
    }                                                                                                                  // 110
                                                                                                                       //
    if (!Accounts.identityServices.saml.isEnabled()) {                                                                 // 112
      next();                                                                                                          // 113
      return;                                                                                                          // 114
    }                                                                                                                  // 115
                                                                                                                       //
    if (!samlObject.actionName) throw new Error("Missing SAML action");                                                // 117
    var service = generateService(); // Skip everything if there's no service set by the saml middleware               // 120
                                                                                                                       //
    if (!service || samlObject.serviceName !== service.provider) throw new Error("Unexpected SAML service " + samlObject.serviceName);
                                                                                                                       //
    if (samlObject.actionName === "authorize") {                                                                       // 126
      service.callbackUrl = Meteor.absoluteUrl("_saml/validate/" + service.provider);                                  // 127
      service.id = samlObject.credentialToken;                                                                         // 128
                                                                                                                       //
      var _saml2 = new SAML(service);                                                                                  // 129
                                                                                                                       //
      _saml2.getAuthorizeUrl(req, function (err, url) {                                                                // 130
        if (err) throw new Error("Unable to generate authorize url");                                                  // 131
        res.writeHead(302, {                                                                                           // 133
          "Location": url                                                                                              // 133
        });                                                                                                            // 133
        res.end();                                                                                                     // 134
      });                                                                                                              // 135
    } else if (samlObject.actionName === "validate") {                                                                 // 136
      var _saml3 = new SAML(service);                                                                                  // 137
                                                                                                                       //
      _saml3.validateResponse(req.body.SAMLResponse, function (err, profile, loggedOut, responseText) {                // 138
        if (err) {                                                                                                     // 140
          console.error("Error validating SAML response:", err.toString(), "\nFull SAML response XML:\n", responseText);
          throw new Error("Unable to validate SAML response.");                                                        // 143
        } // Do NOT use samlObject.credentialToken; it isn't signed!                                                   // 144
                                                                                                                       //
                                                                                                                       //
        var credentialToken = profile.inResponseToId || profile.InResponseTo;                                          // 147
                                                                                                                       //
        if (!credentialToken) {                                                                                        // 148
          throw new Error("SAML response missing InResponseTo attribute. Sandstorm does not support " + "IdP-initiated authentication; authentication requests must start " + "from the user choosing SAML login in the Sandstorm UI.");
        }                                                                                                              // 153
                                                                                                                       //
        _loginResultForCredentialToken[credentialToken] = {                                                            // 155
          profile: profile                                                                                             // 156
        };                                                                                                             // 155
        closePopup(res);                                                                                               // 159
      });                                                                                                              // 160
    } else {                                                                                                           // 161
      throw new Error("Unexpected SAML action " + samlObject.actionName);                                              // 162
    }                                                                                                                  // 163
  } catch (err) {                                                                                                      // 164
    console.error(err.stack);                                                                                          // 165
    closePopup(res, err);                                                                                              // 166
  }                                                                                                                    // 167
}; // Listen to incoming OAuth http requests                                                                           // 168
                                                                                                                       //
                                                                                                                       //
WebApp.connectHandlers.use(connect.urlencoded()).use(function (req, res, next) {                                       // 171
  // Need to create a Fiber since we're using synchronous http calls and nothing                                       // 172
  // else is wrapping this in a fiber automatically                                                                    // 173
  Fiber(function () {                                                                                                  // 174
    middleware(req, res, next);                                                                                        // 175
  }).run();                                                                                                            // 176
});                                                                                                                    // 177
Meteor.methods({                                                                                                       // 179
  generateSamlLogout: function () {                                                                                    // 180
    var service = generateService();                                                                                   // 181
                                                                                                                       //
    if (!service.logoutUrl) {                                                                                          // 182
      throw new Meteor.Error(500, "No SAML logout url specified");                                                     // 183
    }                                                                                                                  // 184
                                                                                                                       //
    var _saml = new SAML(service);                                                                                     // 186
                                                                                                                       //
    var identity = void 0;                                                                                             // 187
    Meteor.user().loginIdentities.forEach(function (_identity) {                                                       // 188
      var currIdentity = Meteor.users.findOne({                                                                        // 189
        _id: _identity.id                                                                                              // 189
      });                                                                                                              // 189
                                                                                                                       //
      if (currIdentity.services.saml) {                                                                                // 190
        identity = currIdentity;                                                                                       // 191
      }                                                                                                                // 192
    }); // TODO(someday): handle user having more than one SAML identity                                               // 193
                                                                                                                       //
    return Meteor.wrapAsync(_saml.getLogoutUrl.bind(_saml))({                                                          // 196
      user: {                                                                                                          // 197
        nameID: identity.services.saml.id,                                                                             // 198
        nameIDFormat: identity.services.saml.nameIDFormat || "urn:oasis:names:tc:SAML:2.0:nameid-format:persistent"    // 199
      }                                                                                                                // 197
    });                                                                                                                // 196
  },                                                                                                                   // 203
  validateSamlLogout: function (samlRequest) {                                                                         // 205
    check(samlRequest, String);                                                                                        // 206
    var db = this.connection.sandstormDb;                                                                              // 208
    var userId = Meteor.userId();                                                                                      // 209
                                                                                                                       //
    if (!userId) {                                                                                                     // 210
      return new Meteor.Error(403, "Non-logged in users can't logout.");                                               // 211
    }                                                                                                                  // 212
                                                                                                                       //
    var service = generateService();                                                                                   // 214
                                                                                                                       //
    var _saml = new SAML(service);                                                                                     // 215
                                                                                                                       //
    if (samlRequest) {                                                                                                 // 216
      var buf = new Buffer(samlRequest, "base64");                                                                     // 217
      var xml = zlib.inflateRawSync(buf).toString();                                                                   // 218
                                                                                                                       //
      _saml.parseLogoutRequest(xml, function (err, nameId) {                                                           // 219
        if (err) {                                                                                                     // 220
          console.error("Error validating SAML logout response:", err.toString(), "\nFull SAML response XML:\n", responseText);
          throw new Error("Unable to validate SAML logout response.");                                                 // 223
        }                                                                                                              // 224
                                                                                                                       //
        check(nameId, String);                                                                                         // 226
        var identity = db.collections.users.findOne({                                                                  // 227
          "services.saml.id": nameId                                                                                   // 227
        }, {                                                                                                           // 227
          fields: {                                                                                                    // 228
            _id: 1                                                                                                     // 228
          }                                                                                                            // 228
        });                                                                                                            // 228
                                                                                                                       //
        if (!identity) {                                                                                               // 229
          return new Meteor.Error(400, "No identity found matching SAML nameID.");                                     // 230
        }                                                                                                              // 231
                                                                                                                       //
        var user = db.collections.users.findOne({                                                                      // 233
          "loginIdentities.id": identity._id                                                                           // 233
        }, {                                                                                                           // 233
          fields: {                                                                                                    // 234
            _id: 1                                                                                                     // 234
          }                                                                                                            // 234
        });                                                                                                            // 234
                                                                                                                       //
        if (!user) {                                                                                                   // 235
          return new Meteor.Error(403, "No user found for expected SAML identity.");                                   // 236
        }                                                                                                              // 237
                                                                                                                       //
        if (user._id !== userId) {                                                                                     // 239
          var txt = "SAML logout requested for wrong user: " + nameId + ", " + userId;                                 // 240
          console.error(txt);                                                                                          // 241
          throw new Error(txt);                                                                                        // 242
        }                                                                                                              // 243
      });                                                                                                              // 244
    }                                                                                                                  // 245
  }                                                                                                                    // 246
});                                                                                                                    // 179
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"accounts-server.js":["meteor/accounts-base","/imports/server/accounts/picture.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/accounts-server.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Accounts = void 0;                                                                                                 // 1
module.import("meteor/accounts-base", {                                                                                // 1
  "Accounts": function (v) {                                                                                           // 1
    Accounts = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var fetchPicture = void 0,                                                                                             // 1
    userPictureUrl = void 0;                                                                                           // 1
module.import("/imports/server/accounts/picture.js", {                                                                 // 1
  "fetchPicture": function (v) {                                                                                       // 1
    fetchPicture = v;                                                                                                  // 1
  },                                                                                                                   // 1
  "userPictureUrl": function (v) {                                                                                     // 1
    userPictureUrl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 20
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 21
                                                                                                                       //
var ValidHandle = Match.Where(function (handle) {                                                                      // 23
  check(handle, String);                                                                                               // 24
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                                         // 25
});                                                                                                                    // 26
Accounts.onCreateUser(function (options, user) {                                                                       // 28
  if (user.loginIdentities) {                                                                                          // 29
    // it's an account                                                                                                 // 30
    check(user, {                                                                                                      // 31
      _id: String,                                                                                                     // 31
      createdAt: Date,                                                                                                 // 32
      isAdmin: Match.Optional(Boolean),                                                                                // 33
      hasCompletedSignup: Match.Optional(Boolean),                                                                     // 34
      signupKey: Match.Optional(String),                                                                               // 35
      signupNote: Match.Optional(String),                                                                              // 36
      signupEmail: Match.Optional(String),                                                                             // 37
      expires: Match.Optional(Date),                                                                                   // 38
      appDemoId: Match.Optional(String),                                                                               // 39
      loginIdentities: [{                                                                                              // 40
        id: String                                                                                                     // 40
      }],                                                                                                              // 40
      nonloginIdentities: [{                                                                                           // 41
        id: String                                                                                                     // 41
      }]                                                                                                               // 41
    });                                                                                                                // 31
                                                                                                                       //
    if (globalDb.isReferralEnabled()) {                                                                                // 43
      user.experiments = user.experiments || {};                                                                       // 44
      user.experiments = {                                                                                             // 45
        freeGrainLimit: Math.random() < 0.1 ? 100 : "control"                                                          // 46
      };                                                                                                               // 45
                                                                                                                       //
      if (!("expires" in user)) {                                                                                      // 48
        globalDb.sendReferralProgramNotification(user._id);                                                            // 49
      }                                                                                                                // 50
    }                                                                                                                  // 51
                                                                                                                       //
    globalDb.preinstallAppsForUser(user._id);                                                                          // 53
    return user;                                                                                                       // 55
  }                                                                                                                    // 56
                                                                                                                       //
  if (globalDb.getOrganizationDisallowGuests() && !globalDb.isIdentityInOrganization(user)) {                          // 58
    throw new Meteor.Error(400, "User not in organization.");                                                          // 60
  } // Check profile.                                                                                                  // 61
                                                                                                                       //
                                                                                                                       //
  if (options.profile) {                                                                                               // 64
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                               // 65
    check(options.profile, Match.ObjectIncluding({                                                                     // 66
      name: Match.OneOf(null, Match.Optional(String)),                                                                 // 67
      handle: Match.Optional(ValidHandle),                                                                             // 68
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot"))                                       // 69
    }));                                                                                                               // 66
  }                                                                                                                    // 71
                                                                                                                       //
  check(options.unverifiedEmail, Match.Optional(String));                                                              // 73
                                                                                                                       //
  if (options.unverifiedEmail) {                                                                                       // 75
    user.unverifiedEmail = options.unverifiedEmail;                                                                    // 76
  }                                                                                                                    // 77
                                                                                                                       //
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronoun"); // Try downloading avatar.                // 79
                                                                                                                       //
  var url = userPictureUrl(user);                                                                                      // 82
                                                                                                                       //
  if (url) {                                                                                                           // 83
    var assetId = fetchPicture(globalDb, url);                                                                         // 84
                                                                                                                       //
    if (assetId) {                                                                                                     // 85
      user.profile.picture = assetId;                                                                                  // 86
    }                                                                                                                  // 87
  }                                                                                                                    // 88
                                                                                                                       //
  var serviceUserId = void 0;                                                                                          // 90
                                                                                                                       //
  if (user.services && user.services.dev) {                                                                            // 91
    check(user.services.dev, {                                                                                         // 92
      name: String,                                                                                                    // 92
      isAdmin: Boolean,                                                                                                // 92
      hasCompletedSignup: Boolean                                                                                      // 92
    });                                                                                                                // 92
    serviceUserId = user.services.dev.name;                                                                            // 93
    user.profile.service = "dev";                                                                                      // 94
  } else if ("expires" in user) {                                                                                      // 95
    serviceUserId = user._id;                                                                                          // 96
    user.profile.service = "demo";                                                                                     // 97
  } else if (user.services && user.services.email) {                                                                   // 98
    check(user.services.email, {                                                                                       // 99
      email: String,                                                                                                   // 100
      tokens: [{                                                                                                       // 101
        digest: String,                                                                                                // 103
        algorithm: String,                                                                                             // 104
        createdAt: Date,                                                                                               // 105
        secureBox: Match.Optional({                                                                                    // 106
          version: Number,                                                                                             // 107
          salt: String,                                                                                                // 108
          iv: String,                                                                                                  // 109
          boxedValue: String                                                                                           // 110
        })                                                                                                             // 106
      }]                                                                                                               // 102
    });                                                                                                                // 99
    serviceUserId = user.services.email.email;                                                                         // 115
    user.profile.service = "email";                                                                                    // 116
  } else if (user.services && "google" in user.services) {                                                             // 117
    serviceUserId = user.services.google.id;                                                                           // 118
    user.profile.service = "google";                                                                                   // 119
  } else if (user.services && "github" in user.services) {                                                             // 120
    serviceUserId = user.services.github.id;                                                                           // 121
    user.profile.service = "github";                                                                                   // 122
  } else if (user.services && "ldap" in user.services) {                                                               // 123
    serviceUserId = user.services.ldap.id;                                                                             // 124
    user.profile.service = "ldap";                                                                                     // 125
  } else if (user.services && "saml" in user.services) {                                                               // 126
    serviceUserId = user.services.saml.id;                                                                             // 127
    user.profile.service = "saml";                                                                                     // 128
  } else {                                                                                                             // 129
    throw new Meteor.Error(400, "user does not have a recognized identity provider: " + JSON.stringify(user));         // 130
  }                                                                                                                    // 132
                                                                                                                       //
  user._id = Crypto.createHash("sha256").update(user.profile.service + ":" + serviceUserId).digest("hex");             // 134
  return user;                                                                                                         // 137
});                                                                                                                    // 138
Accounts.validateLoginAttempt(function (attempt) {                                                                     // 140
  if (!attempt.allowed) {                                                                                              // 141
    return false;                                                                                                      // 142
  }                                                                                                                    // 143
                                                                                                                       //
  var db = attempt.connection.sandstormDb;                                                                             // 145
  var user = attempt.user;                                                                                             // 146
                                                                                                                       //
  if (user.suspended && user.suspended.admin) {                                                                        // 148
    throw new Meteor.Error(403, "User has been suspended. Please contact the administrator " + "of this server if you believe this to be in error.");
  }                                                                                                                    // 151
                                                                                                                       //
  if (user.loginIdentities) {                                                                                          // 153
    // it's an account                                                                                                 // 154
    if (db.getOrganizationDisallowGuests() && !db.isUserInOrganization(user)) {                                        // 155
      throw new Meteor.Error(403, "User not in organization.");                                                        // 157
    }                                                                                                                  // 158
                                                                                                                       //
    db.updateUserQuota(user); // This is a no-op if settings aren't enabled                                            // 160
  } else {                                                                                                             // 161
    if (db.getOrganizationDisallowGuests() && !db.isIdentityInOrganization(user)) {                                    // 162
      throw new Meteor.Error(403, "User not in organization.");                                                        // 164
    }                                                                                                                  // 165
  }                                                                                                                    // 166
                                                                                                                       //
  return true;                                                                                                         // 168
});                                                                                                                    // 169
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"accounts-ui-methods.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/accounts-ui-methods.js                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file contains method definitions which are available on both client and server (on the                         // 17
// client for prediction purposes, on the server for actual execution).                                                // 18
var ValidHandle = Match.Where(function (handle) {                                                                      // 20
  check(handle, String);                                                                                               // 21
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                                         // 22
});                                                                                                                    // 23
Meteor.methods({                                                                                                       // 25
  updateProfile: function (identityId, profile) {                                                                      // 26
    // TODO(cleanup): This check also appears in sandstorm-db/users.js.                                                // 27
    check(identityId, String);                                                                                         // 28
    check(profile, {                                                                                                   // 30
      name: String,                                                                                                    // 31
      handle: ValidHandle,                                                                                             // 32
      pronoun: Match.OneOf("male", "female", "neutral", "robot"),                                                      // 33
      unverifiedEmail: Match.Optional(String)                                                                          // 34
    });                                                                                                                // 30
                                                                                                                       //
    if (!this.userId) {                                                                                                // 37
      throw new Meteor.Error(403, "not logged in");                                                                    // 38
    }                                                                                                                  // 39
                                                                                                                       //
    var userToUpdate = Meteor.users.findOne({                                                                          // 41
      _id: identityId                                                                                                  // 41
    });                                                                                                                // 41
                                                                                                                       //
    if (!this.isSimulation && !this.connection.sandstormDb.userHasIdentity(this.userId, identityId)) {                 // 43
      throw new Meteor.Error(403, "identity is not linked to current user: " + identityId);                            // 45
    }                                                                                                                  // 46
                                                                                                                       //
    var newValues = {                                                                                                  // 48
      "profile.name": profile.name,                                                                                    // 49
      "profile.handle": profile.handle,                                                                                // 50
      "profile.pronoun": profile.pronoun                                                                               // 51
    };                                                                                                                 // 48
    Meteor.users.update({                                                                                              // 54
      _id: userToUpdate._id                                                                                            // 54
    }, {                                                                                                               // 54
      $set: newValues                                                                                                  // 54
    });                                                                                                                // 54
                                                                                                                       //
    if (!Meteor.user().hasCompletedSignup) {                                                                           // 56
      Meteor.users.update({                                                                                            // 57
        _id: this.userId                                                                                               // 57
      }, {                                                                                                             // 57
        $set: {                                                                                                        // 57
          hasCompletedSignup: true                                                                                     // 57
        }                                                                                                              // 57
      });                                                                                                              // 57
    }                                                                                                                  // 58
  },                                                                                                                   // 59
  testFirstSignup: function (profile) {                                                                                // 61
    if (!this.userId) {                                                                                                // 62
      throw new Meteor.Error(403, "not logged in");                                                                    // 63
    }                                                                                                                  // 64
                                                                                                                       //
    Meteor.users.update(this.userId, {                                                                                 // 66
      $unset: {                                                                                                        // 66
        hasCompletedSignup: ""                                                                                         // 66
      }                                                                                                                // 66
    });                                                                                                                // 66
  },                                                                                                                   // 67
  uploadProfilePicture: function (identityId) {                                                                        // 69
    check(identityId, String);                                                                                         // 70
                                                                                                                       //
    if (!this.userId || !this.connection.sandstormDb.userHasIdentity(this.userId, identityId)) {                       // 71
      throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);                               // 72
    }                                                                                                                  // 73
                                                                                                                       //
    return this.connection.sandstormDb.newAssetUpload({                                                                // 75
      profilePicture: {                                                                                                // 76
        userId: this.userId,                                                                                           // 76
        identityId: identityId                                                                                         // 76
      }                                                                                                                // 76
    });                                                                                                                // 75
  },                                                                                                                   // 78
  cancelUploadProfilePicture: function (id) {                                                                          // 80
    check(id, String);                                                                                                 // 81
    this.connection.sandstormDb.fulfillAssetUpload(id);                                                                // 82
  },                                                                                                                   // 83
  setPrimaryEmail: function (email) {                                                                                  // 85
    check(email, String);                                                                                              // 86
                                                                                                                       //
    if (!this.userId) {                                                                                                // 87
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 88
    }                                                                                                                  // 89
                                                                                                                       //
    var emails = SandstormDb.getUserEmails(Meteor.user());                                                             // 91
                                                                                                                       //
    if (!_.findWhere(emails, {                                                                                         // 92
      email: email,                                                                                                    // 92
      verified: true                                                                                                   // 92
    })) {                                                                                                              // 92
      throw new Meteor.Error(403, "Not a verified email of the current user: " + email);                               // 93
    }                                                                                                                  // 94
                                                                                                                       //
    Meteor.users.update({                                                                                              // 96
      _id: this.userId                                                                                                 // 96
    }, {                                                                                                               // 96
      $set: {                                                                                                          // 96
        primaryEmail: email                                                                                            // 96
      }                                                                                                                // 96
    });                                                                                                                // 96
  }                                                                                                                    // 97
});                                                                                                                    // 25
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"accounts-ui-server.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/accounts/accounts-ui-server.js                                                                               //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// Meteor permits users to modify their own profile by default, for some reason.                                       // 17
Meteor.users.deny({                                                                                                    // 18
  insert: function () {                                                                                                // 19
    return true;                                                                                                       // 19
  },                                                                                                                   // 19
  update: function () {                                                                                                // 21
    return true;                                                                                                       // 21
  },                                                                                                                   // 21
  remove: function () {                                                                                                // 23
    return true;                                                                                                       // 23
  },                                                                                                                   // 23
  fetch: []                                                                                                            // 25
});                                                                                                                    // 18
Meteor.publish("getMyUsage", function () {                                                                             // 28
  var db = this.connection.sandstormDb;                                                                                // 29
                                                                                                                       //
  if (this.userId) {                                                                                                   // 30
    // TODO(someday): Make this reactive.                                                                              // 31
    var user = Meteor.users.findOne(this.userId);                                                                      // 32
    var usage = db.getMyUsage(user);                                                                                   // 33
    this.added("users", this.userId, {                                                                                 // 34
      pseudoUsage: usage                                                                                               // 35
    });                                                                                                                // 34
  }                                                                                                                    // 37
                                                                                                                       //
  this.ready();                                                                                                        // 39
});                                                                                                                    // 40
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}},"admin":{"admin-alert-server.js":["meteor/meteor","meteor/check","/imports/server/auth.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/admin-alert-server.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Match = void 0,                                                                                                    // 1
    check = void 0;                                                                                                    // 1
module.import("meteor/check", {                                                                                        // 1
  "Match": function (v) {                                                                                              // 1
    Match = v;                                                                                                         // 1
  },                                                                                                                   // 1
  "check": function (v) {                                                                                              // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var checkAuth = void 0;                                                                                                // 1
module.import("/imports/server/auth.js", {                                                                             // 1
  "checkAuth": function (v) {                                                                                          // 1
    checkAuth = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var maintenanceMessageShape = {                                                                                        // 5
  text: String,                                                                                                        // 6
  time: Match.OneOf(Date, undefined, null),                                                                            // 7
  url: Match.Optional(String)                                                                                          // 8
};                                                                                                                     // 5
Meteor.methods({                                                                                                       // 11
  setMaintenanceMessage: function (params) {                                                                           // 12
    checkAuth(undefined);                                                                                              // 13
    check(params, maintenanceMessageShape);                                                                            // 14
    var db = this.connection.sandstormDb; // TODO(soon): make this a single write to a single settings object          // 15
                                                                                                                       //
    db.collections.settings.upsert({                                                                                   // 17
      _id: "adminAlertTime"                                                                                            // 17
    }, {                                                                                                               // 17
      value: params.time                                                                                               // 17
    });                                                                                                                // 17
    db.collections.settings.upsert({                                                                                   // 18
      _id: "adminAlertUrl"                                                                                             // 18
    }, {                                                                                                               // 18
      value: params.url                                                                                                // 18
    });                                                                                                                // 18
    db.collections.settings.upsert({                                                                                   // 19
      _id: "adminAlert"                                                                                                // 19
    }, {                                                                                                               // 19
      value: params.text                                                                                               // 19
    });                                                                                                                // 19
  }                                                                                                                    // 20
});                                                                                                                    // 11
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"admin-user-invite.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/admin-user-invite.js                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
Meteor.publish("allInviteTokens", function () {                                                                        // 1
  var db = this.connection.sandstormDb;                                                                                // 2
                                                                                                                       //
  if (!db.isAdminById(this.userId)) {                                                                                  // 3
    throw new Meteor.Error(403, "User must be admin to list invite tokens.");                                          // 4
  }                                                                                                                    // 5
                                                                                                                       //
  return db.collections.signupKeys.find({});                                                                           // 7
});                                                                                                                    // 8
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"network-capabilities-server.js":["meteor/meteor","meteor/check","/imports/server/auth.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/network-capabilities-server.js                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var check = void 0;                                                                                                    // 1
module.import("meteor/check", {                                                                                        // 1
  "check": function (v) {                                                                                              // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var checkAuth = void 0;                                                                                                // 1
module.import("/imports/server/auth.js", {                                                                             // 1
  "checkAuth": function (v) {                                                                                          // 1
    checkAuth = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
Meteor.publish("adminGrains", function (grainIds) {                                                                    // 5
  // If the caller is an admin, publishes the Grains referred to by the provided list of grain IDs.                    // 6
  // Otherwise, does nothing.                                                                                          // 7
  if (!isAdminById(this.userId)) return [];                                                                            // 8
  check(grainIds, [String]);                                                                                           // 9
  var db = this.connection.sandstormDb;                                                                                // 11
  return db.collections.grains.find({                                                                                  // 12
    _id: {                                                                                                             // 13
      $in: grainIds                                                                                                    // 14
    }                                                                                                                  // 13
  }, {                                                                                                                 // 12
    fields: {                                                                                                          // 17
      identityId: 1,                                                                                                   // 18
      title: 1,                                                                                                        // 19
      packageId: 1                                                                                                     // 20
    }                                                                                                                  // 17
  });                                                                                                                  // 16
});                                                                                                                    // 23
Meteor.publish("adminPackages", function (packageIds) {                                                                // 25
  // If the caller is an admin, publishes the Packages referred to by the provided list of package                     // 26
  // IDs.  Otherwise, does nothing.                                                                                    // 27
  if (!isAdminById(this.userId)) return [];                                                                            // 28
  check(packageIds, [String]);                                                                                         // 29
  var db = this.connection.sandstormDb;                                                                                // 31
  return db.collections.packages.find({                                                                                // 32
    _id: {                                                                                                             // 33
      $in: packageIds                                                                                                  // 34
    }                                                                                                                  // 33
  }, {                                                                                                                 // 32
    fields: {                                                                                                          // 37
      manifest: 1                                                                                                      // 38
    }                                                                                                                  // 37
  });                                                                                                                  // 36
});                                                                                                                    // 41
Meteor.publish("adminIdentities", function (identityIds) {                                                             // 43
  // If the caller is an admin, publishes the identity Users listed by ID in identityIds.                              // 44
  // Otherwise, does nothing.                                                                                          // 45
  if (!isAdminById(this.userId)) return [];                                                                            // 46
  check(identityIds, [String]);                                                                                        // 47
  var db = this.connection.sandstormDb;                                                                                // 49
  return db.collections.users.find({                                                                                   // 50
    _id: {                                                                                                             // 51
      $in: identityIds                                                                                                 // 52
    }                                                                                                                  // 51
  });                                                                                                                  // 50
});                                                                                                                    // 55
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"personalization-server.js":["meteor/meteor","meteor/check","/imports/server/auth.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/personalization-server.js                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var check = void 0;                                                                                                    // 1
module.import("meteor/check", {                                                                                        // 1
  "check": function (v) {                                                                                              // 1
    check = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var checkAuth = void 0;                                                                                                // 1
module.import("/imports/server/auth.js", {                                                                             // 1
  "checkAuth": function (v) {                                                                                          // 1
    checkAuth = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var personalizationMessageShape = {                                                                                    // 5
  serverTitle: String,                                                                                                 // 6
  splashUrl: String,                                                                                                   // 7
  signupDialog: String,                                                                                                // 8
  termsOfServiceUrl: String,                                                                                           // 9
  privacyPolicyUrl: String,                                                                                            // 10
  whitelabelCustomLoginProviderName: String,                                                                           // 12
  whitelabelHideSendFeedback: Boolean,                                                                                 // 13
  whitelabelHideTroubleshooting: Boolean,                                                                              // 14
  whiteLabelHideAbout: Boolean,                                                                                        // 15
  whitelabelUseServerTitleForHomeText: Boolean                                                                         // 16
};                                                                                                                     // 5
Meteor.methods({                                                                                                       // 19
  setPersonalizationSettings: function (params) {                                                                      // 20
    checkAuth(undefined);                                                                                              // 21
    check(params, personalizationMessageShape);                                                                        // 22
    var db = this.connection.sandstormDb; // TODO(soon): make this a single write to a single settings object          // 23
                                                                                                                       //
    db.collections.settings.upsert({                                                                                   // 25
      _id: "serverTitle"                                                                                               // 25
    }, {                                                                                                               // 25
      value: params.serverTitle                                                                                        // 25
    });                                                                                                                // 25
    db.collections.settings.upsert({                                                                                   // 26
      _id: "splashUrl"                                                                                                 // 26
    }, {                                                                                                               // 26
      value: params.splashUrl                                                                                          // 26
    });                                                                                                                // 26
    db.collections.settings.upsert({                                                                                   // 27
      _id: "signupDialog"                                                                                              // 27
    }, {                                                                                                               // 27
      value: params.signupDialog                                                                                       // 27
    });                                                                                                                // 27
    db.collections.settings.upsert({                                                                                   // 28
      _id: "termsUrl"                                                                                                  // 28
    }, {                                                                                                               // 28
      value: params.termsOfServiceUrl                                                                                  // 28
    });                                                                                                                // 28
    db.collections.settings.upsert({                                                                                   // 29
      _id: "privacyUrl"                                                                                                // 29
    }, {                                                                                                               // 29
      value: params.privacyPolicyUrl                                                                                   // 29
    });                                                                                                                // 29
    db.collections.settings.upsert({                                                                                   // 31
      _id: "whitelabelCustomLoginProviderName"                                                                         // 31
    }, {                                                                                                               // 31
      value: params.whitelabelCustomLoginProviderName                                                                  // 32
    });                                                                                                                // 32
    db.collections.settings.upsert({                                                                                   // 33
      _id: "whitelabelHideSendFeedback"                                                                                // 33
    }, {                                                                                                               // 33
      value: params.whitelabelHideSendFeedback                                                                         // 34
    });                                                                                                                // 34
    db.collections.settings.upsert({                                                                                   // 35
      _id: "whitelabelHideTroubleshooting"                                                                             // 35
    }, {                                                                                                               // 35
      value: params.whitelabelHideTroubleshooting                                                                      // 36
    });                                                                                                                // 36
    db.collections.settings.upsert({                                                                                   // 37
      _id: "whiteLabelHideAbout"                                                                                       // 37
    }, {                                                                                                               // 37
      value: params.whiteLabelHideAbout                                                                                // 38
    });                                                                                                                // 38
    db.collections.settings.upsert({                                                                                   // 39
      _id: "whitelabelUseServerTitleForHomeText"                                                                       // 39
    }, {                                                                                                               // 39
      value: params.whitelabelUseServerTitleForHomeText                                                                // 40
    });                                                                                                                // 40
  },                                                                                                                   // 41
  getWhitelabelLogoUploadToken: function () {                                                                          // 43
    checkAuth(undefined);                                                                                              // 44
    var db = this.connection.sandstormDb;                                                                              // 45
    return db.newAssetUpload({                                                                                         // 46
      loginLogo: {}                                                                                                    // 46
    });                                                                                                                // 46
  },                                                                                                                   // 47
  resetWhitelabelLogo: function () {                                                                                   // 49
    checkAuth(undefined);                                                                                              // 50
    var db = this.connection.sandstormDb;                                                                              // 51
    var result = globalDb.collections.settings.findAndModify({                                                         // 52
      query: {                                                                                                         // 53
        _id: "whitelabelCustomLogoAssetId"                                                                             // 53
      },                                                                                                               // 53
      remove: true,                                                                                                    // 54
      fields: {                                                                                                        // 55
        value: 1                                                                                                       // 55
      }                                                                                                                // 55
    });                                                                                                                // 52
                                                                                                                       //
    if (result.ok) {                                                                                                   // 58
      var old = result.value;                                                                                          // 59
                                                                                                                       //
      if (old) {                                                                                                       // 60
        db.unrefStaticAsset(old.value);                                                                                // 61
      }                                                                                                                // 62
    } else {                                                                                                           // 63
      throw new Meteor.Error(500, "Couldn't remove whitelabelCustomLogoAssetId");                                      // 64
    }                                                                                                                  // 65
  }                                                                                                                    // 66
});                                                                                                                    // 19
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"system-status-server.js":["meteor/meteor","meteor/iron:router","meteor/random","crypto","fs","/imports/server/auth.js","/imports/server/constants.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin/system-status-server.js                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Router = void 0;                                                                                                   // 1
module.import("meteor/iron:router", {                                                                                  // 1
  "Router": function (v) {                                                                                             // 1
    Router = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Random = void 0;                                                                                                   // 1
module.import("meteor/random", {                                                                                       // 1
  "Random": function (v) {                                                                                             // 1
    Random = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var Crypto = void 0;                                                                                                   // 1
module.import("crypto", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    Crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var Fs = void 0;                                                                                                       // 1
module.import("fs", {                                                                                                  // 1
  "default": function (v) {                                                                                            // 1
    Fs = v;                                                                                                            // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var checkAuth = void 0;                                                                                                // 1
module.import("/imports/server/auth.js", {                                                                             // 1
  "checkAuth": function (v) {                                                                                          // 1
    checkAuth = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var SANDSTORM_LOGDIR = void 0;                                                                                         // 1
module.import("/imports/server/constants.js", {                                                                        // 1
  "SANDSTORM_LOGDIR": function (v) {                                                                                   // 1
    SANDSTORM_LOGDIR = v;                                                                                              // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
                                                                                                                       //
var hashSessionId = function (sessionId) {                                                                             // 10
  return Crypto.createHash("sha256").update(sessionId).digest("base64");                                               // 11
}; // An in-memory map of random id -> timestamp, used to authorize requests to download the system                    // 12
// log.                                                                                                                // 15
                                                                                                                       //
                                                                                                                       //
var SYSTEM_LOG_DOWNLOAD_TOKENS = {};                                                                                   // 16
var SYSTEM_LOG_DOWNLOAD_TOKEN_VALIDITY_DURATION = 60000; // 60 seconds                                                 // 17
                                                                                                                       //
Meteor.methods({                                                                                                       // 19
  adminGetServerLogDownloadToken: function (setupToken) {                                                              // 20
    checkAuth(setupToken);                                                                                             // 21
    var token = Random.id(22);                                                                                         // 23
    SYSTEM_LOG_DOWNLOAD_TOKENS[token] = Date.now();                                                                    // 24
    setTimeout(function () {                                                                                           // 25
      if (SYSTEM_LOG_DOWNLOAD_TOKENS[token]) {                                                                         // 26
        delete SYSTEM_LOG_DOWNLOAD_TOKENS[token];                                                                      // 27
      }                                                                                                                // 28
    }, SYSTEM_LOG_DOWNLOAD_TOKEN_VALIDITY_DURATION * 2);                                                               // 29
    return token;                                                                                                      // 30
  }                                                                                                                    // 31
});                                                                                                                    // 19
Router.map(function () {                                                                                               // 34
  this.route("adminDownloadServerLog", {                                                                               // 35
    where: "server",                                                                                                   // 36
    path: "/admin/status/server-log/:tokenId",                                                                         // 37
    action: function () {                                                                                              // 38
      var token = this.params.tokenId;                                                                                 // 39
      var response = this.response;                                                                                    // 40
      var issueDate = SYSTEM_LOG_DOWNLOAD_TOKENS[token];                                                               // 41
      delete SYSTEM_LOG_DOWNLOAD_TOKENS[token]; // Clear the token from the in-memory map.                             // 42
                                                                                                                       //
      if (issueDate === undefined || issueDate + SYSTEM_LOG_DOWNLOAD_TOKEN_VALIDITY_DURATION < Date.now()) {           // 43
        // Require download to start within 60 seconds of the adminGetServerLogDownloadToken                           // 45
        // method being called.                                                                                        // 46
        response.writeHead(404, {                                                                                      // 47
          "Content-Type": "text/plain"                                                                                 // 47
        });                                                                                                            // 47
        return response.end("Invalid server log download token.");                                                     // 48
      }                                                                                                                // 49
                                                                                                                       //
      var logFilePath = SANDSTORM_LOGDIR + "/sandstorm.log";                                                           // 51
      var fd = Fs.openSync(logFilePath, "r");                                                                          // 52
      var stats = Fs.fstatSync(fd);                                                                                    // 53
      var initialSize = stats.size;                                                                                    // 54
      var readStream = Fs.createReadStream(undefined, {                                                                // 55
        fd: fd,                                                                                                        // 56
        start: 0,                                                                                                      // 57
        end: initialSize                                                                                               // 58
      });                                                                                                              // 55
      response.writeHead(200, {                                                                                        // 61
        "Content-Length": initialSize,                                                                                 // 62
        "Content-Type": "text/plain",                                                                                  // 63
        "content-Disposition": "attachment;filename=\"sandstorm.log\""                                                 // 64
      });                                                                                                              // 61
      readStream.pipe(response);                                                                                       // 66
    }                                                                                                                  // 67
  });                                                                                                                  // 35
});                                                                                                                    // 69
Meteor.publish("systemStatus", function () {                                                                           // 71
  // A pseudocollection containing the number of active grain sessions and unique                                      // 72
  // userIds associated with these sessions                                                                            // 73
  var db = this.connection.sandstormDb;                                                                                // 74
                                                                                                                       //
  if (!db.isAdminById(this.userId)) {                                                                                  // 75
    throw new Meteor.Error(403, "User must be admin to view system status.");                                          // 76
  }                                                                                                                    // 77
                                                                                                                       //
  var _this = this;                                                                                                    // 79
                                                                                                                       //
  var userIdToSessionHashes = {}; // Maps userId => list of (sha256(sessionId))                                        // 80
                                                                                                                       //
  var grainIdToSessionHashes = {}; // Maps grainId => list of (sha256(sessionId))                                      // 81
                                                                                                                       //
  var userIdCount = 0;                                                                                                 // 83
  var grainIdCount = 0;                                                                                                // 84
  this.added("systemStatus", "globalStatus", {                                                                         // 86
    activeUsers: userIdCount,                                                                                          // 87
    activeGrains: grainIdCount                                                                                         // 88
  });                                                                                                                  // 86
  var query = db.collections.sessions.find();                                                                          // 91
  var handle = query.observe({                                                                                         // 92
    added: function (session) {                                                                                        // 93
      var hashedSessionId = hashSessionId(session._id);                                                                // 94
      var changed = {}; // Update userId cache.                                                                        // 95
                                                                                                                       //
      if (session.userId) {                                                                                            // 98
        if (userIdToSessionHashes[session.userId] === undefined) {                                                     // 99
          userIdToSessionHashes[session.userId] = [];                                                                  // 100
          userIdCount = userIdCount + 1;                                                                               // 101
          changed.activeUsers = userIdCount;                                                                           // 102
        }                                                                                                              // 103
                                                                                                                       //
        userIdToSessionHashes[session.userId] = _.union(userIdToSessionHashes[session.userId], [hashedSessionId]);     // 105
      } // Update grainId cache.                                                                                       // 107
                                                                                                                       //
                                                                                                                       //
      if (grainIdToSessionHashes[session.grainId] === undefined) {                                                     // 110
        grainIdToSessionHashes[session.grainId] = [];                                                                  // 111
        grainIdCount = grainIdCount + 1;                                                                               // 112
        changed.activeGrains = grainIdCount;                                                                           // 113
      }                                                                                                                // 114
                                                                                                                       //
      grainIdToSessionHashes[session.grainId] = _.union(grainIdToSessionHashes[session.grainId], [hashedSessionId]);   // 116
                                                                                                                       //
      if (!_.isEmpty(changed)) {                                                                                       // 119
        _this.changed("systemStatus", "globalStatus", changed);                                                        // 120
      }                                                                                                                // 121
    },                                                                                                                 // 122
    removed: function (session) {                                                                                      // 124
      var hashedSessionId = hashSessionId(session._id);                                                                // 125
      var changed = {};                                                                                                // 126
                                                                                                                       //
      if (session.userId) {                                                                                            // 128
        userIdToSessionHashes[session.userId] = _.without(userIdToSessionHashes[session.userId], hashedSessionId);     // 129
                                                                                                                       //
        if (userIdToSessionHashes[session.userId].length === 0) {                                                      // 132
          delete userIdToSessionHashes[session.userId];                                                                // 133
          userIdCount = userIdCount - 1;                                                                               // 134
          changed.activeUsers = userIdCount;                                                                           // 135
        }                                                                                                              // 136
      }                                                                                                                // 137
                                                                                                                       //
      grainIdToSessionHashes[session.grainId] = _.without(grainIdToSessionHashes[session.grainId], hashedSessionId);   // 139
                                                                                                                       //
      if (grainIdToSessionHashes[session.grainId].length === 0) {                                                      // 142
        delete grainIdToSessionHashes[session.grainId];                                                                // 143
        grainIdCount = grainIdCount - 1;                                                                               // 144
        changed.activeGrains = grainIdCount;                                                                           // 145
      }                                                                                                                // 146
                                                                                                                       //
      if (!_.isEmpty(changed)) {                                                                                       // 148
        _this.changed("systemStatus", "globalStatus", changed);                                                        // 149
      }                                                                                                                // 150
    }                                                                                                                  // 151
  });                                                                                                                  // 92
  this.onStop(function () {                                                                                            // 154
    handle.stop();                                                                                                     // 155
  });                                                                                                                  // 156
  this.ready();                                                                                                        // 157
});                                                                                                                    // 158
Meteor.publish("adminDemoUsers", function () {                                                                         // 160
  // Publishes expiry information about demo accounts to admins, so demo users can be counted                          // 161
  // clientside and reactively updated.                                                                                // 162
  var db = this.connection.sandstormDb;                                                                                // 163
                                                                                                                       //
  if (!db.isAdminById(this.userId)) {                                                                                  // 164
    throw new Meteor.Error(403, "User must be admin to view system status.");                                          // 165
  }                                                                                                                    // 166
                                                                                                                       //
  return db.collections.users.find({                                                                                   // 168
    expires: {                                                                                                         // 169
      $gt: new Date()                                                                                                  // 170
    },                                                                                                                 // 169
    loginIdentities: {                                                                                                 // 172
      $exists: true                                                                                                    // 173
    }                                                                                                                  // 172
  }, {                                                                                                                 // 168
    expires: 1,                                                                                                        // 176
    loginIdentities: 1                                                                                                 // 177
  });                                                                                                                  // 175
});                                                                                                                    // 179
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"drivers":{"external-ui-view.js":["babel-runtime/helpers/possibleConstructorReturn","babel-runtime/helpers/inherits","babel-runtime/helpers/classCallCheck","/imports/server/persistent.js","/imports/server/networking.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/drivers/external-ui-view.js                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var PersistentImpl = void 0;                                                                                           // 1
module.import("/imports/server/persistent.js", {                                                                       // 1
  "PersistentImpl": function (v) {                                                                                     // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var ssrfSafeLookup = void 0;                                                                                           // 1
module.import("/imports/server/networking.js", {                                                                       // 1
  "ssrfSafeLookup": function (v) {                                                                                     // 1
    ssrfSafeLookup = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 20
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 21
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 22
                                                                                                                       //
var Http = Npm.require("http");                                                                                        // 23
                                                                                                                       //
var Https = Npm.require("https");                                                                                      // 24
                                                                                                                       //
var ApiSession = Capnp.importSystem("sandstorm/api-session.capnp").ApiSession;                                         // 25
var PersistentApiSession = Capnp.importSystem("sandstorm/api-session-impl.capnp").PersistentApiSession;                // 26
                                                                                                                       //
ExternalUiView = function () {                                                                                         // 29
  function ExternalUiView(url, token) {                                                                                // 30
    (0, _classCallCheck3.default)(this, ExternalUiView);                                                               // 30
    this.url = url;                                                                                                    // 31
    this.token = token;                                                                                                // 32
  }                                                                                                                    // 33
                                                                                                                       //
  ExternalUiView.prototype.newSession = function () {                                                                  // 29
    function newSession(userInfo, context, sessionType, sessionParams) {                                               // 29
      var _this2 = this;                                                                                               // 35
                                                                                                                       //
      if (sessionType !== ApiSession.typeId) {                                                                         // 36
        throw new Error("SessionType must be ApiSession.");                                                            // 37
      }                                                                                                                // 38
                                                                                                                       //
      var options = {};                                                                                                // 40
                                                                                                                       //
      if (this.token) {                                                                                                // 42
        options.headers = {                                                                                            // 43
          authorization: "Bearer " + this.token                                                                        // 44
        };                                                                                                             // 43
      }                                                                                                                // 46
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 48
        return {                                                                                                       // 49
          session: new Capnp.Capability(new ExternalWebSession(_this2.url, options, globalDb), ApiSession)             // 50
        };                                                                                                             // 49
      });                                                                                                              // 53
    }                                                                                                                  // 54
                                                                                                                       //
    return newSession;                                                                                                 // 29
  }();                                                                                                                 // 29
                                                                                                                       //
  return ExternalUiView;                                                                                               // 29
}();                                                                                                                   // 29
                                                                                                                       //
function getOAuthServiceInfo(url) {                                                                                    // 57
  // TODO(soon): Define a table somewhere (probably in a .capnp file) mapping API hosts to OAuth                       // 58
  //   metadata.                                                                                                       // 59
  if (url.startsWith("https://apidata.googleusercontent.com/") || url.startsWith("https://www.googleapis.com/")) {     // 60
    return {                                                                                                           // 62
      service: "google",                                                                                               // 63
      endpoint: "https://www.googleapis.com/oauth2/v4/token"                                                           // 64
    };                                                                                                                 // 62
  } else if (url.startsWith("https://api.github.com/users")) {                                                         // 66
    return {                                                                                                           // 67
      service: "github",                                                                                               // 68
      endpoint: "https://github.com/login/oauth/access_token"                                                          // 69
    };                                                                                                                 // 67
  } else {                                                                                                             // 71
    return null;                                                                                                       // 72
  }                                                                                                                    // 73
}                                                                                                                      // 74
                                                                                                                       //
function refreshOAuth(url, refreshToken) {                                                                             // 76
  // TODO(perf): Cache access tokens until they expire? Currently we re-do the refresh on every                        // 77
  //   restore. In particular, this means we always drop the first access token returned (which                        // 78
  //   is returned together with the refresh token) and then immediately request a new one.                            // 79
  var serviceInfo = getOAuthServiceInfo(url);                                                                          // 81
                                                                                                                       //
  if (!serviceInfo) {                                                                                                  // 82
    throw new Error("Don't know how to OAuth for: " + url);                                                            // 83
  }                                                                                                                    // 84
                                                                                                                       //
  var config = ServiceConfiguration.configurations.findOne({                                                           // 86
    service: serviceInfo.service                                                                                       // 86
  });                                                                                                                  // 86
                                                                                                                       //
  if (!config) {                                                                                                       // 87
    throw new Error("can't refresh OAuth token for service that isn't configured: " + serviceInfo.service);            // 88
  }                                                                                                                    // 90
                                                                                                                       //
  var response = HTTP.post(serviceInfo.endpoint, {                                                                     // 92
    headers: {                                                                                                         // 93
      "Content-Type": "application/x-www-form-urlencoded",                                                             // 94
      "Accept": "application/json"                                                                                     // 95
    },                                                                                                                 // 93
    content: "client_id=" + encodeURIComponent(config.clientId) + "&client_secret=" + encodeURIComponent(config.secret) + "&refresh_token=" + encodeURIComponent(refreshToken) + "&grant_type=refresh_token"
  });                                                                                                                  // 92
  return response.data;                                                                                                // 103
}                                                                                                                      // 104
                                                                                                                       //
function newExternalHttpSession(url, auth, db, saveTemplate) {                                                         // 106
  // `url` and `auth` are the corresponding members of `ApiToken.frontendRef.http`.                                    // 107
  var createCap = function (authorization) {                                                                           // 109
    return new Capnp.Capability(new ExternalWebSession(url, authorization ? {                                          // 110
      headers: {                                                                                                       // 111
        authorization: authorization                                                                                   // 111
      }                                                                                                                // 111
    } : {}, db, saveTemplate), PersistentApiSession);                                                                  // 111
  };                                                                                                                   // 113
                                                                                                                       //
  if (auth.refresh) {                                                                                                  // 115
    return createCap("Bearer " + refreshOAuth(url, auth.refresh).access_token);                                        // 116
  } else if (auth.bearer) {                                                                                            // 117
    return createCap("Bearer " + auth.bearer);                                                                         // 118
  } else if (auth.basic) {                                                                                             // 119
    var userpass = [auth.basic.username, auth.basic.password].join(":");                                               // 120
    return createCap("Basic " + new Buffer(userpass, "utf8").toString("base64"));                                      // 121
  } else {                                                                                                             // 122
    return createCap(null);                                                                                            // 123
  }                                                                                                                    // 124
}                                                                                                                      // 125
                                                                                                                       //
function registerHttpApiFrontendRef(registry) {                                                                        // 127
  registry.register({                                                                                                  // 128
    frontendRefField: "http",                                                                                          // 129
    typeId: ApiSession.typeId,                                                                                         // 130
    restore: function (db, saveTemplate, value) {                                                                      // 132
      return newExternalHttpSession(value.url, value.auth, db, saveTemplate);                                          // 133
    },                                                                                                                 // 134
    validate: function (db, session, request) {                                                                        // 136
      check(request, {                                                                                                 // 137
        url: String,                                                                                                   // 138
        auth: Match.OneOf({                                                                                            // 139
          none: null                                                                                                   // 140
        }, {                                                                                                           // 140
          bearer: String                                                                                               // 141
        }, {                                                                                                           // 141
          basic: {                                                                                                     // 142
            username: String,                                                                                          // 142
            password: String                                                                                           // 142
          }                                                                                                            // 142
        }, {                                                                                                           // 142
          oauth: {                                                                                                     // 143
            credentialToken: String,                                                                                   // 143
            credentialSecret: String                                                                                   // 143
          }                                                                                                            // 143
        })                                                                                                             // 143
      });                                                                                                              // 137
                                                                                                                       //
      if (!request.url.startsWith("https://") && !request.url.startsWith("http://")) {                                 // 146
        throw new Meteor.Error(400, "URL must be HTTP or HTTPS.");                                                     // 148
      } // Check for URL patterns.                                                                                     // 149
                                                                                                                       //
                                                                                                                       //
      var parsedUrl = Url.parse(request.url);                                                                          // 152
                                                                                                                       //
      if (parsedUrl.auth) {                                                                                            // 154
        var parts = parsedUrl.auth.split(":");                                                                         // 155
                                                                                                                       //
        if (parts.length === 2) {                                                                                      // 156
          if ("none" in request.auth) {                                                                                // 157
            request.auth = {                                                                                           // 158
              basic: {                                                                                                 // 158
                username: parts[0],                                                                                    // 158
                password: parts[1]                                                                                     // 158
              }                                                                                                        // 158
            };                                                                                                         // 158
          } else {                                                                                                     // 159
            throw new Meteor.Error(400, "Can't support multiple authentication mechanisms at once");                   // 160
          }                                                                                                            // 161
        }                                                                                                              // 162
                                                                                                                       //
        parsedUrl.auth = null;                                                                                         // 164
        request.url = Url.format(parsedUrl);                                                                           // 165
      }                                                                                                                // 166
                                                                                                                       //
      if (parsedUrl.hash) {                                                                                            // 168
        if ("none" in request.auth) {                                                                                  // 169
          request.auth = {                                                                                             // 170
            bearer: parsedUrl.hash.slice(1)                                                                            // 170
          };                                                                                                           // 170
        } else {                                                                                                       // 171
          throw new Meteor.Error(400, "Can't support multiple authentication mechanisms at once");                     // 172
        }                                                                                                              // 173
                                                                                                                       //
        parsedUrl.hash = null;                                                                                         // 175
        request.url = Url.format(parsedUrl);                                                                           // 176
      }                                                                                                                // 177
                                                                                                                       //
      if (request.auth.oauth) {                                                                                        // 179
        // We did an OAuth handshake client-side.                                                                      // 180
        var oauthResult = OAuth.retrieveCredential(request.auth.oauth.credentialToken, request.auth.oauth.credentialSecret);
                                                                                                                       //
        if (oauthResult instanceof Error) {                                                                            // 184
          throw oauthResult;                                                                                           // 185
        }                                                                                                              // 186
                                                                                                                       //
        var serviceData = oauthResult.serviceData;                                                                     // 188
                                                                                                                       //
        if (serviceData.refreshToken) {                                                                                // 189
          request.auth = {                                                                                             // 190
            refresh: serviceData.refreshToken                                                                          // 190
          };                                                                                                           // 190
        } else {                                                                                                       // 191
          request.auth = {                                                                                             // 192
            bearer: serviceData.accessToken                                                                            // 192
          };                                                                                                           // 192
        } // TODO(security): We could maybe add a MembraneRequirement that this user account                           // 193
        //   possesses credentials for this OAuth service. Conversely, perhaps if an authCode was                      // 196
        //   specified, we should automatically add the associated credential to the user's                            // 197
        //   account? (As a non-login credential.)                                                                     // 198
                                                                                                                       //
      }                                                                                                                // 199
                                                                                                                       //
      var descriptor = {                                                                                               // 201
        tags: [{                                                                                                       // 201
          id: ApiSession.typeId                                                                                        // 201
        }]                                                                                                             // 201
      };                                                                                                               // 201
      return {                                                                                                         // 202
        descriptor: descriptor,                                                                                        // 202
        requirements: [],                                                                                              // 202
        frontendRef: request                                                                                           // 202
      };                                                                                                               // 202
    },                                                                                                                 // 203
    query: function (db, userAccountId, tagValue) {                                                                    // 205
      var tag = tagValue ? Capnp.parse(ApiSession.PowerboxTag, tagValue) : {};                                         // 206
      var options = [];                                                                                                // 208
                                                                                                                       //
      if (tag.canonicalUrl && (tag.canonicalUrl.startsWith("https://") || tag.canonicalUrl.startsWith("http://"))) {   // 210
        var serviceInfo = getOAuthServiceInfo(tag.canonicalUrl);                                                       // 213
                                                                                                                       //
        if (serviceInfo && tag.oauthScopes) {                                                                          // 214
          // Note: We don't check if the service is configured, because it's useful to show the                        // 215
          //   user that the option exists but inform them that it will only work if the admin                         // 216
          //   configures this service.                                                                                // 217
          options.push({                                                                                               // 218
            _id: "http-oauth-" + tag.canonicalUrl,                                                                     // 219
            cardTemplate: "httpOAuthPowerboxCard",                                                                     // 220
            configureTemplate: "httpOAuthPowerboxConfiguration",                                                       // 221
            httpUrl: tag.canonicalUrl,                                                                                 // 222
            oauthServiceInfo: serviceInfo,                                                                             // 223
            oauthScopes: tag.oauthScopes                                                                               // 224
          });                                                                                                          // 218
        } else {                                                                                                       // 226
          // TODO(soon): Support tag.authentication.                                                                   // 227
          options.push({                                                                                               // 229
            _id: "http-url-" + tag.canonicalUrl,                                                                       // 230
            frontendRef: {                                                                                             // 231
              http: {                                                                                                  // 231
                url: tag.canonicalUrl,                                                                                 // 231
                auth: {                                                                                                // 231
                  none: null                                                                                           // 231
                }                                                                                                      // 231
              }                                                                                                        // 231
            },                                                                                                         // 231
            cardTemplate: "httpUrlPowerboxCard"                                                                        // 232
          });                                                                                                          // 229
        }                                                                                                              // 234
      } // Always offer the user the option to connect to an arbitrary URL of their choosing, even                     // 235
      // if canonicalUrl exists.                                                                                       // 238
                                                                                                                       //
                                                                                                                       //
      options.push({                                                                                                   // 239
        _id: "http-arbitrary",                                                                                         // 240
        cardTemplate: "httpArbitraryPowerboxCard",                                                                     // 241
        configureTemplate: "httpArbitraryPowerboxConfiguration"                                                        // 242
      });                                                                                                              // 239
      return options;                                                                                                  // 245
    }                                                                                                                  // 246
  });                                                                                                                  // 128
}                                                                                                                      // 248
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 250
  registerHttpApiFrontendRef(globalFrontendRefRegistry);                                                               // 250
}); // =======================================================================================                         // 250
                                                                                                                       //
var responseCodes = {                                                                                                  // 254
  200: {                                                                                                               // 255
    type: "content",                                                                                                   // 255
    code: "ok"                                                                                                         // 255
  },                                                                                                                   // 255
  201: {                                                                                                               // 256
    type: "content",                                                                                                   // 256
    code: "created"                                                                                                    // 256
  },                                                                                                                   // 256
  202: {                                                                                                               // 257
    type: "content",                                                                                                   // 257
    code: "accepted"                                                                                                   // 257
  },                                                                                                                   // 257
  204: {                                                                                                               // 258
    type: "noContent",                                                                                                 // 258
    shouldResetForm: false                                                                                             // 258
  },                                                                                                                   // 258
  205: {                                                                                                               // 259
    type: "noContent",                                                                                                 // 259
    shouldResetForm: true                                                                                              // 259
  },                                                                                                                   // 259
  // Unsupported until something demonstrates need.                                                                    // 261
  // 206: {type: 'noContent'},                                                                                         // 262
  // 300: {type: 'redirect'},                                                                                          // 263
  301: {                                                                                                               // 264
    type: "redirect",                                                                                                  // 264
    switchToGet: true,                                                                                                 // 264
    isPermanent: true                                                                                                  // 264
  },                                                                                                                   // 264
  302: {                                                                                                               // 265
    type: "redirect",                                                                                                  // 265
    switchToGet: true,                                                                                                 // 265
    isPermanent: false                                                                                                 // 265
  },                                                                                                                   // 265
  303: {                                                                                                               // 266
    type: "redirect",                                                                                                  // 266
    switchToGet: true,                                                                                                 // 266
    isPermanent: false                                                                                                 // 266
  },                                                                                                                   // 266
  // Unsupported until something demonstrates need.                                                                    // 268
  // 304: {type: 'redirect'},                                                                                          // 269
  // 305: {type: 'redirect'},                                                                                          // 270
  307: {                                                                                                               // 271
    type: "redirect",                                                                                                  // 271
    switchToGet: false,                                                                                                // 271
    isPermanent: false                                                                                                 // 271
  },                                                                                                                   // 271
  308: {                                                                                                               // 272
    type: "redirect",                                                                                                  // 272
    switchToGet: false,                                                                                                // 272
    isPermanent: true                                                                                                  // 272
  },                                                                                                                   // 272
  400: {                                                                                                               // 273
    type: "clientError",                                                                                               // 273
    clientErrorCode: "badRequest",                                                                                     // 273
    descriptionHtml: "Bad Request"                                                                                     // 273
  },                                                                                                                   // 273
  403: {                                                                                                               // 274
    type: "clientError",                                                                                               // 274
    clientErrorCode: "forbidden",                                                                                      // 274
    descriptionHtml: "Forbidden"                                                                                       // 274
  },                                                                                                                   // 274
  404: {                                                                                                               // 275
    type: "clientError",                                                                                               // 275
    clientErrorCode: "notFound",                                                                                       // 275
    descriptionHtml: "Not Found"                                                                                       // 275
  },                                                                                                                   // 275
  405: {                                                                                                               // 276
    type: "clientError",                                                                                               // 276
    clientErrorCode: "methodNotAllowed",                                                                               // 276
    descriptionHtml: "Method Not Allowed"                                                                              // 276
  },                                                                                                                   // 276
  406: {                                                                                                               // 277
    type: "clientError",                                                                                               // 277
    clientErrorCode: "notAcceptable",                                                                                  // 277
    descriptionHtml: "Not Acceptable"                                                                                  // 277
  },                                                                                                                   // 277
  409: {                                                                                                               // 278
    type: "clientError",                                                                                               // 278
    clientErrorCode: "conflict",                                                                                       // 278
    descriptionHtml: "Conflict"                                                                                        // 278
  },                                                                                                                   // 278
  410: {                                                                                                               // 279
    type: "clientError",                                                                                               // 279
    clientErrorCode: "gone",                                                                                           // 279
    descriptionHtml: "Gone"                                                                                            // 279
  },                                                                                                                   // 279
  413: {                                                                                                               // 280
    type: "clientError",                                                                                               // 280
    clientErrorCode: "requestEntityTooLarge",                                                                          // 280
    descriptionHtml: "Request Entity Too Large"                                                                        // 280
  },                                                                                                                   // 280
  414: {                                                                                                               // 281
    type: "clientError",                                                                                               // 281
    clientErrorCode: "requestUriTooLong",                                                                              // 281
    descriptionHtml: "Request-URI Too Long"                                                                            // 281
  },                                                                                                                   // 281
  415: {                                                                                                               // 282
    type: "clientError",                                                                                               // 282
    clientErrorCode: "unsupportedMediaType",                                                                           // 282
    descriptionHtml: "Unsupported Media Type"                                                                          // 282
  },                                                                                                                   // 282
  418: {                                                                                                               // 283
    type: "clientError",                                                                                               // 283
    clientErrorCode: "imATeapot",                                                                                      // 283
    descriptionHtml: "I'm a teapot"                                                                                    // 283
  },                                                                                                                   // 283
  500: {                                                                                                               // 284
    type: "serverError"                                                                                                // 284
  },                                                                                                                   // 284
  501: {                                                                                                               // 285
    type: "serverError"                                                                                                // 285
  },                                                                                                                   // 285
  502: {                                                                                                               // 286
    type: "serverError"                                                                                                // 286
  },                                                                                                                   // 286
  503: {                                                                                                               // 287
    type: "serverError"                                                                                                // 287
  },                                                                                                                   // 287
  504: {                                                                                                               // 288
    type: "serverError"                                                                                                // 288
  },                                                                                                                   // 288
  505: {                                                                                                               // 289
    type: "serverError"                                                                                                // 289
  }                                                                                                                    // 289
};                                                                                                                     // 254
                                                                                                                       //
ExternalWebSession = function (_PersistentImpl) {                                                                      // 292
  (0, _inherits3.default)(ExternalWebSession, _PersistentImpl);                                                        // 292
                                                                                                                       //
  function ExternalWebSession(url, options, db, saveTemplate) {                                                        // 293
    (0, _classCallCheck3.default)(this, ExternalWebSession);                                                           // 293
                                                                                                                       //
    // TODO(soon): Support HTTP proxy.                                                                                 // 296
    var _this3 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));         // 293
                                                                                                                       //
    var safe = ssrfSafeLookup(db, url);                                                                                // 297
    if (!options) options = {};                                                                                        // 299
    if (!options.headers) options.headers = {};                                                                        // 300
    options.headers.host = safe.host;                                                                                  // 301
    options.servername = safe.host.split(":")[0];                                                                      // 302
                                                                                                                       //
    if (!saveTemplate) {                                                                                               // 304
      // enable backwards-compatibilty tweaks.                                                                         // 305
      _this3.fromHackSession = true;                                                                                   // 306
    }                                                                                                                  // 307
                                                                                                                       //
    var parsedUrl = Url.parse(safe.url);                                                                               // 309
    _this3.host = parsedUrl.hostname;                                                                                  // 310
                                                                                                                       //
    if (_this3.fromHackSession) {// HackSessionContext.getExternalUiView() apparently ignored any path on the URL. Whoops.
    } else {                                                                                                           // 313
      if (parsedUrl.path === "/") {                                                                                    // 314
        // The URL parser says path = "/" for both "http://foo" and "http://foo/". We want to be                       // 315
        // strict, though.                                                                                             // 316
        _this3.path = url.endsWith("/") ? "/" : "";                                                                    // 317
      } else {                                                                                                         // 318
        _this3.path = parsedUrl.path;                                                                                  // 319
      }                                                                                                                // 320
    }                                                                                                                  // 321
                                                                                                                       //
    _this3.port = parsedUrl.port;                                                                                      // 323
    _this3.protocol = parsedUrl.protocol;                                                                              // 324
    _this3.options = options;                                                                                          // 325
    return _this3;                                                                                                     // 293
  }                                                                                                                    // 326
                                                                                                                       //
  ExternalWebSession.prototype.get = function () {                                                                     // 292
    function get(path, context) {                                                                                      // 292
      return this._requestHelper("GET", path, context);                                                                // 329
    }                                                                                                                  // 330
                                                                                                                       //
    return get;                                                                                                        // 292
  }();                                                                                                                 // 292
                                                                                                                       //
  ExternalWebSession.prototype.post = function () {                                                                    // 292
    function post(path, content, context) {                                                                            // 292
      return this._requestHelper("POST", path, context, content.content, content.mimeType);                            // 333
    }                                                                                                                  // 334
                                                                                                                       //
    return post;                                                                                                       // 292
  }();                                                                                                                 // 292
                                                                                                                       //
  ExternalWebSession.prototype.put = function () {                                                                     // 292
    function put(path, content, context) {                                                                             // 292
      return this._requestHelper("PUT", path, context, content.content, content.mimeType);                             // 337
    }                                                                                                                  // 338
                                                                                                                       //
    return put;                                                                                                        // 292
  }();                                                                                                                 // 292
                                                                                                                       //
  ExternalWebSession.prototype.delete = function () {                                                                  // 292
    function _delete(path, context) {                                                                                  // 292
      return this._requestHelper("DELETE", path, context);                                                             // 341
    }                                                                                                                  // 342
                                                                                                                       //
    return _delete;                                                                                                    // 292
  }(); // TODO(someday): implement streaming and websockets for ExternalWebSession                                     // 292
  //postStreaming(path, mimeType, context) {                                                                           // 345
  //}                                                                                                                  // 346
  //putStreaming(path, mimeType, context) {                                                                            // 348
  //}                                                                                                                  // 349
  //openWebSocket(path, context, protocol, clientStream) {                                                             // 351
  //}                                                                                                                  // 352
                                                                                                                       //
                                                                                                                       //
  ExternalWebSession.prototype._requestHelper = function () {                                                          // 292
    function _requestHelper(method, path, context, content, contentType) {                                             // 292
      var _this4 = this;                                                                                               // 354
                                                                                                                       //
      var _this = this;                                                                                                // 355
                                                                                                                       //
      var session = _this;                                                                                             // 356
      return new Promise(function (resolve, reject) {                                                                  // 357
        var options = _.clone(session.options);                                                                        // 358
                                                                                                                       //
        options.headers = options.headers || {};                                                                       // 359
                                                                                                                       //
        if (_this4.fromHackSession) {                                                                                  // 361
          // According to the specification of `WebSession`, `path` should not contain a                               // 362
          // leading slash, and therefore we need to prepend "/". However, for a long time                             // 363
          // this implementation did not in fact prepend a "/". Since some apps might rely on                          // 364
          // that behavior, we only prepend "/" if the path does not start with "/".                                   // 365
          //                                                                                                           // 366
          // TODO(soon): Once apps have updated, prepend "/" unconditionally.                                          // 367
          options.path = path.startsWith("/") ? path : "/" + path;                                                     // 368
        } else {                                                                                                       // 369
          options.path = _this4.path + "/" + path;                                                                     // 370
        }                                                                                                              // 371
                                                                                                                       //
        options.method = method;                                                                                       // 373
                                                                                                                       //
        if (contentType) {                                                                                             // 374
          options.headers["content-type"] = contentType;                                                               // 375
        } // set accept header                                                                                         // 376
                                                                                                                       //
                                                                                                                       //
        if ("accept" in context) {                                                                                     // 379
          options.headers.accept = context.accept.map(function (acceptedType) {                                        // 380
            return acceptedType.mimeType + "; " + acceptedType.qValue;                                                 // 381
          }).join(", ");                                                                                               // 382
        } else if (!("accept" in options.headers)) {                                                                   // 383
          options.headers.accept = "*/*";                                                                              // 384
        } // set cookies                                                                                               // 385
                                                                                                                       //
                                                                                                                       //
        if (context.cookies && context.cookies.length > 0) {                                                           // 388
          options.headers.cookies = options.headers.cookies || "";                                                     // 389
          context.cookies.forEach(function (keyVal) {                                                                  // 390
            options.headers.cookies += keyVal.key + "=" + keyVal.val + ",";                                            // 391
          });                                                                                                          // 392
          options.headers.cookies = options.headers.cookies.slice(0, -1);                                              // 393
        }                                                                                                              // 394
                                                                                                                       //
        options.host = session.host;                                                                                   // 396
        options.port = session.port;                                                                                   // 397
        var requestMethod = Http.request;                                                                              // 399
                                                                                                                       //
        if (session.protocol === "https:") {                                                                           // 400
          requestMethod = Https.request;                                                                               // 401
        }                                                                                                              // 402
                                                                                                                       //
        req = requestMethod(options, function (resp) {                                                                 // 404
          try {                                                                                                        // 405
            var buffers = [];                                                                                          // 406
            var statusInfo = responseCodes[resp.statusCode];                                                           // 407
            var rpcResponse = {};                                                                                      // 409
                                                                                                                       //
            switch (statusInfo ? statusInfo.type : resp.statusCode) {                                                  // 411
              case "content":                                                                                          // 412
                resp.on("data", function (buf) {                                                                       // 413
                  buffers.push(buf);                                                                                   // 414
                });                                                                                                    // 415
                resp.on("end", function () {                                                                           // 417
                  var content = {};                                                                                    // 418
                  rpcResponse.content = content;                                                                       // 419
                  content.statusCode = statusInfo.code;                                                                // 421
                  if ("content-encoding" in resp.headers) content.encoding = resp.headers["content-encoding"];         // 422
                  if ("content-language" in resp.headers) content.language = resp.headers["content-language"];         // 423
                  if ("content-type" in resp.headers) content.language = resp.headers["content-type"];                 // 424
                                                                                                                       //
                  if ("content-disposition" in resp.headers) {                                                         // 425
                    var disposition = resp.headers["content-disposition"];                                             // 426
                    var parts = disposition.split(";");                                                                // 427
                                                                                                                       //
                    if (parts[0].toLowerCase().trim() === "attachment") {                                              // 428
                      parts.forEach(function (part) {                                                                  // 429
                        var splitPart = part.split("=");                                                               // 430
                                                                                                                       //
                        if (splitPart[0].toLowerCase().trim() === "filename") {                                        // 431
                          content.disposition = {                                                                      // 432
                            download: splitPart[1].trim()                                                              // 432
                          };                                                                                           // 432
                        }                                                                                              // 433
                      });                                                                                              // 434
                    }                                                                                                  // 435
                  }                                                                                                    // 436
                                                                                                                       //
                  content.body = {};                                                                                   // 438
                  content.body.bytes = Buffer.concat(buffers);                                                         // 439
                  resolve(rpcResponse);                                                                                // 441
                });                                                                                                    // 442
                break;                                                                                                 // 443
                                                                                                                       //
              case "noContent":                                                                                        // 444
                var noContent = {};                                                                                    // 445
                rpcResponse.noContent = noContent;                                                                     // 446
                noContent.setShouldResetForm = statusInfo.shouldResetForm;                                             // 447
                resolve(rpcResponse);                                                                                  // 448
                break;                                                                                                 // 449
                                                                                                                       //
              case "redirect":                                                                                         // 450
                var redirect = {};                                                                                     // 451
                rpcResponse.redirect = redirect;                                                                       // 452
                redirect.isPermanent = statusInfo.isPermanent;                                                         // 453
                redirect.switchToGet = statusInfo.switchToGet;                                                         // 454
                if ("location" in resp.headers) redirect.location = resp.headers.location;                             // 455
                resolve(rpcResponse);                                                                                  // 456
                break;                                                                                                 // 457
                                                                                                                       //
              case "clientError":                                                                                      // 458
                var clientError = {};                                                                                  // 459
                rpcResponse.clientError = clientError;                                                                 // 460
                clientError.statusCode = statusInfo.clientErrorCode; // TODO(soon): Pass along the body from upstream.
                                                                                                                       //
                clientError.descriptionHtml = statusInfo.descriptionHtml;                                              // 463
                resolve(rpcResponse);                                                                                  // 464
                break;                                                                                                 // 465
                                                                                                                       //
              case "serverError":                                                                                      // 466
                var serverError = {};                                                                                  // 467
                rpcResponse.serverError = serverError; // TODO(soon): Pass along the body from upstream.               // 468
                                                                                                                       //
                clientError.descriptionHtml = statusInfo.descriptionHtml;                                              // 470
                resolve(rpcResponse);                                                                                  // 471
                break;                                                                                                 // 472
              // TODO(soon): Handle token-expired errors by throwing DISCONNECTED -- this will force                   // 474
              //   the client to reload the capability which will refresh the token.                                   // 475
                                                                                                                       //
              default:                                                                                                 // 477
                // ???                                                                                                 // 477
                err = new Error("Invalid status code " + resp.statusCode + " received in response.");                  // 478
                reject(err);                                                                                           // 479
                break;                                                                                                 // 480
            }                                                                                                          // 411
          } catch (err) {                                                                                              // 482
            reject(err);                                                                                               // 483
          }                                                                                                            // 484
        });                                                                                                            // 485
        req.on("error", function (e) {                                                                                 // 487
          reject(e);                                                                                                   // 488
        });                                                                                                            // 489
        req.setTimeout(15000, function () {                                                                            // 491
          req.abort();                                                                                                 // 492
          err = new Error("Request timed out.");                                                                       // 493
          err.kjType = "overloaded";                                                                                   // 494
          reject(err);                                                                                                 // 495
        });                                                                                                            // 496
                                                                                                                       //
        if (content) {                                                                                                 // 498
          req.end(content);                                                                                            // 499
        } else {                                                                                                       // 500
          req.end();                                                                                                   // 501
        }                                                                                                              // 502
      });                                                                                                              // 503
    }                                                                                                                  // 504
                                                                                                                       //
    return _requestHelper;                                                                                             // 292
  }();                                                                                                                 // 292
                                                                                                                       //
  return ExternalWebSession;                                                                                           // 292
}(PersistentImpl);                                                                                                     // 292
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"ip.js":["babel-runtime/helpers/possibleConstructorReturn","babel-runtime/helpers/inherits","babel-runtime/helpers/classCallCheck","bignum","/imports/server/persistent.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/drivers/ip.js                                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var Bignum = void 0;                                                                                                   // 1
module.import("bignum", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    Bignum = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var PersistentImpl = void 0;                                                                                           // 1
module.import("/imports/server/persistent.js", {                                                                       // 1
  "PersistentImpl": function (v) {                                                                                     // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 19
                                                                                                                       //
var Net = Npm.require("net");                                                                                          // 20
                                                                                                                       //
var Tls = Npm.require("tls");                                                                                          // 21
                                                                                                                       //
var Dgram = Npm.require("dgram");                                                                                      // 22
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 23
                                                                                                                       //
var IpRpc = Capnp.importSystem("sandstorm/ip.capnp");                                                                  // 25
                                                                                                                       //
ByteStreamConnection = function () {                                                                                   // 27
  function ByteStreamConnection(connection) {                                                                          // 28
    (0, _classCallCheck3.default)(this, ByteStreamConnection);                                                         // 28
    this.connection = connection;                                                                                      // 29
  }                                                                                                                    // 30
                                                                                                                       //
  ByteStreamConnection.prototype.done = function () {                                                                  // 27
    function done() {                                                                                                  // 27
      this.connection.end();                                                                                           // 33
    }                                                                                                                  // 34
                                                                                                                       //
    return done;                                                                                                       // 27
  }();                                                                                                                 // 27
                                                                                                                       //
  ByteStreamConnection.prototype.write = function () {                                                                 // 27
    function write(data) {                                                                                             // 27
      this.connection.write(data);                                                                                     // 37
    }                                                                                                                  // 38
                                                                                                                       //
    return write;                                                                                                      // 27
  }(); // expectSize not implemented                                                                                   // 27
  // expectSize(size) { }                                                                                              // 41
                                                                                                                       //
                                                                                                                       //
  return ByteStreamConnection;                                                                                         // 27
}();                                                                                                                   // 27
                                                                                                                       //
var IpInterfaceImpl = function (_PersistentImpl) {                                                                     //
  (0, _inherits3.default)(IpInterfaceImpl, _PersistentImpl);                                                           //
                                                                                                                       //
  function IpInterfaceImpl(db, saveTemplate) {                                                                         // 45
    (0, _classCallCheck3.default)(this, IpInterfaceImpl);                                                              // 45
    return (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));               // 45
  }                                                                                                                    // 47
                                                                                                                       //
  IpInterfaceImpl.prototype.listenTcp = function () {                                                                  //
    function listenTcp(portNum, port) {                                                                                //
      return new Promise(function (resolve, reject) {                                                                  // 50
        var resolved = false;                                                                                          // 51
        var server = Net.createServer(function (connection) {                                                          // 52
          var wrappedConnection = new ByteStreamConnection(connection);                                                // 53
          var upstream = port.connect(wrappedConnection).upstream;                                                     // 54
          connection.on("data", function (data) {                                                                      // 56
            upstream.write(data);                                                                                      // 57
          });                                                                                                          // 58
          connection.on("close", function (hadError) {                                                                 // 60
            upstream.done();                                                                                           // 61
          });                                                                                                          // 62
          connection.on("error", function (err) {                                                                      // 64
            if (resolved) {                                                                                            // 65
              connection.write = errorWrite;                                                                           // 66
            } else {                                                                                                   // 67
              // upstream hasn't been resolved yet, so it's safe to reject                                             // 68
              reject(err);                                                                                             // 69
            }                                                                                                          // 70
          });                                                                                                          // 71
        });                                                                                                            // 72
        server.listen(portNum, function () {                                                                           // 74
          resolved = true;                                                                                             // 75
          resolve({                                                                                                    // 76
            handle: server                                                                                             // 76
          }); // server has a close method which is all we want from a handle                                          // 76
        });                                                                                                            // 77
        server.on("error", function (err) {                                                                            // 79
          if (!resolved) {                                                                                             // 80
            reject(err);                                                                                               // 81
          }                                                                                                            // 82
        });                                                                                                            // 83
      });                                                                                                              // 84
    }                                                                                                                  // 85
                                                                                                                       //
    return listenTcp;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  IpInterfaceImpl.prototype.listenUdp = function () {                                                                  //
    function listenUdp(portNum, port) {                                                                                //
      return new Promise(function (resolve, reject) {                                                                  // 88
        var portMap = {};                                                                                              // 89
        var resolved = false;                                                                                          // 90
        var server = Dgram.createSocket("udp4"); // TODO(someday): handle ipv6 sockets too                             // 91
                                                                                                                       //
        server.bind(portNum);                                                                                          // 92
        server.on("listening", function () {                                                                           // 94
          // Although UDP is connectionless, we don't resolve until here so that we can handle bind                    // 95
          // errors such as invalid host                                                                               // 96
          resolved = true;                                                                                             // 97
          resolve({                                                                                                    // 98
            handle: server                                                                                             // 98
          }); // server has a close method which is all we want from a handle                                          // 98
        });                                                                                                            // 99
        server.on("error", function (err) {                                                                            // 101
          // TODO(someday): do something about errors after the promise is resolved                                    // 102
          if (!resolved) {                                                                                             // 103
            reject(err);                                                                                               // 104
          } else {                                                                                                     // 105
            console.error("error in listenUdp: " + err);                                                               // 106
          }                                                                                                            // 107
        });                                                                                                            // 108
        var returnMap = {};                                                                                            // 110
        server.on("message", function (msg, rinfo) {                                                                   // 111
          var address = rinfo.address + "]:" + rinfo.port;                                                             // 112
          var returnPort = returnMap[address];                                                                         // 113
                                                                                                                       //
          if (!returnPort) {                                                                                           // 115
            returnMap[address] = returnPort = new BoundUdpPortImpl(server, rinfo.address, rinfo.port);                 // 116
          }                                                                                                            // 117
                                                                                                                       //
          port.send(msg, returnPort);                                                                                  // 119
        });                                                                                                            // 120
      });                                                                                                              // 121
    }                                                                                                                  // 122
                                                                                                                       //
    return listenUdp;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  return IpInterfaceImpl;                                                                                              //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
; // TODO(cleanup): Meteor.startup() needed because 00-startup.js runs *after* code in subdirectories                  // 123
//   (ugh).                                                                                                            // 126
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 127
  globalFrontendRefRegistry.register({                                                                                 // 128
    frontendRefField: "ipInterface",                                                                                   // 129
    typeId: IpRpc.IpInterface.typeId,                                                                                  // 130
    restore: function (db, saveTemplate) {                                                                             // 132
      return new Capnp.Capability(new IpInterfaceImpl(db, saveTemplate), IpRpc.PersistentIpInterface);                 // 133
    },                                                                                                                 // 135
    validate: function (db, session, value) {                                                                          // 137
      check(value, true);                                                                                              // 138
                                                                                                                       //
      if (!session.userId) {                                                                                           // 140
        throw new Meteor.Error(403, "Not logged in.");                                                                 // 141
      }                                                                                                                // 142
                                                                                                                       //
      return {                                                                                                         // 144
        descriptor: {                                                                                                  // 145
          tags: [{                                                                                                     // 145
            id: IpRpc.IpInterface.typeId                                                                               // 145
          }]                                                                                                           // 145
        },                                                                                                             // 145
        requirements: [{                                                                                               // 146
          userIsAdmin: session.userId                                                                                  // 146
        }],                                                                                                            // 146
        frontendRef: value                                                                                             // 147
      };                                                                                                               // 144
    },                                                                                                                 // 149
    query: function (db, userId, value) {                                                                              // 151
      if (userId && Meteor.users.findOne(userId).isAdmin) {                                                            // 152
        return [{                                                                                                      // 153
          _id: "frontendref-ipinterface",                                                                              // 155
          frontendRef: {                                                                                               // 156
            ipInterface: true                                                                                          // 156
          },                                                                                                           // 156
          cardTemplate: "ipInterfacePowerboxCard"                                                                      // 157
        }];                                                                                                            // 154
      } else {                                                                                                         // 160
        return [];                                                                                                     // 161
      }                                                                                                                // 162
    }                                                                                                                  // 163
  });                                                                                                                  // 128
});                                                                                                                    // 165
                                                                                                                       //
BoundUdpPortImpl = function () {                                                                                       // 167
  function BoundUdpPortImpl(server, address, port) {                                                                   // 168
    (0, _classCallCheck3.default)(this, BoundUdpPortImpl);                                                             // 168
    this.server = server;                                                                                              // 169
    this.address = address;                                                                                            // 170
    this.port = port;                                                                                                  // 171
  }                                                                                                                    // 172
                                                                                                                       //
  BoundUdpPortImpl.prototype.send = function () {                                                                      // 167
    function send(message, returnPort) {                                                                               // 167
      // TODO(someday): this whole class is a hack to deal with the fact that we can't compare                         // 175
      // capabilities or build a map with them. What we should be doing is mapping all ports to                        // 176
      // their raw physical address/port, and using that here                                                          // 177
      this.server.send(message, 0, message.length, this.port, this.address);                                           // 178
    }                                                                                                                  // 179
                                                                                                                       //
    return send;                                                                                                       // 167
  }();                                                                                                                 // 167
                                                                                                                       //
  return BoundUdpPortImpl;                                                                                             // 167
}();                                                                                                                   // 167
                                                                                                                       //
var bits16 = Bignum(1).shiftLeft(16).sub(1);                                                                           // 182
var bits32 = Bignum(1).shiftLeft(32).sub(1);                                                                           // 183
                                                                                                                       //
var intToIpv4 = function (num) {                                                                                       // 185
  var part1 = num & 255;                                                                                               // 186
  var part2 = num >> 8 & 255;                                                                                          // 187
  var part3 = num >> 16 & 255;                                                                                         // 188
  var part4 = num >> 24 & 255;                                                                                         // 189
  return part4 + "." + part3 + "." + part2 + "." + part1;                                                              // 191
};                                                                                                                     // 192
                                                                                                                       //
var addressToString = function (address) {                                                                             // 194
  var ipv6num = Bignum(address.upper64).shiftLeft(64).add(Bignum(address.lower64));                                    // 195
                                                                                                                       //
  if (ipv6num.shiftRight(32).eq(bits16)) {                                                                             // 197
    // this is an ipv4 address, we should return it as such                                                            // 198
    var ipv4num = ipv6num.and(bits32).toNumber();                                                                      // 199
    return intToIpv4(ipv4num);                                                                                         // 200
  }                                                                                                                    // 201
                                                                                                                       //
  var hex = ipv6num.toString(16);                                                                                      // 203
  var numColons = 0;                                                                                                   // 204
  var out = "";                                                                                                        // 205
                                                                                                                       //
  for (var i = 0; i < hex.length; ++i) {                                                                               // 207
    // start with lower bits of address and build the output in reverse                                                // 208
    // this ensures that we can place a colon every 4 characters                                                       // 209
    out += hex[hex.length - 1 - i];                                                                                    // 210
                                                                                                                       //
    if ((i + 1) % 4 === 0) {                                                                                           // 211
      out += ":";                                                                                                      // 212
      ++numColons;                                                                                                     // 213
    }                                                                                                                  // 214
  } // Double colon represents all bits being 0                                                                        // 215
                                                                                                                       //
                                                                                                                       //
  if (numColons < 7) {                                                                                                 // 218
    out += "::";                                                                                                       // 219
  }                                                                                                                    // 220
                                                                                                                       //
  return out.split("").reverse().join("");                                                                             // 222
};                                                                                                                     // 223
                                                                                                                       //
var addressType = function (address) {                                                                                 // 225
  var type = "udp4"; // Check if it's an ipv6 address                                                                  // 226
  // TODO(someday): make this less hacky and change address to explicitly pass this information                        // 229
                                                                                                                       //
  if (address.indexOf(":") != -1) {                                                                                    // 230
    type = "udp6";                                                                                                     // 231
  }                                                                                                                    // 232
                                                                                                                       //
  return type;                                                                                                         // 234
};                                                                                                                     // 235
                                                                                                                       //
var IpNetworkImpl = function (_PersistentImpl2) {                                                                      //
  (0, _inherits3.default)(IpNetworkImpl, _PersistentImpl2);                                                            //
                                                                                                                       //
  function IpNetworkImpl(db, saveTemplate, tls) {                                                                      // 238
    (0, _classCallCheck3.default)(this, IpNetworkImpl);                                                                // 238
                                                                                                                       //
    var _this3 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl2.call(this, db, saveTemplate));        // 238
                                                                                                                       //
    _this3.tls = tls;                                                                                                  // 240
    return _this3;                                                                                                     // 238
  }                                                                                                                    // 241
                                                                                                                       //
  IpNetworkImpl.prototype.getRemoteHost = function () {                                                                //
    function getRemoteHost(address) {                                                                                  //
      // Note that TLS typically authenticates hostnames, not raw IP addresses. So if `this.tls`                       // 244
      // is true, then the returned `IpRemoteHost` will likely refuse to make any connections.                         // 245
      // However, we keep that code path active because there is in principle no reason why it                         // 246
      // should always fail, and we wish to allow for a possible future in which we let users                          // 247
      // specify custom certificate authorities, in which case it might be more likely for TLS                         // 248
      // to be expected to authenticate raw IP addresses.                                                              // 249
      return {                                                                                                         // 251
        host: new IpRemoteHostImpl(addressToString(address), this.tls)                                                 // 251
      };                                                                                                               // 251
    }                                                                                                                  // 252
                                                                                                                       //
    return getRemoteHost;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  IpNetworkImpl.prototype.getRemoteHostByName = function () {                                                          //
    function getRemoteHostByName(address) {                                                                            //
      return {                                                                                                         // 255
        host: new IpRemoteHostImpl(address, this.tls)                                                                  // 255
      };                                                                                                               // 255
    }                                                                                                                  // 256
                                                                                                                       //
    return getRemoteHostByName;                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  return IpNetworkImpl;                                                                                                //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
; // TODO(cleanup): Meteor.startup() needed because 00-startup.js runs *after* code in subdirectories                  // 257
//   (ugh).                                                                                                            // 260
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 261
  globalFrontendRefRegistry.register({                                                                                 // 262
    frontendRefField: "ipNetwork",                                                                                     // 263
    typeId: IpRpc.IpNetwork.typeId,                                                                                    // 264
    restore: function (db, saveTemplate, value) {                                                                      // 266
      return new Capnp.Capability(new IpNetworkImpl(db, saveTemplate, "tls" in value.encryption), IpRpc.PersistentIpNetwork);
    },                                                                                                                 // 270
    validate: function (db, session, value) {                                                                          // 272
      check(value, {                                                                                                   // 273
        encryption: Match.OneOf({                                                                                      // 273
          none: null                                                                                                   // 273
        }, {                                                                                                           // 273
          tls: null                                                                                                    // 273
        })                                                                                                             // 273
      });                                                                                                              // 273
                                                                                                                       //
      if (!session.userId) {                                                                                           // 275
        throw new Meteor.Error(403, "Not logged in.");                                                                 // 276
      }                                                                                                                // 277
                                                                                                                       //
      return {                                                                                                         // 279
        descriptor: {                                                                                                  // 280
          tags: [{                                                                                                     // 281
            id: IpRpc.IpNetwork.typeId,                                                                                // 283
            value: Capnp.serialize(IpRpc.IpNetwork.PowerboxTag, value)                                                 // 284
          }]                                                                                                           // 282
        },                                                                                                             // 280
        requirements: [{                                                                                               // 290
          userIsAdmin: session.userId                                                                                  // 290
        }],                                                                                                            // 290
        frontendRef: value                                                                                             // 291
      };                                                                                                               // 279
    },                                                                                                                 // 293
    query: function (db, userId, value) {                                                                              // 295
      var encryption = {                                                                                               // 296
        none: null                                                                                                     // 296
      };                                                                                                               // 296
                                                                                                                       //
      if (value) {                                                                                                     // 297
        encryption = Capnp.parse(IpRpc.IpNetwork.PowerboxTag, value).encryption || encryption;                         // 298
      }                                                                                                                // 299
                                                                                                                       //
      if (userId && Meteor.users.findOne(userId).isAdmin) {                                                            // 301
        return [{                                                                                                      // 302
          _id: "frontendref-ipnetwork",                                                                                // 304
          frontendRef: {                                                                                               // 305
            ipNetwork: {                                                                                               // 305
              encryption: encryption                                                                                   // 305
            }                                                                                                          // 305
          },                                                                                                           // 305
          cardTemplate: "ipNetworkPowerboxCard"                                                                        // 306
        }];                                                                                                            // 303
      } else {                                                                                                         // 309
        return [];                                                                                                     // 310
      }                                                                                                                // 311
    }                                                                                                                  // 312
  });                                                                                                                  // 262
});                                                                                                                    // 314
                                                                                                                       //
var IpRemoteHostImpl = function () {                                                                                   //
  function IpRemoteHostImpl(address, tls) {                                                                            // 317
    (0, _classCallCheck3.default)(this, IpRemoteHostImpl);                                                             // 317
    this.address = address;                                                                                            // 318
    this.tls = tls;                                                                                                    // 319
  }                                                                                                                    // 320
                                                                                                                       //
  IpRemoteHostImpl.prototype.getTcpPort = function () {                                                                //
    function getTcpPort(portNum) {                                                                                     //
      return {                                                                                                         // 323
        port: new TcpPortImpl(this.address, portNum, this.tls)                                                         // 323
      };                                                                                                               // 323
    }                                                                                                                  // 324
                                                                                                                       //
    return getTcpPort;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  IpRemoteHostImpl.prototype.getUdpPort = function () {                                                                //
    function getUdpPort(portNum) {                                                                                     //
      if (this.tls) {                                                                                                  // 327
        var error = new Error("Datagram Transport Layer Security is not yet supported");                               // 328
        error.kjType = "unimplemented";                                                                                // 329
        throw error;                                                                                                   // 330
      }                                                                                                                // 331
                                                                                                                       //
      return {                                                                                                         // 333
        port: new UdpPortImpl(this.address, portNum)                                                                   // 333
      };                                                                                                               // 333
    }                                                                                                                  // 334
                                                                                                                       //
    return getUdpPort;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  return IpRemoteHostImpl;                                                                                             //
}();                                                                                                                   //
                                                                                                                       //
;                                                                                                                      // 335
                                                                                                                       //
TcpPortImpl = function () {                                                                                            // 337
  function TcpPortImpl(address, portNum, tls) {                                                                        // 338
    (0, _classCallCheck3.default)(this, TcpPortImpl);                                                                  // 338
    this.address = address;                                                                                            // 339
    this.port = portNum;                                                                                               // 340
    this.tls = tls;                                                                                                    // 341
  }                                                                                                                    // 342
                                                                                                                       //
  TcpPortImpl.prototype.connect = function () {                                                                        // 337
    function connect(downstream) {                                                                                     // 337
      var _this = this;                                                                                                // 345
                                                                                                                       //
      var resolved = false;                                                                                            // 346
      var connectMethod = Net.connect;                                                                                 // 347
                                                                                                                       //
      if (this.tls) {                                                                                                  // 348
        connectMethod = Tls.connect;                                                                                   // 349
      }                                                                                                                // 350
                                                                                                                       //
      return new Promise(function (resolve, reject) {                                                                  // 352
        var client = connectMethod({                                                                                   // 353
          host: _this.address,                                                                                         // 353
          port: _this.port                                                                                             // 353
        }, function () {                                                                                               // 353
          resolved = true;                                                                                             // 354
          resolve({                                                                                                    // 355
            upstream: new ByteStreamConnection(client)                                                                 // 355
          });                                                                                                          // 355
        });                                                                                                            // 356
        client.on("data", function (data) {                                                                            // 358
          downstream.write(data);                                                                                      // 359
        });                                                                                                            // 360
        client.on("close", function (hadError) {                                                                       // 362
          downstream.done();                                                                                           // 363
        });                                                                                                            // 364
        client.on("error", function (err) {                                                                            // 366
          if (resolved) {                                                                                              // 367
            client.write = errorWrite;                                                                                 // 368
          } else {                                                                                                     // 369
            // upstream hasn't been resolved yet, so it's safe to reject                                               // 370
            reject(err);                                                                                               // 371
          }                                                                                                            // 372
        });                                                                                                            // 373
      });                                                                                                              // 374
    }                                                                                                                  // 375
                                                                                                                       //
    return connect;                                                                                                    // 337
  }();                                                                                                                 // 337
                                                                                                                       //
  return TcpPortImpl;                                                                                                  // 337
}();                                                                                                                   // 337
                                                                                                                       //
var errorWrite = function (data) {                                                                                     // 378
  throw new Error("error occurred in connection");                                                                     // 379
};                                                                                                                     // 380
                                                                                                                       //
UdpPortImpl = function () {                                                                                            // 382
  function UdpPortImpl(address, portNum) {                                                                             // 383
    (0, _classCallCheck3.default)(this, UdpPortImpl);                                                                  // 383
    this.address = address;                                                                                            // 384
    this.port = portNum;                                                                                               // 385
    var type = addressType(address);                                                                                   // 387
    this.socket = Dgram.createSocket(type); // TODO(someday): close socket after a certain time of inactivity?         // 388
    // This may be pointless since grains are killed frequently when not in use anyways                                // 391
    // Temporary hack. We only expect clients to pass in a single return port, so we'll store it                       // 393
    // and only send replies here.                                                                                     // 394
    // This will be changed to be correct when equality comparisons are added to capabilities.                         // 395
                                                                                                                       //
    this.returnPort = null;                                                                                            // 396
                                                                                                                       //
    var _this = this;                                                                                                  // 398
                                                                                                                       //
    this.socket.on("message", function (msg, rinfo) {                                                                  // 399
      if (_this.returnPort) {                                                                                          // 400
        _this.returnPort.send(msg, _this);                                                                             // 401
      }                                                                                                                // 402
    });                                                                                                                // 403
  }                                                                                                                    // 404
                                                                                                                       //
  UdpPortImpl.prototype.send = function () {                                                                           // 382
    function send(message, returnPort) {                                                                               // 382
      this.returnPort = returnPort;                                                                                    // 407
      this.socket.send(message, 0, message.length, this.port, this.address); // TODO(someday): use callback to catch errors and do something with them
    }                                                                                                                  // 411
                                                                                                                       //
    return send;                                                                                                       // 382
  }();                                                                                                                 // 382
                                                                                                                       //
  return UdpPortImpl;                                                                                                  // 382
}();                                                                                                                   // 382
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"mail.js":["babel-runtime/helpers/classCallCheck","babel-runtime/helpers/possibleConstructorReturn","babel-runtime/helpers/inherits","smtp-server","mailparser","/imports/server/persistent.js","/imports/server/email.js","/imports/server/backend.js","/imports/server/async-helpers.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/drivers/mail.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _this = this;                                                                                                      //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var SMTPServer = void 0;                                                                                               // 1
module.import("smtp-server", {                                                                                         // 1
  "SMTPServer": function (v) {                                                                                         // 1
    SMTPServer = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var MailParser = void 0;                                                                                               // 1
module.import("mailparser", {                                                                                          // 1
  "MailParser": function (v) {                                                                                         // 1
    MailParser = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var PersistentImpl = void 0;                                                                                           // 1
module.import("/imports/server/persistent.js", {                                                                       // 1
  "PersistentImpl": function (v) {                                                                                     // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var rawSend = void 0;                                                                                                  // 1
module.import("/imports/server/email.js", {                                                                            // 1
  "rawSend": function (v) {                                                                                            // 1
    rawSend = v;                                                                                                       // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var shouldRestartGrain = void 0;                                                                                       // 1
module.import("/imports/server/backend.js", {                                                                          // 1
  "shouldRestartGrain": function (v) {                                                                                 // 1
    shouldRestartGrain = v;                                                                                            // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var inMeteor = void 0;                                                                                                 // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 25
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 26
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 27
                                                                                                                       //
var EmailRpc = Capnp.importSystem("sandstorm/email.capnp");                                                            // 29
var EmailImpl = Capnp.importSystem("sandstorm/email-impl.capnp");                                                      // 30
var HackSessionContext = Capnp.importSystem("sandstorm/hack-session.capnp").HackSessionContext;                        // 31
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;                                          // 32
var EmailSendPort = EmailRpc.EmailSendPort;                                                                            // 33
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 35
                                                                                                                       //
var ROOT_URL = Url.parse(process.env.ROOT_URL);                                                                        // 37
var HOSTNAME = ROOT_URL.hostname;                                                                                      // 38
var RECIPIENT_LIMIT = 20;                                                                                              // 40
var CLIENT_TIMEOUT = 15000; // 15s                                                                                     // 42
// Every day, reset all per-user sent counts to zero.                                                                  // 44
// TODO(cleanup): Consider a more granular approach. For example, each user could have a timer                         // 45
//   after which their count will reset. We'd only check the timer when that user is trying to                         // 46
//   send a new message. This avoids a global query.                                                                   // 47
                                                                                                                       //
if (!Meteor.settings.replicaNumber) {                                                                                  // 48
  // only first replica                                                                                                // 48
  SandstormDb.periodicCleanup(86400000, function () {                                                                  // 49
    Meteor.users.update({                                                                                              // 50
      dailySentMailCount: {                                                                                            // 50
        $exists: true                                                                                                  // 50
      }                                                                                                                // 50
    }, {                                                                                                               // 50
      $unset: {                                                                                                        // 51
        dailySentMailCount: ""                                                                                         // 51
      }                                                                                                                // 51
    }, {                                                                                                               // 51
      multi: true                                                                                                      // 52
    });                                                                                                                // 52
  });                                                                                                                  // 53
}                                                                                                                      // 54
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 56
  var server = new SMTPServer({                                                                                        // 57
    banner: "Sandstorm Mail Server",                                                                                   // 58
    authOptional: true,                                                                                                // 59
    disabledCommands: ["STARTTLS", "AUTH"],                                                                            // 60
    onData: function (req, session, callback) {                                                                        // 61
      var mailparser = new MailParser();                                                                               // 62
      req.pipe(mailparser);                                                                                            // 63
      mailparser.on("end", function (mail) {                                                                           // 65
        // Wrap in outer promise for easier error handling.                                                            // 66
        Promise.resolve().then(function () {                                                                           // 67
          // Extract the 'from' address.                                                                               // 68
          var from = void 0;                                                                                           // 69
                                                                                                                       //
          if (mail.from && mail.from.length > 0) {                                                                     // 70
            // It's theoretically possible for the message to have multiple 'from' headers, but this                   // 71
            // never really happens in legitimate practice so we'll just take the first one.                           // 72
            from = mail.from[0];                                                                                       // 73
          } else {                                                                                                     // 74
            // The mail body is missing a 'From:' header. We'll use the bounce address instead,                        // 75
            // with the caveat that this is sometimes *not* the original sender                                        // 76
            // but rather some intermediate agent (e.g. a mailing list daemon). See:                                   // 77
            //   http://en.wikipedia.org/wiki/Bounce_address                                                           // 78
            // TODO(someday): Is this really right, or should we report a blank address instead?                       // 79
            from = {                                                                                                   // 80
              address: session.envelope.mailFrom,                                                                      // 80
              name: false                                                                                              // 80
            };                                                                                                         // 80
          }                                                                                                            // 81
                                                                                                                       //
          var attachments = [];                                                                                        // 83
                                                                                                                       //
          if (mail.attachments) {                                                                                      // 84
            attachments = mail.attachments.map(function (attachment) {                                                 // 85
              var disposition = attachment.contentDisposition || "attachment";                                         // 86
              disposition += ';\n\tfilename="' + (attachment.fileName || attachment.generatedFileName) + '"';          // 87
              return {                                                                                                 // 88
                contentType: attachment.contentType,                                                                   // 89
                contentDisposition: disposition,                                                                       // 90
                contentId: attachment.contentId,                                                                       // 91
                content: attachment.content                                                                            // 92
              };                                                                                                       // 88
            });                                                                                                        // 94
          }                                                                                                            // 95
                                                                                                                       //
          if (mail.replyTo && mail.replyTo.length > 1) {                                                               // 97
            console.error("More than one reply-to address address was received in an email.");                         // 98
          }                                                                                                            // 99
                                                                                                                       //
          var mailMessage = {                                                                                          // 101
            // Note that converting the date to nanoseconds actually goes outside the range of                         // 102
            // integers that Javascript can represent precisely. But this is OK because dates aren't                   // 103
            // precise anyway.                                                                                         // 104
            date: (mail.date || new Date()).getTime() * 1000000,                                                       // 105
            from: from,                                                                                                // 106
            to: mail.to,                                                                                               // 107
            cc: mail.cc || [],                                                                                         // 108
            bcc: mail.bcc || [],                                                                                       // 109
            replyTo: mail.replyTo && mail.replyTo[0] || {},                                                            // 110
            messageId: mail.headers["message-id"] || Meteor.uuid() + "@" + HOSTNAME,                                   // 111
            references: mail.references || [],                                                                         // 112
            inReplyTo: mail.inReplyTo || [],                                                                           // 113
            subject: mail.subject || "",                                                                               // 114
            text: mail.text || null,                                                                                   // 115
            html: mail.html || null,                                                                                   // 116
            attachments: attachments                                                                                   // 117
          }; // Get list of grain IDs.                                                                                 // 101
                                                                                                                       //
          var grainPublicIds = _.uniq(session.envelope.rcptTo.map(function (deliverTo) {                               // 121
            // smtp-server already validates that the address contains an @.                                           // 122
            // To simplify things, we ignore the hostname part of the address and assume that the                      // 123
            // message would not have been sent here if it weren't intended for our host. Usually                      // 124
            // there will be an nginx frontend verifying hostnames anyway. Grain public IDs are                        // 125
            // globally unique anyway, so an e-mail meant for another server presumably won't match                    // 126
            // any ID at this one anyway.                                                                              // 127
            var address = deliverTo.address;                                                                           // 121
            return address.slice(0, address.indexOf("@"));                                                             // 129
          })); // Deliver to each grain in parallel.                                                                   // 130
                                                                                                                       //
                                                                                                                       //
          return Promise.all(grainPublicIds.map(function (publicId) {                                                  // 133
            // Wrap in a function so that we can call it recursively to retry.                                         // 134
            var tryDeliver = function (retryCount) {                                                                   // 135
              var grainId = void 0;                                                                                    // 136
              return inMeteor(function () {                                                                            // 137
                var grain = Grains.findOne({                                                                           // 138
                  publicId: publicId                                                                                   // 138
                }, {                                                                                                   // 138
                  fields: {}                                                                                           // 138
                });                                                                                                    // 138
                                                                                                                       //
                if (grain) {                                                                                           // 139
                  grainId = grain._id;                                                                                 // 140
                  return globalBackend.continueGrain(grainId, retryCount > 0);                                         // 141
                } else {                                                                                               // 142
                  // TODO(someday): We really ought to rig things up so that the 'RCPT TO' SMTP command                // 143
                  // fails in this case, by adding an onRcptTo() callback.                                             // 144
                  throw new Error("No such grain: " + publicId);                                                       // 145
                }                                                                                                      // 146
              }).then(function (grainInfo) {                                                                           // 147
                var supervisor = grainInfo.supervisor;                                                                 // 148
                var uiView = supervisor.getMainView().view; // Create an arbitrary struct to use as the session params. E-mail sessions actually
                // require no params, but node-capnp won't let us pass null and we don't have an                       // 152
                // EmptyStruct type available, so we just use EmailAddress, but any struct type would                  // 153
                // work.                                                                                               // 154
                // TODO(cleanup): Fix node-capnp to accept null.                                                       // 155
                                                                                                                       //
                var emptyParams = Capnp.serialize(EmailRpc.EmailAddress, {}); // Create a new session of type HackEmailSession. This is a short-term hack until
                // persistent capabilities and the Powerbox are implemented. We need to pass along a                   // 159
                // HackSessionContext because old versions of sandstorm-http-bridge always use                         // 160
                // the context of the most recent session; if an old RoundCube grain has a                             // 161
                // WebSession open, receives an email, and then tries to send an email, the request                    // 162
                // will go out on the SessionContext associated with the HackEmailSession that                         // 163
                // delivered the email.                                                                                // 164
                                                                                                                       //
                var session = uiView.newSession({}, makeHackSessionContext(grainId), "0xc3b5ced7344b04a6", emptyParams).session.castAs(EmailSendPort);
                return session.send(mailMessage);                                                                      // 169
              }).catch(function (err) {                                                                                // 170
                if (shouldRestartGrain(err, retryCount)) {                                                             // 171
                  return tryDeliver(retryCount + 1);                                                                   // 172
                } else {                                                                                               // 173
                  throw err;                                                                                           // 174
                }                                                                                                      // 175
              });                                                                                                      // 176
            };                                                                                                         // 177
                                                                                                                       //
            return tryDeliver(0);                                                                                      // 179
          }));                                                                                                         // 180
        }).then(function () {                                                                                          // 181
          callback(null, "Message delivered");                                                                         // 182
        }, function (err) {                                                                                            // 183
          console.error("E-mail delivery failure:", err.stack);                                                        // 184
          var mailErr = new Error(err.message);                                                                        // 185
          err.responseCode = 451;                                                                                      // 186
          callback(err.message);                                                                                       // 187
        });                                                                                                            // 188
      });                                                                                                              // 189
    }                                                                                                                  // 190
  });                                                                                                                  // 57
                                                                                                                       //
  if (global.SANDSTORM_SMTP_LISTEN_HANDLE) {                                                                           // 193
    server.listen(global.SANDSTORM_SMTP_LISTEN_HANDLE);                                                                // 194
  } else {                                                                                                             // 195
    // We must be running `run-dev.sh`.                                                                                // 196
    var BIND_IP = process.env.BIND_IP || "127.0.0.1";                                                                  // 197
    var SMTP_LISTEN_PORT = Meteor.settings.public.smtpListenPort || 30025;                                             // 198
    server.listen(SMTP_LISTEN_PORT, BIND_IP);                                                                          // 199
  }                                                                                                                    // 200
});                                                                                                                    // 201
                                                                                                                       //
hackSendEmail = function (session, email) {                                                                            // 203
  return inMeteor(function () {                                                                                        // 204
    var recipientCount = 0;                                                                                            // 205
    recipientCount += email.to ? email.to.length : 0;                                                                  // 206
    recipientCount += email.cc ? email.cc.length : 0;                                                                  // 207
    recipientCount += email.bcc ? email.bcc.length : 0;                                                                // 208
                                                                                                                       //
    if (recipientCount > RECIPIENT_LIMIT) {                                                                            // 209
      throw new Error("Sorry, Sandstorm currently only allows you to send an e-mail to " + RECIPIENT_LIMIT + " recipients at a time, for spam control. Consider setting up a mailing list. " + "Please feel free to contact us if this is a problem for you.");
    }                                                                                                                  // 214
                                                                                                                       //
    var grainAddress = session._getAddress();                                                                          // 216
                                                                                                                       //
    var userAddress = session._getUserAddress(); // Overwrite the 'from' address with the grain's address.             // 217
                                                                                                                       //
                                                                                                                       //
    if (!email.from) {                                                                                                 // 220
      email.from = {                                                                                                   // 221
        address: grainAddress                                                                                          // 222
      };                                                                                                               // 221
    }                                                                                                                  // 224
                                                                                                                       //
    if (email.from.address !== grainAddress && email.from.address !== userAddress.address) {                           // 226
      throw new Error("FROM header in outgoing emails need to equal either " + grainAddress + " or " + userAddress.address + ". Yours was: " + email.from.address);
    } // Unpack fields                                                                                                 // 230
                                                                                                                       //
                                                                                                                       //
    var from = email.from,                                                                                             // 204
        to = email.to,                                                                                                 // 204
        cc = email.cc,                                                                                                 // 204
        bcc = email.bcc,                                                                                               // 204
        replyTo = email.replyTo,                                                                                       // 204
        subject = email.subject,                                                                                       // 204
        text = email.text,                                                                                             // 204
        html = email.html;                                                                                             // 204
    var options = {                                                                                                    // 235
      from: from,                                                                                                      // 236
      to: to,                                                                                                          // 237
      cc: cc,                                                                                                          // 238
      bcc: bcc,                                                                                                        // 239
      replyTo: replyTo,                                                                                                // 240
      subject: subject,                                                                                                // 241
      text: text,                                                                                                      // 242
      html: html,                                                                                                      // 243
      envelope: {                                                                                                      // 244
        from: grainAddress,                                                                                            // 245
        to: to,                                                                                                        // 246
        cc: cc,                                                                                                        // 247
        bcc: bcc                                                                                                       // 248
      }                                                                                                                // 244
    };                                                                                                                 // 235
    var headers = {};                                                                                                  // 252
                                                                                                                       //
    if (email.messageId) {                                                                                             // 253
      headers["message-id"] = email.messageId;                                                                         // 254
    }                                                                                                                  // 255
                                                                                                                       //
    if (email.references) {                                                                                            // 257
      headers["references"] = email.references; // jscs:ignore requireDotNotation                                      // 258
    }                                                                                                                  // 259
                                                                                                                       //
    if (email.inReplyTo) {                                                                                             // 261
      headers["in-reply-to"] = email.inReplyTo;                                                                        // 262
    }                                                                                                                  // 263
                                                                                                                       //
    if (email.date) {                                                                                                  // 265
      var date = new Date(email.date / 1000000);                                                                       // 266
                                                                                                                       //
      if (!isNaN(date.getTime())) {                                                                                    // 267
        // Check to make sure date is valid                                                                            // 267
        headers["date"] = date.toUTCString(); // jscs:ignore requireDotNotation                                        // 268
      }                                                                                                                // 269
    }                                                                                                                  // 270
                                                                                                                       //
    options.headers = headers;                                                                                         // 272
                                                                                                                       //
    if (email.attachments) {                                                                                           // 274
      var attachments = [];                                                                                            // 275
      email.attachments.forEach(function (attachment) {                                                                // 276
        attachments.push({                                                                                             // 277
          filename: false,                                                                                             // 278
          // Tell nodemailer not to generate a new filename.                                                           // 278
          cid: attachment.contentId,                                                                                   // 279
          contentType: attachment.contentType,                                                                         // 280
          contentDisposition: attachment.contentDisposition,                                                           // 281
          content: attachment.content                                                                                  // 282
        });                                                                                                            // 277
      });                                                                                                              // 284
      options.attachments = attachments;                                                                               // 285
    }                                                                                                                  // 286
                                                                                                                       //
    var grain = Grains.findOne(session.grainId);                                                                       // 288
    if (!grain) throw new Error("Grain does not exist.");                                                              // 289
    globalDb.incrementDailySentMailCount(grain.userId);                                                                // 290
    rawSend(options);                                                                                                  // 292
  }.bind(_this)).catch(function (err) {                                                                                // 293
    console.error("Error sending e-mail:", err.stack);                                                                 // 294
    throw err;                                                                                                         // 295
  });                                                                                                                  // 296
};                                                                                                                     // 297
                                                                                                                       //
var EmailVerifierImpl = function (_PersistentImpl) {                                                                   //
  (0, _inherits3.default)(EmailVerifierImpl, _PersistentImpl);                                                         //
                                                                                                                       //
  function EmailVerifierImpl(db, saveTemplate, params) {                                                               // 300
    (0, _classCallCheck3.default)(this, EmailVerifierImpl);                                                            // 300
                                                                                                                       //
    var _this2 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));         // 300
                                                                                                                       //
    _this2._id = params.id;                                                                                            // 302
    _this2._services = params.services;                                                                                // 303
    return _this2;                                                                                                     // 300
  }                                                                                                                    // 304
                                                                                                                       //
  EmailVerifierImpl.prototype.getId = function () {                                                                    //
    function getId() {                                                                                                 //
      return {                                                                                                         // 307
        id: new Buffer(this._id, "base64")                                                                             // 307
      };                                                                                                               // 307
    }                                                                                                                  // 308
                                                                                                                       //
    return getId;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  EmailVerifierImpl.prototype.verifyEmail = function () {                                                              //
    function verifyEmail(tabId, verification) {                                                                        //
      var _this3 = this;                                                                                               // 310
                                                                                                                       //
      return unwrapFrontendCap(verification, "verifiedEmail", function (verification) {                                // 311
        if (verification.tabId !== tabId.toString("hex")) {                                                            // 312
          throw new Error("VerifiedEmail is from a different tab");                                                    // 313
        }                                                                                                              // 314
                                                                                                                       //
        if (_this3._services) {                                                                                        // 316
          // Since this verifier is restricted to specific services, only indicate a match if the                      // 317
          // VerifiedEmail was for the correct verifier ID. (If our _services is null, then we                         // 318
          // match all services, and therefore all VerifiedEmails.)                                                    // 319
          if (verification.verifierId !== _this3._id) {                                                                // 320
            throw new Error("VerifierEmail is for a different EmailVerifier.");                                        // 321
          }                                                                                                            // 322
        }                                                                                                              // 323
                                                                                                                       //
        return verification.address;                                                                                   // 325
      });                                                                                                              // 326
    }                                                                                                                  // 327
                                                                                                                       //
    return verifyEmail;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  return EmailVerifierImpl;                                                                                            //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
;                                                                                                                      // 328
                                                                                                                       //
var VerifiedEmailImpl = function (_PersistentImpl2) {                                                                  //
  (0, _inherits3.default)(VerifiedEmailImpl, _PersistentImpl2);                                                        //
                                                                                                                       //
  function VerifiedEmailImpl(db, saveTemplate) {                                                                       // 331
    (0, _classCallCheck3.default)(this, VerifiedEmailImpl);                                                            // 331
    return (0, _possibleConstructorReturn3.default)(this, _PersistentImpl2.call(this, db, saveTemplate));              // 331
  }                                                                                                                    // 333
                                                                                                                       //
  return VerifiedEmailImpl;                                                                                            //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
function getVerifiedEmails(db, userId, verifierId) {                                                                   // 336
  // Get all of the email addresses verified as belonging to the given user using the given                            // 337
  // verifier.                                                                                                         // 338
  var services = null;                                                                                                 // 340
  var couldAddAnother = true;                                                                                          // 341
                                                                                                                       //
  if (verifierId) {                                                                                                    // 343
    var verifier = db.collections.apiTokens.findOne({                                                                  // 344
      "frontendRef.emailVerifier.id": verifierId                                                                       // 345
    });                                                                                                                // 345
    if (!verifier) return []; // invalid verifier                                                                      // 346
                                                                                                                       //
    var verifierInfo = verifier.frontendRef.emailVerifier;                                                             // 347
                                                                                                                       //
    if (verifierInfo.services) {                                                                                       // 349
      // Limit to the listed services.                                                                                 // 350
      services = {};                                                                                                   // 351
      verifierInfo.services.forEach(function (service) {                                                               // 352
        return services[service] = true;                                                                               // 352
      });                                                                                                              // 352
      couldAddAnother = !!services.email;                                                                              // 353
    }                                                                                                                  // 354
  }                                                                                                                    // 355
                                                                                                                       //
  var user = Meteor.users.findOne(userId);                                                                             // 357
  var emails = {}; // map address -> true, for uniquification                                                          // 358
                                                                                                                       //
  Meteor.users.find({                                                                                                  // 359
    _id: {                                                                                                             // 359
      $in: SandstormDb.getUserIdentityIds(user)                                                                        // 359
    }                                                                                                                  // 359
  }).forEach(function (identity) {                                                                                     // 359
    if (!services || services[identity.profile.service]) {                                                             // 360
      SandstormDb.getVerifiedEmails(identity).forEach(function (email) {                                               // 361
        emails[email.email] = true;                                                                                    // 361
      });                                                                                                              // 361
    }                                                                                                                  // 362
  });                                                                                                                  // 363
  return {                                                                                                             // 365
    emails: Object.keys(emails),                                                                                       // 365
    couldAddAnother: couldAddAnother                                                                                   // 365
  };                                                                                                                   // 365
} // TODO(cleanup): Meteor.startup() needed because 00-startup.js runs *after* code in subdirectories                  // 366
//   (ugh).                                                                                                            // 369
                                                                                                                       //
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 370
  globalFrontendRefRegistry.register({                                                                                 // 371
    frontendRefField: "emailVerifier",                                                                                 // 372
    typeId: EmailRpc.EmailVerifier.typeId,                                                                             // 373
    restore: function (db, saveTemplate, params) {                                                                     // 375
      return new Capnp.Capability(new EmailVerifierImpl(db, saveTemplate, params), EmailImpl.PersistentEmailVerifier);
    },                                                                                                                 // 378
    validate: function (db, session, value) {                                                                          // 380
      check(value, {                                                                                                   // 381
        services: Match.Optional([String])                                                                             // 381
      });                                                                                                              // 381
      var services = value.services;                                                                                   // 383
                                                                                                                       //
      if (services) {                                                                                                  // 384
        services.forEach(function (service) {                                                                          // 385
          if (!Accounts.identityServices[service]) {                                                                   // 386
            throw new Error("No such identity service: " + service);                                                   // 387
          }                                                                                                            // 388
        });                                                                                                            // 389
      }                                                                                                                // 390
                                                                                                                       //
      value.id = Crypto.randomBytes(16).toString("base64");                                                            // 392
      return {                                                                                                         // 394
        descriptor: {                                                                                                  // 395
          tags: [{                                                                                                     // 395
            id: EmailRpc.EmailVerifier.typeId                                                                          // 395
          }]                                                                                                           // 395
        },                                                                                                             // 395
        requirements: [],                                                                                              // 396
        frontendRef: value                                                                                             // 397
      };                                                                                                               // 394
    },                                                                                                                 // 399
    query: function (db, userId, value) {                                                                              // 401
      var results = [];                                                                                                // 402
      results.push({                                                                                                   // 404
        _id: "emailverifier-all",                                                                                      // 405
        frontendRef: {                                                                                                 // 406
          emailVerifier: {}                                                                                            // 406
        },                                                                                                             // 406
        cardTemplate: "emailVerifierPowerboxCard"                                                                      // 407
      });                                                                                                              // 404
                                                                                                                       //
      for (var name in meteorBabelHelpers.sanitizeForInObject(Accounts.identityServices)) {                            // 410
        if (Accounts.identityServices[name].isEnabled()) {                                                             // 411
          results.push({                                                                                               // 412
            _id: "emailverifier-" + name,                                                                              // 413
            frontendRef: {                                                                                             // 414
              emailVerifier: {                                                                                         // 414
                services: [name]                                                                                       // 414
              }                                                                                                        // 414
            },                                                                                                         // 414
            cardTemplate: "emailVerifierPowerboxCard"                                                                  // 415
          });                                                                                                          // 412
        }                                                                                                              // 417
      }                                                                                                                // 418
                                                                                                                       //
      ;                                                                                                                // 418
      return results;                                                                                                  // 420
    }                                                                                                                  // 421
  });                                                                                                                  // 371
  globalFrontendRefRegistry.register({                                                                                 // 424
    frontendRefField: "verifiedEmail",                                                                                 // 425
    typeId: EmailRpc.VerifiedEmail.typeId,                                                                             // 426
    restore: function (db, saveTemplate) {                                                                             // 428
      return new Capnp.Capability(new VerifiedEmailImpl(db, saveTemplate), EmailImpl.PersistentVerifiedEmail);         // 429
    },                                                                                                                 // 431
    validate: function (db, session, value) {                                                                          // 433
      check(value, {                                                                                                   // 434
        verifierId: Match.Optional(String),                                                                            // 434
        address: String                                                                                                // 434
      });                                                                                                              // 434
                                                                                                                       //
      if (!session.userId) {                                                                                           // 436
        throw new Meteor.Error(403, "Not logged in.");                                                                 // 437
      } // Verify that the address actually belongs to the user.                                                       // 438
                                                                                                                       //
                                                                                                                       //
      if (!_.contains(getVerifiedEmails(db, session.userId, value.verifierId).emails, value.address)) {                // 442
        throw new Meteor.Error(403, "User has no such verified address");                                              // 444
      } // Add the session's tabId.                                                                                    // 445
                                                                                                                       //
                                                                                                                       //
      value.tabId = session.tabId; // Build the descriptor, which contains the verifier ID.                            // 448
                                                                                                                       //
      var tagValue = value.verifierId && Capnp.serialize(EmailRpc.VerifiedEmail.PowerboxTag, {                         // 451
        verifierId: new Buffer(value.verifierId, "hex")                                                                // 453
      });                                                                                                              // 453
      return {                                                                                                         // 455
        descriptor: {                                                                                                  // 456
          tags: [{                                                                                                     // 456
            id: EmailRpc.VerifiedEmail.typeId,                                                                         // 456
            value: tagValue                                                                                            // 456
          }]                                                                                                           // 456
        },                                                                                                             // 456
        requirements: [],                                                                                              // 457
        frontendRef: value                                                                                             // 458
      };                                                                                                               // 455
    },                                                                                                                 // 460
    query: function (db, userId, value) {                                                                              // 462
      var verifierId = value && Capnp.parse(EmailRpc.VerifiedEmail.PowerboxTag, value).verifierId.toString("base64");  // 463
      var verified = getVerifiedEmails(db, userId, verifierId);                                                        // 465
      var result = verified.emails.map(function (address) {                                                            // 466
        return {                                                                                                       // 466
          _id: "email-" + address,                                                                                     // 467
          frontendRef: {                                                                                               // 468
            verifiedEmail: {                                                                                           // 468
              verifierId: verifierId,                                                                                  // 468
              address: address                                                                                         // 468
            }                                                                                                          // 468
          },                                                                                                           // 468
          cardTemplate: "verifiedEmailPowerboxCard",                                                                   // 469
          searchTerms: [address]                                                                                       // 470
        };                                                                                                             // 466
      });                                                                                                              // 466
                                                                                                                       //
      if (verified.couldAddAnother) {                                                                                  // 473
        result.push({                                                                                                  // 474
          _id: "email-add-a-new-one",                                                                                  // 475
          frontendRef: {                                                                                               // 476
            verifiedEmail: {                                                                                           // 476
              verifierId: verifierId                                                                                   // 476
            }                                                                                                          // 476
          },                                                                                                           // 476
          cardTemplate: "addNewVerifiedEmailPowerboxCard",                                                             // 477
          configureTemplate: "addNewVerifiedEmailPowerboxConfiguration"                                                // 478
        });                                                                                                            // 474
      }                                                                                                                // 480
                                                                                                                       //
      return result;                                                                                                   // 482
    }                                                                                                                  // 483
  });                                                                                                                  // 424
});                                                                                                                    // 485
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"00-startup.js":["/imports/db-deprecated.js","/imports/server/frontend-ref.js","/imports/server/persistent.js","/imports/server/migrations.js","/imports/constants.js","/imports/server/async-helpers.js","/imports/server/networking.js","fibers","url","capnp",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/00-startup.js                                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
module.import("/imports/db-deprecated.js");                                                                            // 1
var FrontendRefRegistry = void 0;                                                                                      // 1
module.import("/imports/server/frontend-ref.js", {                                                                     // 1
  "FrontendRefRegistry": function (v) {                                                                                // 1
    FrontendRefRegistry = v;                                                                                           // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var PersistentImpl = void 0;                                                                                           // 1
module.import("/imports/server/persistent.js", {                                                                       // 1
  "PersistentImpl": function (v) {                                                                                     // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var migrateToLatest = void 0;                                                                                          // 1
module.import("/imports/server/migrations.js", {                                                                       // 1
  "migrateToLatest": function (v) {                                                                                    // 1
    migrateToLatest = v;                                                                                               // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var ACCOUNT_DELETION_SUSPENSION_TIME = void 0;                                                                         // 1
module.import("/imports/constants.js", {                                                                               // 1
  "ACCOUNT_DELETION_SUSPENSION_TIME": function (v) {                                                                   // 1
    ACCOUNT_DELETION_SUSPENSION_TIME = v;                                                                              // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var onInMeteor = void 0;                                                                                               // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "onInMeteor": function (v) {                                                                                         // 1
    onInMeteor = v;                                                                                                    // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var monkeyPatchHttp = void 0;                                                                                          // 1
module.import("/imports/server/networking.js", {                                                                       // 1
  "monkeyPatchHttp": function (v) {                                                                                    // 1
    monkeyPatchHttp = v;                                                                                               // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
var Fiber = void 0;                                                                                                    // 1
module.import("fibers", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    Fiber = v;                                                                                                         // 1
  }                                                                                                                    // 1
}, 7);                                                                                                                 // 1
                                                                                                                       //
var url = require("url");                                                                                              // 24
                                                                                                                       //
var Capnp = require("capnp");                                                                                          // 25
                                                                                                                       //
globalFrontendRefRegistry = new FrontendRefRegistry();                                                                 // 27
SandstormPowerbox.registerUiViewQueryHandler(globalFrontendRefRegistry);                                               // 29
                                                                                                                       //
if (global.BlackrockPayments && BlackrockPayments.registerPaymentsApi) {                                               // 31
  // TODO(cleanup): Meteor.startup() needed because unwrapFrontendCap is not defined yet when this                     // 32
  //   first runs. Move it into an import.                                                                             // 33
  Meteor.startup(function () {                                                                                         // 34
    BlackrockPayments.registerPaymentsApi(globalFrontendRefRegistry, PersistentImpl, unwrapFrontendCap);               // 35
  });                                                                                                                  // 37
}                                                                                                                      // 38
                                                                                                                       //
getWildcardOrigin = globalDb.getWildcardOrigin.bind(globalDb);                                                         // 40
Meteor.onConnection(function (connection) {                                                                            // 42
  // TODO(cleanup): This is the best way I've thought of so far to allow methods declared in                           // 43
  //   packages to actually use the DB, but it's pretty sad.                                                           // 44
  connection.sandstormDb = globalDb;                                                                                   // 45
  connection.frontendRefRegistry = globalFrontendRefRegistry;                                                          // 46
});                                                                                                                    // 47
SandstormDb.periodicCleanup(5 * 60 * 1000, SandstormPermissions.cleanupSelfDestructing(globalDb));                     // 48
SandstormDb.periodicCleanup(10 * 60 * 1000, SandstormPermissions.cleanupClientPowerboxTokens(globalDb));               // 49
SandstormDb.periodicCleanup(60 * 60 * 1000, function () {                                                              // 51
  globalDb.cleanupExpiredAssetUploads();                                                                               // 52
});                                                                                                                    // 53
SandstormDb.periodicCleanup(24 * 60 * 60 * 1000, function () {                                                         // 54
  SandstormAutoupdateApps.updateAppIndex(globalDb);                                                                    // 55
});                                                                                                                    // 56
var deleteAccount = global.BlackrockPayments && global.BlackrockPayments.deleteAccount;                                // 57
SandstormDb.periodicCleanup(24 * 60 * 60 * 1000, function () {                                                         // 58
  globalDb.deletePendingAccounts(ACCOUNT_DELETION_SUSPENSION_TIME, globalBackend, deleteAccount);                      // 59
});                                                                                                                    // 61
monkeyPatchHttp(globalDb, HTTP);                                                                                       // 63
Meteor.startup(function () {                                                                                           // 65
  migrateToLatest(globalDb, globalBackend);                                                                            // 65
}); // If there are multiple replicas, prefix every log message with our replica number.                               // 65
                                                                                                                       //
if ("replicaNumber" in Meteor.settings) {                                                                              // 68
  var patchConsole = function (name) {                                                                                 // 68
    var old = console[name];                                                                                           // 72
                                                                                                                       //
    console[name] = function () {                                                                                      // 73
      // Meteor in dev mode writes "LISTENING" to tell the dev runner that it's ready to accept                        // 74
      // connections.                                                                                                  // 75
      if (arguments.length == 1 && arguments[0] == "LISTENING") {                                                      // 76
        old.apply(this, arguments);                                                                                    // 77
      } else {                                                                                                         // 78
        old.apply(this, [prefix].concat(Array.prototype.slice.call(arguments)));                                       // 79
      }                                                                                                                // 80
    };                                                                                                                 // 81
  };                                                                                                                   // 82
                                                                                                                       //
  var prefix = "replica" + Meteor.settings.replicaNumber.toString() + ":";                                             // 69
  patchConsole("log");                                                                                                 // 84
  patchConsole("info");                                                                                                // 85
  patchConsole("warn");                                                                                                // 86
  patchConsole("error");                                                                                               // 87
} // Make the fiber pool size infinite(ish)!                                                                           // 88
//                                                                                                                     // 91
// Each fiber created adds an entry to `v8::Isolate::ThreadDataTable`. Unfortunatley, items are                        // 92
// never deleted from this table. So if we create and delete a lot of fibers, then we leak memory.                     // 93
// Worse yet, the table is represented as a linked list, and v8 performs a linear scan of this                         // 94
// linked list every time we switch fibers. We've seen cases where Sandstorm was spending 65% of                       // 95
// its CPU time just scanning this list! The v8 people say this is "working as intended".                              // 96
//                                                                                                                     // 97
// The fibers package has implemented a work-around by maintaining a fiber pool. Fibers are not                        // 98
// deleted; they are returned to the pool. Unfortunately the pool has a default size of 120. So                        // 99
// if we pass 120 simultaneous fibers, we start leaking and slowing down. It's very easy to hit                        // 100
// this number with a few dozen users present.                                                                         // 101
//                                                                                                                     // 102
// Up until it hits the limit, the fibers module will grow the pool dynamically, starting with an                      // 103
// empty pool and adding each new fiber to it. Therefore, if we set the pool size to an impossibly                     // 104
// large number, we effectively get a pool size equal to the maximum number of simultaneous fibers                     // 105
// seen. This is exactly what we want! Now no fibers are ever deleted, so we never leak. A                             // 106
// Sandstorm server that sees a brief surge of traffic may end up holding on to unused RAM                             // 107
// long-term, but this is a relatively obscure problem.                                                                // 108
//                                                                                                                     // 109
// I initially tried to use the value `Infinity` here, but somehow when this made its way down into                    // 110
// the C++ code and was converted to an integer, that integer ended up being zero. So instead we                       // 111
// use 1e9 (1 billion), which ought to be enough for anyone.                                                           // 112
//                                                                                                                     // 113
// Third-party issues, for reference:                                                                                  // 114
//                                                                                                                     // 115
//    https://bugs.chromium.org/p/v8/issues/detail?id=5338                                                             // 116
//    https://bugs.chromium.org/p/v8/issues/detail?id=3777                                                             // 117
//    https://github.com/laverdet/node-fibers/issues/305                                                               // 118
                                                                                                                       //
                                                                                                                       //
Fiber.poolSize = 1e9; // Special debugging enabled on Blackrock only.                                                  // 120
                                                                                                                       //
if ("replicaNumber" in Meteor.settings) {                                                                              // 123
  var sampleStackIfTooManyFibers = function () {                                                                       // 123
    if (Fiber.fibersCreated > TOO_MANY_FIBERS && sampleCounter++ % FIBER_SAMPLING_PERIOD == 0) {                       // 148
      var stack = new Error().stack;                                                                                   // 149
      console.warn("There are too many fibers!", stack);                                                               // 150
    }                                                                                                                  // 151
  }; // Sample calls to inMeteor().                                                                                    // 152
                                                                                                                       //
                                                                                                                       //
  // Also take samples from various Meteor infrastructure that schedules fibers.                                       // 157
  var addFiberSampling = function (prototype, functionName) {                                                          // 123
    var old = prototype[functionName];                                                                                 // 159
                                                                                                                       //
    prototype[functionName] = function () {                                                                            // 160
      sampleStackIfTooManyFibers();                                                                                    // 161
      return old.apply(this, arguments);                                                                               // 162
    };                                                                                                                 // 163
  };                                                                                                                   // 164
                                                                                                                       //
  console.warn("Fiber bomb defense enabled."); // TEMPORARY: Monitor the number of fibers created and kill the process any time it goes over
  //   2000. Unfortunately, due to the aforementioned linked list in ThreadDataTable, the process                      // 127
  //   will become unreasonably slow once the list gets this big. It's better to kill the process                      // 128
  //   so that it restarts fresh rather than to let the Sandstorm server become unresponsive.                          // 129
  // TODO(soon): Remove this when the bug is fixed.                                                                    // 130
                                                                                                                       //
  setInterval(function () {                                                                                            // 131
    if (Fiber.fibersCreated > 2000) {                                                                                  // 132
      console.error("Process has allocated more than 2000 concurrent fibers. Due to " + "https://bugs.chromium.org/p/v8/issues/detail?id=5338 it will become extremely slow " + "unless we restart it. ABORTING");
      process.abort();                                                                                                 // 137
    }                                                                                                                  // 138
  }, 5000); // Let's also log some stack traces when fiber count gets high so we can find out what's                   // 139
  // happening.                                                                                                        // 142
                                                                                                                       //
  var TOO_MANY_FIBERS = 1000; // Start sampling when we have more than this many fibers.                               // 143
                                                                                                                       //
  var FIBER_SAMPLING_PERIOD = 100; // Sample one stack per this many fibers created.                                   // 144
                                                                                                                       //
  var sampleCounter = 0;                                                                                               // 146
  onInMeteor(sampleStackIfTooManyFibers);                                                                              // 155
  addFiberSampling(Meteor._SynchronousQueue.prototype, "runTask");                                                     // 166
  addFiberSampling(Meteor._SynchronousQueue.prototype, "queueTask");                                                   // 167
  addFiberSampling(Meteor, "bindEnvironment");                                                                         // 168
}                                                                                                                      // 169
                                                                                                                       //
if ("replicaNumber" in Meteor.settings) {                                                                              // 171
  // On Blackrock, dump debug info about local capability counts every minute.                                         // 172
  Meteor.setInterval(Capnp.dumpLocalCapTypeCounts, 60000);                                                             // 173
}                                                                                                                      // 174
                                                                                                                       //
OAuth._checkRedirectUrlOrigin = function (redirectUrl) {                                                               // 176
  // Mostly copied from meteor/packages/oauth/oauth_server.js                                                          // 177
  // We override this method in order to support login from stand-alone grain domains.                                 // 178
  var appHost = Meteor.absoluteUrl();                                                                                  // 179
  var appHostReplacedLocalhost = Meteor.absoluteUrl(undefined, {                                                       // 180
    replaceLocalhost: true                                                                                             // 181
  });                                                                                                                  // 180
  var redirectParsed = url.parse(redirectUrl);                                                                         // 184
  return !(redirectUrl.substr(0, appHost.length) === appHost || redirectUrl.substr(0, appHostReplacedLocalhost.length) === appHostReplacedLocalhost || globalDb.hostIsStandalone(redirectParsed.hostname));
};                                                                                                                     // 191
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"account-suspension.js":["/imports/server/email.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/account-suspension.js                                                                                        //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var send = void 0;                                                                                                     // 1
module.import("/imports/server/email.js", {                                                                            // 1
  "send": function (v) {                                                                                               // 1
    send = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
function sendDeletionEmails(db, deletedUserId, byAdminUserId, feedback) {                                              // 19
  var deletedUser = db.getUser(deletedUserId);                                                                         // 20
                                                                                                                       //
  var userEmail = _.findWhere(SandstormDb.getUserEmails(deletedUser), {                                                // 22
    primary: true                                                                                                      // 22
  });                                                                                                                  // 22
                                                                                                                       //
  if (!byAdminUserId) {                                                                                                // 23
    // This was initiated by the user, send them an email                                                              // 23
    if (!userEmail) {                                                                                                  // 24
      console.error("Couldn't send deletion email. No address found for user with userId:", deletedUser._id);          // 25
    } else {                                                                                                           // 27
      var _emailOptions = {                                                                                            // 28
        from: db.getReturnAddress(),                                                                                   // 29
        subject: "Your account on " + db.getServerTitle() + " will be deleted in 7 days.",                             // 30
        text: "You have requested that your Sandstorm account on " + db.getServerTitle() + " be deleted. Your account has been suspended and will be fully deleted in seven days. If you change your mind, log into " + process.env.ROOT_URL + " to cancel the process.\n\nIf you did not request this deletion, please contact the server administrator immediately."
      };                                                                                                               // 28
                                                                                                                       //
      try {                                                                                                            // 35
        _emailOptions.to = userEmail.email;                                                                            // 36
        send(_emailOptions);                                                                                           // 37
      } catch (err) {                                                                                                  // 38
        console.error("Failed to send deletion email to user (id=" + deletedUser._id + ") with error: " + err);        // 39
      }                                                                                                                // 41
    }                                                                                                                  // 42
  }                                                                                                                    // 43
                                                                                                                       //
  var deleteUserString = userEmail && userEmail.email || "userId=" + deletedUser._id;                                  // 45
  var emailOptions = {                                                                                                 // 46
    from: db.getReturnAddress(),                                                                                       // 47
    subject: "Account for " + deleteUserString + " on " + db.getServerTitle() + " will be deleted in 7 days"           // 48
  };                                                                                                                   // 46
                                                                                                                       //
  if (byAdminUserId) {                                                                                                 // 51
    var initiatingAdmin = db.getUser(byAdminUserId);                                                                   // 52
    var adminName = db.getIdentity(initiatingAdmin.loginIdentities[0].id).profile.name;                                // 53
    emailOptions.text = adminName + " has requested that the Sandstorm account held by " + deleteUserString + " on " + db.getServerTitle() + " be deleted. The account has been suspended and will be fully deleted in seven days. To cancel the deletion, go to: " + process.env.ROOT_URL + "/admin/users/" + deletedUser._id;
  } else {                                                                                                             // 55
    emailOptions.text = deleteUserString + " has requested that their account be deleted on " + db.getServerTitle() + ". The account has been suspended and will be fully deleted in seven days. To cancel the deletion, go to: " + process.env.ROOT_URL + "/admin/users/" + deletedUser._id;
                                                                                                                       //
    if (feedback) {                                                                                                    // 57
      emailOptions.text += "\nUser gave the following feedback: " + feedback;                                          // 58
    }                                                                                                                  // 59
  }                                                                                                                    // 60
                                                                                                                       //
  Meteor.users.find({                                                                                                  // 62
    isAdmin: true                                                                                                      // 62
  }).forEach(function (user) {                                                                                         // 62
    var email = _.findWhere(SandstormDb.getUserEmails(user), {                                                         // 63
      primary: true                                                                                                    // 63
    });                                                                                                                // 63
                                                                                                                       //
    if (!email) {                                                                                                      // 64
      console.error("No email found for admin with userId:", user._id);                                                // 65
      return;                                                                                                          // 66
    }                                                                                                                  // 67
                                                                                                                       //
    try {                                                                                                              // 69
      emailOptions.to = email.email;                                                                                   // 70
      send(emailOptions);                                                                                              // 71
    } catch (err) {                                                                                                    // 72
      console.error("Failed to send deletion email to admin (id=" + user._id + ") with error: " + err);                // 73
    }                                                                                                                  // 75
  });                                                                                                                  // 76
}                                                                                                                      // 77
                                                                                                                       //
Meteor.methods({                                                                                                       // 79
  suspendAccount: function (userId, willDelete) {                                                                      // 80
    check(userId, String);                                                                                             // 81
    check(willDelete, Boolean);                                                                                        // 82
                                                                                                                       //
    if (!isAdmin()) {                                                                                                  // 84
      throw new Meteor.Error(403, "Only admins can suspend other users.");                                             // 85
    }                                                                                                                  // 86
                                                                                                                       //
    if (userId === Meteor.userId()) {                                                                                  // 88
      throw new Meteor.Error(400, "Admins cannot suspend their own accounts from the admin page. Please go to your account setttings.");
    }                                                                                                                  // 90
                                                                                                                       //
    var db = this.connection.sandstormDb;                                                                              // 92
                                                                                                                       //
    if (global.BlackrockPayments) {                                                                                    // 94
      BlackrockPayments.suspendAccount(db, userId);                                                                    // 95
    }                                                                                                                  // 96
                                                                                                                       //
    db.suspendAccount(userId, Meteor.userId(), willDelete);                                                            // 98
                                                                                                                       //
    if (willDelete) {                                                                                                  // 100
      sendDeletionEmails(db, userId, Meteor.userId());                                                                 // 101
    }                                                                                                                  // 102
  },                                                                                                                   // 103
  deleteOwnAccount: function (feedback) {                                                                              // 105
    var db = this.connection.sandstormDb;                                                                              // 106
                                                                                                                       //
    if (!Meteor.userId()) {                                                                                            // 107
      throw new Meteor.Error(403, "Must be logged in to delete an account");                                           // 108
    }                                                                                                                  // 109
                                                                                                                       //
    if (db.isUserInOrganization(Meteor.user())) {                                                                      // 111
      throw new Meteor.Error(403, "Users in an organization cannot delete their own account. " + "Please ask your admin to do it for you.");
    }                                                                                                                  // 114
                                                                                                                       //
    if (global.BlackrockPayments) {                                                                                    // 116
      BlackrockPayments.suspendAccount(db, Meteor.userId());                                                           // 117
    }                                                                                                                  // 118
                                                                                                                       //
    db.suspendAccount(Meteor.userId(), null, true);                                                                    // 120
    sendDeletionEmails(db, Meteor.userId(), null, feedback);                                                           // 122
  },                                                                                                                   // 123
  unsuspendAccount: function (userId) {                                                                                // 125
    check(userId, String);                                                                                             // 126
                                                                                                                       //
    if (!isAdmin()) {                                                                                                  // 128
      throw new Meteor.Error(403, "Only admins can unsuspend other users.");                                           // 129
    }                                                                                                                  // 130
                                                                                                                       //
    this.connection.sandstormDb.unsuspendAccount(userId, Meteor.userId());                                             // 132
  },                                                                                                                   // 133
  unsuspendOwnAccount: function () {                                                                                   // 135
    if (!Meteor.userId()) {                                                                                            // 136
      throw new Meteor.Error(403, "Must be logged in to unsuspend an account");                                        // 137
    }                                                                                                                  // 138
                                                                                                                       //
    this.connection.sandstormDb.unsuspendAccount(Meteor.userId());                                                     // 140
  }                                                                                                                    // 141
});                                                                                                                    // 79
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"admin-server.js":["babel-runtime/helpers/objectWithoutProperties","meteor/meteor","fs","crypto","heapdump","/imports/server/constants.js","/imports/server/auth.js","/imports/server/email.js","/imports/server/observe-helpers.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/admin-server.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _objectWithoutProperties2 = require("babel-runtime/helpers/objectWithoutProperties");                              //
                                                                                                                       //
var _objectWithoutProperties3 = _interopRequireDefault(_objectWithoutProperties2);                                     //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Fs = void 0;                                                                                                       // 1
module.import("fs", {                                                                                                  // 1
  "default": function (v) {                                                                                            // 1
    Fs = v;                                                                                                            // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Crypto = void 0;                                                                                                   // 1
module.import("crypto", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    Crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var Heapdump = void 0;                                                                                                 // 1
module.import("heapdump", {                                                                                            // 1
  "default": function (v) {                                                                                            // 1
    Heapdump = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var SANDSTORM_LOGDIR = void 0;                                                                                         // 1
module.import("/imports/server/constants.js", {                                                                        // 1
  "SANDSTORM_LOGDIR": function (v) {                                                                                   // 1
    SANDSTORM_LOGDIR = v;                                                                                              // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var clearAdminToken = void 0,                                                                                          // 1
    checkAuth = void 0,                                                                                                // 1
    tokenIsValid = void 0,                                                                                             // 1
    tokenIsSetupSession = void 0;                                                                                      // 1
module.import("/imports/server/auth.js", {                                                                             // 1
  "clearAdminToken": function (v) {                                                                                    // 1
    clearAdminToken = v;                                                                                               // 1
  },                                                                                                                   // 1
  "checkAuth": function (v) {                                                                                          // 1
    checkAuth = v;                                                                                                     // 1
  },                                                                                                                   // 1
  "tokenIsValid": function (v) {                                                                                       // 1
    tokenIsValid = v;                                                                                                  // 1
  },                                                                                                                   // 1
  "tokenIsSetupSession": function (v) {                                                                                // 1
    tokenIsSetupSession = v;                                                                                           // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var sendEmail = void 0;                                                                                                // 1
module.import("/imports/server/email.js", {                                                                            // 1
  "send": function (v) {                                                                                               // 1
    sendEmail = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
var fillUndefinedForChangedDoc = void 0;                                                                               // 1
module.import("/imports/server/observe-helpers.js", {                                                                  // 1
  "fillUndefinedForChangedDoc": function (v) {                                                                         // 1
    fillUndefinedForChangedDoc = v;                                                                                    // 1
  }                                                                                                                    // 1
}, 7);                                                                                                                 // 1
var publicAdminSettings = ["google", "github", "ldap", "saml", "emailToken", "splashUrl", "signupDialog", "adminAlert", "adminAlertTime", "adminAlertUrl", "termsUrl", "privacyUrl", "appMarketUrl", "appIndexUrl", "appUpdatesEnabled", "serverTitle", "returnAddress", "ldapNameField", "organizationMembership", "organizationSettings", "whitelabelCustomLoginProviderName", "whitelabelCustomLogoAssetId", "whitelabelHideSendFeedback", "whitelabelHideTroubleshooting", "whiteLabelHideAbout", "whitelabelUseServerTitleForHomeText", "quotaEnabled", "quotaLdapEnabled", "billingPromptUrl"];
var smtpConfigShape = {                                                                                                // 43
  hostname: String,                                                                                                    // 44
  port: Number,                                                                                                        // 45
  auth: {                                                                                                              // 46
    user: String,                                                                                                      // 47
    pass: String                                                                                                       // 48
  },                                                                                                                   // 46
  returnAddress: String                                                                                                // 50
};                                                                                                                     // 43
Meteor.methods({                                                                                                       // 53
  setAccountSetting: function (token, serviceName, value) {                                                            // 54
    checkAuth(token);                                                                                                  // 55
    check(serviceName, String);                                                                                        // 56
    check(value, Boolean); // TODO(someday): currently this relies on the fact that an account is tied to a single     // 57
    // identity, and thus has only that entry in "services". This will need to be looked at when                       // 60
    // multiple login methods/identities are allowed for a single account.                                             // 61
                                                                                                                       //
    if (!value && !tokenIsValid(token) && !tokenIsSetupSession(token) && serviceName in Meteor.user().services) {      // 62
      throw new Meteor.Error(403, "You can not disable the login service that your account uses.");                    // 63
    } // Only check configurations for OAuth services.                                                                 // 65
                                                                                                                       //
                                                                                                                       //
    var oauthServices = ["google", "github"];                                                                          // 68
                                                                                                                       //
    if (value && oauthServices.indexOf(serviceName) != -1) {                                                           // 69
      var config = ServiceConfiguration.configurations.findOne({                                                       // 70
        service: serviceName                                                                                           // 70
      });                                                                                                              // 70
                                                                                                                       //
      if (!config) {                                                                                                   // 71
        throw new Meteor.Error(403, "You must configure the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
      }                                                                                                                // 74
                                                                                                                       //
      if (!config.clientId || !config.secret) {                                                                        // 76
        throw new Meteor.Error(403, "You must provide a non-empty clientId and secret for the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
      }                                                                                                                // 79
    }                                                                                                                  // 80
                                                                                                                       //
    Settings.upsert({                                                                                                  // 82
      _id: serviceName                                                                                                 // 82
    }, {                                                                                                               // 82
      $set: {                                                                                                          // 82
        value: value                                                                                                   // 82
      }                                                                                                                // 82
    });                                                                                                                // 82
                                                                                                                       //
    if (value) {                                                                                                       // 83
      Settings.update({                                                                                                // 84
        _id: serviceName                                                                                               // 84
      }, {                                                                                                             // 84
        $unset: {                                                                                                      // 84
          automaticallyReset: 1                                                                                        // 84
        }                                                                                                              // 84
      });                                                                                                              // 84
    }                                                                                                                  // 85
  },                                                                                                                   // 86
  setSmtpConfig: function (token, config) {                                                                            // 88
    checkAuth(token);                                                                                                  // 89
    check(config, smtpConfigShape);                                                                                    // 90
    Settings.upsert({                                                                                                  // 92
      _id: "smtpConfig"                                                                                                // 92
    }, {                                                                                                               // 92
      $set: {                                                                                                          // 92
        value: config                                                                                                  // 92
      }                                                                                                                // 92
    });                                                                                                                // 92
  },                                                                                                                   // 93
  disableEmail: function (token) {                                                                                     // 95
    checkAuth(token);                                                                                                  // 96
    var db = this.connection.sandstormDb;                                                                              // 98
    db.collections.settings.update({                                                                                   // 99
      _id: "smtpConfig"                                                                                                // 99
    }, {                                                                                                               // 99
      $set: {                                                                                                          // 99
        "value.hostname": ""                                                                                           // 99
      }                                                                                                                // 99
    });                                                                                                                // 99
  },                                                                                                                   // 100
  setSetting: function (token, name, value) {                                                                          // 102
    checkAuth(token);                                                                                                  // 103
    check(name, String);                                                                                               // 104
    check(value, Match.OneOf(null, String, Date, Boolean));                                                            // 105
    Settings.upsert({                                                                                                  // 107
      _id: name                                                                                                        // 107
    }, {                                                                                                               // 107
      $set: {                                                                                                          // 107
        value: value                                                                                                   // 107
      }                                                                                                                // 107
    });                                                                                                                // 107
  },                                                                                                                   // 108
  saveOrganizationSettings: function (token, params) {                                                                 // 110
    checkAuth(token);                                                                                                  // 111
    check(params, {                                                                                                    // 112
      membership: {                                                                                                    // 113
        emailToken: {                                                                                                  // 114
          enabled: Boolean,                                                                                            // 115
          domain: String                                                                                               // 116
        },                                                                                                             // 114
        google: {                                                                                                      // 118
          enabled: Boolean,                                                                                            // 119
          domain: String                                                                                               // 120
        },                                                                                                             // 118
        ldap: {                                                                                                        // 122
          enabled: Boolean                                                                                             // 123
        },                                                                                                             // 122
        saml: {                                                                                                        // 125
          enabled: Boolean                                                                                             // 126
        }                                                                                                              // 125
      },                                                                                                               // 113
      settings: {                                                                                                      // 129
        disallowGuests: Boolean,                                                                                       // 130
        shareContacts: Boolean                                                                                         // 131
      }                                                                                                                // 129
    });                                                                                                                // 112
    this.connection.sandstormDb.collections.settings.upsert({                                                          // 135
      _id: "organizationMembership"                                                                                    // 135
    }, {                                                                                                               // 135
      value: params.membership                                                                                         // 135
    });                                                                                                                // 135
    this.connection.sandstormDb.collections.settings.upsert({                                                          // 136
      _id: "organizationSettings"                                                                                      // 136
    }, {                                                                                                               // 136
      value: params.settings                                                                                           // 136
    });                                                                                                                // 136
  },                                                                                                                   // 137
  adminConfigureLoginService: function (token, options) {                                                              // 139
    checkAuth(token);                                                                                                  // 140
    check(options, Match.ObjectIncluding({                                                                             // 141
      service: String                                                                                                  // 141
    }));                                                                                                               // 141
    ServiceConfiguration.configurations.upsert({                                                                       // 143
      service: options.service                                                                                         // 143
    }, options);                                                                                                       // 143
  },                                                                                                                   // 144
  clearResumeTokensForService: function (token, serviceName) {                                                         // 146
    checkAuth(token);                                                                                                  // 147
    check(serviceName, String);                                                                                        // 148
    var query = {};                                                                                                    // 150
    query["services." + serviceName] = {                                                                               // 151
      $exists: true                                                                                                    // 151
    };                                                                                                                 // 151
    Meteor.users.find(query).forEach(function (identity) {                                                             // 152
      if (identity.services.resume && identity.services.resume.loginTokens && identity.services.resume.loginTokens.length > 0) {
        Meteor.users.update({                                                                                          // 155
          _id: identity._id                                                                                            // 155
        }, {                                                                                                           // 155
          $set: {                                                                                                      // 155
            "services.resume.loginTokens": []                                                                          // 155
          }                                                                                                            // 155
        });                                                                                                            // 155
      }                                                                                                                // 156
                                                                                                                       //
      Meteor.users.update({                                                                                            // 158
        "loginIdentities.id": identity._id                                                                             // 158
      }, {                                                                                                             // 158
        $set: {                                                                                                        // 159
          "services.resume.loginTokens": []                                                                            // 159
        }                                                                                                              // 159
      });                                                                                                              // 159
    });                                                                                                                // 160
  },                                                                                                                   // 161
  adminUpdateUser: function (token, userInfo) {                                                                        // 163
    checkAuth(token);                                                                                                  // 164
    check(userInfo, {                                                                                                  // 165
      userId: String,                                                                                                  // 166
      signupKey: Boolean,                                                                                              // 167
      isAdmin: Boolean                                                                                                 // 168
    });                                                                                                                // 165
    var userId = userInfo.userId;                                                                                      // 171
                                                                                                                       //
    if (userId === Meteor.userId() && !userInfo.isAdmin) {                                                             // 172
      throw new Meteor.Error(403, "User cannot remove admin permissions from itself.");                                // 173
    }                                                                                                                  // 174
                                                                                                                       //
    Meteor.users.update({                                                                                              // 176
      _id: userId                                                                                                      // 176
    }, {                                                                                                               // 176
      $set: _.omit(userInfo, ["_id", "userId"])                                                                        // 176
    });                                                                                                                // 176
  },                                                                                                                   // 177
  testSend: function (token, smtpConfig, to) {                                                                         // 179
    checkAuth(token);                                                                                                  // 180
    check(smtpConfig, smtpConfigShape);                                                                                // 181
    check(to, String);                                                                                                 // 182
    var returnAddress = smtpConfig.returnAddress,                                                                      // 179
        restConfig = (0, _objectWithoutProperties3.default)(smtpConfig, ["returnAddress"]);                            // 179
                                                                                                                       //
    try {                                                                                                              // 185
      sendEmail({                                                                                                      // 186
        to: to,                                                                                                        // 187
        from: {                                                                                                        // 188
          name: globalDb.getServerTitle(),                                                                             // 188
          address: returnAddress                                                                                       // 188
        },                                                                                                             // 188
        subject: "Testing your Sandstorm's SMTP setting",                                                              // 189
        text: "Success! Your outgoing SMTP is working.",                                                               // 190
        smtpConfig: restConfig                                                                                         // 191
      });                                                                                                              // 186
    } catch (e) {                                                                                                      // 193
      // Attempt to give more accurate error messages for a variety of known failure modes,                            // 194
      // and the actual exception data in the event a user hits a new failure mode.                                    // 195
      if (e.syscall === "getaddrinfo") {                                                                               // 196
        if (e.code === "EIO" || e.code === "ENOTFOUND") {                                                              // 197
          throw new Meteor.Error("getaddrinfo " + e.code, "Couldn't resolve \"" + smtpConfig.hostname + "\" - check for typos or broken DNS.");
        }                                                                                                              // 199
      } else if (e.syscall === "connect") {                                                                            // 200
        if (e.code === "ECONNREFUSED") {                                                                               // 201
          throw new Meteor.Error("connect ECONNREFUSED", "Server at " + smtpConfig.hostname + ":" + smtpConfig.port + " refused connection.  Check your settings, firewall rules, and that your mail server is up.");
        }                                                                                                              // 203
      } else if (e.name === "AuthError") {                                                                             // 204
        throw new Meteor.Error("auth error", "Authentication failed.  Check your credentials.  Message from " + smtpConfig.hostname + ": " + e.data);
      }                                                                                                                // 207
                                                                                                                       //
      throw new Meteor.Error("other-email-sending-error", "Error while trying to send test email: " + JSON.stringify(e));
    }                                                                                                                  // 210
  },                                                                                                                   // 211
  createSignupKey: function (token, note, quota) {                                                                     // 213
    checkAuth(token);                                                                                                  // 214
    check(note, String);                                                                                               // 215
    check(quota, Match.OneOf(undefined, null, Number));                                                                // 216
    var key = Random.id();                                                                                             // 218
    var content = {                                                                                                    // 219
      _id: key,                                                                                                        // 219
      used: false,                                                                                                     // 219
      note: note                                                                                                       // 219
    };                                                                                                                 // 219
    if (typeof quota === "number") content.quota = quota;                                                              // 220
    SignupKeys.insert(content);                                                                                        // 221
    return key;                                                                                                        // 222
  },                                                                                                                   // 223
  sendInvites: function (token, origin, from, list, subject, message, quota) {                                         // 225
    checkAuth(token);                                                                                                  // 226
    check(from, {                                                                                                      // 227
      name: String,                                                                                                    // 227
      address: String                                                                                                  // 227
    });                                                                                                                // 227
    check([origin, list, subject, message], [String]);                                                                 // 228
    check(quota, Match.OneOf(undefined, null, Number));                                                                // 229
                                                                                                                       //
    if (!from.address.trim()) {                                                                                        // 231
      throw new Meteor.Error(403, "Must enter 'from' address.");                                                       // 232
    }                                                                                                                  // 233
                                                                                                                       //
    if (!list.trim()) {                                                                                                // 235
      throw new Meteor.Error(403, "Must enter 'to' addresses.");                                                       // 236
    }                                                                                                                  // 237
                                                                                                                       //
    this.unblock();                                                                                                    // 239
    list = list.split("\n");                                                                                           // 241
                                                                                                                       //
    for (var i in meteorBabelHelpers.sanitizeForInObject(list)) {                                                      // 242
      var email = list[i].trim();                                                                                      // 243
                                                                                                                       //
      if (email) {                                                                                                     // 245
        var key = Random.id();                                                                                         // 246
        var content = {                                                                                                // 248
          _id: key,                                                                                                    // 249
          used: false,                                                                                                 // 250
          note: "E-mail invite to " + email,                                                                           // 251
          email: email,                                                                                                // 252
          definitelySent: false                                                                                        // 253
        };                                                                                                             // 248
        if (typeof quota === "number") content.quota = quota;                                                          // 255
        SignupKeys.insert(content);                                                                                    // 256
        sendEmail({                                                                                                    // 257
          to: email,                                                                                                   // 258
          from: from,                                                                                                  // 259
          envelopeFrom: globalDb.getReturnAddress(),                                                                   // 260
          subject: subject,                                                                                            // 261
          text: message.replace(/\$KEY/g, origin + "/signup/" + key)                                                   // 262
        });                                                                                                            // 257
        SignupKeys.update(key, {                                                                                       // 264
          $set: {                                                                                                      // 264
            definitelySent: true                                                                                       // 264
          }                                                                                                            // 264
        });                                                                                                            // 264
      }                                                                                                                // 265
    }                                                                                                                  // 266
                                                                                                                       //
    return {                                                                                                           // 268
      sent: true                                                                                                       // 268
    };                                                                                                                 // 268
  },                                                                                                                   // 269
  adminToggleDisableCap: function (token, capId, value) {                                                              // 271
    checkAuth(token);                                                                                                  // 272
    check(capId, String);                                                                                              // 273
    check(value, Boolean);                                                                                             // 274
                                                                                                                       //
    if (value) {                                                                                                       // 276
      ApiTokens.update({                                                                                               // 277
        _id: capId                                                                                                     // 277
      }, {                                                                                                             // 277
        $set: {                                                                                                        // 277
          revoked: true                                                                                                // 277
        }                                                                                                              // 277
      });                                                                                                              // 277
    } else {                                                                                                           // 278
      ApiTokens.update({                                                                                               // 279
        _id: capId                                                                                                     // 279
      }, {                                                                                                             // 279
        $set: {                                                                                                        // 279
          revoked: false                                                                                               // 279
        }                                                                                                              // 279
      });                                                                                                              // 279
    }                                                                                                                  // 280
  },                                                                                                                   // 281
  updateQuotas: function (token, list, quota) {                                                                        // 283
    checkAuth(token);                                                                                                  // 284
    check(list, String);                                                                                               // 285
    check(quota, Match.OneOf(undefined, null, Number));                                                                // 286
                                                                                                                       //
    if (!list.trim()) {                                                                                                // 288
      throw new Meteor.Error(400, "Must enter addresses.");                                                            // 289
    }                                                                                                                  // 290
                                                                                                                       //
    var items = list.split("\n");                                                                                      // 292
    var invalid = [];                                                                                                  // 293
                                                                                                                       //
    for (var i in meteorBabelHelpers.sanitizeForInObject(items)) {                                                     // 294
      var modifier = typeof quota === "number" ? {                                                                     // 295
        $set: {                                                                                                        // 295
          quota: quota                                                                                                 // 295
        }                                                                                                              // 295
      } : {                                                                                                            // 295
        $unset: {                                                                                                      // 296
          quota: ""                                                                                                    // 296
        }                                                                                                              // 296
      };                                                                                                               // 296
      var n = SignupKeys.update({                                                                                      // 297
        email: items[i]                                                                                                // 297
      }, modifier, {                                                                                                   // 297
        multi: true                                                                                                    // 297
      });                                                                                                              // 297
      n += Meteor.users.update({                                                                                       // 298
        signupEmail: items[i]                                                                                          // 298
      }, modifier, {                                                                                                   // 298
        multi: true                                                                                                    // 298
      });                                                                                                              // 298
      if (n < 1) invalid.push(items[i]);                                                                               // 300
    }                                                                                                                  // 301
                                                                                                                       //
    if (invalid.length > 0) {                                                                                          // 303
      throw new Meteor.Error(404, "These addresses did not map to any user nor invite: " + invalid.join(", "));        // 304
    }                                                                                                                  // 306
  },                                                                                                                   // 307
  dismissAdminStatsNotifications: function (token) {                                                                   // 309
    checkAuth(token);                                                                                                  // 310
    globalDb.collections.notifications.remove({                                                                        // 311
      "admin.type": "reportStats"                                                                                      // 311
    });                                                                                                                // 311
  },                                                                                                                   // 312
  signUpAsAdmin: function (token) {                                                                                    // 314
    check(token, String);                                                                                              // 315
    checkAuth(token);                                                                                                  // 316
                                                                                                                       //
    if (!this.userId) {                                                                                                // 317
      throw new Meteor.Error(403, "Must be logged in to sign up as admin.");                                           // 318
    }                                                                                                                  // 319
                                                                                                                       //
    if (!Meteor.user().loginIdentities) {                                                                              // 321
      throw new Meteor.Error(403, "Must be logged into an account to sign up as admin.");                              // 322
    }                                                                                                                  // 323
                                                                                                                       //
    Meteor.users.update({                                                                                              // 325
      _id: this.userId                                                                                                 // 325
    }, {                                                                                                               // 325
      $set: {                                                                                                          // 325
        isAdmin: true,                                                                                                 // 325
        signupKey: "admin"                                                                                             // 325
      }                                                                                                                // 325
    });                                                                                                                // 325
    clearAdminToken(token);                                                                                            // 326
  },                                                                                                                   // 327
  redeemSetupToken: function (token) {                                                                                 // 329
    // Redeem an admin token into a setup session.                                                                     // 330
    check(token, String);                                                                                              // 331
                                                                                                                       //
    if (tokenIsValid(token)) {                                                                                         // 332
      var sessId = Random.secret();                                                                                    // 333
      var creationDate = new Date();                                                                                   // 334
      var hashedSessionId = Crypto.createHash("sha256").update(sessId).digest("base64");                               // 335
      this.connection.sandstormDb.collections.setupSession.upsert({                                                    // 336
        _id: "current-session"                                                                                         // 337
      }, {                                                                                                             // 336
        creationDate: creationDate,                                                                                    // 339
        hashedSessionId: hashedSessionId                                                                               // 340
      }); // Then, invalidate the token, so one one else can use it.                                                   // 338
                                                                                                                       //
      clearAdminToken(token);                                                                                          // 343
      return sessId;                                                                                                   // 344
    } else {                                                                                                           // 345
      throw new Meteor.Error(401, "Invalid setup token");                                                              // 346
    }                                                                                                                  // 347
  },                                                                                                                   // 348
  heapdump: function () {                                                                                              // 350
    // Requests a heap dump. Intended for use by Sandstorm developers. Requires admin.                                 // 351
    //                                                                                                                 // 352
    // Call this from the JS console like:                                                                             // 353
    //   Meteor.call("heapdump");                                                                                      // 354
    checkAuth(); // We use /var/log because it's a location in the container to which the front-end is allowed         // 356
    // to write.                                                                                                       // 359
                                                                                                                       //
    var name = "/var/log/" + Date.now() + ".heapsnapshot";                                                             // 360
    Heapdump.writeSnapshot(name);                                                                                      // 361
    console.log("Wrote heapdump: /opt/sandstorm" + name);                                                              // 362
    return name;                                                                                                       // 363
  },                                                                                                                   // 364
  setPreinstalledApps: function (appAndPackageIds) {                                                                   // 366
    checkAuth();                                                                                                       // 367
    check(appAndPackageIds, [{                                                                                         // 368
      appId: String,                                                                                                   // 368
      packageId: String                                                                                                // 368
    }]);                                                                                                               // 368
    this.connection.sandstormDb.setPreinstalledApps(appAndPackageIds);                                                 // 370
  }                                                                                                                    // 371
});                                                                                                                    // 53
                                                                                                                       //
var authorizedAsAdmin = function (token, userId) {                                                                     // 374
  return Match.test(token, Match.OneOf(undefined, null, String)) && (userId && isAdminById(userId) || tokenIsValid(token) || tokenIsSetupSession(token));
};                                                                                                                     // 377
                                                                                                                       //
Meteor.publish("admin", function (token) {                                                                             // 379
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 380
  return Settings.find();                                                                                              // 381
});                                                                                                                    // 382
Meteor.publish("adminServiceConfiguration", function (token) {                                                         // 384
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 385
  return ServiceConfiguration.configurations.find();                                                                   // 386
});                                                                                                                    // 387
Meteor.publish("publicAdminSettings", function () {                                                                    // 389
  return Settings.find({                                                                                               // 390
    _id: {                                                                                                             // 390
      $in: publicAdminSettings                                                                                         // 390
    }                                                                                                                  // 390
  });                                                                                                                  // 390
});                                                                                                                    // 391
Meteor.publish("adminToken", function (token) {                                                                        // 393
  check(token, String);                                                                                                // 394
  this.added("adminToken", "adminToken", {                                                                             // 395
    tokenIsValid: tokenIsValid(token) || tokenIsSetupSession(token)                                                    // 395
  });                                                                                                                  // 395
  this.ready();                                                                                                        // 396
});                                                                                                                    // 397
Meteor.publish("allUsers", function (token) {                                                                          // 399
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 400
  return Meteor.users.find();                                                                                          // 401
});                                                                                                                    // 402
Meteor.publish("adminUserDetails", function (userId) {                                                                 // 404
  var _this2 = this;                                                                                                   // 404
                                                                                                                       //
  if (!authorizedAsAdmin(undefined, this.userId)) return []; // Reactive publish of any identities owned by the account with id userId,
  // as well as that user object itself.                                                                               // 408
                                                                                                                       //
  var identitySubs = {};                                                                                               // 409
  var accountId = userId;                                                                                              // 410
                                                                                                                       //
  var unrefIdentity = function (identityId) {                                                                          // 412
    if (!identitySubs[identityId]) {                                                                                   // 413
      // should never happen, but if somehow you attempt to unref an identity that we don't have a                     // 414
      // subscription to, then don't crash                                                                             // 415
      console.error("attempted to unref untracked identity id:", identityId);                                          // 416
      return;                                                                                                          // 417
    }                                                                                                                  // 418
                                                                                                                       //
    var observeHandle = identitySubs[identityId];                                                                      // 420
    delete identitySubs[identityId];                                                                                   // 421
    observeHandle.stop();                                                                                              // 422
                                                                                                                       //
    _this2.removed("users", identityId);                                                                               // 423
  };                                                                                                                   // 424
                                                                                                                       //
  var refIdentity = function (identityId) {                                                                            // 426
    if (identitySubs[identityId]) {                                                                                    // 427
      // should never happen, but if somehow an account wound up with a duplicate identity ID,                         // 428
      // avoid leaking a subscription                                                                                  // 429
      console.error("duplicate identity id:", identityId);                                                             // 430
      return;                                                                                                          // 431
    }                                                                                                                  // 432
                                                                                                                       //
    var cursor = Meteor.users.find({                                                                                   // 434
      _id: identityId                                                                                                  // 434
    });                                                                                                                // 434
    var observeHandle = cursor.observe({                                                                               // 435
      added: function (doc) {                                                                                          // 436
        _this2.added("users", doc._id, doc);                                                                           // 437
      },                                                                                                               // 438
      changed: function (newDoc, oldDoc) {                                                                             // 440
        fillUndefinedForChangedDoc(newDoc, oldDoc);                                                                    // 441
                                                                                                                       //
        _this2.changed("users", newDoc._id, newDoc);                                                                   // 442
      },                                                                                                               // 443
      removed: function (oldDoc) {                                                                                     // 445
        _this2.removed("users", oldDoc._id);                                                                           // 446
      }                                                                                                                // 447
    });                                                                                                                // 435
    identitySubs[identityId] = observeHandle;                                                                          // 450
  };                                                                                                                   // 451
                                                                                                                       //
  var accountCursor = Meteor.users.find({                                                                              // 453
    _id: accountId                                                                                                     // 453
  });                                                                                                                  // 453
  var accountSubHandle = accountCursor.observe({                                                                       // 454
    added: function (newDoc) {                                                                                         // 455
      var newIdentities = SandstormDb.getUserIdentityIds(newDoc);                                                      // 456
      newIdentities.forEach(function (identityId) {                                                                    // 457
        refIdentity(identityId);                                                                                       // 458
      });                                                                                                              // 459
                                                                                                                       //
      _this2.added("users", newDoc._id, newDoc);                                                                       // 461
    },                                                                                                                 // 462
    changed: function (newDoc, oldDoc) {                                                                               // 464
      var newIdentities = SandstormDb.getUserIdentityIds(newDoc);                                                      // 465
      var oldIdentities = SandstormDb.getUserIdentityIds(oldDoc); // Those in newDoc - oldDoc, ref.                    // 466
                                                                                                                       //
      var identitiesAdded = _.difference(newIdentities, oldIdentities);                                                // 469
                                                                                                                       //
      identitiesAdded.forEach(function (identityId) {                                                                  // 470
        refIdentity(identityId);                                                                                       // 471
      }); // Those in oldDoc - newDoc, unref.                                                                          // 472
                                                                                                                       //
      var identitiesRemoved = _.difference(oldIdentities, newIdentities);                                              // 475
                                                                                                                       //
      identitiesRemoved.forEach(function (identityId) {                                                                // 476
        unrefIdentity(identityId);                                                                                     // 477
      });                                                                                                              // 478
      fillUndefinedForChangedDoc(newDoc, oldDoc);                                                                      // 480
                                                                                                                       //
      _this2.changed("users", newDoc._id, newDoc);                                                                     // 482
    },                                                                                                                 // 483
    removed: function (oldDoc) {                                                                                       // 485
      _this2.removed("users", oldDoc._id);                                                                             // 486
                                                                                                                       //
      var oldIdentities = SandstormDb.getUserIdentityIds(oldDoc);                                                      // 487
      oldIdentities.forEach(function (identityId) {                                                                    // 488
        unrefIdentity(identityId);                                                                                     // 489
      });                                                                                                              // 490
    }                                                                                                                  // 491
  });                                                                                                                  // 454
  this.onStop(function () {                                                                                            // 494
    accountSubHandle.stop(); // Also stop all the identity subscriptions.                                              // 495
                                                                                                                       //
    var subs = _.values(identitySubs);                                                                                 // 497
                                                                                                                       //
    subs.forEach(function (sub) {                                                                                      // 498
      sub.stop();                                                                                                      // 499
    });                                                                                                                // 500
  }); // Meteor's cursor.observe() will synchronously call all of the added() callbacks from the initial               // 501
  // query, so by the time we get here we can report readiness.                                                        // 504
                                                                                                                       //
  this.ready();                                                                                                        // 505
});                                                                                                                    // 506
Meteor.publish("activityStats", function (token) {                                                                     // 508
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 509
  return ActivityStats.find();                                                                                         // 510
});                                                                                                                    // 511
Meteor.publish("statsTokens", function (token) {                                                                       // 513
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 514
  return StatsTokens.find();                                                                                           // 515
});                                                                                                                    // 516
Meteor.publish("allPackages", function (token) {                                                                       // 518
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 519
  return Packages.find({                                                                                               // 520
    manifest: {                                                                                                        // 520
      $exists: true                                                                                                    // 520
    }                                                                                                                  // 520
  }, {                                                                                                                 // 520
    fields: {                                                                                                          // 521
      appId: 1,                                                                                                        // 521
      "manifest.appVersion": 1,                                                                                        // 521
      "manifest.actions": 1,                                                                                           // 522
      "manifest.appTitle": 1,                                                                                          // 522
      progress: 1,                                                                                                     // 522
      status: 1                                                                                                        // 522
    }                                                                                                                  // 521
  });                                                                                                                  // 521
});                                                                                                                    // 523
Meteor.publish("realTimeStats", function (token) {                                                                     // 525
  if (!authorizedAsAdmin(token, this.userId)) return []; // Last five minutes.                                         // 526
                                                                                                                       //
  this.added("realTimeStats", "now", computeStats(new Date(Date.now() - 5 * 60 * 1000))); // Since last sample.        // 529
                                                                                                                       //
  var lastSample = ActivityStats.findOne({}, {                                                                         // 532
    sort: {                                                                                                            // 532
      timestamp: -1                                                                                                    // 532
    }                                                                                                                  // 532
  });                                                                                                                  // 532
  var lastSampleTime = lastSample ? lastSample.timestamp : new Date(0);                                                // 533
  this.added("realTimeStats", "today", computeStats(lastSampleTime)); // TODO(someday): Update every few minutes?      // 534
                                                                                                                       //
  this.ready();                                                                                                        // 538
});                                                                                                                    // 539
Meteor.publish("adminLog", function (token) {                                                                          // 541
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 542
  var logfile = SANDSTORM_LOGDIR + "/sandstorm.log";                                                                   // 544
  var fd = Fs.openSync(logfile, "r");                                                                                  // 546
  var startSize = Fs.fstatSync(fd).size; // Difference between the current file offset and the subscription offset. Can be non-zero when
  // logs have rotated.                                                                                                // 550
                                                                                                                       //
  var extraOffset = 0;                                                                                                 // 551
                                                                                                                       //
  if (startSize < 8192) {                                                                                              // 553
    // Log size is less than window size. Check for rotated log and grab its tail.                                     // 554
    var logfile1 = SANDSTORM_LOGDIR + "/sandstorm.log.1";                                                              // 555
                                                                                                                       //
    if (Fs.existsSync(logfile1)) {                                                                                     // 556
      var fd1 = Fs.openSync(logfile1, "r");                                                                            // 557
      var startSize1 = Fs.fstatSync(fd1).size;                                                                         // 558
      var amountFromLog1 = Math.min(startSize1, 8192 - startSize);                                                     // 559
      var offset1 = startSize1 - amountFromLog1;                                                                       // 560
      var buf = new Buffer(amountFromLog1);                                                                            // 561
      var n = Fs.readSync(fd1, buf, 0, buf.length, offset);                                                            // 562
                                                                                                                       //
      if (n > 0) {                                                                                                     // 563
        this.added("adminLog", 0, {                                                                                    // 564
          text: buf.toString("utf8", 0, n)                                                                             // 564
        });                                                                                                            // 564
        extraOffset += n;                                                                                              // 565
      }                                                                                                                // 566
    }                                                                                                                  // 567
  } // Start tailing at EOF - 8k.                                                                                      // 568
                                                                                                                       //
                                                                                                                       //
  var offset = Math.max(0, startSize - 8192);                                                                          // 571
                                                                                                                       //
  var _this = this;                                                                                                    // 573
                                                                                                                       //
  function doTail() {                                                                                                  // 574
    if (Fs.fstatSync(fd).size < offset) {                                                                              // 575
      extraOffset += offset;                                                                                           // 576
      offset = 0;                                                                                                      // 577
    }                                                                                                                  // 578
                                                                                                                       //
    for (;;) {                                                                                                         // 580
      var _buf = new Buffer(Math.max(1024, startSize - offset));                                                       // 581
                                                                                                                       //
      var _n = Fs.readSync(fd, _buf, 0, _buf.length, offset);                                                          // 582
                                                                                                                       //
      if (_n <= 0) break;                                                                                              // 583
                                                                                                                       //
      _this.added("adminLog", offset + extraOffset, {                                                                  // 584
        text: _buf.toString("utf8", 0, _n)                                                                             // 584
      });                                                                                                              // 584
                                                                                                                       //
      offset += _n;                                                                                                    // 585
    }                                                                                                                  // 586
  } // Watch the file for changes.                                                                                     // 587
                                                                                                                       //
                                                                                                                       //
  var watcher = Fs.watch(logfile, {                                                                                    // 590
    persistent: false                                                                                                  // 590
  }, Meteor.bindEnvironment(doTail)); // When the subscription stops, stop watching the file.                          // 590
                                                                                                                       //
  this.onStop(function () {                                                                                            // 593
    watcher.close();                                                                                                   // 594
    Fs.closeSync(fd);                                                                                                  // 595
  }); // Read initial 8k tail data immediately.                                                                        // 596
                                                                                                                       //
  doTail(); // Notify ready.                                                                                           // 599
                                                                                                                       //
  this.ready();                                                                                                        // 602
});                                                                                                                    // 603
Meteor.publish("adminApiTokens", function (token) {                                                                    // 605
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 606
  return ApiTokens.find({                                                                                              // 607
    $or: [{                                                                                                            // 608
      "frontendRef.ipNetwork": {                                                                                       // 609
        $exists: true                                                                                                  // 609
      }                                                                                                                // 609
    }, {                                                                                                               // 609
      "frontendRef.ipInterface": {                                                                                     // 610
        $exists: true                                                                                                  // 610
      }                                                                                                                // 610
    }]                                                                                                                 // 610
  }, {                                                                                                                 // 607
    fields: {                                                                                                          // 613
      frontendRef: 1,                                                                                                  // 614
      created: 1,                                                                                                      // 615
      requirements: 1,                                                                                                 // 616
      revoked: 1,                                                                                                      // 617
      owner: 1                                                                                                         // 618
    }                                                                                                                  // 613
  });                                                                                                                  // 612
});                                                                                                                    // 621
Meteor.publish("hasAdmin", function (token) {                                                                          // 623
  var _this3 = this;                                                                                                   // 623
                                                                                                                       //
  // Like hasUsers, but for admins, and with token auth required.                                                      // 624
  if (!authorizedAsAdmin(token, this.userId)) return []; // Query if there are any admin users.                        // 625
                                                                                                                       //
  var cursor = Meteor.users.find({                                                                                     // 628
    isAdmin: true                                                                                                      // 628
  });                                                                                                                  // 628
                                                                                                                       //
  if (cursor.count() > 0) {                                                                                            // 629
    this.added("hasAdmin", "hasAdmin", {                                                                               // 630
      hasAdmin: true                                                                                                   // 630
    });                                                                                                                // 630
  } else {                                                                                                             // 631
    var handle = cursor.observeChanges({                                                                               // 632
      added: function (id) {                                                                                           // 633
        _this3.added("hasAdmin", "hasAdmin", {                                                                         // 634
          hasAdmin: true                                                                                               // 634
        });                                                                                                            // 634
                                                                                                                       //
        handle.stop();                                                                                                 // 635
        handle = null;                                                                                                 // 636
      }                                                                                                                // 637
    });                                                                                                                // 632
    this.onStop(function () {                                                                                          // 639
      if (handle) handle.stop();                                                                                       // 640
    });                                                                                                                // 641
  }                                                                                                                    // 642
                                                                                                                       //
  this.ready();                                                                                                        // 644
});                                                                                                                    // 645
Meteor.publish("appIndexAdmin", function (token) {                                                                     // 647
  if (!authorizedAsAdmin(token, this.userId)) return [];                                                               // 648
  return globalDb.collections.appIndex.find();                                                                         // 649
});                                                                                                                    // 650
                                                                                                                       //
function observeOauthService(name) {                                                                                   // 652
  Settings.find({                                                                                                      // 653
    _id: name,                                                                                                         // 653
    value: true                                                                                                        // 653
  }).observe({                                                                                                         // 653
    added: function () {                                                                                               // 654
      // Tell the oauth library it should accept login attempts from this service.                                     // 655
      Accounts.oauth.registerService(name);                                                                            // 656
    },                                                                                                                 // 657
    removed: function () {                                                                                             // 659
      // Tell the oauth library it should deny login attempts from this service.                                       // 660
      Accounts.oauth.unregisterService(name);                                                                          // 661
    }                                                                                                                  // 662
  });                                                                                                                  // 653
}                                                                                                                      // 664
                                                                                                                       //
observeOauthService("github");                                                                                         // 666
observeOauthService("google");                                                                                         // 667
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"backup.js":["/imports/server/async-helpers.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/backup.js                                                                                                    //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var ChildProcess = Npm.require("child_process");                                                                       // 19
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 20
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 21
                                                                                                                       //
var GrainInfo = Capnp.importSystem("sandstorm/grain.capnp").GrainInfo;                                                 // 23
var TOKEN_CLEANUP_MINUTES = 120; // Give enough time for large uploads on slow connections.                            // 25
                                                                                                                       //
var TOKEN_CLEANUP_TIMER = TOKEN_CLEANUP_MINUTES * 60 * 1000;                                                           // 26
                                                                                                                       //
function cleanupToken(tokenId) {                                                                                       // 28
  check(tokenId, String);                                                                                              // 29
  FileTokens.remove({                                                                                                  // 30
    _id: tokenId                                                                                                       // 30
  });                                                                                                                  // 30
  waitPromise(globalBackend.cap().deleteBackup(tokenId));                                                              // 31
}                                                                                                                      // 32
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 34
  // Cleanup tokens every TOKEN_CLEANUP_MINUTES                                                                        // 35
  SandstormDb.periodicCleanup(TOKEN_CLEANUP_TIMER, function () {                                                       // 36
    var queryDate = new Date(Date.now() - TOKEN_CLEANUP_TIMER);                                                        // 37
    FileTokens.find({                                                                                                  // 39
      timestamp: {                                                                                                     // 39
        $lt: queryDate                                                                                                 // 39
      }                                                                                                                // 39
    }).forEach(function (token) {                                                                                      // 39
      cleanupToken(token._id);                                                                                         // 40
    });                                                                                                                // 41
  });                                                                                                                  // 42
});                                                                                                                    // 43
Meteor.methods({                                                                                                       // 45
  backupGrain: function (grainId) {                                                                                    // 46
    check(grainId, String);                                                                                            // 47
    var grain = Grains.findOne(grainId);                                                                               // 48
                                                                                                                       //
    if (!grain || !this.userId || grain.userId !== this.userId) {                                                      // 49
      throw new Meteor.Error(403, "Unauthorized", "User is not the owner of this grain");                              // 50
    }                                                                                                                  // 51
                                                                                                                       //
    this.unblock();                                                                                                    // 53
    var token = {                                                                                                      // 55
      _id: Random.id(),                                                                                                // 56
      timestamp: new Date(),                                                                                           // 57
      name: grain.title                                                                                                // 58
    }; // TODO(soon): does the grain need to be offline?                                                               // 55
                                                                                                                       //
    var grainInfo = _.pick(grain, "appId", "appVersion", "title");                                                     // 63
                                                                                                                       //
    FileTokens.insert(token);                                                                                          // 65
    waitPromise(globalBackend.cap().backupGrain(token._id, this.userId, grainId, grainInfo));                          // 66
    return token._id;                                                                                                  // 68
  },                                                                                                                   // 69
  newRestoreToken: function () {                                                                                       // 71
    if (!isSignedUpOrDemo()) {                                                                                         // 72
      throw new Meteor.Error(403, "Unauthorized", "Only invited users can restore backups.");                          // 73
    }                                                                                                                  // 74
                                                                                                                       //
    if (isUserOverQuota(Meteor.user())) {                                                                              // 76
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");           // 77
    }                                                                                                                  // 79
                                                                                                                       //
    var token = {                                                                                                      // 81
      _id: Random.id(),                                                                                                // 82
      timestamp: new Date()                                                                                            // 83
    };                                                                                                                 // 81
    FileTokens.insert(token);                                                                                          // 86
    return token._id;                                                                                                  // 88
  },                                                                                                                   // 89
  restoreGrain: function (tokenId, identityId) {                                                                       // 91
    check(tokenId, String);                                                                                            // 92
    check(identityId, String);                                                                                         // 93
    var token = FileTokens.findOne(tokenId);                                                                           // 94
                                                                                                                       //
    if (!token || !isSignedUpOrDemo()) {                                                                               // 95
      throw new Meteor.Error(403, "Unauthorized", "Token was not found, or user cannot create grains");                // 96
    }                                                                                                                  // 98
                                                                                                                       //
    if (isUserOverQuota(Meteor.user())) {                                                                              // 100
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");           // 101
    }                                                                                                                  // 103
                                                                                                                       //
    this.unblock();                                                                                                    // 105
    var grainId = Random.id(22);                                                                                       // 107
                                                                                                                       //
    try {                                                                                                              // 109
      var grainInfo = waitPromise(globalBackend.cap().restoreGrain(tokenId, this.userId, grainId).catch(function (err) {
        console.error("Unzip failure:", err.message);                                                                  // 112
        throw new Meteor.Error(500, "Invalid backup file.");                                                           // 113
      })).info;                                                                                                        // 114
                                                                                                                       //
      if (!grainInfo.appId) {                                                                                          // 115
        globalBackend.deleteGrain(grainId, this.userId);                                                               // 116
        throw new Meteor.Error(500, "Metadata object for uploaded grain has no AppId");                                // 117
      }                                                                                                                // 118
                                                                                                                       //
      var action = UserActions.findOne({                                                                               // 120
        appId: grainInfo.appId,                                                                                        // 120
        userId: this.userId                                                                                            // 120
      }); // Create variables we'll use for later Mongo query.                                                         // 120
                                                                                                                       //
      var packageId = void 0;                                                                                          // 123
      var appVersion = void 0; // DevPackages are system-wide, so we do not check the user ID.                         // 124
                                                                                                                       //
      var devPackage = DevPackages.findOne({                                                                           // 127
        appId: grainInfo.appId                                                                                         // 127
      });                                                                                                              // 127
                                                                                                                       //
      if (devPackage) {                                                                                                // 128
        // If the dev app package exists, it should override the user action.                                          // 129
        packageId = devPackage.packageId;                                                                              // 130
        appVersion = devPackage.manifest.appVersion;                                                                   // 131
      } else if (action) {                                                                                             // 132
        // The app is installed, so we can continue restoring this                                                     // 133
        // grain.                                                                                                      // 134
        packageId = action.packageId;                                                                                  // 135
        appVersion = action.appVersion;                                                                                // 136
      } else {                                                                                                         // 137
        // If the package isn't installed at all, bail out.                                                            // 138
        globalBackend.deleteGrain(grainId, this.userId);                                                               // 139
        throw new Meteor.Error(500, "App id for uploaded grain not installed", "App Id: " + grainInfo.appId);          // 140
      }                                                                                                                // 143
                                                                                                                       //
      if (appVersion < grainInfo.appVersion) {                                                                         // 145
        globalBackend.deleteGrain(grainId, this.userId);                                                               // 146
        throw new Meteor.Error(500, "App version for uploaded grain is newer than any " + "installed version. You need to upgrade your app first", "New version: " + grainInfo.appVersion + ", Old version: " + appVersion);
      }                                                                                                                // 152
                                                                                                                       //
      Grains.insert({                                                                                                  // 154
        _id: grainId,                                                                                                  // 155
        packageId: packageId,                                                                                          // 156
        appId: grainInfo.appId,                                                                                        // 157
        appVersion: appVersion,                                                                                        // 158
        userId: this.userId,                                                                                           // 159
        identityId: identityId,                                                                                        // 160
        title: grainInfo.title,                                                                                        // 161
        "private": true,                                                                                               // 162
        size: 0                                                                                                        // 163
      });                                                                                                              // 154
    } finally {                                                                                                        // 165
      cleanupToken(tokenId);                                                                                           // 166
    }                                                                                                                  // 167
                                                                                                                       //
    return grainId;                                                                                                    // 169
  }                                                                                                                    // 170
});                                                                                                                    // 45
Router.map(function () {                                                                                               // 173
  this.route("downloadBackup", {                                                                                       // 174
    where: "server",                                                                                                   // 175
    path: "/downloadBackup/:tokenId",                                                                                  // 176
    action: function () {                                                                                              // 177
      var token = FileTokens.findOne(this.params.tokenId);                                                             // 178
      var response = this.response;                                                                                    // 179
                                                                                                                       //
      if (!token) {                                                                                                    // 180
        response.writeHead(404, {                                                                                      // 181
          "Content-Type": "text/plain"                                                                                 // 181
        });                                                                                                            // 181
        return response.end("File does not exist");                                                                    // 182
      }                                                                                                                // 183
                                                                                                                       //
      var started = false;                                                                                             // 185
      var encodedFilename = encodeURIComponent(token.name || "backup") + ".zip";                                       // 186
      var sawEnd = false;                                                                                              // 187
      var stream = {                                                                                                   // 189
        expectSize: function (size) {                                                                                  // 190
          if (!started) {                                                                                              // 191
            started = true;                                                                                            // 192
            response.writeHead(200, {                                                                                  // 193
              "Content-Length": size,                                                                                  // 194
              "Content-Type": "application/zip",                                                                       // 195
              "Content-Disposition": "attachment;filename*=utf-8''" + encodedFilename                                  // 196
            });                                                                                                        // 193
          }                                                                                                            // 198
        },                                                                                                             // 199
        write: function (data) {                                                                                       // 201
          if (!started) {                                                                                              // 202
            started = true;                                                                                            // 203
            response.writeHead(200, {                                                                                  // 204
              "Content-Type": "application/zip",                                                                       // 205
              "Content-Disposition": "attachment;filename*=utf-8''" + encodedFilename                                  // 206
            });                                                                                                        // 204
          }                                                                                                            // 208
                                                                                                                       //
          response.write(data);                                                                                        // 210
        },                                                                                                             // 211
        done: function (data) {                                                                                        // 213
          if (!started) {                                                                                              // 214
            started = true;                                                                                            // 215
            response.writeHead(200, {                                                                                  // 216
              "Content-Length": 0,                                                                                     // 217
              "Content-Type": "application/zip",                                                                       // 218
              "Content-Disposition": "attachment;filename*=utf-8''" + encodedFilename                                  // 219
            });                                                                                                        // 216
          }                                                                                                            // 221
                                                                                                                       //
          sawEnd = true;                                                                                               // 223
          response.end();                                                                                              // 224
        }                                                                                                              // 225
      };                                                                                                               // 189
      waitPromise(globalBackend.cap().downloadBackup(this.params.tokenId, stream));                                    // 228
                                                                                                                       //
      if (!sawEnd) {                                                                                                   // 230
        console.error("backend failed to call done() when downloading backup");                                        // 231
                                                                                                                       //
        if (!started) {                                                                                                // 232
          throw new Meteor.Error(500, "backend failed to produce data");                                               // 233
        }                                                                                                              // 234
                                                                                                                       //
        response.end();                                                                                                // 236
      }                                                                                                                // 237
                                                                                                                       //
      cleanupToken(this.params.tokenId);                                                                               // 239
    }                                                                                                                  // 240
  });                                                                                                                  // 174
  this.route("uploadBackup", {                                                                                         // 243
    where: "server",                                                                                                   // 244
    path: "/uploadBackup/:token",                                                                                      // 245
    action: function () {                                                                                              // 246
      if (this.request.method === "POST") {                                                                            // 247
        var token = FileTokens.findOne(this.params.token);                                                             // 248
                                                                                                                       //
        if (!this.params.token || !token) {                                                                            // 249
          this.response.writeHead(403, {                                                                               // 250
            "Content-Type": "text/plain",                                                                              // 251
            "Access-Control-Allow-Origin": "*"                                                                         // 252
          });                                                                                                          // 250
          this.response.write("Invalid upload token.");                                                                // 254
          this.response.end();                                                                                         // 255
          return;                                                                                                      // 256
        }                                                                                                              // 257
                                                                                                                       //
        var request = this.request;                                                                                    // 259
                                                                                                                       //
        try {                                                                                                          // 260
          var stream = globalBackend.cap().uploadBackup(this.params.token).stream;                                     // 261
          waitPromise(new Promise(function (resolve, reject) {                                                         // 263
            request.on("data", function (data) {                                                                       // 264
              stream.write(data);                                                                                      // 265
            });                                                                                                        // 266
            request.on("end", function () {                                                                            // 267
              resolve(stream.done());                                                                                  // 268
            });                                                                                                        // 269
            request.on("error", function (err) {                                                                       // 270
              stream.close();                                                                                          // 271
            });                                                                                                        // 272
          }));                                                                                                         // 273
          this.response.writeHead(204, {                                                                               // 275
            "Access-Control-Allow-Origin": "*"                                                                         // 276
          });                                                                                                          // 275
          this.response.end();                                                                                         // 278
        } catch (error) {                                                                                              // 279
          console.error(error.stack);                                                                                  // 280
          this.response.writeHead(500, {                                                                               // 281
            "Content-Type": "text/plain",                                                                              // 282
            "Access-Control-Allow-Origin": "*"                                                                         // 283
          });                                                                                                          // 281
          this.response.write(error.stack);                                                                            // 285
          this.response.end();                                                                                         // 286
        }                                                                                                              // 287
      } else if (this.request.method == "OPTIONS") {                                                                   // 288
        // Allow cross-origin posts to uploadBackup so that uploads can occur on the DDP host                          // 289
        // rather than the main host. In theory we could have Access-Control-Allow-Origin specify                      // 290
        // the main host rather than "*", but an uploadBackup request already requires a valid                         // 291
        // upload token, which is plenty of access control in itself.                                                  // 292
        var requestedHeaders = this.request.headers["access-control-request-headers"];                                 // 293
                                                                                                                       //
        if (requestedHeaders) {                                                                                        // 294
          this.response.setHeader("Access-Control-Allow-Headers", requestedHeaders);                                   // 295
        }                                                                                                              // 296
                                                                                                                       //
        this.response.writeHead(204, {                                                                                 // 298
          "Access-Control-Allow-Origin": "*",                                                                          // 299
          "Access-Control-Allow-Methods": "POST, OPTIONS",                                                             // 300
          "Access-Control-Max-Age": "3600"                                                                             // 301
        });                                                                                                            // 298
        this.response.end();                                                                                           // 303
      } else {                                                                                                         // 304
        this.response.writeHead(405, {                                                                                 // 305
          "Content-Type": "text/plain"                                                                                 // 306
        });                                                                                                            // 305
        this.response.write("You can only POST here.");                                                                // 308
        this.response.end();                                                                                           // 309
      }                                                                                                                // 310
    }                                                                                                                  // 311
  });                                                                                                                  // 243
});                                                                                                                    // 313
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"contacts-server.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/contacts-server.js                                                                                           //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Meteor.publish("contactProfiles", function (showAll) {                                                                 // 17
  var db = this.connection.sandstormDb;                                                                                // 18
                                                                                                                       //
  var _this = this;                                                                                                    // 19
                                                                                                                       //
  var userId = this.userId; // We maintain a map from identity IDs to live query handles that track profile changes.   // 20
                                                                                                                       //
  var contactIdentities = {};                                                                                          // 23
  var disallowGuests = db.getOrganizationDisallowGuests();                                                             // 24
                                                                                                                       //
  function addIdentityOfContact(identityId) {                                                                          // 26
    if (!(identityId in contactIdentities)) {                                                                          // 27
      var user = Meteor.users.findOne({                                                                                // 28
        _id: identityId                                                                                                // 28
      });                                                                                                              // 28
                                                                                                                       //
      if (disallowGuests && !showAll) {                                                                                // 30
        if (!db.isIdentityInOrganization(user)) {                                                                      // 31
          return;                                                                                                      // 32
        }                                                                                                              // 33
      }                                                                                                                // 34
                                                                                                                       //
      if (user) {                                                                                                      // 36
        SandstormDb.fillInProfileDefaults(user);                                                                       // 37
        SandstormDb.fillInIntrinsicName(user);                                                                         // 38
                                                                                                                       //
        var filteredUser = _.pick(user, "_id", "profile");                                                             // 39
                                                                                                                       //
        _this.added("contactProfiles", user._id, filteredUser);                                                        // 40
      }                                                                                                                // 41
                                                                                                                       //
      contactIdentities[identityId] = Meteor.users.find({                                                              // 43
        _id: identityId                                                                                                // 44
      }, {                                                                                                             // 44
        fields: {                                                                                                      // 44
          profile: 1                                                                                                   // 44
        }                                                                                                              // 44
      }).observeChanges({                                                                                              // 44
        changed: function (id, fields) {                                                                               // 45
          _this.changed("contactProfiles", id, fields);                                                                // 46
        }                                                                                                              // 47
      });                                                                                                              // 44
    }                                                                                                                  // 49
  }                                                                                                                    // 50
                                                                                                                       //
  var cursor = db.collections.contacts.find({                                                                          // 52
    ownerId: userId                                                                                                    // 52
  });                                                                                                                  // 52
  var handle = cursor.observe({                                                                                        // 54
    added: function (contact) {                                                                                        // 55
      addIdentityOfContact(contact.identityId);                                                                        // 56
    },                                                                                                                 // 57
    changed: function (contact) {                                                                                      // 59
      addIdentityOfContact(contact.identityId);                                                                        // 60
    },                                                                                                                 // 61
    removed: function (contact) {                                                                                      // 63
      _this.removed("contactProfiles", contact.identityId);                                                            // 64
                                                                                                                       //
      var contactIdentity = contactIdentities[contact.identityId];                                                     // 65
      if (contactIdentity) contactIdentities[contact.identityId].stop();                                               // 66
      delete contactIdentities[contact.identityId];                                                                    // 67
    }                                                                                                                  // 68
  });                                                                                                                  // 54
  var orgHandle = void 0;                                                                                              // 71
                                                                                                                       //
  if (db.getOrganizationShareContacts() && db.isUserInOrganization(db.collections.users.findOne({                      // 73
    _id: userId                                                                                                        // 74
  }))) {                                                                                                               // 74
    var orgCursor = db.collections.users.find({                                                                        // 75
      profile: {                                                                                                       // 75
        $exists: 1                                                                                                     // 75
      }                                                                                                                // 75
    }); // TODO(perf): make a mongo query that can find all identities in an organization and add                      // 75
    // indices for it. Currently, we do some case insensitive matching which mongo can't                               // 77
    // handle well.                                                                                                    // 78
                                                                                                                       //
    orgHandle = orgCursor.observe({                                                                                    // 80
      added: function (user) {                                                                                         // 81
        if (db.isIdentityInOrganization(user) && !db.userHasIdentity(userId, user._id)) {                              // 82
          addIdentityOfContact(user._id);                                                                              // 83
        }                                                                                                              // 84
      },                                                                                                               // 85
      changed: function (user) {                                                                                       // 87
        if (db.isIdentityInOrganization(user) && !db.userHasIdentity(userId, user._id)) {                              // 88
          addIdentityOfContact(user._id);                                                                              // 89
        }                                                                                                              // 90
      },                                                                                                               // 91
      removed: function (user) {                                                                                       // 93
        if (db.isIdentityInOrganization(user) && !db.userHasIdentity(userId, user._id)) {                              // 94
          _this.removed("contactProfiles", user._id);                                                                  // 95
                                                                                                                       //
          var contactIdentity = contactIdentities[contact.identityId];                                                 // 96
          if (contactIdentity) contactIdentities[contact.identityId].stop();                                           // 97
          delete contactIdentities[user._id];                                                                          // 98
        }                                                                                                              // 99
      }                                                                                                                // 100
    });                                                                                                                // 80
  }                                                                                                                    // 102
                                                                                                                       //
  this.ready();                                                                                                        // 104
  this.onStop(function () {                                                                                            // 106
    handle.stop();                                                                                                     // 107
                                                                                                                       //
    if (orgHandle) {                                                                                                   // 108
      orgHandle.stop();                                                                                                // 109
    }                                                                                                                  // 110
                                                                                                                       //
    Object.keys(contactIdentities).forEach(function (identityId) {                                                     // 112
      contactIdentities[identityId].stop();                                                                            // 113
      delete contactIdentities[identityId];                                                                            // 114
    });                                                                                                                // 115
  });                                                                                                                  // 116
});                                                                                                                    // 117
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"core.js":["babel-runtime/helpers/possibleConstructorReturn","babel-runtime/helpers/inherits","babel-runtime/helpers/classCallCheck","/imports/server/async-helpers.js","/imports/server/static-asset.js","/imports/server/persistent.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/core.js                                                                                                      //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var inMeteor = void 0,                                                                                                 // 1
    waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var StaticAssetImpl = void 0,                                                                                          // 1
    IdenticonStaticAssetImpl = void 0;                                                                                 // 1
module.import("/imports/server/static-asset.js", {                                                                     // 1
  "StaticAssetImpl": function (v) {                                                                                    // 1
    StaticAssetImpl = v;                                                                                               // 1
  },                                                                                                                   // 1
  "IdenticonStaticAssetImpl": function (v) {                                                                           // 1
    IdenticonStaticAssetImpl = v;                                                                                      // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var PersistentImpl = void 0,                                                                                           // 1
    hashSturdyRef = void 0,                                                                                            // 1
    generateSturdyRef = void 0,                                                                                        // 1
    checkRequirements = void 0,                                                                                        // 1
    fetchApiToken = void 0,                                                                                            // 1
    insertApiToken = void 0;                                                                                           // 1
module.import("/imports/server/persistent.js", {                                                                       // 1
  "PersistentImpl": function (v) {                                                                                     // 1
    PersistentImpl = v;                                                                                                // 1
  },                                                                                                                   // 1
  "hashSturdyRef": function (v) {                                                                                      // 1
    hashSturdyRef = v;                                                                                                 // 1
  },                                                                                                                   // 1
  "generateSturdyRef": function (v) {                                                                                  // 1
    generateSturdyRef = v;                                                                                             // 1
  },                                                                                                                   // 1
  "checkRequirements": function (v) {                                                                                  // 1
    checkRequirements = v;                                                                                             // 1
  },                                                                                                                   // 1
  "fetchApiToken": function (v) {                                                                                      // 1
    fetchApiToken = v;                                                                                                 // 1
  },                                                                                                                   // 1
  "insertApiToken": function (v) {                                                                                     // 1
    insertApiToken = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 19
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 20
                                                                                                                       //
var PersistentHandle = Capnp.importSystem("sandstorm/supervisor.capnp").PersistentHandle;                              // 24
var SandstormCore = Capnp.importSystem("sandstorm/supervisor.capnp").SandstormCore;                                    // 25
var SandstormCoreFactory = Capnp.importSystem("sandstorm/backend.capnp").SandstormCoreFactory;                         // 26
var PersistentOngoingNotification = Capnp.importSystem("sandstorm/supervisor.capnp").PersistentOngoingNotification;    // 27
var PersistentUiView = Capnp.importSystem("sandstorm/persistentuiview.capnp").PersistentUiView;                        // 28
var StaticAsset = Capnp.importSystem("sandstorm/util.capnp").StaticAsset;                                              // 29
var SystemPersistent = Capnp.importSystem("sandstorm/supervisor.capnp").SystemPersistent;                              // 30
                                                                                                                       //
var SandstormCoreImpl = function () {                                                                                  //
  function SandstormCoreImpl(db, grainId) {                                                                            // 33
    (0, _classCallCheck3.default)(this, SandstormCoreImpl);                                                            // 33
    this.db = db;                                                                                                      // 34
    this.grainId = grainId;                                                                                            // 35
  }                                                                                                                    // 36
                                                                                                                       //
  SandstormCoreImpl.prototype.restore = function () {                                                                  //
    function restore(sturdyRef) {                                                                                      //
      var _this = this;                                                                                                // 38
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 39
        sturdyRef = sturdyRef.toString("utf8");                                                                        // 40
        var token = fetchApiToken(_this.db, sturdyRef, {                                                               // 41
          "owner.grain.grainId": _this.grainId                                                                         // 42
        });                                                                                                            // 42
                                                                                                                       //
        if (!token) {                                                                                                  // 43
          throw new Error("no such token");                                                                            // 44
        }                                                                                                              // 45
                                                                                                                       //
        if (token.owner.grain.introducerIdentity) {                                                                    // 47
          throw new Error("Cannot restore grain-owned sturdyref that contains the obsolete " + "introducerIdentity field. Please request a new capability.");
        }                                                                                                              // 50
                                                                                                                       //
        return restoreInternal(_this.db, sturdyRef, {                                                                  // 52
          grain: Match.ObjectIncluding({                                                                               // 53
            grainId: _this.grainId                                                                                     // 53
          })                                                                                                           // 53
        }, [], token);                                                                                                 // 53
      });                                                                                                              // 55
    }                                                                                                                  // 56
                                                                                                                       //
    return restore;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.drop = function () {                                                                     //
    function drop(sturdyRef) {                                                                                         //
      var _this2 = this;                                                                                               // 58
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 59
        sturdyRef = sturdyRef.toString("utf8");                                                                        // 60
        return dropInternal(_this2.db, sturdyRef, {                                                                    // 61
          grain: Match.ObjectIncluding({                                                                               // 61
            grainId: _this2.grainId                                                                                    // 61
          })                                                                                                           // 61
        });                                                                                                            // 61
      });                                                                                                              // 62
    }                                                                                                                  // 63
                                                                                                                       //
    return drop;                                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.makeToken = function () {                                                                //
    function makeToken(ref, owner, requirements) {                                                                     //
      var _this3 = this;                                                                                               // 65
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 66
        var sturdyRef = insertApiToken(_this3.db, {                                                                    // 67
          grainId: _this3.grainId,                                                                                     // 68
          objectId: ref,                                                                                               // 69
          owner: owner,                                                                                                // 70
          created: new Date(),                                                                                         // 71
          requirements: requirements                                                                                   // 72
        });                                                                                                            // 67
        return {                                                                                                       // 75
          token: new Buffer(sturdyRef)                                                                                 // 76
        };                                                                                                             // 75
      });                                                                                                              // 78
    }                                                                                                                  // 79
                                                                                                                       //
    return makeToken;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.makeChildToken = function () {                                                           //
    function makeChildToken(parent, owner, requirements) {                                                             //
      var _this4 = this;                                                                                               // 81
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 82
        // Compute the save ApiToken template.                                                                         // 83
        return makeSaveTemplateForChild(_this4.db, parent.toString(), requirements);                                   // 84
      }).then(function (saveTemplate) {                                                                                // 85
        // Create a dummy PersistentImpl and invoke its own save() method.                                             // 86
        return new PersistentImpl(_this4.db, saveTemplate).save({                                                      // 87
          sealFor: owner                                                                                               // 87
        });                                                                                                            // 87
      }).then(function (saveResult) {                                                                                  // 88
        // Transform to expected result structure.                                                                     // 89
        return {                                                                                                       // 90
          token: saveResult.sturdyRef                                                                                  // 90
        };                                                                                                             // 90
      });                                                                                                              // 91
    }                                                                                                                  // 92
                                                                                                                       //
    return makeChildToken;                                                                                             //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.getOwnerNotificationTarget = function () {                                               //
    function getOwnerNotificationTarget() {                                                                            //
      var _this5 = this;                                                                                               // 94
                                                                                                                       //
      var grainId = this.grainId;                                                                                      // 95
      return {                                                                                                         // 96
        owner: {                                                                                                       // 97
          addOngoing: function (displayInfo, notification) {                                                           // 98
            return inMeteor(function () {                                                                              // 99
              var grain = _this5.db.collections.grains.findOne({                                                       // 100
                _id: grainId                                                                                           // 100
              });                                                                                                      // 100
                                                                                                                       //
              if (!grain) {                                                                                            // 101
                throw new Error("Grain not found.");                                                                   // 102
              }                                                                                                        // 103
                                                                                                                       //
              var castedNotification = notification.castAs(PersistentOngoingNotification);                             // 105
              var wakelockToken = waitPromise(castedNotification.save()).sturdyRef.toString("utf8"); // We have to close both the casted cap and the original. Perhaps this should be fixed in
              // node-capnp?                                                                                           // 109
                                                                                                                       //
              castedNotification.close();                                                                              // 110
              notification.close();                                                                                    // 111
                                                                                                                       //
              var notificationId = _this5.db.collections.notifications.insert({                                        // 112
                ongoing: wakelockToken,                                                                                // 113
                grainId: grainId,                                                                                      // 114
                userId: grain.userId,                                                                                  // 115
                text: displayInfo.caption,                                                                             // 116
                timestamp: new Date(),                                                                                 // 117
                isUnread: true                                                                                         // 118
              });                                                                                                      // 112
                                                                                                                       //
              return {                                                                                                 // 121
                handle: globalFrontendRefRegistry.create(_this5.db, {                                                  // 122
                  notificationHandle: notificationId                                                                   // 123
                })                                                                                                     // 123
              };                                                                                                       // 121
            });                                                                                                        // 125
          }                                                                                                            // 126
        }                                                                                                              // 97
      };                                                                                                               // 96
    }                                                                                                                  // 129
                                                                                                                       //
    return getOwnerNotificationTarget;                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.backgroundActivity = function () {                                                       //
    function backgroundActivity(event) {                                                                               //
      var _this6 = this;                                                                                               // 131
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 132
        logActivity(_this6.grainId, null, event);                                                                      // 133
      });                                                                                                              // 134
    }                                                                                                                  // 135
                                                                                                                       //
    return backgroundActivity;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.reportGrainSize = function () {                                                          //
    function reportGrainSize(bytes) {                                                                                  //
      bytes = parseInt(bytes); // int64s are stringified but precision isn't critical here                             // 138
                                                                                                                       //
      var result = this.db.collections.grains.findAndModify({                                                          // 140
        query: {                                                                                                       // 141
          _id: this.grainId                                                                                            // 141
        },                                                                                                             // 141
        update: {                                                                                                      // 142
          $set: {                                                                                                      // 142
            size: bytes                                                                                                // 142
          }                                                                                                            // 142
        },                                                                                                             // 142
        fields: {                                                                                                      // 143
          _id: 1,                                                                                                      // 143
          userId: 1,                                                                                                   // 143
          size: 1                                                                                                      // 143
        }                                                                                                              // 143
      });                                                                                                              // 140
                                                                                                                       //
      if (!result.ok) {                                                                                                // 146
        throw new Error("Grain not found.");                                                                           // 147
      } // If the grain did not have a "size" field before the update, then it is a grain created                      // 148
      // before per-grain size tracking was implemented. In that case, we don't want to update the                     // 151
      // user record because it may already be counting the grain (specifically on Blackrock, where                    // 152
      // whole-user size counting has existed for some time).                                                          // 153
                                                                                                                       //
                                                                                                                       //
      if (this.db.isQuotaEnabled() && "size" in result.value) {                                                        // 154
        // Update the user record, too. Note that we periodically recompute the user's storage usage                   // 155
        // from scratch as well, so this doesn't have to be perfectly reliable.                                        // 156
        var diff = bytes - (result.value.size || 0);                                                                   // 157
        this.db.collections.users.update(result.value.userId, {                                                        // 158
          $inc: {                                                                                                      // 158
            storageUsage: diff                                                                                         // 158
          }                                                                                                            // 158
        });                                                                                                            // 158
      }                                                                                                                // 159
    }                                                                                                                  // 160
                                                                                                                       //
    return reportGrainSize;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  SandstormCoreImpl.prototype.getIdentityId = function () {                                                            //
    function getIdentityId(identity) {                                                                                 //
      return unwrapFrontendCap(identity, "identity", function (identityId) {                                           // 163
        return {                                                                                                       // 164
          id: new Buffer(identityId, "hex")                                                                            // 164
        };                                                                                                             // 164
      });                                                                                                              // 165
    }                                                                                                                  // 166
                                                                                                                       //
    return getIdentityId;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  return SandstormCoreImpl;                                                                                            //
}();                                                                                                                   //
                                                                                                                       //
var makeSandstormCore = function (db, grainId) {                                                                       // 169
  return new Capnp.Capability(new SandstormCoreImpl(db, grainId), SandstormCore);                                      // 170
};                                                                                                                     // 171
                                                                                                                       //
var NotificationHandle = function (_PersistentImpl) {                                                                  //
  (0, _inherits3.default)(NotificationHandle, _PersistentImpl);                                                        //
                                                                                                                       //
  // TODO(cleanup): Move to a different file.                                                                          // 174
  function NotificationHandle(db, saveTemplate, notificationId) {                                                      // 176
    (0, _classCallCheck3.default)(this, NotificationHandle);                                                           // 176
                                                                                                                       //
    var _this7 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));         // 176
                                                                                                                       //
    _this7.notificationId = notificationId;                                                                            // 178
    _this7.db = db;                                                                                                    // 179
    return _this7;                                                                                                     // 176
  }                                                                                                                    // 180
                                                                                                                       //
  NotificationHandle.prototype.close = function () {                                                                   //
    function close() {                                                                                                 //
      var _this8 = this;                                                                                               // 182
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 183
        if (!_this8.isSaved()) {                                                                                       // 184
          dismissNotification(_this8.db, _this8.notificationId);                                                       // 185
        }                                                                                                              // 186
      });                                                                                                              // 187
    }                                                                                                                  // 188
                                                                                                                       //
    return close;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  return NotificationHandle;                                                                                           //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
var PersistentUiViewImpl = function (_PersistentImpl2) {                                                               //
  (0, _inherits3.default)(PersistentUiViewImpl, _PersistentImpl2);                                                     //
                                                                                                                       //
  // TODO(cleanup): Move out of core.js.                                                                               // 192
  function PersistentUiViewImpl(db, saveTemplate, grainId) {                                                           // 194
    (0, _classCallCheck3.default)(this, PersistentUiViewImpl);                                                         // 194
                                                                                                                       //
    var _this9 = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl2.call(this, db, saveTemplate));        // 194
                                                                                                                       //
    check(grainId, String);                                                                                            // 196
    _this9._db = db;                                                                                                   // 197
    _this9._grainId = grainId;                                                                                         // 198
    return _this9;                                                                                                     // 194
  }                                                                                                                    // 199
                                                                                                                       //
  PersistentUiViewImpl.prototype.getViewInfo = function () {                                                           //
    function getViewInfo() {                                                                                           //
      var _this10 = this;                                                                                              // 201
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 202
        var grain = _this10._db.getGrain(_this10._grainId);                                                            // 203
                                                                                                                       //
        if (!grain || grain.trashed) {                                                                                 // 204
          throw new Error("grain no longer exists");                                                                   // 205
        }                                                                                                              // 206
                                                                                                                       //
        var pkg = _this10._db.getPackage(grain.packageId) || DevPackages.findOne({                                     // 208
          appId: grain.appId                                                                                           // 209
        }) || {};                                                                                                      // 209
        var manifest = pkg.manifest || {};                                                                             // 212
        var viewInfo = grain.cachedViewInfo || {};                                                                     // 214
                                                                                                                       //
        if (!viewInfo.appTitle) {                                                                                      // 216
          viewInfo.appTitle = manifest.appTitle || {};                                                                 // 217
        }                                                                                                              // 218
                                                                                                                       //
        if (!viewInfo.grainIcon) {                                                                                     // 220
          var grainIcon = ((manifest.metadata || {}).icons || {}).grain;                                               // 221
                                                                                                                       //
          if (grainIcon) {                                                                                             // 222
            viewInfo.grainIcon = new Capnp.Capability(new StaticAssetImpl(grainIcon.assetId), StaticAsset);            // 223
          } else {                                                                                                     // 225
            var hash = Identicon.hashAppIdForIdenticon(pkg.appId);                                                     // 226
            viewInfo.grainIcon = new Capnp.Capability(new IdenticonStaticAssetImpl(hash, 24), StaticAsset);            // 227
          }                                                                                                            // 229
        }                                                                                                              // 230
                                                                                                                       //
        return viewInfo;                                                                                               // 232
      });                                                                                                              // 233
    }                                                                                                                  // 234
                                                                                                                       //
    return getViewInfo;                                                                                                //
  }(); // All other UiView methods are currently unimplemented, which, while not strictly correct,                     //
  // results in the same overall behavior, since users can't call restore() on a PersistentUiView,                     // 237
  // and grains can't call methods on UiViews because they lack the "is human" pseudopermission.                       // 238
                                                                                                                       //
                                                                                                                       //
  return PersistentUiViewImpl;                                                                                         //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
var makePersistentUiView = function (db, saveTemplate, grainId) {                                                      // 241
  check(grainId, String); // Verify that the grain exists and hasn't been trashed.                                     // 242
                                                                                                                       //
  var grain = db.getGrain(grainId);                                                                                    // 245
                                                                                                                       //
  if (!grain || grain.trashed) {                                                                                       // 246
    throw new Meteor.Error(404, "grain not found");                                                                    // 247
  }                                                                                                                    // 248
                                                                                                                       //
  return new Capnp.Capability(new PersistentUiViewImpl(db, saveTemplate, grainId), PersistentUiView);                  // 250
};                                                                                                                     // 252
                                                                                                                       //
globalFrontendRefRegistry.register({                                                                                   // 254
  frontendRefField: "notificationHandle",                                                                              // 255
  restore: function (db, saveTemplate, notificationId) {                                                               // 257
    return new Capnp.Capability(new NotificationHandle(db, saveTemplate, notificationId), PersistentHandle);           // 258
  }                                                                                                                    // 260
});                                                                                                                    // 254
                                                                                                                       //
function dismissNotification(db, notificationId, callCancel) {                                                         // 263
  var notification = db.collections.notifications.findOne({                                                            // 264
    _id: notificationId                                                                                                // 264
  });                                                                                                                  // 264
                                                                                                                       //
  if (notification) {                                                                                                  // 265
    db.collections.notifications.remove({                                                                              // 266
      _id: notificationId                                                                                              // 266
    });                                                                                                                // 266
                                                                                                                       //
    if (notification.ongoing) {                                                                                        // 267
      var id = notification.ongoing;                                                                                   // 268
                                                                                                                       //
      if (!callCancel) {                                                                                               // 270
        dropInternal(db, id, {                                                                                         // 271
          frontend: null                                                                                               // 271
        });                                                                                                            // 271
      } else {                                                                                                         // 272
        var notificationCap = restoreInternal(db, id, {                                                                // 273
          frontend: null                                                                                               // 273
        }, []).cap;                                                                                                    // 273
        var castedNotification = notificationCap.castAs(PersistentOngoingNotification);                                // 274
        dropInternal(db, id, {                                                                                         // 275
          frontend: null                                                                                               // 275
        });                                                                                                            // 275
                                                                                                                       //
        try {                                                                                                          // 276
          waitPromise(castedNotification.cancel());                                                                    // 277
          castedNotification.close();                                                                                  // 278
          notificationCap.close();                                                                                     // 279
        } catch (err) {                                                                                                // 280
          if (err.kjType !== "disconnected") {                                                                         // 281
            // ignore disconnected errors, since cancel may shutdown the grain before the supervisor                   // 282
            // responds.                                                                                               // 283
            throw err;                                                                                                 // 284
          }                                                                                                            // 285
        }                                                                                                              // 286
      }                                                                                                                // 287
    } else if (notification.appUpdates) {                                                                              // 288
      _.forEach(notification.appUpdates, function (app, appId) {                                                       // 289
        db.deleteUnusedPackages(appId);                                                                                // 290
      });                                                                                                              // 291
    }                                                                                                                  // 292
  }                                                                                                                    // 293
}                                                                                                                      // 294
                                                                                                                       //
Meteor.methods({                                                                                                       // 296
  dismissNotification: function (notificationId) {                                                                     // 297
    // This will remove notifications from the database and from view of the user.                                     // 298
    // For ongoing notifications, it will begin the process of cancelling and dropping them from                       // 299
    // the app.                                                                                                        // 300
    check(notificationId, String);                                                                                     // 302
    var db = this.connection.sandstormDb;                                                                              // 304
    var notification = db.collections.notifications.findOne({                                                          // 306
      _id: notificationId                                                                                              // 306
    });                                                                                                                // 306
                                                                                                                       //
    if (!notification) {                                                                                               // 307
      throw new Meteor.Error(404, "Notification id not found.");                                                       // 308
    } else if (notification.userId !== Meteor.userId()) {                                                              // 309
      throw new Meteor.Error(403, "Notification does not belong to current user.");                                    // 310
    } else {                                                                                                           // 311
      dismissNotification(db, notificationId, true);                                                                   // 312
    }                                                                                                                  // 313
  },                                                                                                                   // 314
  readAllNotifications: function () {                                                                                  // 316
    // Marks all notifications as read for the current user.                                                           // 317
    if (!Meteor.userId()) {                                                                                            // 318
      throw new Meteor.Error(403, "User not logged in.");                                                              // 319
    }                                                                                                                  // 320
                                                                                                                       //
    var db = this.connection.sandstormDb;                                                                              // 322
    db.collections.notifications.update({                                                                              // 323
      userId: Meteor.userId()                                                                                          // 324
    }, {                                                                                                               // 324
      $set: {                                                                                                          // 325
        isUnread: false                                                                                                // 325
      }                                                                                                                // 325
    }, {                                                                                                               // 325
      multi: true                                                                                                      // 326
    });                                                                                                                // 326
  },                                                                                                                   // 327
  acceptPowerboxOffer: function (sessionId, sturdyRef, sessionToken) {                                                 // 329
    var db = this.connection.sandstormDb;                                                                              // 330
    check(sessionId, String);                                                                                          // 331
    check(sturdyRef, String);                                                                                          // 332
    check(sessionToken, Match.OneOf(String, null, undefined));                                                         // 333
    var sessionQuery = {                                                                                               // 335
      _id: sessionId                                                                                                   // 335
    };                                                                                                                 // 335
                                                                                                                       //
    if (sessionToken) {                                                                                                // 336
      sessionQuery.hashedToken = hashSturdyRef(sessionToken);                                                          // 337
    } else {                                                                                                           // 338
      sessionQuery.userId = this.userId;                                                                               // 339
    }                                                                                                                  // 340
                                                                                                                       //
    var session = db.collections.sessions.findOne(sessionQuery);                                                       // 342
                                                                                                                       //
    if (!session) {                                                                                                    // 343
      throw new Meteor.Error(404, "No matching session found.");                                                       // 344
    }                                                                                                                  // 345
                                                                                                                       //
    var apiToken = fetchApiToken(db, sturdyRef, {                                                                      // 347
      "owner.clientPowerboxOffer.sessionId": sessionId                                                                 // 348
    });                                                                                                                // 348
                                                                                                                       //
    if (!apiToken) {                                                                                                   // 349
      throw new Meteor.Error(404, "No such token.");                                                                   // 350
    }                                                                                                                  // 351
                                                                                                                       //
    var tokenId = apiToken._id;                                                                                        // 353
    var newSturdyRef = void 0;                                                                                         // 355
                                                                                                                       //
    if (sessionToken && apiToken.parentToken) {                                                                        // 356
      // An anonymous user is being offered a child token. To avoid bloating the database,                             // 357
      // we deterministically derive the sturdyref from the session token and the hashed parent                        // 358
      // token.                                                                                                        // 359
      //                                                                                                               // 360
      // Note that an attacker with read access to the database might be able to derive                                // 361
      // `newSturdyRef` from `sessionToken` and gain access to the capability without ever having                      // 362
      // been explicitly offered it by the grain. Therefore, grains must assume that offering                          // 363
      // a capability to an anonymous user can possibly make that capability accessible to any                         // 364
      // other anonymous user who has connected through the same URL.                                                  // 365
      newSturdyRef = Crypto.createHash("sha256").update(sessionToken).update(apiToken.parentToken).digest("base64").slice(0, -1) // removing trailing "="
      .replace(/\+/g, "-").replace(/\//g, "_"); // make URL-safe                                                       // 367
                                                                                                                       //
      if (fetchApiToken(db, newSturdyRef)) {                                                                           // 374
        // We have already generated this token.                                                                       // 375
        db.removeApiTokens({                                                                                           // 376
          _id: tokenId                                                                                                 // 376
        });                                                                                                            // 376
        return newSturdyRef;                                                                                           // 377
      }                                                                                                                // 378
    } else {                                                                                                           // 379
      newSturdyRef = generateSturdyRef();                                                                              // 380
    }                                                                                                                  // 381
                                                                                                                       //
    apiToken.owner = {                                                                                                 // 383
      webkey: null                                                                                                     // 383
    };                                                                                                                 // 383
    insertApiToken(db, apiToken, newSturdyRef);                                                                        // 385
    db.removeApiTokens({                                                                                               // 386
      _id: tokenId                                                                                                     // 386
    });                                                                                                                // 386
    return newSturdyRef;                                                                                               // 387
  }                                                                                                                    // 388
});                                                                                                                    // 296
                                                                                                                       //
var makeSaveTemplateForChild = function (db, parentToken, requirements, parentTokenInfo) {                             // 392
  // Constructs (part of) an ApiToken record appropriate to be used when save()ing a capability                        // 393
  // that was originally created by restore()ing `parentToken`. This fills in everything that is                       // 394
  // appropriate to fill in based only on the parent. Some fields -- especially `owner`, `created`,                    // 395
  // and `_id` -- obviously cannot be filled in until `save()` is called.                                              // 396
  //                                                                                                                   // 397
  // `parentToken` is the raw SturdyRef of the parent.                                                                 // 398
  //                                                                                                                   // 399
  // `requirements` is a list of MembraneRequirements that should be added to any children.                            // 400
  //                                                                                                                   // 401
  // `parentTokenInfo` is the ApiToken record for `parentToken`. Provide this only if you have                         // 402
  // it handy; if omitted it will be looked up.                                                                        // 403
  parentTokenInfo = parentTokenInfo || fetchApiToken(db, parentToken);                                                 // 405
                                                                                                                       //
  if (!parentTokenInfo) {                                                                                              // 406
    throw new Error("no such token");                                                                                  // 407
  }                                                                                                                    // 408
                                                                                                                       //
  var parentOwner = parentTokenInfo.owner || {};                                                                       // 410
  var saveTemplate = void 0;                                                                                           // 412
                                                                                                                       //
  if (parentOwner.clientPowerboxRequest || parentOwner.clientPowerboxOffer) {                                          // 413
    // Saving this token should make a copy of the restored token, rather than make a child                            // 414
    // token.                                                                                                          // 415
    saveTemplate = _.clone(parentTokenInfo); // Don't copy over fields that should be determined at save() time.       // 417
                                                                                                                       //
    delete saveTemplate._id;                                                                                           // 420
    delete saveTemplate.owner;                                                                                         // 421
    delete saveTemplate.created;                                                                                       // 422
  } else {                                                                                                             // 423
    if (parentTokenInfo.identityId) {                                                                                  // 424
      // A UiView token. Need to denormalize some fields from the parent.                                              // 425
      saveTemplate = _.pick(parentTokenInfo, "grainId", "identityId", "accountId"); // By default, a save()d copy should have the same permissions, so set an allAccess role
      // assignment.                                                                                                   // 429
                                                                                                                       //
      saveTemplate.roleAssignment = {                                                                                  // 430
        allAccess: null                                                                                                // 430
      };                                                                                                               // 430
    } else {                                                                                                           // 431
      // A non-UiView child token.                                                                                     // 432
      saveTemplate = {};                                                                                               // 433
    } // Saved token should be a child of the restored token.                                                          // 434
                                                                                                                       //
                                                                                                                       //
    saveTemplate.parentToken = parentTokenInfo._id; // Will be encrypted on save().                                    // 437
                                                                                                                       //
    saveTemplate.parentTokenKey = parentToken;                                                                         // 440
  }                                                                                                                    // 441
                                                                                                                       //
  if (requirements) {                                                                                                  // 443
    // Append additional requirements requested by caller.                                                             // 444
    saveTemplate.requirements = (saveTemplate.requirements || []).concat(requirements);                                // 445
  }                                                                                                                    // 446
                                                                                                                       //
  return saveTemplate;                                                                                                 // 448
};                                                                                                                     // 449
                                                                                                                       //
restoreInternal = function (db, originalToken, ownerPattern, requirements, originalTokenInfo, currentTokenId, currentTokenKey) {
  // Restores the token `originalToken`, which is a Buffer.                                                            // 453
  //                                                                                                                   // 454
  // `ownerPattern` is a match pattern (i.e. used with check()) that the token's owner must match.                     // 455
  // This is used to enforce than an entity can't use tokens owned by some other entity.                               // 456
  //                                                                                                                   // 457
  // `requirements` is a list of additional MembraneRequirements to add to the returned capability,                    // 458
  // beyond what's already stored in ApiTokens. This is often an empty list.                                           // 459
  //                                                                                                                   // 460
  // `originalTokenInfo` is optional. If specified, it should be the ApiToken record associated                        // 461
  // with `originalToken`. Provide this if you happen to have looked it up already.                                    // 462
  //                                                                                                                   // 463
  // `currentTokenId` is only provided when this function calls itself recursively. It is the                          // 464
  // _id (hashed token) of an ancestor of originalToken. The function recurses until it gets to                        // 465
  // the top-level token.                                                                                              // 466
  requirements = requirements || [];                                                                                   // 468
                                                                                                                       //
  if (!originalTokenInfo) {                                                                                            // 470
    originalTokenInfo = fetchApiToken(db, originalToken);                                                              // 471
                                                                                                                       //
    if (!originalTokenInfo) {                                                                                          // 472
      throw new Meteor.Error(403, "No token found to restore");                                                        // 473
    }                                                                                                                  // 474
  }                                                                                                                    // 475
                                                                                                                       //
  var token = !currentTokenId ? originalTokenInfo : typeof currentTokenKey === "string" ? fetchApiToken(db, currentTokenKey) : db.collections.apiTokens.findOne(currentTokenId);
                                                                                                                       //
  if (!token) {                                                                                                        // 480
    if (!originalTokenInfo) {                                                                                          // 481
      throw new Meteor.Error(403, "Couldn't restore token because parent token has been deleted");                     // 482
    }                                                                                                                  // 483
  }                                                                                                                    // 484
                                                                                                                       //
  if (token.revoked) {                                                                                                 // 486
    throw new Meteor.Error(403, "Token has been revoked");                                                             // 487
  } // The ownerPattern should specify the appropriate user or grain involved, if appropriate.                         // 488
                                                                                                                       //
                                                                                                                       //
  check(token.owner, ownerPattern); // Check requirements on the token.                                                // 491
                                                                                                                       //
  checkRequirements(db, token.requirements); // Check expiration.                                                      // 494
                                                                                                                       //
  if (token.expires && token.expires.getTime() <= Date.now()) {                                                        // 497
    throw new Meteor.Error(403, "Authorization token expired");                                                        // 498
  }                                                                                                                    // 499
                                                                                                                       //
  if (token.expiresIfUnused) {                                                                                         // 501
    if (token.expiresIfUnused.getTime() <= Date.now()) {                                                               // 502
      throw new Meteor.Error(403, "Authorization token expired");                                                      // 503
    } else {                                                                                                           // 504
      // It's getting used now, so clear the expiresIfUnused field.                                                    // 505
      db.collections.apiTokens.update(token._id, {                                                                     // 506
        $unset: {                                                                                                      // 506
          expiresIfUnused: ""                                                                                          // 506
        }                                                                                                              // 506
      });                                                                                                              // 506
    }                                                                                                                  // 507
  } // If this token has a parent, go ahead and recurse to it now.                                                     // 508
                                                                                                                       //
                                                                                                                       //
  if (token.parentToken) {                                                                                             // 511
    // A token which chains to some parent token.  Restore the parent token (possibly recursively),                    // 512
    // checking requirements on the way up.                                                                            // 513
    return restoreInternal(db, originalToken, Match.Any, requirements, originalTokenInfo, token.parentToken, token.parentTokenKey);
  } // Check the passed-in `requirements`.                                                                             // 516
                                                                                                                       //
                                                                                                                       //
  checkRequirements(db, requirements);                                                                                 // 519
                                                                                                                       //
  if (token.objectId) {                                                                                                // 521
    // A token which represents a specific capability exported by a grain.                                             // 522
    // Fix Mongo converting Buffers to Uint8Arrays.                                                                    // 524
    if (token.objectId.appRef) {                                                                                       // 525
      token.objectId.appRef = new Buffer(token.objectId.appRef);                                                       // 526
    } // Ensure the grain is running, then restore the capability.                                                     // 527
                                                                                                                       //
                                                                                                                       //
    return waitPromise(globalBackend.useGrain(token.grainId, function (supervisor) {                                   // 530
      // Note that in this case it is the supervisor's job to implement SystemPersistent, so we                        // 531
      // don't generate a saveTemplate here.                                                                           // 532
      return supervisor.restore(token.objectId, requirements, new Buffer(originalToken, "utf8"));                      // 533
    }));                                                                                                               // 534
  } else {                                                                                                             // 535
    // Construct a template ApiToken for use if the restored capability is save()d later.                              // 536
    var saveTemplate = makeSaveTemplateForChild(db, originalToken, requirements, originalTokenInfo);                   // 537
                                                                                                                       //
    if (token.frontendRef) {                                                                                           // 539
      // A token which represents a capability implemented by a pseudo-driver.                                         // 540
      var cap = globalFrontendRefRegistry.restore(db, saveTemplate, token.frontendRef);                                // 542
      return {                                                                                                         // 543
        cap: cap                                                                                                       // 543
      };                                                                                                               // 543
    } else if (token.grainId) {                                                                                        // 544
      // It's a UiView.                                                                                                // 545
      // If a grain is attempting to restore a UiView, it gets a UiView which filters out all                          // 547
      // the method calls.  In the future, we may allow grains to restore UiViews that pass along the                  // 548
      // "is human" pseudopermission (say, to allow an app to proxy all requests to some grain and                     // 549
      // do some transformation), which will return a different capability.                                            // 550
      return {                                                                                                         // 551
        cap: makePersistentUiView(db, saveTemplate, token.grainId)                                                     // 551
      };                                                                                                               // 551
    } else {                                                                                                           // 552
      throw new Meteor.Error(500, "Unknown token type. ID: " + token._id);                                             // 553
    }                                                                                                                  // 554
  }                                                                                                                    // 555
};                                                                                                                     // 556
                                                                                                                       //
function dropInternal(db, sturdyRef, ownerPattern) {                                                                   // 558
  // Drops `sturdyRef`, checking first that its owner matches `ownerPattern`.                                          // 559
  var token = fetchApiToken(db, sturdyRef);                                                                            // 561
                                                                                                                       //
  if (!token) {                                                                                                        // 562
    return;                                                                                                            // 563
  }                                                                                                                    // 564
                                                                                                                       //
  var hashedSturdyRef = token._id;                                                                                     // 566
                                                                                                                       //
  if (!Match.test(token.owner, ownerPattern)) {                                                                        // 568
    // The caller of `drop()` does not own the token. From the caller's perspective, this means                        // 569
    // that there is actually nothing to be dropped. For example, perhaps a grain got backed up                        // 570
    // and restored, and now is trying to drop one of its pre-backup tokens. The call to                               // 571
    // `SandstormApi.drop()` should succeed, behaving exactly as if the token was not found.                           // 572
    return;                                                                                                            // 573
  } // TODO(soon): Revoke OAuth tokens.                                                                                // 574
                                                                                                                       //
                                                                                                                       //
  if (token.frontendRef && token.frontendRef.notificationHandle) {                                                     // 578
    var notificationId = token.frontendRef.notificationHandle;                                                         // 579
    db.removeApiTokens({                                                                                               // 580
      _id: hashedSturdyRef                                                                                             // 580
    });                                                                                                                // 580
    var anyToken = db.collections.apiTokens.findOne({                                                                  // 581
      "frontendRef.notificationHandle": notificationId                                                                 // 582
    });                                                                                                                // 582
                                                                                                                       //
    if (!anyToken) {                                                                                                   // 583
      // No other tokens referencing this notification exist, so dismiss the notification                              // 584
      dismissNotification(db, notificationId);                                                                         // 585
    }                                                                                                                  // 586
  } else if (token.objectId) {                                                                                         // 587
    waitPromise(globalBackend.useGrain(token.grainId, function (supervisor) {                                          // 588
      return supervisor.drop(token.objectId);                                                                          // 589
    }));                                                                                                               // 590
    db.removeApiTokens({                                                                                               // 592
      _id: hashedSturdyRef                                                                                             // 592
    });                                                                                                                // 592
  } else {                                                                                                             // 593
    db.removeApiTokens({                                                                                               // 594
      _id: hashedSturdyRef                                                                                             // 594
    });                                                                                                                // 594
  }                                                                                                                    // 595
}                                                                                                                      // 596
                                                                                                                       //
var SandstormCoreFactoryImpl = function () {                                                                           //
  function SandstormCoreFactoryImpl(db) {                                                                              // 599
    (0, _classCallCheck3.default)(this, SandstormCoreFactoryImpl);                                                     // 599
    this.db = db;                                                                                                      // 600
  }                                                                                                                    // 601
                                                                                                                       //
  SandstormCoreFactoryImpl.prototype.getSandstormCore = function () {                                                  //
    function getSandstormCore(grainId) {                                                                               //
      return {                                                                                                         // 604
        core: makeSandstormCore(this.db, grainId)                                                                      // 604
      };                                                                                                               // 604
    }                                                                                                                  // 605
                                                                                                                       //
    return getSandstormCore;                                                                                           //
  }();                                                                                                                 //
                                                                                                                       //
  return SandstormCoreFactoryImpl;                                                                                     //
}();                                                                                                                   //
                                                                                                                       //
makeSandstormCoreFactory = function (db) {                                                                             // 608
  return new Capnp.Capability(new SandstormCoreFactoryImpl(db), SandstormCoreFactory);                                 // 609
};                                                                                                                     // 610
                                                                                                                       //
unwrapFrontendCap = function (cap, type, callback) {                                                                   // 612
  // Expect that `cap` is a Cap'n Proto capability implemented by the frontend as a frontendRef                        // 613
  // with the given type (the name of one of the fields of frontendRef). Unwraps the capability                        // 614
  // and then calls callback() with the `frontendRef[type]` descriptor object as                                       // 615
  // the paramater. The callback runs in a Meteor fiber, but this function can be called from                          // 616
  // anywhere. The function returns a Promise for the result of the callback.                                          // 617
  //                                                                                                                   // 618
  // (The reason for the callback, rather than just returning a Promise for the descriptor, is                         // 619
  // so that you don't have to do your own inMeteor() dance.)                                                          // 620
  // For now, we save() the capability and then dig through ApiTokens to find where it leads.                          // 622
  // TODO(cleanup): In theory we should be using something like CapabilityServerSet, but it is                         // 623
  //   not available in Javascript yet and even if it were, it wouldn't work in the case where                         // 624
  //   there are multiple front-end replicas, since the capability could be on a different                             // 625
  //   replica.                                                                                                        // 626
  return cap.castAs(SystemPersistent).save({                                                                           // 628
    frontend: null                                                                                                     // 628
  }).then(function (saveResult) {                                                                                      // 628
    return inMeteor(function () {                                                                                      // 629
      var tokenInfo = fetchApiToken(globalDb, saveResult.sturdyRef.toString("utf8")); // Delete the token now since it's not needed.
                                                                                                                       //
      ApiTokens.remove(tokenInfo._id);                                                                                 // 633
                                                                                                                       //
      for (;;) {                                                                                                       // 635
        if (!tokenInfo) throw new Error("missing token?");                                                             // 636
        if (!tokenInfo.parentToken) break;                                                                             // 637
                                                                                                                       //
        if (typeof tokenInfo.parentTokenKey === "string") {                                                            // 638
          tokenInfo = fetchApiToken(globalDb, tokenInfo.parentTokenKey);                                               // 639
        } else {                                                                                                       // 640
          // Hmm, parentTokenKey doesn't exist or is still encrypted. We can't decrypt the parent                      // 641
          // but we can still fetch it in encrypted format.                                                            // 642
          tokenInfo = ApiTokens.findOne(tokenInfo.parentToken);                                                        // 643
        }                                                                                                              // 644
      }                                                                                                                // 645
                                                                                                                       //
      if (!tokenInfo.frontendRef || !tokenInfo.frontendRef[type]) {                                                    // 647
        throw new Error("not a " + type + " capability");                                                              // 648
      }                                                                                                                // 649
                                                                                                                       //
      return callback(tokenInfo.frontendRef[type]);                                                                    // 651
    });                                                                                                                // 652
  });                                                                                                                  // 653
};                                                                                                                     // 654
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"demo-server.js":["/imports/server/async-helpers.js","/imports/demo.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/demo-server.js                                                                                               //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var allowDemo = void 0;                                                                                                // 1
module.import("/imports/demo.js", {                                                                                    // 1
  "allowDemo": function (v) {                                                                                          // 1
    allowDemo = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var DEMO_EXPIRATION_MS = 60 * 60 * 1000;                                                                               // 20
var DEMO_GRACE_MS = 10 * 60 * 1000; // time between expiration and deletion                                            // 21
                                                                                                                       //
Accounts.validateLoginAttempt(function (attempt) {                                                                     // 23
  // Enforce expiration times for demo accounts.                                                                       // 24
  if (attempt.user && attempt.user.expires) {                                                                          // 26
    var expireIn = attempt.user.expires.getTime() - Date.now() + DEMO_GRACE_MS;                                        // 27
                                                                                                                       //
    if (expireIn < 0) {                                                                                                // 28
      throw new Meteor.Error(403, "This demo account has expired.");                                                   // 29
    }                                                                                                                  // 30
                                                                                                                       //
    if (!allowDemo) {                                                                                                  // 32
      // An expiration time implies that this is a demo account, but those have been disabled.                         // 33
      // Perhaps the config was just changed.                                                                          // 34
      throw new Meteor.Error(403, "Demo accounts have been disabled.");                                                // 35
    } // Force connection close when account expires, so that the client reconnects and                                // 36
    // re-authenticates, which fails.                                                                                  // 39
                                                                                                                       //
                                                                                                                       //
    var connection = attempt.connection;                                                                               // 40
    var handle = Meteor.setTimeout(function () {                                                                       // 41
      connection.close();                                                                                              // 41
    }, expireIn);                                                                                                      // 41
    connection.onClose(function () {                                                                                   // 43
      Meteor.clearTimeout(handle);                                                                                     // 43
    });                                                                                                                // 43
  }                                                                                                                    // 44
                                                                                                                       //
  return true;                                                                                                         // 46
});                                                                                                                    // 47
                                                                                                                       //
function cleanupExpiredUsers() {                                                                                       // 49
  // Delete expired demo accounts and all their grains.                                                                // 50
  var now = new Date(Date.now() - DEMO_GRACE_MS);                                                                      // 52
  Meteor.users.find({                                                                                                  // 53
    expires: {                                                                                                         // 53
      $lt: now                                                                                                         // 53
    },                                                                                                                 // 53
    loginIdentities: {                                                                                                 // 53
      $exists: true                                                                                                    // 53
    }                                                                                                                  // 53
  }, {                                                                                                                 // 53
    fields: {                                                                                                          // 54
      _id: 1,                                                                                                          // 54
      loginIdentities: 1,                                                                                              // 54
      lastActive: 1,                                                                                                   // 54
      appDemoId: 1,                                                                                                    // 54
      experiments: 1                                                                                                   // 54
    }                                                                                                                  // 54
  }).forEach(function (user) {                                                                                         // 54
    console.log("delete demo user: " + user._id);                                                                      // 56
    globalDb.deleteAccount(user._id, globalBackend); // Record stats about demo accounts.                              // 57
                                                                                                                       //
    var deleteStatsType = "demoUser";                                                                                  // 60
    var isAppDemoUser = !!user.appDemoId;                                                                              // 61
                                                                                                                       //
    if (isAppDemoUser) {                                                                                               // 62
      deleteStatsType = "appDemoUser";                                                                                 // 63
    } // Intentionally record deleted users at time of deletion to avoid miscounting users that                        // 64
    // were demoing just before the day rolled over.                                                                   // 67
                                                                                                                       //
                                                                                                                       //
    var record = {                                                                                                     // 68
      type: deleteStatsType,                                                                                           // 68
      lastActive: new Date(),                                                                                          // 68
      appId: user.appDemoId                                                                                            // 68
    };                                                                                                                 // 68
                                                                                                                       //
    if (user.experiments) {                                                                                            // 69
      record.experiments = user.experiments;                                                                           // 70
    }                                                                                                                  // 71
                                                                                                                       //
    DeleteStats.insert(record);                                                                                        // 73
  }); // All demo identities should have been deleted as part of deleting the demo users, but just in                  // 74
  // case, check for them too.                                                                                         // 77
                                                                                                                       //
  Meteor.users.find({                                                                                                  // 78
    expires: {                                                                                                         // 78
      $lt: now                                                                                                         // 78
    },                                                                                                                 // 78
    loginIdentities: {                                                                                                 // 78
      $exists: false                                                                                                   // 78
    }                                                                                                                  // 78
  }, {                                                                                                                 // 78
    fields: {                                                                                                          // 79
      _id: 1                                                                                                           // 79
    }                                                                                                                  // 79
  }).forEach(function (user) {                                                                                         // 79
    console.log("delete demo identity: " + user._id);                                                                  // 81
    globalDb.deleteIdentity(user._id);                                                                                 // 82
  });                                                                                                                  // 83
}                                                                                                                      // 84
                                                                                                                       //
if (allowDemo) {                                                                                                       // 86
  Meteor.methods({                                                                                                     // 87
    createDemoUser: function (displayName, appDemoId) {                                                                // 88
      // This is a login method that creates a new temporary user                                                      // 89
      // every time it is used.                                                                                        // 90
      //                                                                                                               // 91
      // appDemoId is important for stats; see cleanupExpiredUsers().                                                  // 92
      check(displayName, String);                                                                                      // 93
      check(appDemoId, Match.OneOf(undefined, null, String)); // Create the new user.                                  // 94
                                                                                                                       //
      var newUser = {                                                                                                  // 97
        expires: new Date(Date.now() + DEMO_EXPIRATION_MS)                                                             // 97
      };                                                                                                               // 97
                                                                                                                       //
      if (appDemoId) {                                                                                                 // 98
        newUser.appDemoId = appDemoId;                                                                                 // 99
      }                                                                                                                // 100
                                                                                                                       //
      var userId = Accounts.insertUserDoc({                                                                            // 102
        profile: {                                                                                                     // 102
          name: displayName                                                                                            // 102
        }                                                                                                              // 102
      }, newUser); // Log them in on this connection.                                                                  // 102
                                                                                                                       //
      return Accounts._loginMethod(this, "createDemoUser", arguments, "demo", function () {                            // 105
        return {                                                                                                       // 106
          userId: userId                                                                                               // 106
        };                                                                                                             // 106
      });                                                                                                              // 106
    },                                                                                                                 // 107
    testExpireDemo: function () {                                                                                      // 109
      if (!isDemoUser()) throw new Meteor.Error(403, "not a demo user");                                               // 110
      var newExpires = new Date(Date.now() + 15000);                                                                   // 112
                                                                                                                       //
      if (Meteor.user().expires.getTime() < newExpires.getTime()) {                                                    // 113
        throw new Meteor.Error(403, "can't exend demo");                                                               // 114
      }                                                                                                                // 115
                                                                                                                       //
      Meteor.users.update(this.userId, {                                                                               // 117
        $set: {                                                                                                        // 117
          expires: newExpires                                                                                          // 117
        }                                                                                                              // 117
      });                                                                                                              // 117
    }                                                                                                                  // 118
  }); // If demo mode is enabled, we permit the client to subscribe to information about an app by                     // 87
  // appId. If this were available in non-demo mode, then anonymous users could effectively ask the                    // 122
  // server which apps are installed.                                                                                  // 123
                                                                                                                       //
  Meteor.publish("appDemoInfo", function (appId) {                                                                     // 124
    // This publishes info about an app, including the latest version of it. Once you log in, it                       // 125
    // also publishes your list of UserActions.                                                                        // 126
    check(appId, String); // Get data about this app from the app index. Note that the app index cache on this server  // 127
    // is typically 6-24 hours delayed from reality, so if an app is newly-available                                   // 130
    // in the app market, it won't be in the app index. Therefore we don't bail-out if the app                         // 131
    // is missing from the AppIndex collection.                                                                        // 132
                                                                                                                       //
    var appIndexData = globalDb.collections.appIndex.findOne({                                                         // 133
      appId: appId                                                                                                     // 133
    }); // Prepare a helper function we can use to transform a package cursor into an appropriate return               // 133
    // value for this function.                                                                                        // 136
                                                                                                                       //
    var packageCursorAndMaybeUserActions = function (userId, appId, packageCursor) {                                   // 137
      // This allows us to avoid creating duplicate UserActions.                                                       // 138
      if (userId) {                                                                                                    // 139
        return [packageCursor, UserActions.find({                                                                      // 140
          userId: userId,                                                                                              // 142
          appId: appId                                                                                                 // 142
        })];                                                                                                           // 142
      }                                                                                                                // 144
                                                                                                                       //
      return packageCursor;                                                                                            // 146
    }; // If the app is in the app index, and the current version is installed, always return that. If                 // 147
    // that package isn't installed, store the version number so we can filter on it later.                            // 150
                                                                                                                       //
                                                                                                                       //
    var packageQuery = {};                                                                                             // 151
                                                                                                                       //
    if (appIndexData) {                                                                                                // 152
      var appIndexPackageQuery = Packages.find({                                                                       // 153
        _id: appIndexData.packageId                                                                                    // 153
      });                                                                                                              // 153
                                                                                                                       //
      if (appIndexPackageQuery.count() > 0) {                                                                          // 154
        return packageCursorAndMaybeUserActions(this.userId, appId, appIndexPackageQuery);                             // 155
      } // If the app index version isn't present, insist on a lower version than the app index                        // 156
      // version. This avoids accidentally catching some development version of the app that has the                   // 159
      // same version number as the app market version but isn't the app market version.                               // 160
                                                                                                                       //
                                                                                                                       //
      packageQuery["manifest.appVersion"] = {                                                                          // 161
        $lt: appIndexData.versionNumber                                                                                // 161
      };                                                                                                               // 161
    } // If the specific package from the app index isn't installed, or the app isn't there at all, do                 // 162
    // our best.                                                                                                       // 165
                                                                                                                       //
                                                                                                                       //
    packageQuery.appId = appId;                                                                                        // 166
    var packageCursor = Packages.find(packageQuery, {                                                                  // 167
      sort: {                                                                                                          // 169
        "manifest.appVersion": -1                                                                                      // 169
      },                                                                                                               // 169
      limit: 1                                                                                                         // 170
    });                                                                                                                // 169
    return packageCursorAndMaybeUserActions(this.userId, appId, packageCursor);                                        // 172
  });                                                                                                                  // 173
  SandstormDb.periodicCleanup(DEMO_EXPIRATION_MS, cleanupExpiredUsers);                                                // 175
} else {                                                                                                               // 176
  // Just run once, in case the config just changed from allowing demos to prohibiting them.                           // 177
  Meteor.setTimeout(cleanupExpiredUsers, DEMO_EXPIRATION_MS + DEMO_GRACE_MS);                                          // 178
}                                                                                                                      // 179
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"desktop-notifications-server.js":["meteor/meteor",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/desktop-notifications-server.js                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var Meteor = void 0;                                                                                                   // 1
module.import("meteor/meteor", {                                                                                       // 1
  "Meteor": function (v) {                                                                                             // 1
    Meteor = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
SandstormDb.periodicCleanup(120000, function () {                                                                      // 19
  // Remove old desktop notfications regularly.                                                                        // 20
  // periodicCleanup doesn't like runs more frequent than once every two minutes                                       // 21
  var now = Date.now();                                                                                                // 22
  var then = new Date(now - 30000); // Clear tokens older than 30 seconds.                                             // 23
                                                                                                                       //
  globalDb.collections.desktopNotifications.find({                                                                     // 24
    creationDate: {                                                                                                    // 25
      $lt: then                                                                                                        // 25
    }                                                                                                                  // 25
  }).forEach(function (doc) {                                                                                          // 24
    globalDb.collections.desktopNotifications.remove({                                                                 // 27
      _id: doc._id                                                                                                     // 28
    });                                                                                                                // 27
  });                                                                                                                  // 30
});                                                                                                                    // 31
Meteor.publish("desktopNotifications", function () {                                                                   // 33
  var _this = this;                                                                                                    // 33
                                                                                                                       //
  var subscribeTime = new Date();                                                                                      // 34
                                                                                                                       //
  if (!this.userId) {                                                                                                  // 36
    // No desktop notifications for anonymous users.                                                                   // 37
    return [];                                                                                                         // 38
  }                                                                                                                    // 39
                                                                                                                       //
  var db = this.connection.sandstormDb;                                                                                // 41
  var callbacks = {                                                                                                    // 42
    added: function (doc) {                                                                                            // 43
      _this.added("desktopNotifications", doc._id, doc);                                                               // 44
    },                                                                                                                 // 45
    changed: function (newDoc, oldDoc) {                                                                               // 47
      _this.changed("desktopNotifications", newDoc._id, newDoc);                                                       // 48
    },                                                                                                                 // 49
    removed: function (doc) {                                                                                          // 51
      _this.removed("desktopNotifications", doc._id);                                                                  // 52
    }                                                                                                                  // 53
  };                                                                                                                   // 42
  var sub = db.collections.desktopNotifications.find({                                                                 // 56
    userId: this.userId,                                                                                               // 57
    creationDate: {                                                                                                    // 58
      $gt: subscribeTime                                                                                               // 58
    }                                                                                                                  // 58
  }).observe(callbacks);                                                                                               // 56
  this.onStop(function () {                                                                                            // 61
    if (sub) {                                                                                                         // 62
      sub.stop();                                                                                                      // 63
    }                                                                                                                  // 64
  });                                                                                                                  // 65
  this.ready();                                                                                                        // 67
});                                                                                                                    // 68
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"dev-accounts-server.js":function(require){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/dev-accounts-server.js                                                                                       //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Crypto = Npm.require("crypto");                                                                                    // 17
                                                                                                                       //
Meteor.methods({                                                                                                       // 19
  createDevAccount: function (displayName, isAdmin, profile, unverifiedEmail) {                                        // 20
    // This is a login method that creates or logs in a dev account with the given displayName                         // 21
    check(displayName, String);                                                                                        // 23
    check(isAdmin, Match.OneOf(undefined, null, Boolean));                                                             // 24
    check(profile, Match.OneOf(undefined, null, Object));                                                              // 25
    check(unverifiedEmail, Match.OneOf(undefined, null, String));                                                      // 26
                                                                                                                       //
    if (!this.connection.sandstormDb.allowDevAccounts()) {                                                             // 28
      throw new Meteor.Error(404, "Dev accounts are not enabled on this server");                                      // 29
    }                                                                                                                  // 30
                                                                                                                       //
    isAdmin = isAdmin || false;                                                                                        // 32
    profile = profile || {};                                                                                           // 34
    profile.name = profile.name || displayName;                                                                        // 35
    var hasCompletedSignup = !!unverifiedEmail && !!profile.pronoun && !!profile.handle;                               // 36
    var user = Meteor.users.findOne({                                                                                  // 38
      "services.dev.name": displayName                                                                                 // 38
    });                                                                                                                // 38
    var userId = void 0;                                                                                               // 39
                                                                                                                       //
    if (user) {                                                                                                        // 41
      userId = user._id;                                                                                               // 42
    } else {                                                                                                           // 43
      userId = Accounts.insertUserDoc({                                                                                // 44
        profile: profile,                                                                                              // 45
        unverifiedEmail: unverifiedEmail                                                                               // 46
      }, {                                                                                                             // 44
        services: {                                                                                                    // 48
          dev: {                                                                                                       // 49
            name: displayName,                                                                                         // 50
            isAdmin: isAdmin,                                                                                          // 51
            hasCompletedSignup: hasCompletedSignup                                                                     // 52
          }                                                                                                            // 49
        }                                                                                                              // 48
      });                                                                                                              // 47
    } // Log them in on this connection.                                                                               // 56
                                                                                                                       //
                                                                                                                       //
    return Accounts._loginMethod(this, "createDevAccount", arguments, "dev", function () {                             // 58
      return {                                                                                                         // 59
        userId: userId                                                                                                 // 59
      };                                                                                                               // 59
    });                                                                                                                // 59
  }                                                                                                                    // 60
});                                                                                                                    // 19
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"grain-server.js":["babel-runtime/helpers/typeof","/imports/server/email.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/grain-server.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _typeof2 = require("babel-runtime/helpers/typeof");                                                                //
                                                                                                                       //
var _typeof3 = _interopRequireDefault(_typeof2);                                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var sendEmail = void 0;                                                                                                // 1
module.import("/imports/server/email.js", {                                                                            // 1
  "send": function (v) {                                                                                               // 1
    sendEmail = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Crypto = Npm.require("crypto");                                                                                    // 17
                                                                                                                       //
var ROOT_URL = process.env.ROOT_URL;                                                                                   // 20
                                                                                                                       //
var emailLinkWithInlineStyle = function (url, text) {                                                                  // 22
  return "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#762F87;color:white'>" + text + "</a>";
};                                                                                                                     // 27
                                                                                                                       //
Meteor.publish("grainTopBar", function (grainId) {                                                                     // 29
  check(grainId, String);                                                                                              // 30
  var result = [Grains.find({                                                                                          // 32
    _id: grainId,                                                                                                      // 34
    $or: [{                                                                                                            // 35
      userId: this.userId                                                                                              // 36
    }, {                                                                                                               // 36
      "private": {                                                                                                     // 37
        $ne: true                                                                                                      // 37
      }                                                                                                                // 37
    }]                                                                                                                 // 37
  }, {                                                                                                                 // 33
    fields: {                                                                                                          // 40
      title: 1,                                                                                                        // 41
      userId: 1,                                                                                                       // 42
      identityId: 1,                                                                                                   // 43
      "private": 1                                                                                                     // 44
    }                                                                                                                  // 40
  })];                                                                                                                 // 39
                                                                                                                       //
  if (this.userId) {                                                                                                   // 48
    var myIdentityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));                                 // 49
    result.push(ApiTokens.find({                                                                                       // 50
      grainId: grainId,                                                                                                // 51
      $or: [{                                                                                                          // 52
        "owner.user.identityId": {                                                                                     // 53
          $in: myIdentityIds                                                                                           // 53
        }                                                                                                              // 53
      }, {                                                                                                             // 53
        identityId: {                                                                                                  // 54
          $in: myIdentityIds                                                                                           // 54
        }                                                                                                              // 54
      }]                                                                                                               // 54
    }));                                                                                                               // 50
  }                                                                                                                    // 57
                                                                                                                       //
  return result;                                                                                                       // 59
}); // We allow users to learn package information about a grain they own.                                             // 60
// This is used for obtaining icon and app title information for grains                                                // 63
// you own, which is used in the sidebar. It is not a security/privacy                                                 // 64
// risk since it only exposes this information for grains the user owns.                                               // 65
                                                                                                                       //
Meteor.publish("packageByGrainId", function (grainId) {                                                                // 66
  check(grainId, String);                                                                                              // 67
  var publishThis = []; // We need to publish the packageId so that client-side code can                               // 68
  // find the right package.                                                                                           // 70
                                                                                                                       //
  var thisGrainCursor = Grains.find({                                                                                  // 71
    _id: grainId,                                                                                                      // 72
    userId: this.userId                                                                                                // 73
  }, {                                                                                                                 // 71
    fields: {                                                                                                          // 75
      packageId: 1                                                                                                     // 75
    }                                                                                                                  // 75
  });                                                                                                                  // 74
  publishThis.push(thisGrainCursor);                                                                                   // 77
                                                                                                                       //
  if (thisGrainCursor.count()) {                                                                                       // 79
    var thisGrain = thisGrainCursor.fetch()[0];                                                                        // 80
    var thisPackageCursor = Packages.find({                                                                            // 81
      _id: thisGrain.packageId                                                                                         // 81
    });                                                                                                                // 81
    publishThis.push(thisPackageCursor);                                                                               // 82
  }                                                                                                                    // 83
                                                                                                                       //
  return publishThis;                                                                                                  // 85
});                                                                                                                    // 86
Meteor.publish("tokenInfo", function (token, isStandalone) {                                                           // 88
  // Allows the client side to map a raw token to its entry in ApiTokens, and the additional                           // 89
  // metadata that it will need to display the app icon and title.  We do not care about making                        // 90
  // the metadata reactive.                                                                                            // 91
  check(token, String);                                                                                                // 92
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");                                        // 94
  var apiToken = ApiTokens.findOne({                                                                                   // 95
    _id: hashedToken                                                                                                   // 96
  }, {                                                                                                                 // 95
    fields: {                                                                                                          // 98
      grainId: 1,                                                                                                      // 99
      identityId: 1,                                                                                                   // 100
      owner: 1,                                                                                                        // 101
      revoked: 1                                                                                                       // 102
    }                                                                                                                  // 98
  });                                                                                                                  // 97
                                                                                                                       //
  if (!apiToken) {                                                                                                     // 105
    this.added("tokenInfo", token, {                                                                                   // 106
      invalidToken: true                                                                                               // 106
    });                                                                                                                // 106
  } else if (apiToken.revoked) {                                                                                       // 107
    this.added("tokenInfo", token, {                                                                                   // 108
      revoked: true                                                                                                    // 108
    });                                                                                                                // 108
  } else {                                                                                                             // 109
    var grainId = apiToken.grainId;                                                                                    // 110
    var grain = Grains.findOne({                                                                                       // 111
      _id: grainId                                                                                                     // 112
    }, {                                                                                                               // 111
      fields: {                                                                                                        // 114
        packageId: 1,                                                                                                  // 115
        appId: 1,                                                                                                      // 116
        userId: 1                                                                                                      // 117
      }                                                                                                                // 114
    });                                                                                                                // 113
                                                                                                                       //
    if (!grain) {                                                                                                      // 120
      this.added("tokenInfo", token, {                                                                                 // 121
        invalidToken: true                                                                                             // 121
      });                                                                                                              // 121
    } else {                                                                                                           // 122
      if (apiToken.owner && apiToken.owner.user) {                                                                     // 123
        var identity = globalDb.getIdentity(apiToken.owner.user.identityId);                                           // 124
        var metadata = apiToken.owner.user.denormalizedGrainMetadata;                                                  // 125
                                                                                                                       //
        if (identity && metadata) {                                                                                    // 126
          SandstormDb.fillInLoginId(identity);                                                                         // 127
          this.added("tokenInfo", token, {                                                                             // 128
            identityOwner: _.pick(identity, "_id", "profile", "loginId"),                                              // 129
            grainId: grainId,                                                                                          // 130
            grainMetadata: metadata                                                                                    // 131
          });                                                                                                          // 128
        } else {                                                                                                       // 133
          this.added("tokenInfo", token, {                                                                             // 134
            invalidToken: true                                                                                         // 134
          });                                                                                                          // 134
        }                                                                                                              // 135
      } else if (!apiToken.owner || "webkey" in apiToken.owner) {                                                      // 136
        if (this.userId && !isStandalone) {                                                                            // 137
          var user = Meteor.users.findOne({                                                                            // 138
            _id: this.userId                                                                                           // 138
          });                                                                                                          // 138
          var identityIds = SandstormDb.getUserIdentityIds(user);                                                      // 139
          var childToken = ApiTokens.findOne({                                                                         // 140
            "owner.user.identityId": {                                                                                 // 141
              $in: identityIds                                                                                         // 141
            },                                                                                                         // 141
            parentToken: apiToken._id                                                                                  // 142
          });                                                                                                          // 140
                                                                                                                       //
          if (childToken || this.userId === grain.userId || identityIds.indexOf(apiToken.identityId) >= 0) {           // 144
            this.added("tokenInfo", token, {                                                                           // 146
              alreadyRedeemed: true,                                                                                   // 146
              grainId: apiToken.grainId                                                                                // 146
            });                                                                                                        // 146
            this.ready();                                                                                              // 147
            return;                                                                                                    // 148
          }                                                                                                            // 149
        }                                                                                                              // 150
                                                                                                                       //
        var pkg = Packages.findOne({                                                                                   // 152
          _id: grain.packageId                                                                                         // 152
        }, {                                                                                                           // 152
          fields: {                                                                                                    // 152
            manifest: 1                                                                                                // 152
          }                                                                                                            // 152
        });                                                                                                            // 152
        var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || {                                               // 153
          defaultText: ""                                                                                              // 153
        };                                                                                                             // 153
        var appIcon = undefined;                                                                                       // 154
                                                                                                                       //
        if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {                             // 155
          var icons = pkg.manifest.metadata.icons;                                                                     // 156
          appIcon = icons.grain || icons.appGrid;                                                                      // 157
        }                                                                                                              // 158
                                                                                                                       //
        var denormalizedGrainMetadata = {                                                                              // 160
          appTitle: appTitle,                                                                                          // 161
          icon: appIcon,                                                                                               // 162
          appId: appIcon ? undefined : grain.appId                                                                     // 163
        };                                                                                                             // 160
        this.added("tokenInfo", token, {                                                                               // 165
          webkey: true,                                                                                                // 166
          grainId: grainId,                                                                                            // 167
          grainMetadata: denormalizedGrainMetadata                                                                     // 168
        });                                                                                                            // 165
      } else {                                                                                                         // 170
        this.added("tokenInfo", token, {                                                                               // 171
          invalidToken: true                                                                                           // 171
        });                                                                                                            // 171
      }                                                                                                                // 172
    }                                                                                                                  // 173
  }                                                                                                                    // 174
                                                                                                                       //
  this.ready();                                                                                                        // 176
  return;                                                                                                              // 177
});                                                                                                                    // 178
Meteor.publish("requestingAccess", function (grainId) {                                                                // 180
  check(grainId, String);                                                                                              // 181
                                                                                                                       //
  if (!this.userId) {                                                                                                  // 183
    throw new Meteor.Error(403, "Must be logged in to request access.");                                               // 184
  }                                                                                                                    // 185
                                                                                                                       //
  var grain = globalDb.getGrain(grainId);                                                                              // 187
                                                                                                                       //
  if (!grain) {                                                                                                        // 188
    throw new Meteor.Error(404, "Grain not found.");                                                                   // 189
  }                                                                                                                    // 190
                                                                                                                       //
  if (grain.userId === this.userId) {                                                                                  // 192
    this.added("grantedAccessRequests", Random.id(), {                                                                 // 193
      grainId: grainId,                                                                                                // 194
      identityId: grain.identityId                                                                                     // 194
    });                                                                                                                // 194
  }                                                                                                                    // 195
                                                                                                                       //
  var identityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({                                              // 197
    _id: this.userId                                                                                                   // 197
  }));                                                                                                                 // 197
  var ownerIdentityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({                                         // 198
    _id: grain.userId                                                                                                  // 199
  }));                                                                                                                 // 199
                                                                                                                       //
  var _this = this;                                                                                                    // 201
                                                                                                                       //
  var query = ApiTokens.find({                                                                                         // 202
    grainId: grainId,                                                                                                  // 203
    identityId: {                                                                                                      // 204
      $in: ownerIdentityIds                                                                                            // 204
    },                                                                                                                 // 204
    parentToken: {                                                                                                     // 205
      $exists: false                                                                                                   // 205
    },                                                                                                                 // 205
    "owner.user.identityId": {                                                                                         // 206
      $in: identityIds                                                                                                 // 206
    },                                                                                                                 // 206
    revoked: {                                                                                                         // 207
      $ne: true                                                                                                        // 207
    }                                                                                                                  // 207
  });                                                                                                                  // 202
  var handle = query.observe({                                                                                         // 209
    added: function (apiToken) {                                                                                       // 210
      _this.added("grantedAccessRequests", Random.id(), {                                                              // 211
        grainId: grainId,                                                                                              // 212
        identityId: apiToken.owner.user.identityId                                                                     // 212
      });                                                                                                              // 212
    }                                                                                                                  // 213
  });                                                                                                                  // 209
  this.onStop(function () {                                                                                            // 216
    return handle.stop();                                                                                              // 216
  });                                                                                                                  // 216
});                                                                                                                    // 217
var GRAIN_DELETION_MS = 1000 * 60 * 60 * 24 * 30; // thirty days                                                       // 219
                                                                                                                       //
SandstormDb.periodicCleanup(86400000, function () {                                                                    // 220
  var trashExpiration = new Date(Date.now() - GRAIN_DELETION_MS);                                                      // 221
  globalDb.removeApiTokens({                                                                                           // 222
    trashed: {                                                                                                         // 222
      $lt: trashExpiration                                                                                             // 222
    }                                                                                                                  // 222
  });                                                                                                                  // 222
  globalDb.deleteGrains({                                                                                              // 223
    trashed: {                                                                                                         // 223
      $lt: trashExpiration                                                                                             // 223
    }                                                                                                                  // 223
  }, globalBackend, "grain");                                                                                          // 223
});                                                                                                                    // 224
Meteor.methods({                                                                                                       // 226
  updateGrainPreferredIdentity: function (grainId, identityId) {                                                       // 227
    check(grainId, String);                                                                                            // 228
    check(identityId, String);                                                                                         // 229
                                                                                                                       //
    if (!this.userId) {                                                                                                // 230
      throw new Meteor.Error(403, "Must be logged in.");                                                               // 231
    }                                                                                                                  // 232
                                                                                                                       //
    var grain = globalDb.getGrain(grainId) || {};                                                                      // 234
                                                                                                                       //
    if (!grain.userId === this.userId) {                                                                               // 235
      throw new Meteor.Error(403, "Grain not owned by current user.");                                                 // 236
    }                                                                                                                  // 237
                                                                                                                       //
    Grains.update({                                                                                                    // 239
      _id: grainId                                                                                                     // 239
    }, {                                                                                                               // 239
      $set: {                                                                                                          // 239
        identityId: identityId                                                                                         // 239
      }                                                                                                                // 239
    });                                                                                                                // 239
  },                                                                                                                   // 240
  updateGrainTitle: function (grainId, newTitle, identityId) {                                                         // 242
    check(grainId, String);                                                                                            // 243
    check(newTitle, String);                                                                                           // 244
    check(identityId, String);                                                                                         // 245
                                                                                                                       //
    if (this.userId) {                                                                                                 // 246
      var grain = Grains.findOne(grainId);                                                                             // 247
                                                                                                                       //
      if (grain) {                                                                                                     // 248
        if (grain.userId === this.userId) {                                                                            // 249
          Grains.update({                                                                                              // 250
            _id: grainId,                                                                                              // 250
            userId: this.userId                                                                                        // 250
          }, {                                                                                                         // 250
            $set: {                                                                                                    // 250
              title: newTitle                                                                                          // 250
            }                                                                                                          // 250
          }); // Denormalize new title out to all sharing tokens.                                                      // 250
                                                                                                                       //
          ApiTokens.update({                                                                                           // 253
            grainId: grainId,                                                                                          // 253
            "owner.user": {                                                                                            // 253
              $exists: true                                                                                            // 253
            }                                                                                                          // 253
          }, {                                                                                                         // 253
            $set: {                                                                                                    // 254
              "owner.user.upstreamTitle": newTitle                                                                     // 254
            }                                                                                                          // 254
          }, {                                                                                                         // 254
            multi: true                                                                                                // 255
          });                                                                                                          // 255
        } else {                                                                                                       // 256
          if (!globalDb.userHasIdentity(this.userId, identityId)) {                                                    // 257
            throw new Meteor.Error(403, "Current user does not have identity " + identityId);                          // 258
          }                                                                                                            // 259
                                                                                                                       //
          var token = ApiTokens.findOne({                                                                              // 261
            grainId: grainId,                                                                                          // 262
            objectId: {                                                                                                // 263
              $exists: false                                                                                           // 263
            },                                                                                                         // 263
            "owner.user.identityId": identityId                                                                        // 264
          }, {                                                                                                         // 261
            sort: {                                                                                                    // 266
              created: 1                                                                                               // 266
            }                                                                                                          // 266
          });                                                                                                          // 265
                                                                                                                       //
          if (token && token.owner.user.title !== newTitle) {                                                          // 268
            if (token.owner.user.upstreamTitle === newTitle) {                                                         // 269
              // User renamed grain to match upstream title. Act like they never renamed it at                         // 270
              // all.                                                                                                  // 271
              ApiTokens.update({                                                                                       // 272
                grainId: grainId,                                                                                      // 273
                "owner.user.identityId": identityId                                                                    // 274
              }, {                                                                                                     // 272
                $set: {                                                                                                // 276
                  "owner.user.title": newTitle                                                                         // 276
                },                                                                                                     // 276
                $unset: {                                                                                              // 277
                  "owner.user.upstreamTitle": 1,                                                                       // 277
                  "owner.user.renamed": 1                                                                              // 277
                }                                                                                                      // 277
              }, {                                                                                                     // 275
                multi: true                                                                                            // 279
              });                                                                                                      // 278
            } else {                                                                                                   // 281
              var modification = {                                                                                     // 282
                "owner.user.title": newTitle,                                                                          // 283
                "owner.user.renamed": true                                                                             // 284
              };                                                                                                       // 282
                                                                                                                       //
              if (!token.owner.user.upstreamTitle) {                                                                   // 287
                // If `upstreamTitle` isn't present then it is equal to the old title.                                 // 288
                modification["owner.user.upstreamTitle"] = token.owner.user.title;                                     // 289
              }                                                                                                        // 290
                                                                                                                       //
              ApiTokens.update({                                                                                       // 292
                grainId: grainId,                                                                                      // 292
                "owner.user.identityId": identityId                                                                    // 292
              }, {                                                                                                     // 292
                $set: modification                                                                                     // 293
              }, {                                                                                                     // 293
                multi: true                                                                                            // 294
              });                                                                                                      // 294
            }                                                                                                          // 295
          }                                                                                                            // 296
        }                                                                                                              // 297
      }                                                                                                                // 298
    }                                                                                                                  // 299
  },                                                                                                                   // 300
  privatizeGrain: function (grainId) {                                                                                 // 302
    check(grainId, String);                                                                                            // 303
                                                                                                                       //
    if (this.userId) {                                                                                                 // 304
      Grains.update({                                                                                                  // 305
        _id: grainId,                                                                                                  // 305
        userId: this.userId                                                                                            // 305
      }, {                                                                                                             // 305
        $set: {                                                                                                        // 305
          "private": true                                                                                              // 305
        }                                                                                                              // 305
      });                                                                                                              // 305
    }                                                                                                                  // 306
  },                                                                                                                   // 307
  inviteUsersToGrain: function (_origin, identityId, grainId, title, roleAssignment, contacts, message) {              // 309
    if ((typeof message === "undefined" ? "undefined" : (0, _typeof3.default)(message)) === "object") {                // 311
      // Older versions of the client passed an object here, but we only care about the `text`                         // 312
      // parameter. (This block can eventually be removed.)                                                            // 313
      message = message.text;                                                                                          // 314
    }                                                                                                                  // 315
                                                                                                                       //
    if (!this.isSimulation) {                                                                                          // 317
      check(_origin, String);                                                                                          // 318
      check(identityId, String);                                                                                       // 319
      check(grainId, String);                                                                                          // 320
      check(title, String);                                                                                            // 321
      check(roleAssignment, roleAssignmentPattern);                                                                    // 322
      check(contacts, [{                                                                                               // 323
        _id: String,                                                                                                   // 325
        isDefault: Match.Optional(Boolean),                                                                            // 326
        profile: Match.ObjectIncluding({                                                                               // 327
          service: String,                                                                                             // 328
          name: String,                                                                                                // 329
          intrinsicName: String                                                                                        // 330
        })                                                                                                             // 327
      }]);                                                                                                             // 324
      check(message, String);                                                                                          // 334
                                                                                                                       //
      if (!this.userId) {                                                                                              // 335
        throw new Meteor.Error(403, "Must be logged in to share by email.");                                           // 336
      }                                                                                                                // 337
                                                                                                                       //
      if (!globalDb.userHasIdentity(this.userId, identityId)) {                                                        // 339
        throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);                             // 340
      }                                                                                                                // 341
                                                                                                                       //
      if (contacts.length === 0) {                                                                                     // 343
        throw new Meteor.Error(400, "No contacts were provided.");                                                     // 344
      }                                                                                                                // 345
                                                                                                                       //
      if (globalDb.isDemoUser()) {                                                                                     // 347
        throw new Meteor.Error(403, "Demo users are not allowed to share by email.");                                  // 348
      }                                                                                                                // 349
                                                                                                                       //
      var escapedMessage = message.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/>/g, "&gt;").replace(/"/g, "&quot;").replace(/\n/g, "<br>");
      var accountId = this.userId;                                                                                     // 357
      var outerResult = {                                                                                              // 358
        successes: [],                                                                                                 // 358
        failures: []                                                                                                   // 358
      };                                                                                                               // 358
      var fromEmail = globalDb.getReturnAddressWithDisplayName(identityId);                                            // 359
      var replyTo = globalDb.getPrimaryEmail(accountId, identityId);                                                   // 360
      contacts.forEach(function (contact) {                                                                            // 361
        if (contact.isDefault && contact.profile.service === "email") {                                                // 362
          var emailAddress = contact.profile.intrinsicName;                                                            // 363
          var result = SandstormPermissions.createNewApiToken(globalDb, {                                              // 364
            identityId: identityId,                                                                                    // 365
            accountId: accountId                                                                                       // 365
          }, grainId, "email invitation for " + emailAddress, roleAssignment, {                                        // 365
            webkey: {                                                                                                  // 367
              forSharing: true                                                                                         // 367
            }                                                                                                          // 367
          });                                                                                                          // 367
          var url = ROOT_URL + "/shared/" + result.token;                                                              // 368
          var html = escapedMessage + "<br><br>" + emailLinkWithInlineStyle(url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: If you forward this email to other people, they will be able to access " + "the share as well. To prevent this, remove the button before forwarding.</div>";
                                                                                                                       //
          try {                                                                                                        // 374
            globalDb.incrementDailySentMailCount(accountId);                                                           // 375
            sendEmail({                                                                                                // 376
              to: emailAddress,                                                                                        // 377
              from: fromEmail,                                                                                         // 378
              replyTo: replyTo,                                                                                        // 379
              subject: title + " - Invitation to collaborate",                                                         // 380
              text: message + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: If you forward this email to other people, they will be able to " + "access the share as well. To prevent this, remove the link before forwarding.",
              html: html                                                                                               // 384
            });                                                                                                        // 376
          } catch (e) {                                                                                                // 386
            outerResult.failures.push({                                                                                // 387
              contact: contact,                                                                                        // 387
              error: e.toString()                                                                                      // 387
            });                                                                                                        // 387
          }                                                                                                            // 388
        } else {                                                                                                       // 389
          var _result = SandstormPermissions.createNewApiToken(globalDb, {                                             // 390
            identityId: identityId,                                                                                    // 391
            accountId: accountId                                                                                       // 391
          }, grainId, "direct invitation to " + contact.profile.intrinsicName, roleAssignment, {                       // 391
            user: {                                                                                                    // 393
              identityId: contact._id,                                                                                 // 393
              title: title                                                                                             // 393
            }                                                                                                          // 393
          });                                                                                                          // 393
                                                                                                                       //
          var _url = ROOT_URL + "/shared/" + _result.token;                                                            // 394
                                                                                                                       //
          try {                                                                                                        // 395
            var identity = Meteor.users.findOne({                                                                      // 396
              _id: contact._id                                                                                         // 396
            });                                                                                                        // 396
                                                                                                                       //
            var email = _.findWhere(SandstormDb.getVerifiedEmails(identity), {                                         // 397
              primary: true                                                                                            // 398
            });                                                                                                        // 398
                                                                                                                       //
            if (email) {                                                                                               // 399
              var intrinsicName = contact.profile.intrinsicName;                                                       // 400
              var loginNote = void 0;                                                                                  // 401
                                                                                                                       //
              if (contact.profile.service === "google") {                                                              // 402
                loginNote = "Google account with address " + email.email;                                              // 403
              } else if (contact.profile.service === "github") {                                                       // 404
                loginNote = "Github account with username " + intrinsicName;                                           // 405
              } else if (contact.profile.service === "email") {                                                        // 406
                loginNote = "email address " + intrinsicName;                                                          // 407
              } else if (contact.profile.service === "ldap") {                                                         // 408
                loginNote = "LDAP username " + intrinsicName;                                                          // 409
              } else if (contact.profile.service === "saml") {                                                         // 410
                loginNote = "SAML ID " + intrinsicName;                                                                // 411
              } else {                                                                                                 // 412
                throw new Meteor.Error(500, "Unknown service to email share link.");                                   // 413
              }                                                                                                        // 414
                                                                                                                       //
              var _html = escapedMessage + "<br><br>" + emailLinkWithInlineStyle(_url, "Open Shared Grain") + "<div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: You will need to log in with your " + loginNote + " to access this grain.";
                                                                                                                       //
              globalDb.incrementDailySentMailCount(accountId);                                                         // 421
              sendEmail({                                                                                              // 422
                to: email.email,                                                                                       // 423
                from: fromEmail,                                                                                       // 424
                replyTo: replyTo,                                                                                      // 425
                subject: title + " - Invitation to collaborate",                                                       // 426
                text: message + "\n\nFollow this link to open the shared grain:\n\n" + _url + "\n\nNote: You will need to log in with your " + loginNote + " to access this grain.",
                html: _html                                                                                            // 430
              });                                                                                                      // 422
            } else {                                                                                                   // 432
              outerResult.failures.push({                                                                              // 433
                contact: contact,                                                                                      // 433
                warning: "User does not have a " + "verified email, so notification of this share was not sent to them. Please " + "manually share " + _url + " with them."
              });                                                                                                      // 433
            }                                                                                                          // 436
          } catch (e) {                                                                                                // 437
            outerResult.failures.push({                                                                                // 438
              contact: contact,                                                                                        // 438
              error: e.toString(),                                                                                     // 438
              warning: "Share succeeded, but there was an error emailing the user. Please " + "manually share " + _url + " with them."
            });                                                                                                        // 438
          }                                                                                                            // 441
        }                                                                                                              // 442
      });                                                                                                              // 443
      return outerResult;                                                                                              // 445
    }                                                                                                                  // 446
  },                                                                                                                   // 447
  requestAccess: function (_origin, grainId, identityId) {                                                             // 449
    check(_origin, String);                                                                                            // 450
    check(grainId, String);                                                                                            // 451
    check(identityId, String);                                                                                         // 452
                                                                                                                       //
    if (!this.isSimulation) {                                                                                          // 453
      if (!this.userId) {                                                                                              // 454
        throw new Meteor.Error(403, "Must be logged in to request access.");                                           // 455
      }                                                                                                                // 456
                                                                                                                       //
      if (!globalDb.userHasIdentity(this.userId, identityId)) {                                                        // 458
        throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);                             // 459
      }                                                                                                                // 460
                                                                                                                       //
      var grain = Grains.findOne(grainId);                                                                             // 462
                                                                                                                       //
      if (!grain) {                                                                                                    // 463
        throw new Meteor.Error(404, "No such grain");                                                                  // 464
      }                                                                                                                // 465
                                                                                                                       //
      var grainOwner = globalDb.getUser(grain.userId);                                                                 // 467
                                                                                                                       //
      var email = _.findWhere(SandstormDb.getUserEmails(grainOwner), {                                                 // 468
        primary: true                                                                                                  // 468
      });                                                                                                              // 468
                                                                                                                       //
      if (!email) {                                                                                                    // 469
        throw new Meteor.Error("no email", "Grain owner has no email address.");                                       // 470
      }                                                                                                                // 471
                                                                                                                       //
      var emailAddress = email.email;                                                                                  // 473
      var identity = globalDb.getIdentity(identityId);                                                                 // 475
      globalDb.addContact(grainOwner._id, identityId);                                                                 // 476
      var fromEmail = globalDb.getReturnAddressWithDisplayName(identityId);                                            // 478
      var replyTo = globalDb.getPrimaryEmail(Meteor.userId(), identityId); // TODO(soon): In the HTML version, we should display an identity card.
                                                                                                                       //
      var identityNote = "";                                                                                           // 482
                                                                                                                       //
      if (identity.profile.service === "google") {                                                                     // 483
        identityNote = " (" + identity.privateIntrinsicName + ")";                                                     // 484
      } else if (identity.profile.service === "github") {                                                              // 485
        identityNote = " (" + identity.profile.intrinsicName + " on GitHub)";                                          // 486
      } else if (identity.profile.service === "email") {                                                               // 487
        identityNote = " (" + identity.profile.intrinsicName + ")";                                                    // 488
      } else if (identity.profile.service === "ldap") {                                                                // 489
        identityNote = " (" + identity.profile.intrinsicName + " on LDAP)";                                            // 490
      } else if (identity.profile.service === "saml") {                                                                // 491
        identityNote = " (" + identity.profile.intrinsicName + " on SAML)";                                            // 492
      }                                                                                                                // 493
                                                                                                                       //
      var message = identity.profile.name + identityNote + " is requesting access to your grain: " + grain.title + ".";
      var url = ROOT_URL + "/share/" + grainId + "/" + identityId;                                                     // 498
      var html = message + "<br><br>" + emailLinkWithInlineStyle(url, "Open Sharing Menu");                            // 500
      var user = Meteor.user();                                                                                        // 503
      var ACCESS_REQUEST_LIMIT = 10;                                                                                   // 504
      var resetCount = true;                                                                                           // 505
                                                                                                                       //
      if (user.accessRequests) {                                                                                       // 506
        if (user.accessRequests.resetOn < new Date()) {                                                                // 507
          Meteor.users.update({                                                                                        // 508
            _id: user._id                                                                                              // 508
          }, {                                                                                                         // 508
            $unset: {                                                                                                  // 508
              accessRequests: 1                                                                                        // 508
            }                                                                                                          // 508
          });                                                                                                          // 508
        } else if (user.accessRequests.count >= ACCESS_REQUEST_LIMIT) {                                                // 509
          throw new Meteor.Error(403, "For spam control reasons, you are not allowed to make " + "more than " + ACCESS_REQUEST_LIMIT + " access requests per day.");
        } else {                                                                                                       // 513
          resetCount = false;                                                                                          // 514
        }                                                                                                              // 515
      }                                                                                                                // 516
                                                                                                                       //
      var modifier = {                                                                                                 // 518
        $inc: {                                                                                                        // 518
          "accessRequests.count": 1                                                                                    // 518
        }                                                                                                              // 518
      };                                                                                                               // 518
                                                                                                                       //
      if (resetCount) {                                                                                                // 519
        var tomorrow = new Date(Date.now() + 86400000);                                                                // 520
        modifier.$set = {                                                                                              // 521
          "accessRequests.resetOn": tomorrow                                                                           // 521
        };                                                                                                             // 521
      }                                                                                                                // 522
                                                                                                                       //
      Meteor.users.update({                                                                                            // 524
        _id: user._id                                                                                                  // 524
      }, modifier);                                                                                                    // 524
      sendEmail({                                                                                                      // 526
        to: emailAddress,                                                                                              // 527
        from: fromEmail,                                                                                               // 528
        replyTo: replyTo,                                                                                              // 529
        subject: grain.title + " - Request for access",                                                                // 530
        text: message + "\n\nFollow this link to share access:\n\n" + url,                                             // 531
        html: html                                                                                                     // 532
      });                                                                                                              // 526
    }                                                                                                                  // 534
  }                                                                                                                    // 535
});                                                                                                                    // 226
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"hack-session.js":["babel-runtime/helpers/possibleConstructorReturn","babel-runtime/helpers/inherits","babel-runtime/helpers/classCallCheck","/imports/server/persistent.js","/imports/server/async-helpers.js","/imports/server/networking.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/hack-session.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var hashSturdyRef = void 0,                                                                                            // 1
    checkRequirements = void 0,                                                                                        // 1
    fetchApiToken = void 0;                                                                                            // 1
module.import("/imports/server/persistent.js", {                                                                       // 1
  "hashSturdyRef": function (v) {                                                                                      // 1
    hashSturdyRef = v;                                                                                                 // 1
  },                                                                                                                   // 1
  "checkRequirements": function (v) {                                                                                  // 1
    checkRequirements = v;                                                                                             // 1
  },                                                                                                                   // 1
  "fetchApiToken": function (v) {                                                                                      // 1
    fetchApiToken = v;                                                                                                 // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var inMeteor = void 0,                                                                                                 // 1
    waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var ssrfSafeLookup = void 0;                                                                                           // 1
module.import("/imports/server/networking.js", {                                                                       // 1
  "ssrfSafeLookup": function (v) {                                                                                     // 1
    ssrfSafeLookup = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Crypto = Npm.require("crypto");                                                                                    // 17
                                                                                                                       //
var Http = Npm.require("http");                                                                                        // 18
                                                                                                                       //
var Https = Npm.require("https");                                                                                      // 19
                                                                                                                       //
var Net = Npm.require("net");                                                                                          // 20
                                                                                                                       //
var Dgram = Npm.require("dgram");                                                                                      // 21
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 22
                                                                                                                       //
var EmailRpc = Capnp.importSystem("sandstorm/email.capnp");                                                            // 27
var HackSessionContext = Capnp.importSystem("sandstorm/hack-session.capnp").HackSessionContext;                        // 28
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;                                          // 29
var SystemPersistent = Capnp.importSystem("sandstorm/supervisor.capnp").SystemPersistent;                              // 30
var IpRpc = Capnp.importSystem("sandstorm/ip.capnp");                                                                  // 31
var EmailSendPort = EmailRpc.EmailSendPort;                                                                            // 32
var Grain = Capnp.importSystem("sandstorm/grain.capnp");                                                               // 33
var Powerbox = Capnp.importSystem("sandstorm/powerbox.capnp");                                                         // 34
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 36
                                                                                                                       //
ROOT_URL = Url.parse(process.env.ROOT_URL);                                                                            // 38
HOSTNAME = ROOT_URL.hostname;                                                                                          // 39
                                                                                                                       //
SessionContextImpl = function () {                                                                                     // 41
  function SessionContextImpl(grainId, sessionId, identityId, tabId) {                                                 // 42
    (0, _classCallCheck3.default)(this, SessionContextImpl);                                                           // 42
    this.grainId = grainId;                                                                                            // 43
    this.sessionId = sessionId;                                                                                        // 44
    this.identityId = identityId;                                                                                      // 45
    this.tabId = tabId;                                                                                                // 46
  }                                                                                                                    // 47
                                                                                                                       //
  SessionContextImpl.prototype.claimRequest = function () {                                                            // 41
    function claimRequest(sturdyRef, requiredPermissions) {                                                            // 41
      var _this2 = this;                                                                                               // 49
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 50
        var token = fetchApiToken(globalDb, sturdyRef, {                                                               // 51
          "owner.clientPowerboxRequest.sessionId": _this2.sessionId                                                    // 52
        });                                                                                                            // 52
                                                                                                                       //
        if (!token) {                                                                                                  // 54
          throw new Error("no such token");                                                                            // 55
        }                                                                                                              // 56
                                                                                                                       //
        var session = Sessions.findOne({                                                                               // 58
          _id: _this2.sessionId                                                                                        // 58
        });                                                                                                            // 58
                                                                                                                       //
        if (!session) {                                                                                                // 60
          throw new Error("no such session");                                                                          // 61
        } // Honor `requiredPermissions`.                                                                              // 62
                                                                                                                       //
                                                                                                                       //
        var requirements = [];                                                                                         // 65
                                                                                                                       //
        if (session.hashedToken) {                                                                                     // 66
          // Session is authorized by token. Note that it's important to check this before identityId                  // 67
          // e.g. in the case of standalone domains.                                                                   // 68
          requirements.push({                                                                                          // 69
            permissionsHeld: {                                                                                         // 70
              permissions: requiredPermissions || [],                                                                  // 71
              tokenId: session.hashedToken,                                                                            // 72
              grainId: _this2.grainId                                                                                  // 73
            }                                                                                                          // 70
          });                                                                                                          // 69
        } else if (session.identityId) {                                                                               // 76
          // Session is authorized by identity.                                                                        // 77
          requirements.push({                                                                                          // 78
            permissionsHeld: {                                                                                         // 79
              permissions: requiredPermissions || [],                                                                  // 80
              identityId: session.identityId,                                                                          // 81
              grainId: _this2.grainId                                                                                  // 82
            }                                                                                                          // 79
          });                                                                                                          // 78
        } else {                                                                                                       // 85
          // This can only happen with old-style sharing which has been deprecated for years. If                       // 86
          // we wanted to support this, I suppose we could do so as long as requiredPermissions is                     // 87
          // all-false? But probably this will never come up.                                                          // 88
          throw new Error("Cannot accept powerbox request from anonymous session that " + "doesn't have a token.");    // 89
        }                                                                                                              // 91
                                                                                                                       //
        return restoreInternal(globalDb, sturdyRef, {                                                                  // 93
          clientPowerboxRequest: Match.ObjectIncluding({                                                               // 95
            sessionId: _this2.sessionId                                                                                // 95
          })                                                                                                           // 95
        }, requirements, token);                                                                                       // 95
      });                                                                                                              // 97
    }                                                                                                                  // 98
                                                                                                                       //
    return claimRequest;                                                                                               // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  SessionContextImpl.prototype._offerOrFulfill = function () {                                                         // 41
    function _offerOrFulfill(isFulfill, cap, requiredPermissions, descriptor, displayInfo) {                           // 41
      var _this3 = this;                                                                                               // 100
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 101
        var session = Sessions.findOne({                                                                               // 103
          _id: _this3.sessionId                                                                                        // 103
        });                                                                                                            // 103
                                                                                                                       //
        if (!session.identityId && !session.hashedToken) {                                                             // 105
          throw new Error("Session has neither an identityId nor a hashedToken.");                                     // 106
        }                                                                                                              // 107
                                                                                                                       //
        if (isFulfill && !session.powerboxRequest) {                                                                   // 109
          // Not a request session, so treat fulfillRequest() same as offer().                                         // 110
          isFulfill = false;                                                                                           // 111
        }                                                                                                              // 112
                                                                                                                       //
        var castedCap = cap.castAs(SystemPersistent);                                                                  // 114
        var apiTokenOwner = {                                                                                          // 115
          clientPowerboxOffer: {                                                                                       // 115
            sessionId: _this3.sessionId                                                                                // 115
          }                                                                                                            // 115
        };                                                                                                             // 115
        var isUiView = descriptor && descriptor.tags && descriptor.tags.length === 1 && descriptor.tags[0] && descriptor.tags[0].id && descriptor.tags[0].id === Grain.UiView.typeId;
                                                                                                                       //
        if (isUiView) {                                                                                                // 119
          var tagValue = {};                                                                                           // 120
                                                                                                                       //
          if (descriptor.tags[0].value) {                                                                              // 121
            tagValue = Capnp.parse(Grain.UiView.PowerboxTag, descriptor.tags[0].value);                                // 122
          }                                                                                                            // 123
                                                                                                                       //
          if (!tagValue.title) {                                                                                       // 125
            throw new Error("No value provided for UiView.PowerboxTag.title.");                                        // 126
          }                                                                                                            // 127
                                                                                                                       //
          if (session.identityId) {                                                                                    // 129
            apiTokenOwner = {                                                                                          // 130
              user: {                                                                                                  // 131
                identityId: _this3.identityId,                                                                         // 132
                title: tagValue.title                                                                                  // 133
              }                                                                                                        // 131
            };                                                                                                         // 130
          }                                                                                                            // 136
        }                                                                                                              // 137
                                                                                                                       //
        if (isFulfill) {                                                                                               // 139
          // The capability will pass directly to the requesting grain.                                                // 140
          apiTokenOwner = {                                                                                            // 141
            clientPowerboxRequest: {                                                                                   // 142
              sessionId: session.powerboxRequest.requestingSession                                                     // 143
            }                                                                                                          // 142
          };                                                                                                           // 141
        } // TODO(soon): This will eventually use SystemPersistent.addRequirements when membranes                      // 146
        // are fully implemented for supervisors.                                                                      // 149
                                                                                                                       //
                                                                                                                       //
        var permissionsHeld = {                                                                                        // 150
          grainId: _this3.grainId,                                                                                     // 151
          permissions: requiredPermissions                                                                             // 152
        }; // Note that if both hashedToken and identityId are present, this indicates that the session                // 150
        // is authorized by a token which the user has NOT redeemed. Thus the identity does not                        // 156
        // directly have access to the grain; the token does. (This is the case in particular for                      // 157
        // standalone domains.) So clearly we want our requirement to be based on the token, not the                   // 158
        // identity.                                                                                                   // 159
                                                                                                                       //
        if (session.hashedToken) {                                                                                     // 160
          permissionsHeld.tokenId = session.hashedToken;                                                               // 161
        } else if (session.identityId) {                                                                               // 162
          permissionsHeld.identityId = session.identityId;                                                             // 163
        } else {                                                                                                       // 164
          throw new Error("Cannot offer to anonymous session that does not have a token.");                            // 165
        }                                                                                                              // 166
                                                                                                                       //
        var requirement = {                                                                                            // 168
          permissionsHeld: permissionsHeld                                                                             // 168
        };                                                                                                             // 168
        checkRequirements(globalDb, [requirement]);                                                                    // 170
        var save = castedCap.save(apiTokenOwner);                                                                      // 172
        var sturdyRef = waitPromise(save).sturdyRef;                                                                   // 173
        ApiTokens.update({                                                                                             // 174
          _id: hashSturdyRef(sturdyRef)                                                                                // 174
        }, {                                                                                                           // 174
          $push: {                                                                                                     // 174
            requirements: requirement                                                                                  // 174
          }                                                                                                            // 174
        });                                                                                                            // 174
        var powerboxView = void 0;                                                                                     // 176
                                                                                                                       //
        if (isFulfill) {                                                                                               // 177
          powerboxView = {                                                                                             // 178
            fulfill: {                                                                                                 // 179
              token: sturdyRef.toString(),                                                                             // 180
              descriptor: Capnp.serializePacked(Powerbox.PowerboxDescriptor, descriptor).toString("base64")            // 181
            }                                                                                                          // 179
          };                                                                                                           // 178
        } else if (isUiView) {                                                                                         // 185
          if (session.identityId) {                                                                                    // 186
            // Deduplicate.                                                                                            // 187
            var newApiToken = fetchApiToken(globalDb, sturdyRef.toString());                                           // 188
            var tokenId = newApiToken._id;                                                                             // 189
                                                                                                                       //
            var dupeQuery = _.pick(newApiToken, "grainId", "roleAssignment", "requirements", "parentToken", "parentTokenKey", "identityId", "accountId");
                                                                                                                       //
            dupeQuery._id = {                                                                                          // 192
              $ne: newApiToken._id                                                                                     // 192
            };                                                                                                         // 192
            dupeQuery["owner.user.identityId"] = _this3.identityId;                                                    // 193
            dupeQuery.trashed = {                                                                                      // 194
              $exists: false                                                                                           // 194
            };                                                                                                         // 194
            dupeQuery.revoked = {                                                                                      // 195
              $exists: false                                                                                           // 195
            };                                                                                                         // 195
            var dupeToken = ApiTokens.findOne(dupeQuery);                                                              // 197
                                                                                                                       //
            if (dupeToken) {                                                                                           // 198
              globalDb.removeApiTokens({                                                                               // 199
                _id: tokenId                                                                                           // 199
              });                                                                                                      // 199
              tokenId = dupeToken._id;                                                                                 // 200
            }                                                                                                          // 201
                                                                                                                       //
            powerboxView = {                                                                                           // 203
              offer: {                                                                                                 // 203
                uiView: {                                                                                              // 203
                  tokenId: tokenId                                                                                     // 203
                }                                                                                                      // 203
              }                                                                                                        // 203
            };                                                                                                         // 203
          } else {                                                                                                     // 204
            powerboxView = {                                                                                           // 205
              offer: {                                                                                                 // 205
                uiView: {                                                                                              // 205
                  token: sturdyRef.toString()                                                                          // 205
                }                                                                                                      // 205
              }                                                                                                        // 205
            };                                                                                                         // 205
          }                                                                                                            // 206
        } else {                                                                                                       // 207
          powerboxView = {                                                                                             // 208
            offer: {                                                                                                   // 209
              token: sturdyRef.toString()                                                                              // 210
            }                                                                                                          // 209
          };                                                                                                           // 208
        }                                                                                                              // 213
                                                                                                                       //
        Sessions.update({                                                                                              // 215
          _id: _this3.sessionId                                                                                        // 215
        }, {                                                                                                           // 215
          $set: {                                                                                                      // 217
            powerboxView: powerboxView                                                                                 // 218
          }                                                                                                            // 217
        });                                                                                                            // 216
      });                                                                                                              // 222
    }                                                                                                                  // 223
                                                                                                                       //
    return _offerOrFulfill;                                                                                            // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  SessionContextImpl.prototype.offer = function () {                                                                   // 41
    function offer(cap, requiredPermissions, descriptor, displayInfo) {                                                // 41
      return this._offerOrFulfill(false, cap, requiredPermissions, descriptor, displayInfo);                           // 226
    }                                                                                                                  // 227
                                                                                                                       //
    return offer;                                                                                                      // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  SessionContextImpl.prototype.fulfillRequest = function () {                                                          // 41
    function fulfillRequest(cap, requiredPermissions, descriptor, displayInfo) {                                       // 41
      return this._offerOrFulfill(true, cap, requiredPermissions, descriptor, displayInfo);                            // 230
    }                                                                                                                  // 231
                                                                                                                       //
    return fulfillRequest;                                                                                             // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  SessionContextImpl.prototype.activity = function () {                                                                // 41
    function activity(event) {                                                                                         // 41
      var _this4 = this;                                                                                               // 233
                                                                                                                       //
      return inMeteor(function () {                                                                                    // 234
        logActivity(_this4.grainId, _this4.identityId || "anonymous", event);                                          // 235
      });                                                                                                              // 236
    }                                                                                                                  // 237
                                                                                                                       //
    return activity;                                                                                                   // 41
  }();                                                                                                                 // 41
                                                                                                                       //
  return SessionContextImpl;                                                                                           // 41
}();                                                                                                                   // 41
                                                                                                                       //
Meteor.methods({                                                                                                       // 240
  finishPowerboxRequest: function (sessionId, webkeyUrl, saveLabel, identityId, grainId) {                             // 241
    check(sessionId, String);                                                                                          // 242
    check(webkeyUrl, String);                                                                                          // 243
    check(saveLabel, Match.OneOf(undefined, null, String));                                                            // 244
    check(identityId, String);                                                                                         // 245
    check(grainId, String);                                                                                            // 246
    var db = this.connection.sandstormDb;                                                                              // 248
    var userId = Meteor.userId();                                                                                      // 250
                                                                                                                       //
    if (!userId || !db.userHasIdentity(userId, identityId)) {                                                          // 251
      throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);                               // 252
    }                                                                                                                  // 253
                                                                                                                       //
    var parsedWebkey = Url.parse(webkeyUrl.trim());                                                                    // 255
    var hostId = matchWildcardHost(parsedWebkey.host);                                                                 // 256
                                                                                                                       //
    if (!hostId || !db.isApiHostId(hostId)) {                                                                          // 257
      throw new Meteor.Error(500, "Hostname does not match this server. External webkeys are not " + "supported (yet)");
    }                                                                                                                  // 260
                                                                                                                       //
    var token = parsedWebkey.hash;                                                                                     // 262
                                                                                                                       //
    if (!token) {                                                                                                      // 263
      throw new Meteor.Error(400, "Invalid webkey. You must pass in the full webkey, " + "including domain name and hash fragment");
    }                                                                                                                  // 266
                                                                                                                       //
    token = token.slice(1);                                                                                            // 268
                                                                                                                       //
    if (db.isTokenSpecificHostId(hostId) && hostId !== db.apiHostIdForToken(token)) {                                  // 269
      // Note: This case is not security-sensitive. The client could easily compute the correct                        // 270
      //   host ID for this token. But since they've passed one that doesn't match, we assume                          // 271
      //   something is wrong and stop here.                                                                           // 272
      throw new Meteor.Error(400, "Invalid webkey: token doesn't match hostname.");                                    // 273
    }                                                                                                                  // 274
                                                                                                                       //
    var cap = restoreInternal(db, token, Match.Optional({                                                              // 276
      webkey: Match.Optional(Match.Any)                                                                                // 277
    }), []).cap;                                                                                                       // 277
    var castedCap = cap.castAs(SystemPersistent);                                                                      // 278
    var owner = {                                                                                                      // 279
      clientPowerboxRequest: {                                                                                         // 280
        grainId: grainId,                                                                                              // 281
        introducerIdentity: identityId,                                                                                // 282
        sessionId: sessionId                                                                                           // 283
      }                                                                                                                // 280
    };                                                                                                                 // 279
                                                                                                                       //
    if (saveLabel) {                                                                                                   // 286
      grainOwner.saveLabel = {                                                                                         // 287
        defaultText: saveLabel                                                                                         // 288
      };                                                                                                               // 287
    }                                                                                                                  // 290
                                                                                                                       //
    var save = castedCap.save(owner);                                                                                  // 292
    var sturdyRef = waitPromise(save).sturdyRef;                                                                       // 293
    return sturdyRef.toString();                                                                                       // 294
  },                                                                                                                   // 295
  finishPowerboxOffer: function (sessionId) {                                                                          // 297
    check(sessionId, String);                                                                                          // 298
    Sessions.update({                                                                                                  // 300
      _id: sessionId                                                                                                   // 300
    }, {                                                                                                               // 300
      $unset: {                                                                                                        // 300
        powerboxView: null                                                                                             // 300
      }                                                                                                                // 300
    });                                                                                                                // 300
  },                                                                                                                   // 301
  getViewInfoForApiToken: function (apiTokenId) {                                                                      // 303
    check(apiTokenId, String);                                                                                         // 304
    var token = globalDb.collections.apiTokens.findOne(apiTokenId);                                                    // 305
                                                                                                                       //
    if (!token) {                                                                                                      // 306
      throw new Meteor.Error(400, "No such token.");                                                                   // 307
    }                                                                                                                  // 308
                                                                                                                       //
    var grain = globalDb.collections.grains.findOne(token.grainId);                                                    // 310
                                                                                                                       //
    if (!grain) {                                                                                                      // 311
      throw new Meteor.Error(500, "No grain found for token.");                                                        // 312
    }                                                                                                                  // 313
                                                                                                                       //
    return grain.cachedViewInfo;                                                                                       // 315
  }                                                                                                                    // 316
});                                                                                                                    // 240
                                                                                                                       //
HackSessionContextImpl = function (_SessionContextImpl) {                                                              // 319
  (0, _inherits3.default)(HackSessionContextImpl, _SessionContextImpl);                                                // 319
                                                                                                                       //
  function HackSessionContextImpl(grainId, sessionId, identityId, tabId) {                                             // 320
    (0, _classCallCheck3.default)(this, HackSessionContextImpl);                                                       // 320
    return (0, _possibleConstructorReturn3.default)(this, _SessionContextImpl.call(this, grainId, sessionId, identityId, tabId));
  }                                                                                                                    // 322
                                                                                                                       //
  HackSessionContextImpl.prototype._getPublicId = function () {                                                        // 319
    function _getPublicId() {                                                                                          // 319
      // Get the grain's public ID, assigning a new one if it doesn't yet have one.                                    // 325
      //                                                                                                               // 326
      // Must be called in a Meteor context.                                                                           // 327
      while (!this.publicId) {                                                                                         // 329
        // We haven't looked up the public ID yet.                                                                     // 330
        var grain = Grains.findOne(this.grainId, {                                                                     // 331
          fields: {                                                                                                    // 331
            publicId: 1                                                                                                // 331
          }                                                                                                            // 331
        });                                                                                                            // 331
        if (!grain) throw new Error("Grain does not exist.");                                                          // 332
                                                                                                                       //
        if (grain.publicId) {                                                                                          // 334
          this.publicId = grain.publicId;                                                                              // 335
        } else {                                                                                                       // 336
          // The grain doesn't have a public ID yet. Generate one.                                                     // 337
          var candidate = generateRandomHostname(20);                                                                  // 338
                                                                                                                       //
          if (Grains.findOne({                                                                                         // 340
            publicId: candidate                                                                                        // 340
          })) {                                                                                                        // 340
            // This should never ever happen.                                                                          // 341
            console.error("CRITICAL PROBLEM: Public ID collision. " + "CSPRNG is bad or has insufficient entropy.");   // 342
            continue;                                                                                                  // 344
          } // Carefully perform an update that becomes a no-op if anyone else has assigned a public ID                // 345
          // simultaneously.                                                                                           // 348
                                                                                                                       //
                                                                                                                       //
          if (Grains.update({                                                                                          // 349
            _id: this.grainId,                                                                                         // 349
            publicId: {                                                                                                // 349
              $exists: false                                                                                           // 349
            }                                                                                                          // 349
          }, {                                                                                                         // 349
            $set: {                                                                                                    // 350
              publicId: candidate                                                                                      // 350
            }                                                                                                          // 350
          }) > 0) {                                                                                                    // 350
            // We won the race.                                                                                        // 351
            this.publicId = candidate;                                                                                 // 352
          }                                                                                                            // 353
        }                                                                                                              // 354
      }                                                                                                                // 355
                                                                                                                       //
      return this.publicId;                                                                                            // 357
    }                                                                                                                  // 358
                                                                                                                       //
    return _getPublicId;                                                                                               // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype._getAddress = function () {                                                         // 319
    function _getAddress() {                                                                                           // 319
      // Get the grain's outgoing e-mail address.                                                                      // 361
      //                                                                                                               // 362
      // Must be called in a Meteor context.                                                                           // 363
      return this._getPublicId() + "@" + HOSTNAME;                                                                     // 365
    }                                                                                                                  // 366
                                                                                                                       //
    return _getAddress;                                                                                                // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype._getUserAddress = function () {                                                     // 319
    function _getUserAddress() {                                                                                       // 319
      // Get the user's e-mail address.                                                                                // 369
      //                                                                                                               // 370
      // Must be called in a Meteor context.                                                                           // 371
      var grain = Grains.findOne(this.grainId, {                                                                       // 373
        fields: {                                                                                                      // 373
          identityId: 1                                                                                                // 373
        }                                                                                                              // 373
      });                                                                                                              // 373
      var identity = globalDb.getIdentity(grain.identityId);                                                           // 375
                                                                                                                       //
      var primaryEmail = _.findWhere(SandstormDb.getVerifiedEmails(identity), {                                        // 377
        primary: true                                                                                                  // 377
      });                                                                                                              // 377
                                                                                                                       //
      var email = primaryEmail && primaryEmail.email || identity.unverifiedEmail;                                      // 378
      var result = {};                                                                                                 // 380
                                                                                                                       //
      if (email) {                                                                                                     // 381
        result.address = email;                                                                                        // 382
      }                                                                                                                // 383
                                                                                                                       //
      if (identity.profile.name) {                                                                                     // 385
        result.name = identity.profile.name;                                                                           // 386
      }                                                                                                                // 387
                                                                                                                       //
      return result;                                                                                                   // 389
    }                                                                                                                  // 390
                                                                                                                       //
    return _getUserAddress;                                                                                            // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype.send = function () {                                                                // 319
    function send(email) {                                                                                             // 319
      return hackSendEmail(this, email);                                                                               // 393
    }                                                                                                                  // 394
                                                                                                                       //
    return send;                                                                                                       // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype.getPublicId = function () {                                                         // 319
    function getPublicId() {                                                                                           // 319
      return inMeteor(function () {                                                                                    // 397
        var result = {};                                                                                               // 398
        result.publicId = this._getPublicId();                                                                         // 400
        result.hostname = HOSTNAME;                                                                                    // 401
        result.autoUrl = ROOT_URL.protocol + "//" + makeWildcardHost(result.publicId);                                 // 402
        var grain = Grains.findOne(this.grainId, {                                                                     // 404
          fields: {                                                                                                    // 404
            userId: 1                                                                                                  // 404
          }                                                                                                            // 404
        });                                                                                                            // 404
        result.isDemoUser = Meteor.users.findOne(grain.userId).expires ? true : false;                                 // 405
        return result;                                                                                                 // 407
      }.bind(this));                                                                                                   // 408
    }                                                                                                                  // 409
                                                                                                                       //
    return getPublicId;                                                                                                // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype.httpGet = function () {                                                             // 319
    function httpGet(url) {                                                                                            // 319
      var _this = this;                                                                                                // 412
                                                                                                                       //
      var session = _this;                                                                                             // 413
      return inMeteor(function () {                                                                                    // 415
        return ssrfSafeLookup(globalDb, url);                                                                          // 416
      }).then(function (safe) {                                                                                        // 417
        return new Promise(function (resolve, reject) {                                                                // 418
          var requestMethod = Http.request;                                                                            // 419
                                                                                                                       //
          if (safe.url.indexOf("https://") === 0) {                                                                    // 420
            requestMethod = Https.request;                                                                             // 421
          } else if (safe.url.indexOf("http://") !== 0) {                                                              // 422
            err = new Error("Protocol not recognized.");                                                               // 423
            err.nature = "precondition";                                                                               // 424
            reject(err);                                                                                               // 425
          }                                                                                                            // 426
                                                                                                                       //
          var options = Url.parse(safe.url);                                                                           // 428
          options.headers = {                                                                                          // 429
            host: safe.host                                                                                            // 429
          };                                                                                                           // 429
          options.servername = safe.host.split(":")[0];                                                                // 430
          req = requestMethod(options, function (resp) {                                                               // 432
            var buffers = [];                                                                                          // 433
            var err = void 0;                                                                                          // 434
                                                                                                                       //
            switch (Math.floor(resp.statusCode / 100)) {                                                               // 436
              case 2:                                                                                                  // 437
                // 2xx response -- OK.                                                                                 // 437
                resp.on("data", function (buf) {                                                                       // 438
                  buffers.push(buf);                                                                                   // 439
                });                                                                                                    // 440
                resp.on("end", function () {                                                                           // 442
                  resolve({                                                                                            // 443
                    content: Buffer.concat(buffers),                                                                   // 444
                    mimeType: resp.headers["content-type"] || null                                                     // 445
                  });                                                                                                  // 443
                });                                                                                                    // 447
                break;                                                                                                 // 448
                                                                                                                       //
              case 3:                                                                                                  // 449
                // 3xx response -- redirect.                                                                           // 449
                resolve(session.httpGet(resp.headers.location));                                                       // 450
                break;                                                                                                 // 451
                                                                                                                       //
              case 4:                                                                                                  // 452
                // 4xx response -- client error.                                                                       // 452
                err = new Error("Status code " + resp.statusCode + " received in response.");                          // 453
                err.nature = "precondition";                                                                           // 454
                reject(err);                                                                                           // 455
                break;                                                                                                 // 456
                                                                                                                       //
              case 5:                                                                                                  // 457
                // 5xx response -- internal server error.                                                              // 457
                err = new Error("Status code " + resp.statusCode + " received in response.");                          // 458
                err.nature = "localBug";                                                                               // 459
                reject(err);                                                                                           // 460
                break;                                                                                                 // 461
                                                                                                                       //
              default:                                                                                                 // 462
                // ???                                                                                                 // 462
                err = new Error("Invalid status code " + resp.statusCode + " received in response.");                  // 463
                err.nature = "localBug";                                                                               // 464
                reject(err);                                                                                           // 465
                break;                                                                                                 // 466
            }                                                                                                          // 436
          });                                                                                                          // 468
          req.on("error", function (e) {                                                                               // 470
            e.nature = "networkFailure";                                                                               // 471
            reject(e);                                                                                                 // 472
          });                                                                                                          // 473
          req.setTimeout(15000, function () {                                                                          // 475
            req.abort();                                                                                               // 476
            err = new Error("Request timed out.");                                                                     // 477
            err.nature = "localBug";                                                                                   // 478
            err.durability = "overloaded";                                                                             // 479
            reject(err);                                                                                               // 480
          });                                                                                                          // 481
          req.end();                                                                                                   // 483
        });                                                                                                            // 484
      });                                                                                                              // 485
    }                                                                                                                  // 486
                                                                                                                       //
    return httpGet;                                                                                                    // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype.getUserAddress = function () {                                                      // 319
    function getUserAddress() {                                                                                        // 319
      return inMeteor(function () {                                                                                    // 489
        return this._getUserAddress();                                                                                 // 490
      }.bind(this));                                                                                                   // 491
    }                                                                                                                  // 492
                                                                                                                       //
    return getUserAddress;                                                                                             // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype.obsoleteGenerateApiToken = function () {                                            // 319
    function obsoleteGenerateApiToken(petname, userInfo, expires) {                                                    // 319
      throw new Error("generateApiToken() has been removed. Use offer templates instead.");                            // 495
    }                                                                                                                  // 496
                                                                                                                       //
    return obsoleteGenerateApiToken;                                                                                   // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype.obsoleteListApiTokens = function () {                                               // 319
    function obsoleteListApiTokens() {                                                                                 // 319
      throw new Error("listApiTokens() has been removed. Use offer templates instead.");                               // 499
    }                                                                                                                  // 500
                                                                                                                       //
    return obsoleteListApiTokens;                                                                                      // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype.obsoleteRevokeApiToken = function () {                                              // 319
    function obsoleteRevokeApiToken(tokenId) {                                                                         // 319
      throw new Error("revokeApiToken() has been removed. Use offer templates instead.");                              // 503
    }                                                                                                                  // 504
                                                                                                                       //
    return obsoleteRevokeApiToken;                                                                                     // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  HackSessionContextImpl.prototype.getUiViewForEndpoint = function () {                                                // 319
    function getUiViewForEndpoint(url) {                                                                               // 319
      var parsedUrl = Url.parse(url);                                                                                  // 507
                                                                                                                       //
      if (parsedUrl.hash) {                                                                                            // 509
        // Assume that anything with a fragment is a webkey                                                            // 509
        if (parsedUrl.pathname && parsedUrl.pathname !== "/") {                                                        // 510
          throw new Error("Webkey urls cannot contain a path.");                                                       // 511
        }                                                                                                              // 512
                                                                                                                       //
        var token = parsedUrl.hash.slice(1); // Get rid of # which is always the first character                       // 514
                                                                                                                       //
        var hostId = matchWildcardHost(parsedUrl.host); // Connecting to a remote server with a bearer token.          // 515
        // TODO(someday): Negotiate server-to-server Cap'n Proto connection.                                           // 517
                                                                                                                       //
        return {                                                                                                       // 518
          view: new ExternalUiView(url, token)                                                                         // 518
        };                                                                                                             // 518
      } else {                                                                                                         // 519
        return {                                                                                                       // 520
          view: new ExternalUiView(url)                                                                                // 520
        };                                                                                                             // 520
      }                                                                                                                // 521
    }                                                                                                                  // 522
                                                                                                                       //
    return getUiViewForEndpoint;                                                                                       // 319
  }();                                                                                                                 // 319
                                                                                                                       //
  return HackSessionContextImpl;                                                                                       // 319
}(SessionContextImpl);                                                                                                 // 319
                                                                                                                       //
makeHackSessionContext = function (grainId, sessionId, identityId, tabId) {                                            // 525
  return new Capnp.Capability(new HackSessionContextImpl(grainId, sessionId, identityId, tabId), HackSessionContext);  // 526
};                                                                                                                     // 528
                                                                                                                       //
var HOSTNAME_CHARS = "abcdefghijklmnopqrstuvwxyz0123456789";                                                           // 530
                                                                                                                       //
generateRandomHostname = function (length) {                                                                           // 532
  // Generate a random unique name suitable for use in a hostname.                                                     // 533
  var digits = [];                                                                                                     // 535
                                                                                                                       //
  for (var i = 0; i < length; i++) {                                                                                   // 536
    digits[i] = Random.choice(HOSTNAME_CHARS);                                                                         // 537
  }                                                                                                                    // 538
                                                                                                                       //
  return digits.join("");                                                                                              // 540
};                                                                                                                     // 541
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"identity.js":["babel-runtime/helpers/classCallCheck","babel-runtime/helpers/possibleConstructorReturn","babel-runtime/helpers/inherits","/imports/server/persistent.js","/imports/server/static-asset.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/identity.js                                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
var _possibleConstructorReturn2 = require("babel-runtime/helpers/possibleConstructorReturn");                          //
                                                                                                                       //
var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);                                 //
                                                                                                                       //
var _inherits2 = require("babel-runtime/helpers/inherits");                                                            //
                                                                                                                       //
var _inherits3 = _interopRequireDefault(_inherits2);                                                                   //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var PersistentImpl = void 0;                                                                                           // 1
module.import("/imports/server/persistent.js", {                                                                       // 1
  "PersistentImpl": function (v) {                                                                                     // 1
    PersistentImpl = v;                                                                                                // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var StaticAssetImpl = void 0,                                                                                          // 1
    IdenticonStaticAssetImpl = void 0;                                                                                 // 1
module.import("/imports/server/static-asset.js", {                                                                     // 1
  "StaticAssetImpl": function (v) {                                                                                    // 1
    StaticAssetImpl = v;                                                                                               // 1
  },                                                                                                                   // 1
  "IdenticonStaticAssetImpl": function (v) {                                                                           // 1
    IdenticonStaticAssetImpl = v;                                                                                      // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var Capnp = Npm.require("capnp");                                                                                      // 17
                                                                                                                       //
var IdentityRpc = Capnp.importSystem("sandstorm/identity-impl.capnp");                                                 // 18
var Identity = Capnp.importSystem("sandstorm/identity.capnp").Identity;                                                // 19
var StaticAsset = Capnp.importSystem("sandstorm/util.capnp").StaticAsset;                                              // 22
                                                                                                                       //
var IdentityImpl = function (_PersistentImpl) {                                                                        //
  (0, _inherits3.default)(IdentityImpl, _PersistentImpl);                                                              //
                                                                                                                       //
  function IdentityImpl(db, saveTemplate, identityId) {                                                                // 25
    (0, _classCallCheck3.default)(this, IdentityImpl);                                                                 // 25
                                                                                                                       //
    var _this = (0, _possibleConstructorReturn3.default)(this, _PersistentImpl.call(this, db, saveTemplate));          // 25
                                                                                                                       //
    _this.identityId = identityId;                                                                                     // 27
    _this.db = db;                                                                                                     // 28
    return _this;                                                                                                      // 25
  }                                                                                                                    // 29
                                                                                                                       //
  IdentityImpl.prototype.getProfile = function () {                                                                    //
    function getProfile() {                                                                                            //
      var identity = this.db.getIdentity(this.identityId);                                                             // 32
      var profile = {                                                                                                  // 34
        displayName: {                                                                                                 // 35
          defaultText: identity.profile.name                                                                           // 35
        },                                                                                                             // 35
        preferredHandle: identity.profile.handle,                                                                      // 36
        pronouns: identity.profile.pronoun                                                                             // 37
      };                                                                                                               // 34
                                                                                                                       //
      if (identity.profile.picture) {                                                                                  // 40
        profile.picture = new Capnp.Capability(new StaticAssetImpl(identity.profile.picture), StaticAsset);            // 41
      } else {                                                                                                         // 43
        var hash = this.identityId.slice(0, 32);                                                                       // 44
        profile.picture = new Capnp.Capability(new IdenticonStaticAssetImpl(hash, 24), StaticAsset);                   // 45
      }                                                                                                                // 47
                                                                                                                       //
      return {                                                                                                         // 49
        profile: profile                                                                                               // 49
      };                                                                                                               // 49
    }                                                                                                                  // 50
                                                                                                                       //
    return getProfile;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  return IdentityImpl;                                                                                                 //
}(PersistentImpl);                                                                                                     //
                                                                                                                       //
; // TODO(cleanup): Find a better home for this.                                                                       // 51
                                                                                                                       //
var MembraneRequirement = Match.OneOf({                                                                                // 54
  tokenValid: String                                                                                                   // 55
}, {                                                                                                                   // 55
  permissionsHeld: {                                                                                                   // 56
    identityId: String,                                                                                                // 57
    grainId: String,                                                                                                   // 57
    permissions: Match.Optional([Boolean])                                                                             // 57
  }                                                                                                                    // 57
}, {                                                                                                                   // 56
  permissionsHeld: {                                                                                                   // 58
    tokenId: String,                                                                                                   // 59
    grainId: String,                                                                                                   // 59
    permissions: Match.Optional([Boolean])                                                                             // 59
  }                                                                                                                    // 59
}, {                                                                                                                   // 58
  userIsAdmin: String                                                                                                  // 60
});                                                                                                                    // 60
                                                                                                                       //
makeIdentity = function (identityId, requirements) {                                                                   // 62
  var saveTemplate = {                                                                                                 // 63
    frontendRef: {                                                                                                     // 63
      identity: identityId                                                                                             // 63
    }                                                                                                                  // 63
  };                                                                                                                   // 63
                                                                                                                       //
  if (requirements) {                                                                                                  // 64
    check(requirements, [MembraneRequirement]);                                                                        // 65
    saveTemplate.requirements = requirements;                                                                          // 66
  }                                                                                                                    // 67
                                                                                                                       //
  return new Capnp.Capability(new IdentityImpl(globalDb, saveTemplate, identityId), IdentityRpc.PersistentIdentity);   // 69
};                                                                                                                     // 71
                                                                                                                       //
globalFrontendRefRegistry.register({                                                                                   // 73
  frontendRefField: "identity",                                                                                        // 74
  typeId: Identity.typeId,                                                                                             // 75
  restore: function (db, saveTemplate, identityId) {                                                                   // 77
    return new Capnp.Capability(new IdentityImpl(db, saveTemplate, identityId), IdentityRpc.PersistentIdentity);       // 78
  },                                                                                                                   // 80
  validate: function (db, session, value) {                                                                            // 82
    check(value, {                                                                                                     // 83
      id: String,                                                                                                      // 83
      roleAssignment: SandstormDb.prototype.roleAssignmentPattern                                                      // 83
    });                                                                                                                // 83
                                                                                                                       //
    if (!session.userId) {                                                                                             // 85
      throw new Meteor.Error(403, "Not logged in.");                                                                   // 86
    }                                                                                                                  // 87
                                                                                                                       //
    var grain = db.getGrain(session.grainId);                                                                          // 89
    SandstormPermissions.createNewApiToken(db, {                                                                       // 90
      identityId: session.identityId,                                                                                  // 92
      accountId: session.userId                                                                                        // 93
    }, session.grainId, "petname", value.roleAssignment, {                                                             // 92
      user: {                                                                                                          // 98
        identityId: value.id,                                                                                          // 98
        title: grain.title                                                                                             // 98
      }                                                                                                                // 98
    }); // TODO(soon): Somehow notify this user that they now have access.                                             // 98
    // TODO(perf): This permissions computation happens once here and then once again when the                         // 102
    //   `permissionsHeld` requirement is checked. Is there a way to avoid the duplicated work?                        // 103
                                                                                                                       //
    var permissions = SandstormPermissions.grainPermissions(db, {                                                      // 104
      grain: {                                                                                                         // 105
        _id: session.grainId,                                                                                          // 105
        identityId: value.id                                                                                           // 105
      }                                                                                                                // 105
    }, session.viewInfo || {}).permissions;                                                                            // 105
    return {                                                                                                           // 108
      descriptor: {                                                                                                    // 109
        tags: [{                                                                                                       // 110
          id: Identity.typeId,                                                                                         // 112
          value: Capnp.serialize(Identity.PowerboxTag, {                                                               // 113
            permissions: permissions                                                                                   // 116
          })                                                                                                           // 115
        }]                                                                                                             // 111
      },                                                                                                               // 109
      requirements: [{                                                                                                 // 121
        permissionsHeld: {                                                                                             // 123
          identityId: value.id,                                                                                        // 124
          grainId: session.grainId,                                                                                    // 125
          permissions: []                                                                                              // 126
        }                                                                                                              // 123
      }],                                                                                                              // 122
      frontendRef: value.id                                                                                            // 130
    };                                                                                                                 // 108
  },                                                                                                                   // 132
  query: function (db, userId, value) {                                                                                // 134
    var result = [];                                                                                                   // 135
    var requestedPermissions = [];                                                                                     // 136
                                                                                                                       //
    if (value) {                                                                                                       // 137
      requestedPermissions = Capnp.parse(Identity.PowerboxTag, value).permissions || [];                               // 138
    }                                                                                                                  // 139
                                                                                                                       //
    db.collections.contacts.find({                                                                                     // 141
      ownerId: userId                                                                                                  // 141
    }).forEach(function (contact) {                                                                                    // 141
      var identity = db.getIdentity(contact.identityId);                                                               // 142
                                                                                                                       //
      if (identity) {                                                                                                  // 143
        result.push({                                                                                                  // 144
          _id: "frontendref-identity-" + contact.identityId,                                                           // 145
          frontendRef: {                                                                                               // 146
            identity: contact.identityId                                                                               // 146
          },                                                                                                           // 146
          cardTemplate: "identityPowerboxCard",                                                                        // 147
          configureTemplate: "identityPowerboxConfiguration",                                                          // 148
          profile: identity.profile,                                                                                   // 149
          requestedPermissions: requestedPermissions,                                                                  // 150
          searchTerms: [identity.profile.name, identity.profile.handle, identity.profile.intrinsicName]                // 151
        });                                                                                                            // 144
      }                                                                                                                // 157
    });                                                                                                                // 158
    return result;                                                                                                     // 160
  }                                                                                                                    // 161
});                                                                                                                    // 73
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"install-server.js":["/imports/demo.js","/imports/server/async-helpers.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/install-server.js                                                                                            //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var allowDemo = void 0;                                                                                                // 1
module.import("/imports/demo.js", {                                                                                    // 1
  "allowDemo": function (v) {                                                                                          // 1
    allowDemo = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var promiseToFuture = void 0;                                                                                          // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "promiseToFuture": function (v) {                                                                                    // 1
    promiseToFuture = v;                                                                                               // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var TOKEN_CLEANUP_MINUTES = 120; // Give enough time for large uploads on slow connections.                            // 20
                                                                                                                       //
var TOKEN_CLEANUP_TIMER = TOKEN_CLEANUP_MINUTES * 60 * 1000;                                                           // 21
                                                                                                                       //
function cleanupToken(tokenId) {                                                                                       // 23
  check(tokenId, String);                                                                                              // 24
  globalDb.collections.spkTokens.remove({                                                                              // 25
    _id: tokenId                                                                                                       // 25
  });                                                                                                                  // 25
  waitPromise(globalBackend.cap().deleteBackup(tokenId));                                                              // 26
}                                                                                                                      // 27
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 29
  // Cleanup tokens every TOKEN_CLEANUP_MINUTES                                                                        // 30
  SandstormDb.periodicCleanup(TOKEN_CLEANUP_TIMER, function () {                                                       // 31
    var queryDate = new Date(Date.now() - TOKEN_CLEANUP_TIMER);                                                        // 32
    globalDb.collections.spkTokens.find({                                                                              // 34
      timestamp: {                                                                                                     // 34
        $lt: queryDate                                                                                                 // 34
      }                                                                                                                // 34
    }).forEach(function (token) {                                                                                      // 34
      cleanupToken(token._id);                                                                                         // 35
    });                                                                                                                // 36
  });                                                                                                                  // 37
});                                                                                                                    // 38
var localizedTextPattern = {                                                                                           // 40
  defaultText: String,                                                                                                 // 41
  localizations: Match.Optional([{                                                                                     // 42
    locale: String,                                                                                                    // 42
    text: String                                                                                                       // 42
  }])                                                                                                                  // 42
}; // Not all users are allowed to upload apps. We need to manually implement authorization                            // 40
// because Meteor.userId() is not available in server-side routes.                                                     // 46
                                                                                                                       //
Meteor.methods({                                                                                                       // 48
  cancelDownload: function (packageId) {                                                                               // 49
    check(packageId, String); // TODO(security):  Only let user cancel download if they initiated it.                  // 50
                                                                                                                       //
    cancelDownload(packageId);                                                                                         // 53
  },                                                                                                                   // 54
  newUploadToken: function () {                                                                                        // 56
    if (!isSignedUp()) {                                                                                               // 57
      throw new Meteor.Error(403, "Unauthorized", "Only invited users can upload apps.");                              // 58
    }                                                                                                                  // 59
                                                                                                                       //
    if (globalDb.isUninvitedFreeUser()) {                                                                              // 61
      throw new Meteor.Error(403, "Unauthorized", "Only paid users can upload apps.");                                 // 62
    }                                                                                                                  // 63
                                                                                                                       //
    var token = Random.id(22);                                                                                         // 65
    globalDb.collections.spkTokens.insert({                                                                            // 66
      _id: token,                                                                                                      // 66
      timestamp: new Date()                                                                                            // 66
    });                                                                                                                // 66
    return token;                                                                                                      // 68
  },                                                                                                                   // 69
  upgradeGrains: function (appId, version, packageId) {                                                                // 71
    this.connection.sandstormDb.upgradeGrains(appId, version, packageId, globalBackend);                               // 72
  },                                                                                                                   // 73
  ensureInstalled: function (packageId, url, isRetry) {                                                                // 75
    check(packageId, String);                                                                                          // 76
    check(url, Match.OneOf(String, undefined, null));                                                                  // 77
    check(isRetry, Boolean);                                                                                           // 78
                                                                                                                       //
    if (!packageId.match(/^[a-zA-Z0-9]*$/)) {                                                                          // 80
      throw new Meteor.Error(400, "The package ID contains illegal characters.");                                      // 81
    }                                                                                                                  // 82
                                                                                                                       //
    if (!this.userId) {                                                                                                // 84
      if (allowDemo && isSafeDemoAppUrl(url)) {// continue on                                                          // 85
      } else {                                                                                                         // 87
        // jscs:ignore disallowEmptyBlocks                                                                             // 87
        throw new Meteor.Error(403, "You must be logged in to install packages.");                                     // 88
      }                                                                                                                // 89
    } else if (!isSignedUpOrDemo()) {                                                                                  // 90
      throw new Meteor.Error(403, "This Sandstorm server requires you to get an invite before installing apps.");      // 91
    }                                                                                                                  // 93
                                                                                                                       //
    var pkg = Packages.findOne(packageId);                                                                             // 95
                                                                                                                       //
    if (!pkg || pkg.status !== "ready") {                                                                              // 97
      if (!this.userId || isDemoUser() || globalDb.isUninvitedFreeUser()) {                                            // 98
        if (!isSafeDemoAppUrl(url)) {                                                                                  // 99
          // TODO(someday): Billing prompt on client side.                                                             // 100
          throw new Meteor.Error(403, "Sorry, demo and free users cannot upload custom apps; " + "they may only install apps from apps.sandstorm.io.");
        }                                                                                                              // 103
      }                                                                                                                // 104
    }                                                                                                                  // 105
                                                                                                                       //
    if (!pkg || isRetry) {                                                                                             // 107
      globalDb.startInstall(packageId, url, isRetry);                                                                  // 108
    }                                                                                                                  // 109
  }                                                                                                                    // 110
});                                                                                                                    // 48
Meteor.publish("packageInfo", function (packageId) {                                                                   // 113
  check(packageId, String);                                                                                            // 114
  var db = this.connection.sandstormDb;                                                                                // 115
  var pkgCursor = db.collections.packages.find(packageId);                                                             // 116
  var pkg = pkgCursor.fetch()[0];                                                                                      // 117
                                                                                                                       //
  if (pkg && this.userId) {                                                                                            // 118
    return [pkgCursor, db.collections.userActions.find({                                                               // 119
      userId: this.userId,                                                                                             // 121
      appId: pkg.appId                                                                                                 // 121
    }), db.collections.grains.find({                                                                                   // 121
      userId: this.userId,                                                                                             // 122
      appId: pkg.appId                                                                                                 // 122
    })];                                                                                                               // 122
  } else {                                                                                                             // 124
    return pkgCursor;                                                                                                  // 125
  }                                                                                                                    // 126
});                                                                                                                    // 127
Router.map(function () {                                                                                               // 129
  this.route("upload", {                                                                                               // 130
    where: "server",                                                                                                   // 131
    path: "/upload/:token",                                                                                            // 132
    action: function () {                                                                                              // 134
      if (typeof this.params.token !== "string" || !globalDb.collections.spkTokens.findOne(this.params.token)) {       // 135
        this.response.writeHead(403, {                                                                                 // 137
          "Content-Type": "text/plain",                                                                                // 138
          "Access-Control-Allow-Origin": "*"                                                                           // 139
        });                                                                                                            // 137
        this.response.write("Invalid upload token.");                                                                  // 141
        this.response.end();                                                                                           // 142
      } else if (this.request.method === "POST") {                                                                     // 143
        try {                                                                                                          // 144
          var packageId = promiseToFuture(doClientUpload(this.request)).wait();                                        // 145
          this.response.writeHead(200, {                                                                               // 146
            "Content-Length": packageId.length,                                                                        // 147
            "Content-Type": "text/plain",                                                                              // 148
            "Access-Control-Allow-Origin": "*"                                                                         // 149
          });                                                                                                          // 146
          this.response.write(packageId);                                                                              // 151
          this.response.end();                                                                                         // 152
          globalDb.collections.spkTokens.remove(this.params.token);                                                    // 153
        } catch (error) {                                                                                              // 154
          console.error(error.stack);                                                                                  // 155
          this.response.writeHead(500, {                                                                               // 156
            "Content-Type": "text/plain",                                                                              // 157
            "Access-Control-Allow-Origin": "*"                                                                         // 158
          });                                                                                                          // 156
          this.response.write("Unpacking SPK failed; is it valid?");                                                   // 160
          this.response.end();                                                                                         // 161
        }                                                                                                              // 162
                                                                                                                       //
        ;                                                                                                              // 162
      } else if (this.request.method == "OPTIONS") {                                                                   // 163
        // Allow cross-origin posts to upload so that uploads can occur on the DDP host                                // 164
        // rather than the main host. In theory we could have Access-Control-Allow-Origin specify                      // 165
        // the main host rather than "*", but an upload request already requires a valid                               // 166
        // upload token, which is plenty of access control in itself.                                                  // 167
        var requestedHeaders = this.request.headers["access-control-request-headers"];                                 // 168
                                                                                                                       //
        if (requestedHeaders) {                                                                                        // 169
          this.response.setHeader("Access-Control-Allow-Headers", requestedHeaders);                                   // 170
        }                                                                                                              // 171
                                                                                                                       //
        this.response.writeHead(204, {                                                                                 // 173
          "Access-Control-Allow-Origin": "*",                                                                          // 174
          "Access-Control-Allow-Methods": "POST, OPTIONS",                                                             // 175
          "Access-Control-Max-Age": "3600"                                                                             // 176
        });                                                                                                            // 173
        this.response.end();                                                                                           // 178
      } else {                                                                                                         // 179
        this.response.writeHead(405, {                                                                                 // 180
          "Content-Type": "text/plain",                                                                                // 181
          "Access-Control-Allow-Origin": "*"                                                                           // 182
        });                                                                                                            // 180
        this.response.write("You can only POST here.");                                                                // 184
        this.response.end();                                                                                           // 185
      }                                                                                                                // 186
    }                                                                                                                  // 187
  });                                                                                                                  // 130
});                                                                                                                    // 189
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"installer.js":["babel-runtime/helpers/classCallCheck","fs","path","crypto","child_process","url","/imports/server/async-helpers.js","/imports/server/networking.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/installer.js                                                                                                 //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var Fs = void 0;                                                                                                       // 1
module.import("fs", {                                                                                                  // 1
  "default": function (v) {                                                                                            // 1
    Fs = v;                                                                                                            // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var Path = void 0;                                                                                                     // 1
module.import("path", {                                                                                                // 1
  "default": function (v) {                                                                                            // 1
    Path = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var Crypto = void 0;                                                                                                   // 1
module.import("crypto", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    Crypto = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var ChildProcess = void 0;                                                                                             // 1
module.import("child_process", {                                                                                       // 1
  "default": function (v) {                                                                                            // 1
    ChildProcess = v;                                                                                                  // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
var Url = void 0;                                                                                                      // 1
module.import("url", {                                                                                                 // 1
  "default": function (v) {                                                                                            // 1
    Url = v;                                                                                                           // 1
  }                                                                                                                    // 1
}, 4);                                                                                                                 // 1
var inMeteor = void 0,                                                                                                 // 1
    waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 5);                                                                                                                 // 1
var ssrfSafeLookupOrProxy = void 0;                                                                                    // 1
module.import("/imports/server/networking.js", {                                                                       // 1
  "ssrfSafeLookupOrProxy": function (v) {                                                                              // 1
    ssrfSafeLookupOrProxy = v;                                                                                         // 1
  }                                                                                                                    // 1
}, 6);                                                                                                                 // 1
var Request = HTTPInternals.NpmModules.request.module;                                                                 // 26
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 28
                                                                                                                       //
var Manifest = Capnp.importSystem("sandstorm/package.capnp").Manifest;                                                 // 30
var installers = void 0; // set to {} on main replica                                                                  // 32
// To protect against race conditions, we require that each row in the Packages                                        // 33
// collection have at most one writer at a time, as tracked by this `installers`                                       // 34
// map. Each key is a package ID and each value is either an AppInstaller object                                       // 35
// or the string 'uninstalling', indicating that some fiber is working on                                              // 36
// uninstalling the package.                                                                                           // 37
                                                                                                                       //
var verifyIsMainReplica = function () {                                                                                // 39
  if (Meteor.settings.replicaNumber) {                                                                                 // 40
    throw new Error("This can only be called on the main front-end replica.");                                         // 41
  }                                                                                                                    // 42
};                                                                                                                     // 43
                                                                                                                       //
var deletePackageInternal = function (pkg) {                                                                           // 45
  verifyIsMainReplica();                                                                                               // 46
  var packageId = pkg._id;                                                                                             // 48
                                                                                                                       //
  if (packageId in installers) {                                                                                       // 50
    return;                                                                                                            // 51
  }                                                                                                                    // 52
                                                                                                                       //
  installers[packageId] = "uninstalling";                                                                              // 54
                                                                                                                       //
  try {                                                                                                                // 56
    var action = UserActions.findOne({                                                                                 // 57
      packageId: packageId                                                                                             // 57
    });                                                                                                                // 57
    var grain = Grains.findOne({                                                                                       // 58
      packageId: packageId                                                                                             // 58
    });                                                                                                                // 58
    var notificationQuery = {};                                                                                        // 59
    notificationQuery["appUpdates." + pkg.appId + ".packageId"] = packageId;                                           // 60
                                                                                                                       //
    if (!grain && !action && !Notifications.findOne(notificationQuery) && !globalDb.getAppIdForPreinstalledPackage(packageId)) {
      Packages.update({                                                                                                // 63
        _id: packageId                                                                                                 // 64
      }, {                                                                                                             // 63
        $set: {                                                                                                        // 66
          status: "delete"                                                                                             // 66
        },                                                                                                             // 66
        $unset: {                                                                                                      // 67
          shouldCleanup: ""                                                                                            // 67
        }                                                                                                              // 67
      });                                                                                                              // 65
      waitPromise(globalBackend.cap().deletePackage(packageId));                                                       // 69
      Packages.remove(packageId); // Clean up assets (icon, etc).                                                      // 70
                                                                                                                       //
      getAllManifestAssets(pkg.manifest).forEach(function (assetId) {                                                  // 73
        globalDb.unrefStaticAsset(assetId);                                                                            // 74
      });                                                                                                              // 75
    } else {                                                                                                           // 76
      Packages.update({                                                                                                // 77
        _id: packageId                                                                                                 // 77
      }, {                                                                                                             // 77
        $unset: {                                                                                                      // 77
          shouldCleanup: ""                                                                                            // 77
        }                                                                                                              // 77
      });                                                                                                              // 77
    }                                                                                                                  // 78
                                                                                                                       //
    delete installers[packageId];                                                                                      // 80
  } catch (error) {                                                                                                    // 81
    delete installers[packageId];                                                                                      // 82
    throw error;                                                                                                       // 83
  }                                                                                                                    // 84
};                                                                                                                     // 85
                                                                                                                       //
var startInstallInternal = function (pkg) {                                                                            // 87
  verifyIsMainReplica();                                                                                               // 88
                                                                                                                       //
  if (pkg._id in installers) {                                                                                         // 90
    return;                                                                                                            // 91
  }                                                                                                                    // 92
                                                                                                                       //
  var installer = new AppInstaller(pkg._id, pkg.url, pkg.appId, pkg.isAutoUpdated);                                    // 94
  installers[pkg._id] = installer;                                                                                     // 95
  installer.start();                                                                                                   // 96
};                                                                                                                     // 97
                                                                                                                       //
cancelDownload = function (packageId) {                                                                                // 99
  Packages.remove({                                                                                                    // 100
    _id: packageId,                                                                                                    // 100
    status: "download"                                                                                                 // 100
  });                                                                                                                  // 100
};                                                                                                                     // 101
                                                                                                                       //
var cancelDownloadInternal = function (pkg) {                                                                          // 103
  verifyIsMainReplica();                                                                                               // 104
  var installer = installers[pkg._id]; // Don't do anything unless a download is in progress.                          // 106
                                                                                                                       //
  if (installer && installer.downloadRequest) {                                                                        // 109
    // OK, effect cancellation by faking an error.                                                                     // 110
    installer.wrapCallback(function () {                                                                               // 111
      throw new Error("Canceled");                                                                                     // 112
    })();                                                                                                              // 113
  }                                                                                                                    // 114
};                                                                                                                     // 115
                                                                                                                       //
if (!Meteor.settings.replicaNumber) {                                                                                  // 117
  installers = {};                                                                                                     // 118
  Meteor.startup(function () {                                                                                         // 120
    // Restart any deletions that were killed while in-progress.                                                       // 121
    Packages.find({                                                                                                    // 122
      status: "delete"                                                                                                 // 122
    }).forEach(deletePackageInternal); // Watch for new installation requests and fulfill them.                        // 122
                                                                                                                       //
    Packages.find({                                                                                                    // 125
      status: {                                                                                                        // 125
        $in: ["download", "verify", "unpack", "analyze"]                                                               // 125
      }                                                                                                                // 125
    }).observe({                                                                                                       // 125
      added: startInstallInternal,                                                                                     // 126
      removed: cancelDownloadInternal                                                                                  // 127
    }); // Watch for new cleanup requests and fulfill them.                                                            // 125
                                                                                                                       //
    Packages.find({                                                                                                    // 131
      status: "ready",                                                                                                 // 131
      shouldCleanup: true                                                                                              // 131
    }).observe({                                                                                                       // 131
      added: deletePackageInternal                                                                                     // 132
    });                                                                                                                // 131
  });                                                                                                                  // 134
}                                                                                                                      // 135
                                                                                                                       //
doClientUpload = function (stream) {                                                                                   // 137
  return new Promise(function (resolve, reject) {                                                                      // 138
    var id = Random.id();                                                                                              // 139
    var backendStream = globalBackend.cap().installPackage().stream;                                                   // 141
    var hasher = Crypto.createHash("sha256");                                                                          // 142
    stream.on("data", function (chunk) {                                                                               // 144
      try {                                                                                                            // 145
        hasher.update(chunk);                                                                                          // 146
        backendStream.write(chunk);                                                                                    // 147
      } catch (err) {                                                                                                  // 148
        reject(err);                                                                                                   // 149
      }                                                                                                                // 150
    });                                                                                                                // 151
    stream.on("end", function () {                                                                                     // 153
      try {                                                                                                            // 154
        backendStream.done();                                                                                          // 155
        var packageId = hasher.digest("hex").slice(0, 32);                                                             // 156
        resolve(backendStream.saveAs(packageId).then(function () {                                                     // 157
          return packageId;                                                                                            // 158
        }));                                                                                                           // 159
        backendStream.close();                                                                                         // 160
      } catch (err) {                                                                                                  // 161
        reject(err);                                                                                                   // 162
      }                                                                                                                // 163
    });                                                                                                                // 164
    stream.on("error", function (err) {                                                                                // 166
      // TODO(soon):  This event does't seem to fire if the user leaves the page mid-upload.                           // 167
      try {                                                                                                            // 168
        backendStream.close();                                                                                         // 169
        reject(err);                                                                                                   // 170
      } catch (err2) {                                                                                                 // 171
        reject(err2);                                                                                                  // 172
      }                                                                                                                // 173
    });                                                                                                                // 174
  });                                                                                                                  // 175
};                                                                                                                     // 176
                                                                                                                       //
AppInstaller = function () {                                                                                           // 178
  function AppInstaller(packageId, url, appId, isAutoUpdated) {                                                        // 179
    (0, _classCallCheck3.default)(this, AppInstaller);                                                                 // 179
    verifyIsMainReplica();                                                                                             // 180
    this.packageId = packageId;                                                                                        // 182
    this.url = url;                                                                                                    // 183
    this.failed = false;                                                                                               // 184
    this.appId = appId;                                                                                                // 185
    this.isAutoUpdated = isAutoUpdated; // Serializes database writes.                                                 // 186
                                                                                                                       //
    this.writeChain = Promise.resolve();                                                                               // 189
  }                                                                                                                    // 190
                                                                                                                       //
  AppInstaller.prototype.updateProgress = function () {                                                                // 178
    function updateProgress(status, progress, error, manifest) {                                                       // 178
      // TODO(security):  On error, we should actually delete the package from the database and only                   // 193
      //   display the error to whomever was watching at the time.  Otherwise it's easy to confuse                     // 194
      //   people by 'pre-failing' packages.  (Actually, perhaps if a user tries to download an                        // 195
      //   already-downloading package but specifies a different URL, we really should initiate an                     // 196
      //   entirely separate download...  but cancel it if the first download succeeds.)                               // 197
      this.status = status;                                                                                            // 199
      this.progress = progress || -1;                                                                                  // 200
      this.error = error;                                                                                              // 201
      this.manifest = manifest || null;                                                                                // 202
                                                                                                                       //
      var _this = this; // The callback passed to inMeteor() runs in a new fiber. We need to make sure database writes
      // occur in exactly the order in which we generate them, so we use a promise chain to serialize                  // 207
      // them.                                                                                                         // 208
                                                                                                                       //
                                                                                                                       //
      this.writeChain = this.writeChain.then(function () {                                                             // 209
        return inMeteor(function () {                                                                                  // 210
          if (manifest) extractManifestAssets(manifest);                                                               // 211
          Packages.update(_this.packageId, {                                                                           // 213
            $set: {                                                                                                    // 214
              status: _this.status,                                                                                    // 215
              progress: _this.progress,                                                                                // 216
              error: _this.error ? _this.error.message : null,                                                         // 217
              manifest: _this.manifest,                                                                                // 218
              appId: _this.appId,                                                                                      // 219
              authorPgpKeyFingerprint: _this.authorPgpKeyFingerprint                                                   // 220
            }                                                                                                          // 214
          });                                                                                                          // 213
                                                                                                                       //
          if (_this.authorPgpKeyFingerprint) {                                                                         // 224
            globalDb.updateKeybaseProfileAsync(_this.authorPgpKeyFingerprint);                                         // 225
          }                                                                                                            // 226
        }).catch(function (err) {                                                                                      // 227
          console.error(err.stack);                                                                                    // 228
        });                                                                                                            // 229
      });                                                                                                              // 230
    }                                                                                                                  // 231
                                                                                                                       //
    return updateProgress;                                                                                             // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.wrapCallback = function () {                                                                  // 178
    function wrapCallback(method) {                                                                                    // 178
      // Note that the function below must not be an arrow function, since arrow functions do not have                 // 234
      // access to the context's arguments array.                                                                      // 235
      var _this = this;                                                                                                // 236
                                                                                                                       //
      return function () {                                                                                             // 237
        if (_this.failed) return;                                                                                      // 238
                                                                                                                       //
        try {                                                                                                          // 239
          return method.apply(_this, _.toArray(arguments));                                                            // 240
        } catch (err) {                                                                                                // 241
          _this.failed = true;                                                                                         // 242
                                                                                                                       //
          _this.cleanup();                                                                                             // 243
                                                                                                                       //
          _this.updateProgress("failed", 0, err);                                                                      // 244
                                                                                                                       //
          _this.writeChain = _this.writeChain.then(function () {                                                       // 245
            delete installers[_this.packageId];                                                                        // 246
          });                                                                                                          // 247
          console.error("Failed to install app:", err.stack);                                                          // 248
        }                                                                                                              // 249
      };                                                                                                               // 250
    }                                                                                                                  // 251
                                                                                                                       //
    return wrapCallback;                                                                                               // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.cleanup = function () {                                                                       // 178
    function cleanup() {                                                                                               // 178
      if (this.uploadStream) {                                                                                         // 254
        try {                                                                                                          // 255
          this.uploadStream.close();                                                                                   // 256
        } catch (err) {}                                                                                               // 257
                                                                                                                       //
        delete this.uploadStream;                                                                                      // 259
      }                                                                                                                // 260
                                                                                                                       //
      if (this.downloadRequest) {                                                                                      // 262
        try {                                                                                                          // 263
          this.downloadRequest.abort();                                                                                // 264
        } catch (err) {}                                                                                               // 265
                                                                                                                       //
        delete this.downloadRequest;                                                                                   // 267
      }                                                                                                                // 268
    }                                                                                                                  // 269
                                                                                                                       //
    return cleanup;                                                                                                    // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.start = function () {                                                                         // 178
    function start() {                                                                                                 // 178
      var _this2 = this;                                                                                               // 271
                                                                                                                       //
      return this.wrapCallback(function () {                                                                           // 272
        _this2.cleanup();                                                                                              // 273
                                                                                                                       //
        globalBackend.cap().tryGetPackage(_this2.packageId).then(_this2.wrapCallback(function (info) {                 // 275
          if (info.appId) {                                                                                            // 276
            _this2.appId = info.appId;                                                                                 // 277
            _this2.authorPgpKeyFingerprint = info.authorPgpKeyFingerprint;                                             // 278
                                                                                                                       //
            _this2.done(info.manifest);                                                                                // 279
          } else {                                                                                                     // 280
            _this2.doDownload();                                                                                       // 281
          }                                                                                                            // 282
        }), _this2.wrapCallback(function (err) {                                                                       // 283
          throw err;                                                                                                   // 284
        }));                                                                                                           // 285
      })();                                                                                                            // 286
    }                                                                                                                  // 287
                                                                                                                       //
    return start;                                                                                                      // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.doDownload = function () {                                                                    // 178
    function doDownload() {                                                                                            // 178
      if (!this.url) {                                                                                                 // 290
        throw new Error("Unknown package ID, and no URL was provided.");                                               // 291
      }                                                                                                                // 292
                                                                                                                       //
      console.log("Downloading app:", this.url);                                                                       // 294
      this.updateProgress("download");                                                                                 // 295
      this.uploadStream = globalBackend.cap().installPackage().stream;                                                 // 297
      return this.doDownloadTo(this.uploadStream);                                                                     // 298
    }                                                                                                                  // 299
                                                                                                                       //
    return doDownload;                                                                                                 // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.doDownloadTo = function () {                                                                  // 178
    function doDownloadTo(out) {                                                                                       // 178
      inMeteor(this.wrapCallback(function () {                                                                         // 302
        var _this3 = this;                                                                                             // 302
                                                                                                                       //
        var safe = ssrfSafeLookupOrProxy(globalDb, this.url);                                                          // 303
        var bytesExpected = undefined;                                                                                 // 305
        var bytesReceived = 0;                                                                                         // 306
        var hasher = Crypto.createHash("sha256");                                                                      // 307
        var done = false;                                                                                              // 308
                                                                                                                       //
        var updateDownloadProgress = _.throttle(this.wrapCallback(function () {                                        // 309
          if (!done) {                                                                                                 // 310
            if (bytesExpected) {                                                                                       // 311
              _this3.updateProgress("download", bytesReceived / bytesExpected);                                        // 312
            } else {                                                                                                   // 313
              _this3.updateProgress("download", bytesReceived);                                                        // 314
            }                                                                                                          // 315
          }                                                                                                            // 316
        }), 500);                                                                                                      // 317
                                                                                                                       //
        var request = safe.proxy ? Request.get(this.url, {                                                             // 319
          proxy: safe.proxy                                                                                            // 320
        }) : Request.get(safe.url, {                                                                                   // 320
          headers: {                                                                                                   // 322
            host: safe.host                                                                                            // 322
          },                                                                                                           // 322
          servername: safe.host.split(":")[0]                                                                          // 323
        });                                                                                                            // 321
        request.on("response", this.wrapCallback(function (response) {                                                 // 326
          if ("content-length" in response.headers) {                                                                  // 327
            bytesExpected = parseInt(response.headers["content-length"]);                                              // 328
          }                                                                                                            // 329
        }));                                                                                                           // 330
        request.on("data", this.wrapCallback(function (chunk) {                                                        // 332
          hasher.update(chunk);                                                                                        // 333
          out.write(chunk);                                                                                            // 334
          bytesReceived += chunk.length;                                                                               // 335
          updateDownloadProgress();                                                                                    // 336
        }));                                                                                                           // 337
        request.on("end", this.wrapCallback(function () {                                                              // 339
          out.done();                                                                                                  // 340
                                                                                                                       //
          if (hasher.digest("hex").slice(0, 32) !== _this3.packageId) {                                                // 342
            throw new Error("Package hash did not match.");                                                            // 343
          }                                                                                                            // 344
                                                                                                                       //
          done = true;                                                                                                 // 346
          delete _this3.downloadRequest;                                                                               // 347
                                                                                                                       //
          _this3.updateProgress("unpack");                                                                             // 349
                                                                                                                       //
          out.saveAs(_this3.packageId).then(_this3.wrapCallback(function (info) {                                      // 350
            _this3.appId = info.appId;                                                                                 // 351
            _this3.authorPgpKeyFingerprint = info.authorPgpKeyFingerprint;                                             // 352
                                                                                                                       //
            _this3.done(info.manifest);                                                                                // 353
          }), _this3.wrapCallback(function (err) {                                                                     // 354
            throw err;                                                                                                 // 355
          }));                                                                                                         // 356
        }));                                                                                                           // 357
        request.on("error", this.wrapCallback(function (err) {                                                         // 359
          throw err;                                                                                                   // 359
        }));                                                                                                           // 359
        this.downloadRequest = request;                                                                                // 361
      }));                                                                                                             // 362
    }                                                                                                                  // 363
                                                                                                                       //
    return doDownloadTo;                                                                                               // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  AppInstaller.prototype.done = function () {                                                                          // 178
    function done(manifest) {                                                                                          // 178
      console.log("App ready:", this.packageId);                                                                       // 366
      this.updateProgress("ready", 1, undefined, manifest);                                                            // 367
                                                                                                                       //
      var _this = this;                                                                                                // 368
                                                                                                                       //
      _this.writeChain = _this.writeChain.then(function () {                                                           // 369
        return inMeteor(function () {                                                                                  // 370
          if (_this.isAutoUpdated) {                                                                                   // 371
            globalDb.sendAppUpdateNotifications(_this.appId, _this.packageId, manifest.appTitle && manifest.appTitle.defaultText, manifest.appVersion, manifest.appMarketingVersion && manifest.appMarketingVersion.defaultText);
          }                                                                                                            // 375
                                                                                                                       //
          if (globalDb.getPackageIdForPreinstalledApp(_this.appId) && globalDb.collections.appIndex.findOne({          // 377
            _id: _this.appId,                                                                                          // 379
            packageId: _this.packageId                                                                                 // 380
          })) {                                                                                                        // 378
            // Only mark app as preinstall ready if its appId is in the preinstalledApps setting                       // 382
            // and if it's the latest package version in the appIndex. The updateAppIndex function                     // 383
            // will always trigger updates of preinstalled apps, even if a concurrent download of                      // 384
            // an older package is going on.                                                                           // 385
            globalDb.setPreinstallAppAsReady(_this.appId, _this.packageId);                                            // 386
          }                                                                                                            // 387
        });                                                                                                            // 388
      }).then(function () {                                                                                            // 389
        delete installers[_this.packageId];                                                                            // 390
      });                                                                                                              // 391
    }                                                                                                                  // 392
                                                                                                                       //
    return done;                                                                                                       // 178
  }();                                                                                                                 // 178
                                                                                                                       //
  return AppInstaller;                                                                                                 // 178
}();                                                                                                                   // 178
                                                                                                                       //
extractManifestAssets = function (manifest) {                                                                          // 395
  var metadata = manifest.metadata;                                                                                    // 396
  if (!metadata) return;                                                                                               // 397
  var icons = metadata.icons;                                                                                          // 399
                                                                                                                       //
  if (icons) {                                                                                                         // 400
    var handleIcon = function (icon) {                                                                                 // 401
      if (icon.svg) {                                                                                                  // 402
        icon.assetId = globalDb.addStaticAsset({                                                                       // 403
          mimeType: "image/svg+xml"                                                                                    // 403
        }, icon.svg);                                                                                                  // 403
        icon.format = "svg";                                                                                           // 404
        delete icon.svg;                                                                                               // 405
        return true;                                                                                                   // 406
      } else if (icon.png) {                                                                                           // 407
        // Use the 1x version for 'normal' DPI, unless 1x isn't provided, in which case use 2x.                        // 408
        var normalDpi = icon.png.dpi1x || icon.png.dpi2x;                                                              // 409
        if (!normalDpi) return false;                                                                                  // 410
        icon.format = "png";                                                                                           // 411
        icon.assetId = globalDb.addStaticAsset({                                                                       // 412
          mimeType: "image/png"                                                                                        // 412
        }, normalDpi);                                                                                                 // 412
                                                                                                                       //
        if (icon.png.dpi1x && icon.png.dpi2x) {                                                                        // 414
          // Icon specifies both resolutions, so also record a 2x DPI option.                                          // 415
          icon.assetId2xDpi = globalDb.addStaticAsset({                                                                // 416
            mimeType: "image/png"                                                                                      // 416
          }, icon.png.dpi2x);                                                                                          // 416
        }                                                                                                              // 417
                                                                                                                       //
        delete icon.png;                                                                                               // 419
        return true;                                                                                                   // 420
      } else {                                                                                                         // 421
        // Unknown icon. Filter it.                                                                                    // 422
        return false;                                                                                                  // 423
      }                                                                                                                // 424
    };                                                                                                                 // 425
                                                                                                                       //
    if (icons.appGrid && !handleIcon(icons.appGrid)) delete icons.appGrid;                                             // 427
    if (icons.grain && !handleIcon(icons.grain)) delete icons.grain; // We don't need the market icons.                // 428
                                                                                                                       //
    if (icons.market) delete icons.market;                                                                             // 431
    if (icons.marketBig) delete icons.marketBig;                                                                       // 432
  }                                                                                                                    // 433
                                                                                                                       //
  var handleLocalizedText = function (text) {                                                                          // 435
    if (text.defaultText) {                                                                                            // 436
      text.defaultTextAssetId = globalDb.addStaticAsset({                                                              // 437
        mimeType: "text/plain"                                                                                         // 437
      }, text.defaultText);                                                                                            // 437
      delete text.defaultText;                                                                                         // 438
    }                                                                                                                  // 439
                                                                                                                       //
    if (text.localizations) {                                                                                          // 441
      text.localizations.forEach(function (localization) {                                                             // 442
        if (localization.text) {                                                                                       // 443
          localization.assetId = globalDb.addStaticAsset({                                                             // 444
            mimeType: "text/plain"                                                                                     // 445
          }, localization.text);                                                                                       // 445
          delete localization.text;                                                                                    // 446
        }                                                                                                              // 447
      });                                                                                                              // 448
    }                                                                                                                  // 449
  };                                                                                                                   // 450
                                                                                                                       //
  var license = metadata.license;                                                                                      // 452
                                                                                                                       //
  if (license) {                                                                                                       // 453
    if (license.proprietary) license.proprietary = handleLocalizedText(license.proprietary);                           // 454
    if (license.publicDomain) license.publicDomain = handleLocalizedText(license.proprietary);                         // 455
    if (license.notices) license.notices = handleLocalizedText(license.notices);                                       // 456
  }                                                                                                                    // 457
                                                                                                                       //
  var author = metadata.author;                                                                                        // 459
                                                                                                                       //
  if (author) {                                                                                                        // 460
    // We remove the PGP signature since it was already verified down to a key ID in the back-end.                     // 461
    if (author.pgpSignature) delete author.pgpSignature;                                                               // 462
  } // Don't need the keyring either.                                                                                  // 463
                                                                                                                       //
                                                                                                                       //
  if (metadata.pgpKeyring) delete metadata.pgpKeyring; // Perhaps used by the 'about' page?                            // 466
                                                                                                                       //
  if (metadata.description) metadata.description = handleLocalizedText(metadata.description); // Screenshots are for app marketing; we don't use them post-install.
                                                                                                                       //
  if (metadata.screenshots) delete metadata.screenshots; // We might allow the user to view the changelog.             // 472
                                                                                                                       //
  if (metadata.changeLog) metadata.changeLog = handleLocalizedText(metadata.changeLog);                                // 475
};                                                                                                                     // 476
                                                                                                                       //
getAllManifestAssets = function (manifest) {                                                                           // 478
  // Returns a list of all asset IDs in the given manifest.                                                            // 479
  var metadata = manifest.metadata;                                                                                    // 481
  if (!metadata) return [];                                                                                            // 482
  var result = [];                                                                                                     // 484
  var icons = metadata.icons;                                                                                          // 486
                                                                                                                       //
  if (icons) {                                                                                                         // 487
    var handleIcon = function (icon) {                                                                                 // 488
      if (icon.assetId) {                                                                                              // 489
        result.push(icon.assetId);                                                                                     // 490
      }                                                                                                                // 491
    };                                                                                                                 // 492
                                                                                                                       //
    if (icons.appGrid) handleIcon(icons.appGrid);                                                                      // 494
    if (icons.grain) handleIcon(icons.grain);                                                                          // 495
  }                                                                                                                    // 496
                                                                                                                       //
  var handleLocalizedText = function (text) {                                                                          // 498
    if (text.defaultTextAssetId) {                                                                                     // 499
      result.push(defaultTextAssetId);                                                                                 // 500
    }                                                                                                                  // 501
                                                                                                                       //
    if (text.localizations) {                                                                                          // 503
      text.localizations.forEach(function (localization) {                                                             // 504
        if (localization.assetId) {                                                                                    // 505
          result.push(localization.assetId);                                                                           // 506
        }                                                                                                              // 507
      });                                                                                                              // 508
    }                                                                                                                  // 509
  };                                                                                                                   // 510
                                                                                                                       //
  var license = metadata.license;                                                                                      // 512
                                                                                                                       //
  if (license) {                                                                                                       // 513
    if (license.proprietary) handleLocalizedText(license.proprietary);                                                 // 514
    if (license.publicDomain) handleLocalizedText(license.publicDomain);                                               // 515
    if (license.notices) handleLocalizedText(license.notices);                                                         // 516
  }                                                                                                                    // 517
                                                                                                                       //
  if (metadata.description) handleLocalizedText(metadata.description);                                                 // 519
  if (metadata.changeLog) handleLocalizedText(metadata.changeLog);                                                     // 520
  return result;                                                                                                       // 522
};                                                                                                                     // 523
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"notifications-server.js":["/imports/server/async-helpers.js","/imports/server/desktop-notifications.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/notifications-server.js                                                                                      //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var createAppActivityDesktopNotification = void 0;                                                                     // 1
module.import("/imports/server/desktop-notifications.js", {                                                            // 1
  "createAppActivityDesktopNotification": function (v) {                                                               // 1
    createAppActivityDesktopNotification = v;                                                                          // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 20
                                                                                                                       //
var SupervisorCapnp = Capnp.importSystem("sandstorm/supervisor.capnp");                                                // 21
var SystemPersistent = SupervisorCapnp.SystemPersistent;                                                               // 22
                                                                                                                       //
logActivity = function (grainId, identityIdOrAnonymous, event) {                                                       // 24
  // identityIdOrAnonymous is the string "anonymous" for an anonymous user, or is null for a                           // 25
  // non-user-initiated ("background") activity.                                                                       // 26
  check(grainId, String);                                                                                              // 28
  check(identityIdOrAnonymous, Match.Maybe(String)); // `event` is always an ActivityEvent parsed from Cap'n Proto but that's too complicated to check
  // here.                                                                                                             // 31
                                                                                                                       //
  var identityId = identityIdOrAnonymous === "anonymous" ? null : identityIdOrAnonymous; // TODO(perf): A cached copy of the grain from when the session opened would be fine to use
  //   here, rather than looking it up every time.                                                                     // 36
                                                                                                                       //
  grain = Grains.findOne(grainId);                                                                                     // 37
                                                                                                                       //
  if (!grain) {                                                                                                        // 38
    // Shouldn't be possible since activity events come from the grain.                                                // 39
    throw new Error("no such grain");                                                                                  // 40
  } // Look up the event typedef.                                                                                      // 41
                                                                                                                       //
                                                                                                                       //
  var eventType = ((grain.cachedViewInfo || {}).eventTypes || [])[event.type];                                         // 44
                                                                                                                       //
  if (!eventType) {                                                                                                    // 45
    throw new Error("No such event type in app's ViewInfo: " + event.type);                                            // 46
  }                                                                                                                    // 47
                                                                                                                       //
  if (!eventType.suppressUnread) {                                                                                     // 49
    // Clear the "seenAllActivity" bit for all users except the acting user.                                           // 50
    // TODO(perf): Consider throttling? Or should that be the app's responsibility?                                    // 51
    if (identityId != grain.identityId) {                                                                              // 52
      Grains.update(grainId, {                                                                                         // 53
        $unset: {                                                                                                      // 53
          ownerSeenAllActivity: true                                                                                   // 53
        }                                                                                                              // 53
      });                                                                                                              // 53
    } // Also clear on ApiTokens.                                                                                      // 54
                                                                                                                       //
                                                                                                                       //
    ApiTokens.update({                                                                                                 // 57
      "grainId": grainId,                                                                                              // 58
      "owner.user.seenAllActivity": true,                                                                              // 59
      "owner.user.identityId": {                                                                                       // 60
        $ne: identityId                                                                                                // 60
      }                                                                                                                // 60
    }, {                                                                                                               // 57
      $unset: {                                                                                                        // 61
        "owner.user.seenAllActivity": true                                                                             // 61
      }                                                                                                                // 61
    }, {                                                                                                               // 61
      multi: true                                                                                                      // 61
    });                                                                                                                // 61
  }                                                                                                                    // 62
                                                                                                                       //
  if (identityId) {                                                                                                    // 64
    // Apply auto-subscriptions.                                                                                       // 65
    if (eventType.autoSubscribeToGrain) {                                                                              // 66
      globalDb.subscribeToActivity(identityId, grainId);                                                               // 67
    }                                                                                                                  // 68
                                                                                                                       //
    if (event.thread && eventType.autoSubscribeToThread) {                                                             // 70
      globalDb.subscribeToActivity(identityId, grainId, event.thread.path || "");                                      // 71
    }                                                                                                                  // 72
  } // Figure out whom we need to notify.                                                                              // 73
                                                                                                                       //
                                                                                                                       //
  var notifyMap = {};                                                                                                  // 76
                                                                                                                       //
  var addRecipient = function (recipient) {                                                                            // 77
    // Mutes take priority over subscriptions.                                                                         // 78
    if (recipient.mute) {                                                                                              // 79
      notifyMap[recipient.identityId] = false;                                                                         // 80
    } else {                                                                                                           // 81
      if (!(recipient.identityId in notifyMap)) {                                                                      // 82
        notifyMap[recipient.identityId] = true;                                                                        // 83
      }                                                                                                                // 84
    }                                                                                                                  // 85
  };                                                                                                                   // 86
                                                                                                                       //
  if (identityId) {                                                                                                    // 88
    // Don't notify self.                                                                                              // 89
    addRecipient({                                                                                                     // 90
      identityId: identityId,                                                                                          // 90
      mute: true                                                                                                       // 90
    });                                                                                                                // 90
  } // Notify subscribers, if desired.                                                                                 // 91
                                                                                                                       //
                                                                                                                       //
  if (eventType.notifySubscribers) {                                                                                   // 94
    // The grain owner is implicitly subscribed.                                                                       // 95
    addRecipient({                                                                                                     // 96
      identityId: grain.identityId                                                                                     // 96
    }); // Add everyone subscribed to the grain.                                                                       // 96
                                                                                                                       //
    globalDb.getActivitySubscriptions(grainId).forEach(addRecipient);                                                  // 99
                                                                                                                       //
    if (event.thread) {                                                                                                // 101
      // Add everyone subscribed to the thread.                                                                        // 102
      globalDb.getActivitySubscriptions(grainId, event.thread.path || "").forEach(addRecipient);                       // 103
    }                                                                                                                  // 104
  } // Add everyone who is mentioned.                                                                                  // 105
                                                                                                                       //
                                                                                                                       //
  if (event.users && event.users.length > 0) {                                                                         // 108
    var promises = [];                                                                                                 // 109
    event.users.forEach(function (user) {                                                                              // 110
      if (user.identity && (user.mentioned || user.subscribed)) {                                                      // 111
        promises.push(unwrapFrontendCap(user.identity, "identity", function (targetId) {                               // 112
          addRecipient({                                                                                               // 113
            identityId: targetId                                                                                       // 113
          });                                                                                                          // 113
        }));                                                                                                           // 114
      }                                                                                                                // 115
    });                                                                                                                // 116
    waitPromise(Promise.all(promises).then(function (junk) {                                                           // 117
      return undefined;                                                                                                // 117
    }));                                                                                                               // 117
  } // Make a list of everyone to notify.                                                                              // 118
                                                                                                                       //
                                                                                                                       //
  var notify = [];                                                                                                     // 121
                                                                                                                       //
  for (var _identityId in meteorBabelHelpers.sanitizeForInObject(notifyMap)) {                                         // 122
    if (notifyMap[_identityId]) {                                                                                      // 123
      notify.push(_identityId);                                                                                        // 124
    }                                                                                                                  // 125
  }                                                                                                                    // 126
                                                                                                                       //
  if (notify.length > 0) {                                                                                             // 128
    var notification = {                                                                                               // 129
      grainId: grainId,                                                                                                // 130
      path: event.path || ""                                                                                           // 131
    }; // Fields we'll update even if the notification already exists.                                                 // 129
                                                                                                                       //
    var update = {                                                                                                     // 135
      isUnread: true,                                                                                                  // 136
      timestamp: new Date()                                                                                            // 137
    };                                                                                                                 // 135
                                                                                                                       //
    if (event.thread) {                                                                                                // 140
      notification.threadPath = event.thread.path || "";                                                               // 141
    }                                                                                                                  // 142
                                                                                                                       //
    if (identityId) {                                                                                                  // 144
      notification.initiatingIdentity = identityId;                                                                    // 145
    } else if (identityIdOrAnonymous) {                                                                                // 146
      notification.initiatorAnonymous = true;                                                                          // 147
    }                                                                                                                  // 148
                                                                                                                       //
    notification.eventType = event.type;                                                                               // 150
    update.text = eventType.verbPhrase;                                                                                // 151
    var body = event.notification && event.notification.caption || {                                                   // 153
      defaultText: ""                                                                                                  // 153
    };                                                                                                                 // 153
    var appActivity = {                                                                                                // 155
      grainId: notification.grainId,                                                                                   // 156
      path: notification.path,                                                                                         // 157
      body: body,                                                                                                      // 158
      actionText: eventType.verbPhrase                                                                                 // 159
    };                                                                                                                 // 155
                                                                                                                       //
    if (identityId) {                                                                                                  // 162
      // Look up icon urls for the responsible identity and the app                                                    // 163
      var identity = Meteor.users.findOne({                                                                            // 164
        _id: identityId                                                                                                // 164
      });                                                                                                              // 164
                                                                                                                       //
      if (!identity) {                                                                                                 // 165
        throw new Error("no such identity");                                                                           // 166
      }                                                                                                                // 167
                                                                                                                       //
      SandstormDb.fillInProfileDefaults(identity);                                                                     // 169
      SandstormDb.fillInPictureUrl(identity);                                                                          // 170
      appActivity.user = {                                                                                             // 172
        identityId: identity._id,                                                                                      // 173
        name: identity.profile.name,                                                                                   // 174
        avatarUrl: identity.profile.pictureUrl || ""                                                                   // 175
      };                                                                                                               // 172
    } else if (identityIdOrAnonymous) {                                                                                // 177
      appActivity.user = {                                                                                             // 178
        anonymous: true                                                                                                // 178
      };                                                                                                               // 178
    }                                                                                                                  // 179
                                                                                                                       //
    notify.forEach(function (targetId) {                                                                               // 181
      // Notify all accounts connected with this identity.                                                             // 182
      Meteor.users.find({                                                                                              // 183
        $or: [{                                                                                                        // 183
          "loginIdentities.id": targetId                                                                               // 184
        }, {                                                                                                           // 184
          "nonloginIdentities.id": targetId                                                                            // 185
        }]                                                                                                             // 185
      }).forEach(function (account) {                                                                                  // 183
        // We need to know the ID of the inserted/updated document so we can embed it in the                           // 187
        // desktop notification to bind them.                                                                          // 188
        var idIfInserted = Random.id(17);                                                                              // 189
        var result = Notifications.findAndModify({                                                                     // 190
          query: _.extend({                                                                                            // 191
            userId: account._id                                                                                        // 191
          }, notification),                                                                                            // 191
          update: {                                                                                                    // 192
            $set: update,                                                                                              // 193
            $inc: {                                                                                                    // 194
              count: 1                                                                                                 // 194
            },                                                                                                         // 194
            $setOnInsert: {                                                                                            // 195
              _id: idIfInserted                                                                                        // 196
            }                                                                                                          // 195
          },                                                                                                           // 192
          upsert: true                                                                                                 // 199
        });                                                                                                            // 190
                                                                                                                       //
        if (!result.ok) {                                                                                              // 202
          console.error("Couldn't create notification!", result.lastErrorObject);                                      // 203
          return;                                                                                                      // 204
        }                                                                                                              // 205
                                                                                                                       //
        var notificationId = result.value._id || idIfInserted;                                                         // 207
        var desktopNotification = {                                                                                    // 209
          userId: account._id,                                                                                         // 210
          notificationId: notificationId,                                                                              // 211
          appActivity: appActivity                                                                                     // 212
        };                                                                                                             // 209
        createAppActivityDesktopNotification(desktopNotification);                                                     // 215
      });                                                                                                              // 216
    });                                                                                                                // 217
  }                                                                                                                    // 218
};                                                                                                                     // 219
                                                                                                                       //
Meteor.methods({                                                                                                       // 221
  testNotifications: function () {                                                                                     // 222
    // Deliver a test notification of each non-grain-initiated type to the user.                                       // 223
    if (!this.userId) return;                                                                                          // 224
    Notifications.insert({                                                                                             // 226
      admin: {                                                                                                         // 227
        action: "/admin/stats",                                                                                        // 228
        type: "reportStats"                                                                                            // 229
      },                                                                                                               // 227
      userId: this.userId,                                                                                             // 231
      timestamp: new Date(),                                                                                           // 232
      isUnread: true                                                                                                   // 233
    });                                                                                                                // 226
    Notifications.insert({                                                                                             // 236
      userId: this.userId,                                                                                             // 237
      referral: true,                                                                                                  // 238
      timestamp: new Date(),                                                                                           // 239
      isUnread: true                                                                                                   // 240
    });                                                                                                                // 236
                                                                                                                       //
    if (global.BlackrockPayments) {                                                                                    // 243
      Notifications.insert({                                                                                           // 244
        userId: this.userId,                                                                                           // 245
        mailingListBonus: true,                                                                                        // 246
        timestamp: new Date(),                                                                                         // 247
        isUnread: true                                                                                                 // 248
      });                                                                                                              // 244
    }                                                                                                                  // 250
  }                                                                                                                    // 251
});                                                                                                                    // 221
Meteor.publish("notifications", function () {                                                                          // 254
  return Notifications.find({                                                                                          // 255
    userId: this.userId                                                                                                // 255
  });                                                                                                                  // 255
});                                                                                                                    // 256
Meteor.publish("notificationGrains", function (notificationIds) {                                                      // 258
  // Since publishes can't be reactive, we leave it to the client to subscribe to both                                 // 259
  // "notifications" and "notificationGrains" reactively.                                                              // 260
  check(notificationIds, [String]);                                                                                    // 261
  var notifications = Notifications.find({                                                                             // 262
    _id: {                                                                                                             // 263
      $in: notificationIds                                                                                             // 263
    },                                                                                                                 // 263
    userId: this.userId                                                                                                // 264
  }, {                                                                                                                 // 262
    fields: {                                                                                                          // 266
      grainId: 1,                                                                                                      // 266
      initiatingIdentity: 1                                                                                            // 266
    }                                                                                                                  // 266
  });                                                                                                                  // 265
  var grainIds = notifications.map(function (row) {                                                                    // 269
    return row.grainId;                                                                                                // 270
  }).filter(function (x) {                                                                                             // 271
    return x;                                                                                                          // 271
  });                                                                                                                  // 271
  var identities = notifications.map(function (row) {                                                                  // 273
    return row.initiatingIdentity;                                                                                     // 274
  }).filter(function (x) {                                                                                             // 275
    return x;                                                                                                          // 275
  });                                                                                                                  // 275
  return [Meteor.users.find({                                                                                          // 277
    _id: {                                                                                                             // 278
      $in: identities                                                                                                  // 278
    }                                                                                                                  // 278
  }, {                                                                                                                 // 278
    fields: {                                                                                                          // 278
      profile: 1                                                                                                       // 278
    }                                                                                                                  // 278
  })];                                                                                                                 // 278
});                                                                                                                    // 280
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"pre-meteor.js":["/imports/server/async-helpers.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/pre-meteor.js                                                                                                //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var inMeteor = void 0;                                                                                                 // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 21
                                                                                                                       //
var Fs = Npm.require("fs");                                                                                            // 22
                                                                                                                       //
var Dns = Npm.require("dns");                                                                                          // 23
                                                                                                                       //
var Future = Npm.require("fibers/future");                                                                             // 24
                                                                                                                       //
var Http = Npm.require("http");                                                                                        // 25
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 26
                                                                                                                       //
var ByteStream = Capnp.importSystem("sandstorm/util.capnp").ByteStream;                                                // 27
var HOSTNAME = Url.parse(process.env.ROOT_URL).hostname;                                                               // 29
var DDP_HOSTNAME = process.env.DDP_DEFAULT_CONNECTION_URL && Url.parse(process.env.DDP_DEFAULT_CONNECTION_URL).hostname;
var CACHE_TTL_SECONDS = 30; // 30 seconds.  Cache-Control expects units of seconds, not millis.                        // 32
                                                                                                                       //
var DNS_CACHE_TTL = CACHE_TTL_SECONDS * 1000; // DNS cache is in millis.                                               // 33
                                                                                                                       //
var webPublishingHandlers = {}; // Maps grain public IDs to Connect handlers.                                          // 35
// TODO(perf): Garbage-collect this map?                                                                               // 37
// TODO(perf): Implement a more general cache not just of web publishing handlers but of all sorts                     // 39
//   of hosts, in order to avoid database lookups to determine how to handle requests. Host types                      // 40
//   to distinguish between:                                                                                           // 41
//   - Grain hosts                                                                                                     // 42
//   - Web publishing hosts                                                                                            // 43
//   - Standalone grain hosts                                                                                          // 44
                                                                                                                       //
var dnsCache = {}; // Unfortunately, node's DNS library doesn't cache results, so we do our own caching.               // 46
// Unfortunately, node's DNS library also dosen't give us TTLs. So, we'll cache for                                    // 48
// DNS_CACHE_TTL (a relatively small value) and rely on the upstream DNS server to implement                           // 49
// better caching.                                                                                                     // 50
                                                                                                                       //
function isSandstormShell(hostname) {                                                                                  // 52
  // Is this hostname mapped to the Sandstorm shell?                                                                   // 53
  return hostname === HOSTNAME || DDP_HOSTNAME && hostname === DDP_HOSTNAME;                                           // 55
} // We need to use connect. Let's make sure we're using the same version as Meteor's WebApp module                    // 56
// uses. Fortunately, they let us extract it.                                                                          // 59
                                                                                                                       //
                                                                                                                       //
var connect = WebAppInternals.NpmModules.connect.module;                                                               // 60
var mime = connect.static.mime;                                                                                        // 62
                                                                                                                       //
function wwwHandlerForGrain(grainId) {                                                                                 // 64
  return function (request, response) {                                                                                // 65
    var path = request.url; // Strip query.                                                                            // 66
                                                                                                                       //
    path = path.split("?")[0]; // If a directory, open "index.html".                                                   // 69
                                                                                                                       //
    if (path.slice(-1) === "/") {                                                                                      // 72
      path = path + "index.html";                                                                                      // 73
    } // Strip leading "/".                                                                                            // 74
                                                                                                                       //
                                                                                                                       //
    if (path[0] === "/") path = path.slice(1); // URI-decode the rest. Note that this allows filenames to contain spaces and question marks.
                                                                                                                       //
    var encoded = path;                                                                                                // 80
    path = decodeURIComponent(encoded);                                                                                // 81
    var type = mime.lookup(path);                                                                                      // 83
    var charset = mime.charsets.lookup(type);                                                                          // 84
                                                                                                                       //
    if (charset) {                                                                                                     // 85
      type = type + "; charset=" + charset;                                                                            // 86
    } else if (type === "application/json") {                                                                          // 87
      // HACK: Apparently the MIME module does not assume UTF-8 for JSON. :(                                           // 88
      type = type + "; charset=utf-8";                                                                                 // 89
    } // TODO(perf): Automatically gzip text content? Use Express's "compress" middleware for this?                    // 90
    //   Note that nginx will also auto-compress things...                                                             // 93
                                                                                                                       //
                                                                                                                       //
    response.setHeader("Content-Type", type);                                                                          // 95
    response.setHeader("Cache-Control", "public, max-age=" + CACHE_TTL_SECONDS);                                       // 96
                                                                                                                       //
    if (path === "apps/index.json" || path.match(/apps\/[a-z0-9]{52}[.]json/) || path === "experimental/index.json" || path.match(/experimental\/[a-z0-9]{52}[.]json/)) {
      // TODO(cleanup): Extra special terrible hack: The app index needs to serve these JSON files                     // 102
      //   cross-origin. We could almost just make all web sites allow cross-origin since generally                    // 103
      //   web publishing is meant to publish public content. There is one case where this is                          // 104
      //   problematic, though: sites behind a firewall. Those sites could potentially be read                         // 105
      //   by outside sites if CORS is enabled on them. Some day we should make it so apps can                         // 106
      //   explicitly opt-in to allowing cross-origin queries but that day is not today.                               // 107
      response.setHeader("Access-Control-Allow-Origin", "*");                                                          // 108
    }                                                                                                                  // 109
                                                                                                                       //
    var stream = new ResponseStream(response, 200, "OK"); // We hold our own capability to the stream becaues we don't want ResponseStream's close()
    // logic to happen if the response has no data.                                                                    // 114
    //                                                                                                                 // 115
    // TODO(cleanup): This is pretty ugly.                                                                             // 116
                                                                                                                       //
    var streamCap = new Capnp.Capability(stream, ByteStream);                                                          // 117
    globalBackend.useGrain(grainId, function (supervisor) {                                                            // 119
      var tryPath = function (path) {                                                                                  // 120
        return supervisor.getWwwFileHack(path, streamCap).then(function (result) {                                     // 120
          // jscs:disable disallowQuotedKeysInObjects                                                                  // 121
          var status = result.status;                                                                                  // 122
                                                                                                                       //
          if (status === "file") {// Nothing to do.                                                                    // 123
          } else if (status === "directory") {                                                                         // 125
            stream.sendingDirectResponse();                                                                            // 126
            response.writeHead(303, {                                                                                  // 127
              "Content-Type": "text/plain",                                                                            // 128
              "Location": "/" + path + "/",                                                                            // 129
              "Cache-Control": "public, max-age=" + CACHE_TTL_SECONDS                                                  // 130
            });                                                                                                        // 127
            response.end("redirect: /" + path + "/");                                                                  // 132
          } else if (status === "notFound") {                                                                          // 133
            if (path !== encoded) {                                                                                    // 134
              // Compatibility hack: Try getting the URL-encoded version of the path. We used to not                   // 135
              // URL-decode paths at all, hence the need for this fallback.                                            // 136
              // TODO(apibump): Move this hack to the compat layer.                                                    // 137
              return tryPath(encoded);                                                                                 // 138
            }                                                                                                          // 139
                                                                                                                       //
            stream.sendingDirectResponse();                                                                            // 141
            response.writeHead(404, {                                                                                  // 142
              "Content-Type": "text/plain"                                                                             // 143
            });                                                                                                        // 142
            response.end("404 not found: /" + path);                                                                   // 145
          } else {                                                                                                     // 146
            console.error("didn't understand result of getWwwFileHack:", status);                                      // 147
            stream.sendingDirectResponse();                                                                            // 148
            response.writeHead(500, {                                                                                  // 149
              "Content-Type": "text/plain"                                                                             // 150
            });                                                                                                        // 149
            response.end("Internal server error");                                                                     // 152
          }                                                                                                            // 153
                                                                                                                       //
          streamCap.close();                                                                                           // 155
        });                                                                                                            // 156
      };                                                                                                               // 120
                                                                                                                       //
      return tryPath(path);                                                                                            // 157
    }).catch(function (err) {                                                                                          // 158
      console.error(err.stack);                                                                                        // 159
      streamCap.close();                                                                                               // 160
    });                                                                                                                // 161
  };                                                                                                                   // 162
}                                                                                                                      // 163
                                                                                                                       //
function writeErrorResponse(res, err) {                                                                                // 165
  var status = 500;                                                                                                    // 166
                                                                                                                       //
  if (err instanceof Meteor.Error && typeof err.error === "number" && err.error >= 400 && err.error < 600) {           // 167
    status = err.error;                                                                                                // 169
  } else if (err.httpErrorCode) {                                                                                      // 170
    status = err.httpErrorCode;                                                                                        // 171
  } // Log errors that are our fault, but not errors that are the client's fault.                                      // 172
                                                                                                                       //
                                                                                                                       //
  if (status >= 500) console.error(err.stack);                                                                         // 175
  res.writeHead(status, {                                                                                              // 177
    "Content-Type": err.htmlMessage ? "text/html" : "text/plain"                                                       // 177
  });                                                                                                                  // 177
  res.end(err.htmlMessage || err.message);                                                                             // 178
}                                                                                                                      // 179
                                                                                                                       //
var PNG_MAGIC = new Buffer([0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A]);                                          // 181
var JPEG_MAGIC = new Buffer([0xFF, 0xD8, 0xFF]);                                                                       // 182
                                                                                                                       //
function checkMagic(buf, magic) {                                                                                      // 184
  if (buf.length < magic.length) return false;                                                                         // 185
                                                                                                                       //
  for (var i = 0; i < magic.length; i++) {                                                                             // 187
    if (buf[i] != magic[i]) return false;                                                                              // 188
  }                                                                                                                    // 189
                                                                                                                       //
  return true;                                                                                                         // 191
}                                                                                                                      // 192
                                                                                                                       //
function serveSelfTest(req, res) {                                                                                     // 194
  try {                                                                                                                // 195
    if (req.method === "GET" && req.url === "/") {                                                                     // 196
      var content = new Buffer("Self-test OK."); // Convert the ROOT_URL to something that is a valid origin.          // 198
                                                                                                                       //
      var rootUrlAsOrigin = process.env.ROOT_URL;                                                                      // 201
                                                                                                                       //
      if (rootUrlAsOrigin.slice(-1) === "/") {                                                                         // 202
        rootUrlAsOrigin = rootUrlAsOrigin.slice(0, -1);                                                                // 203
      }                                                                                                                // 204
                                                                                                                       //
      res.writeHead(200, {                                                                                             // 206
        "Content-Type": "text/plain",                                                                                  // 207
        "Content-Length": content.length,                                                                              // 208
        "Access-Control-Allow-Origin": rootUrlAsOrigin                                                                 // 209
      });                                                                                                              // 206
      res.end(content);                                                                                                // 211
    } else {                                                                                                           // 212
      res.writeHead(400, {                                                                                             // 213
        "Content-Type": "text/plain"                                                                                   // 213
      });                                                                                                              // 213
      res.end("Bad request to self-test subdomain.");                                                                  // 214
      return;                                                                                                          // 215
    }                                                                                                                  // 216
  } catch (err) {                                                                                                      // 217
    console.log("WARNING: An error occurred while serving self-test; proceeding anyway:", err);                        // 218
  }                                                                                                                    // 219
}                                                                                                                      // 220
                                                                                                                       //
function serveStaticAsset(req, res, hostId) {                                                                          // 222
  inMeteor(function () {                                                                                               // 223
    if (req.method === "GET") {                                                                                        // 224
      var assetCspHeader = "default-src 'none'; style-src 'unsafe-inline'; sandbox";                                   // 225
                                                                                                                       //
      if (req.headers["if-none-match"] === "permanent") {                                                              // 226
        // Cache never invalidates since we use a new URL for every resource.                                          // 227
        res.writeHead(304, {                                                                                           // 228
          "Cache-Control": "public, max-age=31536000",                                                                 // 229
          "ETag": "permanent",                                                                                         // 230
          // To be safe, send these again, although it shouldn't be necessary.                                         // 232
          "Content-Security-Policy": assetCspHeader,                                                                   // 233
          "Access-Control-Allow-Origin": "*",                                                                          // 234
          "X-Content-Type-Options": "nosniff"                                                                          // 235
        });                                                                                                            // 228
        res.end();                                                                                                     // 237
        return;                                                                                                        // 238
      }                                                                                                                // 239
                                                                                                                       //
      var url = Url.parse(req.url, true);                                                                              // 241
      var pathname = url.pathname.slice(1);                                                                            // 242
      var asset = void 0;                                                                                              // 243
                                                                                                                       //
      if (pathname.startsWith("identicon/")) {                                                                         // 244
        var size = parseInt((url.query || {}).s);                                                                      // 245
        asset = new Identicon(pathname.slice("identicon/".length), size).asAsset();                                    // 246
      } else if (pathname.indexOf("/") == -1) {                                                                        // 247
        asset = globalDb.getStaticAsset(pathname);                                                                     // 248
      }                                                                                                                // 249
                                                                                                                       //
      if (asset) {                                                                                                     // 251
        var headers = {                                                                                                // 252
          "Content-Type": asset.mimeType,                                                                              // 253
          "Content-Length": asset.content.length,                                                                      // 254
          // Assets can be cached forever because each one has a unique ID.                                            // 256
          "Cache-Control": "public, max-age=31536000",                                                                 // 257
          // Since different resources get different URLs, we can use a static etag.                                   // 259
          "ETag": "permanent",                                                                                         // 260
          // Set strict Content-Security-Policy to prevent static assets from executing any script                     // 262
          // or doing basically anything when browsed to directly. The static assets host is not                       // 263
          // intended to serve HTML. Mostly, it serves images and javascript -- note that setting                      // 264
          // the CSP header on Javascript files does not prevent other hosts from voluntarily                          // 265
          // specifying these scripts in <script> tags.                                                                // 266
          "Content-Security-Policy": assetCspHeader,                                                                   // 267
          // Allow any host to fetch these assets. This is safe since requests to this host are                        // 269
          // totally side-effect-free and the asset ID acts as a capability to prevent loading                         // 270
          // assets you're not supposed to know about.                                                                 // 271
          "Access-Control-Allow-Origin": "*",                                                                          // 272
          // Extra protection against content type trickery.                                                           // 274
          "X-Content-Type-Options": "nosniff"                                                                          // 275
        };                                                                                                             // 252
                                                                                                                       //
        if (asset.encoding) {                                                                                          // 278
          headers["Content-Encoding"] = asset.encoding;                                                                // 279
        }                                                                                                              // 280
                                                                                                                       //
        res.writeHead(200, headers);                                                                                   // 282
        res.end(asset.content);                                                                                        // 283
      } else {                                                                                                         // 284
        res.writeHead(404, {                                                                                           // 285
          "Content-Type": "text/plain"                                                                                 // 285
        });                                                                                                            // 285
        res.end("not found");                                                                                          // 286
      }                                                                                                                // 287
    } else if (req.method === "POST") {                                                                                // 288
      res.setHeader("Access-Control-Allow-Origin", "*");                                                               // 289
                                                                                                                       //
      var _url = Url.parse(req.url);                                                                                   // 291
                                                                                                                       //
      var purpose = globalDb.fulfillAssetUpload(_url.pathname.slice(1)); // Sanity check the purpose of this upload token.
                                                                                                                       //
      if (!purpose) {                                                                                                  // 295
        res.writeHead(404, {                                                                                           // 296
          "Content-Type": "text/plain"                                                                                 // 296
        });                                                                                                            // 296
        res.end("Upload token not found or expired.");                                                                 // 297
        return;                                                                                                        // 298
      } else if (purpose.profilePicture) {                                                                             // 299
        var userId = purpose.profilePicture.userId;                                                                    // 300
        var identityId = purpose.profilePicture.identityId;                                                            // 301
        check(userId, String);                                                                                         // 302
        check(identityId, String);                                                                                     // 303
      } else if (purpose.loginLogo) {// no additional fields for loginLogo                                             // 304
      } else {                                                                                                         // 306
        res.writeHead(500, {                                                                                           // 307
          "Content-Type": "text/plain"                                                                                 // 307
        });                                                                                                            // 307
        res.end("Found invalid purpose for upload");                                                                   // 308
        return;                                                                                                        // 309
      }                                                                                                                // 310
                                                                                                                       //
      var buffers = [];                                                                                                // 312
      var totalSize = 0;                                                                                               // 313
      var done = new Future();                                                                                         // 314
      req.on("data", function (buf) {                                                                                  // 315
        totalSize += buf.length;                                                                                       // 316
                                                                                                                       //
        if (totalSize <= 64 * 1024) {                                                                                  // 317
          buffers.push(buf);                                                                                           // 318
        }                                                                                                              // 319
      });                                                                                                              // 320
      req.on("end", done.return.bind(done));                                                                           // 321
      req.on("error", done.throw.bind(done));                                                                          // 322
      done.wait();                                                                                                     // 323
                                                                                                                       //
      if (totalSize > 64 * 1024) {                                                                                     // 325
        // TODO(soon): Resize the image ourselves.                                                                     // 326
        res.writeHead(400, {                                                                                           // 327
          "Content-Type": "text/plain"                                                                                 // 327
        });                                                                                                            // 327
        res.end("Picture too large; please use an image under 64 KiB.");                                               // 328
        return;                                                                                                        // 329
      }                                                                                                                // 330
                                                                                                                       //
      var content = Buffer.concat(buffers);                                                                            // 332
      var type = void 0;                                                                                               // 333
                                                                                                                       //
      if (checkMagic(content, PNG_MAGIC)) {                                                                            // 334
        type = "image/png";                                                                                            // 335
      } else if (checkMagic(content, JPEG_MAGIC)) {                                                                    // 336
        type = "image/jpeg";                                                                                           // 337
      } else {                                                                                                         // 338
        res.writeHead(400, {                                                                                           // 339
          "Content-Type": "text/plain"                                                                                 // 339
        });                                                                                                            // 339
        res.end("Image must be PNG or JPEG.");                                                                         // 340
        return;                                                                                                        // 341
      }                                                                                                                // 342
                                                                                                                       //
      var assetId = globalDb.addStaticAsset({                                                                          // 344
        mimeType: type                                                                                                 // 344
      }, content);                                                                                                     // 344
                                                                                                                       //
      if (purpose.profilePicture) {                                                                                    // 346
        var _identityId = purpose.profilePicture.identityId;                                                           // 347
        var result = Meteor.users.findAndModify({                                                                      // 348
          query: {                                                                                                     // 349
            _id: _identityId                                                                                           // 349
          },                                                                                                           // 349
          update: {                                                                                                    // 350
            $set: {                                                                                                    // 350
              "profile.picture": assetId                                                                               // 350
            }                                                                                                          // 350
          },                                                                                                           // 350
          fields: {                                                                                                    // 351
            "profile.picture": 1                                                                                       // 351
          }                                                                                                            // 351
        });                                                                                                            // 348
                                                                                                                       //
        if (result.ok) {                                                                                               // 354
          var old = result.value;                                                                                      // 355
                                                                                                                       //
          if (old && old.profile && old.profile.picture) {                                                             // 356
            globalDb.unrefStaticAsset(old.profile.picture);                                                            // 357
          }                                                                                                            // 358
        } else {                                                                                                       // 359
          res.writeHead(500, {                                                                                         // 360
            "Content-Type": "text/plain"                                                                               // 360
          });                                                                                                          // 360
          res.end("Couldn't update profile picture.");                                                                 // 361
          return;                                                                                                      // 362
        }                                                                                                              // 363
      } else if (purpose.loginLogo) {                                                                                  // 364
        var _result = globalDb.collections.settings.findAndModify({                                                    // 365
          query: {                                                                                                     // 366
            _id: "whitelabelCustomLogoAssetId"                                                                         // 366
          },                                                                                                           // 366
          update: {                                                                                                    // 367
            $setOnInsert: {                                                                                            // 368
              _id: "whitelabelCustomLogoAssetId"                                                                       // 368
            },                                                                                                         // 368
            $set: {                                                                                                    // 369
              value: assetId                                                                                           // 369
            }                                                                                                          // 369
          },                                                                                                           // 367
          upsert: true,                                                                                                // 371
          fields: {                                                                                                    // 372
            value: 1                                                                                                   // 372
          }                                                                                                            // 372
        });                                                                                                            // 365
                                                                                                                       //
        if (_result.ok) {                                                                                              // 375
          var _old = _result.value;                                                                                    // 376
                                                                                                                       //
          if (_old && _old.value) {                                                                                    // 377
            globalDb.unrefStaticAsset(_old.value);                                                                     // 378
          }                                                                                                            // 379
        } else {                                                                                                       // 380
          res.writeHead(500, {                                                                                         // 381
            "Content-Type": "text/plain"                                                                               // 381
          });                                                                                                          // 381
          res.end("Couldn't update login logo.");                                                                      // 382
          return;                                                                                                      // 383
        }                                                                                                              // 384
      }                                                                                                                // 385
                                                                                                                       //
      res.writeHead(204, {});                                                                                          // 387
      res.end();                                                                                                       // 388
    } else if (req.method === "OPTIONS") {                                                                             // 389
      var requestedHeaders = req.headers["access-control-request-headers"];                                            // 390
                                                                                                                       //
      if (requestedHeaders) {                                                                                          // 391
        res.setHeader("Access-Control-Allow-Headers", requestedHeaders);                                               // 392
      }                                                                                                                // 393
                                                                                                                       //
      res.writeHead(204, {                                                                                             // 395
        "Access-Control-Allow-Origin": "*",                                                                            // 396
        "Access-Control-Allow-Methods": "GET, HEAD, POST, OPTIONS",                                                    // 397
        "Access-Control-Max-Age": "3600"                                                                               // 398
      });                                                                                                              // 395
      res.end();                                                                                                       // 400
    } else {                                                                                                           // 401
      res.writeHead(405, "Method Not Allowed", {                                                                       // 402
        "Allow": "GET, POST, OPTIONS",                                                                                 // 403
        "Content-Type": "text/plain"                                                                                   // 404
      });                                                                                                              // 402
      res.end("405 Method Not Allowed: " + req.method);                                                                // 406
    }                                                                                                                  // 407
  }).catch(function (err) {                                                                                            // 408
    writeErrorResponse(res, err);                                                                                      // 409
  });                                                                                                                  // 410
}                                                                                                                      // 411
                                                                                                                       //
var canonicalizeShellOrWildcardUrl = function (hostname, url) {                                                        // 413
  // Start with ROOT_URL, apply host & path from inbound URL, then                                                     // 414
  // redirect.                                                                                                         // 415
  var targetUrl = Url.parse(process.env.ROOT_URL); // Retain the protocol & port from ROOT_URL but use the inbound     // 416
  // hostname.                                                                                                         // 419
                                                                                                                       //
  targetUrl.host = hostname + ":" + targetUrl.port; // The following allows to avoid decoding + re-encoding query      // 420
  // string parameters, if provided.                                                                                   // 423
                                                                                                                       //
  targetUrl = Url.resolve(Url.format(targetUrl), url);                                                                 // 424
  return targetUrl;                                                                                                    // 425
};                                                                                                                     // 426
                                                                                                                       //
var handleNonMeteorRequest = function (req, res, next, redirectIfInWildcard) {                                         // 428
  // if redirectIfInWildcard is true, and if the host is part of the wildcard host, then we will                       // 429
  // prefer to redirect to the canonical url for that request.  If false, we will respond to the                       // 430
  // request directly.                                                                                                 // 431
  var hostname = req.headers.host.split(":")[0];                                                                       // 433
  var publicIdPromise = void 0; // See if the request was for a host in the wildcard.                                  // 435
                                                                                                                       //
  var id = matchWildcardHost(req.headers.host);                                                                        // 437
                                                                                                                       //
  if (id) {                                                                                                            // 438
    // Match!                                                                                                          // 439
    if (redirectIfInWildcard) {                                                                                        // 440
      res.writeHead(302, {                                                                                             // 441
        "Location": canonicalizeShellOrWildcardUrl(hostname, req.url)                                                  // 441
      });                                                                                                              // 441
      res.end();                                                                                                       // 442
      return;                                                                                                          // 443
    }                                                                                                                  // 444
                                                                                                                       //
    if (id === "static") {                                                                                             // 446
      // Static assets domain.                                                                                         // 447
      serveStaticAsset(req, res);                                                                                      // 448
      return;                                                                                                          // 449
    }                                                                                                                  // 450
                                                                                                                       //
    if (id.match(/^selftest-/)) {                                                                                      // 452
      // Self test domain pattern. Starts w/ hyphen to avoid ambiguity with grain session/static                       // 453
      // publishing wildcard hosts.                                                                                    // 454
      serveSelfTest(req, res);                                                                                         // 455
      return;                                                                                                          // 456
    } // Try to route the request to a session.                                                                        // 457
                                                                                                                       //
                                                                                                                       //
    publicIdPromise = tryProxyRequest(id, req, res).then(function (handled) {                                          // 460
      if (handled) {                                                                                                   // 461
        return null;                                                                                                   // 462
      } else {                                                                                                         // 463
        return id;                                                                                                     // 464
      }                                                                                                                // 465
    });                                                                                                                // 466
  } else {                                                                                                             // 467
    // Not a wildcard host. Perhaps it is a custom host.                                                               // 468
    publicIdPromise = lookupPublicIdFromDns(hostname);                                                                 // 469
  }                                                                                                                    // 470
                                                                                                                       //
  publicIdPromise.then(function (publicId) {                                                                           // 472
    if (publicId) {                                                                                                    // 473
      return Promise.resolve(undefined).then(function () {                                                             // 474
        var handler = webPublishingHandlers[publicId];                                                                 // 475
                                                                                                                       //
        if (handler) {                                                                                                 // 476
          return handler;                                                                                              // 477
        } else {                                                                                                       // 478
          // We don't have a handler for this publicId, so look it up in the grain DB.                                 // 479
          return inMeteor(function () {                                                                                // 480
            var grain = Grains.findOne({                                                                               // 481
              publicId: publicId                                                                                       // 481
            }, {                                                                                                       // 481
              fields: {                                                                                                // 481
                _id: 1                                                                                                 // 481
              }                                                                                                        // 481
            });                                                                                                        // 481
                                                                                                                       //
            if (!grain) {                                                                                              // 482
              throw new Meteor.Error(404, "No such grain for public ID: " + publicId);                                 // 483
            }                                                                                                          // 484
                                                                                                                       //
            var grainId = grain._id;                                                                                   // 486
            return webPublishingHandlers[publicId] = wwwHandlerForGrain(grainId);                                      // 487
          });                                                                                                          // 488
        }                                                                                                              // 489
      }).then(function (handler) {                                                                                     // 490
        handler(req, res);                                                                                             // 491
      });                                                                                                              // 492
    } else {                                                                                                           // 493
      // Already handled by tryProxyRequest above, no further action needed.                                           // 494
      return Promise.resolve(undefined);                                                                               // 495
    }                                                                                                                  // 496
  }).catch(function (err) {                                                                                            // 497
    writeErrorResponse(res, err);                                                                                      // 498
  });                                                                                                                  // 499
}; // This function serves responses on Sandstorm's main HTTP/HTTPS                                                    // 500
// port.                                                                                                               // 503
                                                                                                                       //
                                                                                                                       //
var handleNonMeteorRequestDirectly = function (req, res, next) {                                                       // 504
  handleNonMeteorRequest(req, res, next, false);                                                                       // 505
}; // "Alternate ports" are ports other than the main HTTP or HTTPS port.                                              // 506
// They are handled as follows:                                                                                        // 509
//                                                                                                                     // 510
// * Requests to the shell & grains, we redirect to the main port.                                                     // 511
// * For requests to static publishing on hosts within the wildcard, we redirect to the main port.                     // 512
// * For static publishing glued in via DNS TXT record, we serve the request as-is.                                    // 513
//                                                                                                                     // 514
// Alternate ports are bound to FD #5 and higher.                                                                      // 515
                                                                                                                       //
                                                                                                                       //
var getNumberOfAlternatePorts = function () {                                                                          // 516
  var numPorts = process.env.PORT.split(",").length;                                                                   // 517
  var numAlternatePorts = numPorts - 1;                                                                                // 518
  return numAlternatePorts;                                                                                            // 519
};                                                                                                                     // 520
                                                                                                                       //
var handleNonMainPortRequest = function (req, res, next) {                                                             // 522
  try {                                                                                                                // 523
    if (!req.headers.host) {                                                                                           // 524
      res.writeHead(400, {                                                                                             // 525
        "Content-Type": "text/plain"                                                                                   // 525
      });                                                                                                              // 525
      res.end("Missing Host header");                                                                                  // 526
      return;                                                                                                          // 527
    } // If this request is intended for the shell, redirect to the canonical shell URL.                               // 528
                                                                                                                       //
                                                                                                                       //
    var hostname = req.headers.host.split(":")[0];                                                                     // 531
                                                                                                                       //
    if (isSandstormShell(hostname)) {                                                                                  // 532
      res.writeHead(302, {                                                                                             // 533
        "Location": canonicalizeShellOrWildcardUrl(hostname, req.url)                                                  // 533
      });                                                                                                              // 533
      res.end();                                                                                                       // 534
      return;                                                                                                          // 535
    } // TODO(someday): check for standalone hosts and do something reasonable                                         // 536
                                                                                                                       //
                                                                                                                       //
    handleNonMeteorRequest(req, res, next, true);                                                                      // 539
  } catch (e) {                                                                                                        // 540
    // This should never be reached, because all the request handlers should be catching                               // 541
    // exceptions, but you can never be too careful in a top-level request handler.                                    // 542
    console.err("Unhandled exception in request handler:", e.stack);                                                   // 543
    res.writeHead(500, {                                                                                               // 544
      "Content-Type": "text/plain"                                                                                     // 544
    });                                                                                                                // 544
    res.end("Unhandled exception: " + e.stack);                                                                        // 545
  }                                                                                                                    // 546
};                                                                                                                     // 547
                                                                                                                       //
var listenOnAlternatePorts = function () {                                                                             // 549
  for (var i = 0; i < getNumberOfAlternatePorts(); i++) {                                                              // 550
    // Call createServerForSandstorm() to skip our monkey patching.                                                    // 551
    var alternatePortServer = Http.createServerForSandstorm(handleNonMainPortRequest);                                 // 552
    alternatePortServer.listen({                                                                                       // 553
      fd: i + 5                                                                                                        // 553
    });                                                                                                                // 553
  }                                                                                                                    // 554
};                                                                                                                     // 555
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 557
  var meteorRequestListeners = WebApp.httpServer.listeners("request"); // Construct the middleware chain for requests to non-DDP, non-shell hosts.
                                                                                                                       //
  var nonMeteorRequestHandler = connect(); // BlackrockPayments is only defined in the Blackrock build of Sandstorm.   // 561
                                                                                                                       //
  if (global.BlackrockPayments) {                                                                                      // 563
    // Have to check with global, because it could be undefined.                                                       // 563
    nonMeteorRequestHandler.use(BlackrockPayments.makeConnectHandler(globalDb));                                       // 564
  }                                                                                                                    // 565
                                                                                                                       //
  nonMeteorRequestHandler.use(handleNonMeteorRequestDirectly);                                                         // 567
  WebApp.httpServer.removeAllListeners("request");                                                                     // 569
  WebApp.httpServer.on("request", function (req, res) {                                                                // 570
    Promise.resolve(undefined).then(function () {                                                                      // 571
      if (!req.headers.host) {                                                                                         // 572
        res.writeHead(400, {                                                                                           // 573
          "Content-Type": "text/plain"                                                                                 // 573
        });                                                                                                            // 573
        res.end("Missing Host header");                                                                                // 574
        return;                                                                                                        // 575
      }                                                                                                                // 576
                                                                                                                       //
      var hostname = req.headers.host.split(":")[0]; // strip port if it exists                                        // 578
                                                                                                                       //
      if (isSandstormShell(hostname)) {                                                                                // 579
        if (req.url.startsWith("/_oauth/")) {                                                                          // 580
          // Intercept oauth callbacks on the main host to check if the user was actually logging                      // 581
          // into a standalone host, in which case we need to redirect to that host instead.                           // 582
          var parsedUrl = Url.parse(req.url, true);                                                                    // 583
                                                                                                                       //
          if (parsedUrl.query && parsedUrl.query.state) {                                                              // 584
            var rawState = new Buffer(parsedUrl.query.state, "base64");                                                // 585
            var state = JSON.parse(rawState.toString());                                                               // 586
                                                                                                                       //
            if (state.redirectUrl) {                                                                                   // 587
              var parsedRedirect = Url.parse(state.redirectUrl);                                                       // 588
              var redirectHostname = parsedRedirect.hostname;                                                          // 589
                                                                                                                       //
              if (redirectHostname !== HOSTNAME) {                                                                     // 590
                return inMeteor(function () {                                                                          // 591
                  if (globalDb.hostIsStandalone(redirectHostname)) {                                                   // 592
                    res.writeHead(302, {                                                                               // 593
                      "Location": parsedRedirect.protocol + "//" + parsedRedirect.host + req.url                       // 593
                    });                                                                                                // 593
                    res.end();                                                                                         // 595
                    return;                                                                                            // 596
                  } else {                                                                                             // 597
                    throw new Meteor.Error(400, "redirectUrl in OAuth was for an unknown host: " + state.redirectUrl);
                  }                                                                                                    // 600
                });                                                                                                    // 601
              }                                                                                                        // 602
            }                                                                                                          // 603
          }                                                                                                            // 604
        } // If destined for the DDP host or the main host, pass on to Meteor                                          // 605
                                                                                                                       //
                                                                                                                       //
        for (var i = 0; i < meteorRequestListeners.length; i++) {                                                      // 607
          meteorRequestListeners[i](req, res);                                                                         // 608
        }                                                                                                              // 609
      } else {                                                                                                         // 610
        return inMeteor(function () {                                                                                  // 611
          if (globalDb.hostIsStandalone(hostname)) {                                                                   // 612
            // If it's a standalone host, also pass on to meteor                                                       // 613
            for (var _i = 0; _i < meteorRequestListeners.length; _i++) {                                               // 614
              meteorRequestListeners[_i](req, res);                                                                    // 615
            }                                                                                                          // 616
          } else {                                                                                                     // 617
            // Otherwise, dispatch to our own middleware proxy chain.                                                  // 618
            nonMeteorRequestHandler(req, res); // Adjust timeouts on proxied requests to allow apps to long-poll if needed.
                                                                                                                       //
            WebApp._timeoutAdjustmentRequestCallback(req, res);                                                        // 621
          }                                                                                                            // 622
        });                                                                                                            // 623
      }                                                                                                                // 624
    }).catch(function (e) {                                                                                            // 625
      // This should never be reached, because all the request handlers should be catching                             // 626
      // exceptions, but you can never be too careful in a top-level request handler.                                  // 627
      console.error("Unhandled exception in request handler:", e.stack);                                               // 628
      res.writeHead(500, {                                                                                             // 629
        "Content-Type": "text/plain"                                                                                   // 629
      });                                                                                                              // 629
      res.end("Unhandled exception: " + e.stack);                                                                      // 630
    });                                                                                                                // 631
  });                                                                                                                  // 632
  var meteorUpgradeListeners = WebApp.httpServer.listeners("upgrade");                                                 // 634
  WebApp.httpServer.removeAllListeners("upgrade");                                                                     // 635
  WebApp.httpServer.on("upgrade", function (req, socket, head) {                                                       // 637
    Promise.resolve(undefined).then(function () {                                                                      // 638
      if (!req.headers.host) {                                                                                         // 639
        throw new Meteor.Error(400, "Missing Host header");                                                            // 640
      } else if (isSandstormShell(req.headers.host.split(":")[0])) {                                                   // 641
        // Go on to Meteor.                                                                                            // 642
        for (var ii = 0; ii < meteorUpgradeListeners.length; ++ii) {                                                   // 643
          meteorUpgradeListeners[ii](req, socket, head);                                                               // 644
        }                                                                                                              // 645
                                                                                                                       //
        return true;                                                                                                   // 647
      } else {                                                                                                         // 648
        return inMeteor(function () {                                                                                  // 649
          // TODO(someday): This is slightly sad that some requests will now need a Meteor fiber                       // 650
          // If we disallow standalone domains from being in the wildcard host, then we can move                       // 651
          // this check to the bottom.                                                                                 // 652
          if (globalDb.hostIsStandalone(req.headers.host.split(":")[0])) {                                             // 653
            for (var _ii = 0; _ii < meteorUpgradeListeners.length; ++_ii) {                                            // 654
              meteorUpgradeListeners[_ii](req, socket, head);                                                          // 655
            }                                                                                                          // 656
                                                                                                                       //
            return true;                                                                                               // 658
          } else {                                                                                                     // 659
            var id = matchWildcardHost(req.headers.host);                                                              // 660
                                                                                                                       //
            if (id) {                                                                                                  // 661
              return tryProxyUpgrade(id, req, socket, head);                                                           // 662
            } else {                                                                                                   // 663
              return false;                                                                                            // 664
            }                                                                                                          // 665
          }                                                                                                            // 666
        });                                                                                                            // 667
      }                                                                                                                // 668
    }).then(function (handled) {                                                                                       // 669
      if (!handled) socket.destroy();                                                                                  // 670
    }).catch(function (err) {                                                                                          // 671
      console.error("WebSocket event handler failed:", err.stack);                                                     // 672
      socket.destroy();                                                                                                // 673
    });                                                                                                                // 674
  });                                                                                                                  // 675
  listenOnAlternatePorts();                                                                                            // 677
});                                                                                                                    // 678
var errorTxtMapping = {};                                                                                              // 680
errorTxtMapping[Dns.NOTFOUND] = "<p>" + "If you were trying to configure static publishing for a blog or website, powered " + "by a Sandstorm app hosted at this server, you either have not added DNS TXT records " + "correctly, or the DNS cache has not updated yet (may take a while, like 5 minutes to one " + "hour).</p>";
errorTxtMapping[Dns.NODATA] = errorTxtMapping[Dns.NOTFOUND];                                                           // 686
errorTxtMapping[Dns.TIMEOUT] = "<p>" + "The DNS query has timed out, which may be a sign of poorly configured DNS on the server.</p>";
errorTxtMapping[Dns.CONNREFUSED] = "<p>" + "The DNS server refused the connection, which means either your DNS server is down/unreachable, " + "or the server has misconfigured their DNS.</p>";
                                                                                                                       //
function lookupPublicIdFromDns(hostname) {                                                                             // 693
  // Given a hostname, determine its public ID.                                                                        // 694
  // We look for a TXT record indicating the public ID. Unfortunately, according to spec, a single                     // 695
  // hostname cannot have both a CNAME and a TXT record, because a TXT lookup on a CNAME'd host                        // 696
  // should actually be redirected to the CNAME, just like an A lookup would be. In practice lots                      // 697
  // of DNS software actually allows TXT records on CNAMEs, and it seems to work, but some software                    // 698
  // does not allow it and it's explicitly disallowed by the spec. Therefore, we instead look for                      // 699
  // the TXT record on a subdomain.                                                                                    // 700
  //                                                                                                                   // 701
  // I also considered having the CNAME itself point to <publicId>.<hostname>, where                                   // 702
  // *.<hostname> is in turn a CNAME for the Sandstorm server. This approach seemed elegant at                         // 703
  // first, but has a number of problems, the biggest being that it breaks the ability to place a                      // 704
  // CDN like CloudFlare in front of the site.                                                                         // 705
  var cache = dnsCache[hostname];                                                                                      // 707
                                                                                                                       //
  if (cache && Date.now() < cache.expiration) {                                                                        // 708
    return Promise.resolve(cache.value);                                                                               // 709
  }                                                                                                                    // 710
                                                                                                                       //
  return new Promise(function (resolve, reject) {                                                                      // 712
    Dns.resolveTxt("sandstorm-www." + hostname, function (err, records) {                                              // 713
      if (err) {                                                                                                       // 714
        var errorMsg = errorTxtMapping[err.code] || "";                                                                // 715
        var error = new Error('Error looking up DNS TXT records for host "' + hostname + '": ' + err.message);         // 716
        error.htmlMessage = '<style type="text/css">h2, h3, p { max-width: 600px; }</style>' + "<h2>Sandstorm static publishing needs further configuration (or wrong URL)</h2>" + errorMsg + "<p>To visit this Sandstorm server's main interface, go to: <a href='" + process.env.ROOT_URL + "'>" + process.env.ROOT_URL + "</a></p>" + "<h3>DNS details</h3>" + '<p>Error looking up DNS TXT records for host "' + hostname + '": ' + err.message + "</p>" + "<p>If you have the <tt>dig</tt> tool, you can run this command to learn more:</p>" + "<p><tt>dig TXT sandstorm-www." + hostname + "</tt></p>" + "<h3>Changing the server URL, or troubleshooting OAuth login</h3>" + "<p>If you are the server admin and want to use this address as the main interface, " + "edit /opt/sandstorm/sandstorm.conf, modify the BASE_URL setting, and restart " + "Sandstorm.</p>" + "<p>If you got here after trying to log in via OAuth (e.g. through GitHub or Google), " + "the problem is probably that the OAuth callback URL was set wrong. You need to " + "update it through the respective login provider's management console. The " + "easiest way to do that is to run <tt>sudo sandstorm admin-token</tt>, then " + "reconfigure the OAuth provider.</p>";
        error.httpErrorCode = _.contains(["ENOTFOUND", "ENODATA"], err.code) ? 404 : 500;                              // 737
        reject(error);                                                                                                 // 738
      } else if (records.length !== 1) {                                                                               // 739
        reject(new Error('Host "sandstorm-www.' + hostname + '" must have exactly one TXT record.'));                  // 740
      } else {                                                                                                         // 741
        // Under node 0.12+, `records` is an array-of-arrays. Each sub-array contains "chunks" of a                    // 742
        // record, which should be joined.                                                                             // 743
        var result = records[0].join("");                                                                              // 744
        dnsCache[hostname] = {                                                                                         // 745
          value: result,                                                                                               // 745
          expiration: Date.now() + DNS_CACHE_TTL                                                                       // 745
        };                                                                                                             // 745
        resolve(result);                                                                                               // 746
      }                                                                                                                // 747
    });                                                                                                                // 748
  });                                                                                                                  // 749
}                                                                                                                      // 750
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"proxy.js":["babel-runtime/helpers/classCallCheck","bignum","/imports/server/constants.js","/imports/server/backend.js","/imports/server/async-helpers.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/proxy.js                                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var _classCallCheck2 = require("babel-runtime/helpers/classCallCheck");                                                //
                                                                                                                       //
var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);                                                       //
                                                                                                                       //
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }                      //
                                                                                                                       //
var Bignum = void 0;                                                                                                   // 1
module.import("bignum", {                                                                                              // 1
  "default": function (v) {                                                                                            // 1
    Bignum = v;                                                                                                        // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var SANDSTORM_ALTHOME = void 0;                                                                                        // 1
module.import("/imports/server/constants.js", {                                                                        // 1
  "SANDSTORM_ALTHOME": function (v) {                                                                                  // 1
    SANDSTORM_ALTHOME = v;                                                                                             // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
var SandstormBackend = void 0,                                                                                         // 1
    shouldRestartGrain = void 0;                                                                                       // 1
module.import("/imports/server/backend.js", {                                                                          // 1
  "SandstormBackend": function (v) {                                                                                   // 1
    SandstormBackend = v;                                                                                              // 1
  },                                                                                                                   // 1
  "shouldRestartGrain": function (v) {                                                                                 // 1
    shouldRestartGrain = v;                                                                                            // 1
  }                                                                                                                    // 1
}, 2);                                                                                                                 // 1
var inMeteor = void 0,                                                                                                 // 1
    waitPromise = void 0;                                                                                              // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  },                                                                                                                   // 1
  "waitPromise": function (v) {                                                                                        // 1
    waitPromise = v;                                                                                                   // 1
  }                                                                                                                    // 1
}, 3);                                                                                                                 // 1
                                                                                                                       //
var Crypto = Npm.require("crypto");                                                                                    // 21
                                                                                                                       //
var ChildProcess = Npm.require("child_process");                                                                       // 22
                                                                                                                       //
var Fs = Npm.require("fs");                                                                                            // 23
                                                                                                                       //
var Path = Npm.require("path");                                                                                        // 24
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 25
                                                                                                                       //
var Capnp = Npm.require("capnp");                                                                                      // 26
                                                                                                                       //
var Net = Npm.require("net");                                                                                          // 27
                                                                                                                       //
var ByteStream = Capnp.importSystem("sandstorm/util.capnp").ByteStream;                                                // 29
var ApiSession = Capnp.importSystem("sandstorm/api-session.capnp").ApiSession;                                         // 30
var WebSession = Capnp.importSystem("sandstorm/web-session.capnp").WebSession;                                         // 31
var HackSession = Capnp.importSystem("sandstorm/hack-session.capnp");                                                  // 32
var Supervisor = Capnp.importSystem("sandstorm/supervisor.capnp").Supervisor;                                          // 33
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 34
var Powerbox = Capnp.importSystem("sandstorm/powerbox.capnp");                                                         // 35
var PARSED_ROOT_URL = Url.parse(process.env.ROOT_URL);                                                                 // 37
var PROTOCOL = PARSED_ROOT_URL.protocol;                                                                               // 38
var PORT = PARSED_ROOT_URL.port;                                                                                       // 39
                                                                                                                       //
var storeReferralProgramInfoApiTokenCreated = function (db, accountId, identityId, apiTokenAccountId) {                // 41
  // From the Referral program's perspective, if Bob's Account has no referredByComplete, then we                      // 42
  // update Bob's Identity to say it's referredBy Alice's Account (which is apiTokenAccountId).                        // 43
  check(accountId, String);                                                                                            // 44
  check(identityId, String);                                                                                           // 45
  check(apiTokenAccountId, String); // Bail out early if referrals aren't enabled                                      // 46
                                                                                                                       //
  if (!db.isReferralEnabled()) {                                                                                       // 49
    return;                                                                                                            // 50
  }                                                                                                                    // 51
                                                                                                                       //
  var aliceAccountId = apiTokenAccountId;                                                                              // 53
  var bobAccountId = accountId;                                                                                        // 54
  var bobIdentityId = identityId;                                                                                      // 55
                                                                                                                       //
  if (Meteor.users.find({                                                                                              // 57
    _id: bobAccountId,                                                                                                 // 58
    referredByComplete: {                                                                                              // 59
      $exists: true                                                                                                    // 59
    }                                                                                                                  // 59
  }).count() > 0) {                                                                                                    // 57
    return;                                                                                                            // 61
  } // Only actually update Bob's Identity ID if there is no referredBy.                                               // 62
                                                                                                                       //
                                                                                                                       //
  Meteor.users.update({                                                                                                // 65
    _id: bobIdentityId,                                                                                                // 66
    referredBy: {                                                                                                      // 66
      $exists: false                                                                                                   // 66
    }                                                                                                                  // 66
  }, {                                                                                                                 // 66
    $set: {                                                                                                            // 67
      referredBy: aliceAccountId                                                                                       // 67
    }                                                                                                                  // 67
  });                                                                                                                  // 67
};                                                                                                                     // 68
                                                                                                                       //
function referralProgramLogSharingTokenUse(db, bobAccountId) {                                                         // 70
  // Hooray! The sharing token is valid! Someone (let's call them Charlie) is going to get a UiView                    // 71
  // to this grain!  This means that the user who created this apiToken knows how to use the 'share                    // 72
  // access' interface. Let's call them Bob.                                                                           // 73
  //                                                                                                                   // 74
  // If Bob's Account.referredByComplete is not yet set, then look at Bob's Identities and take the                    // 75
  // first referredBy we find -- let's call that Alice.                                                                // 76
  //                                                                                                                   // 77
  // We copy Alice's account ID to Bob's Account.referredByComplete, and then update Alice's                           // 78
  // referredIdentityIds to point at Bob's Identity, and then remove the referredBy from Bob's                         // 79
  // Identity since it has become redundant.                                                                           // 80
  //                                                                                                                   // 81
  // Implementation note: this does mean that Alice can get referral credit for Bob by sharing a                       // 82
  // link with Bob, even if Bob already had an account.                                                                // 83
  // Bail out early if referrals aren't enabled                                                                        // 85
  if (!db.isReferralEnabled()) {                                                                                       // 86
    return;                                                                                                            // 87
  } // Bail out if Bob has a referredByComplete.                                                                       // 88
                                                                                                                       //
                                                                                                                       //
  if (Meteor.users.find({                                                                                              // 91
    _id: bobAccountId,                                                                                                 // 91
    referredByComplete: {                                                                                              // 91
      $exists: true                                                                                                    // 91
    }                                                                                                                  // 91
  }).count() > 0) {                                                                                                    // 91
    return;                                                                                                            // 92
  } // Look for a referredBy on any of Bob's identities.                                                               // 93
                                                                                                                       //
                                                                                                                       //
  var bobIdentityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({                                           // 96
    _id: bobAccountId                                                                                                  // 96
  }));                                                                                                                 // 96
  var bobIdentityWithReferredBy = Meteor.users.findOne({                                                               // 97
    _id: {                                                                                                             // 98
      $in: bobIdentityIds                                                                                              // 98
    },                                                                                                                 // 98
    referredBy: {                                                                                                      // 99
      $exists: true                                                                                                    // 99
    }                                                                                                                  // 99
  });                                                                                                                  // 97
                                                                                                                       //
  if (!bobIdentityWithReferredBy) {                                                                                    // 102
    return;                                                                                                            // 103
  }                                                                                                                    // 104
                                                                                                                       //
  var aliceAccountId = bobIdentityWithReferredBy.referredBy; // Store Bob's Account.referralCompletedBy.               // 106
                                                                                                                       //
  var now = new Date();                                                                                                // 109
  Meteor.users.update({                                                                                                // 110
    _id: bobAccountId,                                                                                                 // 111
    referredByComplete: {                                                                                              // 112
      $exists: false                                                                                                   // 112
    }                                                                                                                  // 112
  }, {                                                                                                                 // 110
    $set: {                                                                                                            // 114
      referredByComplete: bobIdentityWithReferredBy.referredBy,                                                        // 115
      referredCompleteDate: now                                                                                        // 116
    }                                                                                                                  // 114
  }); // Update Alice's Account.referredIdentityIds.                                                                   // 113
                                                                                                                       //
  Meteor.users.update({                                                                                                // 121
    _id: aliceAccountId                                                                                                // 121
  }, {                                                                                                                 // 121
    $push: {                                                                                                           // 122
      referredIdentityIds: bobIdentityWithReferredBy._id                                                               // 122
    }                                                                                                                  // 122
  }); // Remove now-redundant Bob identity referredBy.                                                                 // 121
                                                                                                                       //
  Meteor.users.update({                                                                                                // 126
    _id: bobIdentityWithReferredBy._id                                                                                 // 126
  }, {                                                                                                                 // 126
    $unset: {                                                                                                          // 127
      referredBy: true                                                                                                 // 127
    }                                                                                                                  // 127
  });                                                                                                                  // 126
}                                                                                                                      // 129
                                                                                                                       //
var HeaderWhitelist = function () {                                                                                    //
  function HeaderWhitelist(list) {                                                                                     // 132
    var _this2 = this;                                                                                                 // 132
                                                                                                                       //
    (0, _classCallCheck3.default)(this, HeaderWhitelist);                                                              // 132
    this._headers = {};                                                                                                // 133
    this._prefixes = [];                                                                                               // 134
    list.forEach(function (pattern) {                                                                                  // 136
      if (pattern.endsWith("*")) {                                                                                     // 137
        _this2._prefixes.push(pattern.slice(0, -1));                                                                   // 138
      } else {                                                                                                         // 139
        _this2._headers[pattern] = true;                                                                               // 140
      }                                                                                                                // 141
    });                                                                                                                // 142
  }                                                                                                                    // 143
                                                                                                                       //
  HeaderWhitelist.prototype.matches = function () {                                                                    //
    function matches(header) {                                                                                         //
      header = header.toLowerCase();                                                                                   // 146
      if (this._headers[header]) return true;                                                                          // 147
                                                                                                                       //
      for (var i in meteorBabelHelpers.sanitizeForInObject(this._prefixes)) {                                          // 149
        if (header.startsWith(this._prefixes[i])) {                                                                    // 150
          return true;                                                                                                 // 151
        }                                                                                                              // 152
      }                                                                                                                // 153
                                                                                                                       //
      ;                                                                                                                // 153
      return false;                                                                                                    // 155
    }                                                                                                                  // 156
                                                                                                                       //
    return matches;                                                                                                    //
  }();                                                                                                                 //
                                                                                                                       //
  return HeaderWhitelist;                                                                                              //
}();                                                                                                                   //
                                                                                                                       //
var REQUEST_HEADER_WHITELIST = new HeaderWhitelist(WebSession.Context.headerWhitelist);                                // 159
var RESPONSE_HEADER_WHITELIST = new HeaderWhitelist(WebSession.Response.headerWhitelist); // User-agent strings that should be allowed to use http basic authentication.
// These are regex matches, so ensure they are escaped properly with double                                            // 163
// backslashes. For security reasons, we MUST NOT whitelist any user-agents                                            // 164
// that may render html and execute embedded scripts.                                                                  // 165
                                                                                                                       //
BASIC_AUTH_USER_AGENTS = ["git\\/", "GitHub-Hookshot\\/", "mirall\\/", "Mozilla\\/5\\.0 \\([^\\\\]*\\) mirall\\/", "Mozilla\\/5\\.0 \\(iOS\\) ownCloud-iOS\\/", "Mozilla\\/5\\.0 \\(Android\\) ownCloud-android\\/", "litmus\\/"];
BASIC_AUTH_USER_AGENTS_REGEX = new RegExp("^(" + BASIC_AUTH_USER_AGENTS.join("|") + ")", "");                          // 175
var SESSION_PROXY_TIMEOUT = 60000;                                                                                     // 177
var sandstormCoreFactory = makeSandstormCoreFactory(globalDb);                                                         // 179
var backendAddress = "unix:" + (SANDSTORM_ALTHOME || "") + Backend.socketPath;                                         // 180
var sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);                                  // 181
var sandstormBackend = sandstormBackendConnection.restore(null, Backend); // TODO(cleanup): This initilization belongs with the rest of our package initialization in
//   db-deprecated.js. We can't put it there now because we need to contruct sandstormCoreFactory first.               // 185
                                                                                                                       //
globalBackend = new SandstormBackend(globalDb, sandstormBackend);                                                      // 186
Meteor.onConnection(function (connection) {                                                                            // 187
  connection.sandstormBackend = globalBackend;                                                                         // 188
}); // We've observed a problem in production where occasionally the front-end stops talking to the                    // 189
// back-end. It happens very rarely -- like once a month -- and we've been unable to reproduce it                      // 192
// in testing, making it very hard to debug. The problem appears both on Sandstorm and Blackrock.                      // 193
// Restarting the node process (and only the node process) always fixes the problem.                                   // 194
//                                                                                                                     // 195
// Here, I've added some code that attempts to detect the problem by doing a health check                              // 196
// periodically and then remaking the connection if it seems broken. We'll see if this helps!                          // 197
                                                                                                                       //
var backendHealthy = true;                                                                                             // 198
var disconnectCount = 0;                                                                                               // 199
Meteor.setInterval(function () {                                                                                       // 200
  if (!backendHealthy) {                                                                                               // 201
    if (disconnectCount++ > 2) process.abort();                                                                        // 202
    console.error("error: Backend hasn't responded in 30 seconds! Reconnecting.");                                     // 203
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(true);                                        // 204
    sandstormBackendConnection.close();                                                                                // 205
    sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);                                  // 206
    sandstormBackend = sandstormBackendConnection.restore(null, Backend);                                              // 207
    globalBackend._backendCap = sandstormBackend;                                                                      // 208
  }                                                                                                                    // 209
                                                                                                                       //
  var debugLog = !backendHealthy;                                                                                      // 211
  backendHealthy = false;                                                                                              // 212
  var promise = sandstormBackend.ping().then(function () {                                                             // 213
    backendHealthy = true;                                                                                             // 214
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(false);                                       // 215
  }, function (err) {                                                                                                  // 216
    console.error("error: Backend ping threw error!", err.stack); // The connection will be remade on the next interval. Note that we do NOT normally observe
    // exceptions being thrown for this problem; we see the connection simply stop responding.                         // 219
    // So we don't expect this branch to execute in any case.                                                          // 220
  });                                                                                                                  // 221
                                                                                                                       //
  if (debugLog) {                                                                                                      // 223
    console.log("capnp.js: outer promise:", promise);                                                                  // 224
  }                                                                                                                    // 225
}, 30000); // =======================================================================================                  // 226
// API for creating / starting grains from Meteor methods.                                                             // 229
                                                                                                                       //
var proxiesByHostId = {};                                                                                              // 231
                                                                                                                       //
function parsePowerboxDescriptorList(list) {                                                                           // 233
  return list.map(function (packedDescriptor) {                                                                        // 234
    return Capnp.parse(Powerbox.PowerboxDescriptor, new Buffer(packedDescriptor, "base64"), {                          // 234
      packed: true                                                                                                     // 236
    });                                                                                                                // 236
  });                                                                                                                  // 234
}                                                                                                                      // 237
                                                                                                                       //
Meteor.methods({                                                                                                       // 239
  newGrain: function (packageId, command, title, identityId) {                                                         // 240
    // Create and start a new grain.                                                                                   // 241
    check(packageId, String);                                                                                          // 243
    check(command, Object); // Manifest.Command from package.capnp.                                                    // 244
                                                                                                                       //
    check(title, String);                                                                                              // 245
    check(identityId, String);                                                                                         // 246
                                                                                                                       //
    if (!this.userId) {                                                                                                // 248
      throw new Meteor.Error(403, "Unauthorized", "Must be logged in to create grains.");                              // 249
    }                                                                                                                  // 250
                                                                                                                       //
    if (!globalDb.userHasIdentity(this.userId, identityId)) {                                                          // 252
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);                            // 253
    }                                                                                                                  // 254
                                                                                                                       //
    if (!isSignedUpOrDemo()) {                                                                                         // 256
      throw new Meteor.Error(403, "Unauthorized", "Only invited users or demo users can create grains.");              // 257
    }                                                                                                                  // 259
                                                                                                                       //
    if (isUserOverQuota(Meteor.user())) {                                                                              // 261
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");           // 262
    }                                                                                                                  // 264
                                                                                                                       //
    var pkg = Packages.findOne(packageId);                                                                             // 266
    var isDev = false;                                                                                                 // 267
    var mountProc = false;                                                                                             // 268
                                                                                                                       //
    if (!pkg) {                                                                                                        // 269
      // Maybe they wanted a dev package.  Check there too.                                                            // 270
      pkg = DevPackages.findOne(packageId);                                                                            // 271
      isDev = true;                                                                                                    // 272
      mountProc = pkg.mountProc;                                                                                       // 273
    }                                                                                                                  // 274
                                                                                                                       //
    if (!pkg) {                                                                                                        // 276
      throw new Meteor.Error(404, "Not Found", "No such package is installed.");                                       // 277
    }                                                                                                                  // 278
                                                                                                                       //
    var appId = pkg.appId;                                                                                             // 280
    var manifest = pkg.manifest;                                                                                       // 281
    var grainId = Random.id(22); // 128 bits of entropy                                                                // 282
                                                                                                                       //
    Grains.insert({                                                                                                    // 283
      _id: grainId,                                                                                                    // 284
      packageId: packageId,                                                                                            // 285
      appId: appId,                                                                                                    // 286
      appVersion: manifest.appVersion,                                                                                 // 287
      userId: this.userId,                                                                                             // 288
      identityId: identityId,                                                                                          // 289
      title: title,                                                                                                    // 290
      "private": true,                                                                                                 // 291
      size: 0                                                                                                          // 292
    });                                                                                                                // 283
    globalBackend.startGrainInternal(packageId, grainId, this.userId, command, true, isDev, mountProc);                // 295
    globalBackend.updateLastActive(grainId, this.userId, identityId);                                                  // 297
    return grainId;                                                                                                    // 298
  },                                                                                                                   // 299
  openSession: function (grainId, identityId, cachedSalt, options) {                                                   // 301
    // Open a new UI session on an existing grain.  Starts the grain if it is not already                              // 302
    // running.                                                                                                        // 303
    options = options || {};                                                                                           // 305
    check(grainId, String);                                                                                            // 307
    check(identityId, Match.OneOf(undefined, null, String));                                                           // 308
    check(cachedSalt, Match.OneOf(undefined, null, String));                                                           // 309
    check(options, {                                                                                                   // 310
      powerboxRequest: Match.Optional({                                                                                // 311
        descriptors: [String],                                                                                         // 312
        requestingSession: String                                                                                      // 313
      })                                                                                                               // 311
    });                                                                                                                // 310
                                                                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {                             // 317
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);                            // 318
    }                                                                                                                  // 319
                                                                                                                       //
    var grain = Grains.findOne({                                                                                       // 321
      _id: grainId                                                                                                     // 321
    });                                                                                                                // 321
                                                                                                                       //
    if (!grain) {                                                                                                      // 322
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + grainId);                                          // 323
    } else if (grain.trashed) {                                                                                        // 324
      throw new Meteor.Error("grain-is-in-trash", "Grain is in trash", "Grain ID: " + grainId);                        // 325
    } else if (grain.suspended) {                                                                                      // 326
      throw new Meteor.Error("grain-owner-suspended", "Grain's owner is suspended", "Grain ID: " + grainId);           // 327
    }                                                                                                                  // 329
                                                                                                                       //
    var db = this.connection.sandstormDb;                                                                              // 331
    check(cachedSalt, Match.OneOf(undefined, null, String));                                                           // 332
                                                                                                                       //
    if (!SandstormPermissions.mayOpenGrain(db, {                                                                       // 333
      grain: {                                                                                                         // 334
        _id: grainId,                                                                                                  // 334
        identityId: identityId                                                                                         // 334
      }                                                                                                                // 334
    })) {                                                                                                              // 334
      throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");                       // 335
    }                                                                                                                  // 336
                                                                                                                       //
    var sessionFields = {};                                                                                            // 338
                                                                                                                       //
    if (options.powerboxRequest) {                                                                                     // 339
      sessionFields.powerboxRequest = {                                                                                // 340
        descriptors: parsePowerboxDescriptorList(options.powerboxRequest.descriptors),                                 // 341
        requestingSession: options.powerboxRequest.requestingSession                                                   // 342
      };                                                                                                               // 340
    }                                                                                                                  // 344
                                                                                                                       //
    var opened = globalBackend.openSessionInternal(grainId, this.userId, identityId, null, null, cachedSalt, sessionFields);
    var result = opened.methodResult;                                                                                  // 348
    var proxy = new Proxy(grain, result.sessionId, result.hostId, result.tabId, identityId, false, null, opened.supervisor);
                                                                                                                       //
    if (sessionFields.powerboxRequest) {                                                                               // 353
      proxy.powerboxRequest = sessionFields.powerboxRequest;                                                           // 354
    }                                                                                                                  // 355
                                                                                                                       //
    proxiesByHostId[result.hostId] = proxy;                                                                            // 357
    return result;                                                                                                     // 358
  },                                                                                                                   // 359
  openSessionFromApiToken: function (params, identityId, cachedSalt, neverRedeem, parentOrigin, options) {             // 361
    // Given an API token, either opens a new WebSession to the underlying grain or returns a                          // 362
    // path to which the client should redirect in order to open such a session.                                       // 363
    //                                                                                                                 // 364
    // `parentOrigin` is the origin of the parent frame of the grain iframe. This may differ from                      // 365
    // ROOT_URL for standalone domains.                                                                                // 366
    neverRedeem = neverRedeem || false;                                                                                // 368
    parentOrigin = parentOrigin || process.env.ROOT_URL;                                                               // 369
    options = options || {};                                                                                           // 370
    check(params, {                                                                                                    // 372
      token: String,                                                                                                   // 373
      incognito: Match.Optional(Boolean)                                                                               // 374
    });                                                                                                                // 372
    check(identityId, Match.OneOf(undefined, null, String));                                                           // 376
    check(cachedSalt, Match.OneOf(undefined, null, String));                                                           // 377
    check(neverRedeem, Boolean);                                                                                       // 378
    check(parentOrigin, String);                                                                                       // 379
    check(options, {                                                                                                   // 380
      powerboxRequest: Match.Optional({                                                                                // 381
        descriptors: [String],                                                                                         // 382
        requestingSession: String                                                                                      // 383
      })                                                                                                               // 381
    });                                                                                                                // 380
                                                                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {                             // 387
      throw new Meteor.Error(403, "Current user does not own the identity: " + identityId);                            // 388
    }                                                                                                                  // 389
                                                                                                                       //
    if (!identityId && globalDb.getOrganizationDisallowGuests()) {                                                     // 391
      throw new Meteor.Error("guestDisallowed", "This Sandstorm server does not allow " + "guests or anonymous users");
    }                                                                                                                  // 394
                                                                                                                       //
    var token = params.token;                                                                                          // 396
    var incognito = !identityId;                                                                                       // 397
    var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");                                      // 398
    var apiToken = ApiTokens.findOne(hashedToken);                                                                     // 399
    validateWebkey(apiToken);                                                                                          // 400
    var grain = Grains.findOne({                                                                                       // 401
      _id: apiToken.grainId                                                                                            // 401
    });                                                                                                                // 401
                                                                                                                       //
    if (!grain) {                                                                                                      // 402
      throw new Meteor.Error(404, "Grain not found", "Grain ID: " + apiToken.grainId);                                 // 403
    }                                                                                                                  // 404
                                                                                                                       //
    if (apiToken.accountId) {                                                                                          // 406
      referralProgramLogSharingTokenUse(globalDb, apiToken.accountId);                                                 // 407
    }                                                                                                                  // 408
                                                                                                                       //
    var title = void 0;                                                                                                // 410
                                                                                                                       //
    if (grain.userId === apiToken.accountId) {                                                                         // 411
      title = grain.title;                                                                                             // 412
    } else {                                                                                                           // 413
      var sharerToken = apiToken.identityId && ApiTokens.findOne({                                                     // 414
        grainId: apiToken.grainId,                                                                                     // 416
        "owner.user.identityId": apiToken.identityId                                                                   // 417
      }, {                                                                                                             // 415
        sort: {                                                                                                        // 419
          lastUsed: -1                                                                                                 // 420
        }                                                                                                              // 419
      });                                                                                                              // 418
                                                                                                                       //
      if (sharerToken) {                                                                                               // 423
        title = sharerToken.owner.user.title;                                                                          // 424
      } else {                                                                                                         // 425
        title = "shared grain";                                                                                        // 426
      }                                                                                                                // 427
    }                                                                                                                  // 428
                                                                                                                       //
    if (this.userId && !incognito && !neverRedeem) {                                                                   // 430
      if (identityId != apiToken.identityId && identityId != grain.identityId && !ApiTokens.findOne({                  // 431
        "owner.user.identityId": identityId,                                                                           // 432
        parentToken: hashedToken                                                                                       // 432
      })) {                                                                                                            // 432
        var owner = {                                                                                                  // 433
          user: {                                                                                                      // 433
            identityId: identityId,                                                                                    // 433
            title: title                                                                                               // 433
          }                                                                                                            // 433
        }; // Create a new API token for the identity redeeming this token.                                            // 433
                                                                                                                       //
        var result = SandstormPermissions.createNewApiToken(globalDb, {                                                // 436
          rawParentToken: token                                                                                        // 437
        }, apiToken.grainId, apiToken.petname || "redeemed webkey", {                                                  // 437
          allAccess: null                                                                                              // 439
        }, owner);                                                                                                     // 439
        globalDb.addContact(apiToken.accountId, identityId); // If the parent API token is forSharing and it has an accountId, then the logged-in user (call
        // them Bob) is about to access a grain owned by someone (call them Alice) and save a reference                // 443
        // to it as a new ApiToken. (For share-by-link, this occurs when viewing the grain. For                        // 444
        // share-by-identity, this happens immediately.)                                                               // 445
                                                                                                                       //
        if (result.parentApiToken) {                                                                                   // 446
          var parentApiToken = result.parentApiToken;                                                                  // 447
                                                                                                                       //
          if (parentApiToken.forSharing && parentApiToken.accountId) {                                                 // 448
            storeReferralProgramInfoApiTokenCreated(globalDb, this.userId, owner.user.identityId, parentApiToken.accountId);
          }                                                                                                            // 451
        }                                                                                                              // 452
      }                                                                                                                // 453
                                                                                                                       //
      return {                                                                                                         // 455
        redirectToGrain: apiToken.grainId                                                                              // 455
      };                                                                                                               // 455
    } else {                                                                                                           // 456
      if (!SandstormPermissions.mayOpenGrain(globalDb, {                                                               // 457
        token: apiToken                                                                                                // 457
      })) {                                                                                                            // 457
        throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");                     // 458
      }                                                                                                                // 460
                                                                                                                       //
      var sessionFields = {};                                                                                          // 462
                                                                                                                       //
      if (options.powerboxRequest) {                                                                                   // 463
        sessionFields.powerboxRequest = {                                                                              // 464
          descriptors: parsePowerboxDescriptorList(options.powerboxRequest.descriptors),                               // 465
          requestingSession: options.powerboxRequest.requestingSession                                                 // 466
        };                                                                                                             // 464
      } // Even in incognito mode, we want to record the user ID on the session. The user ID is not                    // 468
      // leaked to the app in any way -- it is used to decide what options to show in a powerbox                       // 471
      // request. Even if the user is incognito, they should be able to offer their full range of                      // 472
      // capabilities.                                                                                                 // 473
      //                                                                                                               // 474
      // The identity ID passed here IS revealed to the app, but for incognito mode it is always                       // 475
      // null/undefined.                                                                                               // 476
                                                                                                                       //
                                                                                                                       //
      var opened = globalBackend.openSessionInternal(apiToken.grainId, this.userId, identityId, title, apiToken, cachedSalt, sessionFields);
      var _result = opened.methodResult;                                                                               // 480
      var proxy = new Proxy(grain, _result.sessionId, _result.hostId, _result.tabId, identityId, false, parentOrigin, opened.supervisor);
      proxy.apiToken = apiToken;                                                                                       // 484
                                                                                                                       //
      if (sessionFields.powerboxRequest) {                                                                             // 485
        proxy.powerboxRequest = sessionFields.powerboxRequest;                                                         // 486
      }                                                                                                                // 487
                                                                                                                       //
      proxiesByHostId[_result.hostId] = proxy;                                                                         // 489
      return _result;                                                                                                  // 490
    }                                                                                                                  // 491
  },                                                                                                                   // 492
  keepSessionAlive: function (sessionId) {                                                                             // 494
    // TODO(security):  Prevent draining someone else's quota by holding open several grains shared                    // 495
    //   by them.                                                                                                      // 496
    check(sessionId, String);                                                                                          // 497
    var result = Sessions.findAndModify({                                                                              // 499
      query: {                                                                                                         // 500
        _id: sessionId                                                                                                 // 500
      },                                                                                                               // 500
      update: {                                                                                                        // 501
        $set: {                                                                                                        // 501
          timestamp: new Date().getTime()                                                                              // 501
        }                                                                                                              // 501
      },                                                                                                               // 501
      fields: {                                                                                                        // 502
        grainId: 1,                                                                                                    // 502
        identityId: 1,                                                                                                 // 502
        hostId: 1                                                                                                      // 502
      }                                                                                                                // 502
    });                                                                                                                // 499
                                                                                                                       //
    if (!result.ok) {                                                                                                  // 505
      return false;                                                                                                    // 506
    }                                                                                                                  // 507
                                                                                                                       //
    var session = result.value;                                                                                        // 509
                                                                                                                       //
    if (session) {                                                                                                     // 510
      // Session still present in database, so send keep-alive to backend.                                             // 511
      try {                                                                                                            // 512
        var grainId = session.grainId;                                                                                 // 513
        var hostId = session.hostId;                                                                                   // 514
        var supervisor = proxiesByHostId[hostId] && proxiesByHostId[hostId].supervisor;                                // 515
                                                                                                                       //
        if (!supervisor) {                                                                                             // 516
          supervisor = globalBackend.continueGrain(grainId).supervisor;                                                // 517
        }                                                                                                              // 518
                                                                                                                       //
        waitPromise(supervisor.keepAlive());                                                                           // 520
        globalBackend.updateLastActive(grainId, this.userId, session.identityId);                                      // 521
      } catch (err) {                                                                                                  // 522
        // Ignore disconnects, which imply that the grain shut down already. It'll start back up on                    // 523
        // the next request, so whatever.                                                                              // 524
        if (err.kjType !== "disconnected") {                                                                           // 525
          throw err;                                                                                                   // 526
        }                                                                                                              // 527
      }                                                                                                                // 528
                                                                                                                       //
      return true;                                                                                                     // 530
    } else {                                                                                                           // 531
      return false;                                                                                                    // 532
    }                                                                                                                  // 533
  },                                                                                                                   // 534
  shutdownGrain: function (grainId) {                                                                                  // 536
    check(grainId, String);                                                                                            // 537
    var grain = Grains.findOne(grainId);                                                                               // 538
                                                                                                                       //
    if (!grain || !this.userId || grain.userId !== this.userId) {                                                      // 539
      throw new Meteor.Error(403, "Unauthorized", "User is not the owner of this grain");                              // 540
    }                                                                                                                  // 541
                                                                                                                       //
    waitPromise(globalBackend.shutdownGrain(grainId, grain.userId, true));                                             // 543
  }                                                                                                                    // 544
});                                                                                                                    // 239
                                                                                                                       //
var validateWebkey = function (apiToken, refreshedExpiration) {                                                        // 547
  // Validates that `apiToken` is a valid UiView webkey, throwing an exception if it is not. If                        // 548
  // `refreshedExpiration` is set and if the token has an `expiresIfUnused` field, then the                            // 549
  // `expiresIfUnused` field is reset to `refreshedExpiration`.                                                        // 550
  if (!apiToken) {                                                                                                     // 552
    throw new Meteor.Error(403, "Invalid authorization token");                                                        // 553
  }                                                                                                                    // 554
                                                                                                                       //
  if (apiToken.revoked) {                                                                                              // 556
    throw new Meteor.Error(403, "Authorization token has been revoked");                                               // 557
  }                                                                                                                    // 558
                                                                                                                       //
  if (apiToken.owner && !("webkey" in apiToken.owner)) {                                                               // 560
    throw new Meteor.Error(403, "Unauthorized to open non-webkey token.");                                             // 561
  }                                                                                                                    // 562
                                                                                                                       //
  if (apiToken.expires && apiToken.expires.getTime() <= Date.now()) {                                                  // 564
    throw new Meteor.Error(403, "Authorization token expired");                                                        // 565
  }                                                                                                                    // 566
                                                                                                                       //
  if (apiToken.expiresIfUnused) {                                                                                      // 568
    if (apiToken.expiresIfUnused.getTime() <= Date.now()) {                                                            // 569
      throw new Meteor.Error(403, "Authorization token expired");                                                      // 570
    } else if (refreshedExpiration) {                                                                                  // 571
      ApiTokens.update(apiToken._id, {                                                                                 // 572
        $set: {                                                                                                        // 572
          expiresIfUnused: refreshedExpiration                                                                         // 572
        }                                                                                                              // 572
      });                                                                                                              // 572
    } else {                                                                                                           // 573
      // It's getting used now, so clear the expiresIfUnused field.                                                    // 574
      ApiTokens.update(apiToken._id, {                                                                                 // 575
        $set: {                                                                                                        // 575
          expiresIfUnused: null                                                                                        // 575
        }                                                                                                              // 575
      });                                                                                                              // 575
    }                                                                                                                  // 576
  }                                                                                                                    // 577
                                                                                                                       //
  if (apiToken.objectId || apiToken.frontendRef) {                                                                     // 579
    throw new Meteor.Error(403, "ApiToken refers to a non-webview Capability.");                                       // 580
  }                                                                                                                    // 581
};                                                                                                                     // 582
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 584
  var shutdownApp = function (appId) {                                                                                 // 585
    Grains.find({                                                                                                      // 586
      appId: appId                                                                                                     // 586
    }).forEach(function (grain) {                                                                                      // 586
      waitPromise(globalBackend.shutdownGrain(grain._id, grain.userId));                                               // 587
    });                                                                                                                // 588
  };                                                                                                                   // 589
                                                                                                                       //
  DevPackages.find().observe({                                                                                         // 591
    removed: function (devPackage) {                                                                                   // 592
      shutdownApp(devPackage.appId);                                                                                   // 592
    },                                                                                                                 // 592
    changed: function (newDevPackage, oldDevPackage) {                                                                 // 594
      shutdownApp(oldDevPackage.appId);                                                                                // 595
                                                                                                                       //
      if (oldDevPackage.appId !== newDevPackage.appId) {                                                               // 596
        shutdownApp(newDevPackage.appId);                                                                              // 597
      }                                                                                                                // 598
    },                                                                                                                 // 599
    added: function (devPackage) {                                                                                     // 601
      shutdownApp(devPackage.appId);                                                                                   // 601
    }                                                                                                                  // 601
  });                                                                                                                  // 591
  Sessions.find().observe({                                                                                            // 604
    removed: function (session) {                                                                                      // 605
      var proxy = proxiesByHostId[session.hostId];                                                                     // 606
      delete proxiesByHostId[session.hostId];                                                                          // 607
                                                                                                                       //
      if (proxy) {                                                                                                     // 608
        proxy.close();                                                                                                 // 609
      }                                                                                                                // 610
    }                                                                                                                  // 611
  });                                                                                                                  // 604
}); // Kill off sessions idle for >~3 minutes.                                                                         // 613
                                                                                                                       //
var TIMEOUT_MS = 180000;                                                                                               // 616
                                                                                                                       //
var gcSessions = function () {                                                                                         // 617
  var now = new Date().getTime();                                                                                      // 618
  Sessions.remove({                                                                                                    // 619
    timestamp: {                                                                                                       // 619
      $lt: now - TIMEOUT_MS                                                                                            // 619
    }                                                                                                                  // 619
  });                                                                                                                  // 619
};                                                                                                                     // 620
                                                                                                                       //
SandstormDb.periodicCleanup(TIMEOUT_MS, gcSessions);                                                                   // 622
                                                                                                                       //
var getProxyForHostId = function (hostId, isAlreadyOpened) {                                                           // 624
  // Get the Proxy corresponding to the given grain session host, possibly (re)creating it if it                       // 625
  // doesn't already exist. The first request on the session host will always create a new proxy.                      // 626
  // Later requests may create a proxy if they go to a different front-end replica or if the                           // 627
  // front-end was restarted.                                                                                          // 628
  check(hostId, String);                                                                                               // 629
  return Promise.resolve(undefined).then(function () {                                                                 // 631
    var proxy = proxiesByHostId[hostId];                                                                               // 632
                                                                                                                       //
    if (proxy) {                                                                                                       // 633
      return proxy;                                                                                                    // 634
    } else {                                                                                                           // 635
      // Set table entry to null for now so that we can detect if it is concurrently deleted.                          // 636
      proxiesByHostId[hostId] = null;                                                                                  // 637
      return inMeteor(function () {                                                                                    // 639
        var session = Sessions.findOne({                                                                               // 640
          hostId: hostId                                                                                               // 640
        });                                                                                                            // 640
                                                                                                                       //
        if (!session) {                                                                                                // 641
          if (isAlreadyOpened) {                                                                                       // 642
            var observer = void 0;                                                                                     // 643
            return new Promise(function (resolve, reject) {                                                            // 644
              var task = Meteor.setTimeout(function () {                                                               // 645
                reject(new Meteor.Error(504, "Requested session that no longer exists, and " + "timed out waiting for client to restore it. This can happen if you have " + "opened an app's content in a new window and then closed it in the " + "UI. If you see this error *inside* the Sandstorm UI, please report a " + "bug and describe the circumstances of the error."));
              }, SESSION_PROXY_TIMEOUT);                                                                               // 651
              observer = Sessions.find({                                                                               // 652
                hostId: hostId                                                                                         // 652
              }).observe({                                                                                             // 652
                added: function () {                                                                                   // 653
                  Meteor.clearTimeout(task);                                                                           // 654
                  resolve(getProxyForHostId(hostId, false));                                                           // 655
                }                                                                                                      // 656
              });                                                                                                      // 652
            }).then(function (v) {                                                                                     // 658
              observer.stop();                                                                                         // 659
              return v;                                                                                                // 660
            }, function (e) {                                                                                          // 661
              observer.stop();                                                                                         // 662
              throw e;                                                                                                 // 663
            });                                                                                                        // 664
          } else {                                                                                                     // 665
            // Does not appear to be a valid session host.                                                             // 666
            return undefined;                                                                                          // 667
          }                                                                                                            // 668
        }                                                                                                              // 669
                                                                                                                       //
        var apiToken = void 0;                                                                                         // 671
                                                                                                                       //
        if (session.hashedToken) {                                                                                     // 672
          apiToken = ApiTokens.findOne({                                                                               // 673
            _id: session.hashedToken                                                                                   // 673
          }); // We don't have to fully validate the API token here because if it changed the session                  // 673
          // would have been deleted.                                                                                  // 675
                                                                                                                       //
          if (!apiToken) {                                                                                             // 676
            throw new Meteor.Error(410, "ApiToken has been deleted");                                                  // 677
          }                                                                                                            // 678
        }                                                                                                              // 679
                                                                                                                       //
        var grain = Grains.findOne(session.grainId);                                                                   // 681
                                                                                                                       //
        if (!grain) {                                                                                                  // 682
          // Grain was deleted, I guess.                                                                               // 683
          throw new Meteor.Error(410, "Resource has been deleted");                                                    // 684
        } // Note that we don't need to call mayOpenGrain() because the existence of a session                         // 685
        // implies this check was already performed.                                                                   // 688
                                                                                                                       //
                                                                                                                       //
        var proxy = new Proxy(grain, session._id, hostId, session.tabId, session.identityId, false);                   // 690
        if (apiToken) proxy.apiToken = apiToken;                                                                       // 692
                                                                                                                       //
        if (session.powerboxRequest) {                                                                                 // 694
          proxy.powerboxRequest = session.powerboxRequest;                                                             // 695
        } // Only add the proxy to the table if it was not concurrently deleted (which could happen                    // 696
        // e.g. if the user's access was revoked).                                                                     // 699
                                                                                                                       //
                                                                                                                       //
        if (hostId in proxiesByHostId) {                                                                               // 700
          proxiesByHostId[hostId] = proxy;                                                                             // 701
        } else {                                                                                                       // 702
          throw new Meteor.Error(403, "Session was concurrently closed.");                                             // 703
        }                                                                                                              // 704
                                                                                                                       //
        return proxy;                                                                                                  // 706
      });                                                                                                              // 707
    }                                                                                                                  // 708
  });                                                                                                                  // 709
}; // =======================================================================================                          // 710
// API tokens                                                                                                          // 713
                                                                                                                       //
                                                                                                                       //
var ApiSessionProxies = function () {                                                                                  //
  // Class that caches proxies for requests that come in through HTTP API endpoints. Such                              // 716
  // requests are not associated with any entry in the `Sessions` collection, so we need                               // 717
  // some other means of detecting unused proxies and closing them. This class works by keeping                        // 718
  // the proxies in two buckets. When a proxy is used, it is put in the new bucket. Periodically,                      // 719
  // the old bucket is emptied and the contents of new bucket are moved to the old bucket.                             // 720
  //                                                                                                                   // 721
  // Each bucket is a two-level map, keyed by hashes of the tokens and their `ApiSession.Params`                       // 722
  // structs.                                                                                                          // 723
  function ApiSessionProxies(intervalMillis) {                                                                         // 725
    var _this3 = this;                                                                                                 // 725
                                                                                                                       //
    (0, _classCallCheck3.default)(this, ApiSessionProxies);                                                            // 725
    check(intervalMillis, Number);                                                                                     // 726
    this.newBucket = {};                                                                                               // 727
    this.oldBucket = {};                                                                                               // 728
    this.interval = Meteor.setInterval(function () {                                                                   // 729
      for (var oldHashedToken in meteorBabelHelpers.sanitizeForInObject(_this3.oldBucket)) {                           // 730
        for (var oldHashedParams in meteorBabelHelpers.sanitizeForInObject(_this3.oldBucket[oldHashedToken])) {        // 731
          var newProxy = _this3.newBucket[oldHashedToken] && _this3.newBucket[oldHashedToken][oldHashedParams];        // 732
          var oldProxy = _this3.oldBucket[oldHashedToken][oldHashedParams];                                            // 734
                                                                                                                       //
          if (oldProxy && !newProxy) {                                                                                 // 735
            if (Object.keys(oldProxy.websockets).length > 0 || Object.keys(oldProxy.streamingRequests).length > 0 || Object.keys(oldProxy.streamingResponses).length > 0) {
              // A client has an open stream. Keep the proxy around.                                                   // 739
              _this3.put(oldHashedToken, oldHashedParams, oldProxy);                                                   // 740
            } else {                                                                                                   // 741
              // We can close this proxy and forget about it.                                                          // 742
              oldProxy.close();                                                                                        // 743
            }                                                                                                          // 744
          }                                                                                                            // 745
        }                                                                                                              // 746
      }                                                                                                                // 747
                                                                                                                       //
      _this3.oldBucket = _this3.newBucket;                                                                             // 749
      _this3.newBucket = {};                                                                                           // 750
      var keepAlives = [];                                                                                             // 752
                                                                                                                       //
      for (var hashedToken in meteorBabelHelpers.sanitizeForInObject(_this3.oldBucket)) {                              // 753
        for (var hashedParams in meteorBabelHelpers.sanitizeForInObject(_this3.oldBucket[hashedToken])) {              // 754
          var proxy = _this3.oldBucket[hashedToken][hashedParams];                                                     // 755
          keepAlives.push(proxy.supervisor.keepAlive().catch(function (e) {                                            // 756
            console.error("supervisor.keepAlive() failed.", e);                                                        // 757
          }));                                                                                                         // 758
        }                                                                                                              // 759
      }                                                                                                                // 760
                                                                                                                       //
      waitPromise(Promise.all(keepAlives));                                                                            // 762
    }, intervalMillis);                                                                                                // 763
  }                                                                                                                    // 764
                                                                                                                       //
  ApiSessionProxies.prototype.get = function () {                                                                      //
    function get(hashedToken, hashedParams) {                                                                          //
      if (this.newBucket[hashedToken] && this.newBucket[hashedToken][hashedParams]) {                                  // 767
        return this.newBucket[hashedToken][hashedParams];                                                              // 768
      } else if (this.oldBucket[hashedToken] && this.oldBucket[hashedToken][hashedParams]) {                           // 769
        var proxy = this.oldBucket[hashedToken][hashedParams];                                                         // 770
        delete this.oldBucket[hashedToken][hashedParams];                                                              // 771
        this.put(hashedToken, hashedParams, proxy);                                                                    // 772
        return proxy;                                                                                                  // 773
      } else {                                                                                                         // 774
        return null;                                                                                                   // 775
      }                                                                                                                // 776
    }                                                                                                                  // 777
                                                                                                                       //
    return get;                                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  ApiSessionProxies.prototype.put = function () {                                                                      //
    function put(hashedToken, hashedParams, proxy) {                                                                   //
      if (!this.newBucket[hashedToken]) {                                                                              // 780
        this.newBucket[hashedToken] = {};                                                                              // 781
      }                                                                                                                // 782
                                                                                                                       //
      this.newBucket[hashedToken][hashedParams] = proxy;                                                               // 784
    }                                                                                                                  // 785
                                                                                                                       //
    return put;                                                                                                        //
  }();                                                                                                                 //
                                                                                                                       //
  ApiSessionProxies.prototype.removeProxiesOfToken = function () {                                                     //
    function removeProxiesOfToken(hashedToken) {                                                                       //
      if (hashedToken in this.oldBucket) {                                                                             // 788
        for (var hashedParams in meteorBabelHelpers.sanitizeForInObject(this.oldBucket[hashedToken])) {                // 789
          this.oldBucket[hashedToken][hashedParams].close();                                                           // 790
        }                                                                                                              // 791
                                                                                                                       //
        delete this.oldBucket[hashedToken];                                                                            // 793
      }                                                                                                                // 794
                                                                                                                       //
      if (hashedToken in this.newBucket) {                                                                             // 796
        for (var _hashedParams in meteorBabelHelpers.sanitizeForInObject(this.newBucket[hashedToken])) {               // 797
          this.newBucket[hashedToken][_hashedParams].close();                                                          // 798
        }                                                                                                              // 799
                                                                                                                       //
        delete this.newBucket[hashedToken];                                                                            // 801
      }                                                                                                                // 802
    }                                                                                                                  // 803
                                                                                                                       //
    return removeProxiesOfToken;                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  return ApiSessionProxies;                                                                                            //
}(); // The interval of 60 seconds here deliberately matches the `keepSessionAlive()` interval                         //
// in the case of non-API proxies.                                                                                     // 807
                                                                                                                       //
                                                                                                                       //
var apiSessionProxies = new ApiSessionProxies(60 * 1000);                                                              // 808
Meteor.startup(function () {                                                                                           // 810
  Grains.find().observe({                                                                                              // 811
    changed: function (newGrain, oldGrain) {                                                                           // 812
      if (oldGrain.private != newGrain.private) {                                                                      // 813
        Sessions.remove({                                                                                              // 814
          grainId: oldGrain._id,                                                                                       // 814
          identityId: {                                                                                                // 814
            $ne: oldGrain.identityId                                                                                   // 814
          }                                                                                                            // 814
        });                                                                                                            // 814
        ApiTokens.find({                                                                                               // 815
          grainId: oldGrain._id                                                                                        // 815
        }).forEach(function (apiToken) {                                                                               // 815
          apiSessionProxies.removeProxiesOfToken(apiToken._id);                                                        // 816
        });                                                                                                            // 817
      }                                                                                                                // 818
    }                                                                                                                  // 819
  });                                                                                                                  // 811
});                                                                                                                    // 821
                                                                                                                       //
function getApiSessionParams(request) {                                                                                // 823
  var params = {};                                                                                                     // 824
                                                                                                                       //
  if ("x-sandstorm-passthrough" in request.headers) {                                                                  // 825
    var optIns = request.headers["x-sandstorm-passthrough"].split(",").map(function (s) {                              // 826
      return s.trim();                                                                                                 // 828
    }); // The only currently supported passthrough value is 'address', but others could be useful in                  // 828
    // the future.                                                                                                     // 830
                                                                                                                       //
    if (optIns.indexOf("address") !== -1) {                                                                            // 832
      // Sadly, we can't use request.socket.remoteFamily because it's not available in the                             // 833
      // (rather-old) version of node that comes in the Meteor bundle we're using. Hence this                          // 834
      // hackery.                                                                                                      // 835
      var addressToPass = request.socket.remoteAddress;                                                                // 836
                                                                                                                       //
      if (isRfc1918OrLocal(addressToPass) && "x-real-ip" in request.headers) {                                         // 837
        // Allow overriding the socket's remote address with X-Real-IP header if the request comes                     // 838
        // from either localhost or an RFC1918 address. These are not useful for geolocation anyway.                   // 839
        addressToPass = request.headers["x-real-ip"];                                                                  // 840
      }                                                                                                                // 841
                                                                                                                       //
      if (Net.isIPv4(addressToPass)) {                                                                                 // 843
        // Map IPv4 addresses in IPv6.                                                                                 // 844
        // This conveniently comes out to a 48-bit number, which is precisely representable in a                       // 845
        // double (which has 53 mantissa bits). Thus we can avoid using Bignum/strings, which we                       // 846
        // might otherwise need to precisely represent 64-bit fields.                                                  // 847
        var v4Int = 0xFFFF00000000 + addressToPass.split(".").map(function (x) {                                       // 848
          return parseInt(x, 10);                                                                                      // 849
        }).reduce(function (a, b) {                                                                                    // 849
          return 256 * a + b;                                                                                          // 850
        });                                                                                                            // 850
        params.remoteAddress = {                                                                                       // 851
          lower64: v4Int,                                                                                              // 852
          upper64: 0                                                                                                   // 853
        };                                                                                                             // 851
      } else if (Net.isIPv6(addressToPass)) {                                                                          // 855
        // TODO(test): Unit test this                                                                                  // 856
        // Parse a valid v6 address.                                                                                   // 857
        // Split into groups, then insert an appropriate number of 0's if :: was used.                                 // 858
        var groups = addressToPass.split(":"); // Strip extra empty group in the case of a leading or trailing '::'.   // 859
                                                                                                                       //
        if (groups[0] === "") {                                                                                        // 862
          groups.shift();                                                                                              // 863
        }                                                                                                              // 864
                                                                                                                       //
        if (groups[groups.length - 1] === "") {                                                                        // 866
          groups.pop();                                                                                                // 867
        }                                                                                                              // 868
                                                                                                                       //
        var lastGroup = groups[groups.length - 1]; // Handle IPv4-mapped IPv6 addresses.  These end in a dotted-quad IPv4 address, which we
        // should expand into two groups of 4-character hex strings, like the rest of the address.                     // 872
                                                                                                                       //
        if (Net.isIPv4(lastGroup)) {                                                                                   // 873
          groups.pop();                                                                                                // 874
          var quad = lastGroup.split(".").map(function (x) {                                                           // 875
            return parseInt(x, 10);                                                                                    // 875
          });                                                                                                          // 875
          groups.push((quad[0] * 256 + quad[1]).toString(16));                                                         // 876
          groups.push((quad[2] * 256 + quad[3]).toString(16));                                                         // 877
        }                                                                                                              // 878
                                                                                                                       //
        var groupsToAdd = 8 - groups.length;                                                                           // 880
        var emptyGroupIndex = groups.indexOf("");                                                                      // 881
        var cleanGroups = void 0;                                                                                      // 882
                                                                                                                       //
        if (emptyGroupIndex !== -1) {                                                                                  // 883
          var head = groups.slice(0, emptyGroupIndex); // groupsToAdd + 1 because we sliced out the empty element      // 884
                                                                                                                       //
          var mid = Array(groupsToAdd + 1);                                                                            // 886
                                                                                                                       //
          for (var i = 0; i < groupsToAdd + 1; i++) {                                                                  // 887
            mid[i] = "0";                                                                                              // 888
          }                                                                                                            // 889
                                                                                                                       //
          var tail = groups.slice(emptyGroupIndex + 1, groups.length);                                                 // 891
          cleanGroups = [].concat(head, mid, tail);                                                                    // 892
        } else {                                                                                                       // 893
          cleanGroups = groups;                                                                                        // 894
        }                                                                                                              // 895
                                                                                                                       //
        var ints = cleanGroups.map(function (x) {                                                                      // 897
          return parseInt(x, 16);                                                                                      // 897
        }); // We use strings because we'd lose data from loss of precision casting the 64-bit uints                   // 897
        // into 53-bit-mantissa doubles.                                                                               // 899
                                                                                                                       //
        params.remoteAddress = {                                                                                       // 900
          upper64: quadToIntString(ints.slice(0, 4)),                                                                  // 901
          lower64: quadToIntString(ints.slice(4, 8))                                                                   // 902
        };                                                                                                             // 900
      }                                                                                                                // 904
    }                                                                                                                  // 905
  }                                                                                                                    // 906
                                                                                                                       //
  return Capnp.serialize(ApiSession.Params, params);                                                                   // 908
} // Used by server/drivers/external-ui/view.js                                                                        // 909
                                                                                                                       //
                                                                                                                       //
getProxyForApiToken = function (token, request) {                                                                      // 912
  check(token, String);                                                                                                // 913
  var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");                                        // 914
  var tabId = Crypto.createHash("sha256").update("tab:").update(hashedToken).digest("hex").slice(0, 32);               // 915
  var serializedParams = getApiSessionParams(request);                                                                 // 917
  var hashedParams = Crypto.createHash("sha256").update(serializedParams).digest("base64");                            // 918
  return Promise.resolve(undefined).then(function () {                                                                 // 919
    var proxy = apiSessionProxies.get(hashedToken, hashedParams);                                                      // 920
                                                                                                                       //
    if (proxy) {                                                                                                       // 921
      if (proxy.expires && proxy.expires.getTime() <= Date.now()) {                                                    // 922
        throw new Meteor.Error(403, "Authorization token expired");                                                    // 923
      }                                                                                                                // 924
                                                                                                                       //
      return proxy;                                                                                                    // 926
    } else {                                                                                                           // 927
      return inMeteor(function () {                                                                                    // 928
        var tokenInfo = ApiTokens.findOne(hashedToken);                                                                // 929
        validateWebkey(tokenInfo);                                                                                     // 930
        var grain = Grains.findOne(tokenInfo.grainId);                                                                 // 932
                                                                                                                       //
        if (!grain) {                                                                                                  // 933
          // Grain was deleted, I guess.                                                                               // 934
          throw new Meteor.Error(410, "Resource has been deleted");                                                    // 935
        }                                                                                                              // 936
                                                                                                                       //
        var proxy = void 0;                                                                                            // 938
                                                                                                                       //
        if (tokenInfo.userInfo) {                                                                                      // 939
          throw new Error("API tokens created with arbitrary userInfo no longer supported");                           // 940
        } else {                                                                                                       // 941
          var identityId = null;                                                                                       // 942
                                                                                                                       //
          if (tokenInfo.identityId && !tokenInfo.forSharing) {                                                         // 943
            identityId = tokenInfo.identityId;                                                                         // 944
          }                                                                                                            // 945
                                                                                                                       //
          proxy = new Proxy(grain, null, null, tabId, identityId, true);                                               // 947
          proxy.apiToken = tokenInfo;                                                                                  // 948
          proxy.apiSessionParams = serializedParams;                                                                   // 949
        }                                                                                                              // 950
                                                                                                                       //
        if (!SandstormPermissions.mayOpenGrain(globalDb, {                                                             // 952
          token: tokenInfo                                                                                             // 952
        })) {                                                                                                          // 952
          // Note that only public grains may be opened without a user ID.                                             // 953
          throw new Meteor.Error(403, "Unauthorized.");                                                                // 954
        }                                                                                                              // 955
                                                                                                                       //
        if (tokenInfo.expires) {                                                                                       // 957
          proxy.expires = tokenInfo.expires;                                                                           // 958
        }                                                                                                              // 959
                                                                                                                       //
        apiSessionProxies.put(hashedToken, hashedParams, proxy);                                                       // 961
        return proxy;                                                                                                  // 963
      });                                                                                                              // 964
    }                                                                                                                  // 965
  });                                                                                                                  // 966
};                                                                                                                     // 967
                                                                                                                       //
var apiUseBasicAuth = function (req, hostId) {                                                                         // 969
  // If the request was to a token-specific host *and* the request has no Origin header (meaning                       // 970
  // it is not an XHR from a web site), then we permit the use of HTTP basic auth. The reason we                       // 971
  // prohibit XHRs is because we want to discourage people from using basic auth in a browser,                         // 972
  // because the browser will cache the credentials, which exposes the hostname to XSRF attack                         // 973
  // (which is mostly, but not completely, mitigated by the unguessable hostname). The reason we                       // 974
  // want to allow basic auth at all is because a lot of existing client apps support only basic                       // 975
  // auth. New clients and web-based clients should use `Authorization: bearer <token>` instead.                       // 976
  if (globalDb.isTokenSpecificHostId(hostId) && !req.headers.origin) {                                                 // 977
    return true;                                                                                                       // 978
  } // Historically, all API tokens were served from the same host rather than have token-specific                     // 979
  // hosts. In this model, basic auth in browsers was far more dangerous since the hostname was                        // 982
  // not a secret -- in fact, even benign attempts to use two APIs on the same server from the same                    // 983
  // browser could interfere if using basic auth. Hence, basic auth was prohibited except when                         // 984
  // coming from certain whitelisted clients known not to be browsers, e.g. Mirall, the ownCloud                       // 985
  // client app.                                                                                                       // 986
  //                                                                                                                   // 987
  // Since many clients in the wild have already been configured to use the shared API host, we                        // 988
  // must continue to support them, so this logic remains.                                                             // 989
                                                                                                                       //
                                                                                                                       //
  var agent = req.headers["user-agent"];                                                                               // 990
  return agent && agent.match(BASIC_AUTH_USER_AGENTS_REGEX);                                                           // 991
};                                                                                                                     // 992
                                                                                                                       //
var apiTokenForRequest = function (req, hostId) {                                                                      // 994
  // Extract the API token from the request.                                                                           // 995
  var auth = req.headers.authorization;                                                                                // 997
  var token = void 0;                                                                                                  // 998
                                                                                                                       //
  if (auth && auth.slice(0, 7).toLowerCase() === "bearer ") {                                                          // 999
    token = auth.slice(7).trim();                                                                                      // 1000
  } else if (auth && auth.slice(0, 6).toLowerCase() === "basic " && apiUseBasicAuth(req, hostId)) {                    // 1001
    token = new Buffer(auth.slice(6).trim(), "base64").toString().split(":")[1];                                       // 1003
  } else if (hostId !== "api" && req.url.startsWith("/.sandstorm-token/")) {                                           // 1004
    // Disallow clients that use the old "api" hostId from using tokens in the path.                                   // 1005
    // Allowing them would be a security problem in browsers that ignore CSP.                                          // 1006
    var parts = req.url.slice(1).split("/"); // remove leading / and split                                             // 1007
                                                                                                                       //
    if (parts.length < 2) {                                                                                            // 1008
      token = undefined;                                                                                               // 1009
    } else {                                                                                                           // 1010
      token = parts[1];                                                                                                // 1011
      req.url = "/" + parts.slice(2).join("/"); // remove .sandstorm-token/$TOKEN from path                            // 1012
    }                                                                                                                  // 1014
  } else {                                                                                                             // 1015
    token = undefined;                                                                                                 // 1016
  }                                                                                                                    // 1017
                                                                                                                       //
  if (token && hostId !== "api") {                                                                                     // 1019
    // Verify that the token matches the specific host.                                                                // 1020
    if (hostId !== globalDb.apiHostIdForToken(token)) {                                                                // 1021
      token = undefined;                                                                                               // 1022
    }                                                                                                                  // 1023
  }                                                                                                                    // 1024
                                                                                                                       //
  return token;                                                                                                        // 1026
}; // =======================================================================================                          // 1027
// Routing to proxies.                                                                                                 // 1030
//                                                                                                                     // 1031
// pre-meteor.js calls this                                                                                            // 1033
                                                                                                                       //
                                                                                                                       //
tryProxyUpgrade = function (hostId, req, socket, head) {                                                               // 1034
  // Attempt to handle a WebSocket upgrade by dispatching it to a grain. Returns a promise that                        // 1035
  // resolves true if an appropriate grain is found, false if there was no match (but the caller                       // 1036
  // should consider other host types, like static web publishing), or throws an error if the                          // 1037
  // request is definitely invalid.                                                                                    // 1038
  if (globalDb.isApiHostId(hostId)) {                                                                                  // 1040
    var token = apiTokenForRequest(req, hostId);                                                                       // 1041
                                                                                                                       //
    if (token) {                                                                                                       // 1042
      return getProxyForApiToken(token, req).then(function (proxy) {                                                   // 1043
        // Meteor sets the timeout to five seconds. Change that back to two                                            // 1044
        // minutes, which is the default value.                                                                        // 1045
        socket.setTimeout(120000);                                                                                     // 1046
        proxy.upgradeHandler(req, socket, head);                                                                       // 1048
        return true;                                                                                                   // 1049
      });                                                                                                              // 1050
    } else {                                                                                                           // 1051
      return Promise.resolve(false);                                                                                   // 1052
    }                                                                                                                  // 1053
  } else {                                                                                                             // 1054
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;                   // 1055
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {                                          // 1056
      if (proxy) {                                                                                                     // 1057
        // Cross-origin requests are not allowed on UI session hosts.                                                  // 1058
        var origin = req.headers.origin;                                                                               // 1059
                                                                                                                       //
        if (origin !== PROTOCOL + "//" + req.headers.host) {                                                           // 1060
          throw new Meteor.Error(403, "Detected illegal cross-origin WebSocket from: " + origin);                      // 1061
        } // Meteor sets the timeout to five seconds. Change that back to two                                          // 1062
        // minutes, which is the default value.                                                                        // 1065
                                                                                                                       //
                                                                                                                       //
        socket.setTimeout(120000);                                                                                     // 1066
        proxy.upgradeHandler(req, socket, head);                                                                       // 1068
        return true;                                                                                                   // 1069
      } else {                                                                                                         // 1070
        return false;                                                                                                  // 1071
      }                                                                                                                // 1072
    });                                                                                                                // 1073
  }                                                                                                                    // 1074
}; // pre-meteor.js calls this                                                                                         // 1075
                                                                                                                       //
                                                                                                                       //
tryProxyRequest = function (hostId, req, res) {                                                                        // 1078
  // Attempt to handle an HTTP request by dispatching it to a grain. Returns a promise that                            // 1079
  // resolves true if an appropriate grain is found, false if there was no match (but the caller                       // 1080
  // should consider other host types, like static web publishing), or throws an error if the                          // 1081
  // request is definitely invalid.                                                                                    // 1082
  var hostIdHash = globalDb.isApiHostId(hostId);                                                                       // 1084
                                                                                                                       //
  if (hostIdHash) {                                                                                                    // 1085
    // This is a request for the API host.                                                                             // 1086
    res.setHeader("Access-Control-Allow-Origin", "*");                                                                 // 1088
                                                                                                                       //
    if (req.method === "OPTIONS") {                                                                                    // 1090
      // Reply to CORS preflight request.                                                                              // 1091
      // All we want to do is permit APIs to be accessed from arbitrary origins. Since clients                         // 1093
      // must send a valid Authorization header, and since cookies are not used for                                    // 1094
      // authorization, this is perfectly safe. In a sane world, we would only need to send back                       // 1095
      // 'Access-Control-Allow-Origin: *' and be done with it.                                                         // 1096
      //                                                                                                               // 1097
      // However, CORS demands that we explicitly whitelist individual methods and headers for                         // 1098
      // use cross-origin, as if this is somehow useful for implementing any practical security                        // 1099
      // policy (it isn't). To make matters worse, we are REQUIRED to enumerate each one                               // 1100
      // individually. We cannot just write '*' for these lists. WTF, CORS?                                            // 1101
      //                                                                                                               // 1102
      // Luckily, the request tells us exactly what method and headers are being requested, so we                      // 1103
      // only need to copy those over, rather than create an exhaustive list. But this is still                        // 1104
      // overly complicated.                                                                                           // 1105
      var accessControlHeaders = {                                                                                     // 1107
        "Access-Control-Allow-Methods": "GET, HEAD, POST, PUT, PATCH, DELETE",                                         // 1108
        "Access-Control-Max-Age": "3600"                                                                               // 1109
      }; // Copy all requested headers to the allowed headers list.                                                    // 1107
                                                                                                                       //
      var requestedHeaders = req.headers["access-control-request-headers"];                                            // 1113
                                                                                                                       //
      if (requestedHeaders) {                                                                                          // 1114
        accessControlHeaders["Access-Control-Allow-Headers"] = requestedHeaders;                                       // 1115
      } // Add the requested method to the allowed methods list, if it's not there already.                            // 1116
                                                                                                                       //
                                                                                                                       //
      var requestedMethod = req.headers["access-control-request-method"];                                              // 1119
                                                                                                                       //
      if (requestedMethod && !_.contains(["GET", "HEAD", "POST", "PUT", "PATCH", "DELETE"], requestedMethod)) {        // 1120
        accessControlHeaders["Access-Control-Allow-Methods"] += ", " + requestedMethod;                                // 1122
      }                                                                                                                // 1123
                                                                                                                       //
      for (header in meteorBabelHelpers.sanitizeForInObject(accessControlHeaders)) {                                   // 1125
        res.setHeader(header, accessControlHeaders[header]);                                                           // 1126
      }                                                                                                                // 1127
    }                                                                                                                  // 1128
                                                                                                                       //
    var errorHandler = function (err) {                                                                                // 1130
      if (err instanceof Meteor.Error) {                                                                               // 1131
        console.log("error: " + err);                                                                                  // 1132
        res.writeHead(err.error, err.reason, {                                                                         // 1133
          "Content-Type": "text/plain"                                                                                 // 1133
        });                                                                                                            // 1133
      } else {                                                                                                         // 1134
        res.writeHead(500, "Internal Server Error", {                                                                  // 1135
          "Content-Type": "text/plain"                                                                                 // 1135
        });                                                                                                            // 1135
      }                                                                                                                // 1136
                                                                                                                       //
      res.end(err.stack);                                                                                              // 1138
    };                                                                                                                 // 1139
                                                                                                                       //
    var token = apiTokenForRequest(req, hostId);                                                                       // 1141
                                                                                                                       //
    if (token && req.headers["x-sandstorm-token-keepalive"]) {                                                         // 1142
      inMeteor(function () {                                                                                           // 1143
        var keepaliveDuration = parseInt(req.headers["x-sandstorm-token-keepalive"]);                                  // 1144
        check(keepaliveDuration, Match.Integer);                                                                       // 1145
        var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");                                  // 1146
        validateWebkey(ApiTokens.findOne(hashedToken), new Date(Date.now() + keepaliveDuration));                      // 1147
      }).then(function () {                                                                                            // 1148
        res.writeHead(200, {});                                                                                        // 1149
        res.end();                                                                                                     // 1150
      }, errorHandler);                                                                                                // 1151
    } else if (token) {                                                                                                // 1152
      getProxyForApiToken(token, req).then(function (proxy) {                                                          // 1153
        proxy.requestHandler(req, res);                                                                                // 1154
      }, errorHandler);                                                                                                // 1155
    } else {                                                                                                           // 1156
      // No token. Look up static API host info.                                                                       // 1157
      inMeteor(function () {                                                                                           // 1158
        var apiHost = globalDb.collections.apiHosts.findOne(hostIdHash);                                               // 1159
                                                                                                                       //
        if (req.method === "OPTIONS") {                                                                                // 1161
          // OPTIONS request with no authorization token.                                                              // 1162
          var dav = ((apiHost || {}).options || {}).dav || [];                                                         // 1164
                                                                                                                       //
          if (dav.length > 0) {                                                                                        // 1165
            res.setHeader("DAV", dav.join(", "));                                                                      // 1166
            res.setHeader("Access-Control-Expose-Headers", "DAV");                                                     // 1167
          }                                                                                                            // 1168
                                                                                                                       //
          res.writeHead(200, {});                                                                                      // 1170
          res.end();                                                                                                   // 1171
        } else {                                                                                                       // 1172
          var resources = (apiHost || {}).resources || {};                                                             // 1173
          var path = SandstormDb.escapeMongoKey(req.url.split("?")[0]);                                                // 1174
                                                                                                                       //
          if (path in resources) {                                                                                     // 1175
            // Serve a static resource.                                                                                // 1176
            var resource = resources[path];                                                                            // 1177
            console.log(path, resources, resource);                                                                    // 1178
            if (resource.language) res.setHeader("Content-Language", resource.language);                               // 1179
            if (resource.encoding) res.setHeader("Content-Encoding", resource.encoding);                               // 1180
            res.writeHead(200, {                                                                                       // 1181
              "Content-Type": resource.type                                                                            // 1182
            });                                                                                                        // 1181
            res.end(resource.body);                                                                                    // 1184
          } else {                                                                                                     // 1185
            if (apiUseBasicAuth(req, hostId)) {                                                                        // 1186
              res.writeHead(401, {                                                                                     // 1187
                "Content-Type": "text/plain",                                                                          // 1188
                "WWW-Authenticate": "Basic realm='Sandstorm API'"                                                      // 1189
              });                                                                                                      // 1187
            } else {                                                                                                   // 1191
              // TODO(someday): Display some sort of nifty API browser.                                                // 1192
              res.writeHead(403, {                                                                                     // 1193
                "Content-Type": "text/plain"                                                                           // 1193
              });                                                                                                      // 1193
            }                                                                                                          // 1194
                                                                                                                       //
            res.end("Missing or invalid authorization header.\n\n" + "This address serves APIs, which allow external apps (such as a phone app) to\n" + "access data on your Sandstorm server. This address is not meant to be opened\n" + "in a regular browser.");
          }                                                                                                            // 1200
        }                                                                                                              // 1201
      }).then(function () {}, errorHandler);                                                                           // 1202
    }                                                                                                                  // 1203
                                                                                                                       //
    return Promise.resolve(true);                                                                                      // 1205
  } else {                                                                                                             // 1206
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf("sandstorm-sid=") !== -1;                   // 1207
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {                                          // 1210
      if (proxy) {                                                                                                     // 1211
        // Let's do some CSRF defense.                                                                                 // 1212
        var expectedOrigin = PROTOCOL + "//" + req.headers.host;                                                       // 1213
        var mainUrl = proxy.parentOrigin;                                                                              // 1214
        var origin = req.headers.origin;                                                                               // 1215
        var referer = req.headers.referer;                                                                             // 1216
                                                                                                                       //
        if (origin) {                                                                                                  // 1217
          // If an origin header was provided, then it must be accurate. Note that Chrome and                          // 1218
          // Safari always send an Origin header on non-GET requests (even same-origin), and                           // 1219
          // therefore on those browsers this rule will block all CSRF attacks. Unfortunately,                         // 1220
          // a patch to add this behavior to Firefox has stalled:                                                      // 1221
          //     https://bugzilla.mozilla.org/show_bug.cgi?id=446344                                                   // 1222
          // I haven't found any information on IE/Edge's behavior.                                                    // 1223
          if (origin !== expectedOrigin) {                                                                             // 1224
            // TODO(security): Alas, it turns out we have apps that have:                                              // 1225
            //   <meta name="referrer" content="no-referrer">                                                          // 1226
            // as Chrome sends "Origin: null" in these cases. :( These apps need to switch to:                         // 1227
            //   <meta name="referrer" content="same-origin">                                                          // 1228
            // It's important that we don't break apps, so we will accept null origins for now,                        // 1229
            // which of course completely defeats any CSRF protection. We should get the apps to                       // 1230
            // update or apply a whitelist soon.                                                                       // 1231
            if (origin === "null") {                                                                                   // 1232
              if (!proxy.wroteCsrfWarning) {                                                                           // 1233
                console.warn("Note: Observed null Origin header. App needs to be updated so that " + "we can apply automatic CSRF protection. grain ID:", proxy.grainId);
                proxy.wroteCsrfWarning = true;                                                                         // 1237
              }                                                                                                        // 1238
            } else {                                                                                                   // 1239
              throw new Meteor.Error(403, "Blocked illegal cross-origin request from: " + origin);                     // 1240
            }                                                                                                          // 1241
          }                                                                                                            // 1242
        } else if (referer) {                                                                                          // 1243
          // Mark that we've seed a Referer header on this host, which indicates that the user's                       // 1244
          // browser is not suppressing them.                                                                          // 1245
          proxy.seenReferer = true; // Deny requests coming from an external referer, since there is no legitimate use case
          // for these. Note of course that an attacker can trivially suppress the referer header                      // 1249
          // in practice, so this check alone does not give us any security benefit if we accept                       // 1250
          // requests that lack a referer.                                                                             // 1251
                                                                                                                       //
          if (referer != expectedOrigin && !referer.startsWith(expectedOrigin + "/") && referer != mainUrl && !referer.startsWith(mainUrl + "/")) {
            throw new Meteor.Error(403, "Blocked illegal cross-origin referral from: " + referer);                     // 1254
          }                                                                                                            // 1255
        } else {                                                                                                       // 1256
          // We saw neither an Origin nor a Referer header.                                                            // 1257
          if (req.method !== "GET" && req.method !== "HEAD") {                                                         // 1258
            // This is possibly a side-effecting request, meaning we should worry about CSRF.                          // 1259
            //                                                                                                         // 1260
            // Note that we'll never get to this branch for Chrome or Safari since they would have                     // 1261
            // sent an Origin header.                                                                                  // 1262
            if (!proxy.wroteCsrfWarning) {                                                                             // 1264
              console.warn("Note: Observed session for which auto-CSRF-protection couldn't be applied.", "seenReferer:", !!proxy.seenReferer, "method:", req.method, "path:", req.url, "UA:", req.headers["user-agent"]);
              proxy.wroteCsrfWarning = true;                                                                           // 1269
            }                                                                                                          // 1270
          }                                                                                                            // 1271
        }                                                                                                              // 1272
                                                                                                                       //
        proxy.requestHandler(req, res);                                                                                // 1274
        return true;                                                                                                   // 1275
      } else {                                                                                                         // 1276
        return false;                                                                                                  // 1277
      }                                                                                                                // 1278
    });                                                                                                                // 1279
  }                                                                                                                    // 1280
}; // =======================================================================================                          // 1281
// Proxy class                                                                                                         // 1284
//                                                                                                                     // 1285
// Connects to a grain and exports it on a wildcard host.                                                              // 1286
//                                                                                                                     // 1287
                                                                                                                       //
                                                                                                                       //
var Proxy = function () {                                                                                              //
  function Proxy(grain, sessionId, hostId, tabId, identityId, isApi, parentOrigin, supervisor) {                       // 1290
    var _this4 = this;                                                                                                 // 1290
                                                                                                                       //
    (0, _classCallCheck3.default)(this, Proxy);                                                                        // 1290
    // `grain` is an entry in the `Grains` collection.                                                                 // 1291
    this.grainId = grain._id;                                                                                          // 1292
    this.ownerId = grain.userId;                                                                                       // 1293
    this.identityId = identityId;                                                                                      // 1294
    this.supervisor = supervisor; // note: optional parameter; we can reconnect                                        // 1295
                                                                                                                       //
    this.sessionId = sessionId;                                                                                        // 1296
    this.tabId = tabId;                                                                                                // 1297
    this.isApi = isApi;                                                                                                // 1298
    this.hasLoaded = false;                                                                                            // 1299
    this.websockets = {};                                                                                              // 1300
    this.websocketCounter = 0; // Used for generating unique socket IDs.                                               // 1301
                                                                                                                       //
    this.streamingRequests = {};                                                                                       // 1302
    this.streamingRequestCounter = 0; // Used for generating unique request IDs.                                       // 1303
                                                                                                                       //
    this.streamingResponses = {};                                                                                      // 1304
    this.streamingResponseCounter = 0; // Used for generating unique response IDs.                                     // 1305
                                                                                                                       //
    this.parentOrigin = parentOrigin || process.env.ROOT_URL;                                                          // 1306
                                                                                                                       //
    if (sessionId) {                                                                                                   // 1308
      if (!hostId) throw new Error("sessionId must come with hostId");                                                 // 1309
      if (isApi) throw new Error("API proxy shouldn't have sessionId");                                                // 1310
      this.hostId = hostId;                                                                                            // 1311
    } else {                                                                                                           // 1312
      if (!isApi) throw new Error("non-API proxy requires sessionId");                                                 // 1313
      if (hostId) throw new Error("API proxy sholudn't have hostId");                                                  // 1314
    }                                                                                                                  // 1315
                                                                                                                       //
    if (this.identityId) {                                                                                             // 1317
      var identity = globalDb.getIdentity(this.identityId);                                                            // 1318
                                                                                                                       //
      if (!identity) {                                                                                                 // 1319
        throw new Error("identity not found: " + this.identityId);                                                     // 1320
      } // The identity cap becomes invalid if the user no longer has access to the grain.                             // 1321
                                                                                                                       //
                                                                                                                       //
      var idCapRequirement = {                                                                                         // 1324
        permissionsHeld: {                                                                                             // 1325
          identityId: identity._id,                                                                                    // 1325
          grainId: this.grainId                                                                                        // 1325
        }                                                                                                              // 1325
      };                                                                                                               // 1324
      this.userInfo = {                                                                                                // 1328
        displayName: {                                                                                                 // 1329
          defaultText: identity.profile.name                                                                           // 1329
        },                                                                                                             // 1329
        preferredHandle: identity.profile.handle,                                                                      // 1330
        identityId: new Buffer(identity._id, "hex"),                                                                   // 1331
        identity: makeIdentity(identity._id, [idCapRequirement])                                                       // 1332
      };                                                                                                               // 1328
                                                                                                                       //
      if (identity.profile.pictureUrl) {                                                                               // 1334
        this.userInfo.pictureUrl = identity.profile.pictureUrl;                                                        // 1335
      } else {                                                                                                         // 1336
        this.userInfo.pictureUrl = PROTOCOL + "//" + makeWildcardHost("static") + "/identicon/" + this.identityId.slice(0, 32) + "?s=128";
      }                                                                                                                // 1340
                                                                                                                       //
      if (identity.profile.pronoun) this.userInfo.pronouns = identity.profile.pronoun;                                 // 1342
    } else {                                                                                                           // 1343
      this.userInfo = {                                                                                                // 1344
        displayName: {                                                                                                 // 1345
          defaultText: "Anonymous User"                                                                                // 1345
        },                                                                                                             // 1345
        preferredHandle: "anonymous"                                                                                   // 1346
      };                                                                                                               // 1344
    }                                                                                                                  // 1348
                                                                                                                       //
    this.requestHandler = function (request, response) {                                                               // 1350
      if (_this4.sessionId) {                                                                                          // 1351
        // Implement /_sandstorm-init for setting the session cookie.                                                  // 1352
        var url = Url.parse(request.url, true);                                                                        // 1353
                                                                                                                       //
        if (url.pathname === "/_sandstorm-init" && url.query.sessionid === _this4.sessionId) {                         // 1354
          _this4.doSessionInit(request, response, url.query.path);                                                     // 1355
                                                                                                                       //
          return;                                                                                                      // 1356
        }                                                                                                              // 1357
      }                                                                                                                // 1358
                                                                                                                       //
      Promise.resolve(undefined).then(function () {                                                                    // 1360
        var contentLength = request.headers["content-length"];                                                         // 1361
                                                                                                                       //
        if ((request.method === "POST" || request.method === "PUT") && (contentLength === undefined || contentLength > 1024 * 1024)) {
          // The input is either very long, or we don't know how long it is, so use streaming mode.                    // 1364
          return _this4.handleRequestStreaming(request, response, contentLength, 0);                                   // 1365
        } else {                                                                                                       // 1366
          return readAll(request).then(function (data) {                                                               // 1367
            return _this4.handleRequest(request, data, response, 0);                                                   // 1368
          });                                                                                                          // 1369
        }                                                                                                              // 1370
      }).catch(function (err) {                                                                                        // 1371
        _this4.setHasLoaded();                                                                                         // 1372
                                                                                                                       //
        var body = err.stack;                                                                                          // 1374
                                                                                                                       //
        if (err.cppFile) {                                                                                             // 1375
          body += "\nC++ location:" + err.cppFile + ":" + (err.line || "??");                                          // 1376
        }                                                                                                              // 1377
                                                                                                                       //
        if (err.kjType) {                                                                                              // 1379
          body += "\ntype: " + err.kjType;                                                                             // 1380
        }                                                                                                              // 1381
                                                                                                                       //
        if (response.headersSent) {                                                                                    // 1383
          // Unfortunately, it's too late to tell the client what happened.                                            // 1384
          console.error("HTTP request failed after response already sent:", body);                                     // 1385
          response.end();                                                                                              // 1386
        } else {                                                                                                       // 1387
          if (err instanceof Meteor.Error) {                                                                           // 1388
            response.writeHead(err.error, err.reason, {                                                                // 1389
              "Content-Type": "text/plain"                                                                             // 1389
            });                                                                                                        // 1389
          } else {                                                                                                     // 1390
            response.writeHead(500, "Internal Server Error", {                                                         // 1391
              "Content-Type": "text/plain"                                                                             // 1391
            });                                                                                                        // 1391
          }                                                                                                            // 1392
                                                                                                                       //
          response.end(body);                                                                                          // 1394
        }                                                                                                              // 1395
      });                                                                                                              // 1396
    };                                                                                                                 // 1397
                                                                                                                       //
    this.upgradeHandler = function (request, socket, head) {                                                           // 1399
      _this4.handleWebSocket(request, socket, head, 0).catch(function (err) {                                          // 1400
        console.error("WebSocket setup failed:", err.stack); // TODO(cleanup):  Manually send back a 500 response?     // 1401
                                                                                                                       //
        socket.destroy();                                                                                              // 1403
      });                                                                                                              // 1404
    };                                                                                                                 // 1405
  }                                                                                                                    // 1406
                                                                                                                       //
  Proxy.prototype.close = function () {                                                                                //
    function close() {                                                                                                 //
      for (var streamingResponseIdx in meteorBabelHelpers.sanitizeForInObject(this.streamingResponses)) {              // 1409
        this.streamingResponses[streamingResponseIdx].abort();                                                         // 1410
      }                                                                                                                // 1411
                                                                                                                       //
      this.streamingResponses = {};                                                                                    // 1413
                                                                                                                       //
      for (var streamingRequestIdx in meteorBabelHelpers.sanitizeForInObject(this.streamingRequests)) {                // 1415
        this.streamingRequests[streamingRequestIdx].destroy();                                                         // 1416
      }                                                                                                                // 1417
                                                                                                                       //
      this.streamingRequests = {};                                                                                     // 1419
                                                                                                                       //
      for (var socketIdx in meteorBabelHelpers.sanitizeForInObject(this.websockets)) {                                 // 1421
        this.websockets[socketIdx].destroy();                                                                          // 1422
      }                                                                                                                // 1423
                                                                                                                       //
      ;                                                                                                                // 1423
      this.websockets = {};                                                                                            // 1425
                                                                                                                       //
      if (this.session) {                                                                                              // 1427
        this.session.close();                                                                                          // 1428
        delete this.session;                                                                                           // 1429
      }                                                                                                                // 1430
                                                                                                                       //
      if (this.uiView) {                                                                                               // 1432
        this.uiView.close();                                                                                           // 1433
        delete this.uiView;                                                                                            // 1434
      }                                                                                                                // 1435
                                                                                                                       //
      if (this.supervisor) {                                                                                           // 1437
        this.supervisor.close();                                                                                       // 1438
        delete this.supervisor;                                                                                        // 1439
      }                                                                                                                // 1440
                                                                                                                       //
      if (this.permissionsObserver) {                                                                                  // 1442
        this.permissionsObserver.stop();                                                                               // 1443
        delete this.permissionsObserver;                                                                               // 1444
      }                                                                                                                // 1445
    }                                                                                                                  // 1446
                                                                                                                       //
    return close;                                                                                                      //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.getConnection = function () {                                                                        //
    function getConnection() {                                                                                         //
      if (!this.supervisor) {                                                                                          // 1449
        this.supervisor = globalBackend.cap().getGrain(this.ownerId, this.grainId).supervisor;                         // 1450
        this.uiView = null;                                                                                            // 1451
      }                                                                                                                // 1452
                                                                                                                       //
      if (!this.uiView) {                                                                                              // 1454
        this.uiView = this.supervisor.getMainView().view;                                                              // 1455
      }                                                                                                                // 1456
    }                                                                                                                  // 1457
                                                                                                                       //
    return getConnection;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype._callNewWebSession = function () {                                                                   //
    function _callNewWebSession(request, userInfo) {                                                                   //
      var params = Capnp.serialize(WebSession.Params, {                                                                // 1460
        basePath: PROTOCOL + "//" + request.headers.host,                                                              // 1461
        userAgent: "user-agent" in request.headers ? request.headers["user-agent"] : "UnknownAgent/0.0",               // 1462
        acceptableLanguages: "accept-language" in request.headers ? request.headers["accept-language"].split(",").map(function (s) {
          return s.trim();                                                                                             // 1466
        }) : ["en-US", "en"]                                                                                           // 1466
      });                                                                                                              // 1460
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId, this.tabId), WebSession.typeId, params, new Buffer(this.tabId, "hex")).session;
    }                                                                                                                  // 1472
                                                                                                                       //
    return _callNewWebSession;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype._callNewApiSession = function () {                                                                   //
    function _callNewApiSession(request, userInfo) {                                                                   //
      var _this5 = this;                                                                                               // 1474
                                                                                                                       //
      var serializedParams = this.apiSessionParams;                                                                    // 1475
                                                                                                                       //
      if (!serializedParams) {                                                                                         // 1476
        throw new Meteor.Error(500, "Should have already computed apiSessionParams.");                                 // 1477
      } // TODO(someday): We are currently falling back to WebSession if we get any kind of error upon                 // 1478
      // calling newSession with an ApiSession._id.                                                                    // 1481
      // Eventually we'll remove this logic once we're sure apps have updated.                                         // 1482
                                                                                                                       //
                                                                                                                       //
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId, this.tabId), ApiSession.typeId, serializedParams, new Buffer(this.tabId, "hex")).then(function (session) {
        return session.session;                                                                                        // 1487
      }, function (err) {                                                                                              // 1488
        return _this5._callNewWebSession(request, userInfo);                                                           // 1489
      });                                                                                                              // 1490
    }                                                                                                                  // 1491
                                                                                                                       //
    return _callNewApiSession;                                                                                         //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype._callNewRequestSession = function () {                                                               //
    function _callNewRequestSession(request, userInfo) {                                                               //
      var params = Capnp.serialize(WebSession.Params, {                                                                // 1494
        basePath: PROTOCOL + "//" + request.headers.host,                                                              // 1495
        userAgent: "user-agent" in request.headers ? request.headers["user-agent"] : "UnknownAgent/0.0",               // 1496
        acceptableLanguages: "accept-language" in request.headers ? request.headers["accept-language"].split(",").map(function (s) {
          return s.trim();                                                                                             // 1500
        }) : ["en-US", "en"]                                                                                           // 1500
      });                                                                                                              // 1494
      return this.uiView.newRequestSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId, this.tabId), WebSession.typeId, params, this.powerboxRequest.descriptors, new Buffer(this.tabId, "hex")).session;
    }                                                                                                                  // 1507
                                                                                                                       //
    return _callNewRequestSession;                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype._callNewSession = function () {                                                                      //
    function _callNewSession(request, viewInfo) {                                                                      //
      var _this6 = this;                                                                                               // 1509
                                                                                                                       //
      var userInfo = _.clone(this.userInfo);                                                                           // 1510
                                                                                                                       //
      var promise = inMeteor(function () {                                                                             // 1511
        var vertex = void 0;                                                                                           // 1512
                                                                                                                       //
        if (_this6.apiToken) {                                                                                         // 1513
          vertex = {                                                                                                   // 1514
            token: _this6.apiToken                                                                                     // 1514
          };                                                                                                           // 1514
        } else {                                                                                                       // 1515
          // (this.identityId might be null; this is fine)                                                             // 1516
          vertex = {                                                                                                   // 1517
            grain: {                                                                                                   // 1517
              _id: _this6.grainId,                                                                                     // 1517
              identityId: _this6.identityId                                                                            // 1517
            }                                                                                                          // 1517
          };                                                                                                           // 1517
        }                                                                                                              // 1518
                                                                                                                       //
        var onInvalidated = function () {                                                                              // 1520
          Sessions.remove({                                                                                            // 1521
            _id: _this6.sessionId                                                                                      // 1521
          });                                                                                                          // 1521
        };                                                                                                             // 1522
                                                                                                                       //
        if (!_this6.sessionId) {                                                                                       // 1524
          onInvalidated = function () {                                                                                // 1525
            apiSessionProxies.removeProxiesOfToken(_this6.apiToken._id);                                               // 1526
          };                                                                                                           // 1527
        }                                                                                                              // 1528
                                                                                                                       //
        var permissions = SandstormPermissions.grainPermissions(globalDb, vertex, viewInfo, onInvalidated);            // 1530
        if (_this6.permissionsObserver) _this6.permissionsObserver.stop();                                             // 1533
        _this6.permissionsObserver = permissions.observeHandle;                                                        // 1534
                                                                                                                       //
        if (!permissions.permissions) {                                                                                // 1536
          throw new Meteor.Error(403, "Unauthorized", "User is not authorized to open this grain.");                   // 1537
        }                                                                                                              // 1538
                                                                                                                       //
        Sessions.update({                                                                                              // 1540
          _id: _this6.sessionId                                                                                        // 1541
        }, {                                                                                                           // 1540
          $set: {                                                                                                      // 1543
            viewInfo: viewInfo,                                                                                        // 1544
            permissions: permissions.permissions                                                                       // 1545
          }                                                                                                            // 1543
        });                                                                                                            // 1542
        return permissions.permissions;                                                                                // 1549
      });                                                                                                              // 1550
      return promise.then(function (permissions) {                                                                     // 1552
        userInfo.permissions = permissions;                                                                            // 1553
        var numBytes = Math.ceil(permissions.length / 8);                                                              // 1555
        var buf = new Buffer(numBytes);                                                                                // 1556
                                                                                                                       //
        for (var ii = 0; ii < numBytes; ++ii) {                                                                        // 1557
          buf[ii] = 0;                                                                                                 // 1558
        }                                                                                                              // 1559
                                                                                                                       //
        for (var _ii = 0; _ii < permissions.length; ++_ii) {                                                           // 1561
          var byteNum = Math.floor(_ii / 8);                                                                           // 1562
          var bitNum = _ii % 8;                                                                                        // 1563
                                                                                                                       //
          if (permissions[_ii]) {                                                                                      // 1564
            buf[byteNum] = buf[byteNum] | 1 << bitNum;                                                                 // 1565
          }                                                                                                            // 1566
        }                                                                                                              // 1567
                                                                                                                       //
        userInfo.deprecatedPermissionsBlob = buf;                                                                      // 1569
                                                                                                                       //
        if (_this6.isApi) {                                                                                            // 1571
          return _this6._callNewApiSession(request, userInfo);                                                         // 1572
        } else if (_this6.powerboxRequest) {                                                                           // 1573
          return _this6._callNewRequestSession(request, userInfo);                                                     // 1574
        } else {                                                                                                       // 1575
          return _this6._callNewWebSession(request, userInfo);                                                         // 1576
        }                                                                                                              // 1577
      });                                                                                                              // 1578
    }                                                                                                                  // 1579
                                                                                                                       //
    return _callNewSession;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.getSession = function () {                                                                           //
    function getSession(request) {                                                                                     //
      var _this7 = this;                                                                                               // 1581
                                                                                                                       //
      if (!this.session) {                                                                                             // 1582
        this.getConnection(); // make sure we're connected                                                             // 1583
                                                                                                                       //
        var promise = this.uiView.getViewInfo().then(function (viewInfo) {                                             // 1584
          return inMeteor(function () {                                                                                // 1585
            // For now, we don't allow grains to set `appTitle` or `grainIcon`.                                        // 1586
            var cachedViewInfo = _.omit(viewInfo, "appTitle", "grainIcon");                                            // 1587
                                                                                                                       //
            Grains.update(_this7.grainId, {                                                                            // 1588
              $set: {                                                                                                  // 1588
                cachedViewInfo: cachedViewInfo                                                                         // 1588
              }                                                                                                        // 1588
            });                                                                                                        // 1588
          }).then(function () {                                                                                        // 1589
            return _this7._callNewSession(request, viewInfo);                                                          // 1590
          });                                                                                                          // 1591
        }, function (error) {                                                                                          // 1592
          if (error.kjType === "failed" || error.kjType === "unimplemented") {                                         // 1593
            // Method not implemented.                                                                                 // 1594
            // TODO(apibump): Don't treat 'failed' as 'unimplemented'. Unfortunately, old apps built                   // 1595
            //   with old versions of Cap'n Proto don't throw 'unimplemented' exceptions, so we have                   // 1596
            //   to accept 'failed' here at least until the next API bump.                                             // 1597
            return _this7._callNewSession(request, {});                                                                // 1598
          } else {                                                                                                     // 1599
            return Promise.reject(error);                                                                              // 1600
          }                                                                                                            // 1601
        });                                                                                                            // 1602
        this.session = new Capnp.Capability(promise, WebSession);                                                      // 1603
      }                                                                                                                // 1604
                                                                                                                       //
      return this.session;                                                                                             // 1606
    }                                                                                                                  // 1607
                                                                                                                       //
    return getSession;                                                                                                 //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.keepAlive = function () {                                                                            //
    function keepAlive() {                                                                                             //
      this.getConnection();                                                                                            // 1610
      return this.supervisor.keepAlive();                                                                              // 1611
    }                                                                                                                  // 1612
                                                                                                                       //
    return keepAlive;                                                                                                  //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.resetConnection = function () {                                                                      //
    function resetConnection() {                                                                                       //
      if (this.session) {                                                                                              // 1615
        this.session.close();                                                                                          // 1616
        delete this.session;                                                                                           // 1617
      }                                                                                                                // 1618
                                                                                                                       //
      if (this.uiView) {                                                                                               // 1620
        this.uiView.close();                                                                                           // 1621
        delete this.uiView;                                                                                            // 1622
      }                                                                                                                // 1623
                                                                                                                       //
      if (this.supervisor) {                                                                                           // 1625
        this.supervisor.close();                                                                                       // 1626
        delete this.supervisor;                                                                                        // 1627
      }                                                                                                                // 1628
                                                                                                                       //
      if (this.permissionsObserver) {                                                                                  // 1630
        this.permissionsObserver.stop();                                                                               // 1631
        delete this.permissionsObserver;                                                                               // 1632
      }                                                                                                                // 1633
    }                                                                                                                  // 1634
                                                                                                                       //
    return resetConnection;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.maybeRetryAfterError = function () {                                                                 //
    function maybeRetryAfterError(error, retryCount) {                                                                 //
      var _this8 = this;                                                                                               // 1636
                                                                                                                       //
      // If the error may be caused by the grain dying or a network failure, try to restart it,                        // 1637
      // returning a promise that resolves once restarted. Otherwise, just rethrow the error.                          // 1638
      // `retryCount` should be incremented for every successful retry as part of the same request;                    // 1639
      // we only want to retry once.                                                                                   // 1640
      if (shouldRestartGrain(error, retryCount)) {                                                                     // 1641
        this.resetConnection();                                                                                        // 1642
        return inMeteor(function () {                                                                                  // 1643
          _this8.supervisor = globalBackend.continueGrain(_this8.grainId).supervisor;                                  // 1644
        });                                                                                                            // 1645
      } else {                                                                                                         // 1646
        throw error;                                                                                                   // 1647
      }                                                                                                                // 1648
    }                                                                                                                  // 1649
                                                                                                                       //
    return maybeRetryAfterError;                                                                                       //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.doSessionInit = function () {                                                                        //
    function doSessionInit(request, response, requestPath) {                                                           //
      // jscs:disable disallowQuotedKeysInObjects                                                                      // 1652
      var path = requestPath || "/"; // Check that the path is relative (ie. starts with a /).                         // 1653
      // Also ensure that it doesn't start with 2 /, because that is interpreted as non-relative                       // 1656
                                                                                                                       //
      if (path.lastIndexOf("/", 0) !== 0 || path.lastIndexOf("//", 0) === 0) {                                         // 1657
        response.writeHead(400, "Invalid path supplied", {                                                             // 1658
          "Content-Type": "text/plain"                                                                                 // 1658
        });                                                                                                            // 1658
        response.end("Invalid path supplied.");                                                                        // 1659
        return;                                                                                                        // 1660
      } // Set the session ID.                                                                                         // 1661
                                                                                                                       //
                                                                                                                       //
      response.setHeader("Set-Cookie", ["sandstorm-sid=", this.sessionId, "; Max-Age=31536000; HttpOnly"].join(""));   // 1664
      response.setHeader("Cache-Control", "no-cache, private"); // Redirect to the app's root URL.                     // 1666
      // Note:  All browsers support relative locations and the next update to HTTP/1.1 will officially                // 1669
      //   make them valid.  http://tools.ietf.org/html/draft-ietf-httpbis-p2-semantics-26#page-67                     // 1670
                                                                                                                       //
      response.writeHead(303, "See Other", {                                                                           // 1671
        "Location": encodeURI(path)                                                                                    // 1671
      });                                                                                                              // 1671
      response.end();                                                                                                  // 1672
    }                                                                                                                  // 1673
                                                                                                                       //
    return doSessionInit;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.makeContext = function () {                                                                          //
    function makeContext(request, response, callback) {                                                                //
      // Parses the cookies from the request, checks that the session ID is present and valid, then                    // 1676
      // synchronously calls callback() passing it the request context object which contains the                       // 1677
      // other cookies. Throws an exception if the session ID is missing or invalid. The callback                      // 1678
      // must immediately use the context in an RPC. Once the callback returns, the context object                     // 1679
      // will be torn down.                                                                                            // 1680
      //                                                                                                               // 1681
      // `makeContext()` returns the result of `callback()`.                                                           // 1682
      var context = {};                                                                                                // 1684
                                                                                                                       //
      if (this.hostId) {                                                                                               // 1686
        var parseResult = parseCookies(request);                                                                       // 1687
                                                                                                                       //
        if (!parseResult.sessionId || parseResult.sessionId !== this.sessionId) {                                      // 1688
          throw new Meteor.Error(403, "Unauthorized");                                                                 // 1689
        }                                                                                                              // 1690
                                                                                                                       //
        if (parseResult.cookies.length > 0) {                                                                          // 1692
          context.cookies = parseResult.cookies;                                                                       // 1693
        }                                                                                                              // 1694
      } else {// jscs:disable disallowEmptyBlocks                                                                      // 1695
        // This is an API request. Cookies are not supported.                                                          // 1697
      }                                                                                                                // 1698
                                                                                                                       //
      context.accept = parseAcceptHeader(request);                                                                     // 1700
      context.acceptEncoding = parseAcceptEncodingHeader(request);                                                     // 1702
      context.eTagPrecondition = parsePreconditionHeader(request);                                                     // 1704
      context.additionalHeaders = [];                                                                                  // 1706
                                                                                                                       //
      for (var headerName in meteorBabelHelpers.sanitizeForInObject(request.headers)) {                                // 1707
        if (REQUEST_HEADER_WHITELIST.matches(headerName)) {                                                            // 1708
          context.additionalHeaders.push({                                                                             // 1709
            name: headerName,                                                                                          // 1710
            value: request.headers[headerName]                                                                         // 1711
          });                                                                                                          // 1709
        }                                                                                                              // 1713
      }                                                                                                                // 1714
                                                                                                                       //
      if (response) {                                                                                                  // 1716
        var promise = new Promise(function (resolve, reject) {                                                         // 1717
          response.resolveResponseStream = resolve;                                                                    // 1718
          response.rejectResponseStream = reject;                                                                      // 1719
        });                                                                                                            // 1720
        context.responseStream = new Capnp.Capability(promise, ByteStream);                                            // 1722
                                                                                                                       //
        try {                                                                                                          // 1724
          return callback(context);                                                                                    // 1725
        } finally {                                                                                                    // 1726
          context.responseStream.close();                                                                              // 1727
        }                                                                                                              // 1728
      } else {                                                                                                         // 1729
        // WebSocket -- no response object.                                                                            // 1730
        return callback(context);                                                                                      // 1731
      }                                                                                                                // 1732
    }                                                                                                                  // 1733
                                                                                                                       //
    return makeContext;                                                                                                //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.translateResponse = function () {                                                                    //
    function translateResponse(rpcResponse, response, request) {                                                       //
      var _this9 = this;                                                                                               // 1735
                                                                                                                       //
      if (this.hostId) {                                                                                               // 1736
        if (rpcResponse.setCookies && rpcResponse.setCookies.length > 0) {                                             // 1737
          response.setHeader("Set-Cookie", rpcResponse.setCookies.map(makeSetCookieHeader));                           // 1738
        }                                                                                                              // 1739
                                                                                                                       //
        var mainUrl = this.parentOrigin;                                                                               // 1741
        var grainHost = PROTOCOL + "//" + request.headers.host;                                                        // 1742
        response.setHeader("Content-Security-Policy", "frame-ancestors " + mainUrl + " " + grainHost);                 // 1743
        response.setHeader("X-Frame-Options", "ALLOW-FROM " + mainUrl); // TODO(security): Add a Content-Security-Policy header which:
        // (1) Prevents the app from initiating HTTP requests to third parties.                                        // 1747
        // (2) Prevents the app from navigating the parent frame.                                                      // 1748
        // (3) Prevents the app from opening popups.                                                                   // 1749
        // (4) Prohibits anyone other than the Sandstorm shell from framing the app (as a backup                       // 1750
        //   defense vs. clickjacking, though unguessable hostnames already mostly prevent this).                      // 1751
      } else {                                                                                                         // 1752
        // This is an API request. Cookies are not supported.                                                          // 1753
        // We need to make sure caches know that different bearer tokens get totally different results.                // 1755
        response.setHeader("Vary", "Authorization"); // APIs can be called from any origin. Because we ignore cookies, there is no security problem.
                                                                                                                       //
        response.setHeader("Access-Control-Allow-Origin", "*"); // Add a Content-Security-Policy as a backup in case someone finds a way to load this resource
        // in a browser context. This policy should thoroughly neuter it.                                              // 1762
                                                                                                                       //
        response.setHeader("Content-Security-Policy", "default-src 'none'; sandbox");                                  // 1763
      }                                                                                                                // 1764
                                                                                                                       //
      if (rpcResponse.additionalHeaders && rpcResponse.additionalHeaders.length > 0) {                                 // 1766
        // Add any additional prefixed/whitelisted headers to the response.                                            // 1767
        // We cannot trust that the headers in the RPC response are actually                                           // 1768
        // allowed, so we check if they are prefixed or whitelisted.                                                   // 1769
        rpcResponse.additionalHeaders.forEach(function (header) {                                                      // 1770
          if (RESPONSE_HEADER_WHITELIST.matches(header.name)) {                                                        // 1771
            // If header is prefixed with appHeaderPrefix,                                                             // 1772
            // or if the header name is in whitelist, set it.                                                          // 1773
            response.setHeader(header.name, header.value);                                                             // 1774
          }                                                                                                            // 1775
        });                                                                                                            // 1776
      } // On first response, update the session to have hasLoaded=true                                                // 1777
                                                                                                                       //
                                                                                                                       //
      this.setHasLoaded(); // TODO(security): Set X-Content-Type-Options: nosniff?                                     // 1780
                                                                                                                       //
      if ("content" in rpcResponse) {                                                                                  // 1784
        var content = rpcResponse.content;                                                                             // 1785
        var code = successCodes[content.statusCode];                                                                   // 1786
                                                                                                                       //
        if (!code) {                                                                                                   // 1787
          throw new Error("Unknown status code: ", content.statusCode);                                                // 1788
        }                                                                                                              // 1789
                                                                                                                       //
        if (content.mimeType) {                                                                                        // 1791
          response.setHeader("Content-Type", content.mimeType);                                                        // 1792
        }                                                                                                              // 1793
                                                                                                                       //
        if (content.encoding) {                                                                                        // 1795
          response.setHeader("Content-Encoding", content.encoding);                                                    // 1796
        }                                                                                                              // 1797
                                                                                                                       //
        if (content.language) {                                                                                        // 1799
          response.setHeader("Content-Language", content.language);                                                    // 1800
        }                                                                                                              // 1801
                                                                                                                       //
        if (content.eTag) {                                                                                            // 1803
          response.setHeader("ETag", composeETag(content.eTag));                                                       // 1804
        }                                                                                                              // 1805
                                                                                                                       //
        if ("disposition" in content && "download" in content.disposition) {                                           // 1807
          response.setHeader("Content-Disposition", 'attachment; filename="' + content.disposition.download.replace(/([\\"\n])/g, "\\$1") + '"');
        }                                                                                                              // 1810
                                                                                                                       //
        if ("stream" in content.body) {                                                                                // 1812
          if (request.method === "HEAD") {                                                                             // 1813
            content.body.stream.close();                                                                               // 1814
            response.rejectResponseStream(new Error("HEAD request; content doesn't matter."));                         // 1815
          } else if (response.socket.destroyed) {                                                                      // 1816
            // The client has already closed the stream. No additional "close" event will be                           // 1817
            // emitted. Annoyingly, we cannot rely on `response.write()` throwing an exception                         // 1818
            // in this case, so if we don't handle it specially we could end up with a                                 // 1819
            // `ResponseStream` that gets stuck waiting for a "drain" event that never arrives.                        // 1820
            content.body.stream.close();                                                                               // 1821
            response.rejectResponseStream(new Error("client disconnected"));                                           // 1822
          } else {                                                                                                     // 1823
            var idx = this.streamingResponseCounter;                                                                   // 1824
            this.streamingResponseCounter += 1;                                                                        // 1825
            var streamHandle = content.body.stream;                                                                    // 1826
            var rawResponseStream = new ResponseStream(response, code.id, code.title, streamHandle, function () {      // 1827
              delete _this9.streamingResponses[idx];                                                                   // 1829
            });                                                                                                        // 1829
            this.streamingResponses[idx] = rawResponseStream;                                                          // 1830
            var responseStream = new Capnp.Capability(rawResponseStream, ByteStream);                                  // 1831
            response.resolveResponseStream(responseStream);                                                            // 1832
            return {                                                                                                   // 1833
              streamHandle: responseStream                                                                             // 1833
            };                                                                                                         // 1833
          }                                                                                                            // 1834
        } else {                                                                                                       // 1835
          response.rejectResponseStream(new Error("Response content body was not a stream."));                         // 1836
                                                                                                                       //
          if ("bytes" in content.body) {                                                                               // 1839
            response.setHeader("Content-Length", content.body.bytes.length);                                           // 1840
          } else {                                                                                                     // 1841
            throw new Error("Unknown content body type.");                                                             // 1842
          }                                                                                                            // 1843
        }                                                                                                              // 1844
                                                                                                                       //
        response.writeHead(code.id, code.title);                                                                       // 1846
                                                                                                                       //
        if ("bytes" in content.body && request.method !== "HEAD") {                                                    // 1848
          response.write(content.body.bytes);                                                                          // 1849
        }                                                                                                              // 1850
                                                                                                                       //
        response.end();                                                                                                // 1852
      } else if ("noContent" in rpcResponse) {                                                                         // 1853
        var noContent = rpcResponse.noContent;                                                                         // 1854
        var noContentCode = noContentSuccessCodes[noContent.shouldResetForm * 1];                                      // 1855
                                                                                                                       //
        if (noContent.eTag) {                                                                                          // 1856
          response.setHeader("ETag", composeETag(noContent.eTag));                                                     // 1857
        }                                                                                                              // 1858
                                                                                                                       //
        response.writeHead(noContentCode.id, noContentCode.title);                                                     // 1860
        response.end();                                                                                                // 1861
      } else if ("preconditionFailed" in rpcResponse) {                                                                // 1862
        var preconditionFailed = rpcResponse.preconditionFailed;                                                       // 1863
                                                                                                                       //
        if (request.method === "GET" && "if-none-match" in request.headers) {                                          // 1864
          if (preconditionFailed.matchingETag) {                                                                       // 1865
            response.setHeader("ETag", composeETag(preconditionFailed.matchingETag));                                  // 1866
          }                                                                                                            // 1867
                                                                                                                       //
          response.writeHead(304, "Not Modified");                                                                     // 1869
        } else {                                                                                                       // 1870
          response.writeHead(412, "Precondition Failed");                                                              // 1871
        }                                                                                                              // 1872
                                                                                                                       //
        response.end();                                                                                                // 1874
      } else if ("redirect" in rpcResponse) {                                                                          // 1875
        var redirect = rpcResponse.redirect;                                                                           // 1876
        var redirectCode = redirectCodes[redirect.switchToGet * 2 + redirect.isPermanent];                             // 1877
        response.writeHead(redirectCode.id, redirectCode.title, {                                                      // 1878
          "Location": encodeURI(redirect.location)                                                                     // 1879
        });                                                                                                            // 1878
        response.end();                                                                                                // 1881
      } else if ("clientError" in rpcResponse) {                                                                       // 1882
        var clientError = rpcResponse.clientError;                                                                     // 1883
        var errorCode = errorCodes[clientError.statusCode];                                                            // 1884
                                                                                                                       //
        if (!errorCode) {                                                                                              // 1885
          throw new Error("Unknown status code: ", clientError.statusCode);                                            // 1886
        }                                                                                                              // 1887
                                                                                                                       //
        response.writeHead(errorCode.id, errorCode.title, {                                                            // 1889
          "Content-Type": "text/html"                                                                                  // 1890
        });                                                                                                            // 1889
                                                                                                                       //
        if (request.method !== "HEAD") {                                                                               // 1893
          if (clientError.descriptionHtml) {                                                                           // 1894
            response.write(clientError.descriptionHtml);                                                               // 1895
          } else {                                                                                                     // 1896
            // TODO(someday):  Better default error page.                                                              // 1897
            response.write("<html><body><h1>" + errorCode.id + ": " + errorCode.title + "</h1></body></html>");        // 1898
          }                                                                                                            // 1900
        }                                                                                                              // 1901
                                                                                                                       //
        response.end();                                                                                                // 1903
      } else if ("serverError" in rpcResponse) {                                                                       // 1904
        response.writeHead(500, "Internal Server Error", {                                                             // 1905
          "Content-Type": "text/html"                                                                                  // 1906
        });                                                                                                            // 1905
                                                                                                                       //
        if (request.method !== "HEAD") {                                                                               // 1909
          if (rpcResponse.serverError.descriptionHtml) {                                                               // 1910
            response.write(rpcResponse.serverError.descriptionHtml);                                                   // 1911
          } else {                                                                                                     // 1912
            // TODO(someday):  Better default error page.                                                              // 1913
            response.write("<html><body><h1>500: Internal Server Error</h1></body></html>");                           // 1914
          }                                                                                                            // 1915
        }                                                                                                              // 1916
                                                                                                                       //
        response.end();                                                                                                // 1918
      } else {                                                                                                         // 1919
        throw new Error("Unknown HTTP response type:\n" + JSON.stringify(rpcResponse));                                // 1920
      }                                                                                                                // 1921
                                                                                                                       //
      return {};                                                                                                       // 1923
    }                                                                                                                  // 1924
                                                                                                                       //
    return translateResponse;                                                                                          //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.handleRequest = function () {                                                                        //
    function handleRequest(request, data, response, retryCount) {                                                      //
      var _this10 = this;                                                                                              // 1926
                                                                                                                       //
      return Promise.resolve(undefined).then(function () {                                                             // 1927
        return _this10.makeContext(request, response, function (context) {                                             // 1928
          // jscs:disable requireDotNotation                                                                           // 1929
          // Send the RPC.                                                                                             // 1930
          var path = request.url.slice(1); // remove leading '/'                                                       // 1931
                                                                                                                       //
          var session = _this10.getSession(request);                                                                   // 1932
                                                                                                                       //
          var requestContent = function () {                                                                           // 1934
            return {                                                                                                   // 1935
              content: data,                                                                                           // 1936
              encoding: request.headers["content-encoding"],                                                           // 1937
              mimeType: request.headers["content-type"]                                                                // 1938
            };                                                                                                         // 1935
          };                                                                                                           // 1940
                                                                                                                       //
          var xmlContent = function () {                                                                               // 1942
            var type = request.headers["content-type"] || "application/xml;charset=utf-8";                             // 1943
            var match = type.match(/[^/]*\/xml(; *charset *= *([^ ;]*))?/);                                            // 1944
                                                                                                                       //
            if (!match) {                                                                                              // 1945
              response.writeHead(415, "Unsupported media type.", {                                                     // 1946
                "Content-Type": "text/plain"                                                                           // 1947
              });                                                                                                      // 1946
              response.end("expected XML request body");                                                               // 1949
              throw new Error("expected XML request body");                                                            // 1950
            }                                                                                                          // 1951
                                                                                                                       //
            var charset = match[2] || "ISO-8859-1";                                                                    // 1953
            var encoding = request.headers["content-encoding"];                                                        // 1955
                                                                                                                       //
            if (encoding && encoding !== "identity") {                                                                 // 1956
              if (encoding !== "gzip") throw new Error("unknown Content-Encoding: " + encoding);                       // 1957
              data = gunzipSync(data);                                                                                 // 1958
            }                                                                                                          // 1959
                                                                                                                       //
            return data.toString(charset.toLowerCase() === "utf-8" ? "utf8" : "binary");                               // 1961
          };                                                                                                           // 1962
                                                                                                                       //
          var propfindDepth = function () {                                                                            // 1964
            var depth = request.headers["depth"];                                                                      // 1965
            return depth === "0" ? "zero" : depth === "1" ? "one" : "infinity";                                        // 1966
          };                                                                                                           // 1969
                                                                                                                       //
          var shallow = function () {                                                                                  // 1971
            return request.headers["depth"] === "0";                                                                   // 1972
          };                                                                                                           // 1973
                                                                                                                       //
          var noOverwrite = function () {                                                                              // 1975
            return (request.headers["overwrite"] || "").toLowerCase() === "f";                                         // 1976
          };                                                                                                           // 1977
                                                                                                                       //
          var destination = function () {                                                                              // 1979
            var result = request.headers["destination"];                                                               // 1980
            if (!result) throw new Error("missing destination");                                                       // 1981
            return Url.parse(result).path.slice(1); // remove leading '/'                                              // 1982
          };                                                                                                           // 1983
                                                                                                                       //
          if (request.method === "GET" || request.method === "HEAD") {                                                 // 1985
            return session.get(path, context, request.method === "HEAD");                                              // 1986
          } else if (request.method === "POST") {                                                                      // 1987
            return session.post(path, requestContent(), context);                                                      // 1988
          } else if (request.method === "PUT") {                                                                       // 1989
            return session.put(path, requestContent(), context);                                                       // 1990
          } else if (request.method === "PATCH") {                                                                     // 1991
            return session.patch(path, requestContent(), context);                                                     // 1992
          } else if (request.method === "DELETE") {                                                                    // 1993
            return session.delete(path, context);                                                                      // 1994
          } else if (request.method === "PROPFIND") {                                                                  // 1995
            return session.propfind(path, xmlContent(), propfindDepth(), context);                                     // 1996
          } else if (request.method === "PROPPATCH") {                                                                 // 1997
            return session.proppatch(path, xmlContent(), context);                                                     // 1998
          } else if (request.method === "MKCOL") {                                                                     // 1999
            return session.mkcol(path, requestContent(), context);                                                     // 2000
          } else if (request.method === "COPY") {                                                                      // 2001
            return session.copy(path, destination(), noOverwrite(), shallow(), context);                               // 2002
          } else if (request.method === "MOVE") {                                                                      // 2003
            return session.move(path, destination(), noOverwrite(), context);                                          // 2004
          } else if (request.method === "LOCK") {                                                                      // 2005
            return session.lock(path, xmlContent(), shallow(), context);                                               // 2006
          } else if (request.method === "UNLOCK") {                                                                    // 2007
            return session.unlock(path, request.headers["lock-token"], context);                                       // 2008
          } else if (request.method === "ACL") {                                                                       // 2009
            return session.acl(path, xmlContent(), context);                                                           // 2010
          } else if (request.method === "REPORT") {                                                                    // 2011
            return session.report(path, requestContent(), context);                                                    // 2012
          } else if (request.method === "OPTIONS") {                                                                   // 2013
            return session.options(path, context).then(function (options) {                                            // 2014
              var dav = [];                                                                                            // 2015
              if (options.davClass1) dav.push("1");                                                                    // 2016
              if (options.davClass2) dav.push("2");                                                                    // 2017
              if (options.davClass3) dav.push("3");                                                                    // 2018
                                                                                                                       //
              if (options.davExtensions) {                                                                             // 2019
                options.davExtensions.forEach(function (token) {                                                       // 2020
                  if (token.match(/^([a-zA-Z0-9!#$%&'*+.^_`|~-]+|<[\x21-\x7E]*>)$/)) {                                 // 2021
                    dav.push(token);                                                                                   // 2022
                  }                                                                                                    // 2023
                });                                                                                                    // 2024
              }                                                                                                        // 2025
                                                                                                                       //
              if (dav.length > 0) {                                                                                    // 2027
                response.setHeader("DAV", dav.join(", "));                                                             // 2028
                response.setHeader("Access-Control-Expose-Headers", "DAV");                                            // 2029
              }                                                                                                        // 2030
                                                                                                                       //
              response.end(); // Return no response; we already handled everything.                                    // 2032
            }, function (err) {                                                                                        // 2034
              if (err.kjType !== "unimplemented") throw err;                                                           // 2035
              response.end(); // Return no response; we already handled everything.                                    // 2036
            });                                                                                                        // 2038
          } else {                                                                                                     // 2039
            throw new Error("Sandstorm only supports the following methods: GET, POST, PUT, PATCH, DELETE, HEAD, PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK, ACL, REPORT, and OPTIONS.");
          }                                                                                                            // 2041
        });                                                                                                            // 2042
      }).then(function (rpcResponse) {                                                                                 // 2043
        if (rpcResponse !== undefined) {                                                                               // 2044
          // Will be undefined for OPTIONS request.                                                                    // 2044
          return _this10.translateResponse(rpcResponse, response, request);                                            // 2045
        }                                                                                                              // 2046
      }).catch(function (error) {                                                                                      // 2047
        return _this10.maybeRetryAfterError(error, retryCount).then(function () {                                      // 2048
          return _this10.handleRequest(request, data, response, retryCount + 1);                                       // 2049
        });                                                                                                            // 2050
      });                                                                                                              // 2051
    }                                                                                                                  // 2052
                                                                                                                       //
    return handleRequest;                                                                                              //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.handleRequestStreaming = function () {                                                               //
    function handleRequestStreaming(request, response, contentLength, retryCount) {                                    //
      var _this11 = this;                                                                                              // 2054
                                                                                                                       //
      return this.makeContext(request, response, function (context) {                                                  // 2055
        var path = request.url.slice(1); // remove leading '/'                                                         // 2056
                                                                                                                       //
        var session = _this11.getSession(request);                                                                     // 2057
                                                                                                                       //
        var mimeType = request.headers["content-type"] || "application/octet-stream";                                  // 2059
        var encoding = request.headers["content-encoding"];                                                            // 2060
        var requestStreamPromise = void 0;                                                                             // 2062
                                                                                                                       //
        if (request.method === "POST") {                                                                               // 2063
          requestStreamPromise = session.postStreaming(path, mimeType, context, encoding);                             // 2064
        } else if (request.method === "PUT") {                                                                         // 2065
          requestStreamPromise = session.putStreaming(path, mimeType, context, encoding);                              // 2066
        } else {                                                                                                       // 2067
          throw new Error("Sandstorm only supports streaming POST and PUT requests.");                                 // 2068
        } // TODO(perf): We ought to be pipelining the body, but we can't currently, because we have to                // 2069
        //   handle the case where the app doesn't actually support streaming. We could pipeline while                 // 2072
        //   also buffering the data on the side in case we need it again later, but that's kind of                    // 2073
        //   complicated. We should fix the whole protocol to make streaming the standard.                             // 2074
                                                                                                                       //
                                                                                                                       //
        return requestStreamPromise.then(function (requestStreamResult) {                                              // 2075
          var requestStream = requestStreamResult.stream; // Initialized when getResponse() returns, if the response is streaming.
                                                                                                                       //
          var downloadStreamHandle = void 0; // Initialized if an upload-stream method throws.                         // 2079
                                                                                                                       //
          var uploadStreamError = void 0; // We call `getResponse()` immediately so that the app can start streaming data down even
          // while data is still being streamed up. This theoretically allows apps to perform                          // 2085
          // bidirectional streaming, though probably very few actually do that.                                       // 2086
          //                                                                                                           // 2087
          // Note that we need to be able to cancel `responsePromise` below, so it's important that                    // 2088
          // it is the raw Cap'n Proto promise. Hence `translateResponsePromise` is a separate                         // 2089
          // variable.                                                                                                 // 2090
                                                                                                                       //
          var responsePromise = requestStream.getResponse();                                                           // 2091
                                                                                                                       //
          var reportUploadStreamError = function (err) {                                                               // 2093
            // Called when an upload-stream method throws.                                                             // 2094
            if (!uploadStreamError) {                                                                                  // 2096
              uploadStreamError = err; // If we're still waiting on any response stuff, cancel it.                     // 2097
                                                                                                                       //
              responsePromise.cancel();                                                                                // 2100
              requestStream.close();                                                                                   // 2101
                                                                                                                       //
              if (downloadStreamHandle) {                                                                              // 2102
                downloadStreamHandle.close();                                                                          // 2103
              }                                                                                                        // 2104
            }                                                                                                          // 2105
          }; // If we have a Content-Length, pass it along to the app by calling `expectSize()`.                       // 2106
                                                                                                                       //
                                                                                                                       //
          if (contentLength !== undefined) {                                                                           // 2109
            requestStream.expectSize(contentLength).catch(function (err) {                                             // 2110
              // expectSize() is allowed to be unimplemented.                                                          // 2111
              if (err.kjType !== "unimplemented") {                                                                    // 2112
                reportUploadStreamError(err);                                                                          // 2113
              }                                                                                                        // 2114
            });                                                                                                        // 2115
          } // Pipe the input stream to the app.                                                                       // 2116
                                                                                                                       //
                                                                                                                       //
          var writesInFlight = 0;                                                                                      // 2119
          var readingPaused = false;                                                                                   // 2120
          request.on("data", function (buf) {                                                                          // 2121
            // TODO(someday): Coalesce small writes.                                                                   // 2122
            if (buf.length > 4194304) {                                                                                // 2124
              // Whoa, buffer is more than 4MB. We should be splitting it. But does this actually                      // 2125
              // happen? Let's try to detect this problem before we try to do anything about it.                       // 2126
              console.log("whoa, huge upload buffer:", buf.length);                                                    // 2127
              reportUploadStreamError(new Error("upload buffer too big (bug in Sandstorm)"));                          // 2128
              destructor();                                                                                            // 2129
              return;                                                                                                  // 2130
            }                                                                                                          // 2131
                                                                                                                       //
            if (!uploadStreamError) {                                                                                  // 2133
              requestStream.write(buf).then(function () {                                                              // 2134
                writesInFlight--;                                                                                      // 2135
                                                                                                                       //
                if (readingPaused) {                                                                                   // 2136
                  request.resume();                                                                                    // 2137
                  readingPaused = false;                                                                               // 2138
                }                                                                                                      // 2139
              }).catch(reportUploadStreamError);                                                                       // 2140
                                                                                                                       //
              if (writesInFlight++ > 16) {                                                                             // 2142
                request.pause();                                                                                       // 2143
                readingPaused = true;                                                                                  // 2144
              }                                                                                                        // 2145
            }                                                                                                          // 2146
          });                                                                                                          // 2147
          var idx = _this11.streamingRequestCounter;                                                                   // 2149
          _this11.streamingRequestCounter += 1;                                                                        // 2150
          _this11.streamingRequests[idx] = request;                                                                    // 2151
                                                                                                                       //
          destructor = function () {                                                                                   // 2152
            delete _this11.streamingRequests[idx];                                                                     // 2152
          };                                                                                                           // 2152
                                                                                                                       //
          request.on("end", function () {                                                                              // 2154
            if (!uploadStreamError) requestStream.done().catch(reportUploadStreamError); // We're all done making calls to requestStream.
                                                                                                                       //
            requestStream.close();                                                                                     // 2158
            destructor();                                                                                              // 2159
          });                                                                                                          // 2160
          request.on("close", function () {                                                                            // 2162
            reportUploadStreamError(new Error("HTTP connection unexpectedly closed during request."));                 // 2163
            destructor();                                                                                              // 2164
          });                                                                                                          // 2165
          request.on("error", function (err) {                                                                         // 2167
            reportUploadStreamError(err);                                                                              // 2168
            destructor();                                                                                              // 2169
          });                                                                                                          // 2170
          return responsePromise.then(function (rpcResponse) {                                                         // 2172
            // Stop here if the upload stream has already failed.                                                      // 2173
            if (uploadStreamError) throw uploadStreamError;                                                            // 2174
            downloadStreamHandle = _this11.translateResponse(rpcResponse, response, request).streamHandle;             // 2175
          });                                                                                                          // 2176
        }, function (err) {                                                                                            // 2177
          if (err.kjType === "failed" && err.message.indexOf("not implemented") !== -1) {                              // 2178
            // Hack to work around old apps using an old version of Cap'n Proto, before the                            // 2179
            // 'unimplemented' exception type was introduced. :(                                                       // 2180
            // TODO(cleanup): When we transition to API version 2, we can move this into the                           // 2181
            //   compatibility layer.                                                                                  // 2182
            err.kjType = "unimplemented";                                                                              // 2183
          }                                                                                                            // 2184
                                                                                                                       //
          if (shouldRestartGrain(err, 0)) {                                                                            // 2186
            // This is the kind of error that indicates we should retry. Note that we passed 0 for                     // 2187
            // the retry count above because we were just checking if this is a retriable error (vs.                   // 2188
            // possibly a method-not-implemented error); maybeRetryAfterError() will check again with                  // 2189
            // the proper retry count.                                                                                 // 2190
            return _this11.maybeRetryAfterError(err, retryCount).then(function () {                                    // 2191
              return _this11.handleRequestStreaming(request, response, contentLength, retryCount + 1);                 // 2192
            });                                                                                                        // 2193
          } else if (err.kjType === "unimplemented") {                                                                 // 2194
            // Streaming is not implemented. Fall back to non-streaming version.                                       // 2195
            return readAll(request).then(function (data) {                                                             // 2196
              return _this11.handleRequest(request, data, response, 0);                                                // 2197
            });                                                                                                        // 2198
          } else {                                                                                                     // 2199
            throw err;                                                                                                 // 2200
          }                                                                                                            // 2201
        });                                                                                                            // 2202
      });                                                                                                              // 2203
    }                                                                                                                  // 2204
                                                                                                                       //
    return handleRequestStreaming;                                                                                     //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.handleWebSocket = function () {                                                                      //
    function handleWebSocket(request, socket, head, retryCount) {                                                      //
      var _this12 = this;                                                                                              // 2206
                                                                                                                       //
      return Promise.resolve(undefined).then(function () {                                                             // 2207
        return _this12.makeContext(request, undefined, function (context) {                                            // 2208
          var path = request.url.slice(1); // remove leading '/'                                                       // 2209
                                                                                                                       //
          var session = _this12.getSession(request);                                                                   // 2210
                                                                                                                       //
          if (!("sec-websocket-key" in request.headers)) {                                                             // 2212
            throw new Error("Missing Sec-WebSocket-Accept header.");                                                   // 2213
          }                                                                                                            // 2214
                                                                                                                       //
          var magic = request.headers["sec-websocket-key"] + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11";                   // 2216
          var acceptKey = Crypto.createHash("sha1").update(magic).digest("base64");                                    // 2217
          var protocols = [];                                                                                          // 2219
                                                                                                                       //
          if ("sec-websocket-protocol" in request.headers) {                                                           // 2220
            protocols = request.headers["sec-websocket-protocol"].split(",").map(function (s) {                        // 2221
              return s.trim();                                                                                         // 2222
            });                                                                                                        // 2222
          }                                                                                                            // 2223
                                                                                                                       //
          var receiver = new WebSocketReceiver(socket);                                                                // 2225
          var promise = session.openWebSocket(path, context, protocols, receiver);                                     // 2227
                                                                                                                       //
          if (head.length > 0) {                                                                                       // 2229
            promise.serverStream.sendBytes(head);                                                                      // 2230
          }                                                                                                            // 2231
                                                                                                                       //
          var socketIdx = _this12.websocketCounter;                                                                    // 2233
          _this12.websockets[socketIdx] = socket;                                                                      // 2234
          _this12.websocketCounter += 1;                                                                               // 2235
          pumpWebSocket(socket, promise.serverStream, function () {                                                    // 2236
            delete _this12.websockets[socketIdx];                                                                      // 2236
          });                                                                                                          // 2236
          return promise.then(function (response) {                                                                    // 2238
            var headers = ["HTTP/1.1 101 Switching Protocols", "Upgrade: websocket", "Connection: Upgrade", "Sec-WebSocket-Accept: " + acceptKey];
                                                                                                                       //
            if (response.protocol && response.protocol.length > 0) {                                                   // 2245
              headers.push("Sec-WebSocket-Protocol: " + response.protocol.join(", "));                                 // 2246
            }                                                                                                          // 2247
                                                                                                                       //
            headers.push("");                                                                                          // 2249
            headers.push("");                                                                                          // 2250
            socket.write(headers.join("\r\n"));                                                                        // 2252
            receiver.go(); // Note:  At this point errors are out of our hands.                                        // 2253
          });                                                                                                          // 2256
        });                                                                                                            // 2257
      }).catch(function (error) {                                                                                      // 2258
        return _this12.maybeRetryAfterError(error, retryCount).then(function () {                                      // 2259
          return _this12.handleWebSocket(request, socket, head, retryCount + 1);                                       // 2260
        });                                                                                                            // 2261
      });                                                                                                              // 2262
    }                                                                                                                  // 2263
                                                                                                                       //
    return handleWebSocket;                                                                                            //
  }();                                                                                                                 //
                                                                                                                       //
  Proxy.prototype.setHasLoaded = function () {                                                                         //
    function setHasLoaded() {                                                                                          //
      if (!this.hasLoaded) {                                                                                           // 2266
        this.hasLoaded = true;                                                                                         // 2267
        var sessionId = this.sessionId;                                                                                // 2268
        inMeteor(function () {                                                                                         // 2269
          Sessions.update({                                                                                            // 2270
            _id: sessionId                                                                                             // 2270
          }, {                                                                                                         // 2270
            $set: {                                                                                                    // 2270
              hasLoaded: true                                                                                          // 2270
            }                                                                                                          // 2270
          });                                                                                                          // 2270
        });                                                                                                            // 2271
      }                                                                                                                // 2272
    }                                                                                                                  // 2273
                                                                                                                       //
    return setHasLoaded;                                                                                               //
  }();                                                                                                                 //
                                                                                                                       //
  return Proxy;                                                                                                        //
}();                                                                                                                   //
                                                                                                                       //
var isRfc1918OrLocal = function (address) {                                                                            // 2276
  if (Net.isIPv4(address)) {                                                                                           // 2277
    var quad = address.split(".").map(function (x) {                                                                   // 2278
      return parseInt(x, 10);                                                                                          // 2278
    });                                                                                                                // 2278
    return quad[0] === 127 || quad[0] === 10 || quad[0] === 192 && quad[1] === 168 || quad[0] === 172 && quad[1] >= 16 && quad[1] < 32;
  } else if (Net.isIPv6(address)) {                                                                                    // 2282
    // IPv6 specifies ::1 as localhost and fd:: as reserved for private networks                                       // 2283
    return address === "::1" || address.lastIndexOf("fd", 0) === 0;                                                    // 2284
  } else {                                                                                                             // 2285
    // Ignore things that are neither IPv4 nor IPv6                                                                    // 2286
    return false;                                                                                                      // 2287
  }                                                                                                                    // 2288
};                                                                                                                     // 2289
                                                                                                                       //
var quadToIntString = function (quad) {                                                                                // 2291
  var num = Bignum(quad[0]).shiftLeft(16).add(quad[1]).shiftLeft(16).add(quad[2]).shiftLeft(16).add(quad[3]);          // 2292
  return num.toString();                                                                                               // 2296
}; // -----------------------------------------------------------------------------                                    // 2297
// Session cookie management                                                                                           // 2300
                                                                                                                       //
                                                                                                                       //
var parseCookies = function (request) {                                                                                // 2302
  // jscs:disable requireDotNotation                                                                                   // 2303
  var header = request.headers["cookie"];                                                                              // 2304
  var result = {                                                                                                       // 2306
    cookies: []                                                                                                        // 2306
  };                                                                                                                   // 2306
                                                                                                                       //
  if (header) {                                                                                                        // 2307
    var reqCookies = header.split(";");                                                                                // 2308
                                                                                                                       //
    for (var i in meteorBabelHelpers.sanitizeForInObject(reqCookies)) {                                                // 2309
      var reqCookie = reqCookies[i];                                                                                   // 2310
      var equalsPos = reqCookie.indexOf("=");                                                                          // 2311
      var cookie = void 0;                                                                                             // 2312
                                                                                                                       //
      if (equalsPos === -1) {                                                                                          // 2313
        cookie = {                                                                                                     // 2314
          key: reqCookie.trim(),                                                                                       // 2314
          value: ""                                                                                                    // 2314
        };                                                                                                             // 2314
      } else {                                                                                                         // 2315
        cookie = {                                                                                                     // 2316
          key: reqCookie.slice(0, equalsPos).trim(),                                                                   // 2316
          value: reqCookie.slice(equalsPos + 1)                                                                        // 2316
        };                                                                                                             // 2316
      }                                                                                                                // 2317
                                                                                                                       //
      if (cookie.key === "sandstorm-sid") {                                                                            // 2319
        if (result.sessionId) {                                                                                        // 2320
          throw new Error("Multiple sandstorm session IDs?");                                                          // 2321
        }                                                                                                              // 2322
                                                                                                                       //
        result.sessionId = cookie.value;                                                                               // 2324
      } else {                                                                                                         // 2325
        result.cookies.push(cookie);                                                                                   // 2326
      }                                                                                                                // 2327
    }                                                                                                                  // 2328
  }                                                                                                                    // 2329
                                                                                                                       //
  return result;                                                                                                       // 2331
};                                                                                                                     // 2332
                                                                                                                       //
var parsePreconditionHeader = function (request) {                                                                     // 2334
  if (request.headers["if-match"]) {                                                                                   // 2335
    if (request.headers["if-match"].trim() === "*") {                                                                  // 2336
      return {                                                                                                         // 2337
        exists: null                                                                                                   // 2337
      };                                                                                                               // 2337
    }                                                                                                                  // 2338
                                                                                                                       //
    var matches = parseETagList(request.headers["if-match"]);                                                          // 2340
                                                                                                                       //
    if (matches.length > 0) {                                                                                          // 2341
      return {                                                                                                         // 2342
        matchesOneOf: matches                                                                                          // 2342
      };                                                                                                               // 2342
    }                                                                                                                  // 2343
  }                                                                                                                    // 2344
                                                                                                                       //
  if (request.headers["if-none-match"]) {                                                                              // 2346
    if (request.headers["if-none-match"].trim() === "*") {                                                             // 2347
      return {                                                                                                         // 2348
        doesntExist: null                                                                                              // 2348
      };                                                                                                               // 2348
    }                                                                                                                  // 2349
                                                                                                                       //
    var noneMatches = parseETagList(request.headers["if-none-match"]);                                                 // 2351
                                                                                                                       //
    if (noneMatches.length > 0) {                                                                                      // 2352
      return {                                                                                                         // 2353
        matchesNoneOf: noneMatches                                                                                     // 2353
      };                                                                                                               // 2353
    }                                                                                                                  // 2354
  }                                                                                                                    // 2355
                                                                                                                       //
  return {                                                                                                             // 2357
    none: null                                                                                                         // 2357
  };                                                                                                                   // 2357
};                                                                                                                     // 2358
                                                                                                                       //
var parseETagList = function (input) {                                                                                 // 2360
  // An ETag is a quoted, \-escaped string, possibly prefixed with W/ (outside the quotes) to                          // 2361
  // indicate that it is weak. We are parsing a list of comma-delimited etags.                                         // 2362
  input = input.trim();                                                                                                // 2364
  var results = [];                                                                                                    // 2365
                                                                                                                       //
  while (input.length > 0) {                                                                                           // 2367
    var match = input.match(/^\s*(W\/)?"(([^"\\]|\\.)*)"\s*($|,)/);                                                    // 2368
    if (!match) throw new Meteor.Error(400, "invalid etag");                                                           // 2369
    input = input.slice(match[0].length).trim();                                                                       // 2371
    results.push({                                                                                                     // 2372
      weak: !!match[1],                                                                                                // 2372
      value: match[2].replace(/\\(.)/g, "$1")                                                                          // 2372
    });                                                                                                                // 2372
  }                                                                                                                    // 2373
                                                                                                                       //
  return results;                                                                                                      // 2375
};                                                                                                                     // 2376
                                                                                                                       //
var composeETag = function (tag) {                                                                                     // 2378
  var result = '"' + (tag.value || "").replace(/([\\"])/g, "\\$1") + '"';                                              // 2379
  if (tag.weak) result = "W/" + result;                                                                                // 2380
  return result;                                                                                                       // 2381
};                                                                                                                     // 2382
                                                                                                                       //
var parseAcceptHeaderString = function (header, field) {                                                               // 2384
  var result = [];                                                                                                     // 2385
                                                                                                                       //
  if (header) {                                                                                                        // 2386
    var acceptList = header.split(",");                                                                                // 2387
                                                                                                                       //
    for (var i in meteorBabelHelpers.sanitizeForInObject(acceptList)) {                                                // 2388
      var acceptStr = acceptList[i];                                                                                   // 2389
      var tokensList = acceptStr.split(";");                                                                           // 2390
      var temp = {};                                                                                                   // 2392
      temp[field] = tokensList[0].trim();                                                                              // 2393
      var tokensListRest = tokensList.slice(1);                                                                        // 2395
                                                                                                                       //
      for (var j in meteorBabelHelpers.sanitizeForInObject(tokensListRest)) {                                          // 2396
        var token = tokensListRest[j];                                                                                 // 2397
        var equalsPos = token.indexOf("=");                                                                            // 2398
                                                                                                                       //
        if (equalsPos) {                                                                                               // 2399
          var key = token.slice(0, equalsPos).trim();                                                                  // 2400
          var value = token.slice(equalsPos + 1).trim();                                                               // 2401
                                                                                                                       //
          if (key === "q") {                                                                                           // 2403
            temp.qValue = +value;                                                                                      // 2404
          }                                                                                                            // 2405
        }                                                                                                              // 2406
      }                                                                                                                // 2407
                                                                                                                       //
      result.push(temp);                                                                                               // 2409
    }                                                                                                                  // 2410
  }                                                                                                                    // 2411
                                                                                                                       //
  return result;                                                                                                       // 2413
};                                                                                                                     // 2414
                                                                                                                       //
var parseAcceptHeader = function (request) {                                                                           // 2416
  // jscs:disable requireDotNotation                                                                                   // 2417
  var header = request.headers["accept"];                                                                              // 2418
  return parseAcceptHeaderString(header, "mimeType");                                                                  // 2419
};                                                                                                                     // 2420
                                                                                                                       //
var parseAcceptEncodingHeader = function (request) {                                                                   // 2422
  // jscs:disable requireDotNotation                                                                                   // 2423
  var header = request.headers["accept-encoding"];                                                                     // 2424
  return parseAcceptHeaderString(header, "contentCoding");                                                             // 2425
}; // -----------------------------------------------------------------------------                                    // 2426
// Regular HTTP request handling                                                                                       // 2429
                                                                                                                       //
                                                                                                                       //
var readAll = function (stream) {                                                                                      // 2431
  return new Promise(function (resolve, reject) {                                                                      // 2432
    var buffers = [];                                                                                                  // 2433
    stream.on("data", function (buf) {                                                                                 // 2434
      buffers.push(buf);                                                                                               // 2435
    });                                                                                                                // 2436
    stream.on("end", function () {                                                                                     // 2438
      resolve(Buffer.concat(buffers));                                                                                 // 2439
    });                                                                                                                // 2440
    stream.on("error", reject);                                                                                        // 2442
  });                                                                                                                  // 2443
};                                                                                                                     // 2444
                                                                                                                       //
var makeSetCookieHeader = function (cookie) {                                                                          // 2446
  var result = [cookie.name, "=", cookie.value];                                                                       // 2447
                                                                                                                       //
  if ("absolute" in cookie.expires) {                                                                                  // 2449
    result.push("; Expires=");                                                                                         // 2450
    result.push(new Date(cookie.expires.absolute * 1000).toUTCString());                                               // 2451
  } else if ("relative" in cookie.expires) {                                                                           // 2452
    result.push("; Max-Age=" + cookie.expires.relative);                                                               // 2453
  }                                                                                                                    // 2454
                                                                                                                       //
  if (cookie.path) {                                                                                                   // 2456
    result.push("; Path=" + cookie.path);                                                                              // 2457
  }                                                                                                                    // 2458
                                                                                                                       //
  if (cookie.httpOnly) {                                                                                               // 2460
    result.push("; HttpOnly");                                                                                         // 2461
  }                                                                                                                    // 2462
                                                                                                                       //
  return result.join("");                                                                                              // 2464
}; // TODO(cleanup):  Auto-generate based on annotations in web-session.capnp.                                         // 2465
                                                                                                                       //
                                                                                                                       //
var successCodes = {                                                                                                   // 2468
  ok: {                                                                                                                // 2469
    id: 200,                                                                                                           // 2469
    title: "OK"                                                                                                        // 2469
  },                                                                                                                   // 2469
  created: {                                                                                                           // 2470
    id: 201,                                                                                                           // 2470
    title: "Created"                                                                                                   // 2470
  },                                                                                                                   // 2470
  accepted: {                                                                                                          // 2471
    id: 202,                                                                                                           // 2471
    title: "Accepted"                                                                                                  // 2471
  },                                                                                                                   // 2471
  multiStatus: {                                                                                                       // 2472
    id: 207,                                                                                                           // 2472
    title: "Multi-Status"                                                                                              // 2472
  }                                                                                                                    // 2472
};                                                                                                                     // 2468
var noContentSuccessCodes = [// Indexed by shouldResetForm * 1                                                         // 2474
{                                                                                                                      // 2476
  id: 204,                                                                                                             // 2476
  title: "No Content"                                                                                                  // 2476
}, {                                                                                                                   // 2476
  id: 205,                                                                                                             // 2477
  title: "Reset Content"                                                                                               // 2477
}];                                                                                                                    // 2477
var redirectCodes = [// Indexed by switchToGet * 2 + isPermanent                                                       // 2479
{                                                                                                                      // 2481
  id: 307,                                                                                                             // 2481
  title: "Temporary Redirect"                                                                                          // 2481
}, {                                                                                                                   // 2481
  id: 308,                                                                                                             // 2482
  title: "Permanent Redirect"                                                                                          // 2482
}, {                                                                                                                   // 2482
  id: 303,                                                                                                             // 2483
  title: "See Other"                                                                                                   // 2483
}, {                                                                                                                   // 2483
  id: 301,                                                                                                             // 2484
  title: "Moved Permanently"                                                                                           // 2484
}];                                                                                                                    // 2484
var errorCodes = {                                                                                                     // 2486
  badRequest: {                                                                                                        // 2487
    id: 400,                                                                                                           // 2487
    title: "Bad Request"                                                                                               // 2487
  },                                                                                                                   // 2487
  forbidden: {                                                                                                         // 2488
    id: 403,                                                                                                           // 2488
    title: "Forbidden"                                                                                                 // 2488
  },                                                                                                                   // 2488
  notFound: {                                                                                                          // 2489
    id: 404,                                                                                                           // 2489
    title: "Not Found"                                                                                                 // 2489
  },                                                                                                                   // 2489
  methodNotAllowed: {                                                                                                  // 2490
    id: 405,                                                                                                           // 2490
    title: "Method Not Allowed"                                                                                        // 2490
  },                                                                                                                   // 2490
  notAcceptable: {                                                                                                     // 2491
    id: 406,                                                                                                           // 2491
    title: "Not Acceptable"                                                                                            // 2491
  },                                                                                                                   // 2491
  conflict: {                                                                                                          // 2492
    id: 409,                                                                                                           // 2492
    title: "Conflict"                                                                                                  // 2492
  },                                                                                                                   // 2492
  gone: {                                                                                                              // 2493
    id: 410,                                                                                                           // 2493
    title: "Gone"                                                                                                      // 2493
  },                                                                                                                   // 2493
  requestEntityTooLarge: {                                                                                             // 2494
    id: 413,                                                                                                           // 2494
    title: "Request Entity Too Large"                                                                                  // 2494
  },                                                                                                                   // 2494
  requestUriTooLong: {                                                                                                 // 2495
    id: 414,                                                                                                           // 2495
    title: "Request-URI Too Long"                                                                                      // 2495
  },                                                                                                                   // 2495
  unsupportedMediaType: {                                                                                              // 2496
    id: 415,                                                                                                           // 2496
    title: "Unsupported Media Type"                                                                                    // 2496
  },                                                                                                                   // 2496
  imATeapot: {                                                                                                         // 2497
    id: 418,                                                                                                           // 2497
    title: "I'm a teapot"                                                                                              // 2497
  },                                                                                                                   // 2497
  unprocessableEntity: {                                                                                               // 2498
    id: 422,                                                                                                           // 2498
    title: "Unprocessable Entity"                                                                                      // 2498
  }                                                                                                                    // 2498
};                                                                                                                     // 2486
                                                                                                                       //
ResponseStream = function () {                                                                                         // 2501
  // Note: This class is used in pre-meteor.js as well as in this file.                                                // 2502
  function ResponseStream(response, httpCode, httpStatus, upstreamHandle, idempotentCleanup) {                         // 2504
    var _this13 = this;                                                                                                // 2504
                                                                                                                       //
    (0, _classCallCheck3.default)(this, ResponseStream);                                                               // 2504
    // This is stupidly complicated, because:                                                                          // 2505
    // - New versions of sandstorm-http-bridge (and other well-behaved apps) wait on write()                           // 2506
    //   completion for flow control, so we want write() to complete only when there is space                          // 2507
    //   available for more writes.                                                                                    // 2508
    // - Old versions of sandstorm-http-bridge would make a zillion simultaneous calls to write()                      // 2509
    //   with all of the data they wanted to send without waiting for previous calls to complete.                      // 2510
    //   While in-flight, these writes can consume a lot of RAM at multiple points in the system.                      // 2511
    // - Old versions of sandstorm-http-bridge would print any thrown error to stderr and then                         // 2512
    //   keep going. So if canceling a download early on causes all future write()s to fail, this                      // 2513
    //   can lead to very excessive log spam.                                                                          // 2514
    // - If a write() fails to complete (it returns a promise that never resolves), then later                         // 2515
    //   upon garbage collection the RPC will actually throw an exception about a PromiseFulfiller                     // 2516
    //   never having been fulfilled. Hence, if we simply stop responding to write()s after some                       // 2517
    //   point, we again get log spam.                                                                                 // 2518
    // - If the connection closes during a write, Node never calls the completion callback at all.                     // 2519
    //   We have to listen for the "close" event.                                                                      // 2520
    this.response = response;                                                                                          // 2522
    this.httpCode = httpCode;                                                                                          // 2523
    this.httpStatus = httpStatus;                                                                                      // 2524
    this.upstreamHandle = upstreamHandle;                                                                              // 2525
    this.idempotentCleanup = idempotentCleanup;                                                                        // 2526
    this.started = false;                                                                                              // 2528
    this.ended = false; // Invariant: if `ended` is true, then `started` is also true.                                 // 2529
                                                                                                                       //
    this.waiting = [];                                                                                                 // 2530
    response.on("drain", function () {                                                                                 // 2532
      _this13.waiting.forEach(function (f) {                                                                           // 2533
        return f();                                                                                                    // 2533
      });                                                                                                              // 2533
                                                                                                                       //
      _this13.waiting = [];                                                                                            // 2534
    });                                                                                                                // 2535
    response.on("close", function () {                                                                                 // 2537
      _this13.abort();                                                                                                 // 2537
    });                                                                                                                // 2537
  }                                                                                                                    // 2538
                                                                                                                       //
  ResponseStream.prototype.abort = function () {                                                                       // 2501
    function abort() {                                                                                                 // 2501
      this.aborted = true; // Resolve all outstanding writes. This is OK even if they didn't actually go through because
      // an incomplete download is the client's problem, not the server's. The important thing is                      // 2544
      // to cancel the stream if it is still being generated; to that end, the next new write() call                   // 2545
      // after this point will throw.                                                                                  // 2546
                                                                                                                       //
      this.waiting.forEach(function (f) {                                                                              // 2547
        return f();                                                                                                    // 2547
      });                                                                                                              // 2547
      this.waiting = [];                                                                                               // 2548
      this.close();                                                                                                    // 2550
    }                                                                                                                  // 2551
                                                                                                                       //
    return abort;                                                                                                      // 2501
  }();                                                                                                                 // 2501
                                                                                                                       //
  ResponseStream.prototype.expectSize = function () {                                                                  // 2501
    function expectSize(size) {                                                                                        // 2501
      if (!this.started) {                                                                                             // 2554
        this.started = true;                                                                                           // 2555
        this.response.writeHead(this.httpCode, this.httpStatus, {                                                      // 2556
          "Content-Length": size                                                                                       // 2556
        });                                                                                                            // 2556
      }                                                                                                                // 2557
    }                                                                                                                  // 2558
                                                                                                                       //
    return expectSize;                                                                                                 // 2501
  }();                                                                                                                 // 2501
                                                                                                                       //
  ResponseStream.prototype.write = function () {                                                                       // 2501
    function write(data) {                                                                                             // 2501
      var _this14 = this;                                                                                              // 2560
                                                                                                                       //
      if (!this.started) {                                                                                             // 2561
        this.started = true;                                                                                           // 2562
        this.response.writeHead(this.httpCode, this.httpStatus);                                                       // 2563
      }                                                                                                                // 2564
                                                                                                                       //
      if (this.aborted) {                                                                                              // 2566
        // Connection has been aborted.                                                                                // 2567
        if (this.alreadyThrew) {                                                                                       // 2568
          // We already threw an error from a previous write(). Do not throw more errors, because                      // 2569
          // older versions of sandstorm-http-bridge do not stop on the first error and will print                     // 2570
          // every error to the log, wasting space. Newer versions stop on the first failure.                          // 2571
          return;                                                                                                      // 2572
        } else {                                                                                                       // 2573
          this.alreadyThrew = true;                                                                                    // 2574
          throw new Error("client disconnected");                                                                      // 2575
        }                                                                                                              // 2576
      } else {                                                                                                         // 2577
        if (this.response.write(data)) {                                                                               // 2578
          // All written.                                                                                              // 2579
          return;                                                                                                      // 2580
        } else {                                                                                                       // 2581
          // Write buffer is full. Don't complete until it has more space.                                             // 2582
          return new Promise(function (resolve, reject) {                                                              // 2583
            _this14.waiting.push(resolve);                                                                             // 2584
          });                                                                                                          // 2585
        }                                                                                                              // 2586
      }                                                                                                                // 2587
    }                                                                                                                  // 2588
                                                                                                                       //
    return write;                                                                                                      // 2501
  }();                                                                                                                 // 2501
                                                                                                                       //
  ResponseStream.prototype.done = function () {                                                                        // 2501
    function done() {                                                                                                  // 2501
      if (!this.started) {                                                                                             // 2591
        this.started = true;                                                                                           // 2592
        this.response.writeHead(this.httpCode, this.httpStatus, {                                                      // 2593
          "Content-Length": 0                                                                                          // 2593
        });                                                                                                            // 2593
      }                                                                                                                // 2594
                                                                                                                       //
      if (!this.ended) {                                                                                               // 2596
        this.ended = true;                                                                                             // 2597
        this.response.end();                                                                                           // 2598
      }                                                                                                                // 2599
                                                                                                                       //
      this.close();                                                                                                    // 2601
    }                                                                                                                  // 2602
                                                                                                                       //
    return done;                                                                                                       // 2501
  }();                                                                                                                 // 2501
                                                                                                                       //
  ResponseStream.prototype.close = function () {                                                                       // 2501
    function close() {                                                                                                 // 2501
      if (this.upstreamHandle) {                                                                                       // 2605
        this.upstreamHandle.close();                                                                                   // 2606
      }                                                                                                                // 2607
                                                                                                                       //
      if (this.idempotentCleanup) {                                                                                    // 2609
        this.idempotentCleanup();                                                                                      // 2610
      } // If `this.done()` has not been called, we should report an error to the client. Unfortunately,               // 2611
      // the only case in which we can actually report the error is when we haven't done any writes                    // 2614
      // yet. So we handle that case and ignore the error otherwise. Note that `!this.started`                         // 2615
      // implies `!this.ended`.                                                                                        // 2616
                                                                                                                       //
                                                                                                                       //
      if (!this.started) {                                                                                             // 2617
        this.started = true;                                                                                           // 2618
        this.response.writeHead(500, "done() never called on response stream", {                                       // 2619
          "Content-Type": "text/plain"                                                                                 // 2620
        });                                                                                                            // 2620
      }                                                                                                                // 2621
                                                                                                                       //
      if (!this.ended) {                                                                                               // 2623
        this.ended = true;                                                                                             // 2624
        this.response.end();                                                                                           // 2625
      }                                                                                                                // 2626
    }                                                                                                                  // 2627
                                                                                                                       //
    return close;                                                                                                      // 2501
  }();                                                                                                                 // 2501
                                                                                                                       //
  ResponseStream.prototype.sendingDirectResponse = function () {                                                       // 2501
    function sendingDirectResponse() {                                                                                 // 2501
      // For internal front-end use: Indicates that the caller is going to fill in the response                        // 2630
      // instead. If a response has already been started, this throws an exception.                                    // 2631
      //                                                                                                               // 2632
      // This is used in particular in pre-meteor.js, when dealing with static web publishing.                         // 2633
      if (this.started) {                                                                                              // 2635
        throw new Error("can't sendDirectResponse() because response already started");                                // 2636
        this.response.end();                                                                                           // 2637
      } else if (this.ended) {                                                                                         // 2638
        throw new Error("can't sendDirectResponse() because response already sent");                                   // 2639
      } else {                                                                                                         // 2640
        this.started = true;                                                                                           // 2641
        this.ended = true;                                                                                             // 2642
      }                                                                                                                // 2643
    }                                                                                                                  // 2644
                                                                                                                       //
    return sendingDirectResponse;                                                                                      // 2501
  }();                                                                                                                 // 2501
                                                                                                                       //
  return ResponseStream;                                                                                               // 2501
}(); // TODO(cleanup): Node 0.12 has a `gunzipSync` but 0.10 (which Meteor still uses) does not.                       // 2501
                                                                                                                       //
                                                                                                                       //
var Zlib = Npm.require("zlib");                                                                                        // 2648
                                                                                                                       //
var gunzipSync = Meteor.wrapAsync(Zlib.gunzip, Zlib); // -----------------------------------------------------------------------------
// WebSocket handling                                                                                                  // 2652
                                                                                                                       //
WebSocketReceiver = function () {                                                                                      // 2654
  function WebSocketReceiver(socket) {                                                                                 // 2655
    (0, _classCallCheck3.default)(this, WebSocketReceiver);                                                            // 2655
    this.socket = socket;                                                                                              // 2656
    this.queue = [];                                                                                                   // 2657
  }                                                                                                                    // 2658
                                                                                                                       //
  WebSocketReceiver.prototype.go = function () {                                                                       // 2654
    function go() {                                                                                                    // 2654
      for (var i in meteorBabelHelpers.sanitizeForInObject(this.queue)) {                                              // 2661
        this.socket.write(this.queue[i]);                                                                              // 2662
      }                                                                                                                // 2663
                                                                                                                       //
      this.queue = null;                                                                                               // 2665
    }                                                                                                                  // 2666
                                                                                                                       //
    return go;                                                                                                         // 2654
  }();                                                                                                                 // 2654
                                                                                                                       //
  WebSocketReceiver.prototype.sendBytes = function () {                                                                // 2654
    function sendBytes(message) {                                                                                      // 2654
      // TODO(someday):  Flow control of some sort?                                                                    // 2669
      if (this.queue === null) {                                                                                       // 2670
        this.socket.write(message);                                                                                    // 2671
      } else {                                                                                                         // 2672
        this.queue.push(message);                                                                                      // 2673
      }                                                                                                                // 2674
    }                                                                                                                  // 2675
                                                                                                                       //
    return sendBytes;                                                                                                  // 2654
  }();                                                                                                                 // 2654
                                                                                                                       //
  WebSocketReceiver.prototype.close = function () {                                                                    // 2654
    function close() {                                                                                                 // 2654
      this.socket.end();                                                                                               // 2678
    }                                                                                                                  // 2679
                                                                                                                       //
    return close;                                                                                                      // 2654
  }();                                                                                                                 // 2654
                                                                                                                       //
  return WebSocketReceiver;                                                                                            // 2654
}();                                                                                                                   // 2654
                                                                                                                       //
function pumpWebSocket(socket, rpcStream, destructor) {                                                                // 2682
  var writesInFlight = 0;                                                                                              // 2683
  var readingPaused = false;                                                                                           // 2684
  socket.on("data", function (chunk) {                                                                                 // 2685
    rpcStream.sendBytes(chunk).then(function () {                                                                      // 2686
      writesInFlight--;                                                                                                // 2687
                                                                                                                       //
      if (readingPaused) {                                                                                             // 2688
        socket.resume();                                                                                               // 2689
        readingPaused = false;                                                                                         // 2690
      }                                                                                                                // 2691
    }).catch(function (err) {                                                                                          // 2692
      if (err.kjType !== "disconnected") {                                                                             // 2693
        console.error("WebSocket sendBytes failed: " + err.stack);                                                     // 2694
      }                                                                                                                // 2695
                                                                                                                       //
      socket.destroy();                                                                                                // 2697
    });                                                                                                                // 2698
                                                                                                                       //
    if (writesInFlight++ > 16) {                                                                                       // 2700
      socket.pause();                                                                                                  // 2701
      readingPaused = true;                                                                                            // 2702
    }                                                                                                                  // 2703
  });                                                                                                                  // 2704
  socket.on("end", function (chunk) {                                                                                  // 2706
    rpcStream.close();                                                                                                 // 2707
  });                                                                                                                  // 2708
  socket.on("close", function () {                                                                                     // 2710
    destructor();                                                                                                      // 2711
  });                                                                                                                  // 2712
}                                                                                                                      // 2713
                                                                                                                       //
; // =======================================================================================                           // 2713
// Debug log access                                                                                                    // 2716
                                                                                                                       //
Meteor.publish("grainLog", function (grainId) {                                                                        // 2718
  check(grainId, String);                                                                                              // 2719
  var id = 0;                                                                                                          // 2720
  var grain = Grains.findOne(grainId);                                                                                 // 2721
                                                                                                                       //
  if (!grain || !this.userId || grain.userId !== this.userId) {                                                        // 2722
    this.added("grainLog", id++, {                                                                                     // 2723
      text: "Only the grain owner can view the debug log."                                                             // 2723
    });                                                                                                                // 2723
    this.ready();                                                                                                      // 2724
    return;                                                                                                            // 2725
  }                                                                                                                    // 2726
                                                                                                                       //
  var connected = false;                                                                                               // 2728
                                                                                                                       //
  var _this = this;                                                                                                    // 2729
                                                                                                                       //
  var receiver = {                                                                                                     // 2731
    write: function (data) {                                                                                           // 2732
      connected = true;                                                                                                // 2733
                                                                                                                       //
      _this.added("grainLog", id++, {                                                                                  // 2734
        text: data.toString("utf8")                                                                                    // 2734
      });                                                                                                              // 2734
    },                                                                                                                 // 2735
    close: function () {                                                                                               // 2737
      if (connected) {                                                                                                 // 2738
        _this.added("grainLog", id++, {                                                                                // 2739
          text: "*** lost connection to grain (probably because it shut down) ***"                                     // 2740
        });                                                                                                            // 2739
      }                                                                                                                // 2742
    }                                                                                                                  // 2743
  };                                                                                                                   // 2731
                                                                                                                       //
  try {                                                                                                                // 2746
    var handle = waitPromise(globalBackend.useGrain(grainId, function (supervisor) {                                   // 2747
      return supervisor.watchLog(8192, receiver);                                                                      // 2748
    })).handle;                                                                                                        // 2749
    connected = true;                                                                                                  // 2750
    this.onStop(function () {                                                                                          // 2751
      handle.close();                                                                                                  // 2752
    });                                                                                                                // 2753
  } catch (err) {                                                                                                      // 2754
    if (!connected) {                                                                                                  // 2755
      this.added("grainLog", id++, {                                                                                   // 2756
        text: "*** couldn't connect to grain (" + err + ") ***"                                                        // 2757
      });                                                                                                              // 2756
    }                                                                                                                  // 2759
  } // Notify ready.                                                                                                   // 2760
                                                                                                                       //
                                                                                                                       //
  this.ready();                                                                                                        // 2763
});                                                                                                                    // 2764
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"sandcats.js":["node-forge","/imports/server/constants.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/sandcats.js                                                                                                  //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var pki = void 0,                                                                                                      // 1
    asn1 = void 0;                                                                                                     // 1
module.import("node-forge", {                                                                                          // 1
  "pki": function (v) {                                                                                                // 1
    pki = v;                                                                                                           // 1
  },                                                                                                                   // 1
  "asn1": function (v) {                                                                                               // 1
    asn1 = v;                                                                                                          // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var SANDSTORM_ALTHOME = void 0;                                                                                        // 1
module.import("/imports/server/constants.js", {                                                                        // 1
  "SANDSTORM_ALTHOME": function (v) {                                                                                  // 1
    SANDSTORM_ALTHOME = v;                                                                                             // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Sandcats = {};                                                                                                         // 17
                                                                                                                       //
var querystring = Npm.require("querystring");                                                                          // 20
                                                                                                                       //
var https = Npm.require("https");                                                                                      // 21
                                                                                                                       //
var fs = Npm.require("fs");                                                                                            // 22
                                                                                                                       //
var dgram = Npm.require("dgram");                                                                                      // 23
                                                                                                                       //
var Url = Npm.require("url");                                                                                          // 24
                                                                                                                       //
var SANDCATS_HOSTNAME = Meteor.settings && Meteor.settings.public && Meteor.settings.public.sandcatsHostname;          // 28
var SANDCATS_VARDIR = (SANDSTORM_ALTHOME || "") + "/var/sandcats"; // Figure out what IP address to send Sandcats requests from. For machines with multiple IPs, it
// is important to use the IP to which we're binding. However, some people set BIND_IP to 127.0.0.1                    // 33
// and put sniproxy in front of Sandstorm. In those cases, it won't work to send from BIND_IP;                         // 34
// we'll have to let the system choose.                                                                                // 35
                                                                                                                       //
var BIND_IP = process.env.BIND_IP && process.env.BIND_IP.startsWith("127.") ? null : process.env.BIND_IP;              // 36
var ROOT_URL = Url.parse(process.env.ROOT_URL);                                                                        // 39
var HOSTNAME = ROOT_URL.hostname;                                                                                      // 40
var SANDCATS_NAME = void 0; // Look at `startup` below to see where this is set                                        // 41
                                                                                                                       //
var updateSandcatsIp = function () {                                                                                   // 43
  var responseCallback = function (res) {                                                                              // 44
    if (res.statusCode === 200) {                                                                                      // 45
      console.log("Successfully updated sandcats IP");                                                                 // 46
    } else {                                                                                                           // 47
      console.error("Failed to update sandcats IP:", res.headers);                                                     // 48
    }                                                                                                                  // 49
  };                                                                                                                   // 50
                                                                                                                       //
  var errorCallback = function (err) {                                                                                 // 52
    console.error("Couldn't send update sandcats hostname", err);                                                      // 53
  };                                                                                                                   // 54
                                                                                                                       //
  performSandcatsRequest("/update", SANDCATS_HOSTNAME, {                                                               // 56
    rawHostname: SANDCATS_NAME                                                                                         // 56
  }, errorCallback, responseCallback);                                                                                 // 56
};                                                                                                                     // 58
                                                                                                                       //
var pingUdp = function () {                                                                                            // 60
  var socket = dgram.createSocket("udp4");                                                                             // 61
  var secret = Random.secret(16);                                                                                      // 62
  message = new Buffer(SANDCATS_NAME + " " + secret);                                                                  // 64
  socket.on("message", function (buf) {                                                                                // 65
    if (buf.toString() === secret) {                                                                                   // 66
      updateSandcatsIp();                                                                                              // 67
    } else {                                                                                                           // 68
      console.error("Received unexpected response in UDP sandcats ping:", buf.toString());                             // 69
    }                                                                                                                  // 70
  });                                                                                                                  // 71
  socket.on("error", function (err) {                                                                                  // 73
    throw err;                                                                                                         // 74
  });                                                                                                                  // 75
                                                                                                                       //
  var callback = function () {                                                                                         // 77
    socket.send(message, 0, message.length, 8080, SANDCATS_HOSTNAME, function (err) {                                  // 78
      if (err) {                                                                                                       // 79
        console.error("Couldn't send UDP sandcats ping", err);                                                         // 80
      }                                                                                                                // 81
    });                                                                                                                // 82
    setTimeout(function () {                                                                                           // 84
      socket.close();                                                                                                  // 85
    }, 10 * 1000);                                                                                                     // 86
  };                                                                                                                   // 87
                                                                                                                       //
  if (BIND_IP) {                                                                                                       // 89
    socket.bind({                                                                                                      // 90
      address: BIND_IP                                                                                                 // 90
    }, callback);                                                                                                      // 90
  } else {                                                                                                             // 91
    callback();                                                                                                        // 92
  }                                                                                                                    // 93
};                                                                                                                     // 94
                                                                                                                       //
var performSandcatsRequest = function (path, hostname, postData, errorCallback, responseCallback) {                    // 96
  var options = {                                                                                                      // 97
    hostname: hostname,                                                                                                // 98
    path: path,                                                                                                        // 99
    method: "POST",                                                                                                    // 100
    agent: false,                                                                                                      // 101
    key: fs.readFileSync(SANDCATS_VARDIR + "/id_rsa"),                                                                 // 102
    cert: fs.readFileSync(SANDCATS_VARDIR + "/id_rsa.pub"),                                                            // 103
    headers: {                                                                                                         // 104
      "X-Sand": "cats",                                                                                                // 105
      "Content-Type": "application/x-www-form-urlencoded"                                                              // 106
    }                                                                                                                  // 104
  };                                                                                                                   // 97
                                                                                                                       //
  if (BIND_IP) {                                                                                                       // 110
    options.localAddress = BIND_IP;                                                                                    // 111
  }                                                                                                                    // 112
                                                                                                                       //
  if (postData.certificateSigningRequest) {                                                                            // 114
    console.log("Submitting certificate request for host", postData.rawHostname, "where the request has length", postData.certificateSigningRequest.length);
  }                                                                                                                    // 118
                                                                                                                       //
  var newPostData = querystring.stringify(postData);                                                                   // 120
                                                                                                                       //
  if (SANDCATS_HOSTNAME === "sandcats-dev.sandstorm.io" && !process.env.NODE_TLS_REJECT_UNAUTHORIZED) {                // 121
    console.log("You are using the Sandcats dev server but have full HTTPS " + "certificate checking enabled. Probably you will get an HTTPS " + "error. Proceeding anyway.");
  }                                                                                                                    // 126
                                                                                                                       //
  var req = https.request(options, responseCallback);                                                                  // 128
  req.write(newPostData);                                                                                              // 129
  req.end();                                                                                                           // 130
                                                                                                                       //
  if (errorCallback) {                                                                                                 // 132
    req.on("error", errorCallback);                                                                                    // 133
  }                                                                                                                    // 134
                                                                                                                       //
  return req;                                                                                                          // 136
};                                                                                                                     // 137
                                                                                                                       //
var generateKeyAndCsr = function (commonName) {                                                                        // 139
  check(commonName, String); // Generate key pair. Using Meteor.wrapAsync because forge supports                       // 140
  // a synchronous as well as an asynchronous API, and the synchronous                                                 // 143
  // one blocks for a while.                                                                                           // 144
                                                                                                                       //
  var wrappedGenerateKeyPair = Meteor.wrapAsync(pki.rsa.generateKeyPair); // I could pick an `e`[xponent] value for the resulting RSA key, but
  // I will refrain.                                                                                                   // 148
                                                                                                                       //
  var keys = wrappedGenerateKeyPair({                                                                                  // 149
    bits: 2048                                                                                                         // 149
  }); // Create a certificate request (CSR).                                                                           // 149
                                                                                                                       //
  var csr = pki.createCertificationRequest();                                                                          // 152
  csr.publicKey = keys.publicKey;                                                                                      // 153
  csr.setSubject([{                                                                                                    // 154
    name: "commonName",                                                                                                // 156
    value: commonName,                                                                                                 // 157
    valueTagClass: asn1.Type.UTF8                                                                                      // 158
  }]);                                                                                                                 // 155
  csr.sign(keys.privateKey);                                                                                           // 164
  console.log("generateKeyAndCsr created new key & certificate request for", commonName);                              // 165
  return {                                                                                                             // 166
    privateKeyAsPem: pki.privateKeyToPem(keys.privateKey),                                                             // 167
    csrAsPem: pki.certificationRequestToPem(csr)                                                                       // 168
  };                                                                                                                   // 166
};                                                                                                                     // 170
                                                                                                                       //
Sandcats.storeNewKeyAndCsr = function (hostname, basePath) {                                                           // 172
  // We use the current JS time (like UNIX timestamp but in                                                            // 173
  // milliseconds) as the key number. Note that the keyNumber is                                                       // 174
  // intended to be an opaque identifier; the only important thing is                                                  // 175
  // that it increases numerically over time. We use the current time                                                  // 176
  // just as a simplistic way to pick a filename that probably no one                                                  // 177
  // else has created yet.                                                                                             // 178
  var keyNumber = new Date().getTime();                                                                                // 179
  var keyFilename = basePath + "/" + keyNumber;                                                                        // 180
  var csrFilename = keyFilename + ".csr";                                                                              // 181
  var responseFilename = keyFilename + ".response-json";                                                               // 182
  var withWildcard = "*." + hostname;                                                                                  // 183
  var keyAndCsr = generateKeyAndCsr(withWildcard);                                                                     // 184
  fs.writeFileSync(keyFilename, keyAndCsr.privateKeyAsPem, {                                                           // 185
    mode: 0400                                                                                                         // 186
  });                                                                                                                  // 186
  fs.writeFileSync(csrFilename, keyAndCsr.csrAsPem);                                                                   // 187
  console.log("storeNewKeyAndCsr successfully saved key and certificate request to", keyFilename, "and", csrFilename, "respectively of length", keyAndCsr.privateKeyAsPem.length, "and", keyAndCsr.csrAsPem.length);
  return {                                                                                                             // 192
    csrFilename: csrFilename,                                                                                          // 193
    keyFilename: keyFilename,                                                                                          // 194
    responseFilename: responseFilename                                                                                 // 195
  };                                                                                                                   // 192
};                                                                                                                     // 197
                                                                                                                       //
Sandcats.renewHttpsCertificateIfNeeded = function () {                                                                 // 199
  var renewHttpsCertificate = function () {                                                                            // 200
    var hostname = Url.parse(process.env.ROOT_URL).hostname;                                                           // 201
    var basePath = "/var/sandcats/https/" + hostname;                                                                  // 202
    var filenames = Sandcats.storeNewKeyAndCsr(hostname, basePath);                                                    // 203
                                                                                                                       //
    var errorCallback = function (err) {                                                                               // 205
      console.error("Error while renewing HTTPS certificate (will continue to retry)", err);                           // 206
    };                                                                                                                 // 207
                                                                                                                       //
    var responseCallback = function (res) {                                                                            // 209
      if (res.statusCode == 200) {                                                                                     // 210
        // Save the response, chunk by chunk, then store it on disk                                                    // 211
        // for later use.                                                                                              // 212
        var responseBody = "";                                                                                         // 213
        res.on("data", function (chunk) {                                                                              // 214
          responseBody += chunk;                                                                                       // 215
        });                                                                                                            // 216
        res.on("end", function () {                                                                                    // 218
          // For sanity, make sure it parses as JSON, since we're                                                      // 219
          // going to need it to do that.                                                                              // 220
          try {                                                                                                        // 221
            JSON.parse(responseBody);                                                                                  // 222
          } catch (e) {                                                                                                // 223
            console.error("JSON parse error receiving new HTTPS certificate. Discarding:", responseBody, "due to exception:", e); // Overwrite the responseBody with the empty JSON object
            // and continue with the process of saving it to disk.                                                     // 229
                                                                                                                       //
            responseBody = "{}";                                                                                       // 230
          }                                                                                                            // 231
                                                                                                                       //
          try {                                                                                                        // 233
            fs.writeFileSync(filenames.responseFilename, responseBody, "utf-8");                                       // 234
          } catch (err) {                                                                                              // 235
            return console.error("Failure while saveing new HTTPS certificate to", filenames.responseFilename, "will continue to retry. Exception was:", err);
          } // Tell Sandcats that now is a good time to update its info                                                // 240
          // about which keys are available.                                                                           // 243
                                                                                                                       //
                                                                                                                       //
          if (!global.sandcats) {                                                                                      // 244
            console.error("When getting new certificate, could not find callback to request re-keying! Certs will probably become invalid soon.");
          } else {                                                                                                     // 246
            // Call the sandcats rekeying function.                                                                    // 247
            global.sandcats.rekey(); // That's that.                                                                   // 248
                                                                                                                       //
            console.log("Successfully renewed HTTPS certificate into", filenames.responseFilename);                    // 250
          }                                                                                                            // 252
        });                                                                                                            // 253
      } else {                                                                                                         // 254
        console.log("Received HTTP error while renewing certificate (will keep retrying)", res.statusCode);            // 255
        res.on("data", function (chunk) {                                                                              // 257
          console.log("Error response contained information", chunk.toString("utf-8"));                                // 258
        });                                                                                                            // 259
      }                                                                                                                // 260
    };                                                                                                                 // 261
                                                                                                                       //
    var wrappedReadFile = Meteor.wrapAsync(fs.readFile);                                                               // 263
    performSandcatsRequest("/getcertificate", SANDCATS_HOSTNAME, {                                                     // 265
      rawHostname: SANDCATS_NAME,                                                                                      // 266
      certificateSigningRequest: wrappedReadFile(filenames.csrFilename, "utf-8")                                       // 267
    }, errorCallback, responseCallback);                                                                               // 265
  };                                                                                                                   // 269
                                                                                                                       //
  if (global.sandcats.shouldGetAnotherCertificate()) {                                                                 // 271
    console.log("renewHttpsCertificateIfNeeded: Happily choosing to renew certificate.");                              // 272
    return renewHttpsCertificate();                                                                                    // 273
  }                                                                                                                    // 274
};                                                                                                                     // 275
                                                                                                                       //
Sandcats.initializeSandcats = function () {                                                                            // 277
  var i = HOSTNAME.lastIndexOf(SANDCATS_HOSTNAME);                                                                     // 278
                                                                                                                       //
  if (i < 0) {                                                                                                         // 279
    console.error("SANDCATS_BASE_DOMAIN is configured but your HOSTNAME doesn't appear to contain it:", SANDCATS_HOSTNAME, HOSTNAME);
  } else {                                                                                                             // 282
    var oneMinute = 60 * 1000;                                                                                         // 283
    var oneHour = 60 * oneMinute;                                                                                      // 284
    var randomIntervalZeroToOneHour = Math.random() * oneHour; // All Sandcats installs need dyndns updating.          // 285
                                                                                                                       //
    SANDCATS_NAME = HOSTNAME.slice(0, i - 1);                                                                          // 287
    Meteor.setInterval(pingUdp, oneMinute); // If process.env.HTTPS_PORT is set, we need to auto-refresh our HTTPS certificate.
                                                                                                                       //
    if (process.env.HTTPS_PORT) {                                                                                      // 290
      // Always do a HTTPS certificate update check on Sandstorm start.                                                // 291
      Sandcats.renewHttpsCertificateIfNeeded(); // After that's done, schedule it for every approx 1-2 hours in        // 292
      // the future.                                                                                                   // 295
                                                                                                                       //
      Meteor.setInterval(Sandcats.renewHttpsCertificateIfNeeded, oneHour + randomIntervalZeroToOneHour);               // 296
    }                                                                                                                  // 298
  }                                                                                                                    // 299
};                                                                                                                     // 300
                                                                                                                       //
if (SANDCATS_HOSTNAME) {                                                                                               // 302
  Meteor.startup(Sandcats.initializeSandcats);                                                                         // 303
}                                                                                                                      // 304
                                                                                                                       //
;                                                                                                                      // 304
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"shell-server.js":["/imports/server/async-helpers.js","/imports/server/constants.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/shell-server.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var inMeteor = void 0;                                                                                                 // 1
module.import("/imports/server/async-helpers.js", {                                                                    // 1
  "inMeteor": function (v) {                                                                                           // 1
    inMeteor = v;                                                                                                      // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var staticAssetHost = void 0;                                                                                          // 1
module.import("/imports/server/constants.js", {                                                                        // 1
  "staticAssetHost": function (v) {                                                                                    // 1
    staticAssetHost = v;                                                                                               // 1
  }                                                                                                                    // 1
}, 1);                                                                                                                 // 1
BrowserPolicy.framing.disallow(); // Disallow framing of the UI.                                                       // 38
                                                                                                                       //
Meteor.startup(function () {                                                                                           // 39
  var frameSetter = function () {                                                                                      // 40
    BrowserPolicy.content.disallowFrame(); // This clears all the old rules                                            // 41
                                                                                                                       //
    BrowserPolicy.content.allowFrameOrigin(getWildcardOrigin());                                                       // 42
    var billingPromptUrl = globalDb.getBillingPromptUrl();                                                             // 43
                                                                                                                       //
    if (billingPromptUrl) {                                                                                            // 44
      BrowserPolicy.content.allowFrameOrigin(billingPromptUrl);                                                        // 45
    }                                                                                                                  // 46
  };                                                                                                                   // 47
                                                                                                                       //
  frameSetter(); // Call once on startup                                                                               // 49
                                                                                                                       //
  globalDb.collections.settings.find({                                                                                 // 50
    _id: "billingPromptUrl"                                                                                            // 50
  }).observe({                                                                                                         // 50
    added: frameSetter,                                                                                                // 51
    changed: frameSetter,                                                                                              // 52
    removed: frameSetter                                                                                               // 53
  });                                                                                                                  // 50
}); // Allow anything to be loaded from the static asset host.                                                         // 55
                                                                                                                       //
BrowserPolicy.content.allowImageOrigin(staticAssetHost);                                                               // 59
BrowserPolicy.content.allowScriptOrigin(staticAssetHost);                                                              // 60
BrowserPolicy.content.allowFontOrigin(staticAssetHost);                                                                // 61
BrowserPolicy.content.allowConnectOrigin(staticAssetHost);                                                             // 62
BrowserPolicy.content.allowConnectOrigin("wss:");                                                                      // 63
BrowserPolicy.content.allowConnectOrigin("ws:");                                                                       // 64
Meteor.publish("grainsMenu", function () {                                                                             // 66
  if (this.userId) {                                                                                                   // 67
    if (globalDb.isQuotaEnabled()) {                                                                                   // 68
      // Hack: Fire off an asynchronous update to the user's storage usage whenever they open the                      // 69
      //   front page.                                                                                                 // 70
      // TODO(someday): Implement the ability to reactively subscribe to storage usage from the                        // 71
      //   back-end?                                                                                                   // 72
      var userId = this.userId;                                                                                        // 73
      globalBackend.cap().getUserStorageUsage(userId).then(function (results) {                                        // 74
        inMeteor(function () {                                                                                         // 75
          Meteor.users.update(userId, {                                                                                // 76
            $set: {                                                                                                    // 76
              storageUsage: parseInt(results.size)                                                                     // 76
            }                                                                                                          // 76
          });                                                                                                          // 76
        });                                                                                                            // 77
      }).catch(function (err) {                                                                                        // 78
        if (err.kjType === "unimplemented") {                                                                          // 79
          // Compute based on sum of grain sizes instead.                                                              // 80
          var total = 0;                                                                                               // 81
          Grains.find({                                                                                                // 82
            userId: userId                                                                                             // 82
          }, {                                                                                                         // 82
            fields: {                                                                                                  // 82
              size: 1                                                                                                  // 82
            }                                                                                                          // 82
          }).forEach(function (grain) {                                                                                // 82
            return total += grain.size || 0;                                                                           // 83
          });                                                                                                          // 83
          Meteor.users.update(userId, {                                                                                // 84
            $set: {                                                                                                    // 84
              storageUsage: total                                                                                      // 84
            }                                                                                                          // 84
          });                                                                                                          // 84
        } else {                                                                                                       // 85
          console.error(err.stack);                                                                                    // 86
        }                                                                                                              // 87
      });                                                                                                              // 88
    }                                                                                                                  // 89
                                                                                                                       //
    var identityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));                                   // 91
    return [UserActions.find({                                                                                         // 92
      userId: this.userId                                                                                              // 93
    }), Grains.find({                                                                                                  // 93
      userId: this.userId                                                                                              // 94
    }), ApiTokens.find({                                                                                               // 94
      "owner.user.identityId": {                                                                                       // 95
        $in: identityIds                                                                                               // 95
      }                                                                                                                // 95
    })];                                                                                                               // 95
  } else {                                                                                                             // 97
    return [];                                                                                                         // 98
  }                                                                                                                    // 99
});                                                                                                                    // 100
Meteor.publish("sessions", function (sessionId) {                                                                      // 102
  // sessionId itself should be secret enough, but they are also not meant to be shared, so as                         // 103
  // a backup we only publish the session to its owner. Note that `userId` can be null if the                          // 104
  // user is not logged in or is using incognito mode.                                                                 // 105
  check(sessionId, String); // We exclude powerboxRequest because the client already has the descriptor list in packed
  // format, and the parsed format can be kind of large.                                                               // 109
                                                                                                                       //
  return Sessions.find({                                                                                               // 110
    _id: sessionId,                                                                                                    // 110
    $or: [{                                                                                                            // 110
      userId: this.userId                                                                                              // 110
    }, {                                                                                                               // 110
      userId: null                                                                                                     // 110
    }]                                                                                                                 // 110
  }, {                                                                                                                 // 110
    fields: {                                                                                                          // 111
      powerboxRequest: 0                                                                                               // 111
    }                                                                                                                  // 111
  });                                                                                                                  // 111
});                                                                                                                    // 112
Meteor.publish("devPackages", function () {                                                                            // 114
  return DevPackages.find();                                                                                           // 115
});                                                                                                                    // 116
Meteor.publish("hasUsers", function () {                                                                               // 118
  var _this = this;                                                                                                    // 118
                                                                                                                       //
  // Publish pseudo-collection which tells the client if there are any users at all.                                   // 119
  //                                                                                                                   // 120
  // TODO(cleanup):  This seems overcomplicated.  Does Meteor have a better way?                                       // 121
  var cursor = Meteor.users.find();                                                                                    // 122
                                                                                                                       //
  if (cursor.count() > 0) {                                                                                            // 123
    this.added("hasUsers", "hasUsers", {                                                                               // 124
      hasUsers: true                                                                                                   // 124
    });                                                                                                                // 124
  } else {                                                                                                             // 125
    var handle = cursor.observeChanges({                                                                               // 126
      added: function (id) {                                                                                           // 127
        _this.added("hasUsers", "hasUsers", {                                                                          // 128
          hasUsers: true                                                                                               // 128
        });                                                                                                            // 128
                                                                                                                       //
        handle.stop();                                                                                                 // 129
        handle = null;                                                                                                 // 130
      }                                                                                                                // 131
    });                                                                                                                // 126
    this.onStop(function () {                                                                                          // 133
      if (handle) handle.stop();                                                                                       // 134
    });                                                                                                                // 135
  }                                                                                                                    // 136
                                                                                                                       //
  this.ready();                                                                                                        // 138
});                                                                                                                    // 139
Meteor.publish("referralInfoPseudo", function () {                                                                     // 141
  var _this2 = this;                                                                                                   // 141
                                                                                                                       //
  // This publishes a pseudo-collection called referralInfo whose documents have the following                         // 142
  // form:                                                                                                             // 143
  //                                                                                                                   // 144
  // - id: (String) same as the User._id of an identity this user has referred                                         // 145
  // - name: (String) the profile.name from that identity                                                              // 146
  // - completed: (Boolean) if this referral is complete                                                               // 147
  //  If the user is not logged in, then we have no referralInfo.                                                      // 149
  if (!this.userId) {                                                                                                  // 150
    return [];                                                                                                         // 151
  } // Implementation note:                                                                                            // 152
  //                                                                                                                   // 155
  // This pseudo-collection is populated very differently for (1) the completed: false case versus                     // 156
  // the (2) completed: true case.                                                                                     // 157
  // Case 1. Publish information about not-yet-complete referrals.                                                     // 159
                                                                                                                       //
                                                                                                                       //
  var notCompletedReferralIdentitiesCursor = Meteor.users.find({                                                       // 160
    referredBy: this.userId,                                                                                           // 161
    "profile.name": {                                                                                                  // 162
      $exists: true                                                                                                    // 162
    }                                                                                                                  // 162
  }, {                                                                                                                 // 160
    fields: {                                                                                                          // 164
      _id: 1,                                                                                                          // 165
      referredBy: 1,                                                                                                   // 166
      "profile.name": 1                                                                                                // 167
    }                                                                                                                  // 164
  });                                                                                                                  // 163
  var notCompletedReferralIdentitiesHandle = notCompletedReferralIdentitiesCursor.observeChanges({                     // 170
    // The added function gets called with the id of Bob when Alice refers Bob.                                        // 171
    added: function (id, fields) {                                                                                     // 172
      _this2.added("referralInfo", id, {                                                                               // 173
        name: fields.profile.name,                                                                                     // 173
        completed: false                                                                                               // 173
      });                                                                                                              // 173
    },                                                                                                                 // 174
    // The removed function gets called when Bob is no longer an uncompleted referral.  Note that                      // 175
    // this will get more complicated once we support sending completed referrals to the client.                       // 176
    removed: function (id) {                                                                                           // 177
      _this2.removed("referralInfo", id);                                                                              // 178
    },                                                                                                                 // 179
    // The modified function gets called when Bob's profile.name changed.                                              // 180
    modified: function (id, fields) {                                                                                  // 181
      _this2.modified("referralInfo", id, {                                                                            // 182
        name: fields.profile.name,                                                                                     // 182
        completed: false                                                                                               // 182
      });                                                                                                              // 182
    }                                                                                                                  // 183
  }); // Case 2. Handle completed referrals.                                                                           // 170
  //                                                                                                                   // 187
  // - Do a query for the current list of completed identities.                                                        // 188
  //                                                                                                                   // 189
  // - Every time we see a new such identity, we create a query that watches that one identity in                      // 190
  //   case its profile.name changes.                                                                                  // 191
  //                                                                                                                   // 192
  // - Also watch the first query, since the list of completed identities might change.                                // 193
                                                                                                                       //
  var handleForProfileNameByIdentityId = {};                                                                           // 194
                                                                                                                       //
  var stopWatchingAllIdentities = function () {                                                                        // 195
    Object.keys(handleForProfileNameByIdentityId).forEach(function (identityId) {                                      // 196
      stopWatchingIdentity(identityId);                                                                                // 197
    });                                                                                                                // 198
  };                                                                                                                   // 199
                                                                                                                       //
  var stopWatchingIdentity = function (identityId) {                                                                   // 201
    var handleForProfileName = handleForProfileNameByIdentityId[identityId];                                           // 202
                                                                                                                       //
    if (handleForProfileName) {                                                                                        // 203
      _this2.removed("referralInfo", identityId);                                                                      // 204
                                                                                                                       //
      handleForProfileName.stop(); // delete is safe because we iterate across `Object.keys()` which returns a copy.   // 205
                                                                                                                       //
      delete handleForProfileNameByIdentityId[identityId];                                                             // 207
    }                                                                                                                  // 208
  };                                                                                                                   // 209
                                                                                                                       //
  var watchIdentityAndPublishReferralSuccess = function (identityId) {                                                 // 211
    var handleForProfileName = handleForProfileNameByIdentityId[identityId];                                           // 212
                                                                                                                       //
    if (handleForProfileName) {                                                                                        // 213
      return;                                                                                                          // 214
    }                                                                                                                  // 215
                                                                                                                       //
    handleForProfileName = Meteor.users.find({                                                                         // 217
      _id: identityId                                                                                                  // 218
    }, {                                                                                                               // 217
      fields: {                                                                                                        // 220
        "profile.name": 1                                                                                              // 221
      }                                                                                                                // 220
    }).observeChanges({                                                                                                // 219
      added: function (id, fields) {                                                                                   // 224
        _this2.added("referralInfo", id, {                                                                             // 225
          name: fields.profile.name,                                                                                   // 225
          completed: true                                                                                              // 225
        });                                                                                                            // 225
      },                                                                                                               // 226
      changed: function (id, fields) {                                                                                 // 228
        _this2.changed("referralInfo", id, {                                                                           // 229
          name: fields.profile.name,                                                                                   // 229
          completed: true                                                                                              // 229
        });                                                                                                            // 229
      },                                                                                                               // 230
      removed: function (id) {                                                                                         // 232
        stopWatchingIdentity(id);                                                                                      // 233
      }                                                                                                                // 234
    });                                                                                                                // 223
    handleForProfileNameByIdentityId[identityId] = handleForProfileName;                                               // 237
  };                                                                                                                   // 238
                                                                                                                       //
  var completedIdentityIdsHandle = Meteor.users.find({                                                                 // 240
    _id: this.userId,                                                                                                  // 241
    referredIdentityIds: {                                                                                             // 242
      $exists: true                                                                                                    // 242
    }                                                                                                                  // 242
  }, {                                                                                                                 // 240
    fields: {                                                                                                          // 244
      referredIdentityIds: true                                                                                        // 245
    }                                                                                                                  // 244
  }).observeChanges({                                                                                                  // 243
    // `added` gets called when a user gets their first completed referral.                                            // 248
    added: function (id, fields) {                                                                                     // 249
      for (var i = 0; i < fields.referredIdentityIds.length; i++) {                                                    // 250
        // Unconditionally mark these as successful referrals and start watching.                                      // 251
        watchIdentityAndPublishReferralSuccess(fields.referredIdentityIds[i]);                                         // 252
      }                                                                                                                // 254
    },                                                                                                                 // 255
    // `changed` gets called when a user adds/removes referredIdentityIds, usually when a                              // 256
    // referral becomes complete.                                                                                      // 257
    changed: function (id, fields) {                                                                                   // 258
      // Two major tasks.                                                                                              // 259
      //                                                                                                               // 260
      // 1. Look for identityIds to unsubscribe from & send removed notices to the client.                             // 261
      //                                                                                                               // 262
      // 2. Look for identityIds to subscribe to.                                                                      // 263
      // Task 1. Unsubscribe where needed.                                                                             // 265
      var referredIdentityIdsAsObject = {};                                                                            // 266
      fields.referredIdentityIds.forEach(function (i) {                                                                // 267
        referredIdentityIdsAsObject[i] = true;                                                                         // 267
      });                                                                                                              // 267
      Object.keys(handleForProfileNameByIdentityId).forEach(function (identityId) {                                    // 269
        // If the handle doesn't show up in the new list of referredIdentityIds, then remove                           // 270
        // info from the client & stop it on the server & make it null.                                                // 271
        var handleForProfileName = handleForProfileNameByIdentityId[identityId];                                       // 272
                                                                                                                       //
        if (referredIdentityIdsAsObject.hasOwnProperty(identityId)) {                                                  // 273
          stopWatchingIdentity(identityId);                                                                            // 274
        }                                                                                                              // 275
      }); // Task 2. Subscribe where needed.                                                                           // 276
                                                                                                                       //
      for (var i = 0; i < fields.referredIdentityIds.length; i++) {                                                    // 279
        // The watch... function will avoid double-creating subscriptions, so this is safe.                            // 280
        watchIdentityAndPublishReferralSuccess(fields.referredIdentityIds[i]);                                         // 281
      }                                                                                                                // 282
    },                                                                                                                 // 283
    // `removed` gets called when a User suddenly has no referredIdentityIds.                                          // 284
    removed: function () {                                                                                             // 285
      // Remove all data from client; stop all handles.                                                                // 286
      stopWatchingAllIdentities();                                                                                     // 287
    }                                                                                                                  // 288
  }); // With cases 1 and 2 handled, register a cleanup function, then declare victory.                                // 247
                                                                                                                       //
  this.onStop(function () {                                                                                            // 292
    stopWatchingAllIdentities();                                                                                       // 293
    notCompletedReferralIdentitiesHandle.stop();                                                                       // 294
    completedIdentityIdsHandle.stop();                                                                                 // 295
  });                                                                                                                  // 296
  this.ready();                                                                                                        // 298
});                                                                                                                    // 299
Meteor.publish("backers", function () {                                                                                // 301
  var _this3 = this;                                                                                                   // 301
                                                                                                                       //
  var backers = Assets.getText("backers.txt");                                                                         // 302
  var anonCount = 0;                                                                                                   // 303
  var counter = 0;                                                                                                     // 304
  backers.split("\n").forEach(function (name) {                                                                        // 306
    name = name.trim();                                                                                                // 307
                                                                                                                       //
    if (name === "") {                                                                                                 // 308
      ++anonCount;                                                                                                     // 309
    } else {                                                                                                           // 310
      _this3.added("backers", counter++, {                                                                             // 311
        name: name                                                                                                     // 311
      });                                                                                                              // 311
    }                                                                                                                  // 312
  }); // Text file ends in \n but that shouldn't count.                                                                // 313
                                                                                                                       //
  --anonCount;                                                                                                         // 316
  this.added("backers", "anonymous", {                                                                                 // 318
    count: anonCount - 1                                                                                               // 318
  });                                                                                                                  // 318
  this.ready();                                                                                                        // 320
});                                                                                                                    // 321
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}],"signup-server.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/signup-server.js                                                                                             //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file covers creation and consumption of invite keys (i.e. to invite people to become                           // 17
// users of the Sandstorm server).                                                                                     // 18
Meteor.publish("signupKey", function (key) {                                                                           // 20
  check(key, String);                                                                                                  // 21
  return SignupKeys.find(key);                                                                                         // 22
});                                                                                                                    // 23
Meteor.methods({                                                                                                       // 25
  useSignupKey: function (key) {                                                                                       // 26
    check(key, String);                                                                                                // 27
                                                                                                                       //
    if (!this.userId) {                                                                                                // 29
      throw new Meteor.Error(403, "Must be signed in.");                                                               // 30
    }                                                                                                                  // 31
                                                                                                                       //
    var user = Meteor.user();                                                                                          // 33
                                                                                                                       //
    if (user.signupKey) {                                                                                              // 34
      // Don't waste it.                                                                                               // 35
      return;                                                                                                          // 36
    }                                                                                                                  // 37
                                                                                                                       //
    if (isDemoUser()) {                                                                                                // 39
      throw new Meteor.Error(403, "Demo users cannot accept invite keys. Please sign in as a real user.");             // 40
    }                                                                                                                  // 42
                                                                                                                       //
    if (!user.loginIdentities) {                                                                                       // 44
      // Don't consume signup key as identity user.                                                                    // 45
      return;                                                                                                          // 46
    }                                                                                                                  // 47
                                                                                                                       //
    var keyInfo = SignupKeys.findOne(key);                                                                             // 49
                                                                                                                       //
    if (!keyInfo || keyInfo.used) {                                                                                    // 50
      throw new Meteor.Error(403, "Invalid key or already used.");                                                     // 51
    }                                                                                                                  // 52
                                                                                                                       //
    if (isSignedUp() && user.payments && user.payments.id) {                                                           // 54
      // This user is already signed up with a payment account. Possibly, they signed up before                        // 55
      // using their invite, and then went back and clicked on the invite. As a result they                            // 56
      // probably now have two payment accounts. Mark this invite as used but also add a special                       // 57
      // flag so we can find it later and cancel the dupe payment account. Record who tried to                         // 58
      // use it so that we can transfer credits over if needed.                                                        // 59
      SignupKeys.update(key, {                                                                                         // 60
        $set: {                                                                                                        // 60
          used: true,                                                                                                  // 60
          rejectedBy: this.userId                                                                                      // 60
        }                                                                                                              // 60
      });                                                                                                              // 60
      return;                                                                                                          // 61
    }                                                                                                                  // 62
                                                                                                                       //
    var userFields = {                                                                                                 // 64
      signupKey: key,                                                                                                  // 65
      signupNote: keyInfo.note                                                                                         // 66
    };                                                                                                                 // 64
                                                                                                                       //
    if (keyInfo.email) {                                                                                               // 68
      userFields.signupEmail = keyInfo.email;                                                                          // 69
    }                                                                                                                  // 70
                                                                                                                       //
    if ("quota" in keyInfo) {                                                                                          // 72
      userFields.quota = keyInfo.quota;                                                                                // 73
    }                                                                                                                  // 74
                                                                                                                       //
    if (keyInfo.plan) {                                                                                                // 76
      userFields.plan = keyInfo.plan;                                                                                  // 77
    }                                                                                                                  // 78
                                                                                                                       //
    if (keyInfo.payments) {                                                                                            // 80
      userFields["payments.id"] = keyInfo.payments.id;                                                                 // 81
    }                                                                                                                  // 82
                                                                                                                       //
    Meteor.users.update(this.userId, {                                                                                 // 84
      $set: userFields                                                                                                 // 84
    });                                                                                                                // 84
    SignupKeys.update(key, {                                                                                           // 85
      $set: {                                                                                                          // 85
        used: true                                                                                                     // 85
      }                                                                                                                // 85
    });                                                                                                                // 85
  }                                                                                                                    // 86
});                                                                                                                    // 25
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"standalone-server.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/standalone-server.js                                                                                         //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Meteor.publish("standaloneDomain", function (domain) {                                                                 // 17
  check(domain, String);                                                                                               // 18
  return globalDb.collections.standaloneDomains.find({                                                                 // 20
    _id: domain                                                                                                        // 20
  });                                                                                                                  // 20
});                                                                                                                    // 21
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"startup.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/startup.js                                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// This file is for various startup code that doesn't fit neatly anywhere else                                         // 17
var ROOT_URL = process.env.ROOT_URL;                                                                                   // 19
Meteor.startup(function () {                                                                                           // 21
  var baseUrlRow = Misc.findOne({                                                                                      // 22
    _id: "BASE_URL"                                                                                                    // 22
  });                                                                                                                  // 22
                                                                                                                       //
  if (!baseUrlRow) {                                                                                                   // 24
    // Fill data with current value in the case this is a first run                                                    // 25
    baseUrlRow = {                                                                                                     // 26
      _id: "BASE_URL",                                                                                                 // 26
      value: ROOT_URL                                                                                                  // 26
    };                                                                                                                 // 26
    Misc.insert(baseUrlRow);                                                                                           // 27
  } else if (baseUrlRow.value !== ROOT_URL) {                                                                          // 28
    // If the only thing that happened is some trailing slashes got removed, then don't reset OAuth                    // 29
    // providers.                                                                                                      // 30
    var resetOAuth = true;                                                                                             // 31
                                                                                                                       //
    if (baseUrlRow.value.replace(/[/]*$/, "") === ROOT_URL) {                                                          // 32
      console.log("Not resetting OAuth because old ROOT_URL and new ROOT_URL differ only by trailing slashes.");       // 33
      resetOAuth = false;                                                                                              // 34
    }                                                                                                                  // 35
                                                                                                                       //
    if (resetOAuth) {                                                                                                  // 37
      console.log("resetting oauth");                                                                                  // 38
      Settings.find({                                                                                                  // 39
        _id: {                                                                                                         // 39
          $in: ["google", "github"]                                                                                    // 39
        }                                                                                                              // 39
      }).forEach(function (setting) {                                                                                  // 39
        if (!!setting.value) {                                                                                         // 40
          Settings.update({                                                                                            // 41
            _id: setting._id                                                                                           // 41
          }, {                                                                                                         // 41
            $set: {                                                                                                    // 42
              value: false,                                                                                            // 42
              automaticallyReset: {                                                                                    // 43
                baseUrlChangedFrom: baseUrlRow.value                                                                   // 43
              }                                                                                                        // 43
            }                                                                                                          // 42
          });                                                                                                          // 42
        }                                                                                                              // 44
      });                                                                                                              // 45
    }                                                                                                                  // 46
                                                                                                                       //
    baseUrlRow = {                                                                                                     // 48
      _id: "BASE_URL",                                                                                                 // 48
      value: ROOT_URL                                                                                                  // 48
    };                                                                                                                 // 48
    Misc.update({                                                                                                      // 49
      _id: "BASE_URL"                                                                                                  // 49
    }, {                                                                                                               // 49
      $set: {                                                                                                          // 49
        value: ROOT_URL                                                                                                // 49
      }                                                                                                                // 49
    });                                                                                                                // 49
  }                                                                                                                    // 50
});                                                                                                                    // 51
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"stats-server.js":["/imports/demo.js",function(require,exports,module){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// server/stats-server.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
var allowDemo = void 0;                                                                                                // 1
module.import("/imports/demo.js", {                                                                                    // 1
  "allowDemo": function (v) {                                                                                          // 1
    allowDemo = v;                                                                                                     // 1
  }                                                                                                                    // 1
}, 0);                                                                                                                 // 1
var DAY_MS = 24 * 60 * 60 * 1000;                                                                                      // 19
                                                                                                                       //
if (Mongo.Collection.prototype.aggregate) {                                                                            // 21
  throw new Error("Looks like Meteor wrapped the Collection.aggregate() call. Make sure it " + "works then delete our own wrapper.");
}                                                                                                                      // 24
                                                                                                                       //
Mongo.Collection.prototype.aggregate = function () {                                                                   // 26
  // Meteor doesn't wrapp Mongo's aggregate() method.                                                                  // 27
  var raw = this.rawCollection();                                                                                      // 28
  return Meteor.wrapAsync(raw.aggregate, raw).apply(raw, arguments);                                                   // 29
};                                                                                                                     // 30
                                                                                                                       //
computeStats = function (since) {                                                                                      // 32
  // We'll need this for a variety of queries.                                                                         // 33
  var timeConstraint = {                                                                                               // 34
    $gt: since                                                                                                         // 34
  }; // This calculates the number of user accounts that have been used                                                // 34
  // during the requested time period.                                                                                 // 37
                                                                                                                       //
  var currentlyActiveUsersCount = Meteor.users.find({                                                                  // 38
    expires: {                                                                                                         // 39
      $exists: false                                                                                                   // 39
    },                                                                                                                 // 39
    loginIdentities: {                                                                                                 // 40
      $exists: true                                                                                                    // 40
    },                                                                                                                 // 40
    lastActive: timeConstraint                                                                                         // 41
  }).count(); // This calculates the number of grains that have been used during                                       // 38
  // the requested time period.                                                                                        // 45
                                                                                                                       //
  var activeGrainsCount = Grains.find({                                                                                // 46
    lastUsed: timeConstraint                                                                                           // 46
  }).count(); // If Meteor.settings.allowDemoAccounts is true, DeleteStats                                             // 46
  // contains records of type `user` and `appDemoUser`, indicating                                                     // 49
  // the number of those types of accounts that were created and                                                       // 50
  // then auto-expired through the demo mode's auto-account-expiry.                                                    // 51
                                                                                                                       //
  var deletedDemoUsersCount = DeleteStats.find({                                                                       // 52
    type: "demoUser",                                                                                                  // 53
    lastActive: timeConstraint                                                                                         // 53
  }).count();                                                                                                          // 53
  var deletedAppDemoUsersCount = DeleteStats.find({                                                                    // 54
    type: "appDemoUser",                                                                                               // 55
    lastActive: timeConstraint                                                                                         // 55
  }).count(); // Similarly, if the demo is enabled, we auto-delete grains; we store that                               // 55
  // fact in DeleteStats with type: "grain".                                                                           // 58
                                                                                                                       //
  var deletedGrainsCount = DeleteStats.find({                                                                          // 59
    type: "grain",                                                                                                     // 60
    lastActive: timeConstraint                                                                                         // 60
  }).count();                                                                                                          // 60
  var apps = Grains.aggregate([{                                                                                       // 62
    $match: {                                                                                                          // 63
      lastUsed: timeConstraint                                                                                         // 63
    }                                                                                                                  // 63
  }, {                                                                                                                 // 63
    $group: {                                                                                                          // 65
      _id: "$appId",                                                                                                   // 66
      grains: {                                                                                                        // 67
        $sum: 1                                                                                                        // 67
      },                                                                                                               // 67
      userIds: {                                                                                                       // 68
        $addToSet: "$userId"                                                                                           // 68
      }                                                                                                                // 68
    }                                                                                                                  // 65
  }, {                                                                                                                 // 64
    $project: {                                                                                                        // 72
      grains: 1,                                                                                                       // 73
      owners: {                                                                                                        // 74
        $size: "$userIds"                                                                                              // 74
      }                                                                                                                // 74
    }                                                                                                                  // 72
  }]);                                                                                                                 // 71
  apps = _.indexBy(apps, "_id");                                                                                       // 78
                                                                                                                       //
  for (var appId in meteorBabelHelpers.sanitizeForInObject(apps)) {                                                    // 80
    // We need to count ApiTokens, which don't have appId denormalized into them. We therefore                         // 81
    // have to fetch a list of grainIds first.                                                                         // 82
    // TODO(perf): If stats are getting slow, denormalize appId into ApiTokens. Note that it is                        // 83
    //   already denormalized for the specific case of apps that don't have icons, but the data                        // 84
    //   for that use case is NOT safe to use here because it's intended that we might allow an                        // 85
    //   app to mimic another app's icon by spoofing that app ID. In other words, the existing                         // 86
    //   denormalization of appId should be considered "app ID for identicon purposes only". We'll                     // 87
    //   need to add a new denormalization for stats purposes -- and make sure that it is not                          // 88
    //   revealed to the user.                                                                                         // 89
    var app = apps[appId];                                                                                             // 90
    delete app._id;                                                                                                    // 91
    var grains = Grains.find({                                                                                         // 92
      lastUsed: timeConstraint,                                                                                        // 93
      appId: appId                                                                                                     // 94
    }, {                                                                                                               // 92
      fields: {                                                                                                        // 96
        _id: 1                                                                                                         // 96
      }                                                                                                                // 96
    }).fetch();                                                                                                        // 95
                                                                                                                       //
    var grainIds = _.pluck(grains, "_id");                                                                             // 98
                                                                                                                       //
    var counts = ApiTokens.aggregate([{                                                                                // 100
      $match: {                                                                                                        // 102
        "owner.user": {                                                                                                // 103
          $exists: true                                                                                                // 103
        },                                                                                                             // 103
        lastUsed: timeConstraint,                                                                                      // 104
        grainId: {                                                                                                     // 105
          $in: grainIds                                                                                                // 105
        }                                                                                                              // 105
      }                                                                                                                // 102
    }, {                                                                                                               // 101
      $group: {                                                                                                        // 108
        _id: "$owner.user.identityId"                                                                                  // 108
      }                                                                                                                // 108
    }, {                                                                                                               // 108
      $group: {                                                                                                        // 110
        _id: "count",                                                                                                  // 111
        count: {                                                                                                       // 112
          $sum: 1                                                                                                      // 112
        }                                                                                                              // 112
      }                                                                                                                // 110
    }]);                                                                                                               // 109
                                                                                                                       //
    if (counts.length > 0) {                                                                                           // 117
      if (counts.length !== 1) {                                                                                       // 118
        console.error("error: sharedUsers aggregation returned multiple rows");                                        // 119
      }                                                                                                                // 120
                                                                                                                       //
      app.sharedUsers = counts[0].count;                                                                               // 122
    }                                                                                                                  // 123
  } // Count per-app appdemo users and deleted grains.                                                                 // 124
                                                                                                                       //
                                                                                                                       //
  DeleteStats.aggregate([{                                                                                             // 127
    $match: {                                                                                                          // 129
      lastActive: timeConstraint,                                                                                      // 130
      appId: {                                                                                                         // 131
        $exists: true                                                                                                  // 131
      }                                                                                                                // 131
    }                                                                                                                  // 129
  }, {                                                                                                                 // 128
    $group: {                                                                                                          // 135
      _id: {                                                                                                           // 136
        appId: "$appId",                                                                                               // 137
        type: "$type"                                                                                                  // 138
      },                                                                                                               // 136
      count: {                                                                                                         // 140
        $sum: 1                                                                                                        // 140
      }                                                                                                                // 140
    }                                                                                                                  // 135
  }]).forEach(function (deletion) {                                                                                    // 134
    var app = apps[deletion._id.appId];                                                                                // 144
                                                                                                                       //
    if (!app) {                                                                                                        // 145
      app = apps[deletion.appId] = {};                                                                                 // 146
    }                                                                                                                  // 147
                                                                                                                       //
    if (deletion._id.type === "appDemoUser") {                                                                         // 149
      app.appDemoUsers = deletion.count;                                                                               // 150
    } else if (deletion._id.type === "grain") {                                                                        // 151
      app.deleted = deletion.count;                                                                                    // 152
    } else if (deletion._id.type === "demoGrain") {                                                                    // 153
      app.demoed = deletion.count;                                                                                     // 154
    }                                                                                                                  // 155
  });                                                                                                                  // 156
  return {                                                                                                             // 158
    activeUsers: currentlyActiveUsersCount,                                                                            // 159
    demoUsers: deletedDemoUsersCount,                                                                                  // 160
    appDemoUsers: deletedAppDemoUsersCount,                                                                            // 161
    activeGrains: activeGrainsCount + deletedGrainsCount,                                                              // 162
    apps: apps                                                                                                         // 163
  };                                                                                                                   // 158
};                                                                                                                     // 165
                                                                                                                       //
function recordStats() {                                                                                               // 167
  var postStats = function (record) {                                                                                  // 168
    HTTP.post("https://alpha-api.sandstorm.io/data", {                                                                 // 169
      data: record,                                                                                                    // 170
      headers: {                                                                                                       // 171
        Authorization: "Bearer aT-mGyNwsgwZBbZvd5FWr0Ma79O9IehI4NiEO94y_oR",                                           // 172
        "Content-Type": "application/json"                                                                             // 173
      }                                                                                                                // 171
    });                                                                                                                // 169
  };                                                                                                                   // 176
                                                                                                                       //
  var now = new Date();                                                                                                // 178
                                                                                                                       //
  var planStats = _.countBy(Meteor.users.find({                                                                        // 180
    expires: {                                                                                                         // 181
      $exists: false                                                                                                   // 181
    },                                                                                                                 // 181
    "payments.id": {                                                                                                   // 181
      $exists: true                                                                                                    // 181
    }                                                                                                                  // 181
  }, {                                                                                                                 // 181
    fields: {                                                                                                          // 182
      plan: 1                                                                                                          // 182
    }                                                                                                                  // 182
  }).fetch(), "plan");                                                                                                 // 182
                                                                                                                       //
  var record = {                                                                                                       // 186
    timestamp: now,                                                                                                    // 187
    daily: computeStats(new Date(now.getTime() - DAY_MS)),                                                             // 188
    weekly: computeStats(new Date(now.getTime() - 7 * DAY_MS)),                                                        // 189
    monthly: computeStats(new Date(now.getTime() - 30 * DAY_MS)),                                                      // 190
    forever: computeStats(new Date(0)),                                                                                // 191
    plans: planStats                                                                                                   // 192
  };                                                                                                                   // 186
  record.computeTime = Date.now() - now;                                                                               // 194
                                                                                                                       //
  if (global.BlackrockPayments && global.BlackrockPayments.getTotalCharges) {                                          // 195
    // This only exists under Blackrock                                                                                // 196
    record.totalCharges = global.BlackrockPayments.getTotalCharges();                                                  // 197
  }                                                                                                                    // 198
                                                                                                                       //
  ActivityStats.insert(record);                                                                                        // 200
  var age = ActivityStats.find().count(); // The stats page which the user agreed we can send actually displays the whole history
  // of the server, but we're only sending stats from the last day. Let's also throw in the                            // 203
  // length of said history. This is still strictly less information than what the user said                           // 204
  // we're allowed to send.                                                                                            // 205
                                                                                                                       //
  record.serverAge = age;                                                                                              // 206
                                                                                                                       //
  if (age > 3) {                                                                                                       // 208
    var reportSetting = Settings.findOne({                                                                             // 209
      _id: "reportStats"                                                                                               // 209
    });                                                                                                                // 209
                                                                                                                       //
    if (!reportSetting) {                                                                                              // 211
      // Setting not set yet, send out notifications and set it to false                                               // 212
      globalDb.sendAdminNotification("reportStats", "/admin/stats");                                                   // 213
      Settings.insert({                                                                                                // 214
        _id: "reportStats",                                                                                            // 214
        value: "unset"                                                                                                 // 214
      });                                                                                                              // 214
    } else if (reportSetting.value === true) {                                                                         // 215
      postStats(record);                                                                                               // 216
    }                                                                                                                  // 217
  }                                                                                                                    // 218
}                                                                                                                      // 219
                                                                                                                       //
if (!Meteor.settings.replicaNumber) {                                                                                  // 221
  // Wait until 10:00 UTC (2:00 PST / 5:00 EST), then start recording stats every 24 hours.                            // 222
  // (Only on the first replica to avoid conflicts.)                                                                   // 223
  Meteor.setTimeout(function () {                                                                                      // 224
    Meteor.setInterval(function () {                                                                                   // 225
      recordStats();                                                                                                   // 226
    }, DAY_MS);                                                                                                        // 227
    recordStats();                                                                                                     // 229
  }, DAY_MS - (Date.now() - 10 * 60 * 60 * 1000) % DAY_MS);                                                            // 230
  Meteor.startup(function () {                                                                                         // 232
    if (StatsTokens.find().count() === 0) {                                                                            // 233
      StatsTokens.remove({});                                                                                          // 234
      StatsTokens.insert({                                                                                             // 235
        _id: Random.id(22)                                                                                             // 235
      });                                                                                                              // 235
    }                                                                                                                  // 236
  });                                                                                                                  // 237
}                                                                                                                      // 238
                                                                                                                       //
Meteor.methods({                                                                                                       // 240
  regenerateStatsToken: function () {                                                                                  // 241
    if (!isAdmin()) {                                                                                                  // 242
      throw new Meteor.Error(403, "Unauthorized", "User must be admin");                                               // 243
    }                                                                                                                  // 244
                                                                                                                       //
    StatsTokens.remove({});                                                                                            // 246
    var token = StatsTokens.insert({                                                                                   // 247
      _id: Random.id(22)                                                                                               // 247
    });                                                                                                                // 247
    return token._id;                                                                                                  // 248
  }                                                                                                                    // 249
});                                                                                                                    // 240
Router.map(function () {                                                                                               // 252
  this.route("fetchStats", {                                                                                           // 253
    where: "server",                                                                                                   // 254
    path: "/fetchStats/:tokenId",                                                                                      // 255
    action: function () {                                                                                              // 256
      var token = StatsTokens.findOne({                                                                                // 257
        _id: this.params.tokenId                                                                                       // 257
      });                                                                                                              // 257
                                                                                                                       //
      if (!token) {                                                                                                    // 259
        this.response.writeHead(404, {                                                                                 // 260
          "Content-Type": "text/plain"                                                                                 // 261
        });                                                                                                            // 260
        this.response.write("Token not found");                                                                        // 263
        return this.response.end();                                                                                    // 264
      }                                                                                                                // 265
                                                                                                                       //
      try {                                                                                                            // 267
        var stats = ActivityStats.find().fetch();                                                                      // 268
        var statsString = JSON.stringify(stats);                                                                       // 269
        this.response.writeHead(200, {                                                                                 // 271
          "Content-Type": "application/json"                                                                           // 272
        });                                                                                                            // 271
        this.response.write(statsString);                                                                              // 274
      } catch (error) {                                                                                                // 275
        console.error(error.stack);                                                                                    // 276
        this.response.writeHead(500, {                                                                                 // 277
          "Content-Type": "text/plain"                                                                                 // 278
        });                                                                                                            // 277
        this.response.write(error.stack);                                                                              // 280
      }                                                                                                                // 281
                                                                                                                       //
      return this.response.end();                                                                                      // 283
    }                                                                                                                  // 284
  });                                                                                                                  // 253
});                                                                                                                    // 286
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}]},"shared":{"admin.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/admin.js                                                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
// TODO(someday): split out fields by locus                                                                            // 17
// isEnabled is available on both client and server                                                                    // 18
// getLoginId is only used on server                                                                                   // 19
// initiateLogin and loginTemplate are only used on client                                                             // 20
function serviceEnabled(name) {                                                                                        // 22
  var setting = Settings.findOne({                                                                                     // 23
    _id: name                                                                                                          // 23
  });                                                                                                                  // 23
  return setting && !!setting.value;                                                                                   // 24
}                                                                                                                      // 25
                                                                                                                       //
Accounts.identityServices.github = {                                                                                   // 27
  isEnabled: function () {                                                                                             // 28
    return serviceEnabled("github");                                                                                   // 29
  },                                                                                                                   // 30
  getLoginId: function (identity) {                                                                                    // 32
    return identity.services.github.username;                                                                          // 33
  },                                                                                                                   // 34
  initiateLogin: function (loginId) {                                                                                  // 36
    Meteor.loginWithGithub();                                                                                          // 37
    return {                                                                                                           // 38
      oneClick: true                                                                                                   // 38
    };                                                                                                                 // 38
  },                                                                                                                   // 39
  loginTemplate: {                                                                                                     // 41
    name: "oauthLoginButton",                                                                                          // 42
    priority: 1,                                                                                                       // 43
    data: {                                                                                                            // 44
      method: "loginWithGithub",                                                                                       // 45
      name: "github",                                                                                                  // 46
      displayName: "GitHub",                                                                                           // 47
      linkingNewIdentity: false                                                                                        // 48
    }                                                                                                                  // 44
  }                                                                                                                    // 41
};                                                                                                                     // 27
Accounts.identityServices.google = {                                                                                   // 53
  isEnabled: function () {                                                                                             // 54
    return serviceEnabled("google");                                                                                   // 55
  },                                                                                                                   // 56
  getLoginId: function (identity) {                                                                                    // 58
    return identity.services.google.email;                                                                             // 59
  },                                                                                                                   // 60
  initiateLogin: function (loginId) {                                                                                  // 62
    Meteor.loginWithGoogle({                                                                                           // 63
      loginHint: loginId                                                                                               // 63
    });                                                                                                                // 63
    return {                                                                                                           // 64
      oneClick: true                                                                                                   // 64
    };                                                                                                                 // 64
  },                                                                                                                   // 65
  loginTemplate: {                                                                                                     // 67
    name: "oauthLoginButton",                                                                                          // 68
    priority: 2,                                                                                                       // 69
    data: {                                                                                                            // 70
      method: "loginWithGoogle",                                                                                       // 71
      name: "google",                                                                                                  // 72
      displayName: "Google",                                                                                           // 73
      linkingNewIdentity: false                                                                                        // 74
    }                                                                                                                  // 70
  }                                                                                                                    // 67
};                                                                                                                     // 53
Accounts.identityServices.email = {                                                                                    // 79
  isEnabled: function () {                                                                                             // 80
    return serviceEnabled("emailToken");                                                                               // 81
  },                                                                                                                   // 82
  getLoginId: function (identity) {                                                                                    // 84
    return identity.services.email.email;                                                                              // 85
  },                                                                                                                   // 86
  initiateLogin: function (loginId) {                                                                                  // 88
    return {                                                                                                           // 89
      form: true                                                                                                       // 89
    };                                                                                                                 // 89
  },                                                                                                                   // 90
  loginTemplate: {                                                                                                     // 92
    name: "emailLoginForm",                                                                                            // 93
    priority: 10                                                                                                       // 94
  }                                                                                                                    // 92
};                                                                                                                     // 79
Accounts.identityServices.ldap = {                                                                                     // 98
  isEnabled: function () {                                                                                             // 99
    return serviceEnabled("ldap");                                                                                     // 100
  },                                                                                                                   // 101
  getLoginId: function (identity) {                                                                                    // 103
    return identity.services.ldap.username;                                                                            // 104
  },                                                                                                                   // 105
  initiateLogin: function (loginId) {                                                                                  // 107
    return {                                                                                                           // 108
      form: true                                                                                                       // 108
    };                                                                                                                 // 108
  },                                                                                                                   // 109
  loginTemplate: {                                                                                                     // 111
    name: "ldapLoginForm",                                                                                             // 112
    priority: 21                                                                                                       // 113
  }                                                                                                                    // 111
};                                                                                                                     // 98
Accounts.identityServices.saml = {                                                                                     // 117
  isEnabled: function () {                                                                                             // 118
    return serviceEnabled("saml");                                                                                     // 119
  },                                                                                                                   // 120
  getLoginId: function (identity) {                                                                                    // 122
    return identity.services.saml.id;                                                                                  // 123
  },                                                                                                                   // 124
  initiateLogin: function (loginId) {                                                                                  // 126
    Meteor.loginWithSaml();                                                                                            // 127
    return {                                                                                                           // 128
      oneClick: true                                                                                                   // 128
    };                                                                                                                 // 128
  },                                                                                                                   // 129
  loginTemplate: {                                                                                                     // 131
    name: "samlLoginForm",                                                                                             // 132
    priority: 20                                                                                                       // 133
  }                                                                                                                    // 131
};                                                                                                                     // 117
Accounts.identityServices.demo = {                                                                                     // 137
  isEnabled: function () {                                                                                             // 138
    return false; // You can never authenticate as a demo user.                                                        // 139
  },                                                                                                                   // 140
  getLoginId: function (identity) {                                                                                    // 142
    return identity._id;                                                                                               // 143
  },                                                                                                                   // 144
  initiateLogin: function (loginId) {                                                                                  // 146
    return {                                                                                                           // 147
      demoCannotLogin: true                                                                                            // 147
    };                                                                                                                 // 147
  },                                                                                                                   // 148
  loginTemplate: {}                                                                                                    // 150
};                                                                                                                     // 137
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"dev-accounts.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/dev-accounts.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Accounts.identityServices.dev = {                                                                                      // 17
  isEnabled: function () {                                                                                             // 18
    return globalDb.allowDevAccounts();                                                                                // 19
  },                                                                                                                   // 20
  getLoginId: function (identity) {                                                                                    // 22
    return identity.services.dev.name;                                                                                 // 23
  },                                                                                                                   // 24
  initiateLogin: function (loginId) {                                                                                  // 26
    loginDevAccount(loginId);                                                                                          // 27
    return {                                                                                                           // 28
      oneClick: true                                                                                                   // 28
    };                                                                                                                 // 28
  },                                                                                                                   // 29
  loginTemplate: {                                                                                                     // 31
    name: "devLoginForm",                                                                                              // 32
    priority: -10                                                                                                      // 33
  }                                                                                                                    // 31
};                                                                                                                     // 17
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"grain-shared.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/grain-shared.js                                                                                              //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
Meteor.methods({                                                                                                       // 17
  // Methods defined in this file have meaningful latency compensation (client-side prediction)                        // 18
  // potential.                                                                                                        // 19
  //                                                                                                                   // 20
  // Methods for which latency compensation makes no sense are defined in grain-server.js.                             // 21
  markActivityReadByOwner: function (grainId) {                                                                        // 23
    check(grainId, String);                                                                                            // 24
    check(this.userId, String);                                                                                        // 25
    Grains.update({                                                                                                    // 27
      _id: grainId,                                                                                                    // 27
      userId: this.userId                                                                                              // 27
    }, {                                                                                                               // 27
      $set: {                                                                                                          // 28
        "ownerSeenAllActivity": true                                                                                   // 28
      }                                                                                                                // 28
    });                                                                                                                // 28
  },                                                                                                                   // 29
  markActivityRead: function (grainId, identityId) {                                                                   // 31
    check(grainId, String);                                                                                            // 32
    check(identityId, String);                                                                                         // 33
                                                                                                                       //
    if (!this.userId || !globalDb.userHasIdentity(this.userId, identityId)) {                                          // 35
      throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);                               // 36
    }                                                                                                                  // 37
                                                                                                                       //
    ApiTokens.update({                                                                                                 // 39
      "grainId": grainId,                                                                                              // 39
      "owner.user.identityId": identityId                                                                              // 39
    }, {                                                                                                               // 39
      $set: {                                                                                                          // 40
        "owner.user.seenAllActivity": true                                                                             // 40
      }                                                                                                                // 40
    }, {                                                                                                               // 40
      multi: true                                                                                                      // 40
    });                                                                                                                // 40
  },                                                                                                                   // 41
  moveGrainsToTrash: function (grainIds) {                                                                             // 43
    var _this = this;                                                                                                  // 43
                                                                                                                       //
    check(grainIds, [String]);                                                                                         // 44
                                                                                                                       //
    if (this.userId) {                                                                                                 // 46
      Grains.update({                                                                                                  // 47
        userId: {                                                                                                      // 47
          $eq: this.userId                                                                                             // 47
        },                                                                                                             // 47
        _id: {                                                                                                         // 48
          $in: grainIds                                                                                                // 48
        },                                                                                                             // 48
        trashed: {                                                                                                     // 49
          $exists: false                                                                                               // 49
        }                                                                                                              // 49
      }, {                                                                                                             // 47
        $set: {                                                                                                        // 50
          trashed: new Date()                                                                                          // 50
        }                                                                                                              // 50
      }, {                                                                                                             // 50
        multi: true                                                                                                    // 51
      });                                                                                                              // 51
      var identityIds = SandstormDb.getUserIdentityIds(Meteor.user());                                                 // 53
      ApiTokens.update({                                                                                               // 55
        grainId: {                                                                                                     // 55
          $in: grainIds                                                                                                // 55
        },                                                                                                             // 55
        "owner.user.identityId": {                                                                                     // 56
          $in: identityIds                                                                                             // 56
        },                                                                                                             // 56
        trashed: {                                                                                                     // 57
          $exists: false                                                                                               // 57
        }                                                                                                              // 57
      }, {                                                                                                             // 55
        $set: {                                                                                                        // 58
          "trashed": new Date()                                                                                        // 58
        }                                                                                                              // 58
      }, {                                                                                                             // 58
        multi: true                                                                                                    // 59
      });                                                                                                              // 59
                                                                                                                       //
      if (!this.isSimulation) {                                                                                        // 61
        var grainsOwned = Grains.find({                                                                                // 62
          userId: {                                                                                                    // 63
            $eq: this.userId                                                                                           // 63
          },                                                                                                           // 63
          _id: {                                                                                                       // 64
            $in: grainIds                                                                                              // 64
          }                                                                                                            // 64
        }, {                                                                                                           // 62
          fields: {                                                                                                    // 65
            _id: 1                                                                                                     // 65
          }                                                                                                            // 65
        });                                                                                                            // 65
        grainsOwned.forEach(function (grain) {                                                                         // 67
          Sessions.remove({                                                                                            // 68
            grainId: grain._id                                                                                         // 68
          });                                                                                                          // 68
                                                                                                                       //
          try {                                                                                                        // 69
            _this.connection.sandstormBackend.shutdownGrain(grain._id, _this.userId).await();                          // 70
          } catch (err) {                                                                                              // 71
            console.error("Failed to shutdown trashed grain", grain._id, err);                                         // 72
          }                                                                                                            // 73
        });                                                                                                            // 74
      }                                                                                                                // 75
    }                                                                                                                  // 76
  },                                                                                                                   // 77
  moveGrainsOutOfTrash: function (grainIds) {                                                                          // 79
    check(grainIds, [String]);                                                                                         // 80
                                                                                                                       //
    if (this.userId) {                                                                                                 // 82
      Grains.update({                                                                                                  // 83
        userId: {                                                                                                      // 83
          $eq: this.userId                                                                                             // 83
        },                                                                                                             // 83
        _id: {                                                                                                         // 84
          $in: grainIds                                                                                                // 84
        },                                                                                                             // 84
        trashed: {                                                                                                     // 85
          $exists: true                                                                                                // 85
        }                                                                                                              // 85
      }, {                                                                                                             // 83
        $unset: {                                                                                                      // 86
          trashed: 1                                                                                                   // 86
        }                                                                                                              // 86
      }, {                                                                                                             // 86
        multi: true                                                                                                    // 87
      });                                                                                                              // 87
      var identityIds = SandstormDb.getUserIdentityIds(Meteor.user());                                                 // 89
      ApiTokens.update({                                                                                               // 91
        grainId: {                                                                                                     // 91
          $in: grainIds                                                                                                // 91
        },                                                                                                             // 91
        "owner.user.identityId": {                                                                                     // 92
          $in: identityIds                                                                                             // 92
        },                                                                                                             // 92
        "trashed": {                                                                                                   // 93
          $exists: true                                                                                                // 93
        }                                                                                                              // 93
      }, {                                                                                                             // 91
        $unset: {                                                                                                      // 94
          "trashed": 1                                                                                                 // 94
        }                                                                                                              // 94
      }, {                                                                                                             // 94
        multi: true                                                                                                    // 95
      });                                                                                                              // 95
    }                                                                                                                  // 96
  },                                                                                                                   // 97
  deleteGrain: function (grainId) {                                                                                    // 99
    check(grainId, String);                                                                                            // 100
                                                                                                                       //
    if (this.userId) {                                                                                                 // 102
      var grainsQuery = {                                                                                              // 103
        _id: grainId,                                                                                                  // 104
        userId: this.userId,                                                                                           // 105
        trashed: {                                                                                                     // 106
          $exists: true                                                                                                // 106
        }                                                                                                              // 106
      };                                                                                                               // 103
      var numDeleted = 0;                                                                                              // 109
                                                                                                                       //
      if (this.isSimulation) {                                                                                         // 110
        numDeleted = Grains.remove(grainsQuery);                                                                       // 111
      } else {                                                                                                         // 112
        numDeleted = globalDb.deleteGrains(grainsQuery, globalBackend, isDemoUser() ? "demoGrain" : "grain");          // 113
      } // Usually we don't automatically remove user-owned tokens that have become invalid,                           // 115
      // because if we did their owner might become confused as to why they have mysteriously                          // 118
      // disappeared. In this particular case, however, for tokens held by the grain owner,                            // 119
      // there should be no confusion. Indeed, it would be more confusing *not* to remove these                        // 120
      // tokens, because then the grain could still show up in the trash bin as a "shared with me"                     // 121
      // grain after the owner clicks "delete permanently".                                                            // 122
      //                                                                                                               // 123
      // Note that these tokens may be visible to other accounts if there are identities shared                        // 124
      // between the accounts; by only removing 'trashed' tokens, we minimize confusion in that                        // 125
      // case too.                                                                                                     // 126
                                                                                                                       //
                                                                                                                       //
      var apiTokensQuery = {                                                                                           // 127
        grainId: grainId,                                                                                              // 128
        "owner.user.identityId": {                                                                                     // 129
          $in: SandstormDb.getUserIdentityIds(Meteor.user())                                                           // 129
        },                                                                                                             // 129
        "trashed": {                                                                                                   // 130
          $exists: true                                                                                                // 130
        }                                                                                                              // 130
      };                                                                                                               // 127
                                                                                                                       //
      if (numDeleted > 0) {                                                                                            // 133
        if (this.isSimulation) {                                                                                       // 134
          ApiTokens.remove(apiTokensQuery);                                                                            // 135
        } else {                                                                                                       // 136
          globalDb.removeApiTokens(apiTokensQuery);                                                                    // 137
        }                                                                                                              // 138
      }                                                                                                                // 139
    }                                                                                                                  // 140
  },                                                                                                                   // 141
  forgetGrain: function (grainId, identityId) {                                                                        // 143
    check(grainId, String);                                                                                            // 144
    check(identityId, String);                                                                                         // 145
                                                                                                                       //
    if (!this.userId) {                                                                                                // 147
      throw new Meteor.Error(403, "Must be logged in to forget a grain.");                                             // 148
    }                                                                                                                  // 149
                                                                                                                       //
    if (!globalDb.userHasIdentity(this.userId, identityId)) {                                                          // 151
      throw new Meteor.Error(403, "Current user does not have the identity " + identityId);                            // 152
    }                                                                                                                  // 153
                                                                                                                       //
    var query = {                                                                                                      // 155
      grainId: grainId,                                                                                                // 156
      "owner.user.identityId": identityId,                                                                             // 157
      "trashed": {                                                                                                     // 158
        $exists: true                                                                                                  // 158
      }                                                                                                                // 158
    };                                                                                                                 // 155
                                                                                                                       //
    if (this.isSimulation) {                                                                                           // 161
      ApiTokens.remove(query);                                                                                         // 162
    } else {                                                                                                           // 163
      globalDb.removeApiTokens(query);                                                                                 // 164
    }                                                                                                                  // 165
  }                                                                                                                    // 166
});                                                                                                                    // 17
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"install.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/install.js                                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
isSafeDemoAppUrl = function () {                                                                                       // 17
    function isSafeDemoAppUrl(url) {                                                                                   // 17
        // For demo accounts, we allow using a bare hash with no URL (which will never upload a new app)               // 18
        // and we allow specifying a sandstorm.io URL.                                                                 // 19
        return !url || url.lastIndexOf("http://sandstorm.io/", 0) === 0 || url.lastIndexOf("https://sandstorm.io/", 0) === 0 || url.lastIndexOf("https://alpha-j7uny7u376jnimcsx34c.sandstorm.io/", 0) === 0 || url.lastIndexOf("https://app-index.sandstorm.io/", 0) === 0;
    }                                                                                                                  // 25
                                                                                                                       //
    return isSafeDemoAppUrl;                                                                                           // 17
}();                                                                                                                   // 17
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

},"testing.js":function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// shared/testing.js                                                                                                   //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
var isTesting = Meteor.settings && Meteor.settings.public && Meteor.settings.public.isTesting; // jscs:disable requireCamelCaseOrUpperCaseIdentifiers
                                                                                                                       //
if (isTesting) {                                                                                                       // 21
  if (Meteor.isServer) {                                                                                               // 22
    var clearUser = function (id) {                                                                                    // 22
      UserActions.remove({                                                                                             // 24
        userId: id                                                                                                     // 24
      });                                                                                                              // 24
      globalDb.removeApiTokens({                                                                                       // 25
        userId: id                                                                                                     // 25
      });                                                                                                              // 25
      Grains.find({                                                                                                    // 26
        userId: id                                                                                                     // 26
      }).forEach(function (grain) {                                                                                    // 26
        globalBackend.deleteGrain(grain._id);                                                                          // 27
      });                                                                                                              // 28
      Grains.remove({                                                                                                  // 30
        userId: id                                                                                                     // 30
      });                                                                                                              // 30
      Meteor.users.remove({                                                                                            // 31
        _id: id                                                                                                        // 31
      });                                                                                                              // 31
    };                                                                                                                 // 32
                                                                                                                       //
    Meteor.methods({                                                                                                   // 34
      createMockGithubUser: function () {                                                                              // 35
        Meteor.users.update({                                                                                          // 36
          _id: "Py8fwsaryQNGBuiXb"                                                                                     // 36
        }, {                                                                                                           // 36
          $set: {                                                                                                      // 37
            createdAt: new Date("2014-08-11T21:44:04.147Z"),                                                           // 37
            isAdmin: true,                                                                                             // 37
            lastActive: new Date("2014-08-19T09:58:39.676Z"),                                                          // 37
            profile: {                                                                                                 // 37
              name: "Github User"                                                                                      // 37
            },                                                                                                         // 37
            services: {                                                                                                // 37
              github: {                                                                                                // 37
                accessToken: "sometoken",                                                                              // 37
                id: 1595880,                                                                                           // 37
                username: "testuser"                                                                                   // 37
              },                                                                                                       // 37
              resume: {                                                                                                // 37
                loginTokens: [{                                                                                        // 37
                  when: new Date("2099-08-13T05:16:02.356Z"),                                                          // 37
                  hashedToken: "GriUSDp+uN/K4HptwSl1wsdWfHEpS8c9KjjdqwKNo0k="                                          // 37
                }]                                                                                                     // 37
              }                                                                                                        // 37
            },                                                                                                         // 37
            signupKey: "admin"                                                                                         // 37
          }                                                                                                            // 37
        }, {                                                                                                           // 37
          upsert: true                                                                                                 // 38
        });                                                                                                            // 38
      },                                                                                                               // 39
      clearMockGithubUser: function () {                                                                               // 41
        clearUser("Py8fwsaryQNGBuiXb");                                                                                // 42
      },                                                                                                               // 43
      createMockGoogleUser: function () {                                                                              // 45
        Meteor.users.update({                                                                                          // 46
          _id: "6WJcRo2gg2Ysuxsok"                                                                                     // 46
        }, {                                                                                                           // 46
          $set: {                                                                                                      // 47
            createdAt: new Date("2014-08-21T07:52:55.581Z"),                                                           // 47
            profile: {                                                                                                 // 47
              name: "Google User"                                                                                      // 47
            },                                                                                                         // 47
            services: {                                                                                                // 47
              google: {                                                                                                // 47
                accessToken: "sometoken",                                                                              // 47
                expiresAt: 4562182723000,                                                                              // 47
                id: "116893057283177439912",                                                                           // 47
                verified_email: true,                                                                                  // 47
                name: "Google User",                                                                                   // 47
                given_name: "Google",                                                                                  // 47
                family_name: "User",                                                                                   // 47
                picture: "https://lh3.googleusercontent.com/-XdUIqdMkCWA/AAAAAAAAAAI/AAAAAAAAAAA/4252rscbv5M/photo.jpg",
                locale: "en",                                                                                          // 47
                gender: "male"                                                                                         // 47
              },                                                                                                       // 47
              resume: {                                                                                                // 47
                loginTokens: [{                                                                                        // 47
                  when: new Date("2099-08-21T07:52:55.592Z"),                                                          // 47
                  hashedToken: "cbJGxLGKW3f0j7Ehit77hdK58W7xuPjzZhGHgKhyddo="                                          // 47
                }]                                                                                                     // 47
              }                                                                                                        // 47
            },                                                                                                         // 47
            signupKey: "admin"                                                                                         // 47
          }                                                                                                            // 47
        }, {                                                                                                           // 47
          upsert: true                                                                                                 // 48
        });                                                                                                            // 48
      },                                                                                                               // 49
      clearMockGoogleUser: function () {                                                                               // 51
        clearUser("6WJcRo2gg2Ysuxsok");                                                                                // 52
      },                                                                                                               // 53
      fetchAppIndexTest: function () {                                                                                 // 55
        globalDb.collections.appIndex.remove({});                                                                      // 56
        SandstormAutoupdateApps.updateAppIndex(this.connection.sandstormDb);                                           // 57
      }                                                                                                                // 58
    });                                                                                                                // 34
  }                                                                                                                    // 60
                                                                                                                       //
  mockLoginGithub = function () {                                                                                      // 62
    Meteor.call("createMockGithubUser", function (err) {                                                               // 63
      if (err) {                                                                                                       // 64
        console.log(err);                                                                                              // 65
      }                                                                                                                // 66
                                                                                                                       //
      window.localStorage.setItem("Meteor.loginToken", "F0S6luPxIJV--y_GAkBKOMuWMYnnqgG3UMv9M-DIs2f");                 // 68
      window.localStorage.setItem("Meteor.loginTokenExpires", "Mon Nov 10 2099 21:16:02 GMT-0800 (PST)");              // 69
      window.localStorage.setItem("Meteor.userId", "Py8fwsaryQNGBuiXb");                                               // 70
      window.location.reload();                                                                                        // 71
    });                                                                                                                // 72
  };                                                                                                                   // 73
                                                                                                                       //
  clearMockGithubUser = function () {                                                                                  // 75
    Meteor.call("clearMockGithubUser", function (err) {                                                                // 76
      if (err) {                                                                                                       // 77
        console.log(err);                                                                                              // 78
      }                                                                                                                // 79
    });                                                                                                                // 80
  };                                                                                                                   // 81
                                                                                                                       //
  mockLoginGoogle = function () {                                                                                      // 83
    Meteor.call("createMockGoogleUser", function (err) {                                                               // 84
      if (err) {                                                                                                       // 85
        console.log(err);                                                                                              // 86
      }                                                                                                                // 87
                                                                                                                       //
      window.localStorage.setItem("Meteor.loginToken", "P3ffUfVJtptyVX2IPUfNDZY0F3b-GIZ-WQf7w3GdL21");                 // 89
      window.localStorage.setItem("Meteor.loginTokenExpires", "Tue Nov 18 2099 23:52:55 GMT-0800 (PST)");              // 90
      window.localStorage.setItem("Meteor.userId", "6WJcRo2gg2Ysuxsok");                                               // 91
      window.location.reload();                                                                                        // 92
    });                                                                                                                // 93
  };                                                                                                                   // 94
                                                                                                                       //
  clearMockGoogleUser = function () {                                                                                  // 96
    Meteor.call("clearMockGoogleUser", function (err) {                                                                // 97
      if (err) {                                                                                                       // 98
        console.log(err);                                                                                              // 99
      }                                                                                                                // 100
    });                                                                                                                // 101
  };                                                                                                                   // 102
}                                                                                                                      // 103
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}}},{"extensions":[".js",".json"]});
require("./server/accounts/email-token/token-server.js");
require("./server/accounts/identity/identity-server.js");
require("./server/accounts/ldap/ldap-server.js");
require("./server/accounts/saml/saml-server.js");
require("./server/accounts/accounts-server.js");
require("./server/accounts/accounts-ui-methods.js");
require("./server/accounts/accounts-ui-server.js");
require("./server/admin/admin-alert-server.js");
require("./server/admin/admin-user-invite.js");
require("./server/admin/network-capabilities-server.js");
require("./server/admin/personalization-server.js");
require("./server/admin/system-status-server.js");
require("./server/drivers/external-ui-view.js");
require("./server/drivers/ip.js");
require("./server/drivers/mail.js");
require("./server/00-startup.js");
require("./server/account-suspension.js");
require("./server/admin-server.js");
require("./server/backup.js");
require("./server/contacts-server.js");
require("./server/core.js");
require("./server/demo-server.js");
require("./server/desktop-notifications-server.js");
require("./server/dev-accounts-server.js");
require("./server/grain-server.js");
require("./server/hack-session.js");
require("./server/identity.js");
require("./server/install-server.js");
require("./server/installer.js");
require("./server/notifications-server.js");
require("./server/pre-meteor.js");
require("./server/proxy.js");
require("./server/sandcats.js");
require("./server/shell-server.js");
require("./server/signup-server.js");
require("./server/standalone-server.js");
require("./server/startup.js");
require("./server/stats-server.js");
require("./shared/admin.js");
require("./shared/dev-accounts.js");
require("./shared/grain-shared.js");
require("./shared/install.js");
require("./shared/testing.js");
//# sourceMappingURL=app.js.map
